(* ---------------------------------------------------------------
Title         Q&D Delimited Text Extractor
Overview      see help
Usage         see help
Notes         very, very, very quick & dirty... :-(

Bugs
Wish List     procfile() is overcomplicated

--------------------------------------------------------------- *)

MODULE Delims;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT SYSTEM;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileSupportLFN;

(* shortened list *)

FROM QD_LFN IMPORT unicodeConversionFlagType, findDataRecordType,
w9XfindFirst, w9XfindClose, w9XfindNext, w9XshortToLong;

FROM IO IMPORT WrStr,WrLn,WrCard,WrLngCard;

FROM Storage IMPORT ALLOCATE, DEALLOCATE, Available,
HeapTotalAvail, MainHeap;

(* ------------------------------------------------------------ *)

CONST
    ProgEXEname     = "DELIMS";
    ProgTitle       = "Q&D Delimited Text Extractor";
    ProgVersion     = "v1.0a";
    ProgCopyright   = "by PhG";
    Banner          = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    extBAK        = ".BK!";
CONST
    extCOM      = ".COM";
    extEXE      = ".EXE";
    extDLL      = ".DLL";
    extOVR      = ".OVR";
    extOVL      = ".OVL";
    extDRV      = ".DRV";
    extVXD      = ".VXD";
    extBIN      = ".BIN";
    extZIP      = ".ZIP";
    extARJ      = ".ARJ";
    extLZH      = ".LZH";
    skippedextensions = extCOM+delim+extEXE+delim+
                        extDLL+delim+extOVR+delim+extOVL+delim+extDRV+delim+
                        extVXD+delim+extBIN+delim+
                        extZIP+delim+extARJ+delim+extLZH;
CONST
    sINFO         = "::: ";
CONST
    codeON  = "<";
    codeOFF = ">";
CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errTooManyParms = 3;
    errMissingSpec  = 4;
    errTooMany      = 5;
    errNoMatch      = 6;
    errBadSpec      = 7;
    errBadDelims    = 8;
    errNotTogether  = 9;
    errAborted      = 10;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)

errmsg =
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" <file(s)> [option]..."+nl+
nl+
"This program extracts delimited text from a file."+nl+
nl+
"    -dq      use double quotes as delimiters (default)"+nl+
"    -sq      use single quotes as delimiters"+nl+
'    -k[f|k]  use "'+codeON+'" and "'+codeOFF+'" as delimiters (-kf = -k -f)'+nl+
"    -n       show line number"+nl+
"    -f       filter out delimiter"+nl+
"    -c       do not filter control characters [$00..$1f]"+nl+
"    -u:??    case sensitive user-defined delimiters"+nl+
"    -<i|r|-> reverse filter"+nl+
"    -t       show filename"+nl+
"    -x       disable LFN support even if available"+nl+
nl+
"a) Program will not process files with any of these extensions :"+nl+
"   "+skippedextensions+" for safety."+nl+
"b) Note user-defined delimiters are case sensitive."+nl+
"c) -n and -i options are mutually exclusive."+nl+
"d) This program only supports these line separators :"+nl+
"   CRLF sequence ($0d$a), LF character ($a)."+nl+
nl+
"Examples : "+ProgEXEname+" -n *.MOD"+nl+
"           "+ProgEXEname+' -u:"&;" *.htm'+nl+
"           "+ProgEXEname+" -k *.htm"+nl;

VAR
    S : str1024; (* was str256 but in case we get a LFN... *)
BEGIN
    CASE e OF
    | errHelp :
        WrStr(errmsg);
    | errOption:       S:='Unknown "~" option !';
    | errTooManyParms: S:='"~" parameter is one too many !';
    | errMissingSpec:  S:= "Missing file specification !";
    | errTooMany:      S:='Too many files match "~" specification !';
    | errNoMatch:      S:='No file matches "~" specification !';
    | errBadSpec:      S:='Illegal "~" specification !';
    | errBadDelims:    S:='Illegal "~" user-defined delimiters !';
    | errNotTogether:  S:="~ options are mutually exclusive !";
    | errAborted:      S:="Aborted by user !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp:
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType  = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn    : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    pFname = POINTER TO fnameType;
    fnameType = RECORD
        next      : pFname;
        slen      : CARDINAL; (* a SHORTCARD could do but who knows ? *)
        str       : CHAR;     (* variable length *)
    END;

PROCEDURE initList (VAR anchor : pFname );
BEGIN
    anchor := NIL;
END initList;

PROCEDURE freeList (anchor : pFname);
VAR
    needed : CARDINAL;
    p      : pFname;
BEGIN
    (* p:=anchor; *)
    WHILE anchor # NIL DO
        needed := SIZE(fnameType) - SIZE(anchor^.str) + anchor^.slen;
        p := anchor^.next;
        DEALLOCATE(anchor,needed);
        anchor:=p;
    END
END freeList;

PROCEDURE buildNewPtr (VAR anchor,p:pFname; len:CARDINAL):BOOLEAN;
VAR
    needed : CARDINAL;
BEGIN
    needed := SIZE(fnameType) - SIZE(p^.str) + len;
    IF Available(needed)=FALSE THEN RETURN FALSE; END;
    IF anchor = NIL THEN
        ALLOCATE(anchor,needed);
        p:=anchor;
    ELSE
        p:=anchor;
        WHILE p^.next # NIL DO
            p:=p^.next;
        END;
        ALLOCATE(p^.next,needed);
        p:=p^.next;
    END;
    p^.next := NIL;
    RETURN TRUE;
END buildNewPtr;

(* assume p is valid *)

PROCEDURE getStr (VAR S : pathtype; p:pFname);
VAR
    len:CARDINAL;
BEGIN
    len := p^.slen;
    Lib.FastMove( ADR(p^.str),ADR(S),len);
    S[len] := nullchar; (* REQUIRED safety ! *)
END getStr;

(* ------------------------------------------------------------ *)

(* Str.Match is not case sensitive *)

PROCEDURE isReservedPattern (S,skipthem:ARRAY OF CHAR):BOOLEAN;
VAR
    e3 : str16;
    n:CARDINAL;
    rc:BOOLEAN;
BEGIN
    rc:=FALSE;
    n:=0;
    LOOP
        isoleItemS(e3, skipthem,delim,n);
        IF same(e3,"") THEN EXIT; END;
        Str.Prepend(e3,"*");
        IF Str.Match(S,e3) THEN rc:=TRUE;EXIT; END;
        INC(n);
    END;
    RETURN rc;
END isReservedPattern;

PROCEDURE isReservedEntry (S:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE buildFileList (VAR anchor:pFname;
                        useLFN,DEBUG :BOOLEAN;spec:pathtype):CARDINAL;
VAR
    count:CARDINAL; (* should do ! *)
    ok,found:BOOLEAN;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    entry : FIO.DirEntry;
    dosattr:FIO.FileAttr;
    entryname:pathtype;
    len : CARDINAL;
    pp:pFname;
    excludeme1,excludeme2:BOOLEAN;
BEGIN
    count:=0;
    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,entry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(entryname,w9Xentry.fullfilename);
        ELSE
            Str.Copy(entryname,entry.Name);
        END;
        excludeme1 := isReservedEntry   (entryname);  (* skip "." and ".." *)
        excludeme2 := isReservedPattern (entryname,skippedextensions );
        IF NOT(excludeme1 OR excludeme2) THEN
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=entry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                (* if file has no extension, add it as a marker *)
                IF Str.RCharPos(entryname,".")=MAX(CARDINAL) THEN
                    Str.Append(entryname,".");
                END;
                IF DEBUG THEN WrStr("Included : ");WrStr(entryname);WrLn; END;
                len:=Str.Length(entryname);
                IF buildNewPtr(anchor,pp,len)=FALSE THEN
                    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL); (* errStorage *)
                END;
                INC(count);
                pp^.slen      := len;
                Lib.FastMove ( ADR(entryname),ADR(pp^.str),len );
            ELSE
                IF DEBUG THEN WrStr("Ignored  : ");WrStr(entryname);WrLn;END;
            END;
        ELSE
            IF DEBUG THEN WrStr("Excluded : ");WrStr(entryname);WrLn;END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(entry);
        END;
    END;
    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
    RETURN count;
END buildFileList;

PROCEDURE doGetCurrent (useLFN:BOOLEAN;drive:SHORTCARD;
                       VAR unit:str2; VAR current:pathtype);
VAR
    rc:CARDINAL;
    longform:pathtype;
BEGIN
    Str.Concat(unit, CHR( ORD("A")-1+ORD(drive) ),colon);

    FIO.GetDir(drive,current); (* \path without u: nor trailing \ except at root *)
    IF current[1] # colon THEN Str.Prepend(current,unit); END; (* safety *)
    IF useLFN THEN
        IF w9XshortToLong(current,rc,longform) THEN (* if error, keep DOS current *)
            Str.Copy(current,longform);
        END;
    END;
    (* LFN function seems to always return "u:\*" form except at root *)
    IF current[1] = colon THEN Str.Delete(current,0,2);END; (* safety *)
    fixDirectory(current);
END doGetCurrent;

PROCEDURE makebase (useLFN:BOOLEAN;spec:pathtype;VAR basepath:pathtype);
VAR
    u,d,n,e,current:pathtype;
    unit:str2;
    drive:SHORTCARD;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Str.Concat(basepath,u,d);
    IF same(basepath,"") THEN
        drive:=FIO.GetDrive(); (* yes we could use 0 as default drive *)
        doGetCurrent(useLFN,drive,  unit,current); (* "u:" and "\" or "\*\" *)
        Str.Concat(basepath, unit,current);
    END;
END makebase;

PROCEDURE chkUD (S:pathtype):BOOLEAN ;
VAR
    u,d,n,e:pathtype;
    pb:CARDINAL;
BEGIN
    qdSplitAllPath(S , u,d,n,e);
    pb:=0;
    IF chkJoker(u) THEN INC(pb);END;
    IF chkJoker(d) THEN INC(pb);END;
    RETURN (pb=0);
END chkUD;

PROCEDURE WrQ (S:ARRAY OF CHAR  );
BEGIN
    WrStr(dquote);WrStr(S);WrStr(dquote);
END WrQ;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fmtnum (v:LONGCARD;wi:CARDINAL;padchar:CHAR):str16;
VAR
    N:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(v,N,10,ok);
    IF ok THEN
        FOR i:= Str.Length(N)+1 TO wi DO
            Str.Prepend(N,padchar);
        END;
    ELSE
        N:="???";
    END;
    RETURN N;
END fmtnum;

PROCEDURE procFile (useLFN,DEBUG,
                   inverse,showdelims,showlinenum,showctrl:BOOLEAN;
                   opendelim,closedelim:CHAR;f:pathtype);
CONST
    wanted = SIZE(CHAR);
    winum  = 9;
    padwith= "0";
VAR
    got:CARDINAL;
    currline : LONGCARD;
    hin      : FIO.File;
    ch       : CHAR;
    state:(waiting,indelim);
    dmpme,dmpnl,issep:BOOLEAN;
BEGIN
    currline:=1;

    hin := fileOpenRead(useLFN,f);
    FIO.AssignBuffer(hin,ioBufferIn);
    state:=waiting;
    LOOP
        got:=FIO.RdBin(hin,ch,wanted);
        IF got # wanted THEN EXIT;END;
        IF ch = lf THEN INC(currline); END;
        dmpme:=FALSE;
        dmpnl:=FALSE;
        issep:=FALSE;
        CASE state OF
        | waiting:
            IF ch=opendelim THEN
                IF showdelims THEN dmpme:=TRUE;END;
                IF showlinenum THEN
                    WrStr(fmtnum(currline,winum,padwith)); WrStr(" : ");
                END;
                issep:=TRUE;
                state:=indelim;
            END;
        | indelim:
            IF ch=closedelim THEN
                IF showdelims THEN dmpme:=TRUE;END;
                dmpnl := TRUE;
                issep:=TRUE;
                state:=waiting;
            ELSE
                dmpme := TRUE;
            END;
        END;
        IF inverse THEN
            IF issep THEN
                dmpme:=FALSE;
            ELSE
                dmpme:=NOT(dmpme);
            END;
            dmpnl:=FALSE;
        END;
        IF ORD(ch) < ORD(space) THEN
            CASE ch OF
            | cr,lf: ;
            ELSE
                dmpme:=showctrl;
            END;
        END;
        IF dmpme THEN WrStr(ch);END;
        IF dmpnl THEN WrLn;END;
    END;
    fileClose(useLFN,hin);
END procFile;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    parmcount,i,opt : CARDINAL;
    S,R,filespec    : pathtype; (* was str128 *)
    state           : (waiting,gotparm1);
VAR
    countFile       : CARDINAL;
    file,basedir    : pathtype;
    ptr,anchor      : pFname;
VAR
    DEBUG,useLFN:BOOLEAN;
    opendelim,closedelim:CHAR;
    inverse,showdelims,showlinenum,showheader,showctrl : BOOLEAN;
    poll,currpoll:CARDINAL;
    userdelims : str16;
BEGIN
    Lib.DisableBreakCheck();
    FIO.ShareMode:=FIO.ShareDenyNone;

    WrLn;

    DEBUG       := FALSE;
    inverse     := FALSE;
    showdelims  := TRUE;
    showlinenum := FALSE;
    showheader  := FALSE;
    showctrl    := FALSE;
    useLFN      := TRUE;
    opendelim   := dquote;
    closedelim  := dquote;

    state   := waiting;
    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "I"+delim+"R"+delim+"-"+delim+
                                   "F"+delim+"NODELIMS"+delim+
                                   "N"+delim+"SHOWLINENUM"+delim+
                                   "K"+delim+"I8HTML"+delim+
                                   "KK"+delim+"KF"+delim+
                                   "U:"+delim+"USER:"+delim+
                                   "T"+delim+"FILENAME"+delim+
                                   "C"+delim+"CTRL"+delim+
                                   "X"+delim+"LFN"+delim+
                                   "DQ"+delim+"DOUBLEQUOTES"+delim+
                                   "SQ"+delim+"SINGLEQUOTES"
                               );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5,6 : inverse     := TRUE;
            | 7,8   : showdelims  := FALSE;
            | 9,10  : showlinenum := TRUE;
            | 11,12 : opendelim   := codeON;
                      closedelim  := codeOFF;
            | 13,14 : opendelim   := codeON;
                      closedelim  := codeOFF;
                      showdelims  := FALSE;
            | 15,16 : GetString(S,userdelims);
                      IF Str.Length(userdelims) # 2 THEN abort(errBadDelims,S);END;
                      opendelim   := userdelims[0];
                      closedelim  := userdelims[1];
            | 17,18 : showheader  := TRUE;
            | 19,20 : showctrl    := TRUE;
            | 21,22 : useLFN      := FALSE;
            | 23,24 : opendelim   := dquote;
                      closedelim  := dquote;
            | 25,26 : opendelim   := singlequote;
                      closedelim  := singlequote;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting : Str.Copy(filespec,S);
            | gotparm1: abort(errTooManyParms,S);
            END;
            INC(state);
        END;
    END;
    (* check nonsense *)
    IF inverse AND showlinenum THEN abort(errNotTogether,"-i and -n");END;
    IF state=waiting THEN abort(errMissingSpec,"");END;
    useLFN:=( useLFN AND fileSupportLFN() );

    (* a few q&d sanity checks *)

    IF chkUD(filespec)=FALSE THEN abort(errBadSpec,filespec);END;
    IF same(filespec,dot) THEN Str.Copy(filespec,stardotstar); END;
    IF same(filespec,star) THEN Str.Copy(filespec,stardotstar); END;
    IF Str.Match(filespec,"*"+antislash) THEN Str.Append(filespec,stardotstar); END;
    IF chkJoker(filespec)=FALSE THEN
        IF fileIsDirectorySpec ( useLFN,filespec) THEN
            fixDirectory(filespec);
            Str.Append(filespec,stardotstar);
        END;
    END;
    makebase(useLFN,filespec,basedir);

    initList(anchor);
    countFile := buildFileList(anchor,useLFN,DEBUG,filespec);

    CASE countFile OF
    | 0 :            abort(errNoMatch,filespec);
    | 1 : ;
    | MAX(CARDINAL): abort(errTooMany,filespec);
    END;

    IF useLFN THEN
        poll := 50; (* vindoze does not like frequent calls TO chkEscape() *)
    ELSE
        poll := 5;
    END;
    currpoll:=0;

    ptr:=anchor;
    WHILE ptr # NIL DO
        getStr(filespec,ptr);
        Str.Concat(file,basedir,filespec);

        IF showheader THEN WrStr(sINFO);WrQ(file);WrLn; WrLn; END;

        procFile (useLFN,DEBUG,inverse,showdelims,showlinenum,showctrl,
                 opendelim,closedelim,file);

        INC(currpoll);
        IF (currpoll MOD poll) = 0 THEN
            IF ChkEscape() THEN
                freeList(anchor);
                abort(errAborted,"");
            END;
            currpoll:=0;
        END;

        ptr:=ptr^.next;
    END;
    freeList(anchor);

    abort(errNone,"");
END Delims.






(*

# 1 <htm1> "string1" text1a text1b
# 2
# 3 "string3a" <htm3a> <htm3b> "string3b" text3

# 5 <htm5>
# 6 "string6"
# 7 text7a text7b text7c <htm7>
# 8
# 9 <htm9a> "string9" <htm9b>

*)


(*

@echo off
set f=delims.txt
set p=delims

cls
type %f%
echo.
echo // DEFAULT
%p% %f%
echo.
echo // SHOWLINES
%p% %f% /n

pause
cls
type %f%
echo.
echo // HTML
%p% %f% /k
echo.
echo // HTML INVERSE
%p% %f% /k /i

pause
cls
type %f%
echo.
echo // HTML NODELIM
%p% %f% /k /f
echo.
echo // HTML NODELIM INVERSE
%p% %f% /k /f /i



pause
cls
type %f%
echo.
echo // TITLE
%p% %f% /n /t
echo.

set p=
set f=


*)
