(* ---------------------------------------------------------------
Title         Q&D Flip File Data
Overview      see help
Usage         see help
Notes         yes, yes, CMP* was better a name
              as EVENODD, can be used to fool overzealous virus false alarms
Bugs
Wish List

--------------------------------------------------------------- *)

MODULE flip;

IMPORT Lib;
IMPORT Str;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn, WrLngCard,WrCard;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    errNone            = 0;
    errHelp            = 1;
    errOption          = 2;
    errExpecting       = 3;
    errParmOverflow    = 4;
    errJoker           = 5;
    errDir             = 6;
    errNotFound        = 7;
    errAlreadyExists   = 8;
    errEmpty           = 9;
CONST
    exe                = "FLIP";
    title              = "Q&D Flip File Contents";
    version            = "1.0a";
    author             = "by PhG";

PROCEDURE abort (rc:CARDINAL;einfo:ARRAY OF CHAR);
CONST
    msgHelp     =
title+" "+version+" "+author+nl+
nl+
"Syntax : "+exe+" <source> <target> [option]..."+nl+
nl+
"  -l line mode (default is binary)"+nl+
"  -o overwrite existing <target>"+nl+
"  -z remove ASCII code 26 (Ctrl-Z)"+nl+
"  -n append CRLF at end of <target>"+nl+
"  -t shortcut for -z -n"+nl+
"  -x disable LFN support even if available"+nl+
nl+
"This program flips <source> contents."+nl+
nl+
"a) Maximum line length supported by -l option is 24576 characters."+nl;

VAR
    S:str1024;
BEGIN
    CASE rc OF
    | errNone:             S := "";
    | errHelp:             WrStr(msgHelp);
    | errOption:           S := 'Unknown "|" option !';
    | errExpecting:        S := "Expecting | !";
    | errParmOverflow:     S := '"|" is one parameter too many !';
    | errJoker:            S := '| cannot contain any joker !';
    | errDir:              S := '| must be a file !';
    | errNotFound:         S := '"|" does not exist !';
    | errAlreadyExists:    S := '| already exists !';
    | errEmpty:            S := '"|" is empty !';
    ELSE
                           S := "This is illogical, Captain !";
    END;
    CASE rc OF
    | errNone,errHelp:
        ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(exe+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode( SHORTCARD(rc) );
    HALT;
END abort;

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBufferSrc : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
    ioBufferDst : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)

CONST
    ctrlZ          = BYTE(26);
    databuffersize = 65000;
    hugestrsize    = 24576;
CONST
    minparm = 1;
    maxparm = 2;
VAR
    databuffer : ARRAY [0..databuffersize-1] OF BYTE;
    Z : str1024;
    S,R,src,dst:pathtype;
    opt,parmcount,lastparm,i,wanted,got:CARDINAL;
    hin,hout:FIO.File;
    fsize,fpos:LONGCARD;
    b:BYTE;
    dbg,appendCRLF,skipCtrlZ,overwrite,useLFN:BOOLEAN;
    parm:ARRAY[minparm..maxparm] OF pathtype;
    flipmode : (binary,line);
    hugestr : ARRAY [0..hugestrsize] OF CHAR;
    ch:CHAR;
    j:CARDINAL;
BEGIN
    FIO.IOcheck:=FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone;

    WrLn;

    useLFN     := TRUE;
    overwrite  := FALSE;
    skipCtrlZ  := FALSE;
    appendCRLF := FALSE;
    flipmode   := binary;
    dbg        := FALSE;
    lastparm   := minparm-1;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "DEBUG"+delim+"DBG"+delim+
                                  "Z"+delim+"EOF"+delim+
                                  "N"+delim+"NL"+delim+"CRLF"+delim+
                                  "T"+delim+"TXT"+delim+"TEXT"+delim+
                                  "L"+delim+"LINE"
                              );
            CASE opt OF
            | 1,2,3    : abort(errHelp,"");
            | 4,5      : useLFN     := FALSE;
            | 6,7      : overwrite  := TRUE;
            | 8,9      : dbg        := TRUE;
            | 10,11    : skipCtrlZ  := TRUE;
            | 12,13,14 : appendCRLF := TRUE;
            | 15,16,17 : skipCtrlZ  := TRUE;
                         appendCRLF := TRUE;
            | 18,19    : flipmode   := line;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy( parm[lastparm] , S);
        END;
    END;
    CASE lastparm OF
    | minparm-1 : abort(errHelp,"");
    | minparm   : abort(errExpecting,"<target>");
    | maxparm   : ; (* ok *)
    ELSE
        ; (* trapped supra *)
    END;

    useLFN:=( useLFN AND fileSupportLFN() );

    src := parm[minparm];
    IF chkJoker(src) THEN abort(errJoker,"<source>");END;
    IF fileExists(useLFN,src) THEN
        IF fileIsDirectorySpec(useLFN,src) THEN abort(errDir,"<source>");END;
    ELSE
        abort(errNotFound,src);
    END;

    fsize:=fileGetFileSize(useLFN,src);
    IF fsize = 0 THEN abort(errEmpty,src);END;

    dst := parm[minparm+1];
    IF chkJoker(dst) THEN abort(errJoker,"<target>"); END;
    IF fileExists(useLFN,dst) THEN
        IF fileIsDirectorySpec(useLFN,dst) THEN abort(errDir,"<target>");END;
        IF NOT(overwrite) THEN abort(errAlreadyExists,"<target>");END;
    END;

    hin:=fileOpenRead(useLFN,src);
    FIO.AssignBuffer(hin,ioBufferSrc);
    hout:=fileCreate(useLFN,dst);
    FIO.AssignBuffer(hout,ioBufferDst);

    Z:='::: Flipping "|" data, please wait...';
    Str.Subst(Z,"|",src);
    WrStr(Z);WrLn;

    CASE flipmode OF
    | binary:
        wanted:=CARDINAL ( fsize MOD LONGCARD(databuffersize) );
        IF wanted=0 THEN wanted:=databuffersize;END;
        fpos:=fsize-1+1; (* -1 because 0-based, +1 because we need data AFTER this position *)
        LOOP
            DEC(fpos,LONGCARD(wanted));
            IF dbg THEN
                WrStr("fpos = ");WrLngCard(fpos,12);
                WrStr("  wanted = ");WrCard(wanted,8);WrLn;
            END;
            FIO.Seek(hin,fpos);
            got:=FIO.RdBin(hin,databuffer,wanted);
            i:=wanted;
            LOOP
                DEC(i);
                b:=databuffer[i];
                IF skipCtrlZ THEN
                    IF b # ctrlZ THEN FIO.WrBin(hout,b,SIZE(b));END;
                ELSE
                    FIO.WrBin(hout,b,SIZE(b));
                END;
                IF i=0 THEN EXIT; END;
            END;
            IF fpos=0 THEN EXIT; END;
            wanted:=databuffersize;
         END;
    | line:
        LOOP
            IF FIO.GetPos(hin) >= fsize THEN EXIT; END;
            FIO.RdStr(hin,hugestr);
            j:=Str.Length(hugestr);
            IF j > 1 THEN
                i:=0;
                LOOP
                    ch:=hugestr[i];
                    hugestr[i]:=hugestr[j-1];
                    hugestr[j-1]:=ch;
                    INC(i);
                    DEC(j);
                    IF i >= j THEN EXIT;END;
                END;
            END;
            FIO.WrStr(hout,hugestr);FIO.WrLn(hout);
        END;
    END;

    IF appendCRLF THEN FIO.WrLn(hout);END;

    fileClose(useLFN,hout);
    fileClose(useLFN,hin);

    Z:='+++ "|" created.';
    Str.Subst(Z,"|",dst);
    WrStr(Z);WrLn;

    abort(errNone,"");
END flip.

