
(*
    we ignore entries without a pattern
    we handle multiple patterns in an entry (todo : index ?)

*)

MODULE lifeLEX;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT IO;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
    extTXT = ".TXT";
    exe    = "LIFELEX";
    title  = "Q&D Life Lexicon to VLIFE.DAT converter";
    version= "v1.0";
    culprit= "by PhG";
    msghelp = title+" "+version+" "+culprit+nl+
              nl+
              "Syntax : "+exe+" <file["+extTXT+"]>"+nl+
              nl+
              "This program extracts names and patterns from Life Lexicon by Stephen Silver."+nl;
    header  = "#"+nl+
              "# Life patterns extracted from Life Lexicon by Stephen Silver"+nl+
              "#"+nl;

PROCEDURE abort (errcode:CARDINAL; S:ARRAY OF CHAR);
BEGIN
    CASE errcode OF
    | 0 : ;
    | 1 : WrStr(msghelp);
    ELSE
          WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(errcode));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procFile(src:ARRAY OF CHAR);
CONST
    DEBUG=FALSE;
    allowed = dot+star;
VAR
    hin:FIO.File;
    S:str256; (* safety *)
    id:str128;
    state:(waitforid,waitforpat);
    p,patlines:CARDINAL;
    fpos:LONGCARD;
    ok:BOOLEAN;
BEGIN
    Str.Copy(id,"::: THIS CANNOT HAPPEN ! :::");
    state:=waitforid;
    hin:=FIO.OpenRead(src);
    LOOP
        fpos:=FIO.GetPos(hin);
        FIO.RdStr(hin,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        ok:= Str.Length(S) # 0;
        IF ok THEN
            IF DEBUG THEN WrStr("::: ");WrStr(S);WrLn;END;
            CASE state OF
            | waitforid:
                IF DEBUG THEN WrStr("### waitforid");WrLn;END;
                IF S[0] = colon THEN
                    Str.Delete(S,0,1);
                    p:=Str.CharPos(S,colon);
                    IF p # MAX(CARDINAL) THEN
                        Str.Slice(id,S,0,p);
                        patlines:=0;
                        state:=waitforpat;
                    END;
                END;
            | waitforpat:
                IF DEBUG THEN WrStr("### waitforpat");WrLn;END;
                IF verifyString(S,allowed) THEN
                    IF patlines = 0 THEN
                        WrLn;WrStr(id);
                        WrLn;
                    END;
                    WrStr(S);WrLn;
                    INC(patlines);
                ELSE
                    IF S[0] = colon THEN
                        FIO.Seek(hin,fpos);
                        state:=waitforid;
                    ELSE
                        patlines:=0;
                    END;
                END;
            END;
        END;
        IF FIO.EOF THEN EXIT; END;
    END;
    WrLn; (* better safe than sorry *)
    FIO.Close(hin);
END procFile;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    S    : str128;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;

    WrLn;

    CASE Lib.ParamCount() OF
    | 1 :
        Lib.ParamStr(S,1);
        CASE S[0] OF
        | "-","/": Str.Delete(S,0,1);
        END;
        IF same(S,"?") THEN abort(1,msghelp);END;
    ELSE
        abort(1,msghelp);
    END;

    Lib.ParamStr(S,1);
    IF Str.RCharPos(S,".")=MAX(CARDINAL) THEN Str.Append(S,extTXT);END;
    IF chkJoker(S) THEN abort(1,exe+" : Jokers are not supported !");END;
    IF FIO.Exists(S)=FALSE THEN abort(2,exe+" : <file> does not exist !");END;

    WrStr(header);WrLn;
    procFile(S);

    abort(0,"");
END lifeLEX.

