'--------------------------------------------------------------------
' Title         MorfMsg
' Version       1.0a
' Author        PhG
' Overview
' Future        convert a bw picture to morfmsg.def format?
' Notes         based on EGAclock v1.0
'               use BWBMPDEF to convert from bw BMP to DEF / DAT format
' Usage         MorfMsg msgfile [options]
'
'--------------------------------------------------------------------

$CPU 8086

$OPTIMIZE SIZE
$COMPILE EXE

$DEBUG MAP OFF
$DEBUG PBDEBUG OFF

$LIB COM        OFF
$LIB CGA        OFF
$LIB EGA        ON
$LIB VGA        OFF
$LIB HERC       OFF
$LIB LPT        OFF
$LIB IPRINT     OFF
$LIB FULLFLOAT  ON

$ERROR BOUNDS   Off
$ERROR NUMERIC  Off
$ERROR OVERFLOW Off
$ERROR STACK    Off

$FLOAT PROCEDURE

$COM    0
$STRING 16
$STACK  8192 ' big stack for big local array in SUB
$SOUND  1

$DIM ARRAY

$DYNAMIC

$OPTION CNTLBREAK OFF ' OFF in final EXE

'--------------------------------------------------------------------

DEFINT A-Z
OPTION ARRAY BASE 0
OPTION BINARY BASE 0
'
%False = 0
%True = NOT %False

%Debug = %false ' false in final exe

'--------------------------------------------------------------------
' external functions

$LINK "QDBOX.PBU"
$INCLUDE "QDBOX.DEF"

'--------------------------------------------------------------------
' internal functions

DECLARE SUB GetStrLoc() ' PowerBasic runtime library

SUB StdOut ( BYVAL Text AS STRING )
! push DS                    ; save DS FOR PowerBASIC
! push WORD Ptr Text         ; push STRING handle on stack
! CALL GetStrLoc
! jcxz ExitStdOut
! mov  DS, DX
! mov  DX, AX
! mov  AH, &H40              ; DOS WRITE TO file
! mov  BX, 1                 ; file handle 1 is CONS
! INT  &H21
ExitStdOut:
! pop  DS
END SUB

SUB StdOutLn( BYVAL s$ )
CALL StdOut (s$ + CHR$(13, 10) )
END SUB

'--------------------------------------------------------------------
' error handling

%eNone = 100
%eUsage = 101
%eTooManySame = 102
%eUnknownOpt = 103
%eNeedEGA = 104
%eBadRange = 105
%eDosTooOld = 106
%eBadExeName = 107
%eVertices= 108
%ecannotfitx = 109
%ecannotfity = 110
%ePaperInk=111
%eFileNotFound=112
%eFileIllegal=113
%eSectionNotFound=114
%eBadDElimiter=115
%eNotEnoughShapes=116
%emissinglastquote=117
%eautofit=118
%eCannotAutofit=119
%etest=163


'--------------------------------------------------------------------
' constants

%minX=0
%minY=0
%maxX=639
%maxY=349
%hires=9 ' EGA 640x350
' requirements for screen mode 9 ega 640x350
%PixelsPerBytePerPlane=8
%Planes=4

%page1=0
%page2=1

%text=0
%cga=2 ' 640x200 CGA

%Circle = 0
%Square = 1
%Disk = 2
%Box = 3

%Circle2 = 4
%Square2=5

%MinDosVersion = 310 ' 3.10 or later

'--------------------------------------------------------------------
' global variables

SHARED einfo$ ' error info

SHARED vShape$,vOuterDot$,vInnerDot$,vPaper$,vInk$
shared vMotionSteps$,vDelay$,vAutoFit$

shared viewpage,workpage
shared etapes
shared pause

'--------------------------------------------------------------------

DIM DYNAMIC Icon(1) AS INTEGER ' defined at run time
SHARED Icon()

DIM dynamic Eraser(1) as integer ' when erasing force
shared Eraser()

%firstVertex=1
%lastVertex=8000 ' should be enough! and cannot exceed 64Kb anyway...

type VertexType ' coordinate of a dot
	x as integer
    y as integer
end type

' vertex contains coordinates of a dot from (0,0)++
' i.e. gives position of each "*" in shape profile
' will have to convert them later using plotsize for relative screen coors
DIM STATIC Vertex (%FirstVertex:%LastVertex) AS VertexType ' compile time
SHARED Vertex()


%FirstShape = 1
%MaxShape = 10

TYPE ShapeInfoType
	offset AS INTEGER ' offset of first dot
    count AS INTEGER ' number of dots
	maxwide as integer
    maxheight as integer
END TYPE

'shapeinfo contains infos about a shape
DIM STATIC ShapeInfo (%FirstShape:%MaxShape) AS ShapeInfoType ' compile time
SHARED ShapeInfo()

%Origin=0
%Target=1

' points qui se transforment pendant le morphing
DIM STATIC Path(%FirstVertex:%LastVertex,%Origin:%Target) AS VertexType ' compile time
SHARED Path()

'--------------------------------------------------------------------

ProgramName$="MorfMsg"
EXEName$="MORFMSG"
Version$="v1.0c"
Copyright$="par PhG"
Banner$=ProgramName$+" "+Version$+" "+Copyright$

'
$if %debug
$else
ON ERROR GOTO Abort
$endif

GOTO Start: ' jump to main() ;-)

'--------------------------------------------------------------------

Abort:
IF ERR = %eUsage THEN
	wi=39
	stdoutln Banner$
	stdoutln ""
        stdoutln "Syntaxe : "+ExeName$+" [msgfile[.DEF]] [option]..."
  	stdoutln       "                                          par dfaut"
  	call Dmp(wi,"/S:# Motif (0..5)",vshape$)
  	CALL Dmp(wi,"/W:# Rayon extrieur du motif",vOuterDot$)
  	CALL Dmp(wi,"/R:# Rayon intrieur du motif",vInnerDot$)
  	CALL Dmp(wi,"/P:# Couleur du fond",vPaper$)
  	CALL Dmp(wi,"/I:# Couleur du motif",vInk$)
  	CALL Dmp(wi,"/N:# Nombre d'tapes dans une animation",vMotionSteps$)
  	CALL Dmp(wi,"/D:# Temporisation (1/10s)",vDelay$)
    call dmp(wi,"/A:# Ajustement automatique",vAutoFit$)
    stdoutln ""
    stdoutln "Le nombre maximal de points est "+LTRIM$(RTRIM$(STR$(%lastvertex)))+"."
    stdoutln ""
    stdoutln "L'utilitaire BWBMPDEF permettra de crer un fichier .DEF"
    stdoutln " partir d'une image monochrome au format BMP."
    END %eUsage-%eNone
END IF
SELECT CASE ERR
CASE %eUsage
    e$="Impossible (!) ;  signaler au programmeur"
CASE %eTooManySame
	e$="Au moins une option rpte sans raison"
CASE %eUnknownOpt
	e$="Au moins une option imprvue"
case %eNeedEGA
	e$="Ce programme exige une carte graphique EGA ou plus"
case %eBadRange
	e$=einfo$
case %eDosTooOld
	e$="Ce programme exige un DOS version "+einfo$+" ou plus"
case %eBadexeName
	e$=einfo$+" n'est pas le nom original du fichier excutable"
case %eVertices
	e$="Trop de points sont dfinis"
CASE %ecannotfitx
    e$="La largeur est trop grande pour l'cran"
cASE %ecannotfity
    e$="La hauteur est trop grande pour l'cran"
case %epaperink
	e$="Encre et papier doivent diffrer"
case %etest
	e$="Jusque l, a va..."
case %eFileNotFound
	e$="Le fichier "+einfo$+" est introuvable"
case %eFileIllegal
	e$="Le nom du fichier "+einfo$+" est incorrect"
case %eSectionNotFound
	e$="La section d'en-tte "+einfo$+" est introuvable dans le fichier"
case %eNotEnoughShapes
	e$="Au moins deux images sont ncessaires pour une animation"
case %eBadDelimiter
	e$="Le caractre "+einfo$+" n'est pas cens commencer une ligne"
case %emissinglastquote
	e$="Une ligne ne finit pas par un guillemet"
case %eautofit
	e$="La valeur de l'ajustement automatique doit tre 0 ou 1"
case %eCannotAutofit
	e$="Impossible d'ajuster automatiquement l'animation"
CASE ELSE
    e$=HEX$(ERADR) ' ERADR is a longint (7fFFffFF)
    e$=MID$("00000000",1,8-LEN(e$))+e$
    e$= "Erreur #"+MID$(STR$(ERR),2)+"  l'adresse $"+e$
END SELECT
e$=ProgramName$+" : "+e$+" !"
stdoutln e$
'hnd=freefile
'open "a",#hnd,"morfmsg.err"
'print #hnd,e$
'close #hnd
END ERR-%eNone

'--------------------------------------------------------------------

SUB WaitVGARetrace
!   mov dx,&h03DA
l1:
!   in al,dx
!   and al,&h08
!   jnz l1
l2:
!   in al,dx
!   and al,&h08
!   jz  l2
END SUB

FUNCTION IsEGA                'test for EGA
  IF BIT (pbvScrnCard,3) = 0 THEN
  	IsEGA = %False
  ELSE
  	IsEGA = %True
  END IF
END FUNCTION

' *should* test for wide < 2 ! ;-)
SUB Dmp (BYVAL wide,BYVAL s1$,BYVAL s2$)
pad$=STRING$(wide," ")
length=LEN(s1$)
IF length > wide THEN
	s$=LEFT$(s1$,wide-2)+".."
ELSE
	s$=s1$+RIGHT$(pad$,wide-length)
END IF
s$=s$+" | "
s2$=ltrim$(rtrim$(s2$))
s2$=MID$("    ",1,4-LEN(s2$))+s2$ ' 4 digits
stdoutln s$+s2$ ' do not check if bigger than 80!
END SUB

' returns %True or %False if opt found or not, keep count of opt found
' vardest$ is not changed unless optstr$ found AND param <> ""

FUNCTION ChkOptStr (BYVAL cli$,BYVAL optstr$,BYVAL exactflag,vardest$,count)
var$=""
count=FindOptStr(cli$,optstr$,exactflag,var$)
SELECT CASE count
CASE 0
	ChkOptStr=%False
CASE 1
	if var$="" then
    	ChkOptStr=%False ' do not accept empty string as parameter
    else
		vardest$=var$
		ChkOptStr=%True
    end if
CASE ELSE
	ERROR %eTooManySame
END SELECT
END FUNCTION

SUB ChkRange (BYVAL param$,BYVAL value$,BYVAL minrange,BYVAL maxrange)
v#=VAL(value$)
IF v# < minrange OR v# > maxrange THEN
	range$="["+MID$(STR$(minrange),2)+".."+MID$(STR$(maxrange),2)+"]"
	einfo$="Le paramtre "+param$+" doit appartenir  l'intervalle "+range$
	ERROR %eBadRange
END IF
END SUB

function booltostr$(byval tf%)
if tf% = %true then
	s$="oui"
else
	s$="non"
end if
function=s$
end function


'--------------------------------------------------------------------

SUB InitDefault
vShape$=str$(%Disk)
vOuterDot$="6" ' outer half-width
vInnerDot$="4" ' inner half-width
vPaper$="0"
vInk$="10"
vMotionSteps$=str$(20)
vDelay$="10"
vAutoFit$="oui"
END SUB

'--------------------------------------------------------------------

SUB TextMode
SCREEN %cga ' reset width to 80
SCREEN %text
END SUB

sub ViewWork (byval vpage, byval wpage)
screen %hires,,wpage,vpage
CALL WaitVGARetrace ' must be AFTER and hopla, no longer flicker! ;-)
end sub

sub SetHires(byval viewpage, byval workpage,byval ink,byval paper)
call viewwork(viewpage,workpage)
COLOR ink,paper
CLS
end sub

sub InitIcon (BYVAL numicon,byval ink,byval paper,byval plotsize,byval radius)
cx=%maxX \ 2
cy=%maxY \ 2
diameter=radius*2
wleft=cx-radius
wtop=cy-radius
wright=wleft+diameter-1 ' substract 1 to have just width in pixels
wbot=wtop+diameter-1 ' idem
SELECT CASE numicon
CASE %Circle
	CIRCLE (cx,cy),radius
CASE %Square
    LINE (wleft,wtop)-(wright,wbot),,B
CASE %Disk
	CIRCLE (cx,cy),radius
    PAINT (cx,cy)
CASE %Box
    LINE (wleft,wtop)-(wright,wbot),,BF
CASE %Circle2
	CIRCLE (cx,cy),radius
    circle (cx,cy),radius-0.5
    circle (cx,cy),radius-1
CASE %Square2
    LINE (wleft,wtop)-(wright,wbot),,B
    line (wleft+1,wtop+1)-(wright-1,wbot-1),,B
CASE ELSE ' if undefined, disk
	CIRCLE (cx,cy),radius
    PAINT (cx,cy)
END SELECT
dotwidthX=plotSize*2+1 ' better safe than sorry! ;-)
dotwidthY=plotSize*2+1 ' idem
needed=(dotwidthX\%PixelsPerBytePerPlane)+1 ' idem... and avoid CEIL function
BytesRequested=4+needed*%Planes*dotwidthY
REDIM Icon(BytesRequested\2) AS SHARED INTEGER ' more just in case...
wleft=cx-plotSize
wtop=cy-plotsize
wright=cx+plotsize-1 ' must substract in order to have just width in pixels
wbot=cy+plotsize-1 ' idem
GET (wleft,wtop)-(wright,wbot),Icon
end sub

' create a shape as an eraser with screen already blanked
sub InitEraser (BYVAL plotsize)
cx=%maxX \ 2
cy=%maxY \ 2
dotwidthX=plotSize*2+1 ' better safe than sorry! ;-)
dotwidthY=plotSize*2+1 ' idem
needed=(dotwidthX\%PixelsPerBytePerPlane)+1 ' idem... and avoid CEIL function
BytesRequested=4+needed*%Planes*dotwidthY
REDIM Eraser(BytesRequested\2) AS SHARED INTEGER ' more just in case...
wleft=cx-plotSize
wtop=cy-plotsize
wright=cx+plotsize-1 ' must substract in order to have just width in pixels
wbot=cy+plotsize-1 ' idem
GET (wleft,wtop)-(wright,wbot),Eraser
end sub


'--------------------------------------------------------------------
' normally read from a file
' return lastshape
SUB InitShapes(byval f$,byval section$, lastshape)
' store vertices temporiraly here before we modify the plotting order
DIM DYNAMIC TmpVertex (%FirstVertex:%LastVertex) AS VertexType ' compile time
LOCAL TmpVertex()

'stdoutln "Lecture du fichier..."
'''call working(%initworking,2)

section$=Upper$(section$)
hnd=freefile
open "i",#hnd,f$
foundit=%False
DO UNTIL EOF(hnd)
    LINE INPUT #hnd,li$
    '''call working(%showworking,0)
    REPLACE CHR$(9) WITH " " IN li$     ' remove tabs
    li$=LTRIM$(RTRIM$(li$)) ' just in case...
    li$=Upper$(li$)         ' case insensitive
	IF li$=section$ THEN
		foundit=%True
		EXIT LOOP
	END IF
LOOP
IF foundit=%False THEN
	'''call working(%endworking,0)
    stdoutln ""
	CLOSE #hnd
	einfo$=section$
	ERROR %eSectionNotFound
END IF

'''call working(%progressworking,0)

endOfShape$="$"
doublequote$=chr$(34)
RANDOMIZE TIMER
currvertex=%FirstVertex
currshape=%firstshape-1

' begin a new entry so init parms
startofdef=currvertex
maxWidth=0
y=0

do until eof(hnd)
    LINE INPUT #hnd,li$
    '''call working(%showworking,0)
   	REPLACE CHR$(9) WITH " " IN li$     ' remove tabs
   	li$=LTRIM$(RTRIM$(li$)) ' just in case...
   	li$=Upper$(li$)         ' case insensitive
    first$=LEFT$(li$,1)
   	SELECT CASE first$
    CASE "",";","#"  		' ignore empty line and comment

    CASE EndOfShape$
        if (currshape+1) > %maxshape then
        	exit loop ' we ignore other defs if any
        end if
        incr currshape
	    ' now we have finished reading shape definition so actualize infos
    	ShapeInfo(currshape).offset=startofdef
	    count=currvertex-startofdef
		ShapeInfo(currshape).count=count
	    ShapeInfo(currshape).maxwide=maxWidth
	    ShapeInfo(currshape).maxheight=y
   		'
	    ' we now have each vertex making shape stored as defined
		' we can build new order of sequence and reset Vertex accordingly
	    '
        ' warning!!! this limits # of dots to 0..255 interval!!!
        ' this is not a problem for simple digits but can be with big picture
        ' this is a hack
        '
		s$=""
        plen=0
    	FOR k = 1 TO count
            '''call working(%showworking,0)
        	thisdot=k-1 ' important!
	    	DO
    			' int((upper-lower+1)*rnd+lower)
				newpos=INT((count-1+1)*RND+1) ' [1..num] -> chr$(1)..chr$(255)
				newpos=INT((count-1+1)*RND+1)
				newpos=INT((count-1+1)*RND+1)
				newposlo=(newpos-1) mod 256 ' compute a double-byte char
                newposhi=(newpos-1) \ 256
				db$=chr$(newposlo)+chr$(newposhi)
                p=1
                do
                	foundit=instr(p,s$,db$)
                    if foundit = 0 then
                    	exit loop ' db$ was NOT found so we can stop randoming
                    else
                    	if (foundit mod 2) = 1 then ' found at impair, we must try again
                    		exit loop
                        else ' was found at even pos so look in rest of s$
	                    	p=foundit+1 ' next starting position at ODD
                        end if
                    end if
                loop until p >= plen
		   	LOOP UNTIL foundit= 0 ' find a NEW position
       		s$=s$+CHR$(newposlo)+chr$(newposhi)
            incr plen
            incr plen
	    	TmpVertex(newpos)=Vertex(startofdef+thisdot)
		NEXT
	    ' now we can change order of sequence in array of positions
		FOR k = 1 TO count
    		thisdot=k-1
	    	Vertex(startofdef+thisdot)=TmpVertex(k)
		NEXT
        ' now we can reinit new entry parms
		startofdef=currvertex
		maxWidth=0
		y=0
	CASE DoubleQuote$ ' a definition line in current entry
		if right$(li$,1) <> DoubleQuote$ then
        	'''call working(%endworking,0)
            stdoutln ""
        	close #hnd
            error %eMissingLastQuote
        end if
        s$=mid$(li$,2,len(li$)-2)
       	x=0
       	FOR ndx=1 TO len(s$)
      		IF MID$(s$,ndx,1)<>" " THEN
      			Vertex(currvertex).x=x
               	Vertex(currvertex).y=y
               	INCR currvertex
               	IF currvertex > %LastVertex THEN
					'''call working(%endworking,0)
                    stdoutln ""
                    close #hnd
                   	ERROR %eVertices
                end if
           	END IF
           	INCR x
      	NEXT
       	IF x > maxWidth THEN maxWidth=x
       	INCR y
	CASE ELSE
		'''call working(%endworking,0)
        stdoutln ""
       	CLOSE #hnd
        einfo$=first$
        ERROR %eBadDelimiter
    END SELECT
LOOP
CLOSE #hnd
'''call working(%endworking,0)
'stdoutln ""
lastshape=currshape
if lastshape < (%firstshape+1) then error %eNotenoughshapes
end sub

'--------------------------------------------------------------------

sub flushkeyboard
do until inkey$ = ""
loop
end sub

'--------------------------------------------------------------------

' build list of vertices stored in Path, while morphing, returns this #
function buildMorphState (byval sourceshape, byval targetshape)
ptrSource =   ShapeInfo(sourceshape).offset
countSource = ShapeInfo(sourceshape).count
ptrTarget=    ShapeInfo(targetshape).offset
countTarget = ShapeInfo(targetshape).count
nVertex=max%(countsource,counttarget) ' number of vertices while morphing
zx2=nVertex
zy2=    min%(countSource,countTarget)
deltax = zx2 - 1
deltay = zy2 - 1
zd = (2 * deltay) - deltax
dinc1 = deltay * 2
dinc2 = (deltay - deltax) * 2
IF countSource < countTarget THEN
' use bresenham algorithm as if from 1,1 to big,small
	FOR p = %firstvertex TO nVertex
        Path(p,%Origin)=Vertex(ptrSource)
    	Path(p,%Target)=vertex(ptrTarget)
	    IF zd < 0 THEN
       		zd = zd + dinc1
		ELSE
       		zd = zd + dinc2
      		INCR ptrsource
		END IF
        INCR ptrTarget
	NEXT
ELSE
	FOR p = %firstvertex TO nVertex
        Path(p,%Origin)=vertex(ptrsource)
   		Path(p,%Target)=vertex(ptrtarget)
	    IF zd < 0 THEN
       		zd = zd + dinc1
		ELSE
       		zd = zd + dinc2
       		INCR ptrtarget
		END IF
        INCR ptrsource
	NEXT
END IF
buildmorphstate=nvertex
end function

'--------------------------------------------------------------------

function ScaleShapes (byval plotsize,byval lastshape)
scale = plotsize*2 ' radius * 2 !!!
limx=%maxX+1 ' because width begins at 1
limy=%maxY+1
FOR currshape=%FirstShape TO lastshape
    startofdef=ShapeInfo(currshape).offset
	count=ShapeInfo(currshape).count
    for k = 1 to count
    	currvertex=startofdef+k-1
    	x=vertex(currvertex).x
        y=vertex(currvertex).y
		vertex(currvertex).x=x*scale
        vertex(currvertex).y=y*scale
    next
    maxwide=scale*ShapeInfo(currshape).maxwide
	maxheight=scale*ShapeInfo(currshape).maxheight
    if maxwide > limx then function= %ecannotfitx : exit function
    if maxheight > limy then function =  %ecannotfity : exit function
    shapeinfo(currshape).maxwide=maxwide
    shapeinfo(currshape).maxheight=maxheight
    ' now center the whole shape by adjusting coordinates
    fromx=(limx-maxwide) \ 2
	fromy=(limy-maxheight) \ 2
    for k = 1 to count
    	currvertex=startofdef+k-1
    	x=vertex(currvertex).x
        y=vertex(currvertex).y
		vertex(currvertex).x=x+fromx
        vertex(currvertex).y=y+fromy
    next
next
function = %eNone
end function

'--------------------------------------------------------------------
' plot shape
SUB drawshapexor(byval shapenum)
origin=shapeinfo(shapenum).offset
count =shapeinfo(shapenum).count
FOR k = 1 to count
	currdot=k-1
	wleft=Vertex(origin+currdot).x
	wtop=vertex(origin+currdot).y
	PUT (wleft,wtop),Icon,XOR
NEXT
END SUB

'--------------------------------------------------------------------

' program code begins here : main() ;-)

Start:
stdoutln ""
if isEGA = %False then error %eNeedEGA
einfo$=MID$(STR$(%MinDosVersion \ 100),2)+"."+MID$(STR$(%MinDosVersion MOD 100),2)
IF DosVersion < %MinDosVersion THEN ERROR %eDosTooOld
exepath$= GetArg0$
exepath$= Upper$(exepath$)
CALL SplitPath(exepath$,currunit$,currpath$,currfile$)
CALL SplitName(currfile$,filename$,ext$)
einfo$=filename$+".EXE"

$if %debug
$else
IF filename$ <> ExeName$ THEN ERROR %eBadExeName
$endif

' now check command line for help

CALL InitDefault ' do it NOW!

cli$=COMMAND$
foundSOSarg=FindArg(cli$,"?|H|HELP|SOS|AIDE",%False)
foundSOSopt=FindOpt(cli$,"?|H|HELP|SOS|AIDE",%False)
SELECT CASE (foundSOSarg+foundSOSopt)
CASE 0
	' no help call
CASE 1
	ERROR %eUsage
CASE ELSE
	ERROR %eTooManySame
END SELECT

' check input filename
vInput$=ArgV$(cli$,1)
vInput$=Upper$(vInput$)
einfo$=vInput$
IF INSTR("-/",LEFT$(vInput$,1)) = 0 THEN
	IF INSTR(vInput$,ANY "*?") > 0 THEN ERROR %eFileIllegal
	CALL SplitPath(vInput$,vunit$,vpath$,vfile$)
	CALL SplitName(vfile$,vfilename$,vext$)
	IF vext$="" THEN vext$="DEF"
	vInput$=vunit$+vpath$+vfilename$+"."+vext$
    nfile=1
ELSE ' was an option, take default
	vInput$=filename$+".DEF"
	nfile=0
END IF
einfo$=vInput$
iF Exist(vInput$) = %False THEN ERROR %eFileNotFound


' check command line for options

if ChkOptStr(cli$,"S|SHAPE",%False,v$,n1)=%True THEN vShape$=v$
if ChkOptStr(cli$,"W|WIDTH",%False,v$,n3)=%True THEN vOuterDot$=v$
if ChkOptStr(cli$,"R|RADIUS",%False,v$,n4)=%True THEN vInnerDot$=v$
if ChkOptStr(cli$,"P|PAPER",%False,v$,n5)=%True THEN vPaper$=v$
if ChkOptStr(cli$,"I|INK",%False,v$,n6)=%True THEN vInk$=v$
if ChkOptStr(cli$,"N|STEPS",%False,v$,n9)=%True THEN vMotionSteps$=v$
if ChkOptStr(cli$,"D|DELAY|TEMPO",%False,v$,n10)=%True THEN vDelay$=v$
if ChkOptStr(cli$,"A|AUTOFIT",%false,v$,n11)=%true then vAutoFit$=v$
' check for unknown commands/options on command line

optcount=n1+n3+n4+n5+n6+n9+n10+n11+nfile
IF optcount <> ArgC(cli$) THEN ERROR %eUnknownOpt

' perform some checks

CALL ChkRange ("/S",vShape$,%circle,100)
CALL ChkRange ("/W",vOuterDot$,1,44)
call ChkRange ("/R",vInnerDot$,1,44)
CALL ChkRange ("/P",vPaper$,0,15)
call ChkRange ("/I",vInk$,0,15)
CALL ChkRange ("/N",vMotionSteps$,2,1000)
call ChkRange ("/D",vDelay$,1,100)
s$=lcase$(vAutoFit$)
select case s$
case "oui","o","yes","y","1"
	vautofit$="1"
case "non","n","no","n","0"
	vautofit$="0"
case else
    error %eAutofit
end select

ink=val(vink$)
paper=val(vpaper$)
if ink=paper then error %ePaperInk
fontnum=val(vfont$)
numicon=val (vShape$)
plotsize=VAL(vOuterDot$)
radius=VAL(vInnerDot$)
etapes=val(vmotionsteps$)
pause=val(vdelay$) / 10
call InitShapes(vInput$,"[MorfMsg]",lastshape) ' return # of shapes
i%= ScaleShapes(plotsize,lastshape)
if i% <> %eNone then
	if vautofit$="0" then
    	error i%
    else
    	do
    		decr plotsize
        	if plotsize < 1 then error %eCannotAutofit
			call InitShapes(vInput$,"[MorfMsg]",lastshape) ' yes we must reinit
            i%= scaleshapes(plotsize,lastshape)
        	if i% = %eNone then exit do
    	loop
	end if
end if

call SetHires(%page1,%page1,ink,paper) ' clear screen 1
call SetHires(%page1,%page2,ink,paper) ' clear screen 2
call InitIcon(numicon,ink,paper,plotsize,radius)
call SetHires(%page1,%page2,ink,paper) ' clear screen 2 again
call initEraser(plotsize) ' create shape to recreate background
call FlushKeyboard

' init pages

viewpage=%page1
workpage=%page2
call viewwork(viewpage,workpage)

' draw current time on both pages to begin in a known state

currshape=%firstshape

' display on hidden page
call drawshapexor (currshape) ' always the first time drawn
swap viewpage,workpage ' toggle pages now
' show updated page
call viewwork(viewpage,workpage)
' now display on newly hidden page too
call drawshapexor (currshape)

' update display now
do
	prevshape=currshape
	incr currshape
    if currshape > lastshape then currshape=%firstshape
    ' perform the page flipping in routines thanks to global vars
	call updatemov (currshape,prevshape)
    'for i%=1 to pause
    '	call waitvgaretrace
    'next
    delay pause ' wait a little after animation
loop while not instat ' use mouse, flush kbd before?
call SetHires(%page1,%page1,ink,paper) ' clear screen 1
CALL TextMode
END %eNone-%eNone

'--------------------------------------------------------------------
'
sub updateMov (byval currshape,byval prevshape)
for currstep=1 to etapes+1 ' take care of overhead
	CALL DrawMoving(currshape,prevshape,currstep)
	swap viewpage,workpage ' toggle pages now
	call viewwork(viewpage,workpage) ' show updated page
next
END sub

SUB DrawMoving(byval currshape,byval prevshape,byval currstep)
' build polygon while morphing, return # of vertices stored in Path
vertices=buildMorphState(prevshape,currshape)
select case currstep
case 1
	' we erase previous shape
    CALL PlotState(prevshape,currstep-1,vertices,0) ' erase source
    ' and set currstep 1
	cALL PlotState(currshape,currstep,vertices,1)
case (etapes+1)
	' we erase previous number
    CALL PlotState(prevshape,currstep-2,vertices,0)
    ' draw new on workpage at currstep
    CALL PlotState(currshape,currstep-1,vertices,1) ' draw target
case else
	' erase old on workpage
    ' we erase previous previous drawn page
    CALL PlotState(prevshape,currstep-2,vertices,0)
    ' draw new on workpage at currstep
    CALL PlotState(currshape,currstep,vertices,1)
end select
END SUB

' plot digit at coordinates fromx,fromy
SUB PlotState(byval shapenum,byval here, byval vertices, byval operation)
FOR k = %firstvertex to vertices
	wleft1=Path(k,%Origin).x
	wtop1=Path(k,%Origin).y
	wleft2=Path(k,%Target).x
    wtop2=Path(k,%Target).y
    deltax=wleft2-wleft1
    deltay=wtop2-wtop1
	wleft=wleft1+deltax*here/etapes ' BERK algorithm !!! ;-)
	wtop=wtop1+deltay*here/etapes ' must perform 0..etapes
    if operation = 0 then ' erase
		PUT (wleft,wtop),Eraser,PSET
    else ' set
		PUT (wleft,wtop),Icon,OR ' avoid blocks on blocks!
    end if
NEXT
end sub