(* ---------------------------------------------------------------
Title         Q&D reboot
Author        PhG
Overview      self-explanatory !
Usage         see help
Notes

Bugs
Wish List

--------------------------------------------------------------- *)

MODULE reboot;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

IMPORT Lib;
IMPORT Str;
IMPORT IO;
IMPORT SYSTEM;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
    FINAL         = TRUE; (* while false, reboot code is never called *)

CONST
    ProgEXEname   = "REBOOT";
    ProgTitle     = "Q&D Reboot";
    ProgVersion   = "v1.0";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    errNone             = 0;
    errHelp             = 1;
    errOption           = 2;
    errParameter        = 3;
    errRedefinedCmd     = 4;
    errExpectingCmd     = 5;
    errBadVal           = 6;
    errRange            = 7;
    errTest             = 255; (* never displayed *)
CONST
    defaultgracedelay   = 10;
    mingracedelay       = 0;
    maxgracedelay       = 30;
    strDefaultGrace     = "10";
    strGraceRange       = "[0..30]";

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msgHelp =
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" [-]<cold|warm|io>[!!!] [-yes|-!!!] [-grace:#]"+nl+
nl+
"a) For safety and by design, actual reboot requires -yes option or !!! form"+nl+
"   (return code is 255 is reboot was not performed)."+nl+
"b) Grace delay (in seconds) should belong to "+strGraceRange+" range, default is "+strDefaultGrace+"."+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :        WrStr(msgHelp);
    | errOption :      S:='Unknown "|" option !';
    | errParameter :   S:='Unknown "|" parameter !';
    | errRedefinedCmd: S:="Only one of [-]<cold|warm|io> commands can be specified !";
    | errExpectingCmd: S:="Expecting one of [-]<cold|warm|io> commands !";
    | errBadVal:       S:='Illegal "|" value !';
    | errRange:        S:="|";
    | errTest:         S:="System was not rebooted !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp,errTest: ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE fmtlc (v:LONGCARD;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str80;
VAR
    R : str80;
    ok: BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    FOR i:= Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,ch);
        ELSE
            Str.Prepend(R,ch);
        END;
    END;
    IF base=16 THEN Str.Lows(R);END;
    Str.Prepend(R,prefix);
    RETURN R;
END fmtlc;

PROCEDURE fmt (v:CARDINAL;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str80;
BEGIN
    RETURN fmtlc(LONGCARD(v),base,wi,ch,prefix);
END fmt;

(* ------------------------------------------------------------ *)

PROCEDURE waitgrace (showcountdown:BOOLEAN; grace:CARDINAL);
VAR
    prev,curr:LONGINT;
    h,m,s,ss:CARDINAL;
BEGIN
    prev := MAX(LONGINT);
    LOOP
        Lib.GetTime(h,m,s,ss);
        curr := (LONGINT(h)*60+LONGINT(m))*60+LONGINT(s);
        IF curr # prev THEN
            IF showcountdown THEN WrStr( fmt (grace,10,1,"",blank) ) ; END;
            IF grace = 0 THEN EXIT;END;
            DEC(grace);
            prev:=curr;
        END;
    END;
END waitgrace;

(* ------------------------------------------------------------ *)

TYPE
    cmdtype = (donada,docold,dowarm,doio);

PROCEDURE newcmd (VAR cmd:cmdtype;wanted:cmdtype ):BOOLEAN;
VAR
    ok:BOOLEAN;
BEGIN
    ok:= ( (cmd = donada) OR (cmd=wanted) );
    IF ok THEN cmd:=wanted;END;
    RETURN ok;
END newcmd;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    code16 = ARRAY [1..16] OF BYTE;

(*# save *)
(*# call(inline_max=>66) *)


(* cold I/O
mov		al,fe
out		64,al

b0 fe e6 64 ea 90 90 90 90 90 90 ea 00 00 ff ff
*)

INLINE PROCEDURE rebootIO ()   = code16  (
0B0H,0FEH,0E6H,064H,0EAH,090H,090H,090H,
090H,090H,090H,0EAH,000H,000H,0FFH,0FFH );


(* cold
mov		ax,40
mov		ds,ax
mov		ax,0000
mov		[72],ax
jmp		ffff:0000

b8 40 00 8e d8 b8 00 00 a3 72 00 ea 00 00 ff ff
*)

INLINE PROCEDURE rebootCold () = code16  (
0B8H,040H,000H,08EH,0D8H,0B8H,000H,000H,
0A3H,072H,000H,0EAH,000H,000H,0FFH,0FFH );


(* warm
mov		ax,40
mov		ds,ax
mov		ax,1234
mov		[72],ax
jmp		ffff:0000

b8 40 00 8e d8 b8 34 12 a3 72 00 ea 00 00 ff ff
*)

INLINE PROCEDURE rebootWarm () = code16  (
0B8H,040H,000H,08EH,0D8H,0B8H,034H,012H,
0A3H,072H,000H,0EAH,000H,000H,0FFH,0FFH );


(*# call(inline=>off) *)
(*# restore *)


PROCEDURE procReboot (gracedelay:CARDINAL;cmd:cmdtype;testmode:BOOLEAN;Z:ARRAY OF CHAR);
VAR
    S:str128;
BEGIN
    S:="::: Reboot mode : |";
    Str.Subst(S,"|",Z);
    WrStr(S);WrLn;
    S:="::: Countdown   :";  (* yes, no trailing space *)
    WrStr(S);
    waitgrace(TRUE,gracedelay);
    WrLn;
    IF testmode THEN
        WrStr("::: System was not rebooted !");WrLn;
    ELSE
        WrStr("::: System trying to reboot !");WrLn;
(*%T FINAL *)
        CASE cmd OF
        | docold: rebootCold();
        | dowarm: rebootWarm();
        | doio:   rebootIO();    (* SYSTEM.Out ( 064H, BYTE( 0FEH ) ); *)
        ELSE
            ; (* NEVER !!! *)
        END;
(*%E  *)
    END;
END procReboot;

(* ------------------------------------------------------------ *)

VAR
    parmcount,i,opt : CARDINAL;
    S,U       : str128;
    cmd       : cmdtype;
    testmode  : BOOLEAN;
    gracedelay: CARDINAL;
    v         : LONGCARD;
BEGIN
    Lib.DisableBreakCheck();

    WrLn;

    cmd       := donada;
    testmode  := TRUE;
    gracedelay:= defaultgracedelay;

    parmcount := Lib.ParamCount();

    IF parmcount = 0 THEN abort(errHelp,"");END;

    FOR i := 1 TO parmcount DO (* for future extension ! *)
        Lib.ParamStr(S,i);
        Str.Copy(U,S);
        UpperCase(U);
        IF NOT(isOption(U)) THEN
            opt:= getStrIndex(delim,U,
                  "COLD,WARM,IO,"+
                  "C,W,IO,"+
                  "COLD!!!,WARM!!!,IO!!!,"+
                  "C!!!,W!!!,IO!!!"
                  );
            IF opt # 0 THEN Str.Prepend(U,"-");END;
        END;
        IF isOption(U) THEN
            opt := GetOptIndex(U, "?"+delim+"H"+delim+"HELP"+delim+
                                  "Y"+delim+"YES"+delim+"!!!"+delim+
                                  "COLD"+delim+"C"+delim+
                                  "WARM"+delim+"W"+delim+
                                  "IO"+delim+
                                  "COLD!!!"+delim+"C!!!"+delim+
                                  "WARM!!!"+delim+"W!!!"+delim+
                                  "IO!!!"+delim+
                                  "G:"+delim+"GRACE:"
                                  );
            CASE opt OF
            | 1,2,3 :  abort(errHelp,"");
            | 4,5,6 :  testmode:=FALSE;
            | 7,8   :  IF newcmd(cmd,docold)=FALSE THEN abort(errRedefinedCmd,"");END;
            | 9,10  :  IF newcmd(cmd,dowarm)=FALSE THEN abort(errRedefinedCmd,"");END;
            | 11    :  IF newcmd(cmd,doio)=FALSE THEN abort(errRedefinedCmd,"");END;
            | 12,13 :  IF newcmd(cmd,docold)=FALSE THEN abort(errRedefinedCmd,"");END;
                       testmode:=FALSE;
            | 14,15 :  IF newcmd(cmd,dowarm)=FALSE THEN abort(errRedefinedCmd,"");END;
                       testmode:=FALSE;
            | 16    :  IF newcmd(cmd,doio)=FALSE THEN abort(errRedefinedCmd,"");END;
                       testmode:=FALSE;
            | 17,18 :  IF GetLongCard(U,v)=FALSE THEN abort(errBadVal,S);END;
                       IF ( (v < mingracedelay) OR (v > maxgracedelay) ) THEN
                           abort(errRange,"Grace delay should belong to "+strGraceRange+" range !");
                       END;
                       gracedelay:=CARDINAL(v);
            ELSE
                       abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            abort(errParameter,S);
        END;
    END;

    CASE cmd OF
    | donada : abort(errExpectingCmd,"");
    | docold : S:="cold";
    | dowarm : S:="warm";
    | doio   : S:="cold using I/O port";
    END;
    procReboot(gracedelay,cmd,testmode,S);
    abort(errTest,"");

    (* should, of course, never pass this point ! *)

    abort(errNone,"");
END reboot.


(*

--------B-19---------------------------------
INT 19 - SYSTEM - BOOTSTRAP LOADER

	to accomplish a warm boot equivalent to Ctrl-Alt-Del, store 1234h in
	  0040h:0072h and jump to FFFFh:0000h.	For a cold boot equivalent to
	  a reset, store 0000h at 0040h:0072h before jumping.

----------M00400072--------------------------
MEM 0040h:0072h - POST RESET FLAG
Size:	WORD
Desc:	specify the action the BIOS should take at the beginning of the
	  power-on self-test when the machine is reset
SeeAlso: INT 19,MEM F000h:FFF0h

(Table M0021)
Values for POST reset flag:
 0000h	cold boot
 0064h	Burn-in mode
 1234h	to bypass memory test (warm boot)
 4321h	[PS/2 except Mod 25,30] to preserve memory
 5678h	[Conv] system suspended
 9ABCh	[Conv] manufacturing test mode
 ABCDh	[Conv] POST loop mode

--------H-@F000FFF0--------------------------
CALL F000h:FFF0h - COLD-BOOT ADDRESS
InstallCheck:	always present
SeeAlso: @FFFFh:0000h,INT 19

--------H-@FFFF0000--------------------------
CALL FFFFh:0000h - COLD-BOOT ADDRESS
InstallCheck:	always present
Note:	this address is another way of expressing F000h:FFF0h
SeeAlso: @F000h:FFF0h,INT 19

*)










