
(* ---------------------------------------------------------------
Title         Q&D REG2TXT
Author        PhG
Overview      see help
Usage         see help
Notes
              tentative pour faciliter la comparaison des monstrueux fichiers
              exports par regedit win9X !

              Q&D and BFI rule !

              we handled trailing "\" by merging those huge hex values
              but strings would become too huge : now, we index them

              we don't keep sections without value

              this is really Q&D but this *is* a useful small tool
              (compare before/after registry dumps, look for entries...)

Bugs          unwanted minor side-effect a.k.a. bug :
              "[]:::(01)REGEDIT 4" is always generated
Wish List

--------------------------------------------------------------- *)

MODULE reg2txt;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
    exe     = "REG2TXT";
    version = "1.0a";
    culprit = "PhG";

    msgHelp = "Q&D Registry dump to text "+version+" by "+culprit+nl+
               nl+
               "Syntax : "+exe+" <infile> <outfile>"+nl+
               nl+
               "a) <infile> must be a REGEDIT exported dump (Win9X format)."+nl+
               "b) Note sections without value are ignored."+nl;

CONST
    errNone     = 0;
    errHelp     = 1;
    errNotFound = 2;
    errOption   = 3;
    errAborted  = 255;

PROCEDURE abort (rc:CARDINAL;einfo:ARRAY OF CHAR);
VAR
    S : str128;
BEGIN
    CASE rc OF
    | errNone :     ;
    | errHelp :     WrStr(msgHelp);
    | errNotFound:  S:='"|" does not exist !';
    | errOption:    S:='Unexpected "|" option !';
    | errAborted:   S:="Aborted by user !";
    ELSE
                    S:="This is illogical, Captain !";
    END;
    CASE rc OF
    | errNone,errHelp:
        ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(exe+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(rc));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType    = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

PROCEDURE num2str (num,fix:CARDINAL;padchar:CHAR   ):str16;
VAR
    R:str16;
    ok:BOOLEAN;
BEGIN
    Str.CardToStr( LONGCARD(num),R,10,ok);
    WHILE Str.Length(R) < fix DO
        Str.Prepend(R,padchar);
    END;
    RETURN R;
END num2str;

(* ------------------------------------------------------------ *)

CONST
    gluecode    = "\";
    msg         = "Processing... ";
    sep         = ":::";
VAR
    hin,hout:FIO.File;
    ioBufferIn,ioBufferOut:ioBufferType;
    infile,outfile:pathtype;
    section,value,sndx:str1024; (* probably oversized but "who knows what evil lurks in the hearts of men ?" *)
    (* S,SS:ARRAY [0..24576-1] OF CHAR; *)
    S:str4096;
    len,index,p : CARDINAL;
    state : (gotsection,whatnext);
    zesize:LONGCARD;
    useLFN,alcatraz:BOOLEAN;
BEGIN
    WrLn;
    FIO.IOcheck:=FALSE;

    useLFN := TRUE;

    CASE Lib.ParamCount() OF
    | 2 :
         Lib.ParamStr(infile,1);
         IF isOption(infile) THEN abort(errOption,infile);END;
         Lib.ParamStr(outfile,2);
         IF isOption(outfile) THEN abort(errOption,outfile);END;
    ELSE
         abort(errHelp,"");
    END;

    useLFN := ( useLFN AND fileSupportLFN() );

    IF fileExists(useLFN,infile)=FALSE THEN abort(errNotFound,infile); END;

    video(msg,TRUE);
    zesize:=fileGetFileSize(useLFN,infile);
    completed(completedInit,zesize);

    hin:=fileOpenRead(useLFN,infile);
    FIO.AssignBuffer(hin,ioBufferIn);
    hout:=fileCreate(useLFN,outfile);
    FIO.AssignBuffer(hout,ioBufferOut);

    section:="[]"; (* forced for "REGEDIT 4" header *)
    state := gotsection;

    index := 1-1; (* safety for unwanted "REGEDIT 4" *)
    LOOP
        completed(completedShow,FIO.GetPos(hin));
        LOOP
            FIO.RdStr(hin,S);
            LtrimBlanks(S);
            RtrimBlanks(S);
            IF FIO.EOF THEN EXIT; END;
            len:=Str.Length(S);
            IF len > 0 THEN EXIT; END;
        END;

        IF FIO.EOF THEN EXIT; END;

        IF S[len-1] = gluecode THEN (* multiline *)
            S[len-1]:=nullchar; (* remove trailng "\" *)
            p:=Str.CharPos(S,equal);
            IF p # MAX(CARDINAL) THEN
                Str.Slice(value,S,0,p+1);
                Str.Delete(S,0,p+1);
                index:=1;
            ELSE
                INC(index);
            END;
            Str.Concat(sndx,value,"(");
            Str.Append(sndx,num2str(index,2,"0"));
            Str.Append(sndx,")");
            Str.Prepend(S,sndx);
        ELSE
            IF Str.CharPos(S,openbracket)=MAX(CARDINAL) THEN (* not section *)
                IF Str.CharPos(S,equal)=MAX(CARDINAL) THEN (* last of multiline *)
                    INC(index);
                    Str.Concat(sndx,value,"(");
                    Str.Append(sndx,num2str(index,2,"0"));
                    Str.Append(sndx,")");
                    Str.Prepend(S,sndx);
                END;
            END;
        END;

        CASE state OF
        | gotsection:
            CASE S[0] OF
            | openbracket:
                Str.Copy(section,S);
                state:=whatnext;
            ELSE
                Str.Prepend(S,sep);
                Str.Prepend(S,section);
                FIO.WrStr(hout,S);FIO.WrLn(hout);
            END;
        | whatnext:
            CASE S[0] OF
            | openbracket:
                Str.Copy(section,S);
                state:=gotsection;
            ELSE
                Str.Prepend(S,sep);
                Str.Prepend(S,section);
                FIO.WrStr(hout,S);FIO.WrLn(hout);
            END;
        END;
        alcatraz:=ChkEscape();
        IF alcatraz THEN EXIT; END;
    END;
    fileClose(useLFN,hout);
    fileClose(useLFN,hin);

    completed(completedEnd,0);
    video(msg,FALSE);

    IF alcatraz THEN abort(errAborted,""); END;

    WrStr("Done !");WrLn;

    abort(errNone,"");
END reg2txt.


(*

REGEDIT4

[HKEY_LOCAL_MACHINE]

[HKEY_LOCAL_MACHINE\Software]

[HKEY_LOCAL_MACHINE\Software\Microsoft]

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows]

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion]
"InstallType"=hex:03,00
"SetupFlags"=hex:08,05,00,00
"DevicePath"="C:\\WINDOWS\\INF"
"DigitalProductId"=hex:00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,\
  00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,\
  00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,\
  00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,\
  00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,\
  00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,\
  00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00
"AuditMode"=hex:00,00,00,00
"ProgramFilesPath"="C:\\Program Files"

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Setup]
"BootDir"="C:\\"

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Setup\Migration]

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Setup\Migration\600]

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Setup\Migration\650]

[HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Setup\Migration\450]
"DirectXMigration"="C:\\WINDOWS\\SYSTEM\\dxmigr.dll,DxMigration,dxwin9x"

*)


