(* ---------------------------------------------------------------
Title
Overview
Usage
Notes
Bugs
Wish list

--------------------------------------------------------------- *)

MODULE unApple;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;

FROM IO IMPORT WrStr,WrLn;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    cr          = CHR(13);
    lf          = CHR(10);
    nl          = cr+lf;
    dot         = ".";
    extDSK      = ".DSK";
    extCAT      = ".CAT";
    letters     = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    digits      = "0123456789";
    legal       = letters+digits;

CONST
    ProgEXEname   = "UNAPPLE";
    ProgTitle     = "Q&D Apple ][ DOS 3.3 disk image to PC files extractor";
    ProgVersion   = "v1.0a";
    Banner        = ProgTitle+" "+ProgVersion;
CONST
    errNone             = 0;
    errHelp             = 1;
    errOption           = 2;
    errParmOverflow     = 3;
    errJoker            = 4;
    errExpecting        = 5;
    errNotFound         = 6;
    errExt              = 7;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <disk_image["+extDSK+"]> [-lfn]"+nl+
"Syntax 2 : "+ProgEXEname+" <-list> <tokenized_program>"+nl+
nl+
"With syntax 1, this program extracts files from an Apple ][ DOS 3.3 disk image."+nl+
"With syntax 2, it converts a tokenized Applesoft BASIC program to ASCII."+nl;

VAR
    S : pathtype;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(helpmsg);
    | errOption :
        Str.Concat(S,"Unknown ",einfo);Str.Append(S," option !");
    | errParmOverflow :
        Str.Concat(S,"Uneeded ",einfo);Str.Append(S," parameter !");
    | errJoker:
        Str.Concat(S,"At least one illegal joker in ",einfo);
        Str.Append(S," specification !");
    | errExpecting:
        Str.Concat(S,einfo," expected !");
    | errNotFound :
        Str.Concat(S,einfo," does not exist !");
    | errExt:
        Str.Concat(S,einfo," should have "+extDSK+" extension !");
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp :
        ; (* nada *)
    ELSE
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE fmtnum (v:LONGCARD;base:CARDINAL;padwi:INTEGER;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    IF base=16 THEN Str.Lows(R);END;
    FOR i:=Str.Length(R)+1 TO ABS(padwi) DO
        IF padwi < 0 THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
    RETURN R;
END fmtnum;

PROCEDURE padstr (wi:INTEGER;S:ARRAY OF CHAR  ):str1024;
VAR
    i:CARDINAL;
    R:str1024;
BEGIN
    Str.Copy(R,S);
    FOR i:=Str.Length(S)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Prepend(R," ");
        ELSE
            Str.Append(R," ");
        END;
    END;
    RETURN R;
END padstr;

PROCEDURE dmp ( S:ARRAY OF CHAR; v:CARDINAL  );
BEGIN
    WrStr( padstr(14-1,S) );
    WrStr(" : ");
    WrStr( fmtnum( LONGCARD(v),10,5," ") );WrLn;
END dmp;

PROCEDURE dmp2 ( S:ARRAY OF CHAR; v,v2:CARDINAL  );
BEGIN
    WrStr( padstr(14-1,S) );
    WrStr(" : ");
    WrStr( fmtnum( LONGCARD(v),10,5," ") );
    WrStr(" , ");
    WrStr( fmtnum( LONGCARD(v2),10,5," ") );WrLn;
END dmp2;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType  = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn,ioBufferOut : ioBufferType;

(* ------------------------------------------------------------ *)

CONST
    sectorsPerTrack = 16;
    blocksize  = 256;
    firstblockbuffer = 0;
    lastblockbuffer = blocksize-1;
TYPE
    blockBufferType = ARRAY [firstblockbuffer..lastblockbuffer] OF BYTE;
VAR
    vtoc,cat,tslist,data : blockBufferType;

(* ------------------------------------------------------------ *)

PROCEDURE readBlock (hin:FIO.File;track,sector:CARDINAL;
          VAR buf:blockBufferType);
VAR
    offset : LONGCARD;
    got:CARDINAL;
BEGIN
	offset := LONGCARD (track) * blocksize * sectorsPerTrack + LONGCARD (sector) * blocksize;
    FIO.Seek(hin,offset);
    got:=FIO.RdBin(hin,buf,SIZE(buf));
END readBlock;

PROCEDURE getbyte (p:CARDINAL;buf:blockBufferType):CARDINAL;
BEGIN
    RETURN CARDINAL( buf[p] );
END getbyte;

PROCEDURE getword (p:CARDINAL;buf:blockBufferType):CARDINAL;
BEGIN
    RETURN ( getbyte(p,buf) + 256 * getbyte(p+1,buf) );
END getword;

(* ------------------------------------------------------------ *)

PROCEDURE showVTOC (VAR nexttrack,nextsector:CARDINAL;
                   useLFN,show:BOOLEAN;hin:FIO.File;src:pathtype);
VAR
    tracks,sectors,total:CARDINAL;
BEGIN

    readBlock(hin,11H,0,vtoc);

    tracks:=getbyte(34H,vtoc);
    sectors:=getbyte(35H,vtoc);
    total:=getword(36H,vtoc);
    nexttrack:=getbyte(01H,vtoc);
    nextsector:=getbyte(02H,vtoc);
    IF show THEN
        dmp("Tracks"             ,tracks);
        dmp("Sectors"            ,sectors);
        dmp("Total sectors"      ,total);
        dmp("First track"        ,nexttrack);
        dmp("First sector"       ,nextsector);
    END;
END showVTOC;

(* ------------------------------------------------------------ *)

PROCEDURE dmpTSL (useLFN, DEBUG:BOOLEAN; trtsl, setsl, ftype:CARDINAL;
                  hin:FIO.File; nam, dosnam:pathtype);
VAR
    suffix : CHAR;
    hout:FIO.File;
    F:pathtype;
    secnum,trdata,sedata:CARDINAL;
    tslflag:BOOLEAN;
BEGIN
    CASE ftype OF
    | 00H : suffix := "T";
    | 04H : suffix := "B";
    | 02H : suffix := "A";
    ELSE
            suffix := "X";
    END;
    IF useLFN THEN
        Str.Append(nam,".");
        Str.Append(nam,suffix);
        F:=nam;
    ELSE
        Str.Append(dosnam,suffix); (* .##? *)
        F:=dosnam;
    END;
IF DEBUG THEN WrStr("// ");WrStr(F);WrLn;END;
    hout:=fileCreate(useLFN,F);
    FIO.AssignBuffer(hout,ioBufferOut);

    tslflag:=FALSE;
    LOOP
		readBlock( hin, trtsl,setsl,tslist);
		trtsl := getbyte(1,tslist);
		setsl := getbyte(2,tslist);
		secnum := 0;
		LOOP
		    trdata := getbyte ( 0CH + secnum * 2    ,tslist );
		    sedata := getbyte ( 0CH + secnum * 2 + 1,tslist );
		    IF trdata = 0 THEN tslflag := TRUE; EXIT; END;
		    readBlock( hin, trdata,sedata, data);
            IF DEBUG THEN dmp2("tr,se data", trdata,sedata); END;
		    FIO.WrBin(hout,data,SIZE(data));
            INC(secnum);
	 	    IF secnum >= 120 THEN EXIT ;END;
		END;
	    IF trtsl = 0 THEN tslflag:= TRUE; END;
	    IF tslflag THEN EXIT; END;
    END;
    fileClose(useLFN,hout);
END dmpTSL;

(* ------------------------------------------------------------ *)

PROCEDURE procFile(VAR nexttrack,nextsector,thisname:CARDINAL;
                       useLFN,DEBUG:BOOLEAN;hin,hlog:FIO.File):BOOLEAN;
VAR
    nse,c,basen,entry,basentry,trtsl,setsl,ftype,namndx:CARDINAL;
    nam,dosnam,tmpnam:pathtype;
    ch:CHAR;
    S:str1024;
BEGIN
    readBlock(hin,nexttrack,nextsector, cat);
    nexttrack:=getbyte(01H,cat);
    nextsector:=getbyte(02H,cat);

	basen := 0BH;
	entry := 0;
	LOOP
		basentry := basen + entry * 23H;
		trtsl := getbyte (basentry,cat);
	    CASE trtsl OF
	    | 0 : RETURN FALSE;
	    | 1..254 :
		    setsl := getbyte (basentry+1,cat);
		    ftype := getbyte (basentry+2,cat);
		    IF ftype > 127 THEN DEC(ftype,128);END;
		    nam := "";
		    tmpnam := "";
		    FOR namndx := 3 TO 20H DO
			    c:=getbyte (basentry+namndx,cat);
			    DEC(c,128);
			    ch:=CHR(c);
			    Str.Append(nam, ch);
			    ch:=CAP(ch);
			    IF Str.CharPos(legal,ch) # MAX(CARDINAL) THEN Str.Append(tmpnam,ch);END;
     		END;
            Str.Slice(dosnam,tmpnam,0,8);
            LtrimBlanks(nam);RtrimBlanks(nam);
            LtrimBlanks(dosnam);RtrimBlanks(dosnam);
            Str.Append(dosnam,".");
            Str.Append(dosnam, fmtnum (LONGCARD(thisname),10,2,"0") );

		    INC (thisname);
		    nse := getword(basentry+21H,cat);

            Str.Copy(S,fmtnum(LONGCARD(ftype),16,2,"0"));
            Str.Append(S,"  ");
            Str.Append(S,fmtnum(LONGCARD(nse),10,5," "));
            Str.Append(S,"  ");
            Str.Append(S,padstr(-8-1-3,dosnam));
            Str.Append(S,"  ");
            Str.Append(S,nam);
            WrStr(S);WrLn;
	        FIO.WrStr(hlog,S);FIO.WrLn(hlog);
	        IF DEBUG THEN
	            dmp2 ("trtsl,setsl",trtsl,setsl);
	        END;
            dmpTSL(useLFN, DEBUG, trtsl, setsl, ftype, hin, nam, dosnam);
	   END;
	   INC (entry);
	   IF entry >= 7 THEN EXIT; END;
	END;
	RETURN (nexttrack # 0 );
END procFile;

(* track $00..$22, sectors $0..$F *)

PROCEDURE procDiskImage (useLFN,DEBUG:BOOLEAN;src:pathtype  );
VAR
    nexttrack,nextsector,thisname:CARDINAL;
    hin,hlog:FIO.File;
    catalog:pathtype;
BEGIN
    hin:=fileOpenRead(useLFN,src);
    FIO.AssignBuffer(hin,ioBufferIn);
    catalog:=src;
    Str.Subst(catalog,extDSK,extCAT);
    hlog:=fileCreate(useLFN,catalog);

    showVTOC(nexttrack,nextsector, useLFN, DEBUG ,hin, src);

    thisname:=1;
    LOOP
       IF procFile(nexttrack,nextsector,thisname,useLFN,DEBUG,hin,hlog) = FALSE THEN EXIT; END;
    END;

    fileClose(useLFN,hlog);
    fileClose(useLFN,hin);
END procDiskImage;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    hbar = "|";
CONST
    tokens =
    "END"+hbar+    (* $80 *)
    "FOR"+hbar+
    "NEXT"+hbar+
    "DATA"+hbar+
    "INPUT"+hbar+
    "DEL"+hbar+
    "DIM"+hbar+
    "READ"+hbar+
    "GR"+hbar+
    "TEXT"+hbar+
    "PR#"+hbar+
    "IN#"+hbar+
    "CALL"+hbar+
    "PLOT"+hbar+
    "HLIN"+hbar+
    "VLIN"+hbar+
    "HGR2"+hbar+         (* $90 *)
    "HGR"+hbar+
    "HCOLOR="+hbar+
    "HPLOT"+hbar+
    "DRAW"+hbar+
    "XDRAW"+hbar+
    "HTAB"+hbar+
    "HOME"+hbar+
    "ROT="+hbar+
    "SCALE="+hbar+
    "SHLOAD"+hbar+
    "TRACE"+hbar+
    "NOTRACE"+hbar+
    "NORMAL"+hbar+
    "INVERSE"+hbar+
    "FLASH"+hbar+
    "COLOR="+hbar+             (* $a0 *)
    "POP"+hbar+
    "VTAB"+hbar+
    "HIMEM:"+hbar+
    "LOMEM:"+hbar+
    "ONERR"+hbar+
    "RESUME"+hbar+
    "RECALL"+hbar+
    "STORE"+hbar+
    "SPEED="+hbar+
    "LET"+hbar+
    "GOTO"+hbar+
    "RUN"+hbar+
    "IF"+hbar+
    "RESTORE"+hbar+
    "&"+hbar+
    "GOSUB"+hbar+                    (* $b0 *)
    "RETURN"+hbar+
    "REM"+hbar+
    "STOP"+hbar+
    "ON"+hbar+
    "WAIT"+hbar+
    "LOAD"+hbar+
    "SAVE"+hbar+
    "DEF"+hbar+
    "POKE"+hbar+
    "PRINT"+hbar+
    "CONT"+hbar+
    "LIST"+hbar+
    "CLEAR"+hbar+
    "GET"+hbar+
    "NEW"+hbar+
    "TAB("+hbar+            (* $c0 *)
    "TO"+hbar+
    "FN"+hbar+
    "SPC("+hbar+
    "THEN"+hbar+
    "AT"+hbar+
    "NOT"+hbar+
    "STEP"+hbar+
    "+"+hbar+
    "-"+hbar+
    "*"+hbar+
    "/"+hbar+
    "^"+hbar+
    "AND"+hbar+
    "OR"+hbar+
    ">"+hbar+
    "="+hbar+                       (* $d0 *)
    "<"+hbar+
    "SGN"+hbar+
    "INT"+hbar+
    "ABS"+hbar+
    "USR"+hbar+
    "FRE"+hbar+
    "SCRN("+hbar+
    "PDL"+hbar+
    "POS"+hbar+
    "SQR"+hbar+
    "RND"+hbar+
    "LOG"+hbar+
    "EXP"+hbar+
    "COS"+hbar+
    "SIN"+hbar+
    "TAN"+hbar+                           (* $e0 *)
    "ATN"+hbar+
    "PEEK"+hbar+
    "LEN"+hbar+
    "STR$"+hbar+
    "VAL"+hbar+
    "ASC"+hbar+
    "CHR$"+hbar+
    "LEFT$"+hbar+
    "RIGHT$"+hbar+                              (* $e9 *)
    "MID$"+hbar+
    "// UNDEFINED TOKEN //";

PROCEDURE rdw (h:FIO.File  ):CARDINAL;
VAR
    w,got:CARDINAL;
BEGIN
    got:=FIO.RdBin(h,w,SIZE(w));
    RETURN w;
END rdw;

PROCEDURE rdb (h:FIO.File  ):CARDINAL;
VAR
    b:SHORTCARD;
    got:CARDINAL;
BEGIN
    got:=FIO.RdBin(h,b,SIZE(b));
    RETURN CARDINAL(b);
END rdb;

PROCEDURE procBASIC (useLFN,DEBUG:BOOLEAN;src:pathtype  );
VAR
    lastaddr:LONGCARD;
    currline,mystery,token : CARDINAL;
    hin:FIO.File;
    R:str16;
BEGIN
    hin:=fileOpen(useLFN,src);
    FIO.AssignBuffer(hin,ioBufferIn);
    lastaddr := LONGCARD ( rdw(hin) );
    LOOP
        IF FIO.GetPos(hin) > lastaddr THEN EXIT; END; (* safety *)
        mystery := rdw(hin);
        IF mystery = 0 THEN EXIT; END;
        currline := rdw(hin);
        WrStr( fmtnum( LONGCARD(currline),10,5," "));
        WrStr(" ");
        LOOP
            token:=rdb(hin);
            CASE token OF
            | 0 : WrLn; EXIT;
            | 1..31    : WrStr( CHR(token) );
            | 32..127  : WrStr( CHR(token) );
            | 128..255 : isoleItemS(R, tokens, hbar, token-128);
                         WrStr(" ");WrStr(R);WrStr(" ");
            END;
        END;
    END;
    fileClose(useLFN,hin);
END procBASIC;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    parmcount,i,opt  : CARDINAL;
    S,R,src          : pathtype;
    DEBUG,useLFN,untoken : BOOLEAN;
    status           : (waiting,gotsrc);
BEGIN
    FIO.IOcheck := FALSE;
    WrLn;

    DEBUG := FALSE;
    useLFN := TRUE;
    untoken := FALSE;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;
    status := waiting;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S);
        UpperCase(R);  cleantabs(R); (* YATB ! *)
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "X"+delim+"LFN"+delim+
                                   "L"+delim+"LIST"+delim+
                                   "DEBUG"
                               );
            CASE opt OF
            |  1,  2,  3 : abort(errHelp,"");
            |  4,  5     : useLFN := FALSE;
            |  6,  7     : untoken := TRUE;
            |  8         : DEBUG := TRUE;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            CASE status OF
            | waiting  : Str.Copy(src,R); (* keep upper case *)
            | gotsrc   : abort(errParmOverflow,S);
            END;
            INC(status);
        END;
    END;

    CASE status OF
    | waiting :
        IF untoken THEN
            S:="<tokenized_program>";
        ELSE
            S:="<disk_image["+extDSK+"]>";
        END;
        abort(errExpecting,S);
    | gotsrc:
        ;
    END;

    (* yes, we know we won't trap ..\* form but we're getting VERY old *)

    IF chkJoker(src) THEN abort(errJoker,src);END;

    IF NOT(untoken) THEN
        IF Str.RCharPos(src,dot)=MAX(CARDINAL) THEN Str.Append(src,extDSK);END;
        IF Str.Pos(src,extDSK)=MAX(CARDINAL) THEN abort(errExt,src);END;
    END;

    useLFN := (useLFN AND fileSupportLFN() );
    IF fileExists(useLFN,src)=FALSE THEN abort(errNotFound,src);END;

    IF untoken THEN
        procBASIC(useLFN,DEBUG,src);
    ELSE
        procDiskImage(useLFN,DEBUG,src);
    END;

    abort(errNone,"");
END unApple.



