(* ---------------------------------------------------------------
Title         Q&D YouTube transcription to SRT format
Overview      see help
Usage         see help
Notes
              model cannot be small
Bugs
Wish List     force range and text divide ?
              seems out of sync at times !
              persistence according to charcount ?
              persistence limit IF more than a few seconds ?
              user specified target name ?
              handle multilines (seen in a few transcriptions)

--------------------------------------------------------------- *)

MODULE YT2SRT;

IMPORT Lib;
IMPORT Str;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    progEXE    = "YT2SRT";
    progTitle  = "Q&D YT transcription to SRT format converter";
    progVersion= "v1.0e";
    progCulprit= "by PhG";
    extSRT     = ".SRT";
    extTX      = ".TX";
    herelocal  = TRUE; (* create in currdir, else in SRC dir *)
CONST
    (* could be vars *)
    defaultsubwidth=66; strdefaultsubwidth = "66";
    minsubwidth= 0; maxsubwidth = 100;
    finalpersistance = 30; (* in seconds, oversized *)
CONST
    errNone    = 0;
    errHelp    = 1;
    errOpt     = 2;
    errOverflow= 3;
    errJoker   = 4;
    errRsrvdExt= 5;
    errNotFound= 6;
    errExists  = 7;
    errExpected= 8;
    errSrcFmt  = 9;
    errBadNumber=10;
    errRange   = 11;

PROCEDURE abort (errcode:CARDINAL; info:ARRAY OF CHAR);
CONST
    msgHelp = progTitle+" "+progVersion+" "+progCulprit+nl+
nl+
"Syntax : "+progEXE+" <source> [-f] [-w:#] [-t] [-m] [-o]"+nl+
nl+
'  -f   create "'+extSRT+'" target (default is dumping to screen)'+nl+
"  -w:# subtitle line width (default is "+strdefaultsubwidth+", 0=as is)"+nl+
"  -t   keep text only"+nl+
"  -m   allow multiline"+nl+
'  -o   overwrite existing "'+extSRT+'" target'+nl+
nl+
"This program tries and converts <source> YT transcription to SRT format."+nl+
nl+
"a) <source> is made of one or more two-lines sets :"+nl+
nl+
"      [#:]#:#"+nl+
"      $"+nl+
nl+
'   where "[#:]#:#" is time in [hh:]mm:ss format, and "$" a line of text.'+nl+
'b) Note "[music]" and "[applause]" are filtered out.'+nl+
"c) Target will be created in current directory."+nl+
'd) -t option omits numbering and time ; it forces "'+extTX+'" extension.'+nl;

VAR
    S:str1024;
BEGIN
    CASE errcode OF
    | errNone:     ;
    | errHelp:     WrStr(msgHelp);
    | errOpt:      S:='Unknown "~" option !';
    | errOverflow: S:='"~" is one parameter too many !';
    | errJoker:    S:="Jokers are not supported !";
    | errRsrvdExt: S:='Reserved extension in "~" <source> !';
    | errNotFound: S:='"~" does not exist !';
    | errExists:   S:='"~" already exists !';
    | errExpected: S:="<source> expected !";
    | errSrcFmt:   S:='Unexpected file format in "~" <source> !';
    | errBadNumber:S:='"Illegal "~" number !';
    | errRange:    S:='"~" is out of valid range !';
    ELSE
               S:="";
    END;
    CASE errcode OF
    | errNone,errHelp:
    ELSE
        Str.Subst(S,"~",info);
        WrStr(progEXE);WrStr(" : ");WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(errcode));
    HALT;
END abort;

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferSrc,ioBufferDst:ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fmtnum (v:LONGCARD;base,wi:CARDINAL;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr( LONGCARD(v),R,base,ok);
    FOR i:=Str.Length(R)+1 TO wi DO Str.Prepend(R,padchar);END;
    RETURN R;
END fmtnum;

(* [hh:]mm:ss *)

PROCEDURE str2time (S:ARRAY OF CHAR):LONGCARD;
VAR
    i,p:CARDINAL;
    v,hh,mm,ss:LONGCARD;
    ok:BOOLEAN;
    N:str16;
BEGIN
    IF CharCount(S,colon)=1 THEN Str.Prepend(S,"0:");END; (* fake hh *)
    FOR i:=1 TO 3 DO
        p:=Str.CharPos(S,colon);
        IF p # MAX(CARDINAL) THEN  (* safety check //V10A *)
            Str.Slice(N,S,0,p);
            Str.Delete(S,0,p+1);
        ELSE
            Str.Copy(N,S);
        END;
        v:= Str.StrToCard(N,10,ok);
        CASE i OF
        | 1: hh:=v;
        | 2: mm:=v;
        | 3: ss:=v;
        END;
    END;
    v:= hh*3600 + mm*60 + ss; (* more than DOS MAX(CARDINAL) *)
    RETURN v;
END str2time;

(* hh:mm:ss,### *)

PROCEDURE time2str (v:LONGCARD;zeromilli:BOOLEAN):str16;
VAR
    R:str16;
    n:LONGCARD;
BEGIN
    IF zeromilli THEN
        R:="~:~:~,000";
    ELSE
        R:="~:~:~,999";
    END;
    n := v DIV 3600;
    Str.Subst(R,"~",fmtnum(n,10,2,"0"));
    v:= v - n*3600;
    n := v DIV 60;
    Str.Subst(R,"~",fmtnum(n,10,2,"0"));
    v:= v - n*60;
    Str.Subst(R,"~",fmtnum(v,10,2,"0"));
    RETURN R;
END time2str;

(*

0:00
xxxxxxxxxxxxxxxx
0:06
xxxxxxxxxxxxx

...

54:26
xxxxxxxxxxxxxxxxxxx

*)

(*

1
hh:mm:ss,000 --> hh:mm:ss,000
xxxxx
xxxxxxxx
xxxx

2
hh:mm:ss,000 --> hh:mm:ss,000
xxxxxxx

...

63
hh:mm:ss,000 --> hh:mm:ss,000
xxxxxxxxxxxxxxxxx
xxxxx

*)

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procConvert (DEBUG,useLFN,
                      flagVid,flagTextOnly,flagSingleLineOnly:BOOLEAN;
                      subwidth:CARDINAL;SRC,DST:pathtype ):BOOLEAN;
VAR
    hin,hout:FIO.File;
    state:(waiting,gottime,gottext);
    lastsub : CARDINAL;
    N,strHMS,strHMSnext : str16;
    Z,txt:str1024;
    continue,rc:BOOLEAN;
    starttime,endtime:LONGCARD;
    duration : str80;
BEGIN
    hin:=fileOpenRead(useLFN,SRC);
    FIO.AssignBuffer(hin,ioBufferSrc);
    IF flagVid THEN
        hout:=FIO.StandardOutput;
    ELSE
        hout:=fileCreate(useLFN,DST);
        FIO.AssignBuffer(hout,ioBufferDst);
    END;
    rc:=TRUE;
    lastsub := 1-1;
    state:=waiting;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,Z);
        LtrimBlanks(Z);RtrimBlanks(Z);
        IF same(Z,"") = FALSE THEN
            IF DEBUG THEN WrStr("DBG : ");WrStr(Z);WrLn;END;
            CASE state OF
            | waiting:
                IF verifyString(Z,digits+colon) THEN
                    Str.Copy(strHMS,Z);
                    state:=gottime;
                END;
            | gottime:
                Str.Copy(txt,Z);
                IF Str.Match(txt,"[*]") THEN (* music, applause *)
                    state:=waiting;
                ELSE
                    (* //V10D handle ##:##, emptied line, ##:## *)
                    IF verifyString(Z,digits+colon) THEN
                        Str.Copy(strHMS,Z);
                        state:=gottime;
                    ELSE
                        state:=gottext;
                    END;
                END;
                (* handle inner [music] AND [applause] ? *)
            | gottext:
                IF flagSingleLineOnly THEN
                    rc:=verifyString(Z,digits+colon);
                    IF NOT(rc) THEN EXIT;END;
                END;
                Str.Copy(strHMSnext,Z);
                starttime:=str2time(strHMS);
                endtime:=str2time(strHMSnext);
                rc:=(starttime < endtime); (* IF NOT(rc) THEN EXIT;END; //V10C *)
                N:=fmtnum( LONGCARD(lastsub),10,1,"");
                IF NOT (flagTextOnly) THEN
                    FIO.WrStr(hout,N);FIO.WrLn(hout);
                END;

                duration := "~ --> ~";
                Str.Subst(duration,"~",time2str(starttime,TRUE) );
                Str.Subst(duration,"~",time2str(endtime-1,FALSE) );
                IF NOT (flagTextOnly) THEN
                    FIO.WrStr(hout,duration);FIO.WrLn(hout);
                END;

                IF subwidth = 0 THEN
                    FIO.WrStr(hout,txt);
                ELSE
                    continue:=dmpTTX(txt,subwidth,Z,TRUE);
                    WHILE continue DO
                        FIO.WrStr(hout,Z);FIO.WrLn(hout);
                        continue:=dmpTTX(txt,subwidth,Z,FALSE);
                    END;
                END;
                IF NOT (flagTextOnly) THEN FIO.WrLn(hout); END; (* // *)

                INC(lastsub);
                strHMS:=strHMSnext;
                state:=gottime;
            END;
        END;
    END;
    CASE state OF
    | gottext:
                N:=fmtnum( LONGCARD(lastsub),10,1,"");
                IF NOT (flagTextOnly) THEN
                    FIO.WrStr(hout,N);FIO.WrLn(hout);
                END;
                duration := "~ --> ~";
                Str.Subst(duration,"~",time2str(starttime,TRUE));
                Str.Subst(duration,"~",time2str(starttime+finalpersistance-1,FALSE));
                IF NOT (flagTextOnly) THEN
                    FIO.WrStr(hout,duration);FIO.WrLn(hout);
                END;

                IF subwidth = 0 THEN
                    FIO.WrStr(hout,txt);
                ELSE
                    continue:=dmpTTX(txt,subwidth,Z,TRUE);
                    WHILE continue DO
                        FIO.WrStr(hout,Z);FIO.WrLn(hout);
                        continue:=dmpTTX(txt,subwidth,Z,FALSE);
                    END;
                END;
                IF NOT (flagTextOnly) THEN FIO.WrLn(hout); END;
    END;

    IF NOT(flagVid) THEN fileClose(useLFN,hout); END;
    fileClose(useLFN,hin);

    RETURN rc;
END procConvert;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE buildTarget (VAR DST:pathtype;
                       SRC:pathtype;ext:str16;here:BOOLEAN);
VAR
    p:CARDINAL;
    u,d,n,e:pathtype; (* u and e are oversized *)
BEGIN
    p:=Str.RCharPos(SRC,dot);
    IF p = MAX(CARDINAL) THEN
        Str.Concat(DST,SRC,ext);
    ELSE
        Str.Slice(DST,SRC,0,p);
        Str.Append(DST,ext);
    END;
    IF here THEN
        qdSplitAllPath (DST, u,d,n,e);
        Str.Concat(DST,n,e);
    END;
END buildTarget;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
CONST
    minparm = 1;
    maxparm = 1;
VAR
    parm : ARRAY[minparm..maxparm] OF pathtype;
    SRC,DST : pathtype;
    S,R:str128;
    DEBUG,useLFN:BOOLEAN;
    flagVid,flagOverwrite,flagTextOnly,flagSingleLineOnly : BOOLEAN;
    subwidth:CARDINAL;
    i,parmcount,lastparm,opt: CARDINAL;
    lc:LONGCARD;
    fext : str16;
BEGIN
    FIO.IOcheck := FALSE;

    WrLn;

    flagVid      := TRUE;
    flagTextOnly := FALSE;
    flagSingleLineOnly:=TRUE;
    flagOverwrite:= FALSE;
    SRC          := "";
    subwidth     := defaultsubwidth;
    Str.Copy(fext,extSRT);
    DEBUG        := FALSE;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;
    lastparm := minparm-1;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "F"+delim+"FILE"+delim+
                                  "W:"+delim+"WIDTH:"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "T"+delim+"TEXT"+delim+
                                  "M"+delim+"MULTILINE"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3    : abort(errHelp,"");
            | 4,5      : flagVid      := FALSE;
            | 6,7      : IF GetLongCard(R,lc)=FALSE THEN abort(errBadNumber,S);END;
                         IF lc < LONGCARD(minsubwidth) THEN abort(errRange,S);END;
                         IF lc > LONGCARD(maxsubwidth) THEN abort(errRange,S);END;
                         subwidth := CARDINAL(lc);
            | 8,9      : flagOverwrite:= TRUE;
            | 10,11    : flagTextOnly := TRUE; Str.Copy(fext,extTX);
            | 12,13    : flagSingleLineOnly := FALSE;
            | 14       : DEBUG := TRUE;
            ELSE
                abort(errOpt,S);
            END;
        ELSE
            Lib.ParamStr(SRC,i);
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errOverflow,SRC); END;
            Str.Copy(parm[lastparm],SRC);
        END;
    END;
    IF lastparm = minparm-1 THEN abort(errExpected,"");END;
    Str.Copy(SRC,parm[minparm]);

    useLFN := fileSupportLFN();

    IF same(SRC,dot) THEN Str.Copy(SRC,stardotstar); END;
    IF same(SRC,star) THEN Str.Copy(SRC,stardotstar); END;
    IF Str.Match(SRC,"*"+antislash) THEN Str.Append(SRC,stardotstar); END;
    IF chkJoker(SRC) THEN abort(errJoker,SRC);END;
    Str.Concat(R,"*",fext);
    IF Str.Match(SRC,R) THEN abort(errRsrvdExt,SRC);END;

    IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;

    buildTarget(DST,SRC,fext,herelocal);
    IF NOT(flagVid) THEN
        IF fileExists(useLFN,DST) THEN
            IF NOT(flagOverwrite) THEN abort(errExists,DST);END;
        END;
    END;

    IF procConvert (DEBUG,useLFN,flagVid,flagTextOnly,flagSingleLineOnly,
                   subwidth,SRC,DST)=FALSE THEN
        abort(errSrcFmt,SRC);
    END;
    IF NOT(flagVid) THEN
        WrStr('+++ "');WrStr(DST);WrStr('"');WrLn;
    END;

    abort(errNone,"");
END YT2SRT.

