(* ---------------------------------------------------------------
Title         Q&D File Entropy
Overview      see help
Usage         see help
Notes         sfile uses another method
Bugs
Wish List     esc ?

--------------------------------------------------------------- *)

MODULE fentropy;

IMPORT Lib;
IMPORT Str;
IMPORT FIO;
IMPORT MATHLIB;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XopenFile, w9XcloseFile,
w9XsupportLFN;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minfix        = 1;
    maxfix        = 10;
    defaultfix    = 6;
    strdefaultfix = "6";
    strfixrange   = "[1..10]";
CONST
    errNone            = 0;
    errHelp            = 1;
    errOption          = 2;
    errExpecting       = 3;
    errParmOverflow    = 4;
    errJoker           = 5;
    errDir             = 6;
    errNotFound        = 7;
    errBadNumber       = 8;
    errFixRange        = 9;
    errBadSpec         = 10;
CONST
    exe                = "FENTROPY";
    title              = "Q&D File Entropy";
    version            = "1.0";
    author             = "by PhG";

PROCEDURE abort (rc:CARDINAL;einfo:ARRAY OF CHAR);
CONST
    msgHelp     =
title+" "+version+" "+author+nl+
nl+
"Syntax : "+exe+" <filespec> [option]..."+nl+
nl+
"  -f:# precision digits (valid range is "+strfixrange+", default is "+strdefaultfix+")"+nl+
"  -us  American decimal separator (default is French decimal separator)"+nl+
"  -x   disable LFN support even if available"+nl+
nl+
"This program computes <filespec> entropy."+nl+
"Results should match CCOUNT -ent mode."+nl;

VAR
    S:str1024;
BEGIN
    CASE rc OF
    | errNone:         S := "";
    | errHelp:         WrStr(msgHelp);
    | errOption:       S := 'Unknown "|" option !';
    | errExpecting:    S := "Expecting | !";
    | errParmOverflow: S := '"|" is one parameter too many !';
    | errJoker:        S := '| cannot contain any joker !';
    | errDir:          S := '| must be a file !';
    | errNotFound:     S := '"|" does not exist !';
    | errBadNumber:    S := 'Illegal "|" number !';
    | errFixRange:     S := '"|" is out of '+strfixrange+" range !";
    | errBadSpec:      S := 'Illegal "|" <filespec> !';
    ELSE
                       S := "This is illogical, Captain !";
    END;
    CASE rc OF
    | errNone,errHelp:
        ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(exe+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode( SHORTCARD(rc) );
    HALT;
END abort;

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBufferSrc : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)

TYPE
    HUGECARD = LONGCARD; (* or LONGREAL ? *)
CONST
    HUGEZERO = LONGREAL(0.0);
CONST
    minval = MIN(BYTE);
    maxval = MAX(BYTE);
VAR
    freq : ARRAY[minval..maxval] OF HUGECARD;

(* ------------------------------------------------------------ *)

CONST
    databuffersize = 16384;
VAR
    databuffer : ARRAY[0..databuffersize-1] OF BYTE;

PROCEDURE procFile (useLFN,debug:BOOLEAN;src:pathtype  ):LONGREAL;
VAR
    ent,f:LONGREAL;
    fsize:LONGCARD;
    i,wanted,got,v:CARDINAL;
    hin:FIO.File;
BEGIN
    fsize:=fileGetFileSize(useLFN,src);
    IF fsize=0 THEN RETURN HUGEZERO;END;

    FOR i:=minval TO maxval DO freq[i]:=0;END;

    hin:=fileOpenRead(useLFN,src);
    FIO.AssignBuffer(hin,ioBufferSrc);
    LOOP
        wanted:=databuffersize;
        got:=FIO.RdBin(hin,databuffer,wanted);
        IF got = 0 THEN EXIT;END;
        FOR i:=0 TO got-1 DO
            v:=CARDINAL( databuffer[i] );
            INC(freq[v]);
        END;
        IF got # wanted THEN EXIT;END;
    END;
    fileClose(useLFN,hin);

    ent := 0.0;
    FOR i := minval TO maxval DO
        IF freq[i] # 0 THEN
            f := LONGREAL(freq[i]) / LONGREAL(fsize);
            ent := ent - f * (MATHLIB.Log(f) / MATHLIB.Log(2.0));
        END;
    END;
    RETURN ent;
END procFile;

PROCEDURE floattostr (VAR R:ARRAY OF CHAR;
                     commachar:CHAR;fix:CARDINAL;v:LONGREAL);
VAR
    ok:BOOLEAN;
BEGIN
    Str.FixRealToStr(v,fix,R,ok);
    IF NOT(ok) THEN Str.Copy(R,"???");END;
    Str.Subst(R,dot,commachar);
END floattostr;

(* ------------------------------------------------------------ *)

PROCEDURE isReservedEntry (S:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE procDir (useLFN,debug:BOOLEAN;commachar:CHAR;fix:CARDINAL;
                  spec,basedir:pathtype);
VAR
    ok,found:BOOLEAN;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    entry : FIO.DirEntry;
    dosattr:FIO.FileAttr;
    entryname:pathtype;
    len : CARDINAL;
    countmeout:BOOLEAN;
    Z:str1024;
    R:str80;
    ent:LONGREAL;
BEGIN
    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,entry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(entryname,w9Xentry.fullfilename);
        ELSE
            Str.Copy(entryname,entry.Name);
        END;
        countmeout := isReservedEntry   (entryname);  (* skip "." and ".." *)
        IF NOT(countmeout) THEN
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=entry.attr;
            END;
            IF NOT (aD IN dosattr) THEN (* ignore dirs *)

                Str.Prepend(entryname,basedir);
                Z:='Computing "|" entropy, please wait...';
                Str.Subst(Z,"|",entryname);
                video(Z,TRUE);
                ent:=procFile(useLFN,debug,entryname);
                video(Z,FALSE);
                Z:='Entropy is | bits per byte for "|"';
                floattostr(R,commachar,fix,ent);
                Str.Subst(Z,"|",R);
                Str.Subst(Z,"|",entryname);
                WrStr(Z);WrLn;

            END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(entry);
        END;
    END;
    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
END procDir;

(* ------------------------------------------------------------ *)

PROCEDURE doGetCurrent (useLFN:BOOLEAN;drive:SHORTCARD;
                       VAR unit:str2; VAR current:pathtype);
VAR
    rc:CARDINAL;
    longform:pathtype;
BEGIN
    Str.Concat(unit, CHR( ORD("A")-1+ORD(drive) ),colon);

    FIO.GetDir(drive,current); (* \path without u: nor trailing \ except at root *)
    IF current[1] # colon THEN Str.Prepend(current,unit); END; (* safety *)
    IF useLFN THEN
        IF w9XshortToLong(current,rc,longform) THEN (* if error, keep DOS current *)
            Str.Copy(current,longform);
        END;
    END;
    (* LFN function seems to always return "u:\*" form except at root *)
    IF current[1] = colon THEN Str.Delete(current,0,2);END; (* safety *)
    fixDirectory(current);
END doGetCurrent;

PROCEDURE makebase (useLFN:BOOLEAN;spec:pathtype;VAR basepath:pathtype);
VAR
    u,d,n,e,current:pathtype;
    unit:str2;
    drive:SHORTCARD;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Str.Concat(basepath,u,d);
    IF same(basepath,"") THEN
        drive:=FIO.GetDrive(); (* yes we could use 0 as default drive *)
        doGetCurrent(useLFN,drive,  unit,current); (* "u:" and "\" or "\*\" *)
        Str.Concat(basepath, unit,current);
    END;
END makebase;

PROCEDURE chkUD (S:pathtype):BOOLEAN ;
VAR
    u,d,n,e:pathtype;
    pb:CARDINAL;
BEGIN
    qdSplitAllPath(S , u,d,n,e);
    pb:=0;
    IF chkJoker(u) THEN INC(pb);END;
    IF chkJoker(d) THEN INC(pb);END;
    RETURN (pb=0);
END chkUD;

(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 1;
    commacharFR = ",";
    commacharUS = ".";
    defaultcommachar = commacharFR;
VAR
    S,R,spec,basedir:pathtype;
    opt,parmcount,lastparm,i:CARDINAL;
    useLFN,debug:BOOLEAN;
    parm:ARRAY[minparm..maxparm] OF pathtype;
    fix:CARDINAL;
    commachar:CHAR;
    lc:LONGCARD;
BEGIN
    FIO.IOcheck:=FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone;

    WrLn;

    useLFN     := TRUE;
    fix        := defaultfix;
    commachar  := defaultcommachar;
    debug      := FALSE;
    lastparm   := minparm-1;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "F:"+delim+"FIX:"+delim+
                                  "US"+delim+"UK"+delim+
                                  "DEBUG"+delim+"DBG"
                              );
            CASE opt OF
            | 1,2,3    : abort(errHelp,"");
            | 4,5      : useLFN     := FALSE;
            | 6,7      : IF GetLongCard(R,lc)=FALSE THEN abort(errBadNumber,S);END;
                         IF lc < LONGCARD(minfix) THEN abort(errFixRange,S);END;
                         IF lc > LONGCARD(maxfix) THEN abort(errFixRange,S);END;
                         fix := CARDINAL(lc);
            | 8,9      : commachar  := commacharUS;
            | 10,11    : debug      := TRUE;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy( parm[lastparm] , S);
        END;
    END;
    CASE lastparm OF
    | minparm-1 : abort(errHelp,"");
    | maxparm   : ; (* ok *)
    ELSE
        ; (* trapped supra *)
    END;

    useLFN:=( useLFN AND fileSupportLFN() );

    spec := parm[minparm];

    IF chkUD(spec)=FALSE THEN abort(errBadSpec,spec);END;

    IF same(spec,dot) THEN Str.Copy(spec,stardotstar); END;
    IF same(spec,star) THEN Str.Copy(spec,stardotstar); END;
    IF Str.Match(spec,"*"+antislash) THEN Str.Append(spec,stardotstar); END;

    IF chkJoker(spec)=FALSE THEN
        IF fileIsDirectorySpec ( useLFN,spec) THEN
            fixDirectory(spec);
            Str.Append(spec,stardotstar);
        END;
    END;
    IF fileExists(useLFN,spec)=FALSE THEN abort(errNotFound,spec);END;

    makebase(useLFN,spec,basedir);

    procDir(useLFN,debug,commachar,fix,spec,basedir);

    abort(errNone,"");
END fentropy.










