(* ---------------------------------------------------------------
Title         Q&D fix PE executable
Overview      see help
Usage         see help
Notes         A few months ago, while checking a few PE docs
              so I could enhance DLLFIND,
              I had the idea to patch by hand certain "interesting" PE fields
              so patched programs (games, screensavers, utilities...)
              would run on older Vindoze OSes such as Win98SE
              (possibly with the excellent KernelEx libraries) or WinXP.
              Patching versions often did the job, and patching CPU flags
              would sometimes allow a program to run on a newer multicore PC,
              provided patch executable requires a single procesor.

              While I dared think this was rather original an idea,
              I happened to find recently an IMAGECFG.EXE tool
              (seemingly from a M$ Win32 SDK) performing similar tasks. :-(

              I still find my own FIXPE (although more limited in scope)
              is much more user friendly. ;-)

              note KernelEx libraries may be more useful than this program...
              except for single processor setting

              -a display is compatible with DLLFIND /fff display

              port to TS XDS would require carful check for types
Bugs

Wish List     patch "PE plus" Win64 binaries ?
              as I've decided to stop this OS "upgrade" madness with WinXP... ;-)

              list only files matching a version OR a minimum version ?
              bah, just grep log !

              an option to fix header CRC ?

--------------------------------------------------------------- *)

MODULE fixPE;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;

FROM IO IMPORT WrStr, WrLn;

FROM Storage IMPORT ALLOCATE, DEALLOCATE, Available;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XopenFile, w9XcloseFile,
w9XsupportLFN;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    SHOWCULPRIT   = TRUE;
CONST
    progEXEname   = "FIXPE";
    progTitle     = "Q&D fix PE executable";
    progVersion   = "v1.0a";
(*%T SHOWCULPRIT *)
    progCopyright = "by Philippe Guiochon";
(*%E  *)
(*%F SHOWCULPRIT *)
    progCopyright = "by PhG";
(*%E  *)
    banner        = progTitle+" "+progVersion+" "+progCopyright;

CONST
    extCOM        = ".COM";
    sINFO         = "::: ";
    sERR          = "--- ";
    sOK           = "+++ ";
    sPrefix       = "    ";
    sINFOVER      = "=== ";
    sGOODNEWS     = ":+: ";
    sBADNEWS      = ":-: ";
    sNOTABENE     = ":!: ";
CONST
    extLOG        = ".LOG";
    extCD         = ".CD";   (* .RO was better a choice but well, now... *)
    LOGFILE       = progEXEname+extLOG;
    LOGFIXFILE    = progEXEname+extCD;
    specSWP       = "*.SWP";
    specPAR       = "*.PAR";
    specPAGEFILE  = "*\PAGEFILE.SYS"; (* was "?:\PAGEFILE.SYS" *)
    TOOMANY       = MAX(CARDINAL);
CONST
    sBit14on      = "$4000";
    sBit14off     = "$BFFF";
    bit14on       =  04000H; (* 0100000000000000 IN characteristics flag *)
    bit14off      =  0BFFFH; (* 1011111111111111 IN characteristics flag *)
    anyCPU        =  00000H; (* 0000000000000000 IN affinity mask *)



CONST
    errNone            = 0;
    errHelp            = 1;
    errUnknownOption   = 2;
    errTooManyParms    = 3;
    errMissingSpec     = 4;
    errNotFound        = 5;
    errNotFile         = 6;
    errJoker           = 7;
    errCmd             = 8;
    errJokerPath       = 9;
    errTooMany         = 10;
    errMajMin          = 11;
    errRedefined       = 12;
    errCardVal         = 13;
    errMask            = 14;

    errROtarget        = 32;
    errNoWeCant        = 64;
    errBadPEinfos      = 128;
    errUnsupported     = 192;
    errEscape          = 255;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)

errmsg = banner+nl+
nl+
"Syntax : "+progEXEname+" <file> [option]..."+nl+
nl+
'This program shows a few "interesting" values in a PE 32-bits executable.'+nl+
"It can also modify them so patched binary *MAY* run on older Win OSes :"+nl+
"Win98SE (sometimes with the excellent KernelEx libraries) or WinXP."+nl+
'Beware ! Running a patched "newer" binary on an "older" O.S. may *NOT* work.'+nl+

(*
"Just ask your local guru about KernelEX, Unicode, libraries with alternate functions and the like, eh eh... ;-)"
*)

nl+
"  -v          view <file> interesting values (default)"+nl+
"  -c:#        set characteristics flag"+nl+
'  -s          set "single CPU required" ("OR '+sBit14on+'" in characteristics flag)'+nl+
'  -m          set "multiple CPUs allowed" ("AND '+sBit14off+'" in characteristics flag)'+nl+
"  -m:#        set processor affinity mask"+nl+
"  -p:#[,#]... allow specified processor(s) (CPU range is [0..31], *=all)"+nl+
"  -<n|o>:#.#  set NT OS version"+nl+
"  -<w|s>:#.#  set Win32 subsystem version"+nl+
"  -<a|i>:#.#  set application image version"+nl+
"  -v:#        set Win32 version value"+nl+
nl+
"  -a          alternate display (similar to DLLFIND -fff display)"+nl+
"  -n          keep CRLFs to a minimum"+nl+
"  -r          update read-only <file>"+nl+
"  -c          create a new "+LOGFILE+" instead of appending data to existing one"+nl+
"  -l          prevent operations from being reported to "+LOGFILE+" log"+nl+
"  -d          update <file> date/time stamp (default is not to update it) "+nl+
"  -x          disable LFN support even if available"+nl+
nl+
"a) When patching, filename jokers are not supported : this is by design."+nl+
"b) PE+ Win64 executables are not supported : this is by lack of interest."+nl+
'c) Please note "affinity mask" field is not always defined : this is NOT a bug.'+nl+
'd) ".0" is appended to version if specified without ".#" part.'+nl+
'e) Number parsing handles "$#", "0x#", "H#" and "#H" forms for hexadecimal,'+nl+
'   "B#" and "#B" for binary.'+nl+
"f) Modifications are written to "+LOGFILE+" log in current directory :"+nl+
"   this allows easy undoing of any change."+nl+
"   Note -l option can be forced if "+LOGFIXFILE+" exists in executable directory."+nl+
"g) Please note creating a backup of original file is *STRONGLY* advised."+nl+
"h) For obvious reasons, program will not try and analyze files matching"+nl+
"   "+specSWP+", "+specPAR+" and "+specPAGEFILE+" specifications."+nl+
"i) This DOS program was written with J.P.I. TopSpeed Modula-2 compiler."+nl+
nl+
"Examples : "+progEXEname+" calc.exe -single -p:0"+nl+
"           "+progEXEname+" ribbons.scr -n:4.0"+nl+
"           "+progEXEname+" c:\windows\system\*.* /a"+nl;



VAR
    S  : str1024; (* we may get a LFN *)
BEGIN
    CASE e OF
    | errHelp :         WrStr(errmsg);
    | errUnknownOption: S := 'Unknown "|" option !';
    | errTooManyParms:  S := '"|" is one parameter too many !';
    | errMissingSpec:   S := "Missing <file> specification !";
    | errNotFound:      S := '"|" does not exist !';
    | errNotFile:       S := '"|" looks like a directory !';
    | errJoker:         S := 'Jokers are not allowed when patching a file !';
    | errCmd:           S := '"|" option conflicts with an already set command !';
    | errJokerPath:     S := "Jokers are not allowed in path !";
    | errTooMany:       S := "Storage.ALLOCATE() failure !"; (* or 65535 matches ! *)
                        S := "Too many files !";
    | errMajMin:        S := 'Bad "|" version number !';
    | errRedefined:     S := 'Field specified with "|" has already been redefined !';
    | errCardVal:       S := 'Illegal "|" value !';
    | errMask:          S := 'Illegal "|" mask !';

    | errROtarget:      S := "";
    | errNoWeCant:      S := "";
    | errBadPEinfos:    S := "";
    | errUnsupported:   S := "";
    | errEscape:        S := "Aborted by user !";
    ELSE
                        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp:
        ;
    | errUnsupported,errBadPEinfos,errNoWeCant,errROtarget : (* already displayed *)
        ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(progEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ioBufferSize      = (8 * 512) + FIO.BufferOverhead;
    firstioBufferByte = 1;
    lastioBufferByte  = ioBufferSize;
TYPE
    ioBufferType  = ARRAY [firstioBufferByte..lastioBufferByte] OF BYTE;
VAR
    ioBufferList,ioBufferIn : ioBufferType; (* iobufferlist used for ini and list *)

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE initRptLog (SKIPLOG,NEWLOG:BOOLEAN;F:ARRAY OF CHAR);
BEGIN
    IF SKIPLOG THEN RETURN; END;
    IF FIO.Exists(LOGFILE) THEN
        IF isReadOnly(LOGFILE) THEN setReadWrite(LOGFILE); END; (* safety *)
        IF NEWLOG THEN FIO.Erase(LOGFILE); END;
    END;
END initRptLog;

PROCEDURE flushRptLog (SKIPLOG:BOOLEAN;F:ARRAY OF CHAR);
BEGIN
END flushRptLog;

PROCEDURE rptwr (SKIPLOG:BOOLEAN;F, prefix,S:ARRAY OF CHAR);
VAR
    hnd:FIO.File;
BEGIN
    WrStr(prefix);
    WrStr(S);
    WrLn;
    IF SKIPLOG THEN RETURN;END;
    IF FIO.Exists(F) THEN
        hnd:=FIO.Append(F);
    ELSE
        hnd:=FIO.Create(F);
    END;
    FIO.WrStr(hnd,prefix);
    FIO.WrStr(hnd,S);
    FIO.WrLn(hnd);
    FIO.Close(hnd);
END rptwr;

PROCEDURE foundFixFile(ext:ARRAY OF CHAR):BOOLEAN; (* ".???" uppercase *)
VAR
    S:str128;
    p:CARDINAL;
BEGIN
    Lib.ParamStr(S,0); (* retrieve executable location so LOGFILE is EXE.LOG *)
    p:=Str.RCharPos(S,dot); (* always found ! *)
    S[p]:=charnull;
    Str.Append(S,ext);
    RETURN FIO.Exists(S);
END foundFixFile;

(* ------------------------------------------------------------ *)

PROCEDURE chkJokerPath (spec:pathtype  ):BOOLEAN;
VAR
    S,u,d,n,e:pathtype;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Lib.MakeAllPath(S,u,d,"","");
    RETURN chkJoker(S);
END chkJokerPath;

(* Str.Match is not case sensitive *)

PROCEDURE isReservedPattern (S:ARRAY OF CHAR):BOOLEAN;
BEGIN
    IF Str.Match(S,specSWP) THEN RETURN TRUE; END;
    IF Str.Match(S,specPAR) THEN RETURN TRUE; END;
    RETURN Str.Match(S,specPAGEFILE);
END isReservedPattern;

PROCEDURE isReservedEntry (S : ARRAY OF CHAR) : BOOLEAN; (* assume uppercase *)
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    IF same(S,dotdot) THEN RETURN TRUE; END;
    RETURN isReservedPattern(S);
END isReservedEntry;

(* ------------------------------------------------------------ *)

CONST
    firstEntry = 1; (* 1..count *)
TYPE
    ptrToEntry = POINTER TO entryType;
    entryType = RECORD
        next   : ptrToEntry;
        slen   : CARDINAL; (* a SHORTCARD would do fine in almost all cases  *)
        string : CHAR;      (* here, after other data, because variable length *)
    END;

PROCEDURE store (VAR newInList,anchor:ptrToEntry; VAR lastEntry:CARDINAL;
                S:pathtype):BOOLEAN ;
VAR
    len,needed:CARDINAL;
    rc:BOOLEAN;
BEGIN
    INC(lastEntry);
    IF lastEntry=TOOMANY THEN RETURN FALSE;END; (* too many files but fake storage ALLOCATE failure here *)

    len:=Str.Length(S);
    needed:=SIZE(entryType)-SIZE(CHAR)+len;
    rc:= Available(needed);
    IF rc THEN
        CASE lastEntry OF
        | firstEntry :
            ALLOCATE( anchor,needed);
            newInList := anchor;
        ELSE
            ALLOCATE(newInList^.next,needed);
            newInList :=newInList^.next;
        END;
        Lib.FastMove( ADR(S),ADR(newInList^.string),len);
        newInList^.slen := len;
        newInList^.next := NIL;
    END;
    RETURN rc;
END store;

PROCEDURE buildMatchList (VAR anchor:ptrToEntry;
                          useLFN:BOOLEAN;spec:pathtype):CARDINAL;
VAR
    S,u,d,n,e,dirbase,filespec:pathtype; (* some are oversized but safety first ! *)
    newInList : ptrToEntry;
    lastEntry: CARDINAL;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    DOSentry     : FIO.DirEntry;
    rc,found:BOOLEAN;
    dosattr:FIO.FileAttr;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Lib.MakeAllPath(dirbase,u,d,"","");
    fixDirectory(dirbase);
    Lib.MakeAllPath(filespec,"","",n,e);

    lastEntry:=firstEntry-1;
    anchor:=NIL;

    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,DOSentry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(S,w9Xentry.fullfilename);
        ELSE
            Str.Copy(S,DOSentry.Name);
        END;
        IF isReservedEntry(S) THEN (* skip "." ".." "*.SWP" "*.PAR" *)
            ; (* silently ignore this spec *)
        ELSE
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=DOSentry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                Str.Prepend(S,dirbase);
                IF store(newInList,anchor,lastEntry,S)=FALSE THEN
                    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN TOOMANY; (* storage ALLOCATE failure or too many files *)
                END;
            END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(DOSentry);
        END;
    END;
    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
    RETURN lastEntry;
END buildMatchList;

PROCEDURE freeMatchList (anchor:ptrToEntry);
VAR
    len,needed      : CARDINAL;
    firstInList,newInList   : ptrToEntry;
BEGIN
    firstInList := anchor;
    newInList := firstInList;
    WHILE newInList # NIL DO
        len         := CARDINAL(newInList^.slen);
        needed      := SIZE(entryType)-SIZE(CHAR)+len;
        firstInList := firstInList^.next;
        DEALLOCATE (newInList,needed);
        newInList := firstInList;
    END
END freeMatchList;

PROCEDURE getMatchEntry (VAR R:pathtype;
                         n:CARDINAL; anchor:ptrToEntry);
VAR
    i,len:CARDINAL;
    newInList:ptrToEntry;
    S:pathtype;
BEGIN
    newInList := anchor;
    DEC(n); (* trick to force to anchor if 1, and locate correct string if > 1 *)
    FOR i:=firstEntry TO n DO
         newInList := newInList^.next;
    END;
    len         := newInList^.slen;
    Lib.FastMove( ADR(newInList^.string),ADR(S),len);
    S[len]      := nullchar; (* REQUIRED safety ! *)
    Str.Copy(R,S); (* yep, compiler won't let us fill R directly *)
END getMatchEntry;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    dttype = RECORD
        CASE : BOOLEAN OF
        | TRUE  :
            t : CARDINAL; (* hms is low  *)
            d : CARDINAL; (* ymd is high *)
        | FALSE :
            dt : LONGCARD;
        END;
    END;

PROCEDURE getfilestamp (VAR h:dttype  ; useLFN:BOOLEAN;S:pathtype):BOOLEAN;
VAR
    hnd:FIO.File;
    rc:BOOLEAN;
BEGIN
    hnd:=fileOpenRead(useLFN,S);
    IF hnd = MAX(CARDINAL) THEN (* not trapped by fileGetFileStamp() *)
        rc:=FALSE;
    ELSE
        h.dt := FIO.GetFileDate(hnd);
        rc:=TRUE;
    END;
    fileClose(useLFN,hnd);
    RETURN rc;
END getfilestamp;

PROCEDURE setFileStamp (useLFN:BOOLEAN;h:dttype;S:pathtype):BOOLEAN;
VAR
    hnd:FIO.File;
    rc:BOOLEAN;
BEGIN
    hnd:=fileOpen(useLFN,S);
    IF hnd=MAX(CARDINAL) THEN
        rc:=FALSE;
    ELSE
        FIO.SetFileDate(hnd,h.dt);
        rc:=TRUE;
    END;
    fileClose(useLFN,hnd);
    RETURN rc;
END setFileStamp;

PROCEDURE enquote (S:pathtype):pathtype;
VAR
    R: pathtype;
BEGIN
    Str.Concat(R,dquote,S);
    Str.Append(R,dquote);
    RETURN R;
END enquote;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fmtbignum (v:LONGCARD; base:CARDINAL;wi:INTEGER; pad:CHAR ):str16;
VAR
    S:str16;
    ok:BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr( v, S, base,ok);
    FOR i:=Str.Length(S)+1 TO ABS(wi) DO
         IF wi < 0 THEN
             Str.Append(S,pad);
         ELSE
             Str.Prepend(S,pad);
         END;
    END;
    IF base=16 THEN Str.Lows(S);END;
    RETURN S;
END fmtbignum;

PROCEDURE fmtnum ( v:CARDINAL; base:CARDINAL;wi:INTEGER; pad:CHAR ):str16;
BEGIN
    RETURN fmtbignum( LONGCARD(v),base,wi,pad);
END fmtnum;

(* ------------------------------------------------------------ *)

TYPE
    byteArrayType = ARRAY [0..7] OF SHORTCARD; (* 8 bytes should DO FOR now *)

PROCEDURE fmtnummem ( v:CARDINAL):str16;
VAR
    R:str16;
    b:byteArrayType;
    i,n:CARDINAL;
BEGIN
    n:=SIZE(v);
    Lib.Move( ADR(v), ADR(b), n );
    R:="";
    FOR i:=1 TO n DO
        Str.Append(R,fmtnum( CARDINAL(b[i-1]),16,2,"0"));
        IF i < n THEN Str.Append(R," "); END;
    END;
    RETURN R;
END fmtnummem;

PROCEDURE fmtbignummem ( v:LONGCARD):str16;
VAR
    R:str16;
    b:byteArrayType;
    i,n:CARDINAL;
BEGIN
    n:=SIZE(v);
    Lib.Move( ADR(v), ADR(b), n );
    R:="";
    FOR i:=1 TO n DO
        Str.Append(R,fmtnum( CARDINAL(b[i-1]),16,2,"0"));
        IF i < n THEN Str.Append(R," "); END;
    END;
    RETURN R;
END fmtbignummem;

(* ------------------------------------------------------------ *)

PROCEDURE fmtval ( base:CARDINAL; v:LONGCARD;S:ARRAY OF CHAR  ):str128;
VAR
    R:str128;
BEGIN
    IF base=16 THEN
        Str.Concat(R,"$", fmtbignum(v,base,8,"0") );
    ELSE
        Str.Concat(R,"",fmtbignum(v,base,1,"") );
    END;
    RETURN R;
END fmtval;

PROCEDURE fmtaddr ( base:CARDINAL;h:FIO.File; S:ARRAY OF CHAR ):str128;
VAR
    addr:LONGCARD;
BEGIN
    addr:=FIO.GetPos(h);
    RETURN fmtval(base,addr,S);
END fmtaddr;

PROCEDURE fmtver ( major,minor:CARDINAL):str16;
VAR
    R:str16;
BEGIN
    R:="~.~";
    Str.Subst(R,"~",fmtbignum(LONGCARD(major),10,1,""));
    Str.Subst(R,"~",fmtbignum(LONGCARD(minor),10,1,""));
    RETURN R;
END fmtver;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE readStringZ (hin:FIO.File;pos:LONGCARD;VAR R:ARRAY OF CHAR);
VAR
    wanted,got : CARDINAL;
    ch  : CHAR;
BEGIN
    wanted:=SIZE(ch);
    FIO.Seek(hin,pos);
    Str.Copy(R,"");
    LOOP
        got:=FIO.RdBin(hin,ch,wanted);
        IF got # wanted THEN ch:=nullchar;END; (* hide error *)
        Str.Append(R,ch);
        IF ch=nullchar THEN EXIT; END;
    END;
END readStringZ;

PROCEDURE readStringZU (hin:FIO.File;pos:LONGCARD;VAR R:ARRAY OF CHAR);
VAR
    wanted,got : CARDINAL;
    ch  : CHAR;
BEGIN
    wanted:=SIZE(ch);
    FIO.Seek(hin,pos);
    Str.Copy(R,"");
    LOOP
        got:=FIO.RdBin(hin,ch,wanted);
        IF got # wanted THEN ch:=nullchar;END; (* hide error *)
        Str.Append(R,ch);
        IF ch=nullchar THEN EXIT; END;
        got:=FIO.RdBin(hin,ch,wanted); (* $00 *)
        IF got # wanted THEN EXIT;END;
    END;
END readStringZU;

PROCEDURE readString (hin:FIO.File;pos:LONGCARD;len:CARDINAL;VAR R:ARRAY OF CHAR);
VAR
    got : CARDINAL;
BEGIN
    FIO.Seek(hin,pos);
    got:=FIO.RdBin(hin,R,len);
    IF got = len THEN
        R[got]:=nullchar; (* brutal ! *)
    ELSE
        Str.Copy(R,""); (* error *)
    END;
END readString;

PROCEDURE readByte (hin:FIO.File; pos:LONGCARD):BYTE;
VAR
    b:BYTE;
    wanted,got:CARDINAL;
BEGIN
    wanted:=SIZE(b);
    FIO.Seek(hin,pos);
    got:=FIO.RdBin(hin,b,wanted);
    IF got # wanted THEN b:=0;END; (* ignore errors *)
    RETURN b;
END readByte;

PROCEDURE readWord (hin:FIO.File; pos:LONGCARD):WORD;
VAR
    w:WORD;
    wanted,got:CARDINAL;
BEGIN
    wanted:=SIZE(w);
    FIO.Seek(hin,pos);
    got:=FIO.RdBin(hin,w,wanted);
    IF got # wanted THEN w:=0;END; (* ignore errors *)
    RETURN w;
END readWord;

PROCEDURE readDword (hin:FIO.File; pos:LONGCARD):LONGWORD;
VAR
    d:LONGWORD;
    wanted,got:CARDINAL;
BEGIN
    wanted:=SIZE(d);
    FIO.Seek(hin,pos);
    got:=FIO.RdBin(hin,d,wanted);
    IF got # wanted THEN d:=0;END; (* ignore errors *)
    RETURN d;
END readDword;

(* ------------------------------------------------------------ *)

PROCEDURE readCurrStringZU (hin:FIO.File;DEBUG:BOOLEAN; VAR R:ARRAY OF CHAR);
VAR
    wanted,got : CARDINAL;
    ch,unicrap  : CHAR;
BEGIN
IF DEBUG THEN WrStr( fmtval( 16,FIO.GetPos(hin),"readstrZU") );WrLn;END;
    wanted:=SIZE(ch);
    Str.Copy(R,"");
    LOOP
        got:=FIO.RdBin(hin,ch,wanted);
        IF got # wanted THEN ch:=nullchar;END; (* hide error *)
        Str.Append(R,ch);
        got:=FIO.RdBin(hin,unicrap,wanted); (* $00 *)
        IF got # wanted THEN EXIT;END;
        IF ch=nullchar THEN EXIT; END;
    END;
END readCurrStringZU;

PROCEDURE readCurrStringU (hin:FIO.File;DEBUG:BOOLEAN; len:CARDINAL; VAR R:ARRAY OF CHAR);
VAR
    i,wanted,got : CARDINAL;
    ch,unicrap  : CHAR;
BEGIN
IF DEBUG THEN WrStr( fmtval(  16,FIO.GetPos(hin),"readstrU start") );WrLn;END;
    wanted:=SIZE(ch);
    Str.Copy(R,"");
    FOR i:= 1 TO len DO
        got:=FIO.RdBin(hin,ch,wanted);
        IF got # wanted THEN ch:=nullchar;END; (* hide error *)
        Str.Append(R,ch);
        got:=FIO.RdBin(hin,unicrap,wanted); (* $00 *)
        IF got # wanted THEN RETURN;END;
        IF ch=nullchar THEN RETURN; END;
    END;
IF DEBUG THEN WrStr( fmtval(  16,FIO.GetPos(hin),"readstrU end") );WrLn;END;
END readCurrStringU;

PROCEDURE readCurrWord (hin:FIO.File;DEBUG:BOOLEAN):WORD;
VAR
    w:WORD;
    wanted,got:CARDINAL;
BEGIN
IF DEBUG THEN WrStr( fmtval(  16,FIO.GetPos(hin),"readcurrword") );WrLn;END;
    wanted:=SIZE(w);
    got:=FIO.RdBin(hin,w,wanted);
    IF got # wanted THEN w:=0;END; (* ignore errors *)
    RETURN w;
END readCurrWord;

PROCEDURE readCurrByte (hin:FIO.File):BYTE;
VAR
    b:BYTE;
    wanted,got:CARDINAL;
BEGIN
    wanted:=SIZE(b);
    got:=FIO.RdBin(hin,b,wanted);
    IF got # wanted THEN b:=0;END; (* ignore errors *)
    RETURN b;
END readCurrByte;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    PEfileHeaderType = RECORD
        signature            : LONGCARD; (* 50 45 00 00 at $40 *)
        Machine              : CARDINAL;
        NumberOfSections     : CARDINAL; (* useful  *)
        TimeDateStamp        : LONGCARD;
        PointerToSymbolTable : LONGCARD;
        NumberOfSymbols      : LONGCARD;
        SizeOfOptionalHeader : CARDINAL;
        CharacteristicsFlag  : CARDINAL; (* bit 14 = UP_SYSTEM_ONLY = ONE CPU *)
    END;

TYPE
    PEoptionalHeaderType = RECORD
        Magic                       : CARDINAL;  (* +0 PE is $010B, PE+ is $020B *)
        MajorLinkerVersion          : SHORTCARD;
        MinorLinkerVersion          : SHORTCARD;
        SizeOfCode                  : LONGCARD;  (* +4 *)
        SizeOfInitializedData       : LONGCARD;
        SizeOfUninitializedData     : LONGCARD;
        AddressOfEntryPoint         : LONGCARD;  (* +16 *)
        BaseOfCode                  : LONGCARD;
        BaseOfData                  : LONGCARD;
        ImageBase                   : LONGCARD;
        SectionAlignment            : LONGCARD;  (* +32 *)
        FileAlignment               : LONGCARD;
        MajorOperatingSystemVersion : CARDINAL;  (* +40 *)
        MinorOperatingSystemVersion : CARDINAL;
        MajorImageVersion           : CARDINAL;  (* +44 *)
        MinorImageVersion           : CARDINAL;
        MajorSubsystemVersion       : CARDINAL;  (* +48 *)
        MinorSubsystemVersion       : CARDINAL;
        Win32VersionValue           : LONGCARD;  (* +52 *)
        SizeOfImage                 : LONGCARD;
        SizeOfHeaders               : LONGCARD;
        CheckSum                    : LONGCARD;  (* //FIXME algo ? *)
        SubsystemFlag               : CARDINAL;  (* +68 *)
        DllCharacteristics          : CARDINAL;
        SizeOfStackReserve          : LONGCARD;
        SizeOfStackCommit           : LONGCARD;
        SizeOfHeapReserve           : LONGCARD;
        SizeOfHeapCommit            : LONGCARD;
        LoaderFlags                 : LONGCARD;
        NumberOfRvaAndSizes         : LONGCARD;
    END;
TYPE
    PEimageDirectoryType = RECORD
        strucaddr : LONGCARD;
        strucsize : LONGCARD;
    END;

CONST
    (* from WIN32 API headers *)
    IMAGE_DIRECTORY_ENTRY_EXPORT       =  0;   (* Export Directory *)
    IMAGE_DIRECTORY_ENTRY_IMPORT       =  1;   (* Import Directory // *)
    IMAGE_DIRECTORY_ENTRY_RESOURCE     =  2;   (* Resource Directory // *)
    IMAGE_DIRECTORY_ENTRY_EXCEPTION    =  3;   (* Exception Directory *)
    IMAGE_DIRECTORY_ENTRY_SECURITY     =  4;   (* Security Directory *)
    IMAGE_DIRECTORY_ENTRY_BASERELOC    =  5;   (* Base Relocation Table *)
    IMAGE_DIRECTORY_ENTRY_DEBUG        =  6;   (* Debug Directory *)
    IMAGE_DIRECTORY_ENTRY_COPYRIGHT    =  7;   (* Description String *)
    IMAGE_DIRECTORY_ENTRY_GLOBALPTR    =  8;   (* Machine Value (MIPS GP) *)
    IMAGE_DIRECTORY_ENTRY_TLS          =  9;   (* TLS Directory *)
    IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG  = 10;   (* Load Configuration Directory // *)
    IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT = 11;   (* Bound Import Directory in headers *)
    IMAGE_DIRECTORY_ENTRY_IAT          = 12;   (* Import Address Table *)
    IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT   = 13; (* Delay Load Import Descriptors *)
    IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR = 14; (* COM Runtime descriptor *)

CONST
    (* image directory entries *)
    firstdirindex   = 0;
    lastdirindex    = 15; (* not 14 ! *)
    IMPORTindex     = IMAGE_DIRECTORY_ENTRY_IMPORT;
    RESOURCEindex   = IMAGE_DIRECTORY_ENTRY_RESOURCE;
    LOADCONFIGindex = IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG;
TYPE
    PEdataDirectoriesType = ARRAY [firstdirindex..lastdirindex] OF PEimageDirectoryType;

TYPE
    PEsectionHeaderType = RECORD
        Name                 : ARRAY [0..7] OF CHAR; (* 0 terminated or 8 chars *)
        VirtualSize          : LONGCARD;
        VirtualAddress       : LONGCARD;
        SizeOfRawData        : LONGCARD;
        PointerToRawData     : LONGCARD; (* file offset *)
        PointerToRelocations : LONGCARD;
        PointerToLinenumbers : LONGCARD;
        NumberOfRelocations  : CARDINAL;
        NumberOfLinenumbers  : CARDINAL;
        Characteristics      : LONGCARD;
    END;

TYPE
    PEimageImportDescriptorType = RECORD
        OriginalFirstThunk : LONGCARD;
        TimeDateStamp      : LONGCARD;
        ForwarderChain     : LONGCARD;
        Name               : LONGCARD; (* // ASCIIZ *)
        FirstThunk         : LONGCARD;
    END;

(*
//FIXME search doc about this !
tiny C compiler WINNT.H file
discrepancies with digitalmars C compiler WINNT.H and PEFILE.PY code !
we should try AND locate a reliable source FOR this damn PE format !
*)

TYPE
    PEloadConfig32type = RECORD
        Size                          :  LONGCARD; (* +0 *)
        TimeDateStamp                 :  LONGCARD;
        MajorVersion                  :  CARDINAL; (* +8 *)
        MinorVersion                  :  CARDINAL;
        GlobalFlagsClear              :  LONGCARD; (* +12 *)
        GlobalFlagsSet                :  LONGCARD;
        CriticalSectionDefaultTimeout :  LONGCARD;
        DeCommitFreeBlockThreshold    :  LONGCARD;
        DeCommitTotalFreeThreshold    :  LONGCARD;
        LockPrefixTable               :  LONGCARD;
        MaximumAllocationSize         :  LONGCARD;
        VirtualMemoryThreshold        :  LONGCARD; (* +40 *)
        ProcessHeapFlags              :  LONGCARD;
        ProcessAffinityMask           :  LONGCARD; (* +48 //FIXME check *)
        CSDVersion                    :  CARDINAL;
        Reserved1                     :  CARDINAL;
        EditList                      :  LONGCARD;
        SecurityCookie                :  LONGCARD;
        SEHandlerTable                :  LONGCARD;
        SEHandlerCount                :  LONGCARD;
        (* here PEFILE.PY *)
        GuardCFCheckFunctionPointer   : LONGCARD;
        Reserved2                     : LONGCARD;
        GuardCFFunctionTable          : LONGCARD;
        GuardCFFunctionCoun           : LONGCARD;
        GuardFlags                    : LONGCARD;
    END;

(* ------------------------------------------------------------ *)

(* TS kludge as not to use additions of SIZE() *)

CONST  (* PEfileHeaderType *)
    ofsCharacteristicsFlag         = 22; (* bit 14 UP_SYSTEM_ONLY ONE CPU *)
CONST  (* PEoptionalHeaderType *)
    ofsMajorOperatingSystemVersion = 40;
    ofsMajorImageVersion           = 44;
    ofsMajorSubsystemVersion       = 48;
    ofsWin32VersionValue           = 52;
    ofsSubsystemFlag               = 68;
CONST  (* PEloadConfig32type *)
    ofsProcessorAffinityMask       = 48;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    idNotEXE          = 1; (* possibly COM format *)
    idNotPackedDOSexe = 2; (* almost certainly a DOS executable *)
    idUnknownWinEXE   = 3; (* almost certainly a packed DOS executable *)
    idNE              = 4;
    idLE              = 5;
    idW3              = 6;
    idW4              = 7;
    idPE              = 8;   (* supported *)
    idCOM             = 9;
    idLZ090           = 10;
    idLZ091           = 11;
    idPKLITE          = 12;
    idLHASFX          = 13;
    idTScrunch        = 14;
    idUPXdos          = 15;
    idUPXwin32        = 16;  (* supported *)
    idPEplus          = 17;
    idDIET            = 18;
    idGloups          = 19;

PROCEDURE getdesc (id:CARDINAL; VAR R:ARRAY OF CHAR);
VAR
    S:str128;
BEGIN
    CASE id OF
    | idNotEXE :         S:="Not an executable";
    | idNotPackedDOSexe: S:="DOS executable (probably not packed)";
    | idUnknownWinEXE:   S:="DOS executable (probably packed)";
    | idNE:              S:="NE Win16 executable";
    | idLE:              S:="LE Win16 VxD";
    | idW3:              S:="W3 Win16 collection of VxDs";
    | idW4:              S:="W4 Win9x VxD";
    | idPE:              S:="PE Win32 executable";
    | idCOM:             S:="DOS program (COM format)"; (* assumption based upon file extension ! *)
    | idLZ090:           S:="DOS executable (packed with LZEXE 0.90)";
    | idLZ091:           S:="DOS executable (packed with LZEXE 0.91)";
    | idPKLITE:          S:="DOS executable (packed with PKLITE)";
    | idLHASFX:          S:="DOS executable (LHA 2.x self-extracting archive)";
    | idTScrunch:        S:="DOS executable (packed with TopSpeed CRUNCH)";
    | idUPXdos:          S:="DOS executable (packed with UPX)";
    | idUPXwin32:        S:="PE Win32 executable (packed with UPX)";
    | idPEplus:          S:="PE+ Win64 executable";
    | idDIET:            S:="DOS executable (packed with DIET)";
    ELSE
                         S:="Unknown";
    END;
    Str.Copy(R,S);
END getdesc;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
values for IMAGE_FILE_HEADER.Machine field according to M$ PE/COFF dox :

UNKNOWN      0x0        applicable to any machine type
AM33         0x1d3      Matsushita AM33
AMD64        0x8664     x64                                     ///
ARM          0x1c0      ARM little endian
EBC          0xebc      EFI byte code
I386         0x14c      Intel 386 or later                      ///
IA64         0x200      Intel Itanium processor family          /// ???
M32R         0x9041     Mitsubishi M32R little endian
MIPS16       0x266      MIPS16
MIPSFPU      0x366      MIPS with FPU
MIPSFPU16    0x466      MIPS16 with FPU
POWERPC      0x1f0      Power PC little endian
POWERPCFP    0x1f1      Power PC with floating point support
R4000        0x166      MIPS little endian
SH3          0x1a2      Hitachi SH3
SH3DSP       0x1a3      Hitachi SH3 DSP
SH4          0x1a6      Hitachi SH4
SH5          0x1a8      Hitachi SH5
THUMB        0x1c2      Thumb
WCEMIPSV2    0x169      MIPS little-endian WCE v2

values for "PE architecture" i.e. Magic field :
UNKNOWN      0x0000
ANYCPU       0x0001
X86          0x010B
x64          0x020B
*)

PROCEDURE chkUPXwin32 (VAR seems64bits:BOOLEAN;
                       hin:FIO.File;winheaderpos:LONGCARD;DEBUG:BOOLEAN):BOOLEAN;
CONST
    strUPX          = "UPX"; (* normally, UPX0, UPX1, UPX2 *)
CONST
    magicUnknown    = 0000H;
    magicAnyCPU     = 0001H;
    magic32         = 010BH;
    magic64         = 020BH;
CONST
    machineX86      = 014CH;
    machineX64      = 8664H;
    machineI64      = 0200H;
VAR
    rc:BOOLEAN;
    pefileheader     : PEfileHeaderType;
    peoptionalheader : PEoptionalHeaderType; (* required "optional" : go figure ! *)
    pedirindex       : PEdataDirectoriesType;
    pesectionheader  : PEsectionHeaderType;
    i,got,wanted,sectioncount:CARDINAL;
    filealignment,sectionalignment,importaddr : LONGCARD;
    sectionheadername : str16; (* oversized *)
    chk64a,chk64b:BOOLEAN;
BEGIN
    seems64bits := FALSE; (* assume 32 bits PE *)
    rc:=FALSE; (* assume NOT an UPX Win32 *)

    FIO.Seek(hin,winheaderpos);

    wanted:=SIZE(pefileheader);
    got := FIO.RdBin(hin,pefileheader,wanted);
    IF got # wanted THEN RETURN rc; END;
    sectioncount:=pefileheader.NumberOfSections;
    IF sectioncount=0 THEN RETURN rc; END;

IF DEBUG THEN WrStr(fmtval( 16,LONGCARD( pefileheader.Machine) ,"machine"));WrLn;END;

    CASE pefileheader.Machine OF
    | machineX86 : chk64a := FALSE;
    | machineX64 : chk64a := TRUE;
    | machineI64 : chk64a := TRUE;
    ELSE
                   chk64a := FALSE;
    END;

    wanted:=SIZE(peoptionalheader);
    got := FIO.RdBin(hin,peoptionalheader,wanted);
    IF got # wanted THEN RETURN rc; END;

IF DEBUG THEN WrStr(fmtval( 16,LONGCARD( peoptionalheader.Magic) ,"magic"));WrLn;END;

    CASE peoptionalheader.Magic OF
    | magic32 : chk64b := FALSE;
    | magic64 : chk64b := TRUE;
    ELSE
                chk64b := FALSE;
    END;

    seems64bits := (chk64a AND chk64b); (* better safe than sorry *)

    filealignment := peoptionalheader.FileAlignment;
    sectionalignment:=peoptionalheader.SectionAlignment;

    wanted:=SIZE(pedirindex);
    got := FIO.RdBin(hin,pedirindex,wanted);
    IF got # wanted THEN RETURN rc; END;

    importaddr:=pedirindex[IMPORTindex].strucaddr;

    i:=0;
    LOOP
        IF i >= sectioncount THEN EXIT; END; (* = is enough but... *)
        wanted:=SIZE(pesectionheader);
        got := FIO.RdBin(hin,pesectionheader,wanted);
        IF got # wanted THEN RETURN rc; END; (* file image problem ! abort ! *)

        (* kludge *)
        Str.Copy(sectionheadername, pesectionheader.Name);
        IF Str.Pos(sectionheadername,strUPX) = 0 THEN rc:=TRUE; EXIT; END; (* assume "UPX#" *)

        IF pesectionheader.VirtualAddress<=importaddr THEN
            IF importaddr < (pesectionheader.VirtualAddress+pesectionheader.SizeOfRawData) THEN
                EXIT;
            END;
        END;
        INC(i);
    END;
    RETURN rc; (* true only if one strUPX was found in section header name *)
END chkUPXwin32;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE getIDfileType (VAR winheaderpos : LONGCARD ;
                        hin:FIO.File;DEBUG:BOOLEAN;zefile:pathtype):CARDINAL;
CONST
    sigMZ = ( ORD("M") + (ORD("Z") << 8) );
    sigZM = ( ORD("Z") + (ORD("M") << 8) );
    sigNE = ( ORD("N") + (ORD("E") << 8) );
    sigLE = ( ORD("L") + (ORD("E") << 8) );
    sigW3 = ( ORD("W") + (ORD("3") << 8) );
    sigW4 = ( ORD("W") + (ORD("4") << 8) );
    sigPE = ( ORD("P") + (ORD("E") << 8) );
CONST
    jpimagic1 = 0001H;
    jpimagic2 = 018AH;
    jpimagic3 = 1565H;
VAR
    tsmagic:ARRAY[1..3] OF CARDINAL;
    id,magic,addr:CARDINAL;
    R:str16;
    seems64bits:BOOLEAN;
BEGIN

    magic:=readWord(hin,000H);
    CASE magic OF
    | sigMZ,sigZM:
        addr:=readWord(hin,018H); (* //FIXHARD *)
        IF addr < 040H THEN       (* //FIXHARD *)
            readString (hin, 01CH, 4, R);                 (* //FIXHARD *)
            IF same(R,"LZ09") THEN RETURN idLZ090; END;
            IF same(R,"LZ91") THEN RETURN idLZ091; END;
            IF same(R,"diet") THEN RETURN idDIET;END;   (* assume lowercase signature *)
            readString(hin, 01EH, 6, R);                  (* //FIXHARD *)
            IF same(R,"PKLITE") THEN RETURN idPKLITE; END;
            readString(hin, 024H, 10, R); Str.Caps(R);    (* //FIXHARD *)
            IF same(R,"LHA'S SFX ") THEN RETURN idLHASFX; END;
            readString(hin, 055H,  3, R);                 (* //FIXHARD *)
            IF same(R,"UPX") THEN RETURN idUPXdos; END;

            id:=idNotPackedDOSexe; (* default *)

            tsmagic[1]:=readWord(hin, 01CH);               (* //FIXHARD *)
            tsmagic[2]:=readWord(hin, 01EH);               (* //FIXHARD *)
            tsmagic[3]:=readWord(hin, 020H);               (* //FIXHARD *)
            IF tsmagic[1] = jpimagic1 THEN
                IF tsmagic[2] = jpimagic2 THEN
                    IF tsmagic[3] = jpimagic3 THEN id:=idTScrunch; END;
                END;
            END;
        ELSE
            winheaderpos:=readDword(hin,03CH); (* //FIXHARD *)
            IF winheaderpos = 0 THEN (* unlikely *)
                id:=idNotPackedDOSexe;
            ELSE
                magic:=readWord(hin,winheaderpos);
                CASE magic OF
                | sigNE: id:=idNE; (* NE Windows or OS/2 1.x segmented ("new") executable *)
                | sigLE: id:=idLE; (* LE Windows virtual device driver (VxD) linear executable *)
                | sigW3: id:=idW3; (* W3 Windows WIN386.EXE file; a collection of LE files *)
                | sigW4: id:=idW4; (* W4 Windows95 VMM32.VXD file *)
                | sigPE:
                         IF chkUPXwin32(seems64bits, hin,winheaderpos,DEBUG) THEN
                             id:=idUPXwin32;
                         ELSE
                             IF seems64bits THEN
                                 id:=idPEplus;
                             ELSE
                                 id:=idPE; (* PE Win32 (Windows NT and Win32s) portable executable based on Unix COFF *)
                             END;
                         END;
                ELSE
                    (* a few DOS packers can have addr >= 40H, so we must test them here too *)
                    readString(hin, 01EH, 6, R);   (* //FIXHARD *)
                    IF same(R,"PKLITE") THEN RETURN idPKLITE; END;
                    (* check upx just in case *)
                    readString(hin, 055H,  3, R);  (* //FIXHARD *)
                    IF same(R,"UPX") THEN RETURN idUPXdos; END;
                    id:=idUnknownWinEXE; (* default *)
                    tsmagic[1]:=readWord(hin, 01CH);    (* //FIXHARD *)
                    tsmagic[2]:=readWord(hin, 01EH);    (* //FIXHARD *)
                    tsmagic[3]:=readWord(hin, 020H);    (* //FIXHARD *)
                    IF tsmagic[1] = jpimagic1 THEN
                        IF tsmagic[2] = jpimagic2 THEN
                            IF tsmagic[3] = jpimagic3 THEN id:=idTScrunch; END;
                        END;
                    END;
                END;
            END;
        END;
    ELSE
        id:=idNotEXE;

    END;

    IF id = idNotEXE THEN (* further check needed *)
        (* assume file extension is correct ! we could check first byte is a JMP too *)
        IF Str.Match(zefile,star+extCOM) THEN (* case-insensitive *)
            id:=idCOM;
        END;
    END;

    RETURN id;
END getIDfileType;

PROCEDURE getFileType (VAR winheaderpos : LONGCARD ;
                       useLFN,DEBUG:BOOLEAN;zefile:pathtype):CARDINAL;
VAR
    hin:FIO.File;
    id:CARDINAL;
BEGIN
    hin:=fileOpenRead(useLFN,zefile);
    FIO.AssignBuffer(hin,ioBufferIn);

    id:=getIDfileType (winheaderpos,  hin,DEBUG,zefile) ;

    fileClose(useLFN,hin);
    RETURN id;
END getFileType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    aboutType = (isWORD,isDWORD,isMajMin);
    idType = (idCharacteristicsFlag,
              idProcessorAffinityMask,
              idMajorOperatingSystemVersion,
              idMajorSubsystemVersion,
              idMajorImageVersion,
              idWin32VersionValue);
CONST
    firstPEinfo = ORD(idCharacteristicsFlag);
    lastPEinfo  = ORD(idWin32VersionValue);

CONST
    idString = "characteristics flag,"+
               "processor affinity mask,"+
               "NT OS version,"+
               "Win32 subsystem version,"+
               "application image version,"+
               "Win32 version value";

TYPE
    valueType = RECORD
        CASE : BOOLEAN OF
        | TRUE : hilo  : LONGCARD;
        | FALSE: lo,hi : CARDINAL;
        END;
    END;

TYPE PEinfoType = RECORD
     fpos              : LONGCARD;
     value             : valueType;
     isOptional        : BOOLEAN;
     isPresent         : BOOLEAN;    (* isDefined : affinity mask may NOT exist *)
     about             : aboutType;
     idstr             : str80; (* should do FOR name *)
END;

TYPE
    PEinfostruc = ARRAY[firstPEinfo..lastPEinfo] OF PEinfoType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
     MISSINGDIRENTRY       = LONGCARD(0);
     HIDDENAFFINITYPROBLEM = 0FFFFFFFFH;
     UNDEFINED             = MAX(LONGCARD);

PROCEDURE zeroPEinfos ( defined:BOOLEAN; VAR PEinfo : PEinfostruc );
VAR
    i:CARDINAL;
    S:str80;
BEGIN
    FOR i:=firstPEinfo TO lastPEinfo DO (* default safety *)
        PEinfo[i].fpos              := UNDEFINED;
        PEinfo[i].value.hilo        := UNDEFINED;
        IF i = ORD(idProcessorAffinityMask) THEN
            PEinfo[i].isOptional        := TRUE;
        ELSE
            PEinfo[i].isOptional        := FALSE;
        END;
        PEinfo[i].isPresent         := defined; (* may be FALSE FOR affinitymask *)
        CASE i OF
        | ORD(idCharacteristicsFlag) :
            PEinfo[i].about             := isWORD;
        | ORD(idProcessorAffinityMask), ORD(idWin32VersionValue):
            PEinfo[i].about             := isDWORD;
        ELSE
            PEinfo[i].about             := isMajMin;
        END;
        isoleItemS(S, idString,coma,i); (* 0.. *)
        PEinfo[i].idstr             :=S;
    END;
END zeroPEinfos;


PROCEDURE grabPEdata (VAR PEinfo:PEinfostruc ;
                      winheaderpos:LONGCARD;hin:FIO.File; DEBUG:BOOLEAN):BOOLEAN;
VAR
    i,got,wanted     : CARDINAL;
    sectioncount     : CARDINAL;
    pefileheader     : PEfileHeaderType;
    peoptionalheader : PEoptionalHeaderType; (* required "optional" : go figure ! *)
    pedirindex       : PEdataDirectoriesType;
    pesectionheader  : PEsectionHeaderType;
    peloadconfig     : PEloadConfig32type;
    posloadconfig,sizeloadconfig,importaddr,posheader : LONGCARD;
    rc,missing       : BOOLEAN;
BEGIN
    rc:=FALSE;

    FIO.Seek(hin,winheaderpos);

    wanted:=SIZE(pefileheader);
    got := FIO.RdBin(hin,pefileheader,wanted);
    IF got # wanted THEN RETURN rc; END;
    sectioncount:=pefileheader.NumberOfSections;
    IF sectioncount=0 THEN RETURN rc; END;

    posheader:=FIO.GetPos(hin);

    wanted:=SIZE(peoptionalheader);
    got := FIO.RdBin(hin,peoptionalheader,wanted);
    IF got # wanted THEN RETURN rc; END;

    wanted:=SIZE(pedirindex);
    got := FIO.RdBin(hin,pedirindex,wanted);
    IF got # wanted THEN RETURN rc; END;

    (* may be 0 i.e. undefined !!! *)
    importaddr    := pedirindex[IMPORTindex].strucaddr;
    posloadconfig := pedirindex[LOADCONFIGindex].strucaddr; (* seems optional ! *)
    sizeloadconfig:= pedirindex[LOADCONFIGindex].strucsize;

    i:=0;
    LOOP
        IF i >= sectioncount THEN EXIT; END; (* = is enough but... *)
        wanted:=SIZE(pesectionheader);
        got := FIO.RdBin(hin,pesectionheader,wanted);
        IF got # wanted THEN RETURN rc; END; (* file image problem ! abort ! *)
        IF pesectionheader.VirtualAddress<=importaddr THEN
            IF importaddr < (pesectionheader.VirtualAddress+pesectionheader.SizeOfRawData) THEN
                (* WrStr(".rdata ! TAKE ME !");WrLn; *)
                EXIT;
            END;
        END;
        INC(i);
    END;
    IF i >= sectioncount THEN RETURN rc; END; (* abort because problem ! *)

    (* both 0 means this entry does NOT exist *)
    missing := ( (posloadconfig = MISSINGDIRENTRY) AND (sizeloadconfig = MISSINGDIRENTRY) );
    IF missing THEN
        peloadconfig.ProcessAffinityMask := HIDDENAFFINITYPROBLEM;
    ELSE
        INC(posloadconfig,pesectionheader.PointerToRawData);
        DEC(posloadconfig,pesectionheader.VirtualAddress);
        FIO.Seek(hin,posloadconfig);
        wanted:=SIZE(peloadconfig);
        got := FIO.RdBin(hin,peloadconfig,wanted);
        IF got # wanted THEN RETURN rc;END;
    END;

    zeroPEinfos(TRUE, PEinfo);

    i:=firstPEinfo;
    PEinfo[i].fpos              := winheaderpos+ofsCharacteristicsFlag; (* //FIXME hardcoded *)
    PEinfo[i].value.lo          := pefileheader.CharacteristicsFlag;
    PEinfo[i].value.hi          := 0;

    INC(i);
    IF peloadconfig.ProcessAffinityMask = HIDDENAFFINITYPROBLEM THEN
        PEinfo[i].fpos              := UNDEFINED;
        PEinfo[i].value.hilo        := UNDEFINED;
        PEinfo[i].isPresent         := FALSE;
    ELSE
        PEinfo[i].fpos              := posloadconfig+ofsProcessorAffinityMask; (* //FIXME hardcoded *)
        PEinfo[i].value.hilo        := peloadconfig.ProcessAffinityMask;
        PEinfo[i].isPresent         := TRUE; (* useless *)
    END;

    INC(i);
    PEinfo[i].fpos              := posheader+ofsMajorOperatingSystemVersion; (* //FIXME hardcoded *)
    PEinfo[i].value.hi          := peoptionalheader.MajorOperatingSystemVersion;
    PEinfo[i].value.lo          := peoptionalheader.MinorOperatingSystemVersion;

    INC(i);
    PEinfo[i].fpos              := posheader+ofsMajorSubsystemVersion; (* //FIXME hardcoded *)
    PEinfo[i].value.hi          := peoptionalheader.MajorSubsystemVersion;
    PEinfo[i].value.lo          := peoptionalheader.MinorSubsystemVersion;

    INC(i);
    PEinfo[i].fpos              := posheader+ofsMajorImageVersion; (* //FIXME hardcoded *)
    PEinfo[i].value.hi          := peoptionalheader.MajorImageVersion;
    PEinfo[i].value.lo          := peoptionalheader.MinorImageVersion;

    INC(i);
    PEinfo[i].fpos              := posheader+ofsWin32VersionValue; (* //FIXME hardcoded *)
    PEinfo[i].value.hilo        := peoptionalheader.Win32VersionValue;

    RETURN TRUE; (* ok *)
END grabPEdata;

PROCEDURE grabPEinfos (VAR PEinfo : PEinfostruc;
                       winheaderpos : LONGCARD ;
                       useLFN,DEBUG:BOOLEAN;zefile:pathtype):BOOLEAN;
VAR
    hin:FIO.File;
    rc:BOOLEAN;
BEGIN
    hin:=fileOpenRead(useLFN,zefile);
    FIO.AssignBuffer(hin,ioBufferIn);

    rc:=grabPEdata(PEinfo,  winheaderpos,hin,DEBUG) ;

    fileClose(useLFN,hin);
    RETURN rc;
END grabPEinfos;

(* ------------------------------------------------------------ *)

PROCEDURE dmpPEinfos (SKIPLOG,altdisplay:BOOLEAN;PEinfo:PEinfostruc  );
CONST
    sPadWord            = "     "; (* "?? ??" *)
    sW                  = "@ : ~ "+sPadWord+" = $# (!)";
    sDW                 = "@ : ~ = $# (!)";
    sVer                = "@ : ~ ~ = # (!)";
    sCharacteristicsAlt = "@ : ~ "+sPadWord+" = $# (! - |)";
CONST
    sSepar   = ", ";
    lenSepar = 2;
CONST
    sMultipleCPUsOKalt      = "multiple CPUs OK";
    sSingleCPUrequiredAlt   = "single CPU required";
    sMultipleCPUsOK         = "Multiple CPUs are supported.";
    sSingleCPUrequired      = "A single CPU is required.";
    sUndefinedAffinityMask  = "Affinity mask is undefined.";
    sAnyCPU                 = "All core CPUs are supported.";
    sCPUlistSingular        = "Supported core CPU is ";
    sCPUlistPlural          = "Supported core CPUs are ";
VAR
    i,count,lo,hi:CARDINAL;
    mask,fpos,hilo:LONGCARD;
    about:aboutType;
    undefinedAffinityMask,isPresent,isOptional:BOOLEAN;
    S:str128;
    Z,idstr:str80;
    processorAffinityMask : LONGCARD;
    characteristicsFlag   : CARDINAL;
BEGIN
    FOR i:= firstPEinfo TO lastPEinfo DO
        fpos              := PEinfo[i].fpos;
        hi                := PEinfo[i].value.hi;
        lo                := PEinfo[i].value.lo;
        hilo              := PEinfo[i].value.hilo;
        about             := PEinfo[i].about;
        idstr             := PEinfo[i].idstr;
        isPresent         := PEinfo[i].isPresent;
        isOptional        := PEinfo[i].isOptional;

        CASE i OF
        | ORD(idCharacteristicsFlag)   :
            characteristicsFlag  :=lo;
        | ORD(idProcessorAffinityMask) :
            processorAffinityMask:=hilo;
            undefinedAffinityMask:=NOT(isPresent);
            IF undefinedAffinityMask THEN
                S:=sDW;
                Str.Subst(S,"@","????????");
                Str.Subst(S,"~","?? ?? ?? ??");
                Str.Subst(S,"#","????????");
                Str.Subst(S,"!","undefined !"); (* trick *)
                Str.Subst(S,"!",idstr);
            END;
        END;

        IF isPresent THEN
            CASE about OF
            | isWORD:
                S:=sW;
                Str.Subst(S,"@",fmtbignum(fpos,16,8,"0"));
                Str.Subst(S,"~",fmtnummem(lo));
                Str.Subst(S,"#",fmtnum(lo,16,4,"0"));
            | isDWORD:
                S:=sDW;
                Str.Subst(S,"@",fmtbignum(fpos,16,8,"0"));
                Str.Subst(S,"~",fmtbignummem(hilo));
                Str.Subst(S,"#",fmtbignum(hilo,16,8,"0"));
            | isMajMin:
                S:=sVer;
                Str.Subst(S,"@",fmtbignum(fpos,16,8,"0"));
                Str.Subst(S,"~",fmtnummem(hi)); (* reversed here : major first *)
                Str.Subst(S,"~",fmtnummem(lo));
                Str.Subst(S,"#",fmtver(hi,lo));
            END;
            Str.Subst(S,"!",idstr);

            IF altdisplay THEN
                IF i = ORD(idCharacteristicsFlag) THEN
                    S:=sCharacteristicsAlt;
                    Str.Subst(S,"@",fmtbignum(fpos,16,8,"0"));
                    Str.Subst(S,"~",fmtnummem(lo));
                    Str.Subst(S,"#",fmtnum(lo,16,4,"0"));
                    Str.Subst(S,"!",idstr);
                    IF ( characteristicsFlag AND bit14on ) = bit14on THEN
                        Str.Subst(S,"|",sSingleCPUrequiredAlt);
                    ELSE
                        Str.Subst(S,"|",sMultipleCPUsOKalt);
                    END;
                END;
            END;
        END;
        rptwr(SKIPLOG,LOGFILE,sINFOVER,S);
    END;

    IF altdisplay THEN RETURN; END;

    (* about characteristics bit 14 *)

    IF (characteristicsFlag AND bit14on) = bit14on THEN
        Z:=sNOTABENE; S:=sSingleCPUrequired;
    ELSE
        Z:=sGOODNEWS; S:=sMultipleCPUsOK;
    END;
    rptwr(SKIPLOG,LOGFILE, Z,S);


    (* about core CPUs *)

    IF undefinedAffinityMask THEN
        Z:=sBADNEWS; S:=sUndefinedAffinityMask;
    ELSE
        IF processorAffinityMask = anyCPU THEN
            Z:=sGOODNEWS; S:=sAnyCPU;
        ELSE
            count:=0;
            mask:=01H;
            FOR i:=0 TO 15 DO
                IF (processorAffinityMask AND mask)=mask THEN INC(count);END;
                mask:=(mask << 1);
            END;
            CASE count OF
            | 0 : Z:=sGOODNEWS; S:=sAnyCPU;
            | 1 : Z:=sNOTABENE; S:=sCPUlistSingular;
            ELSE
                  Z:=sNOTABENE; S:=sCPUlistPlural;
            END;
            mask:=01H;
            FOR i:=0 TO 15 DO
                IF (processorAffinityMask AND mask)=mask THEN
                    Str.Append(S,fmtnum(i,10,1,""));Str.Append(S,sSepar);
                END;
                mask:=(mask << 1);
            END;
            IF count # 0 THEN
                i:=Str.Length(sSepar);
                Str.Delete(S,Str.Length(S)-i,i);
                Str.Append(S,dot);
                i:=Str.RCharPos(S,coma);
                IF i # MAX(CARDINAL) THEN
                    Str.Delete(S,i,1);
                    Str.Insert(S," and",i);
                END;
            END;
        END;
    END;
    rptwr(SKIPLOG,LOGFILE,Z,S);
END dmpPEinfos;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    flagVALUE = 0;
    flagOR    = MAX(CARDINAL);
    flagAND   = MAX(CARDINAL)-1;

PROCEDURE createPatch (VAR newPEinfo:PEinfostruc;
                      refPEinfo:PEinfostruc  ):BOOLEAN;
VAR
    i,v:CARDINAL;
BEGIN
    i:=ORD(idProcessorAffinityMask);
    IF newPEinfo[i].isPresent THEN
        IF NOT(refPEinfo[i].isPresent) THEN RETURN FALSE;END;
    END;

    FOR i:=firstPEinfo TO lastPEinfo DO
        IF newPEinfo[i].isPresent THEN (* modified *)
            CASE i OF (* could use .about too *)
            | ORD(idCharacteristicsFlag) :
                CASE newPEinfo[i].value.hi OF
                | flagVALUE : v:= newPEinfo[i].value.lo;
                | flagOR    : v:= (refPEinfo[i].value.lo OR bit14on);
                | flagAND   : v:= (refPEinfo[i].value.lo AND bit14off);
                END;
                refPEinfo[i].value.lo := v;
            | ORD(idProcessorAffinityMask), ORD(idWin32VersionValue):
                refPEinfo[i].value.hilo:=newPEinfo[i].value.hilo;
            ELSE
                refPEinfo[i].value.hi:=newPEinfo[i].value.hi;
                refPEinfo[i].value.lo:=newPEinfo[i].value.lo;
            END;
        END;
    END;
    (* we could also check for no change i.e. useless patch but well, who cares ? *)
    newPEinfo := refPEinfo;
    RETURN TRUE;
END createPatch;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* assume F is rw *)

PROCEDURE doPatch (useLFN:BOOLEAN;F:pathtype;PEinfo:PEinfostruc );
VAR
    i : CARDINAL;
    hnd:FIO.File;
    hilo:LONGCARD;
    lo,hi,count:CARDINAL;
    b : byteArrayType;
BEGIN
    hnd:=fileOpen(useLFN,F);
    FIO.AssignBuffer(hnd,ioBufferIn);

    FOR i:=firstPEinfo TO lastPEinfo DO
        IF PEinfo[i].isPresent THEN (* should do for optional affinity flag *)
            CASE i OF
            | ORD(idCharacteristicsFlag) :
                lo:=PEinfo[i].value.lo;
                count:=SIZE(lo);
                Lib.Move(ADR(lo),ADR(b),count);
            | ORD(idWin32VersionValue) :
                hilo:=PEinfo[i].value.hilo;
                count:=SIZE(hilo);
                Lib.Move(ADR(hilo),ADR(b),count);
            | ORD(idProcessorAffinityMask) :
                hilo:=PEinfo[i].value.hilo;
                count:=SIZE(hilo);
                Lib.Move(ADR(hilo),ADR(b),count);
            ELSE
                hi:=PEinfo[i].value.hi; (* major first *)
                count:=SIZE(hi);
                Lib.Move(ADR(hi),ADR(b),count);
                lo:=PEinfo[i].value.lo;
                count:=SIZE(lo);
                Lib.Move(ADR(lo),ADR(b[SIZE(lo)]),count);
                count:=SIZE(hi)+SIZE(lo);
            END;
            FIO.Seek(hnd,PEinfo[i].fpos);
            FIO.WrBin(hnd,b,count);
        END;
    END;
    fileClose(useLFN,hnd);
END doPatch;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE viewMatches (VAR errcode:CARDINAL; VAR PEinfo : PEinfostruc;
                      useLFN,addNL,SKIPLOG,altdisplay,DEBUG:BOOLEAN;
                      lastfile:CARDINAL;anchor:ptrToEntry):BOOLEAN;
CONST
    sPEheader = "Missing or unsupported PE headers.";
VAR
    zefile:pathtype;
    flagAbort:BOOLEAN;
    i:CARDINAL;
    winheaderpos:LONGCARD;
    R:str128;
    id:CARDINAL;
BEGIN
    errcode:=errNone;
    flagAbort:=FALSE;
    i:=firstEntry-1;
    LOOP
        INC(i);
        IF i > lastfile THEN EXIT; END;
        getMatchEntry(zefile, i,anchor);


        id:=getFileType (winheaderpos,  useLFN,DEBUG,zefile) ; (* must be done first *)
        getdesc(id,R);

        rptwr(SKIPLOG, LOGFILE,sOK,enquote(zefile));
        rptwr(SKIPLOG, LOGFILE,sINFO,R);
        CASE id OF
        | idPE,idUPXwin32 :
            IF grabPEinfos(PEinfo, winheaderpos,useLFN,DEBUG,zefile) THEN
                dmpPEinfos(SKIPLOG,altdisplay, PEinfo);   (* show values current state *)
            ELSE
                IF NOT(altdisplay) THEN
                    rptwr(SKIPLOG, LOGFILE,sERR,sPEheader);
                END;
                errcode:=errBadPEinfos;
                (* abort(errBadPEinfos,""); *)
            END;
        ELSE
            IF NOT(altdisplay) THEN
                rptwr(SKIPLOG, LOGFILE,sERR,"This file does not look like a supported PE 32-bits executable.");
            END;
            errcode:=errUnsupported;
            (* abort(errUnsupported,""); *)
        END;


        IF addNL THEN
            IF i # lastfile THEN WrLn; END;
        END;

        flagAbort:=ChkEscape();
        IF flagAbort THEN EXIT; END;
    END;
    RETURN flagAbort;
END viewMatches;

PROCEDURE paranoDump (useLFN,SKIPLOG,altdisplay:BOOLEAN;F:pathtype);
VAR
    id:CARDINAL;
    winheaderpos : LONGCARD;
    ok:BOOLEAN;
    PEinfo:PEinfostruc;
BEGIN
    id:=getFileType (winheaderpos,  useLFN,FALSE,F) ; (* must be done first *)
    ok:= grabPEinfos(PEinfo, winheaderpos,useLFN,FALSE,F);
    dmpPEinfos(SKIPLOG,altdisplay, PEinfo);
END paranoDump;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    cmdtype = (cmdnone,cmdview,cmdpatch);

PROCEDURE newcmd (VAR cmd:cmdtype; wanted:cmdtype  ):BOOLEAN ;
VAR
    rc:BOOLEAN;
BEGIN
    rc:=TRUE;
    IF    cmd = cmdnone THEN
        cmd:=wanted;
    ELSIF cmd = wanted  THEN
        cmd:=wanted; (* useless, eh ! *)
    ELSE
        rc:=FALSE;
    END;
    RETURN rc;
END newcmd;

(* #.# *)

PROCEDURE getmajmin(VAR major,minor:CARDINAL; R:ARRAY OF CHAR):BOOLEAN ;
VAR
    S:str128;
    Z:str16;
    p:CARDINAL;
    v:LONGCARD;
    ok:BOOLEAN;
BEGIN
    GetString(R,S);
    p:=Str.CharPos(S,dot);
    IF p = MAX(CARDINAL) THEN
        Str.Append(S,dot+"0");
        p:=Str.CharPos(S,dot);
    END;
    Str.Slice(Z,S,0,p);
    v:=Str.StrToCard(Z,10,ok);
    IF NOT(ok) THEN RETURN FALSE;END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE;END;
    major:=CARDINAL(v);

    Str.Delete(S,0,p+1);
    v:=Str.StrToCard(S,10,ok);
    IF NOT(ok) THEN RETURN FALSE;END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE;END;
    minor:=CARDINAL(v);
    RETURN TRUE;
END getmajmin;

PROCEDURE newmajmin (VAR PEinfo:PEinfostruc;
                    major,minor,i:CARDINAL):BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    IF PEinfo[i].isPresent THEN
        rc:=FALSE;
    ELSE
        PEinfo[i].isPresent := TRUE;
        PEinfo[i].value.lo  := minor;
        PEinfo[i].value.hi  := major;
        rc:=TRUE;
    END;
    RETURN rc;
END newmajmin;

(* #, $#, 0x#, [H]#[H], [B]#[B] *)

PROCEDURE str2lc (VAR lc:LONGCARD;S:ARRAY OF CHAR):BOOLEAN;
VAR
    ok:BOOLEAN;
    base:CARDINAL;
BEGIN
    IF Str.Match(S,"$*") THEN
        base:=16;  Str.Delete(S,0,1);
    ELSIF Str.Match(S,"0X*") THEN (* FOR our C f(r)iends... *)
        base:=16;  Str.Delete(S,0,2);
    ELSIF Str.Match(S,"H*") THEN
        base:=16;  Str.Delete(S,0,1);
    ELSIF Str.Match(S,"B*") THEN
        base:=2;   Str.Delete(S,0,1);
    ELSIF Str.Match(S,"*H") THEN
        base:=16;  Str.Delete(S, Str.Length(S)-1,1); (* or brutal charnull *)
    ELSIF Str.Match(S,"*B") THEN
        base:=2;   Str.Delete(S, Str.Length(S)-1,1);
    ELSE
        base:=10;
    END;
    lc:=Str.StrToCard(S,base,ok);
    RETURN ok;
END str2lc;

PROCEDURE getcard (VAR n:CARDINAL;R:ARRAY OF CHAR):BOOLEAN;
VAR
    S:str128;
    lc:LONGCARD;
BEGIN
    GetString(R, S);
    IF str2lc(lc,S)=FALSE THEN RETURN FALSE;END;
    IF lc > MAX(CARDINAL) THEN RETURN FALSE;END;
    n:=CARDINAL(lc);
    RETURN TRUE;
END getcard;

PROCEDURE newcard (VAR PEinfo:PEinfostruc;
                   n,flag,i:CARDINAL):BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    IF PEinfo[i].isPresent THEN
        rc:=FALSE;
    ELSE
        PEinfo[i].isPresent := TRUE;
        PEinfo[i].value.hi  := flag;
        PEinfo[i].value.lo  := n;
        rc:=TRUE;
    END;
    RETURN rc;
END newcard;

PROCEDURE getlongcard (VAR n:LONGCARD;R:ARRAY OF CHAR):BOOLEAN;
VAR
    S:str128;
BEGIN
    GetString(R, S);
    RETURN str2lc(n,S);
END getlongcard;

PROCEDURE newlongcard (VAR PEinfo:PEinfostruc;
                      n : LONGCARD; i:CARDINAL):BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    IF PEinfo[i].isPresent THEN
        rc:=FALSE;
    ELSE
        PEinfo[i].isPresent := TRUE;
        PEinfo[i].value.hilo:= n;
        rc:=TRUE;
    END;
    RETURN rc;
END newlongcard;

PROCEDURE getmask (VAR lc:LONGCARD; R:ARRAY OF CHAR  ):BOOLEAN;
CONST
    minCPU = 0;
    maxCPU = 31; (* //V10A was 7 *)
VAR
    S:str128;
    Z:str16;
    i,j,pb:CARDINAL;
    cpu,bitmask:LONGCARD;
BEGIN
    GetString(R, S);
    IF same(S,"*") THEN
        lc:=anyCPU;
        RETURN TRUE;
    END;
    ReplaceChar(S,dot,coma);
    Str.Append(S,coma); (* trick *)
    lc:=0;
    pb:=0;
    FOR i := 1 TO CharCount(S,coma) DO
         isoleItemS(Z,  S,coma,i-1);
         IF str2lc(cpu,Z) THEN
             IF cpu > maxCPU THEN
                 INC(pb);
             ELSE
                 bitmask := 01H;
                 FOR j:=1 TO CARDINAL(cpu) DO bitmask := (bitmask << 1); END;
                 lc:=(lc OR bitmask);
             END;
         ELSE
             INC(pb);
         END;
    END;
    RETURN (pb=0);
END getmask;

PROCEDURE newmask (VAR PEinfo:PEinfostruc;
                  n : LONGCARD; i:CARDINAL) : BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    IF PEinfo[i].isPresent THEN
        rc:=FALSE;
    ELSE
        PEinfo[i].isPresent := TRUE;
        PEinfo[i].value.hilo:= n;
        rc:=TRUE;
    END;
    RETURN rc;
END newmask;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    sNoAffinityMask = "Missing affinity mask cannot be changed.";
    sReadOnly       = "Read-Only file cannot be patched.";
    sUpdated        = "File has been updated.";
    sVerified       = "File has been verified.";
CONST
    firstparm = 1;
    maxparm   = 1;
VAR
    ndx,major,minor,charac,errcode,lastfile,parmcount,i,opt,lastparm:CARDINAL;
    S,R,zefile:pathtype;
    alcatraz,addNL,useLFN,DEBUG:BOOLEAN;
    parm:ARRAY [firstparm..maxparm] OF pathtype;
    cmd:cmdtype;
    anchor:ptrToEntry;
    wasRO,altdisplay,updateRO,SKIPLOG,NEWLOG,updateStamp:BOOLEAN;
    lc:LONGCARD;
    ok:BOOLEAN;
    orgstamp:dttype;
VAR
    refPEinfo , newPEinfo : PEinfostruc;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE; (* don't let topspeed handle problems *)

    WrLn;   (* beautify display *)

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    lastparm    := firstparm-1;
    cmd         := cmdnone;
    useLFN      := TRUE;
    addNL       := TRUE;

    updateRO    := FALSE;
    SKIPLOG     := FALSE;
    NEWLOG      := FALSE;
    updateStamp := FALSE;
    altdisplay  := FALSE;

    zeroPEinfos (FALSE, newPEinfo); (* SET all data TO NOT present *)

    DEBUG       := FALSE;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "V"+delim+"VIEW"+delim+
                                  "N"+delim+"NL"+delim+
                                  "DEBUG"+delim+

                                  "R"+delim+"READONLY"+delim+
                                  "L"+delim+"LOG"+delim+
                                  "C"+delim+"CREATE"+delim+
                                  "D"+delim+"DATETIME"+delim+
                                  "A"+delim+"ALTERNATE"+delim+

                                  "N:"+delim+"O:"+delim+"NT:"+delim+"OS:"+delim+
                                  "W:"+delim+"S:"+delim+"WIN:"+delim+"SUB:"+delim+
                                  "A:"+delim+"I:"+delim+"APP:"+delim+"IMG:"+delim+

                                  "C:"+delim+"CHARACTERISTICS:"+delim+
                                  "M"+delim+"MULTIPLE"+delim+"BIT14OFF"+delim+
                                            "MULTI"+delim+
                                  "S"+delim+"SINGLE"+delim+"BIT14ON"+delim+

                                  "V:"+delim+"VERSION:"+delim+

                                  "P:"+delim+"CPU:"+delim+"PROCESSOR:"+delim+
                                  "M:"+delim+"MASK:"+delim+"AFFINITY:"


                              );
            CASE opt OF
            | 1,2,3 :    abort(errHelp,"");
            | 4,5:       useLFN       := FALSE;
            | 6,7:       IF newcmd(cmd,cmdview)=FALSE THEN abort(errCmd,S);END;
            | 8,9:       addNL        := FALSE;
            | 10:        DEBUG        := TRUE;

            | 11,12:     updateRO     := TRUE;
            | 13,14:     SKIPLOG      := TRUE;
            | 15,16:     NEWLOG       := TRUE;
            | 17,18:     updateStamp  := TRUE;
            | 19,20:     altdisplay   := TRUE;

            |21,22,23,24:IF getmajmin(major,minor, R)=FALSE THEN abort(errMajMin,S); END;
                         ndx:=ORD(idMajorOperatingSystemVersion);
                         IF newmajmin(newPEinfo,  major,minor,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;
            |25,26,27,28:IF getmajmin(major,minor, R)=FALSE THEN abort(errMajMin,S); END;
                         ndx:=ORD(idMajorSubsystemVersion);
                         IF newmajmin(newPEinfo,  major,minor,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;
            |29,30,31,32:IF getmajmin(major,minor, R)=FALSE THEN abort(errMajMin,S); END;
                         ndx:=ORD(idMajorImageVersion);
                         IF newmajmin(newPEinfo,  major,minor,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;

            |33,34:      IF getcard(charac, R)=FALSE THEN abort(errCardVal,S);END;
                         ndx:=ORD(idCharacteristicsFlag);
                         IF newcard(newPEinfo, charac,flagVALUE,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;
            |35,36,37,38:ndx:=ORD(idCharacteristicsFlag);
                         IF newcard(newPEinfo, bit14off,flagAND,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;
            |39,40,41:   ndx:=ORD(idCharacteristicsFlag);
                         IF newcard(newPEinfo, bit14on,flagOR,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;

            |42,43:      IF getlongcard(lc, R)=FALSE THEN abort(errCardVal,S);END;
                         ndx:=ORD(idWin32VersionValue);
                         IF newlongcard(newPEinfo, lc,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;

            |44,45,46:   IF getmask(lc, R)=FALSE THEN abort(errMask,S);END;
                         ndx:=ORD(idProcessorAffinityMask);
                         IF newmask(newPEinfo, lc,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;
            |47,48,49:   IF getlongcard(lc, R)=FALSE THEN abort(errCardVal,S);END;
                         ndx:=ORD(idProcessorAffinityMask);
                         IF newlongcard(newPEinfo, lc,ndx)=FALSE THEN
                             abort(errRedefined,S);
                         END;
                         IF newcmd(cmd,cmdpatch)=FALSE THEN abort(errCmd,S);END;

            ELSE
                abort(errUnknownOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errTooManyParms,S);END;
            Str.Copy(parm[lastparm],S); (* keep case *)
        END;
    END;
    IF lastparm < firstparm THEN abort(errMissingSpec,"");END;

    IF cmd = cmdnone THEN cmd:=cmdview;END;

    useLFN:=( useLFN AND fileSupportLFN() );

    Str.Copy(zefile,parm[firstparm]);
    CASE cmd OF
    | cmdview:
        IF same(zefile,dot) THEN Str.Copy(zefile,stardotstar);END;
        IF chkJokerPath(zefile) THEN abort(errJokerPath,zefile);END;
    ELSE
        IF chkJoker(zefile) THEN abort(errJoker,zefile);END;
        IF fileIsDirectorySpec(useLFN,zefile) THEN abort(errNotFile,zefile);END;
        IF fileExists(useLFN,zefile)=FALSE THEN abort(errNotFound,zefile);END;

        IF foundFixFile(extCD) THEN SKIPLOG:=TRUE;END; (* .cd found hence force -l *)
        initRptLog(SKIPLOG,NEWLOG,LOGFILE);
    END;

    lastfile:=buildMatchList (anchor, useLFN,zefile);
    CASE lastfile OF
    | firstEntry-1:
        abort(errNotFound,zefile);
    | TOOMANY:
        abort(errTooMany,zefile);
    ELSE
        IF NOT(SKIPLOG) THEN
            SKIPLOG := ( cmd = cmdview ); (* log only if patching ONE file *)
        END;
        alcatraz:=viewMatches (errcode,refPEinfo,
                              useLFN,addNL,SKIPLOG,altdisplay,DEBUG,
                              lastfile,anchor);
        getMatchEntry(zefile, firstEntry,anchor); (* here for later patch use *)
        freeMatchList(anchor);
        IF alcatraz THEN abort(errEscape,"");END;
        IF cmd = cmdview THEN abort(errNone,"");END;
    END;

    (* ------------------------------------------------------------ *)
    (* cmd cannot be cmdview now *)
    (* ------------------------------------------------------------ *)

    (* patch : first, let's check that yes, we can do it *)

    CASE errcode OF
    | errBadPEinfos:  abort(errBadPEinfos,"");
    | errUnsupported: abort(errUnsupported,"");
    END;

    wasRO:= fileIsRO(useLFN,zefile);
    IF wasRO THEN
        IF updateRO THEN
            fileSetRW(useLFN,zefile);
        ELSE
            rptwr(SKIPLOG, LOGFILE,sERR,sReadOnly);
            abort(errROtarget,"");
        END;
    END;

    ok:=getfilestamp(orgstamp, useLFN,zefile);

    IF createPatch(newPEinfo, refPEinfo) THEN
        rptwr(SKIPLOG, LOGFILE,"","");

        doPatch(useLFN,zefile,newPEinfo);
        IF NOT(updateStamp) THEN ok:=setFileStamp(useLFN,orgstamp,zefile);END;
        IF wasRO THEN fileSetRO(useLFN,zefile);END;

        dmpPEinfos(SKIPLOG,altdisplay, newPEinfo); (* we could reread data too *)
        rptwr(SKIPLOG, LOGFILE,"","");
        rptwr(SKIPLOG, LOGFILE,sINFO,sUpdated);
        rptwr(SKIPLOG, LOGFILE,"","");
        IF DEBUG THEN
            paranoDump(useLFN,SKIPLOG,altdisplay,zefile);
            rptwr(SKIPLOG, LOGFILE,"","");
            rptwr(SKIPLOG, LOGFILE,sINFO,sVerified);
            rptwr(SKIPLOG, LOGFILE,"","");
        END;

        i:=errNone;
        S:="";
    ELSE
        rptwr(SKIPLOG, LOGFILE,sERR,sNoAffinityMask);

        i:=errNoWeCant;
        Str.Copy(S,zefile);
    END;

    flushRptLog(SKIPLOG,LOGFILE);

    abort(i,S);
END fixPE.


