MODULE XDSsplit;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT IO;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    str1024 = ARRAY [0..1023] OF CHAR;
CONST
    colon     = ":";
    slash     = "/";
    antislash = "\";
    dot       = ".";
    charnull  = 0C;
    semicolon = ";";
    pound     = "#";

PROCEDURE LtrimBlanks (VAR S : ARRAY OF CHAR);
CONST
    espace=ORD(" ");
VAR
    i,len : CARDINAL;
BEGIN
    len := Str.Length (S);
    i   := 0;
    LOOP
        IF i = len THEN EXIT; END;
        IF ORD(S[i]) > espace THEN EXIT; END;
        INC (i);
    END;
    Str.Delete (S,0,i); (* procedure handles removal of i=0 *)
END LtrimBlanks;

PROCEDURE RtrimBlanks (VAR S : ARRAY OF CHAR);
CONST
    espace=ORD(" ");
VAR
    i,len : CARDINAL;
BEGIN
    len := Str.Length (S);
    i   := len;
    LOOP
        IF i = 0 THEN EXIT; END;
        IF ORD(S[i-1]) > espace THEN EXIT; END;
        DEC (i);
    END;
    Str.Delete (S,i,len-i);
END RtrimBlanks;

PROCEDURE same(S1,S2 : ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF Str.Compare(S1,S2)=0 THEN RETURN TRUE; END;
    RETURN FALSE;
END same;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    exe      = "XDSSPLIT";
    extTXT   = ".TXT";
    datafile = exe+extTXT;

PROCEDURE abort (errcode:CARDINAL; S:ARRAY OF CHAR);
BEGIN
    IF errcode # 0 THEN
        WrStr(exe+" : ");WrStr(S);WrLn;
    END;
    HALT (errcode);
END abort;

(* ------------------------------------------------------------ *)

(*
    handle both "\" and "/" separators
    *:*\[n[.[e]]]

    grab last "\" if any then check for next "."

    we could fix "/" and "\" mix, as well as empty extension ?
*)

PROCEDURE qdSplitAllPath (udne:ARRAY OF CHAR;VAR u,d,n,e:ARRAY OF CHAR);
VAR
    F:str1024; (* bigger than pathtype *)
    p:CARDINAL;
BEGIN
    (*
    Lib.SplitAllPath(F,u,d,n,e);
    Lib.MakeAllPath(F,u,d,n,e);
    IF Str.Compare(udne,F) = 0 THEN RETURN; END;
    *)

    (* let's fix JPI buggy function so we still get "u:" "\*\" "f" ".e" *)

    p:=Str.CharPos(udne,colon);
    IF p = MAX(CARDINAL) THEN
        Str.Copy(u,"");
    ELSE
        Str.Slice(u,udne,0,p+1);
        Str.Delete(udne,0,p+1);
    END;

    Str.Copy(F,udne);
    LOOP
        p:=Str.CharPos(F,slash);
        IF p = MAX(CARDINAL) THEN EXIT;END;
        F[p]:=antislash;
    END;
    p:=Str.RCharPos(F,antislash);
    IF p = MAX(CARDINAL) THEN
        Str.Copy(d,"");
    ELSE
        Str.Slice(d,udne,0,p+1);
        Str.Delete(udne,0,p+1);
    END;

    p:=Str.RCharPos(udne,dot);
    CASE p OF
    | MAX(CARDINAL) :
        Str.Copy(n,udne);
        Str.Copy(e,"");
    | 0 :
        Str.Copy(n,"");
        Str.Copy(e,udne);
    ELSE
        Str.Slice(n,udne,0,p);
        Str.Delete(udne,0,p);
        Str.Copy(e,udne);
    END;

END qdSplitAllPath;

(* ------------------------------------------------------------ *)

VAR
    S,R,udne,u,d,n,e:str1024;
    uu,dd,nn,ee:str1024;
    hin:FIO.File;
    ok:BOOLEAN;
BEGIN
    FIO.IOcheck := FALSE;
    WrLn;

    IF Lib.ParamCount() # 0 THEN abort(1,"XDS Lib.SplitAllPath()");END;
    IF FIO.Exists(datafile)=FALSE THEN abort(2,datafile+" does not exist !");END;

    hin:=FIO.OpenRead(datafile);
    WHILE NOT (FIO.EOF) DO
        FIO.RdStr(hin,S);
        LtrimBlanks(S);RtrimBlanks(S);
        CASE S[0] OF
        | charnull,semicolon,pound:
            ;
        ELSE
            R:='::: "$"';
            Str.Subst(R,"$",S);
            WrStr(R);WrLn;

            Lib.SplitAllPath(S,u,d,n,e);
            Lib.MakeAllPath(R,u,d,n,e);
            IF Str.Compare(S,R) # 0 THEN
                WrStr("--- rebuilt Lib seems wrong");WrLn;
                WrLn;
            END;
            qdSplitAllPath(S,uu,dd,nn,ee);

            ok:=same(u,uu);
            ok:=(ok AND same(d,dd));
            ok:=(ok AND same(n,nn));
            ok:=(ok AND same(e,ee));
            IF ok THEN
                WrStr("+++ Lib and QD results match");WrLn;
                R:='    U = "$"';  Str.Subst(R,"$",u);  WrStr(R);WrLn;
                R:='    D = "$"';  Str.Subst(R,"$",d);  WrStr(R);WrLn;
                R:='    N = "$"';  Str.Subst(R,"$",n);  WrStr(R);WrLn;
                R:='    E = "$"';  Str.Subst(R,"$",e);  WrStr(R);WrLn;
                WrLn;
            ELSE
                R:='    u = "$"';  Str.Subst(R,"$",u);  WrStr(R);WrLn;
                R:='    d = "$"';  Str.Subst(R,"$",d);  WrStr(R);WrLn;
                R:='    n = "$"';  Str.Subst(R,"$",n);  WrStr(R);WrLn;
                R:='    e = "$"';  Str.Subst(R,"$",e);  WrStr(R);WrLn;
                WrLn;
                R:='    uu= "$"';  Str.Subst(R,"$",uu);  WrStr(R);WrLn;
                R:='    dd= "$"';  Str.Subst(R,"$",dd);  WrStr(R);WrLn;
                R:='    nn= "$"';  Str.Subst(R,"$",nn);  WrStr(R);WrLn;
                R:='    ee= "$"';  Str.Subst(R,"$",ee);  WrStr(R);WrLn;
                WrLn;
            END;
        END;
    END;
    FIO.Close(hin);

    abort(0,"");
END XDSsplit.




(*


; classic cases

c:\z
foo.bar
foo
foo.
foo\bar.e
foo\bar
\foo\bar.baz

; inner dots

\foo.bar/baz.ext
\foobar/baz.ext

cd:foo.bar/baz.
cd:foobar/baz.

cd:foo.bar\baz
cd:foobar\baz

c:foo.bar/baz.bix\
c:foobar/bazbix\

c:foo.bar/baz.bix/eh
c:foobar/bazbix/eh

c:foo.bar/baz.bix/eh.
c:foobar/bazbix/eh.

*)
