(* ---------------------------------------------------------------
Title         Q&D Factorize
Author        PhG
Overview      see help
Notes         chkPrime is about 4 times slower than factorize()
              for ADW, we're removed chkEscape() calls
Bugs
Wish List     create a VERYLONGCARD type for 64 bits ?

--------------------------------------------------------------- *)

MODULE Fact_ADW;

IMPORT Strings;
IMPORT STextIO;
IMPORT WholeStr;
IMPORT Environment;
IMPORT ConvTypes;
IMPORT LWholeStr;
IMPORT WIN32;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    tabul         = CHR(9);
    bs            = CHR(8);
    space         = " ";
    blank         = space;
    slash         = "/";
    dash          = "-";
    delim         = ",";
    equal         = "=";
    colon         = ":";
    bsblankbs     = bs+blank+bs;

CONST
    cr            = CHR(13);
    lf            = CHR(10);
    nl            = cr+lf;
    digits        = "0123456789";
    nullchar      = 0C; (* CHR(0) *)
    charnull      = nullchar;
    doublequote   = '"';
    singlequote   = "'";

TYPE
    str128 = ARRAY [0..127] OF CHAR;
    str80  = ARRAY [0..79] OF CHAR;
    str2048= ARRAY [0..2047] OF CHAR;
    str1024= ARRAY [0..1023] OF CHAR;
    str256 = ARRAY [0..255] OF CHAR;

PROCEDURE WrStr(S:ARRAY OF CHAR);
BEGIN
    STextIO.WriteString(S);    (* //ADW WriteString but WriteLn : go figure ! *)
END WrStr;

PROCEDURE WrLn();
BEGIN
    STextIO.WriteLn();
END WrLn;

PROCEDURE Prepend(VAR R : ARRAY OF CHAR; S:ARRAY OF CHAR);
VAR
    Z : str1024; (* hack should do -- we don't feel like really programming with ADW *)
BEGIN
    Z:=S;
    Strings.Append(R,Z);
    R:=Z;
END Prepend;

PROCEDURE Concat (S1,S2:ARRAY OF CHAR; VAR R : ARRAY OF CHAR);
BEGIN

END Concat;

(* note we want TS mode : replacing fist occurrence, not all occurrences *)
PROCEDURE Subst (VAR S : ARRAY OF CHAR; oldstr,newstr : ARRAY OF CHAR);
VAR
    found:BOOLEAN;
    foundat, p,len : CARDINAL;
BEGIN
    Strings.FindNext(oldstr,S,0,found,foundat);
    IF found THEN
        len:=Strings.Length(oldstr);
        Strings.Delete(S,foundat,len);
        Strings.Insert(newstr,foundat,S); (* //ADW reversed logic here *)
    END;
END Subst;

PROCEDURE CharPos(S : ARRAY OF CHAR; c:CHAR) : CARDINAL;
VAR
    found : BOOLEAN;
    foundat : CARDINAL;
BEGIN
    (* //ADW yet another overcomplicated library forcing us to reinvent wheel *)
    Strings.FindNext (c, S, 0, found, foundat);
    IF NOT(found) THEN foundat := MAX(CARDINAL); END; (* let's hope it's okay with ADW *)
    RETURN foundat;
END CharPos;

PROCEDURE Pos(S,wanted : ARRAY OF CHAR) : CARDINAL;
VAR
    found : BOOLEAN;
    foundat : CARDINAL;
BEGIN
    (* //ADW yet another overcomplicated library forcing us to reinvent wheel *)
    Strings.FindNext (wanted, S, 0, found, foundat);
    IF NOT(found) THEN foundat := MAX(CARDINAL); END; (* let's hope it's okay with ADW *)
    RETURN foundat;
END Pos;

TYPE
    SetOfChars = SET OF CHAR;

PROCEDURE isoleItem(VAR R: ARRAY OF CHAR;
                    S: ARRAY OF CHAR; T: SetOfChars; N: CARDINAL);

VAR
    len,i,j,k:CARDINAL;
BEGIN
    len := Strings.Length(S);
    i   := 0;
    WHILE ( (i<len) AND (N > 0) ) DO
        IF (S[i] IN T) THEN DEC(N); END;
        INC(i);
    END;
    (* i is at first char after separator *)
    j := i;
    WHILE ( (j<len) AND NOT (S[j] IN T) ) DO
        INC(j);
    END;
    k := 0;
    WHILE (i<j) DO
        R[k] := S[i];
        INC(k);
        INC(i);
    END;
    R[k]:=CHR(0);
END isoleItem;

(* now we can safely parse "a,b,c,,e" ! what about ",,c,d,e" ? *)
(* first element is 0 *)

PROCEDURE isoleItemS(VAR R: ARRAY OF CHAR;
                     S: ARRAY OF CHAR; T: ARRAY OF CHAR; N: CARDINAL);
VAR
    CS  : SetOfChars;
    len : CARDINAL;
BEGIN
    len := Strings.Length(T);
    CS := SetOfChars{};
    WHILE len > 0 DO
        DEC(len);
        INCL(CS,T[len]);
    END;
    isoleItem(R,S,CS,N);
END isoleItemS;

PROCEDURE ItemS(VAR str:ARRAY OF CHAR; options:ARRAY OF CHAR;delim:CHAR;i:CARDINAL);
BEGIN
    (* //TODO *)
    isoleItemS(str, options,delim,i);
END ItemS;

PROCEDURE Belongs (S : ARRAY OF CHAR; c:CHAR  ) : BOOLEAN;
BEGIN
    RETURN ( CharPos(S,c) # MAX(CARDINAL) );
END Belongs;

PROCEDURE verifyString (S,allowed:ARRAY OF CHAR):BOOLEAN;
VAR
    i,mismatches : CARDINAL;
    ch : CHAR;
BEGIN
    mismatches := 0;
    FOR i := 1 TO Strings.Length(S) DO
        ch := S[i-1];
        IF Belongs(allowed,ch)=FALSE THEN INC(mismatches);END;
    END;
    RETURN (mismatches=0);
END verifyString;

PROCEDURE same(S1,S2 : ARRAY OF CHAR) : BOOLEAN;
BEGIN
    RETURN ( Strings.Equal(S1,S2) );
END same;

PROCEDURE map (c : CHAR ;table : ARRAY OF CHAR) : CHAR;
BEGIN
    RETURN table [ ORD (c) ];
END map;

CONST
    ignore = " ";
    (* darn, DOS and OEM are different but this should do for our test *)
    Tupper = "                                "+
             "                                "+
             " ABCDEFGHIJKLMNOPQRSTUVWXYZ     "+
             " ABCDEFGHIJKLMNOPQRSTUVWXYZ     "+
             "CUEAAAACEEEIIIAAE  OOOUUYOU     "+ (* use "ae" *)
             "AIOUNN                          "+
             "                                "+
             "                              Y ";
    Tlower = "                                "+
             "                                "+
             " abcdefghijklmnopqrstuvwxyz     "+
             " abcdefghijklmnopqrstuvwxyz     "+
             "cueaaaaceeeoooaae  ooouuyou     "+ (* use "ae" *)
             "aiounn                          "+
             "                                "+
             "                              y ";

PROCEDURE UpperCase (VAR S : ARRAY OF CHAR);
VAR
    len : CARDINAL;
    i   : CARDINAL;
    c   : CHAR;
    ch  : CHAR;
BEGIN
    len := Strings.Length(S);
    IF len = 0 THEN RETURN; END;         (* this WAS a problem ! *)
    FOR i := 0 TO (len - 1) DO
        c  := S[i];
        ch := map(c,Tupper);
        IF ch # ignore THEN S[i] := ch; END;
    END;
END UpperCase;

PROCEDURE LowerCase (VAR S : ARRAY OF CHAR);
VAR
    len : CARDINAL;
    i   : CARDINAL;
    c   : CHAR;
    ch  : CHAR;
BEGIN
    len := Strings.Length(S);
    IF len = 0 THEN RETURN; END;         (* this WAS a problem ! *)
    FOR i := 0 TO (len - 1) DO
        c  := S[i];
        ch := map(c,Tlower);
        IF ch # ignore THEN S[i] := ch; END;
    END;
END LowerCase;


PROCEDURE isOption (S : ARRAY OF CHAR) : BOOLEAN;
BEGIN
CASE S[0] OF
| slash, dash :
    RETURN TRUE;
ELSE
    RETURN FALSE;
END;
END isOption;

PROCEDURE ReplaceChar (VAR S : ARRAY OF CHAR;
                       oldchar, newchar : CHAR);
VAR
    p : CARDINAL;
BEGIN
    IF oldchar=newchar THEN RETURN; END; (* silly... *)
    LOOP
        p := CharPos(S,oldchar);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        IF newchar = "" THEN
            Strings.Delete(S,p,1);
        ELSE
            S[p]:=newchar;
        END;
    END;
END ReplaceChar;

PROCEDURE GetOptIndex (S : ARRAY OF CHAR; options : ARRAY OF CHAR) : CARDINAL;
VAR
    i,p   : CARDINAL;
    str : str80; (* option tested against s *)
    tmp : str256; (* oversized *)
BEGIN
    Strings.Delete(S,0,1); (* remove leading - or / because we arrive if option *)
    Strings.Append(delim, options); (* //ADW reversed logic *)

    Subst(S,equal,colon); (* command line option xxx= becomes xxx: *)
    ReplaceChar(options,equal,colon); (* all possibles option xxx= become xxx: *)

    i := 0;
    LOOP
        ItemS(str,options,delim,i);
        IF Strings.Length(str)=0 THEN RETURN 0; END;
        p:=CharPos(str,colon);
        IF p = MAX(CARDINAL) THEN (* normal option *)
            IF same(str,S) THEN RETURN (i+1); END;
        ELSE (* option with a number : cut S at colon *)
            Strings.Extract(S,0,p+1,tmp);
            IF same(str,tmp) THEN RETURN (i+1); END;
        END;
        INC(i);
    END;
END GetOptIndex;

(* //ADW passim : reversed logic for Strings.Append() *)

PROCEDURE argc (cli : ARRAY OF CHAR) : CARDINAL;
VAR
    state    : (empty,intoken,instring);
    cliLen   : CARDINAL;
    cliPos   : CARDINAL;
    argCount : CARDINAL;
    ch       : CHAR;
    code     : CARDINAL;
    arg      : str128;
    sdelim   : CARDINAL;
BEGIN
    cliLen   := Strings.Length(cli);
    cliPos   := 1;
    argCount := 0;
    state    := empty;
    LOOP
        IF cliPos > cliLen THEN EXIT; END;
        ch := cli[cliPos-1];
        code := ORD(ch);
        CASE state OF
        | empty :
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN
                sdelim:= code;
                state := instring;        (* begin new string *)
                (* arg:=ch; //ADW does not want us to concatenate a string and a char *)
                Strings.Assign(ch,arg);
                INC(argCount);
            ELSIF code > ORD(blank) THEN  (* quote already trapped *)
                state := intoken;         (* begin new token *)
                Strings.Assign(ch,arg);
                INC(argCount);
            END;
        | intoken :
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN (* quote in token *)
                sdelim:= code; (* fix *)
                state := instring;        (* if string in parameter *)
                Strings.Append(ch,arg); (* //ADW reversed logic *)
            ELSIF code > ORD(blank) THEN
                Strings.Append(ch,arg);
            ELSE
                state := empty;           (* end of token *)
            END;
        | instring :
            IF code = sdelim THEN
                Strings.Append(ch,arg);
                state := empty;           (* end of string *)
            ELSIF code > ORD(blank) THEN
                Strings.Append(ch,arg);
            ELSE
                Strings.Append(blank,arg);    (* remove TAB and controls if any *)
            END;
        END;
        INC(cliPos);
    END;
    RETURN argCount;
END argc;

PROCEDURE argv (VAR argument : ARRAY OF CHAR;
                cli : ARRAY OF CHAR; n : CARDINAL);
VAR
    state    : (empty,intoken,instring);
    cliLen   : CARDINAL;
    cliPos   : CARDINAL;
    argCount : CARDINAL;
    ch     : CHAR;
    code     : CARDINAL;
    arg      : str128;
    sdelim   : CARDINAL;
BEGIN
    IF ( (n < 1) OR (n > argc(cli)) ) THEN
        argument:="";
        RETURN;
    END;
    cliLen   := Strings.Length(cli);
    cliPos   := 1;
    argCount := 0;
    state    := empty;
    LOOP
        IF cliPos > cliLen THEN EXIT; END;
        ch := cli[cliPos-1];
        code := ORD(ch);
        CASE state OF
        | empty :
            IF n = argCount THEN EXIT; END; (* argV$ test *)
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN
                sdelim:= code;
                state := instring;        (* begin new string *)
                Strings.Assign(ch,arg);
                INC(argCount);
            ELSIF code > ORD(blank) THEN  (* quote already trapped *)
                state := intoken;         (* begin new token *)
                Strings.Assign(ch,arg);
                INC(argCount);
            END;
        | intoken :
            IF ( (code = ORD(doublequote)) OR (code=ORD(singlequote)) ) THEN (* quote in token *)
                sdelim:= code; (* fix *)
                state := instring;        (* if string in parameter *)
                Strings.Append(ch,arg);
            ELSIF code > ORD(blank) THEN
                Strings.Append(ch,arg);
            ELSE
                state := empty;           (* end of token *)
            END;
        | instring :
            IF code = sdelim THEN
                Strings.Append(ch,arg);
                state := empty;           (* end of string *)
            ELSIF code > ORD(blank) THEN
                Strings.Append(ch,arg);
            ELSE
                Strings.Append(blank,arg);    (* remove TAB and controls if any *)
            END;
        END;
        INC(cliPos);
    END;
    argument:=arg;
END argv;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE ChkEscape (  ) : BOOLEAN;
BEGIN
    RETURN FALSE;
END ChkEscape;

PROCEDURE video (S : ARRAY OF CHAR; flag:BOOLEAN);
VAR
    len : CARDINAL;
    i   : CARDINAL;
BEGIN
    len := Strings.Length(S);
    CASE flag OF
    | TRUE :
        WrStr(S);
    | FALSE:
        FOR i := 1 TO len DO
            WrStr(bsblankbs);
        END;
    END;
END video;

(* try and fix TopSpeed bug when tabs are in command line *)
(* try and fix our NEVER using tabs in any command line too ! *)

PROCEDURE cleantabs (VAR S : ARRAY OF CHAR);
VAR
    p : CARDINAL;
BEGIN
    LOOP
        p:=CharPos(S,tabul);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        (* S[p]:=Str.Delete(S,p,1); (* delete unwanted tabulation *) *)
        S[p]:=space;
    END;
END cleantabs;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    winum         = 13;  (* "#.###.###.###" *)
    dot           = ".";
    coma          = ",";
    filteredChars = dot+coma;
CONST
    msgWait = "Computing, please wait... ";
    kEsc    = 64;       (* keep 9X happy, was 1024 *)

(* ------------------------------------------------------------ *)

PROCEDURE filter (VAR R:ARRAY OF CHAR; filtered:ARRAY OF CHAR );
VAR
    i:CARDINAL;
    S:str128;
    ch:CHAR;
BEGIN
    S:="";
    FOR i:=1 TO Strings.Length(R) DO
        ch:=R[i-1];
        IF CharPos(filtered,ch)=MAX(CARDINAL) THEN Strings.Append(ch,S);END; (* silly ADW reverse logic *)
    END;
    R:=S;
END filter;

(* darn, we no longer can process hex *)
PROCEDURE getvallc (S:ARRAY OF CHAR;VAR v : LONGCARD ):BOOLEAN;
VAR
    lc   : LONGCARD;
    ok   : LWholeStr.ConvResults;
BEGIN
    LWholeStr.StrToLongCard(S,lc,ok);
    IF lc > MAX(LONGCARD) THEN RETURN FALSE; END;
    v := lc;
    RETURN TRUE;
END getvallc;

PROCEDURE getval (S:ARRAY OF CHAR;VAR v : CARDINAL ):BOOLEAN;
VAR
    lc:LONGCARD;
BEGIN
    lc := VAL(LONGCARD,v);
    IF getvallc(S, lc) THEN
        IF lc > MAX(CARDINAL) THEN
            RETURN FALSE;
        ELSE
            v := VAL(CARDINAL,lc);
            RETURN TRUE;
        END;
    ELSE
        RETURN FALSE;
    END;
END getval;

(* darn, we can no longer process hex *)

PROCEDURE fmtlc (v:LONGCARD;wi:INTEGER;ch,prefix:CHAR) : str80;
VAR
    R : str80;
    ok: BOOLEAN;
    i,le,limit : CARDINAL;
BEGIN
    LWholeStr.LongCardToStr(v,R);
    le := Strings.Length(R) + 1;
    limit:=ABS(wi);
    FOR i:= le TO limit  DO
        IF wi < 0 THEN
            Strings.Append(R,ch);
        ELSE
            Prepend(R,ch);
        END;
    END;
    Prepend(R,prefix);
    RETURN R;
END fmtlc;

PROCEDURE fmt (v:CARDINAL;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str80;
BEGIN
    RETURN fmtlc( VAL(LONGCARD,v),wi,ch,prefix);
END fmt;

PROCEDURE beautifiedlc (v : LONGCARD;pad:CHAR; sep:CHAR; field:INTEGER) : str80;
VAR
    S,R   : str80;
    len,i : CARDINAL;
    ok  : BOOLEAN;
    ch  : CHAR;
    le,limit:CARDINAL;
BEGIN
    LWholeStr.LongCardToStr(v,S);
    len:=Strings.Length(S);
    R := "";
    FOR i := 1 TO len DO
        Prepend(R,S[len-i]);
        IF i < len THEN
            IF (i MOD 3) = 0 THEN
                Prepend(R,sep);
            END;
        END;
    END;
    LOOP
        le:=Strings.Length(R);
        limit:=ABS(field);
        IF le >= limit THEN EXIT; END;
        IF field < 0 THEN
            Strings.Append(R,pad);  (* right alignment *)
        ELSE
            Prepend(R,pad);
        END;
    END;
    RETURN R;
END beautifiedlc;

PROCEDURE beautified (v : CARDINAL;pad:CHAR; sep:CHAR; field:INTEGER) : str80;
BEGIN
    RETURN beautifiedlc ( VAL(LONGCARD,v),pad,sep,field);
END beautified;

PROCEDURE fmtme (v:LONGCARD;wi:CARDINAL; BEAUTIFY:BOOLEAN):str80;
VAR
    S:str80;
BEGIN
    IF BEAUTIFY THEN
        S:=beautifiedlc (v,blank,dot,wi);
    ELSE
        S:=fmtlc (v,wi,blank,"");
    END;
    RETURN S;
END fmtme;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    progTitle     = "Q&D Factorize";
    progVersion   = "v1.0b";
    progCopyright = "by PhG";
    Banner        = progTitle+" "+progVersion+" "+progCopyright;
    progEXEname   = "FACTORIZ";

(* ------------------------------------------------------------ *)

CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParameter    = 3;
    errLongCard     = 4;
    errTHSfailure   = 5;
    errAborted      = 6;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);

    PROCEDURE msg2 (VAR R : ARRAY OF CHAR;S1,S2:ARRAY OF CHAR);
    BEGIN
        Concat(R,S1,S2);
    END msg2;

    PROCEDURE msg3 (VAR R : ARRAY OF CHAR;S1,S2,S3:ARRAY OF CHAR);
    BEGIN
         msg2(R,S1,S2);Strings.Append(R,S3);
    END msg3;

CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msgHelp =
Banner+nl+
nl+
"Syntax : "+progEXEname+" <number>... [option]..."+nl+
nl+
"   -p   (very slowly) check <number> found to be prime with alternate method"+nl+
"   -l   show one factor per line"+nl+
"   -r   do not beautify factors"+nl+
"   -ths try and build TxHxS from <number>"+nl+
"   -e   do not poll keyboard for Escape"+nl+
nl+
"This program (slowly) factorizes specified <number>."+nl+
"<number> must belong to [0..~] range."+nl+
"["+filteredChars+"] characters are filtered out from <number>."+nl+
nl+
"Examples : "+progEXEname+" 16,481,808 20,016,864"+nl+
"           "+progEXEname+" -ths 160,086,528"+nl+
"           "+progEXEname+" -p 01.02.1963"+nl;

VAR
    S : str128;
    H : str2048;
BEGIN
    CASE e OF
    | errHelp :       H := msgHelp;
                      Subst(H,"~",fmtlc( MAX(LONGCARD),1,"","" ));
                      WrStr(H);
    | errOption :     msg3(S,'Unknown "',einfo,'" option !');
    | errParameter :  msg3(S,'Unexpected "',einfo,'" parameter !');
    | errLongCard :   msg3(S,'Illegal "',einfo,'" number !');
    | errTHSfailure : msg2(S,einfo," would not translate to valid TxHxS !");
    | errAborted :    S := "Aborted by user !"
    ELSE
                      S := "How did you get THERE ???";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        WrStr(progEXEname+" : ");WrStr(S);WrLn;
    END;

    WIN32.ExitProcess ( VAL(WIN32.UINT, e) );
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE trydiv (VAR q : LONGCARD ;v,d:LONGCARD; DEBUG:BOOLEAN):BOOLEAN ;
BEGIN
    q := v DIV d;
IF DEBUG THEN
WrStr("value = ");      WrStr(fmtlc(v,winum,blank,""));
WrStr("  divisor = ");  WrStr(fmtlc(d,winum,blank,""));
WrStr("  quotient = "); WrStr(fmtlc(q,winum,blank,""));WrLn;
END;
    RETURN ( v = (q * d) );
END trydiv;

CONST
    minfactentry = 0;
    maxfactentry = 0;
    minfactor    = 1;    (* so we can flag using minfactor-1=0 *)
    maxfactor    = 32+1; (* should do for at most 2^32 ! *)
    ABORTED      = MAX(CARDINAL);
TYPE
    factortype = RECORD
       k    : LONGCARD;
       flag : BOOLEAN;
    END;
VAR
    factor : ARRAY[minfactentry..maxfactentry] , [minfactor..maxfactor] OF factortype;

PROCEDURE factorize (where:CARDINAL; v:LONGCARD;POLLESC, DEBUG:BOOLEAN):CARDINAL;
VAR
    i,ndx:CARDINAL;
    quotient, divisor,added : LONGCARD;
BEGIN
    ndx:=minfactor-1;
    (* handle trivial cases *)
    IF v = 0 THEN RETURN ndx; END;
    IF v = 1 THEN
        INC(ndx);
        factor[where][ndx].k:=v;
        factor[where][ndx].flag:=TRUE;
        RETURN ndx;
    END;

    FOR i:=1 TO 2 DO
        CASE i OF
        | 1: divisor := 2;
        | 2: divisor := 3;
        END;
        LOOP
            IF trydiv (quotient, v,divisor,DEBUG) THEN
                INC(ndx);
                IF ndx > maxfactor THEN RETURN minfactor-1; END; (* should never happen *)
                factor[where][ndx].k:=divisor;
                factor[where][ndx].flag:=TRUE;
                v:=quotient;
                IF quotient < divisor THEN RETURN ndx; END;
            ELSE
                EXIT;
            END;
            IF POLLESC THEN
                IF (divisor MOD kEsc) = 0 THEN
                    IF ChkEscape() THEN RETURN ABORTED; END;
                END;
            END;
        END;
    END;

    (*
    after 2 and 3, gen next prime (5, 7, 11, 13, 17, 19, 23, 25, 29, 31...)
    by adding alternatively 2 then 4 to previous value
    *)

    added := 2;
    INC(divisor,added); (* 3+2=5 *)
    LOOP
        IF trydiv (quotient,v,divisor,DEBUG) THEN
            INC(ndx);
            IF ndx  > maxfactor THEN RETURN minfactor-1; END;
            factor[where][ndx].k   :=divisor;
            factor[where][ndx].flag:=TRUE;
            v:=quotient;
            IF quotient < divisor THEN RETURN ndx; END;
        ELSE
            INC(divisor,added);
            IF added=2 THEN
                added:=4;
            ELSE (* is 4 *)
                added:=2;
            END;
        END;
        IF POLLESC THEN
            IF (divisor MOD kEsc)=0 THEN
                 IF ChkEscape() THEN ndx:=ABORTED;EXIT; END;
            END;
        END;
    END;

    RETURN ndx;
END factorize;

(* ------------------------------------------------------------ *)

PROCEDURE showfactors (what,ndx:CARDINAL; num:LONGCARD;ONEPERLINE,BEAUTIFY:BOOLEAN);
CONST
    equals = " = ";
VAR
    S:str80;
    v:LONGCARD;
    i,wi:CARDINAL;
BEGIN
    IF ONEPERLINE THEN
        wi:= winum;
    ELSE
        wi:= 1;
    END;
    S:=fmtme(num,wi,BEAUTIFY);
    WrStr(S);
    IF ndx = minfactor THEN
        WrStr(" is prime.");WrLn; RETURN;
    ELSE
        WrStr(equals);
        IF ONEPERLINE THEN WrLn;END;
    END;
    FOR i:=minfactor TO ndx DO
        v:=factor[what][i].k;
        S:=fmtme(v,wi,BEAUTIFY);
        WrStr(S);
        IF ONEPERLINE THEN
            IF i < ndx THEN WrLn; END;
        ELSE
            IF i < ndx THEN WrStr(" x ");END;
        END;
    END;
    WrLn;
END showfactors;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE chkPrime (num:LONGCARD; POLLESC,DEBUG:BOOLEAN):CARDINAL;
VAR
    n  : LONGCARD;
    rc : CARDINAL;
BEGIN
    rc:= minfactor;
    IF num=2 THEN RETURN rc; END; (* 2 is prime *)
    video(msgWait,TRUE);
    n:=2; (* BFI : not the fastest way but it works *)
    LOOP
        IF ( num MOD n ) = 0 THEN rc:=maxfactor; EXIT; END;
        INC(n);
        IF n >= num THEN EXIT; END;
        IF POLLESC THEN
            IF (n MOD kEsc) = 0 THEN
                IF ChkEscape() THEN rc:=ABORTED;EXIT; END;
            END;
        END;
    END;
    video(msgWait,FALSE);
    RETURN rc;
END chkPrime;

PROCEDURE showPrimeStatus (ndx:CARDINAL;num:LONGCARD;ONEPERLINE,BEAUTIFY:BOOLEAN);
VAR
    S : str80;
    wi: CARDINAL;
BEGIN
    IF ONEPERLINE THEN
        wi:= winum;
    ELSE
        wi:= 1;
    END;
    S:=fmtme(num,wi,BEAUTIFY);
    WrStr(S);
    IF ndx = minfactor THEN
        S:=" is prime indeed. :-)";
    ELSE
        S:=" does not look so prime at second sight. :-(";
    END;
    WrStr(S);WrLn;
END showPrimeStatus;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE findinfact (where,ndx:CARDINAL; limit:LONGCARD):LONGCARD;
VAR
    v, k: LONGCARD;
    i : CARDINAL;
BEGIN
    (*
    (* lazy eval : we don't find highest value here *)
    v := 1;
    i := minfactor-1;
    LOOP
        INC(i);
        IF i > ndx THEN EXIT; END;
        IF factor[where][i].flag THEN
            k:=factor[where][i].k;
            IF v * k <= limit THEN
                factor[where][i].flag := FALSE; (* no longer useable *)
                v := v * k;
            ELSE
                EXIT;
            END;
        END;
    END;
    *)
    v := 1;
    i := ndx+1;
    LOOP
        DEC(i);
        IF i < minfactor THEN EXIT; END;
        IF factor[where][i].flag THEN
            k:=factor[where][i].k;
            IF v * k <= limit THEN
                factor[where][i].flag := FALSE; (* no longer useable *)
                v := v * k;
            END;
        END;
    END;
    RETURN v;
END findinfact;

CONST
    mintrack = 0;
    minhead  = 0;
    minsector= 1;
    FORCEBEST= 0;
    FORCE255 = 255;
    FORCE240 = 240;

PROCEDURE buildTHSfromTOTAL (VAR maxTrack,maxHead,maxSector:CARDINAL;
                            BESTFIT:CARDINAL;totalSectors:LONGCARD;
                            POLLESC,DEBUG:BOOLEAN):BOOLEAN;
CONST
    maxsectorcount    = 63;
    maxsectorcountalt = 255; (* desperate try to keep CARDINALs everywhere *)
    maxheadcount      = 255;
    maxtrackcount     = MAX(LONGCARD);
VAR
    rebuilt,v,quotient:LONGCARD;
    t,h,s,smax:LONGCARD;
    headcount,trackcount,sectorcount:CARDINAL;
    retry,ndx,ndxref:CARDINAL;
    rc:BOOLEAN;
BEGIN
    rc:=TRUE;
    retry := 1;
    LOOP
        v:=totalSectors;
        CASE retry OF
        | 1 :
            s:=maxsectorcount; (* try 63 first *)
            IF trydiv(quotient, v,s,DEBUG) THEN
                v:=quotient;
            ELSE
                s:=0;
            END;
            smax := maxsectorcount;
        | 2 :
            s:=0;
            smax := maxsectorcountalt;
        END;
        (*
        FOR now, don't care about BESTFIT which is a BAD idea here
        CASE BESTFIT OF (* should always be FORCEBEST *)
        | FORCEBEST : h:=16;
        | FORCE255:   h:=255;
        | FORCE240:   h:=240;
        ELSE
                      h:=16; (* safety *)
        END;
        *)
        video(msgWait,TRUE);
        ndx:=factorize(minfactentry,v,POLLESC,DEBUG);
        video(msgWait,FALSE);
        IF ndx=ABORTED THEN
            maxTrack := ABORTED;
            maxHead  := ABORTED;
            maxSector:= ABORTED;
            RETURN FALSE;
        END;

        IF s = 0 THEN s:=findinfact(minfactentry,ndx,smax);END;
        h:=findinfact(minfactentry,ndx,maxheadcount);
        t:=findinfact(minfactentry,ndx,maxtrackcount);

        trackcount  := VAL(CARDINAL,t);
        headcount   := VAL(CARDINAL,h);
        sectorcount := VAL(CARDINAL,s);

        rebuilt     := VAL(LONGCARD,trackcount) * VAL(LONGCARD,headcount) * VAL(LONGCARD,sectorcount); (* safety *)
        IF rebuilt = totalSectors THEN EXIT; END;
        INC(retry);
        IF retry > 2 THEN rc:=FALSE; EXIT; END;
    END;
    maxTrack    := trackcount + mintrack -1;
    maxHead     := headcount  + minhead  -1;
    maxSector   := sectorcount+ minsector-1;
    RETURN rc;
END buildTHSfromTOTAL;

PROCEDURE showTHS (t,h,s:CARDINAL; v:LONGCARD; BEAUTIFY:BOOLEAN );
VAR
    R : str128;
    wi:CARDINAL;
BEGIN
    R := "~ = ~ x ~ x ~";
    wi:=1;
    Subst(R,"~",fmtme(v,wi,BEAUTIFY));
    Subst(R,"~",fmtme( VAL(LONGCARD,t-mintrack+1 ),wi,BEAUTIFY));
    Subst(R,"~",fmtme( VAL(LONGCARD,h-minhead+1  ),wi,BEAUTIFY));
    Subst(R,"~",fmtme( VAL(LONGCARD,s-minsector+1),wi,BEAUTIFY));
    WrStr(R);WrLn;
END showTHS;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    firstparm = 1;
    maxparm   = 10;
VAR
    ONEPERLINE,BEAUTIFY,THSMODE,PARANOIA,POLLESC,DEBUG : BOOLEAN;
    lastparm,parmcount,i,opt:CARDINAL;
    S,R:str128;
    parm : ARRAY[firstparm..maxparm] OF str128;
    v:LONGCARD;
    ndx:CARDINAL;
    maxTrack,maxHead,maxSector:CARDINAL;
    cli:str1024;
BEGIN
    WrLn;

    Environment.GetCommandLine(cli);
    parmcount := argc(cli);
    IF parmcount=0 THEN abort(errHelp,"");END;

    PARANOIA   := FALSE;
    ONEPERLINE := FALSE;
    BEAUTIFY   := TRUE;
    POLLESC    := TRUE;
    THSMODE    := FALSE;
    DEBUG      := FALSE;
    lastparm   := firstparm-1;

    FOR i := 1 TO parmcount DO
        argv(S,cli,i); cleantabs(S);
        R:=S;
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "L"+delim+"ONEPERLINE"+delim+
                                  "R"+delim+"RAW"+delim+
                                  "THS"+delim+
                                  "P"+delim+"PRIME"+delim+"PARANOIA"+delim+
                                  "E"+delim+"ESC"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3 :   abort(errHelp,"");
            | 4,5 :     ONEPERLINE := TRUE;
            | 6,7 :     BEAUTIFY   := FALSE;
            | 8:        THSMODE    := TRUE;
            | 9,10,11 : PARANOIA  := TRUE;
            | 12,13 :   POLLESC    := FALSE;
            | 14 :      DEBUG      := TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParameter,S);END;
            parm[lastparm] := R;
        END;
    END;

    FOR i:= firstparm TO lastparm DO
        S := parm[i];
        filter(S,filteredChars);
        IF getvallc (S, v) THEN
            IF THSMODE THEN
                IF buildTHSfromTOTAL (maxTrack,maxHead,maxSector,
                                     FORCEBEST,v,POLLESC,DEBUG) THEN
                    showTHS(maxTrack,maxHead,maxSector,v,BEAUTIFY);
                ELSE
                    ndx:=0;
                    IF maxTrack= ABORTED THEN INC(ndx);END;
                    IF maxHead=  ABORTED THEN INC(ndx);END;
                    IF maxSector=ABORTED THEN INC(ndx);END;
                    IF ndx=3 THEN abort(errAborted,"");END;
                    abort(errTHSfailure,S);
                END;
            ELSE
                video(msgWait,TRUE);
                ndx:=factorize(minfactentry,v,POLLESC,DEBUG);
                video(msgWait,FALSE);
                IF ndx=ABORTED THEN abort(errAborted,"");END;
                showfactors(minfactentry,ndx,v,ONEPERLINE,BEAUTIFY);
                IF PARANOIA THEN
                    IF ndx=minfactor THEN
                        ndx:=chkPrime(v,POLLESC,DEBUG);
                        IF ndx=ABORTED THEN abort(errAborted,"");END;
                        showPrimeStatus(ndx,v,ONEPERLINE,BEAUTIFY);
                    END;
                END;
            END;
        ELSE
            abort(errLongCard,S);
        END;
        IF ONEPERLINE THEN
            IF i < lastparm THEN WrLn;END;
        END;
    END;

    abort(errNone,"");
END Fact_ADW.
