(* ---------------------------------------------------------------
Title         Q&D Base 64
Overview
Usage
Notes
Bugs
Wish List     save date/time stamp ? bah...

--------------------------------------------------------------- *)

MODULE base64;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;
IMPORT IO;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    progEXEname   = "BASE64";
    progTitle     = "Q&D Base 64";
    progVersion   = "v1.0a";
    progCopyright = "by PhG";
    progBanner    = progTitle+" "+progVersion+" "+progCopyright;
CONST
    errNone            = 0;
    errHelp            = 1;
    errExclusive       = 2;
    errUnknownOption   = 3;
    errTooManyParms    = 4;
    errExpecting       = 5;
    errNoMatch         = 6;
    errMissingCommand  = 7;
    errJoker           = 8;
    errTargetRO        = 9;
    errIsDir           = 10;
    errTargetHere      = 11;
    errWidth           = 12;
    errNonsenseDecode  = 13;
    errByFour          = 14;
    errOp              = 15;
    errNotMultipleOfFour=16;

PROCEDURE abort(e:CARDINAL;einfo:ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)

errmsg =
progBanner+nl+
nl+
"Syntax 1 : "+progEXEname+" <-e> <original> [MIME_target] [option]..."+nl+
"Syntax 2 : "+progEXEname+" <-d> <MIME_source> <original> [option]..."+nl+
nl+
"This program either converts a file to MIME BASE64 format (syntax 1),"+nl+
"or converts a MIME BASE64 file back to its original binary (syntax 2)."+nl+
nl+
"  -e[e] encode (-ee = -e -o)"+nl+
"  -d[d} decode (-dd = -d -o)"+nl+
"  -w:#  line width ([4..256], default is 60)"+nl+
"  -o[o] overwrite existing target file (-oo = -o -r)"+nl+
"  -r    ignore existing read-only attribute when overwriting or erasing a file"+nl+
"  -q    quiet (no message on operation completion)"+nl+
"  -i    include optional header (-e[e] only)"+nl+
"  -n|-m no final CRLF (-e[e] only)"+nl+
"  -x    disable LFN support even if available"+nl+
nl+
"a) With syntax 1, should MIME_target be omitted, output will go to screen."+nl+
"b) Jokers are not supported : this is by design."+nl+
"c) -ii = -nn = -mm = -i -n"+nl+
nl+
"Examples : "+progEXEname+" -w:60 -e -o archive.zip archive.64"+nl+
"           "+progEXEname+" -d archive.64 archive.zip"+nl;

VAR
    S : str1024; (* oversized for possible LFN *)
BEGIN
    CASE e OF
    | errHelp :            WrStr(errmsg);
    | errExclusive :       S := "-encode and -decode options are mutually exclusive !";
    | errUnknownOption:    S := 'Unknown "~" option !';
    | errTooManyParms:     S := 'Useless "~"  parameter !';
    | errExpecting:        S := 'Expecting "~" !';
    | errNoMatch:          S := 'No file matches "~" specification !';
    | errMissingCommand:   S := "Missing -encode or -decode command !";
    | errJoker:            S := '"~" should not contain any joker !';
    | errTargetRO:         S := '"~" target is read-only !';
    | errIsDir:            S := '"~" cannot be a directory !';
    | errTargetHere:       S := '"~" target already exists !';
    | errWidth:            S := 'Illegal -w:# value or out of "~" range !';
    | errNonsenseDecode:   S := '"~" option is a nonsense with -decode option !';
    | errByFour:           S := '"~" is not a multiple of 4 !';
    | errOp:               S := "Any of -encode or -decode options is required !";
    | errNotMultipleOfFour:S := "Length of at least one input line is not a multiple of 4 !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (16 * 512) + FIO.BufferOverhead; (* was 8 *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn  : ioBufferType;
    ioBufferOut : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
     strmimeheader = "Content-Type: text/plain; charset=US-ASCII; name=|"+nl+
                     "Content-transfer-encoding: base64"+nl;

CONST
     (*
     $41..$5a
     $61..$7a
     $30..$39
     $2b $2f
     $3d
     *)
     tochar64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"+
                "abcdefghijklmnopqrstuvwxyz"+
                "0123456789"+
                "+/"+
                "="; (* 65th i.e. index 64 *)
     minndx64 = 1-1;
     maxndx64 = 64-1;
     ndx64EOF = maxndx64+1;
TYPE
    triplettype = RECORD
        b : ARRAY[0..2] OF SHORTCARD;
    END;
    quadtype = RECORD
        c : ARRAY[0..3] OF CHAR;
    END;

(*
    aaaaaabb
            bbbbcccc
                    ccdddddd
*)

PROCEDURE procEncode (useLFN,btofile,bheader,bcrlf:BOOLEAN;wiline:CARDINAL;
                     source,target:pathtype):CARDINAL;
CONST
    wanted = SIZE(triplettype);
VAR
    hin,hout:FIO.File;
    got,currwi,i:CARDINAL;
    t:triplettype;
    q:quadtype;
    done:BOOLEAN;
    ndx:SHORTCARD;
    S:str256;
BEGIN
    hin:=fileOpenRead(useLFN,source);
    FIO.AssignBuffer(hin,ioBufferIn);
    IF btofile THEN
        hout:=fileCreate(useLFN,target);
        FIO.AssignBuffer(hout,ioBufferOut);
    ELSE
        hout:=FIO.StandardOutput;
    END;

    IF bheader THEN
        S:= strmimeheader;
        Str.Subst(S,"|",source);
        IF bcrlf THEN Str.Prepend(S,nl);END;
        FIO.WrStr(hout,S); FIO.WrLn(hout);
    END;

    FIO.EOF := FALSE;
    done:=FALSE;
    currwi:=0;
    LOOP
        got:=FIO.RdBin(hin,t,wanted);
        IF got = 0 THEN EXIT; END;
        IF got # wanted THEN        (* 1 OR 2 hence zero 1..2 or 2..2 *)
            FOR i:= got TO wanted-1 DO t.b[i]:=0;END;
            done:=TRUE;
        END;

        ndx := (t.b[0] >> 2);
        q.c[0]:= tochar64 [ CARDINAL(ndx) ];

        ndx := ((t.b[0] AND 003H) << 4) OR ((t.b[1] AND 0F0H) >> 4);
        q.c[1]:= tochar64 [ CARDINAL(ndx) ];

        IF got < 2 THEN    (* got 1 byte *)
            ndx := ndx64EOF;
        ELSE
            ndx := ((t.b[1] AND 00FH) << 2) OR ((t.b[2] AND 0C0H) >> 6);
        END;
        q.c[2]:= tochar64 [ CARDINAL(ndx) ];

        IF got < 3 THEN    (* got 1 or 2 bytes *)
            ndx := ndx64EOF;
        ELSE
            ndx := (t.b[2] AND 03FH);
        END;
        q.c[3]:= tochar64 [ CARDINAL(ndx) ];

        FOR i:= 0 TO 3 DO FIO.WrChar(hout,q.c[i]);END;

        INC(currwi,4);
        IF (currwi MOD wiline) = 0 THEN FIO.WrLn(hout);currwi:=0;END;
        IF done THEN EXIT; END;
    END;

    FIO.WrLn(hout);
    IF bcrlf THEN FIO.WrLn(hout); END; (* //V10A beautify *)

    IF btofile THEN fileClose(useLFN,hout); END;
    fileClose(useLFN,hin);

    RETURN errNone;
END procEncode;

(*
    aaaaaa
          bbbbbb
                cccccc
                      dddddd
    aaaaaabb
            bbbbcccc
                    ccdddddd
*)

PROCEDURE procDecode (useLFN:BOOLEAN;source,target:pathtype):CARDINAL;
VAR
    hin,hout:FIO.File;
    S:str128;
    len,i,j,count,ndx:CARDINAL;
    pb:BOOLEAN;
    t:triplettype;
    q:quadtype;
    toindex:ARRAY[0..255] OF SHORTCARD;
    ch:CHAR;
BEGIN
    FOR i:=0 TO 255 DO
        toindex[i]:=minndx64;
    END;
    FOR i:= minndx64 TO maxndx64 DO
        toindex[ORD(tochar64[i])]:=SHORTCARD(i);
    END;

    hin :=fileOpenRead(useLFN,source);
    FIO.AssignBuffer(hin,ioBufferIn);
    hout :=fileCreate(useLFN,target);
    FIO.AssignBuffer(hout,ioBufferOut);

    FIO.EOF:=FALSE; (* useless safety *)
    pb:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT;END;
        FIO.RdStr(hin,S);
        IF (S[0]=charnull) AND FIO.EOF THEN EXIT;END;
        LtrimBlanks(S);
        RtrimBlanks(S);
        IF Str.CharPos(S,colon)#MAX(CARDINAL) THEN Str.Copy(S,"");END; (* ignore header *)
        len:=Str.Length(S);
        IF (len MOD 4) # 0 THEN pb:=TRUE; EXIT;END; (* should never happen *)
        IF len # 0 THEN
            DEC(len); (* len is now lastpos *)
            i:=0;
            LOOP
                IF i > len THEN EXIT; END;
                count:=3;
                FOR j:= 0 TO 3 DO
                    ch:=S[i+j];
                    IF ch = tochar64[ndx64EOF] THEN DEC(count);END;
                    q.c[j]:=CHR(toindex[ ORD(ch) ]); (* back to real index *)
                END;
                ndx := (ORD(q.c[0]) << 2) OR (ORD(q.c[1]) >> 4);
                t.b[0]:=SHORTCARD(ndx);

                ndx := (ORD(q.c[1]) << 4) OR (ORD(q.c[2]) >> 2);
                t.b[1]:=SHORTCARD(ndx);

                ndx := (ORD(q.c[2]) << 6) OR (ORD(q.c[3]));
                t.b[2]:=SHORTCARD(ndx);

                FIO.WrBin(hout,t,count);
                INC(i,4);
            END;
        END;
    END;

    fileClose(useLFN,hout);
    fileClose(useLFN,hin);

    IF pb THEN
        i:=errNotMultipleOfFour;
    ELSE
        i:=errNone;
    END;
    RETURN i;
END procDecode;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE msg2 (quoted:BOOLEAN;S1,S2:ARRAY OF CHAR );
CONST
    wi = 10;
VAR
    i:CARDINAL;
BEGIN
    WrStr(S1);
    FOR i:=Str.Length(S1)+1 TO wi DO WrStr(" ");END;WrStr(" : ");
    IF quoted THEN WrStr(dquote);END;
    WrStr(S2);
    IF quoted THEN WrStr(dquote);END;
    WrLn;
END msg2;

PROCEDURE num2str (v,base:CARDINAL;wipad:INTEGER;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(LONGCARD(v),R,base,ok);
    IF ok THEN
        FOR i:=Str.Length(R)+1 TO ABS(wipad) DO
            IF wipad < 0 THEN
                Str.Append(R,padchar);
            ELSE
                Str.Prepend(R,padchar);
            END;
        END;
    ELSE
        R:="*** ERROR ***";
    END;
    RETURN R;
END num2str;

PROCEDURE grabvalue (VAR v:CARDINAL;
                    mini,maxi:CARDINAL;S:ARRAY OF CHAR):BOOLEAN;
VAR
    lc:LONGCARD;
    ok:BOOLEAN;
BEGIN
    ok:=GetLongCard(S,lc);
    IF ok THEN
        ok:=NOT ( lc > LONGCARD(maxi) );
        IF ok THEN
            ok:= ( lc >= LONGCARD(mini) );
            IF ok THEN v:=CARDINAL(lc);END;
        END;
    END;
    RETURN ok;
END grabvalue;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    opNone      = 0;
    opEncode    = 1;
    opDecode    = 2;
    minWidth    = 4;
    maxWidth    = 256; (* should be 72 *)
    defWidth    = 60;
    sWidthRange = "[4..256]";
CONST
    firstparm = 1;
    maxparm   = 2;
VAR
    parm : ARRAY[firstparm..maxparm] OF pathtype;
    source,target,S,R: pathtype;
    lastparm,parmcount,i,got,opt : CARDINAL;
    useLFN,DEBUG:BOOLEAN;
    op,wiline : CARDINAL;
    verbose:BOOLEAN;
    flagTOFILE,flagOVERWRITE,flagOVERWRITERO,flagHEADER,flagCRLF:BOOLEAN;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    FIO.ShareMode := FIO.ShareDenyNone; (* very, very important ! *)
    WrLn;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    DEBUG           := FALSE;
    useLFN          := TRUE;
    op              := opNone;
    wiline          := MAX(CARDINAL);
    flagOVERWRITE   := FALSE;
    flagOVERWRITERO := FALSE;
    flagTOFILE      := TRUE;
    verbose         := TRUE;
    flagHEADER      := FALSE;
    flagCRLF        := TRUE;

    lastparm := firstparm-1;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "E"+delim+"ENCODE"+delim+
                                  "D"+delim+"DECODE"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "DEBUG"+delim+
                                  "W:"+delim+"WIDTH:"+delim+
                                  "L:"+delim+"LINE:"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "R"+delim+"READONLY"+delim+
                                  "OO"+delim+
                                  "EE"+delim+
                                  "DD"+delim+
                                  "Q"+delim+"QUIET"+delim+
                                  "I"+delim+"HEADER"+delim+
                                  "N"+delim+"CRLF"+delim+"NL"+delim+"M"+delim+"MIME"+delim+
                                  "II"+delim+"NN"+delim+"MM"
                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5   : IF op=opDecode THEN abort(errExclusive,"");END;
                      op     := opEncode;
            | 6,7   : IF op=opEncode THEN abort(errExclusive,"");END;
                      op     := opDecode;
            | 8,9   : useLFN := FALSE;
            | 10    : DEBUG  := TRUE;
            | 11,12,13,14 : IF grabvalue(wiline,minWidth,maxWidth,R)=FALSE THEN abort(errWidth,sWidthRange);END;
                      IF (wiline MOD 4) # 0 THEN abort(errByFour,R);END;
            | 15,16 : flagOVERWRITE   := TRUE;
            | 17,186 : flagOVERWRITERO := TRUE;
            | 19:     flagOVERWRITE   := TRUE;
                      flagOVERWRITERO := TRUE;
            | 20 :    IF op=opDecode THEN abort(errExclusive,"");END;
                      op     := opEncode;
                      flagOVERWRITE := TRUE;
            | 21 :    IF op=opEncode THEN abort(errExclusive,"");END;
                      op     := opDecode;
                      flagOVERWRITE := TRUE;
            | 22,23:  verbose:=FALSE;
            | 24,25:  flagHEADER:=TRUE;
            | 26,27,28,29,30:flagCRLF:=FALSE;
            | 31,32,33:flagHEADER:=TRUE;
                      flagCRLF:=FALSE;
            ELSE
                abort(errUnknownOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errTooManyParms,S);END;
            parm[lastparm]:=S;
        END;
    END;
    CASE op OF
    | opNone :
        abort(errOp,"");
    | opDecode:
        IF wiline # MAX(CARDINAL) THEN abort(errNonsenseDecode,"-w:#");END;
        IF flagHEADER THEN abort(errNonsenseDecode,"-i");END;
        IF flagCRLF=FALSE THEN abort(errNonsenseDecode,"-n");END;
    | opEncode:
        IF wiline = MAX(CARDINAL) THEN wiline:=defWidth;END;
    END;

    CASE lastparm OF
    | firstparm-1:
        abort(errHelp,"");
    | firstparm:
        CASE op OF
        | opEncode : source:=parm[firstparm];
                     target:=""; (* useless safety *)
                     flagTOFILE := FALSE;
        | opDecode : abort(errExpecting,"<original>");
        END;
    | maxparm:
        source:=parm[firstparm];
        target:=parm[firstparm+1];
    END;

    useLFN:=( useLFN AND fileSupportLFN() );

    IF DEBUG THEN
        CASE op OF
        | opEncode:S:="encode";
        | opDecode:S:="decode";
        END;
        msg2(FALSE,"Operation",S);
        msg2(TRUE,"Source",source);
        IF flagTOFILE THEN
            S:=target;
        ELSE
            S:="video standard output";
        END;
        msg2( flagTOFILE,"Target",S);
        IF op = opEncode THEN
            msg2(FALSE,"Line width",num2str(wiline,10,1,""));
        END;
    END;

        IF same(source,dot) THEN source:=stardotstar;END;
        IF chkJoker(source) THEN abort(errJoker,source);END;
        IF fileIsDirectorySpec(useLFN,source) THEN abort(errIsDir,source);END;
        IF fileExists(useLFN,source)=FALSE THEN abort(errNoMatch,source);END;

    IF flagTOFILE THEN
        IF same(target,dot) THEN target:=stardotstar;END;
        IF chkJoker(target) THEN abort(errJoker,target);END;
        IF fileIsDirectorySpec(useLFN,target) THEN abort(errIsDir,target);END;

        IF fileExists(useLFN,target) THEN
            IF flagOVERWRITE THEN
                IF fileIsRO(useLFN,target) THEN
                    IF flagOVERWRITERO THEN
                        fileSetRW(useLFN,target);
                        fileErase(useLFN,target);
                    ELSE
                        abort(errTargetRO,target);
                    END;
                ELSE
                    fileErase(useLFN,target);
                END;
            ELSE
                abort(errTargetHere,target);
            END;
        END;
    END;

    S:="";
    CASE op OF
    | opEncode: i:=procEncode (useLFN,flagTOFILE,flagHEADER,flagCRLF,
                              wiline,source,target);
        IF i = errNone THEN
            IF flagTOFILE THEN
                S:='+++ "|" encoded to "|"';
            END;
        END;
    | opDecode: i:=procDecode(useLFN,source,target);
        S:='+++ "|" decoded to "|"';
    END;

    IF NOT(verbose) THEN S:="";END;

    IF same(S,"")=FALSE THEN
        Str.Subst(S,"|",source);
        Str.Subst(S,"|",target);
        WrStr(S);WrLn;
    END;

    abort(i,"");
END base64.


