(* ---------------------------------------------------------------
Title         Q&D Base Converter
Overview
Notes         with p4 2.6MHz, 5 hours to check 0..60000000 in shadok !
Bugs
Wish List

--------------------------------------------------------------- *)

MODULE BaseConv;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

(* ------------------------------------------------------------ *)

TYPE
    str4          = ARRAY[0..3] OF CHAR; (* str3 was too short FOR "Meu," ! *)
CONST
    ProgEXEname   = "BASECONV";
    ProgTitle     = "Q&D Base Converter";
    ProgVersion   = "v1.0g";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    extBANG       = ".!!!";
    FORCESHADOKFILE= ProgEXEname+extBANG;
CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errTooManyNumbers=3;
    errNoNumber     = 4;
    errBadNumber    = 5;
    errBadLC        = 6;
    errFmt          = 7;
    errAborted      = 255;

(* ------------------------------------------------------------ *)

CONST
    foundNowhere   = 0;
    foundInCurrdir = 1;
    foundInEXEdir  = 2;

PROCEDURE foundForceShadokFile (VAR where:CARDINAL;
                               F:ARRAY OF CHAR):BOOLEAN;
VAR
    ok:BOOLEAN;
    R:str128;
    p:CARDINAL;
BEGIN
    Lib.ParamStr(R,0);
    p:=Str.RCharPos(R,antislash); (* yes, we could just check extension *)
    IF p # MAX(CARDINAL) THEN R[p]:=charnull;END;
    fixDirectory(R);
    Str.Append(R,F);
    where:=foundInEXEdir;
    ok:=FIO.Exists(R);
    IF NOT(ok) THEN
        where:=foundInCurrdir;
        ok:=FIO.Exists(F);
        IF NOT(ok) THEN where:=foundNowhere;END;
    END;
    RETURN ok;
END foundForceShadokFile;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" <[$|%|0x|_|!]number>... [option]..."+nl+
nl+
"This program displays LONGCARD numbers in bases 10, 16, 2, 8 and 4"+nl+
"(decimal, hexadecimal, binary, octal, Shadok numeration)."+nl+
nl+
"    -a[a]    dump each base value on its own line (-aa = -a -o)"+nl+
"    -o|-b    show octal (-a forced)"+nl+
"    -4[@]|-! show base 4 (-a forced)"+nl+
"    -@[4]    show Shadok numeration (-a forced)"+nl+
"    -x       shortcut for -o -4 -@"+nl+
"    -c       show DOS ASCII character"+nl+
"    -f[r]    increase values readability (French-style)"+nl+
"    -u[s]    increase values readability (English-style)"+nl+
"    -$|--|-/ show only decimal and Shadok (-a forced)"+nl+
"    -i       show maximum values"+nl+
nl+
"a) -i option ignores any parameter."+nl+
"b) -c option is ignored if number is out of [0..255] ([$00..$ff]) range."+nl+
'c) Inner "'+dot+'" and "'+coma+'" separators are ignored.'+nl+
'd) Although octal "_" and base 4 "!" prefixes are handled,'+nl+
"   octal and base 4 display still require -o or -4 options."+nl+
"e) Shadok numeration does not use any prefix :"+nl+
"   it is base 4 where 0 = Ga, 1 = Bu, 2 = Zo, 3 = Meu."+nl+
"   Alternate forms are 0 = G, 1 = B, 2 = Z, 3 = M."+nl+
"   Normal and alternate forms should not be mixed."+nl+
'f) -$ option is forced if "'+FORCESHADOKFILE+'" file exists'+nl+
"   in executable directory or in current directory."+nl+
"   Please note this file ^ found^."+nl+ (* //FIXME *)
nl+
"Examples : "+ProgEXEname+" %00.00.00.01 0x2 1.963 $13 _55 -o"+nl+
"           "+ProgEXEname+" -4 !3210"+nl+
"           "+ProgEXEname+" -@ MeuZoBuGa MZBG Meu.Zo.Bu.Ga 228"+nl+
"           "+ProgEXEname+" -@ -4 -f MeuZoBuGa 19630201"+nl+
"           "+ProgEXEname+" -$ meuzobuga 228"+nl;

VAR
    S : str256;
    Z : str2048; (* str1024 was too short *)
    rc: CARDINAL;
BEGIN
    CASE e OF
    | errHelp :
        Str.Copy(Z,helpmsg);
        IF foundForceShadokFile(rc,FORCESHADOKFILE) THEN
            Str.Subst(Z,"^","WAS");
            CASE rc OF
            | foundInCurrdir: Str.Subst(Z,"^"," in current directory");
            | foundInEXEdir : Str.Subst(Z,"^"," in executable directory");
            END;
        ELSE
            Str.Subst(Z,"^","was NOT");
            Str.Subst(Z,"^","");
        END;
        WrStr(Z);
    | errOption :
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," option !");
    | errTooManyNumbers:
        S := "Too many numbers !";
    | errNoNumber:
        S := "No number was specified !";
    | errBadNumber:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" value !');
    | errBadLC:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" LONGCARD value !');
    | errFmt:
        S := "-f[r] and -u[s] options are mutually exclusive !";
    | errAborted:
        S := "Aborted by user !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp : ;
    ELSE
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    gremlin   = ".";
    hexprefix = "$"; (* hex *)
    binprefix = "%"; (* bin *)
    octprefix = "_"; (* octal *)
    quadprefix= "!"; (* base 4 *)
CONST
    bindigits = "01";
    decdigits = "0123456789";
    hexdigits = decdigits+"ABCDEF"; (* uppercase *)
    octdigits = "01234567";
    quaddigits= "0123";
CONST
    (* nice luck with letters here ! *)
    numShadokLong  = "Ga,Bu,Zo,Meu";
    numShadokShort = "G,B,Z,M";
    numShadok      = numShadokLong+','+numShadokShort;

CONST
    (* $ffffffff = 4294967295 = %11111111111111111111111111111111 *)
    wibin        = 32; (* $ffffffff x 4 *)
    wihex        = 8;  (* $ffffffff *)
    widec        = 10; (* 4.294.697.295 *)
    wioct        = 11; (* 37.777.777.777 *)
    wiquad       = 16; (* 3333.3333.3333.3333 *)
    wishadok     = 64; (* wiquad x 3 for Meu *)

PROCEDURE chkNumber (VAR R:ARRAY OF CHAR):BOOLEAN; (* note R may change *)
VAR
    rc:BOOLEAN;
    i,j:CARDINAL;
    str:str4;
    ch:CHAR;
    S:str128; (* oversized *)
BEGIN
    (* //V10G filter out FR AND US the Q&D way *)
    ReplaceChar(R,coma,"");
    ReplaceChar(R,dot,"");

    Str.Copy(S,R);
    IF Str.Match(S,"0X*") THEN (* case does not matter *)
        Str.Delete(S,0,2);
        Str.Prepend(S,hexprefix);
    END;

    CASE S[0] OF
    | hexprefix:
        Str.Delete(S,0,1);
        rc:=verifyString(S,hexdigits);
    | binprefix:
        Str.Delete(S,0,1);
        rc:=verifyString(S,bindigits);
    | octprefix:
        Str.Delete(S,0,1);
        rc:=verifyString(S,octdigits);
    | quadprefix:
        Str.Delete(S,0,1);
        rc:=verifyString(S,quaddigits);
    ELSE
        rc:=verifyString(S,decdigits);
        IF NOT(rc) THEN
            Str.Caps(S); (* safety *)
            FOR i:=0 TO 7 DO (* 0..3 AND 4..7 *)
                isoleItemS(str, numShadok,delim,i);
                Str.Caps(str); (* required ! AND yes, just done earlier *)
                CASE i OF
                | 0..3 : j:=i;
                | 4..7 : j:=i-4;
                END;
                ch:=CHR ( ORD("0") + j );
                WHILE Str.Pos(S,str) # MAX(CARDINAL) DO
                    Str.Subst(S,str,ch);
                END;
            END;
            rc:=verifyString(S,quaddigits);
            IF rc THEN Str.Concat(R,quadprefix,S);END;
        END;
    END;
    rc := (rc AND (Str.Length(S) > 0) );
    RETURN rc;
END chkNumber;

PROCEDURE legalLC (VAR v:LONGCARD; S:ARRAY OF CHAR):BOOLEAN;
VAR
    base:CARDINAL;
    ok:BOOLEAN;
BEGIN
    IF Str.Match(S,"0X*") THEN (* case does not matter *)
        Str.Delete(S,0,2);
        Str.Prepend(S,hexprefix);
    END;

    CASE S[0] OF
    | hexprefix:
        Str.Delete(S,0,1);
        base:=16;
    | binprefix:
        Str.Delete(S,0,1);
        base:=2;
    | octprefix:
        Str.Delete(S,0,1);
        base:=8;
    | quadprefix:
        Str.Delete(S,0,1);
        base:=4;
    ELSE
        base:=10;
    END;
    v:=Str.StrToCard(S,base,ok);
    RETURN ok;
END legalLC;

PROCEDURE pad (padchar:CHAR;wi:CARDINAL;VAR R:ARRAY OF CHAR);
VAR
    i:CARDINAL;
BEGIN
    FOR i:=(Str.Length(R)+1) TO wi DO
        Str.Prepend(R,padchar);
    END;
END pad;

PROCEDURE qdusing (padchar,sepch:CHAR;wi,wigroup:CARDINAL;VAR R:ARRAY OF CHAR);
VAR
    i,len:CARDINAL;
    R2:str80;
BEGIN
    IF padchar # blank THEN pad(padchar,wi,R); END;

    IF sepch # charnull THEN
        len:=Str.Length(R);
        Str.Copy(R2,"");
        FOR i := 1 TO len DO
            Str.Prepend(R2,R[len-i]);
            IF i < len THEN
                IF (i MOD wigroup) = 0 THEN
                    Str.Prepend(R2,sepch);
                END;
            END;
        END;
        Str.Copy(R,R2);
    END;

    IF padchar = " " THEN pad(padchar,wi+(wi DIV wigroup),R); END;
END qdusing;

PROCEDURE fmtNum (v:LONGCARD;base:CARDINAL;padchar,decfmt:CHAR; VAR R:ARRAY OF CHAR);
VAR
    ok:BOOLEAN;
BEGIN
    CASE base OF
    | 2,10,16,8,4:
        Str.CardToStr(v,R,base,ok); (* ok should always be TRUE now *)
    ELSE
        Str.Copy(R,"???");
    END;
    CASE base OF
    | 2:  qdusing(padchar,decfmt,wibin,4,R);
          Str.Prepend(R,binprefix);
    | 10: qdusing(padchar,decfmt,widec,3,R);
    | 16: qdusing(padchar,decfmt,wihex,2,R);
          Str.Prepend(R,hexprefix);
          Str.Lows(R);
    | 8:  qdusing(padchar,decfmt,wioct,3,R);
          Str.Prepend(R,octprefix);
    | 4:  qdusing(padchar,decfmt,wiquad,4,R);
          Str.Prepend(R,quadprefix);
    END;

END fmtNum;

PROCEDURE fmtChar (v:LONGCARD;VAR R : ARRAY OF CHAR):BOOLEAN;
VAR
    code:CARDINAL;
BEGIN
    Str.Copy(R,"???"); (* safety *)
    IF v > MAX(SHORTCARD) THEN RETURN FALSE;END;
    code:=CARDINAL(v);
    Str.Copy(R,dquote+vbar+dquote);
    CASE code OF
    | 0..4, 7..0AH, 0DH, 1AH..1BH, 255 :
        Str.Subst(R,vbar,gremlin);
    | ORD(dquote)  :
        Str.Copy(R,quote+dquote+quote);
    ELSE
        Str.Subst(R,vbar,CHR(code));
    END;
    RETURN TRUE;
END fmtChar;

PROCEDURE fmtShadok (v:LONGCARD;base:CARDINAL;padchar,decfmt:CHAR;
                    VAR R:ARRAY OF CHAR);
VAR
    i:CARDINAL;
    ch:CHAR;
    str:str4;
    ok:BOOLEAN;
BEGIN
    CASE base OF
    | 4:
        Str.CardToStr(v,R,base,ok); (* ok should always be TRUE now *)
        qdusing(padchar,charnull,wiquad,4,R); (* no decfmt, no leading "0" *)
    ELSE
        Str.Copy(R,"???");
    END;

    FOR i:=0 TO 3 DO
        ch:=CHR ( ORD("0") + i );
        isoleItemS(str, numShadokLong,delim,i);
        IF decfmt # charnull THEN Str.Append(str,decfmt);END;
        WHILE Str.CharPos(R,ch) # MAX(CARDINAL) DO
            Str.Subst(R,ch,str);
        END;
    END;
    IF decfmt # charnull THEN
        i:=Str.Length(R);
        IF i # 0 THEN
            IF R[i-1] = decfmt THEN R[i-1]:=charnull;END;
        END;
    END;
    IF R[0] = quadprefix THEN Str.Delete(R,0,1);END;
END fmtShadok;

(* ------------------------------------------------------------ *)

PROCEDURE doChkShadok (decfmt:CHAR;first,last:LONGCARD):BOOLEAN;
CONST
    every = 100;
    sep   = " : ";
VAR
    i,v:LONGCARD;
    R:ARRAY[1..4] OF str128;
    ok:BOOLEAN;
    turn:CARDINAL;
    sGBZM:str128;
BEGIN
    ok:=TRUE;
    turn:=0;
    i:=first;
    LOOP
        fmtNum    (i,10,blank,  decfmt, R[1]);
        fmtNum    (i,4,"0",     decfmt, R[3]);
        fmtShadok (i,4,charnull,decfmt, R[4]);

        WrStr(R[1]);WrStr(sep);
        Str.Copy(R[2],R[4]);
        IF chkNumber( R[2] ) = FALSE THEN
            WrStr("--- chkNumber() ");WrStr(R[2]);WrLn;EXIT;
        END;
        IF legalLC( v, R[2] )=FALSE THEN
            WrStr("--- legalLC() ");WrStr(R[2]);WrLn;EXIT;
        END;
        IF v # i THEN
            WrStr("--- values !");WrLn;EXIT;
        END;
        WrStr(R[3]);WrStr(sep);
        WrStr(R[4]);WrLn;


        IF i = last THEN EXIT;END;
        INC(i);

        INC(turn);
        IF turn > every THEN
            IF ChkEscape() THEN ok:=FALSE;EXIT;END;
            turn:=0;
        END;
    END;
    RETURN ok;
END doChkShadok;

(* ------------------------------------------------------------ *)

CONST
    sepequ   = " = ";
    sep      = " : ";
    sDec     = "Decimal    "+sep; (* account for hex/bin one-char prefix ? bah... *)
    sHex     = "Hexadecimal"+sep;
    sBin     = "Binary     "+sep;
    sOct     = "Octal      "+sep;
    sQuad    = "Base 4     "+sep;
    sShadok  = "Shadok     "+sep;
    sChar    = "Character  "+sep;
    firstnum = 1;
    maxnum   = 20; (* more than enough *)
VAR
    snum : ARRAY [firstnum..maxnum] OF str80; (* oversized *)
    num  : ARRAY [firstnum..maxnum] OF LONGCARD;
    lastnum:CARDINAL;
    parmcount,i,opt:CARDINAL;
    S,R:str128;
    charok,oneline,showspecial:BOOLEAN;
    showoctal,showquad,showchar,showmax,showshadok,showniceshadok:BOOLEAN;
    decfmt:CHAR;
    sGBZM,sQ,sD,sH,sB,sO,sC:str128; (* oversized, str80 should DO but who cares ? *)
    v:LONGCARD;
BEGIN
    FIO.IOcheck:=FALSE;
    WrLn;

    parmcount := Lib.ParamCount();
    (* IF parmcount=0 THEN abort(errHelp,"");END; *)

    lastnum     := firstnum-1;
    oneline     := TRUE;
    showoctal   := FALSE;
    showquad    := FALSE;
    showshadok  := FALSE;
    showniceshadok:=FALSE;
    showmax     := FALSE;
    decfmt      := charnull;
    showchar    := FALSE;
    showspecial := FALSE;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "A"+delim+"ALTERNATE"+delim+
                                   "F"+delim+"FR"+delim+
                                   "U"+delim+"US"+delim+"UK"+delim+
                                   "O"+delim+"OCTAL"+delim+"B"+delim+
                                   "AA"+delim+
                                   "I"+delim+"LIMITS"+delim+"RANGES"+delim+
                                   "C"+delim+"CHARS"+delim+
                                   "4"+delim+"!"+delim+
                                   "@"+delim+"SHADOK"+delim+
                                   "X"+delim+
                                   "@@"+delim+"@4"+delim+"44"+delim+"4@"+delim+
                                   "$"+delim+"-"+delim+"/"
                               );
            CASE opt OF
            | 1,2,3 :    abort(errHelp,"");
            | 4,5   :    oneline:=FALSE;
            | 6,7   :    CASE decfmt OF
                         | charnull,dot:decfmt:=dot;
                         ELSE
                             abort(errFmt,"");
                         END;
            | 8,9,10:    CASE decfmt OF
                         | charnull,comma:decfmt:=comma;
                         ELSE
                             abort(errFmt,"");
                         END;
            | 11,12,13:  showoctal := TRUE; oneline := FALSE;
            | 14:        showoctal := TRUE; oneline := FALSE;
            | 15,16,17:  showmax   := TRUE;
            | 18,19:     showchar  := TRUE;
            | 20,21:     showquad  := TRUE; oneline := FALSE;
            | 22,23:     showshadok:= TRUE; oneline := FALSE;
            | 24:        showoctal := TRUE;
                         showquad  := TRUE;
                         showshadok:= TRUE; oneline := FALSE;
            | 25,26,27,28:
                         showshadok:= TRUE;
                         showquad  := TRUE;
                                            oneline := FALSE;
            | 29,30,31:  showspecial:=TRUE;

            (*
            | :
                         IF doChkShadok( decfmt,MIN(LONGCARD),MAX(LONGCARD)) THEN
                             abort(errNone,"");
                         ELSE
                             abort(errAborted,"");
                         END;
            *)

            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastnum);
            IF lastnum > maxnum THEN abort(errTooManyNumbers,"");END;
            Str.Copy(snum[lastnum],R); (* uppercased string *)
        END;
    END;

    IF foundForceShadokFile(i,FORCESHADOKFILE) THEN showspecial:=TRUE;END;
    IF showspecial THEN oneline:=FALSE;END;

    IF showmax THEN
        lastnum:=firstnum;
        num[lastnum]:=MAX(LONGCARD);
        INC(lastnum);
        num[lastnum]:=MAX(LONGINT);
        WrStr("::: Maximum values :");WrLn;
        WrLn;
    ELSE
        IF lastnum < firstnum THEN  abort(errNoNumber,"");END; (* won't happen *)

        FOR i:=firstnum TO lastnum DO
            IF chkNumber( snum[i] )=FALSE THEN abort(errBadNumber,snum[i]);END;
            IF legalLC( num[i],snum[i] )=FALSE THEN abort(errBadLC,snum[i]);END;
        END;
    END;

    FOR i:=firstnum TO lastnum DO
        v:=num[i];
        fmtNum    (v,10," ",decfmt,sD);
        fmtNum    (v,16,"0",decfmt,sH);
        fmtNum    (v, 2,"0",decfmt,sB);
        fmtNum    (v, 8,"0",decfmt,sO);
        fmtNum    (v, 4,"0",decfmt,sQ);
        fmtShadok (v, 4,"" ,decfmt,sGBZM); (* no leading "0" *)
        charok:=fmtChar(v,sC);

        IF oneline THEN
            WrStr(sD);WrStr(sepequ);WrStr(sH);WrStr(sepequ);WrStr(sB);
            IF showchar THEN
                IF charok THEN WrStr(sepequ);WrStr(sC);END;
            END;
            WrLn;
        ELSE
            IF showspecial THEN
                LtrimBlanks(sD);
                WrStr(sDec);WrStr(sD);WrLn;
                WrStr(sShadok);WrStr(sGBZM);WrLn;
            ELSE
                (* yes, it would be more elegant to have a flag for each display *)
                LtrimBlanks(sD);
                WrStr(sDec);WrStr(sD);WrLn;
                WrStr(sHex);WrStr(sH);WrLn;
                WrStr(sBin);WrStr(sB);WrLn;
                IF showoctal THEN
                WrStr(sOct);WrStr(sO);WrLn;
                END;
                IF showquad THEN
                WrStr(sQuad);WrStr(sQ);WrLn;
                END;
                IF showshadok THEN
                WrStr(sShadok);WrStr(sGBZM);WrLn;
                END;
                IF showchar THEN
                IF charok THEN WrStr(sChar);WrStr(sC);WrLn;END;
                END;
            END;

            IF i < lastnum THEN WrLn; END;
        END;
    END;

    IF showmax THEN
        WrLn;
        WrStr("::: MAX(LONGCARD) and MAX(LONGINT).");WrLn;
    END;

    abort(errNone,"");
END BaseConv.
