(* ---------------------------------------------------------------
Title         Q&D Characters Count
Overview
Usage         see help
Notes         MaybeWare TM (C) mouse support has been discarded :
              it was very low a priority anyway !
              ENT utility mode does not fully support useless ISO -f option
              darn, I've fixed a bug in Vernon Buerg's LIST
              and another one in John Walker's ENT : not bad ! ;-)
              -vv could handle frequency too ?
              we're old and lazy thus we use a fixed array for LFNs
Bugs          yes, in ENT utility ! Should file length be less than
              MONTEN (MAXMONTEVAL here), ENT utility crashes
              for montecount is 0 and forces a division by 0
              see //BUGFIXED
              (division by 0 as montecount is undefined or 0)
Wish List     ignore also graphics and the like?

--------------------------------------------------------------- *)

MODULE cCount;

IMPORT IO;
IMPORT Lib;
IMPORT Str;
IMPORT FIO;
IMPORT Graph;
IMPORT SYSTEM;
IMPORT BiosIO;
IMPORT MATHLIB;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileSupportLFN;

(* shortened list *)

FROM QD_LFN IMPORT unicodeConversionFlagType, findDataRecordType,
w9XfindFirst, w9XfindClose, w9XfindNext, w9XshortToLong;

(* ------------------------------------------------------------ *)

CONST
    blankcode   = ORD(space); (* yes, we use character value here *)
    firstcode   = 0;
    maxcode     = 255;
    countcode   = maxcode - firstcode + 1;
CONST
    firstfile = 1;
    maxfile   = 250; (* was 4000 str16 *)
    smaxfile  = "250";

(* ------------------------------------------------------------ *)

CONST
    progEXEname   = "CCOUNT";
    progTitle     = "Q&D Character Counter";
    progVersion   = "v1.1a";
    progCopyright = "by PhG";
    banner        = progTitle+" "+progVersion+" "+progCopyright;
CONST
    errNone          = 0;
    errHelp          = 1;
    errUnknownOption = 2;
    errTooManyParms  = 3;
    errMissingSpec   = 4;
    errTooManyMatches= 5;
    errNoMatch       = 6;
    errNonsenseSort  = 7;
    errBadNumber     = 8;
    errBarWidthRange = 9;
    errVGAneeded     = 10;
    errCannotSet     = 11;
    errCmd           = 12;
    errMoreHelp      = 13;
    errEvenMoreHelp  = 14;
    errBadSpec       = 15;
    errEscape        = 16;

PROCEDURE abort (e : CARDINAL; einfo:ARRAY OF CHAR);
CONST
(*
 0        1         2         3         4         5         6         7         8
 12345678901234567890123456789012345678901234567890123456789012345678901234567890
*)
    msgHelp =
banner+nl+
nl+
"Syntax 1 : "+progEXEname+" <file(s)> [option]..."+nl+
"Syntax 2 : "+progEXEname+" <file(s)> <-v> [-s|-r]"+nl+
"Syntax 3 : "+progEXEname+" <file(s)> <-vv|-xy>"+nl+
"Syntax 4 : "+progEXEname+" <file(s)> <-entropy> [-e<bctlur>] [-dos]"+nl+
nl+
"This program counts bytes (syntax 1), shows bytes distribution (syntax 3),"+nl+
"tries and shows visually randomness (syntax 3),"+nl+
"or computes entropy statistics emulating John Walker's ENT utility (syntax 4)."+nl+
nl+
"-a     show data for all bytes (no restriction to only bytes appearing in file)"+nl+
"-p     add percentage to byte count"+nl+
"-c     compute percentages excluding control characters ([00..31]) from total"+nl+
"-i[i]  ignore case (-ii = ignore case and accents)"+nl+
"-s     sort from higher to lower byte count"+nl+
"-r     sort from lower to higher byte count"+nl+
"-x     prefix each character with its hexadecimal value"+nl+
"-d     prefix each character with its decimal value"+nl+
"-t     terse report (without header)"+nl+
"-g[g]  show global cumulative data (-gg = global only)"+nl+
"-b[b]  show bar graph (-bb = no semigraphics)"+nl+
"-w:#   bar graph width (default is 25 characters)"+nl+
"-e     eye candy"+nl+
"-v     graphical distribution view"+nl+
"-vv    randomness view (-xy = -vv)"+nl+
"-ee    ENT utility emulation mode"+nl+
"-eb    consider input as a stream of bits instead of 8-bits bytes"+nl+
"-ec    show table of occurrences"+nl+
"-et    show table in CSV format"+nl+
"-el    force lowercase (accents are kept)"+nl+
"-eu    force uppercase (accents are lost)"+nl+
"-er    ENT utility original report format"+nl+
"-dos   ENT for DOS Chi square code (default is ENT for Win32 code)"+nl+
"-l[fn] disable LFN support even if available"+nl+
"-??[?] (even) more help (-??? = about ENT utility emulation mode)"+nl;

CONST
    msgMoreHelp = nl+
"a) This program will handle up to "+smaxfile+" files."+nl+
"b) -i[i] option assumes DOS character set."+nl+
"c) Syntax 2 requires a VGA card allowing 320x200x4 standard video mode ;"+nl+
'   it is roughly similar to "-a -p -x -d -gg -b" syntax 1 display ;'+nl+
"   any option other than -<s|r|e> is silently ignored."+nl+
"   Note -t -g -gg -a options are forced while -c option is ignored."+nl+
"   While viewing, Tab key toggles display mode (real or proportional),"+nl+
'   and "*" key changes sort mode (none, ascending, descending).'+nl+
"d) Syntax 3 requires a VGA card allowing 640x480x4 standard video mode ;"+nl+
"   it parses file as a stream of X,Y coordinates ranging each from 0 to 255."+nl+
"   File length should be even for obvious reasons"+nl+
"   (x without an y final pair is discarded). Note any option is ignored."+nl+
'   When display is done, Tab key toggles undefined dots highlighting.'+nl+
"e) Syntax 4 tries and emulates public domain ENT by John Walker :"+nl+
"   This program uses ENT for Win32 Chi square code by Gary Perlman,"+nl+
"   unless -dos option is specified to use older ENT for DOS Chi square code :"+nl+
"   values may be very different. Note -dos option is not recommended."+nl+
"   -el and -eu options assume DOS character set. ENT -f option is unsupported."+nl+
"   Alternate options : -entropy, -bits, -table, -csv, -lower, -upper, -report."+nl+
"   Contrarily to ENT utility, this program will not crash nor show nonsense"+nl+
"   when computing PI using Monte Carlo method if file length is smaller than 6."+nl+ (* //MONTEVAL *)
nl+
"Examples : "+progEXEname+" *.t -a"+nl+
"           "+progEXEname+" e:\fun\fortune\*.txt -v -ii"+nl+
"           "+progEXEname+" entitle.gif -entropy -report"+nl;

CONST
    msgEvenMoreHelp = nl+
"Here are a few notes about ENT emulation mode"+nl+
"adapted from its original documentation as found in fourmilab.ch Web site."+nl+
"Of course, should any error happen in these notes,"+nl+
"ENT's original author, John Walker, cannot be held responsible for it or them !"+nl+
nl+
"1) Entropy is the information density of the contents of the file,"+nl+
"   expressed as a number of bits per character."+nl+
"2) The chi-square test is the most commonly used test"+nl+
"   for the randomness of data, and is extremely sensitive to errors"+nl+
"   in pseudorandom sequence generators."+nl+
"   The chi-square distribution is calculated for the stream of bytes"+nl+
"   in the file and expressed as an absolute number and a percentage"+nl+
"   which indicates how frequently a truly random sequence would exceed"+nl+
"   the value calculated. One may interpret the percentage as the degree"+nl+
"   to which the sequence tested is suspected of being non-random."+nl+
"   If the percentage is greater than 99% or less than 1%,"+nl+
"   the sequence is almost certainly not random."+nl+
"   If the percentage is between 99% and 95% or between 1% and 5%,"+nl+
"   the sequence is suspect. Percentages between 90% and 95% and 5% and 10%"+nl+
'   indicate the sequence is "almost suspect".'+nl+
"3) The arithmetic mean is simply the result of summing all the bytes or bits"+nl+
"   in the file and dividing by the file length."+nl+
"   If the data are close to random, this should be about 127.5 (0.5 for bits)."+nl+
"   If the mean departs from this value,"+nl+
"   the values are consistently high or low."+nl+
"4) Monte Carlo value for Pi assumes file is at least 6 bytes long."+nl+
"   Each successive sequence of six bytes is used"+nl+
"   as 24 bit X and Y co-ordinates within a square."+nl+
"   If the distance of the randomly-generated point is less"+nl+
"   than the radius of a circle inscribed within the square,"+nl+
'   the six-byte sequence is considered a "hit".'+nl+
"   The percentage of hits can be used to calculate the value of Pi."+nl+
"   For very large streams (this approximation converges very slowly),"+nl+
"   the value will approach the correct value of Pi"+nl+
"   if the sequence is close to random."+nl+
"5) Serial correlation coefficient is a quantity"+nl+
"   which measures the extent to which each byte in the file"+nl+
"   depends upon the previous byte. For random sequences,"+nl+
"   this value (which can be positive or negative) will be close to zero."+nl+
"   A non-random byte stream will yield a serial correlation coefficient"+nl+
"   on the order of 0.5."+nl+
"   Wildly predictable data will exhibit serial correlation coefficients"+nl+
"   approaching 1."+nl;

VAR
    S : str1024; (* oversized but we may get a LFN *)
BEGIN
    CASE e OF
    | errHelp,errMoreHelp,errEvenMoreHelp :
                        WrStr(msgHelp);
                        IF e # errHelp THEN
                            WrStr(msgMoreHelp);
                            IF e # errMoreHelp THEN
                                WrStr(msgEvenMoreHelp);
                            END;
                            e:=errHelp;
                        END;
    | errUnknownOption: S:='Unknown "~" option !';
    | errTooManyParms:  S:='Useless "~" parameter !';
    | errMissingSpec:   S:="Missing <file(s)> specification !";
    | errTooManyMatches:S:='Too many files match "~" specification !';
    | errNoMatch:       S:='No file matches "~" specification !';
    | errNonsenseSort:  S:="-s and -r options are mutually exclusive !";
    | errBadNumber:     S:='Illegal number in "~" parameter !';
    | errBarWidthRange: S:="Bar graph width should belong to the [1..200] range !";
    | errVGAneeded :    S:="VGA adapter needed !" ;
    | errCannotSet :    S:="Cannot set ~ video mode !";
    | errCmd:           S:="-v, -xy and -ent options are mutually exclusive !";
    | errBadSpec:       S:='Illegal "~" specification !';
    | errEscape:        S:="Aborted by user !";
    ELSE
                        S:="How did you get THERE ? This is illogical, Captain ! ;-)";
    END;
    CASE e OF
    | errNone, errHelp:
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE sound (freq,duration,pause:CARDINAL);
BEGIN
    Lib.Sound(freq);
    Lib.Delay(duration);
    Lib.NoSound();
    Lib.Delay(pause);
END sound;

PROCEDURE errbip (okbeep:BOOLEAN);
CONST
    freq     = 55;
    duration = 55; (* was 300 *)
    tempo    = 100;
BEGIN
    IF okbeep THEN sound (freq,duration,tempo); END;
END errbip;

(* ------------------------------------------------------------ *)

(* sep ok if pad is space *)

PROCEDURE fmtlc (v : LONGCARD; pad:CHAR; sep:CHAR; field:INTEGER) : str80;
VAR
    S,R   : str80;
    len,i : CARDINAL;
    ok  : BOOLEAN;
    ch  : CHAR;
BEGIN
    Str.CardToStr(v,S,10,ok);
    len:=Str.Length(S);
    R := "";
    FOR i := 1 TO len DO
        Str.Prepend(R,S[len-i]);
        IF i < len THEN
            IF (i MOD 3) = 0 THEN
                Str.Prepend(R,sep);
            END;
        END;
    END;
    LOOP
        IF INTEGER(Str.Length(R)) >= ABS(field) THEN EXIT; END;
        IF field < 0 THEN
            Str.Append(R,pad);  (* right alignment *)
        ELSE
            Str.Prepend(R,pad);
        END;
    END;
    RETURN R;
END fmtlc;

PROCEDURE fmtstring (S:ARRAY OF CHAR; pad:CHAR; n:INTEGER ) : str80;
VAR
    R : str80;
BEGIN
    Str.Copy(R,S);
    LOOP
        IF INTEGER(Str.Length(R)) >= ABS(n) THEN EXIT; END;
        IF n < 0 THEN
            Str.Append(R,pad); (* right alignment *)
        ELSE
            Str.Prepend(R,pad);
        END;
    END;
    RETURN R;
END fmtstring;

PROCEDURE fmtreal (v:LONGREAL;pad:CHAR;sep:CHAR;
          prec:CARDINAL;field:INTEGER):str80;
VAR
    R : str80;
    ok: BOOLEAN;
BEGIN
    Str.FixRealToStr(v,prec,R,ok);
    Str.Subst(R,dot,sep);
    RETURN fmtstring(R,pad,field);
END fmtreal;

PROCEDURE fmtbar (percentused:LONGREAL;used,free:CHAR;field:CARDINAL):str80;
VAR
    R : str80;
    p,i : CARDINAL;
BEGIN
    percentused := (percentused / 100.0 ) * LONGREAL(field);
    p := CARDINAL(percentused + 0.5); (* round ! *)
    R := "";
    FOR i := 1 TO field DO
        IF i <= p THEN
            Str.Append(R,used);
        ELSE
            Str.Append(R,free);
        END;
    END;
    RETURN R;
END fmtbar;

(* ------------------------------------------------------------ *)

CONST
    IObufferSize = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = IObufferSize;
VAR
    ioBuffer : ARRAY [1..IObufferSize] OF BYTE;

(* ------------------------------------------------------------ *)

CONST
    decidigits    = "0123456789";
    hexadigits    = decidigits+"abcdef";

PROCEDURE WriteCharacter ( i : CARDINAL );
VAR
    ok : BOOLEAN;
BEGIN
    CASE i OF
    | 0..31, 127, 255 :
       WrStr("$");
       WrStr(hexadigits[i DIV 16]); (* assume $00..$FF range, of course ! *)
       WrStr(hexadigits[i MOD 16]);
    | ORD(dquote):
       WrStr( singlequote + dquote + singlequote);
    ELSE
       WrStr(dquote);
       IO.WrChar( CHR(i) );
       WrStr(dquote);
    END;
END WriteCharacter;

PROCEDURE showbargraph (semigraphics:BOOLEAN;wi:CARDINAL;count,maxcount:LONGCARD  );
CONST
    chUsed = CHR(219);  chUsedTxt = "+";
    chFree = CHR(176);  chFreeTxt = "-";
VAR
    v      : LONGREAL;
    charUsed,charFree:CHAR;
BEGIN
    IF semigraphics THEN
        charUsed := chUsed ;     charFree := chFree;
    ELSE
        charUsed := chUsedTxt ;  charFree := chFreeTxt;
    END;
    v := ( LONGREAL(count) / LONGREAL(maxcount) ) * 100.0;
    WrStr(fmtbar(v,charUsed,charFree,wi));
END showbargraph;

PROCEDURE wrval (i:CARDINAL;hex:BOOLEAN);
BEGIN
    IF hex THEN
        WrStr("$");
        WrStr(hexadigits[i DIV 16]); (* assume $00..$FF range, of course ! *)
        WrStr(hexadigits[i MOD 16]);
    ELSE
        WrStr(decidigits[i DIV 100]);
        WrStr(decidigits[ (i MOD 100) DIV 10 ]);
        WrStr(decidigits[i MOD  10]);
    END;
END wrval;

(* ------------------------------------------------------------ *)

(* ugly globerks ! *)

VAR
    code      ,gcode     ,orggcode  : ARRAY [firstcode..maxcode] OF CARDINAL;
    count     ,gcount    ,orggcount : ARRAY [firstcode..maxcode] OF LONGCARD;
    total     ,gtotal               : LONGCARD;
    ctrlchars ,gctrlchars           : LONGCARD;
    filesize  ,gfilesize            : LONGCARD;
CONST
    wilc          = (11+3);
    msgRem        = ";";
    msgFile       = msgRem+" File       : ";
    msgBytes      = msgRem+" Bytes read : ";
    msgTotal      = msgRem+" Total      : ";
    msgCtrl       = msgRem+" [00..31]   : ";
    msgRestricted = msgRem+" Show data only for bytes appearing in file";
    msgAll        = msgRem+" Show data for all bytes";
    msgNoCtrl     = msgRem+" Percentages computed excluding [00..31] from bytes read";
    msgWithCtrl   = msgRem+" Percentages computed using all bytes read";
    msgWarning    = msgRem+" Warning ! As total equals 0, percentages will NOT be computed !";
    msgCase       = msgRem+" Warning ! Case is ignored";
    msgCaseAccents= msgRem+" Warning ! Case and accents are ignored";
CONST
    SORTnone      = 0;
    SORTnormal    = 1;
    SORTreverse   = 2;

PROCEDURE process ( path:pathtype;
                  useLFN,excludecontrol,eyecandy,showglobal,
                  ignorecase,ignoreaccents:BOOLEAN;
                  VAR  filesize, total, ctrlchars,
                  gfilesize,gtotal,gctrlchars:LONGCARD);
VAR
    hnd    : FIO.File;
    prompt : str128;
    i,got  : CARDINAL;
    ch     : BYTE;
    charac : str2;
    ok     : BOOLEAN;
BEGIN
    total     := 0;
    ctrlchars := 0;
    filesize  := fileGetFileSize(useLFN,path);
    IF filesize = 0 THEN RETURN; END;

    FOR i := firstcode TO maxcode DO count[i]:=0; END;

    Str.Concat(prompt,"Processing ",path);Str.Append(prompt," ");
    video(prompt,TRUE);
    IF eyecandy THEN Work(cmdInit);END;
    hnd:=fileOpenRead(useLFN,path);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=TRUE; (* safety ! *)
    LOOP
        IF eyecandy THEN Work(cmdShow);END;
        got := FIO.RdBin(hnd,ch,1);
        IF got = 0 THEN EXIT; END;

        ok:=ch # BYTE(0);
        ok:=( ok AND ignorecase );
        IF ok THEN
            Str.Copy(charac, CHR( ch ) );
            IF ignoreaccents THEN
                LowerCase(charac);
            ELSE
                LowerCaseAlt(charac);
            END;
            ch := charac[0];
        END;

        INC( count[CARDINAL(ch)] ); (* could use ORD(ch) if CHAR TYPE *)
        IF showglobal THEN
            INC( gcount[CARDINAL(ch)] );
        END;
    END;
    fileClose(useLFN,hnd);
    IF eyecandy THEN Work(cmdStop);END;

    FOR i := firstcode TO (blankcode-1) DO
        INC(ctrlchars,count[i]);
        IF showglobal THEN INC(gctrlchars,count[i]); END;
    END;
    IF excludecontrol THEN
        total := filesize-ctrlchars;
    ELSE
        total := filesize;
    END;
    IF showglobal THEN INC(gtotal,total); INC(gfilesize,filesize); END;
    video(prompt,FALSE);
END process;

PROCEDURE dump (path:ARRAY OF CHAR;restricted,percentage,excludecontrol,
               hexval,decval,eyecandy,
               showheader,showglobal,showbar,semigraphics,
               ignorecase,ignoreaccents:BOOLEAN;
               filesize,total,ctrlchars:LONGCARD;sortmode,barwi:CARDINAL);
VAR
    i : CARDINAL;
    exchange,test,dmpall : BOOLEAN;
    ln1,ln2:LONGCARD;
    n1,n2:CARDINAL;
    v : LONGREAL;
    S : str16;
    (* alphanum,symbols,graphics,others:LONGCARD; *)
BEGIN
    IF showheader THEN
        WrStr(msgFile);WrStr(path);WrLn;
        WrStr(msgBytes);WrStr(fmtlc(filesize,space,dot,wilc));WrLn;
        WrStr(msgTotal);WrStr(fmtlc(total,space,dot,wilc));WrLn;
        WrStr(msgCtrl);WrStr(fmtlc(ctrlchars,space,dot,wilc));WrLn;
        WrStr(msgRem);WrLn;
        IF restricted THEN
            WrStr(msgRestricted);
        ELSE
            WrStr(msgAll);
        END;
        WrLn;
        IF ignorecase THEN
            IF ignoreaccents THEN
                WrStr(msgCaseAccents);
            ELSE
                WrStr(msgCase);
            END;
            WrLn;
        END;
        IF percentage THEN
            IF excludecontrol THEN
                WrStr(msgNoCtrl);
            ELSE
                WrStr(msgWithCtrl);
            END;
            WrLn;
            IF total=0 THEN
                WrStr(msgWarning);WrLn;
                percentage := FALSE;
            END;
        END;
        WrLn;
    END;

    FOR i := firstcode TO maxcode DO code [i]:=i; END;

    IF sortmode # SORTnone THEN
        IF eyecandy THEN Animation(cmdInit);END;
        exchange := TRUE;
        WHILE exchange DO
            IF eyecandy THEN Animation(cmdShow);END;
            exchange := FALSE;
            FOR i := firstcode TO maxcode-1 DO
                ln1 := count[i];
                ln2 := count[i+1];
                CASE sortmode OF
                | SORTnormal:
                    test := (ln1 < ln2);  (* keep A, B, C,... sequence *)
                | SORTreverse:
                    test := (ln2 < ln1); (* <= instead of < forces A, B, C,... sequence, but would provoke an infinite loop *)
                END;
                IF test THEN
                    count[i]   := ln2;
                    count[i+1] := ln1;
                    n1         := code[i];
                    n2         := code[i+1];
                    code[i]    := n2;
                    code[i+1]  := n1;
                    exchange   := TRUE;
                END;
            END;
        END;
        IF eyecandy THEN Animation(cmdStop);END;
    END;

    dmpall := NOT(restricted);
    FOR i := firstcode TO maxcode DO
        IF ( dmpall OR (count[i] # 0) ) THEN

            IF hexval THEN wrval(ORD(code[i]),TRUE); WrStr(tabul); END;
            IF decval THEN
                IF hexval THEN WrStr("(");END;
                           wrval(ORD(code[i]),FALSE);
                IF hexval THEN WrStr(")");END;
                                                     WrStr(tabul);
            END;

            WriteCharacter(code[i]);
            WrStr(tabul);
            IO.WrLngCard(count[i],11);
            IF percentage THEN
                WrStr(tabul);
                v := ( LONGREAL(count[i]) / LONGREAL(total) ) * 100.0;
                Str.Concat(S, fmtreal(v,space,comma,2, 3+1+2) , "%");
                IF excludecontrol THEN
                    CASE i OF
                    | 0..31 : S := "nonsense";
                    END;
                END;
                WrStr(S);
            END;
            IF showbar THEN
                WrStr(tabul);
                showbargraph (semigraphics, barwi, count[i],total);
            END;
            WrLn;
        END;
    END;
END dump;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE realToCard (v:LONGREAL ):CARDINAL ;
CONST
    k = LONGREAL(0.5); (* was 0.5 *)
BEGIN
    RETURN VAL(CARDINAL, v + k);
END realToCard;

PROCEDURE plot (x,ybase,color:CARDINAL; v:LONGREAL );
VAR
    p,y:CARDINAL;
BEGIN
    p:=realToCard(v);
    y:=ybase-p+1;
    CASE p OF
    | 0 : ;
    (* | 1 : Graph.Plot(x,y,color); *)
    ELSE
        Graph.Line(x,ybase,x,y,color);
    END;
END plot;

CONST
    cBLACK         =0;
    cBLUE          =1;
    cGREEN         =2;
    cCYAN          =3;
    cRED           =4;
    cMAGENTA       =5;
    cBROWN         =6;
    cWHITE         =7;
    cGRAY          =8;
    cLIGHTBLUE     =9;
    cLIGHTGREEN    =10;
    cLIGHTCYAN     =11;
    cLIGHTRED      =12;
    cLIGHTMAGENTA  =13;
    cLIGHTYELLOW   =14;
    cBRIGHTWHITE   =15;
CONST
    COLORFRAME     = cWHITE;
    COLORSTATUS    = cLIGHTGREEN;
    COLORTEXT      = cBRIGHTWHITE;
    COLORBK        = cBLACK;
    COLORBAR1      = cLIGHTRED;
    COLORBAR2      = cLIGHTYELLOW;

PROCEDURE WaitVGAretrace ();
BEGIN
    WHILE (SYSTEM.In(03DAH) AND 08H) # 0 DO
    END;
    WHILE (SYSTEM.In(03DAH) AND 08H) = 0 DO
    END;
END WaitVGAretrace;

PROCEDURE centerme (minhtab,wi:CARDINAL;S:ARRAY OF CHAR):CARDINAL;
VAR
    len,hpos:CARDINAL;
BEGIN
    len := Str.Length(S);
    IF len < wi THEN
        hpos:=minhtab+(wi-len) DIV 2;
    ELSE
        hpos:=minhtab;
    END;
    RETURN hpos;
END centerme;

VAR
    gvideoconfig        : Graph.VideoConfig;
    gvmode              : CARDINAL;
VAR
    mouseXmul           : CARDINAL;
    screenX, screenY    : CARDINAL;
    htabMin,htabWidth   : CARDINAL;
    vtabMin,vtabHeight  : CARDINAL;
    numColors           : CARDINAL;
    xmin,xmax,ymin,ymax : CARDINAL;
    screenCX,screenCY   : CARDINAL;

PROCEDURE setupVideoScreen();
VAR
    k : CARDINAL;
BEGIN
    Graph.GetVideoConfig(gvideoconfig);
    screenX  := gvideoconfig.numxpixels; (* 320 *)
    screenY  := gvideoconfig.numypixels; (* 200 *)
    CASE screenX OF
    | 320 : k :=2;
    | 640 : k :=1;
    ELSE
            k :=1;
    END;
    mouseXmul := k;

    htabMin  := 1;
    htabWidth:= gvideoconfig.numtextcols; (* screenX DIV 8 i.e. 40 here *)
    vtabMin  := 1;
    vtabHeight:=gvideoconfig.numtextrows;

    numColors:= gvideoconfig.numcolors; (* 16 *)

    xmin     := 0;
    xmax     := screenX-1;
    ymin     := 0;
    ymax     := screenY-1;

    screenCX := screenX DIV 2;
    screenCY := screenY DIV 2;
END setupVideoScreen;

PROCEDURE isKeyWaiting (  ) : BOOLEAN;
BEGIN
    RETURN BiosIO.KeyPressed();
END isKeyWaiting;

CONST
    keyEscape = 1B00H;
    keyLeft   = 004BH;
    keyRight  = 004DH;
    keyHome   = 0047H;
    keyEnd    = 004FH;
    keyPgUp   = 0049H;
    keyPgDn   = 0051H;
    keyTab    = 0900H;
    keyStar   = 2A00H;

PROCEDURE ReadKeycode (  ) : CARDINAL;
VAR
    cl,ch : CARDINAL;
BEGIN
    ch := ORD(BiosIO.RdKey());
    IF ch = 0 THEN
        cl:=ORD(BiosIO.RdKey());
    ELSE
        cl:=0;
    END;
    RETURN cl+(ch << 8);
END ReadKeycode;

PROCEDURE nextcolor (VAR c :CARDINAL);
 BEGIN
    CASE c OF
    | COLORBAR1 : c:=COLORBAR2;
    | COLORBAR2 : c:=COLORBAR1;
    END;
END nextcolor;

PROCEDURE newcode (VAR v:CARDINAL; k:INTEGER);
VAR
    i:CARDINAL;
BEGIN
    IF k < 0 THEN
        FOR i:=1 TO ABS(k) DO
            IF v > firstcode THEN
                DEC(v);
            ELSE
                v:=maxcode;
            END;
        END;
    ELSE
        FOR i:=1 TO k DO
            IF v < maxcode THEN
                INC(v);
            ELSE
                v:=firstcode;
            END;
        END;
    END;
END newcode;

PROCEDURE procSort (sortmode:CARDINAL);
VAR
    i : CARDINAL;
    exchange,test:BOOLEAN;
    ln1,ln2:LONGCARD;
    n1,n2:CARDINAL;
BEGIN
    FOR i := firstcode TO maxcode DO gcode [i]:=i; END;
    IF sortmode # SORTnone THEN
        exchange := TRUE;
        WHILE exchange DO
            exchange := FALSE;
            FOR i := firstcode TO maxcode-1 DO
                ln1 := gcount[i];
                ln2 := gcount[i+1];
                CASE sortmode OF
                | SORTnormal:
                    test := (ln1 < ln2);  (* keep A, B, C,... sequence *)
                | SORTreverse:
                    test := (ln2 < ln1); (* <= instead of < forces A, B, C,... sequence, but would provoke an infinite loop *)
                END;
                IF test THEN
                    gcount[i]   := ln2;
                    gcount[i+1] := ln1;
                    n1          := gcode[i];
                    n2          := gcode[i+1];
                    gcode[i]    := n2;
                    gcode[i+1]  := n1;
                    exchange    := TRUE;
                END;
            END;
        END;
    END;
END procSort;

PROCEDURE drawdata (VAR highest:LONGREAL;
                   xbase,ybase,hedata:CARDINAL;realsize,computehighest:BOOLEAN);
VAR
    v:LONGREAL;
    i,x,color:CARDINAL;
BEGIN
    IF computehighest THEN
        highest:=0.0;
        FOR i:=firstcode TO maxcode DO
            v := ( LONGREAL(gcount[i]) / LONGREAL(gtotal) ) * 100.0;
            IF v > highest THEN highest:=v; END;
        END;
    END;
    color := COLORBAR1;
    FOR i:=firstcode TO maxcode DO
        v := ( LONGREAL(gcount[i]) / LONGREAL(gtotal) ) * 100.0;
        IF  NOT(realsize) THEN v:=v * 100.0 / highest; END;
        x:=xbase+i;

        Graph.Line(x,ybase+1,x,ybase-hedata-1,cBLACK); (* erase old bar plus safety margins *)

        plot(x,ybase,color,v);
        nextcolor(color);
    END;
END drawdata;

(* we use gcode and gcount *)

PROCEDURE procView (sortmode:CARDINAL );
CONST
    (*                    0        1         2         3         4 *)
    (*                    1234567890123456789012345678901234567890 *)
    msgHelpMouse       = "Mouse-cursor  Escape-exit";
    msgHelpKbd         = "Arrows Home End PgUp PgDn Tab * Escape";
CONST
    msgHex             = "Hexadecimal : $~";
    msgDec             = "Decimal     : ~";
    msgChar            = "Character   : ~";
    msgPercentage      = "Percentage  : ~%";
    msgCount           = "Count       : ~";
    msgTotal           = "Total       : ~";
    nlines             = 6;
CONST
    COLORDEC   = cWHITE;
    COLORHEX   = cLIGHTYELLOW;
CONST
    widata = countcode; (* codes [0..255] *)
    hedata = 100; (* percentage *)
    k      = 1+2+1+1+1;
VAR
    realcode,prevcode,currcode,hpos,vpos,charcode : CARDINAL;
    color,i,x,mx,my:CARDINAL;
    xbase,ybase:CARDINAL;
    v,highest:LONGREAL;
    S:str128;
    xy:Graph.TextCoords;
    R:str80;
    (* msinfo : MsMouse.MsData; //FIXME L8R *)
    ok,realsize:BOOLEAN;
BEGIN
    orggcode := gcode;   (* save original data *)
    orggcount:= gcount;

    setupVideoScreen();

    xbase := xmin + (screenX-widata) DIV 2;
    ybase := ymin + hedata + k*2;

    color := COLORSTATUS;
    i:=Graph.SetTextColor(color);
    vpos  := vtabMin+vtabHeight-1-1;
    S:=msgHelpKbd;
    hpos:=centerme(htabMin,htabWidth,S);
    xy:=Graph.SetTextPosition(vpos,hpos);
    Graph.OutText(S);
    DEC(vpos,nlines+1);

    color := COLORFRAME;
    Graph.Rectangle(xbase-k,ybase+k,xbase+widata+k,ybase-hedata-k,color,FALSE);
    FOR i:=firstcode TO maxcode DO
        IF i MOD 8 = 0 THEN (* inner hex *)
            color := COLORHEX;
            Graph.Plot(xbase+i,ybase+k-1,color);
        END;
        IF i MOD 10 = 0 THEN (* outer dec *)
            color := COLORDEC;
            Graph.Plot(xbase+i,ybase+k+1,color);
        END;
    END;

    color:=COLORFRAME;
    FOR i:=0 TO hedata DO
        IF i MOD 10 = 0 THEN
            Graph.Plot (xbase-k-1,ybase-i,color);
            Graph.Plot (xbase+widata+k+1,ybase-i,color);
        END;
    END;

    procSort(sortmode);

    realsize:=FALSE; (*  true for 0..100 *)
    drawdata (highest, xbase,ybase,hedata,realsize,TRUE );

    color := COLORTEXT;
    i:=Graph.SetTextColor(color);

    currcode := firstcode;
    LOOP
        prevcode:=currcode;
        IF isKeyWaiting() THEN
            charcode:=ReadKeycode();
            CASE charcode OF
            | keyStar:
                INC(sortmode);
                IF sortmode > SORTreverse THEN sortmode:=SORTnone;END;
                (* could be done for SORTnone only but... *)
                gcode:=orggcode;
                gcount:=orggcount;
                procSort(sortmode);
                drawdata (highest, xbase,ybase,hedata,realsize,FALSE );

            | keyHome: currcode:=firstcode;
            | keyEnd:  currcode:=maxcode;
            | keyEscape:
                EXIT;
            | keyRight: newcode(currcode,1);
            | keyLeft:  newcode(currcode,-1);
            | keyPgUp:  newcode(currcode,-16);
            | keyPgDn:  newcode(currcode,16);
            | keyTab:
                realsize := NOT(realsize);

                drawdata (highest, xbase,ybase,hedata,realsize,FALSE );

                FOR i:=0 TO hedata DO
                    CASE i OF
                    | 0,hedata: ok:=TRUE; color:=COLORFRAME; (* force 0 and 100 *)
                    ELSE
                        ok:=( ( i MOD 10 ) = 0);
                        IF realsize THEN
                            color:=COLORFRAME;
                        ELSE
                            color:=COLORBK;
                        END;
                    END;
                    IF ok THEN
                        Graph.Plot (xbase-k-1,ybase-i,color);
                        Graph.Plot (xbase+widata+k+1,ybase-i,color);
                    END;
                END;
            ELSE
                errbip(TRUE);
            END;
        END;

        (* cursor *)

        FOR i:=1 TO 2 DO
            CASE i OF
            | 1: color:=COLORBK;    x:=prevcode;
            | 2: color:=COLORTEXT;  x:=currcode;
            END;
            Graph.Plot(xbase+x,ybase+2,color);
            Graph.Plot(xbase+x-1,ybase+2+1,color);
            Graph.Plot(xbase+x+1,ybase+2+1,color);
        END;

        realcode:=gcode[currcode];
        FOR i:=1 TO nlines DO
            CASE i OF
            | 1: S:=msgHex;
                Str.Copy(R,hexadigits[realcode DIV 16]); (* assume $00..$FF range, of course ! *)
                Str.Append(R,hexadigits[realcode MOD 16]);
            | 2: S:=msgDec;
                R:=fmtlc ( LONGCARD(realcode), space, "",3);
            | 3: S:=msgChar;
                CASE realcode OF
                | 0..31:
                    Str.Concat(R,"Ctrl-",CHR(realcode+ORD("A")-1));
                | 34:
                    Str.Concat(R,singlequote,CHR(realcode));
                    Str.Append(R,singlequote+"   "); (* same length as "Ctrl-?" *)
                | 255:
                    R:=dquote+" "+dquote+"   "; (* invisible space *)
                ELSE
                    Str.Concat(R,dquote,CHR(realcode));
                    Str.Append(R,dquote+"   "); (* same length as "Ctrl-?" *)
                END;
            | 4: S:=msgPercentage;
                v := ( LONGREAL(gcount[currcode]) / LONGREAL(gtotal) ) * 100.0;
                R:= fmtreal (v, space, comma , 2,3+1+2);
            | 5: S:=msgCount;
                R:=fmtlc ( gcount[currcode], space, dot, wilc);
            | 6: S:=msgTotal;
                R:=fmtlc ( gtotal, space, dot, wilc);
            END;
            Str.Subst(S,"~",R);
            hpos:=htabMin+2;
            xy:=Graph.SetTextPosition(vpos+i-1,hpos);
            Graph.OutText(S);
        END;
    END;
END procView;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
PROCEDURE dmpstr (S1,S2,S3:ARRAY OF CHAR);
CONST
    padwi = 30;
VAR
    i:CARDINAL;
BEGIN
    WrStr(S1);
    FOR i:= Str.Length(S1)+1 TO padwi DO WrStr(" ");END;
    WrStr(" : ");WrStr(S2);
    IF same(S3,"") = FALSE THEN WrStr(S3);END;
    WrLn;
END dmpstr;

PROCEDURE dmplc (S1:ARRAY OF CHAR;v:LONGCARD;S3:ARRAY OF CHAR    );
VAR
    S2:str80; (* slightly oversized *)
BEGIN
    Str.Copy(S2, fmtlc(v,"","",1));
    dmpstr(S1,S2,S3);
END dmplc;
*)

(*
PROCEDURE dbgf ( DEBUG:BOOLEAN;v:LONGREAL;S:ARRAY OF CHAR);
BEGIN
    IF DEBUG THEN
        WrStr("/// ");WrStr(S);WrStr(" : ");
        WrStr( fmtreal(v,"",dot,6,1) );WrLn;
    END;
END dbgf;

PROCEDURE dbg (DEBUG:BOOLEAN;S:ARRAY OF CHAR  );
BEGIN
    IF DEBUG THEN
        WrStr("/// ");WrStr(S);WrLn;
    END;
END dbg;
*)

PROCEDURE printable (ch:CHAR  ):CHAR;
BEGIN
    CASE ORD(ch) OF
    | 0..31 : RETURN ".";
    | 32..254 : RETURN ch;
    | 255 : RETURN " ";
    END;
END printable;

TYPE
    rt_type = RECORD
        ent     : LONGREAL;
        chisq   : LONGREAL;
        mean    : LONGREAL;
        montepi : LONGREAL;
        scc     : LONGREAL;
        chip    : LONGREAL; (* could be computed later but well... *)
    END;

VAR
    PI      : LONGREAL;
    jwtotal : LONGCARD;
    jwcount : ARRAY [firstcode..maxcode] OF LONGCARD;

CONST
    firstmonteval = 1;
    (*
       Bytes used as Monte Carlo co-ordinates.
       This should be no more bits than the mantissa OF LONGREAL type.
    *)
    MAXMONTEVAL   = 6;
    sccUNDEFINED  = -99999.0;
    monteUNDEFINED= -99999.0; (* //BUGFIXED *)
VAR
    lastmonteval : CARDINAL;
    monteval     : ARRAY [firstmonteval..MAXMONTEVAL] OF LONGREAL;
    montecount   : LONGCARD;
    montehit     : LONGCARD;
    incircle     : LONGREAL;
    datasum      : LONGREAL;
    sccfirst     : BOOLEAN;
    scct1, scct2, scct3: LONGREAL;
    sccu0, sccun, scclast : LONGREAL;
VAR
    sx,sx2:LONGREAL;
    sxn:LONGCARD;
VAR
    proba        : ARRAY [firstcode..maxcode] OF LONGREAL; (* Probabilities per bin for entropy *)

(* ------------------------------------------------------------ *)

(* log2of10 = 3.32192809488736234787 = 1/log(2) *)

PROCEDURE log2 (v:LONGREAL ):LONGREAL;
BEGIN
    RETURN MATHLIB.Log(v) / MATHLIB.Log(2.0);
END log2;

PROCEDURE jwISOchar (code:CARDINAL):CHAR;
VAR
    ok:BOOLEAN;
    ch:CHAR;
BEGIN
    ok:=FALSE;
    CASE code OF
    | 0..32 :    ;
    | 127..160 : ;
    (* | 255 :      ; *)
    ELSE
        ok:=TRUE;
    END;
    IF ok THEN
        ch:=CHR(code);
    ELSE
        ch:=" ";
    END;
    RETURN ch;
END jwISOchar;

PROCEDURE getjwLast(jwBitMode:BOOLEAN):CARDINAL;
VAR
    last:CARDINAL;
BEGIN
    IF jwBitMode THEN
        last:=firstcode+1; (* 0+1 *)
    ELSE
        last:=maxcode;     (* 255 *)
    END;
    RETURN last;
END getjwLast;

PROCEDURE getjwBitCount (jwBitMode:BOOLEAN):CARDINAL ;
VAR
    k:CARDINAL;
BEGIN
    IF jwBitMode THEN
        k:=1;
    ELSE
        k:=8;
    END;
    RETURN k;
END getjwBitCount;

PROCEDURE getfjwBitCount (jwBitMode:BOOLEAN):LONGREAL;
BEGIN
    RETURN LONGREAL(getjwBitCount(jwBitMode));
END getfjwBitCount;

PROCEDURE getstrUnit (jwBitMode:BOOLEAN ):str16;
VAR
    units:str16;
BEGIN
    IF jwBitMode THEN
        units:="bit";
    ELSE
        units:="byte";
    END;
    RETURN units;
END getstrUnit;

PROCEDURE doProblem (jwOrgRpt:BOOLEAN;path:ARRAY OF CHAR   );
VAR
    S:str128;
BEGIN
    IF NOT(jwOrgRpt) THEN
        S:='::: Analysis results for "~" file.'+nl+
        nl+
        "--- No analysis can be performed on a 0-length file."+nl;
        Str.Subst(S,"~",path);
        WrStr(S);
    END;
END doProblem;

PROCEDURE doSummaryAlt (ENT32,jwBitMode:BOOLEAN;jwfix:CARDINAL;
                        rt:rt_type;path:ARRAY OF CHAR);
VAR
    S:str1024; (* can be very long a string *)
    Z:str128;
    units:str16;
    v:LONGREAL;
BEGIN
        units:=getstrUnit(jwBitMode);
        S:= '::: Analysis results for "~" file.'+nl+
            nl+
            "+++ Entropy is ~ bits per ~."+nl+
            "+++ Optimum compression would reduce the size of this ~ ~ file"+nl+
            "    by ~ percent."+nl+
            "+++ Chi square distribution for ~ samples is ~"+nl+
            "    and it randomly would exceed this value"+nl+"    ";
        IF NOT(ENT32) THEN
            Str.Append(S,"~ percent of the times."+nl);
        ELSE
            IF rt.chip < 0.0001 THEN
                rt.chip:=0.01 / 100.0;
                Str.Append(S,"less than ~ percent of the times."+nl);
            ELSIF rt.chip > 0.9999 THEN
                rt.chip:=99.99 / 100.0;
                Str.Append(S,"more than ~ percent of the times."+nl); (* //FIXED "than than" *)
            ELSE
                Str.Append(S,"~ percent of the times."+nl);
            END;
        END;
            Str.Append(S,"+++ Arithmetic mean value of data ~s is ~ (~ = random)."+nl);

        Str.Subst(S,"~",path);

        (* entropy *)
        Str.Subst(S,"~",fmtreal(rt.ent,"",dot,jwfix,1) );
        Str.Subst(S,"~",units);

        Str.Subst(S,"~",fmtlc( jwtotal," ","",1));
        Str.Subst(S,"~",units);
        v:= 100.0 * ( getfjwBitCount(jwBitMode) - rt.ent) / getfjwBitCount(jwBitMode);
        Str.Subst(S,"~",fmtlc(LONGCARD(v),"","",1));

        (* chi square *)
        Str.Subst(S,"~",fmtlc(jwtotal,"","",1));
        Str.Subst(S,"~",fmtreal(rt.chisq,"",dot,2,1));
        Str.Subst(S,"~",fmtreal(rt.chip*100.0,"",dot,2,1) );

        (* arithmetic mean *)
        Str.Subst(S,"~",units);
        Str.Subst(S,"~",fmtreal(rt.mean,"",dot,4,1));
        IF jwBitMode THEN
            Str.Subst(S,"~",fmtreal(0.5,"",dot,1,1) );
        ELSE
            Str.Subst(S,"~",fmtreal(127.5,"",dot,1,1) );
        END;
        WrStr(S);

        (* monte carlo *)
        Z:="| Monte Carlo value for Pi is ";
        IF rt.montepi >= monteUNDEFINED THEN
            S:="~ (error ~ percent)."+nl;
            Str.Subst(S,"~",fmtreal(rt.montepi,"",dot,9,1) );
            v:=100.0 * (ABS(PI-rt.montepi) / PI);
            Str.Subst(S,"~",fmtreal(v,"",dot,2,1) );
            Str.Subst(Z,"|","+++");
        ELSE                                              (* //BUGFIXED *)
            S:="undefined (fewer than ~ bytes)."+nl;
            Str.Subst(S,"~",fmtlc(MAXMONTEVAL,"","",1));
            Str.Subst(Z,"|","---");
        END;
        WrStr(Z);
        WrStr(S);

        Z:="| Serial correlation coefficient is ";
        IF rt.scc >= sccUNDEFINED THEN
            S:="~ (totally uncorrelated = 0.0).";
            Str.Subst(S,"~",fmtreal(rt.scc,"",dot,6,1) );
            Str.Subst(Z,"|","+++");
        ELSE
            S:="undefined (all values are equal).";
            Str.Subst(Z,"|","---");
        END;
        WrStr(Z);
        WrStr(S);WrLn;

        (* average AND sigma *)

        IF sxn = 0 THEN
            S:= "--- Average and standard deviation are undefined"+nl+
                "    for 0 byte sample."+nl;
        ELSIF sxn = 1 THEN
            S:= "--- Average is ~ but standard deviation (sigma) is undefined"+nl+
                "    for 1 byte sample."+nl;
            v:= sx ;
            Str.Subst(S,"~",fmtreal(v,"",dot,4,1));
        ELSE
            S:= "+++ Average is ~ and standard deviation (sigma) is ~"+nl+
                "    for ~ byte samples."+nl;
            v:= sx / LONGREAL(sxn);
            Str.Subst(S,"~",fmtreal(v,"",dot,4,1));
            v:= MATHLIB.Sqrt( (sx2 - (sx * sx) / LONGREAL(sxn) ) / LONGREAL(sxn-1) );
            Str.Subst(S,"~",fmtreal(v,"",dot,jwfix,1));
            Str.Subst(S,"~",fmtlc(sxn,"","",1));
        END;

        WrStr(S);

END doSummaryAlt;

PROCEDURE doSummary(ENT32,jwBitMode:BOOLEAN;jwfix:CARDINAL;rt:rt_type);
VAR
    S:str1024; (* can be very long a string *)
    units:str16;
    v:LONGREAL;
BEGIN
        units:=getstrUnit(jwBitMode);
        S:= "Entropy = ~ bits per ~."+nl+
            nl+
            "Optimum compression would reduce the size"+nl+
            "of this ~ ~ file by ~ percent."+nl+
            nl+
            "Chi square distribution for ~ samples is ~, and randomly"+nl+
            "would exceed this value ";
        IF NOT(ENT32) THEN
            Str.Append(S,"~ percent of the times."+nl);
        ELSE
            IF rt.chip < 0.0001 THEN
                rt.chip:=0.01 / 100.0;
                Str.Append(S,"less than ~ percent of the times."+nl);
            ELSIF rt.chip > 0.9999 THEN
                rt.chip:=99.99 / 100.0;
                Str.Append(S,"more than ~ percent of the times."+nl); (* //FIXED "than than" *)
            ELSE
                Str.Append(S,"~ percent of the times."+nl);
            END;
        END;
            Str.Append(S,nl+
            "Arithmetic mean value of data ~s is ~ (~ = random)."+nl);
        (* entropy *)
        Str.Subst(S,"~",fmtreal(rt.ent,"",dot,jwfix,1) );
        Str.Subst(S,"~",units);

        Str.Subst(S,"~",fmtlc( jwtotal," ","",1));
        Str.Subst(S,"~",units);
        v:= 100.0 * ( getfjwBitCount(jwBitMode) - rt.ent) / getfjwBitCount(jwBitMode);
        Str.Subst(S,"~",fmtlc(LONGCARD(v),"","",1));

        (* chi square *)
        Str.Subst(S,"~",fmtlc(jwtotal,"","",1));
        Str.Subst(S,"~",fmtreal(rt.chisq,"",dot,2,1));
        Str.Subst(S,"~",fmtreal(rt.chip*100.0,"",dot,2,1) );

        (* arithmetic mean *)
        Str.Subst(S,"~",units);
        Str.Subst(S,"~",fmtreal(rt.mean,"",dot,4,1));
        IF jwBitMode THEN
            Str.Subst(S,"~",fmtreal(0.5,"",dot,1,1) );
        ELSE
            Str.Subst(S,"~",fmtreal(127.5,"",dot,1,1) );
        END;
        WrStr(S);

        (* monte carlo *)
        S:="Monte Carlo value for Pi is ";
        WrStr(S);
        IF rt.montepi >= monteUNDEFINED THEN
            S:="~ (error ~ percent)."+nl;
            Str.Subst(S,"~",fmtreal(rt.montepi,"",dot,9,1) );
            v:=100.0 * (ABS(PI-rt.montepi) / PI);
            Str.Subst(S,"~",fmtreal(v,"",dot,2,1) );
        ELSE                                              (* //BUGFIXED *)
            S:="undefined (fewer than ~ bytes)."+nl;
            Str.Subst(S,"~",fmtlc(MAXMONTEVAL,"","",1));
        END;
        WrStr(S);

        S:="Serial correlation coefficient is ";
        WrStr(S);
        IF rt.scc >= sccUNDEFINED THEN
            S:="~ (totally uncorrelated = 0.0).";
            Str.Subst(S,"~",fmtreal(rt.scc,"",dot,6,1) );
        ELSE
            S:="undefined (all values equal!).";
        END;
        WrStr(S);WrLn;
END doSummary;

PROCEDURE doReport (ENT32,jwBitMode,jwShowTable,jwCSV,jwOrgRpt:BOOLEAN;
                   rt:rt_type;path:ARRAY OF CHAR);
CONST
    jwfix = 6;
    windx = 3;
    wilc  = 10;
VAR
    units:str16;
    S:str128;
    i:CARDINAL;
    thiscount:LONGCARD;
    percentage:LONGREAL;
BEGIN
    units:=getstrUnit(jwBitMode);
    IF jwShowTable THEN
        IF jwCSV THEN
            S:="0,File-~s,Entropy,Chi-square,Mean,Monte-Carlo-Pi,Serial-Correlation";
            Str.Subst(S,"~",units);
            WrStr(S);WrLn;
            S:="1,~,~,~,~,~,~";
            Str.Subst(S,"~",fmtlc  (jwtotal,"","",1));
            Str.Subst(S,"~",fmtreal(rt.ent,"",dot,jwfix,1));
            Str.Subst(S,"~",fmtreal(rt.chisq,"",dot,jwfix,1));
            Str.Subst(S,"~",fmtreal(rt.mean,"",dot,jwfix,1));
            Str.Subst(S,"~",fmtreal(rt.montepi,"",dot,jwfix,1));
            Str.Subst(S,"~",fmtreal(rt.scc,"",dot,jwfix,1));
            WrStr(S);WrLn;
            S:="2,Value,Occurrences,Fraction";
        ELSE
            S:="Value Char Occurrences Fraction";
        END;
        WrStr(S);WrLn;

        FOR i:= firstcode TO getjwLast(jwBitMode) DO
            thiscount:=jwcount[i];
            percentage := LONGREAL(thiscount) / LONGREAL(jwtotal);
            IF jwCSV THEN
                S:="3,~,~,~";
                Str.Subst(S,"~",fmtlc( LONGCARD(i),"","",1 ) );
                Str.Subst(S,"~",fmtlc( jwcount[i],"","",1) );
                Str.Subst(S,"~",fmtreal(percentage,"",dot,jwfix,1) );
                WrStr(S);WrLn;
            ELSE
                IF thiscount # 0 THEN
                    S:="~   ~";
                    Str.Subst(S,"~",fmtlc( LONGCARD(i)," ","",windx ) );
                    Str.Subst(S,"~",jwISOchar(i));
                    WrStr(S);
                    (* handle char being placeholder *)
                    S:="   ~   ~";
                    Str.Subst(S,"~",fmtlc( jwcount[i]," ","",wilc) );
                    Str.Subst(S,"~",fmtreal(percentage,"",dot,jwfix,1) );
                    WrStr(S);WrLn;
                END;
            END;
        END;
        IF NOT(jwCSV) THEN
            WrLn;
            S:="Total:    ~   ~"; (* same alignment as supra *)
            Str.Subst(S,"~",fmtlc( jwtotal," ","",wilc) );
            Str.Subst(S,"~",fmtreal(1.0,"",dot,jwfix,1) );
            WrStr(S);WrLn;
        END;
    END;
    IF NOT(jwCSV) THEN
        IF jwOrgRpt THEN
            doSummary(ENT32,jwBitMode,jwfix,rt);
        ELSE
            doSummaryAlt(ENT32,jwBitMode,jwfix,rt,path);
        END;
    END;
END doReport;

(* ------------------------------------------------------------ *)

(*
    for gory details about this dark magic code by Gary Perlman,
    check original ENT for Win32 by John Walker !
    mathematica sunt : non leguntur
*)

CONST
    BIGX   = LONGREAL(20.0);
    Z_MAX  = LONGREAL(6.0); (* maximum meaningful z value *)

PROCEDURE poz (z:LONGREAL):LONGREAL;
VAR
    y,x,w:LONGREAL;
BEGIN
    IF z = 0.0 THEN
    	x := 0.0;
    ELSE
	    y := 0.5 * ABS(z);
	END;
	IF (y >= (Z_MAX * 0.5)) THEN
  	    x := 1.0;
  	ELSIF (y < 1.0) THEN
	   w := y * y;
	   x := ((((((((0.000124818987 * w
		   -0.001075204047) * w +0.005198775019) * w
		   -0.019198292004) * w +0.059054035642) * w
		   -0.151968751364) * w +0.319152932694) * w
		   -0.531923007300) * w +0.797884560593) * y * 2.0;
	ELSE
	    y := y - 2.0;
	    x := (((((((((((((-0.000045255659 * y
		    +0.000152529290) * y -0.000019538132) * y
		    -0.000676904986) * y +0.001390604284) * y
		    -0.000794620820) * y -0.002034254874) * y
		    +0.006549791214) * y -0.010557625006) * y
		    +0.011630447319) * y -0.009279453341) * y
		    +0.005353579108) * y -0.002141268741) * y
		    +0.000535310849) * y +0.999936657524;
    END;
    IF z > 0.0 THEN
        RETURN ( (x + 1.0) * 0.5 );
    ELSE
        RETURN ( (1.0 - x) * 0.5 );
    END;
END poz;

PROCEDURE ex (x:LONGREAL   ):LONGREAL ;
BEGIN
    IF x < (-BIGX) THEN
        RETURN 0.0;
    ELSE
        RETURN MATHLIB.Exp(x);
    END;
END ex;

PROCEDURE pochisq (ax:LONGREAL;df:CARDINAL):LONGREAL;
VAR
    Pi,LOG_SQRT_PI,I_SQRT_PI:LONGREAL;
    x:LONGREAL;
    a,y,s:LONGREAL;
    e,c,z:LONGREAL;
    isEven : BOOLEAN;
BEGIN
    Pi := 4.0 * MATHLIB.ATan(1.0); (* better safe than sorry *)
    LOG_SQRT_PI := MATHLIB.Log ( MATHLIB.Sqrt(Pi) ); (* 0.5723649429247000870717135 = log (sqrt (pi)) *)
    I_SQRT_PI := 1.0 / MATHLIB.Sqrt(Pi); (* 0.5641895835477562869480795 = 1 / sqrt (pi) *)

    x:=ax;

    IF ( (x <= 0.0) OR (df < 1) ) THEN RETURN 1.0; END;

    a := 0.5 * x;
    isEven := NOT( ODD(df));
    IF df > 1 THEN y:=ex(-a);END;

    IF isEven THEN
        s:=y;
    ELSE
        s:= 2.0 * poz(-MATHLIB.Sqrt(x));
    END;
    IF df > 2 THEN
	    x := 0.5 * LONGREAL (df - 1);
	    IF isEven THEN
	        z:=1.0;
	    ELSE
	        z:=0.5;
	    END;
	    IF a > BIGX THEN
    	    IF isEven THEN
    	        e:=0.0;
    	    ELSE
    	        e:=LOG_SQRT_PI;
    	    END;
    	    c := MATHLIB.Log(a);
    	    LOOP (* WHILE z <= x *)
    	        IF z > x THEN EXIT; END;
		        e := MATHLIB.Log(z) + e;
		        s := s + ex(c * z - a - e);
		        z := z+ 1.0;
    	    END;
    	    RETURN s;
    	ELSE
	        IF isEven THEN
 	           e := 1.0;
	        ELSE
	            e:= I_SQRT_PI / MATHLIB.Sqrt(a);
	        END;
	        c := 0.0;
	        LOOP (* WHILE (z <= x) *)
	             IF z > x THEN EXIT; END;
		         e := e * (a / z);
		         c := c + e;
		         z := z + 1.0;
    	    END;
	        RETURN (c * y + s);
    	END;
    ELSE
        RETURN s;
    END;
END pochisq;

(* ------------------------------------------------------------ *)

(* RT_INIT  --  Initialise random test counters.  *)

PROCEDURE rt_init (VAR rt:rt_type;jwBitMode:BOOLEAN );
VAR
    i:CARDINAL;
    v:LONGREAL;
BEGIN
    PI := 4.0 * MATHLIB.ATan(1.0);
    jwtotal := 0;
    FOR i:=firstcode TO maxcode DO jwcount[i]:=0;END;

    rt.montepi      := 0.0;

    lastmonteval    := firstmonteval; (* monte carlo accumulator pointer *)
    montecount      := 0; (* monte carlo tries *)
    montehit        := 0; (* monte carlo inside count *)
    v               := MATHLIB.Pow( 256.0, LONGREAL(MAXMONTEVAL DIV 2) );
    incircle        := 65535.0 * 65535.0; (* In-circle distance for Monte Carlo *)
    incircle        := MATHLIB.Pow( v - 1.0,  2.0 );

    rt.ent          := 0.0; (* clean entropy accumulator *)

    datasum         := 0.0; (* clear sum OF bytes for arithmetic mean *)
    rt.chisq        := 0.0; (* clear Chi-Square *)
    rt.chip         := 0.0;

    sccfirst        := TRUE; (* first time for serial correlation*)
    scct1           := 0.0;  (* serial correlation termps *)
    scct2           := 0.0;
    scct3           := 0.0;

    rt.scc          := 0.0;
    rt.mean         := 0.0;

    sx              := 0.0;
    sx2             := 0.0;
    sxn             := 0;

END rt_init;

PROCEDURE rt_add (ch:CHAR; jwBitMode:BOOLEAN);
CONST
    bitsPerByte = 8;
VAR
    i,chb,v:CARDINAL;
    montex,montey:LONGREAL;
BEGIN
    chb := ORD(ch);

    (* avg AND sigma *)

    sx  := sx + LONGREAL(chb);
    sx2 := sx2 + LONGREAL(chb) * LONGREAL(chb);
    INC(sxn);

    IF jwBitMode THEN
        INC(jwtotal,bitsPerByte);
        v:=chb;
        FOR i:=1 TO bitsPerByte DO
            INC ( jwcount[ v AND 01H ] );
            v:=(v >> 1);
        END;
    ELSE
        INC(jwtotal);
        INC(jwcount[chb]);
    END;

    (* Update inside / outside circle counts for Monte Carlo computation of PI *)

    monteval[lastmonteval]:=LONGREAL(chb); (* save char *)
    INC(lastmonteval);
    IF lastmonteval > MAXMONTEVAL THEN (* calculate every MONTEVAL char *)
        INC(montecount);

        lastmonteval := firstmonteval;

        montex := 0.0;
        montey := 0.0;
        FOR i:=firstmonteval TO (MAXMONTEVAL DIV 2) DO
            montex := (montex * 256.0) + monteval[i];
            montey := (montey * 256.0) + monteval[i+ MAXMONTEVAL DIV 2];
        END;
        IF ( montex * montex + montey * montey ) <= incircle THEN
            INC(montehit);
        END;
    END;

    (* Update calculation of serial correlation coefficient *)

    v:=chb;
    FOR i:=1 TO getjwBitCount( NOT(jwBitMode) ) DO
        IF jwBitMode THEN
            chb:=( (v AND 080H) >> 7 );
        END;
        sccun := LONGREAL( chb );
        IF sccfirst THEN
            sccfirst := FALSE;
            scclast  := 0.0;
            sccu0    := sccun;
        ELSE
            scct1    := scct1 + scclast * sccun;
        END;
        scct2        := scct2 + sccun;
        scct3        := scct3 + (sccun * sccun);
        scclast      := sccun;
        IF jwBitMode THEN v:=(v << 1);END;
    END;

END rt_add;

PROCEDURE rt_end (VAR rt:rt_type; ENT32,jwBitMode,DEBUG:BOOLEAN);
TYPE
    ChiXPtableType      = ARRAY [1..10] OF LONGREAL;
    ChiXPprobaTableType = ARRAY [1..10] OF LONGREAL;
CONST
    chisqt = ChiXPtableType (
    0.5,
    0.25,
    0.1,
    0.05,
    0.025,
    0.01,
    0.005,
    0.001,
    0.0005,
    0.0001);

    (* Table of chi-square Xp values versus corresponding probabilities *)

    chisqtproba = ChiXPprobaTableType (
    0.0,
    0.6745,
    1.2816,
    1.6449,
    1.9600,
    2.3263,
    2.5758,
    3.0902,
    3.2905,
    3.7190);
VAR
    i:CARDINAL;
    cexp,a,fjwtotal:LONGREAL;
    degreesOfFreedom:CARDINAL;
BEGIN
    fjwtotal := LONGREAL(jwtotal);

    (* first compute probas *)

    FOR i:=firstcode TO getjwLast(jwBitMode) DO
        proba[i] := LONGREAL(jwcount[i]) / fjwtotal;
    END;

    (* Calculate Monte Carlo value for PI from percentage of hits within the circle *)
(* dbgf(DEBUG,LONGREAL(montecount),"montecount"); *)
    IF montecount = 0 THEN
        rt.montepi := monteUNDEFINED -1.0; (* //BUGFIXED *)
    ELSE
        rt.montepi := 4.0 * LONGREAL ( montehit ) / LONGREAL( montecount );
    END;

    (* Calculate entropy *)

    FOR i:=firstcode TO getjwLast(jwBitMode) DO
        IF proba[i] > 0.0 THEN
(* dbgf(DEBUG,proba[i],"proba[i]"); *)
            rt.ent := rt.ent + proba[i] * log2(1.0 / proba[i]);
        END;
    END;

    (*
        Scan bins and calculate probability for each bin and
        Chi-Square distribution
    *)

    cexp := fjwtotal / LONGREAL ( getjwLast(jwBitMode) +1 );
    FOR i:=firstcode TO getjwLast(jwBitMode) DO
        a:= LONGREAL(jwcount[i]) - cexp;
(* dbgf(DEBUG,cexp,"cexp"); *)
        rt.chisq := rt.chisq + (a * a) / cexp;
        datasum:=datasum + LONGREAL(i) * LONGREAL(jwcount[i]);
    END;

IF NOT(ENT32) THEN
    (*
        Calculate probability of observed distribution occurring from
        the results of the Chi-Square test
    *)

    rt.chip := MATHLIB.Sqrt(2.0 * rt.chisq) - MATHLIB.Sqrt(2.0 * LONGREAL(getjwLast(jwBitMode))- 1.0);
    a:= ABS(rt.chip);
    i:=10+1;
    LOOP
        DEC(i);
        IF i < 1 THEN EXIT; END;
        IF chisqtproba[i] < a THEN EXIT; END;
    END;
    IF rt.chip >= 0.0 THEN
        rt.chip := chisqt[i];
    ELSE
        rt.chip := 1.0 - chisqt[i];
    END;
ELSE
    degreesOfFreedom:= getjwLast(jwBitMode);
    rt.chip := pochisq(rt.chisq,degreesOfFreedom); (* already percentage *)
END;
    (* Complete calculation of serial correlation coefficient *)

    scct1 := scct1 + scclast * sccu0;
    scct2 := scct2 * scct2;
    rt.scc   := fjwtotal * scct3 - scct2;
    IF rt.scc = 0.0 THEN
       rt.scc := sccUNDEFINED -1.0;
    ELSE
(* dbgf(DEBUG,rt.scc,"rt.scc"); *)
       rt.scc := (fjwtotal * scct1 - scct2) / rt.scc;
    END;

    rt.mean := datasum / fjwtotal;

END rt_end;

(* ------------------------------------------------------------ *)

PROCEDURE procEntropy (path:pathtype;
                      useLFN,ENT32,jwBitMode,jwShowTable,jwCSV,
                      jwLowercase,jwUppercase,jwOrgRpt,DEBUG:BOOLEAN):BOOLEAN;

CONST
    msg = "Working, please wait... ";
VAR
    got    : CARDINAL;
    hnd    : FIO.File;
    ch     : CHAR;
    rt     : rt_type;
BEGIN
(* dbg(DEBUG,"rt_init"); *)
    rt_init(rt,jwBitMode); (* Initialise for calculations *)

    hnd:=fileOpenRead(useLFN,path);
    IF FIO.Size(hnd) = 0 THEN
        fileClose(useLFN,hnd);
        RETURN FALSE; (* empty file *)
    END;
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=TRUE; (* useless safety ! *)
    video(msg,TRUE);
    LOOP
        got := FIO.RdBin(hnd,ch,1);
        IF got # 1 THEN EXIT; END;
        IF jwUppercase THEN
            IF ch # CHR(0) THEN UpperCase(ch);END;
        END;
        IF jwLowercase THEN
            IF ch # CHR(0) THEN LowerCaseAlt(ch);END;
        END;
(* dbg(DEBUG,printable(ch)); *)
(* dbg(DEBUG,"rt_add"); *)
        rt_add(ch,jwBitMode); (* Update counter for this bin *)
    END;
    video(msg,FALSE);
    fileClose(useLFN,hnd);
(* dbg(DEBUG,"rt_end"); *)
    rt_end(rt,ENT32,jwBitMode,DEBUG); (* Complete calculation and return sequence metrics *)

    doReport(ENT32,jwBitMode,jwShowTable,jwCSV,jwOrgRpt,rt,path);
    RETURN TRUE;
END procEntropy;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    widataxy      = countcode;
    hedataxy      = countcode;
    kxy           = 2;
    COLORFRAMEALT1= cLIGHTGREEN;
    COLORFRAMEALT2= cLIGHTRED;
    COLORLIT      = cWHITE;
    COLORHI       = cBRIGHTWHITE;
    COLORSTATUSALT= cLIGHTYELLOW;
VAR
    globx,globy:SHORTCARD;

PROCEDURE statusline (S:ARRAY OF CHAR   );
VAR
    xy:Graph.TextCoords;
    hpos,vpos,color,i:CARDINAL;
BEGIN
    color := COLORSTATUSALT;
    i:=Graph.SetTextColor(color);
    vpos  := vtabMin+vtabHeight-4;
    hpos:=centerme(htabMin,htabWidth,S);
    xy:=Graph.SetTextPosition(vpos,hpos);
    Graph.OutText(S);
END statusline;

CONST
    msgWorkingXY = "Working, please wait...";
    msgAnykeyXY  = "[Tab]-toggle unlit  [Escape]-quit"; (* at least same lenght as above *)

PROCEDURE openScreenXY;
VAR
    xbase,ybase,ink:CARDINAL;
BEGIN
    setupVideoScreen();

    xbase := screenCX - (widataxy DIV 2);
    ybase := screenCY - (hedataxy DIV 2);

    ink   := COLORFRAMEALT1;
    Graph.Rectangle(xbase-kxy,ybase-kxy,xbase+widataxy+kxy-1,ybase+hedataxy+kxy-1,ink,FALSE);
    statusline(msgWorkingXY);
END openScreenXY;

PROCEDURE closeScreenXY;
VAR
    charcode:CARDINAL;
    hilite:BOOLEAN;
    x,y,xx,yy:CARDINAL;
    xbase,ybase:CARDINAL;
    ink:CARDINAL;
BEGIN
    hilite:=FALSE;
    statusline(msgAnykeyXY);
    LOOP
        IF isKeyWaiting() THEN
            charcode:=ReadKeycode();
            CASE charcode OF
            | keyTab :

                IF hilite THEN
                    ink:=COLORFRAMEALT1;
                ELSE
                    ink:=COLORFRAMEALT2;
                END;
                xbase := screenCX - (widataxy DIV 2);
                ybase := screenCY - (hedataxy DIV 2);
                Graph.Rectangle(xbase-kxy,ybase-kxy,xbase+widataxy+kxy-1,ybase+hedataxy+kxy-1,ink,FALSE);

                hilite:=NOT(hilite);
                FOR yy := firstcode TO maxcode DO
                    FOR xx:=firstcode TO maxcode DO
                        x:=screenCX - (widataxy DIV 2) + CARDINAL(xx);
                        y:=screenCY - (hedataxy DIV 2) + CARDINAL(yy);
                        CASE Graph.Point(x,y) OF
                        | COLORBK: Graph.Plot(x,y,COLORHI);
                        | COLORHI: Graph.Plot(x,y,COLORBK);
                        END;
                    END;
                END;

            ELSE
                EXIT;
            END;
        END;
    END;
END closeScreenXY;

PROCEDURE processxy (VAR okpair:BOOLEAN;
                    useLFN:BOOLEAN; path:pathtype);
VAR
    hnd:FIO.File;
    got,x,y,color:CARDINAL;
BEGIN
    hnd:=fileOpenRead(useLFN,path);
    IF FIO.Size(hnd) = 0 THEN
        fileClose(useLFN,hnd);
        RETURN;
    END;
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.EOF:=TRUE; (* useless safety ! *)
    LOOP
        IF okpair THEN
            got:=FIO.RdBin(hnd,globx,1);
            IF got # 1 THEN EXIT; END;
            got:=FIO.RdBin(hnd,globy,1);
            IF got # 1 THEN okpair:=FALSE; EXIT; END;
        ELSE
            got:=FIO.RdBin(hnd,globy,1);
            IF got # 1 THEN okpair:=FALSE; EXIT; END;
            okpair:=TRUE;
        END;
        x:=screenCX - (widataxy DIV 2) + CARDINAL(globx);
        y:=screenCY - (hedataxy DIV 2) + CARDINAL(globy);
        color:=COLORLIT;
        Graph.Plot(x,y,color);
    END;
    fileClose(useLFN,hnd);
END processxy;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* yes, yes, we could have used our "clever" dynamic allocation but we don't care *)

VAR
    filename  : ARRAY [firstfile..maxfile] OF pathtype;

PROCEDURE WrQuoted (S:ARRAY OF CHAR);
BEGIN
    WrStr(dquote);  WrStr(S);  WrStr(dquote);
END WrQuoted;

PROCEDURE WrFname (useLFN:BOOLEAN;S:pathtype );
BEGIN
    IF useLFN THEN
        WrQuoted(S);
    ELSE
        WrStr(S);
    END;
END WrFname;

PROCEDURE isReservedEntry (S:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE buildFileList (useLFN,DEBUG :BOOLEAN;spec:pathtype):CARDINAL;
VAR
    last:CARDINAL; (* should do ! *)
    ok,found:BOOLEAN;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    entry : FIO.DirEntry;
    dosattr:FIO.FileAttr;
    entryname:pathtype;
    excludeme1:BOOLEAN;
BEGIN
    last:=firstfile-1;
    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,entry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(entryname,w9Xentry.fullfilename);
        ELSE
            Str.Copy(entryname,entry.Name);
        END;
        excludeme1 := isReservedEntry   (entryname);  (* skip "." and ".." *)
        IF NOT(excludeme1) THEN
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=entry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                (* if file has no extension, add it as a marker *)
                IF Str.RCharPos(entryname,".")=MAX(CARDINAL) THEN
                    Str.Append(entryname,".");
                END;
                IF DEBUG THEN WrStr("Included : ");WrFname(useLFN,entryname);WrLn; END;
                INC(last);
                IF last > maxfile THEN
                    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL);
                END;
                Str.Copy(filename[last],entryname);
            ELSE
                IF DEBUG THEN WrStr("Ignored  : ");WrFname(useLFN,entryname);WrLn;END;
            END;
        ELSE
            IF DEBUG THEN WrStr("Excluded : ");WrFname(useLFN,entryname);WrLn;END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(entry);
        END;
    END;
    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
    RETURN last;
END buildFileList;

PROCEDURE chkUD (S:pathtype):BOOLEAN ;
VAR
    u,d,n,e:pathtype;
    pb:CARDINAL;
BEGIN
    qdSplitAllPath(S , u,d,n,e);
    pb:=0;
    IF chkJoker(u) THEN INC(pb);END;
    IF chkJoker(d) THEN INC(pb);END;
    RETURN (pb=0);
END chkUD;

PROCEDURE doGetCurrent (useLFN:BOOLEAN;drive:SHORTCARD;
                       VAR unit:str2; VAR current:pathtype);
VAR
    rc:CARDINAL;
    longform:pathtype;
BEGIN
    Str.Concat(unit, CHR( ORD("A")-1+ORD(drive) ),colon);

    FIO.GetDir(drive,current); (* \path without u: nor trailing \ except at root *)
    IF current[1] # colon THEN Str.Prepend(current,unit); END; (* safety *)
    IF useLFN THEN
        IF w9XshortToLong(current,rc,longform) THEN (* if error, keep DOS current *)
            Str.Copy(current,longform);
        END;
    END;
    (* LFN function seems to always return "u:\*" form except at root *)
    IF current[1] = colon THEN Str.Delete(current,0,2);END; (* safety *)
    fixDirectory(current);
END doGetCurrent;

PROCEDURE makebase (useLFN:BOOLEAN;spec:pathtype;VAR basepath:pathtype);
VAR
    u,d,n,e,current:pathtype;
    unit:str2;
    drive:SHORTCARD;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Lib.MakeAllPath(basepath,u,d,"","");
    IF same(basepath,"") THEN
        drive:=FIO.GetDrive(); (* yes we could use 0 as default drive *)
        doGetCurrent(useLFN,drive,  unit,current); (* "u:" and "\" or "\*\" *)
        Str.Concat(basepath, unit,current);
    END;
END makebase;

(* ------------------------------------------------------------ *)

TYPE
    cmdtype = (cmdUndefined,cmdCount,cmdHiresView,cmdEntropy,cmdXY);

PROCEDURE newcmd (VAR cmd:cmdtype;wanted:cmdtype  ):BOOLEAN;
VAR
    ok:BOOLEAN;
BEGIN
    IF ( (cmd = cmdUndefined) OR (cmd = wanted) ) THEN
        cmd:=wanted;
        ok:=TRUE;
    ELSE
        ok:=FALSE;
    END;
    RETURN ok;
END newcmd;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    defaultBarWidth = 25;
    minBarWidth     = 1;
    maxBarWidth     = 200;
    polleveryDOS    = 5;
    polleveryWin    = 25;
VAR
    currpoll,pollevery:CARDINAL;
    restricted,percentage,excludecontrol,semigraphics : BOOLEAN;
    eyecandy,hexval,decval,showheader,showglobal,showall,showbar:BOOLEAN;
    sortmode,barwi:CARDINAL;
    ignorecase,ignoreaccents : BOOLEAN;
    jwBitMode,jwShowTable,jwCSV,jwLowercase,jwUppercase,jwOrgRpt,DEBUG:BOOLEAN;
    ENT32,okcoor,alcatraz,useLFN,ok:BOOLEAN;
    cmd : cmdtype;
VAR
    parmcount : CARDINAL;
    i,opt     : CARDINAL;
    S,R       : pathtype;
    status    : (waiting,gotspec);
    filespec,basedir,path : pathtype;
    lastfile:CARDINAL;
    lc:LONGCARD;
BEGIN
    FIO.IOcheck := FALSE;
    WrLn; (* yes, here now for help, errors, etc. *)

    restricted     := TRUE;
    percentage     := FALSE;
    excludecontrol := FALSE;
    sortmode       := SORTnone;
    eyecandy       := FALSE;
    hexval         := FALSE;
    decval         := FALSE;
    showheader     := TRUE;
    showglobal     := FALSE;
    showall        := TRUE;
    showbar        := FALSE;
    semigraphics   := TRUE;
    ignorecase     := FALSE;
    ignoreaccents  := FALSE;
    barwi          := defaultBarWidth;
    useLFN         := TRUE;
    ENT32          := TRUE;
    DEBUG          := FALSE;

    jwBitMode      := FALSE; (* ent -b *)
    jwShowTable    := FALSE; (* ent -c *)
    jwCSV          := FALSE; (* ent -t *)
    jwLowercase    := FALSE; (* ent -f NOT ! *)
    jwUppercase    := FALSE;
    jwOrgRpt       := FALSE; (* EN original format for results *)

    cmd            := cmdUndefined;

    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;

    status:=waiting;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "A"+delim+"ALL"+delim+
                                  "P"+delim+"PERCENTAGE"+delim+
                                  "C"+delim+"CONTROL"+delim+
                                  "E"+delim+"EYECANDY"+delim+
                                  "S"+delim+"SORT"+delim+
                                  "R"+delim+"REVERSE"+delim+
                                  "X"+delim+"HEX"+delim+
                                  "D"+delim+"DEC"+delim+
                                  "T"+delim+"TERSE"+delim+
                                  "G"+delim+"GLOBAL"+delim+
                                  "GG"+delim+
                                  "B"+delim+"BARGRAPH"+delim+
                                  "BB"+delim+
                                  "W:"+delim+"WIDTH:"+delim+
                                  "V"+delim+"VIEW"+delim+
                                  "I"+delim+"CASE"+delim+
                                  "II"+delim+"ACCENTS"+delim+

                                  "EE"+delim+"ENTROPY"+delim+"ENT"+delim+
                                  "EB"+delim+"BITS"+delim+
                                  "EC"+delim+"TABLE"+delim+
                                  "ET"+delim+"CSV"+delim+
                                  "EL"+delim+"LOWER"+delim+
                                  "EU"+delim+"UPPER"+delim+
                                  "ER"+delim+"REPORT"+delim+"RPT"+delim+"ALT"+delim+
                                  "??"+delim+
                                  "???"+delim+
                                  "VV"+delim+"XY"+delim+
                                  "DOS"+delim+"OLD"+delim+
                                  "L"+delim+"LFN"+delim+
                                  "DEBUG"
                                  );
            CASE opt OF
            | 1,2,3 :   abort(errHelp,"");
            | 4,5:      restricted     := FALSE;
            | 6,7:      percentage     := TRUE;
            | 8,9:      excludecontrol := TRUE;
            | 10,11:    eyecandy       := TRUE;
            | 12,13:    CASE sortmode OF
                        | SORTnone,SORTnormal: sortmode:=SORTnormal;
                        ELSE
                            abort(errNonsenseSort,"");
                        END;
            | 14,15:    CASE sortmode OF
                        | SORTnone,SORTreverse: sortmode:=SORTreverse;
                        ELSE
                            abort(errNonsenseSort,"");
                        END;
            | 16,17:    hexval         := TRUE;
            | 18,19:    decval         := TRUE;
            | 20,21:    showheader     := FALSE;
            | 22,23:    showglobal     := TRUE;
            | 24:       showglobal     := TRUE; showall:=FALSE;
            | 25,26:    showbar        := TRUE;
            | 27:       showbar        := TRUE; semigraphics:=FALSE;
            | 28,29:    IF GetLongCard(S,lc)=FALSE THEN abort(errBadNumber,S);END;
                        IF (lc < LONGCARD(minBarWidth)) OR (lc > LONGCARD(maxBarWidth)) THEN
                            abort(errBarWidthRange,"");
                        END;
                        barwi:=CARDINAL(lc);
            | 30,31:    IF newcmd(cmd,cmdHiresView)=FALSE THEN abort(errCmd,"");END;
            | 32,33:    ignorecase := TRUE;
            | 34,35:    ignorecase := TRUE; ignoreaccents:=TRUE;

            | 36,37,38: IF newcmd(cmd,cmdEntropy)=FALSE THEN abort(errCmd,"");END;
            | 39,40:    jwBitMode    := TRUE;
            | 41,42:    jwShowTable  := TRUE;
            | 43,44:    jwCSV        := TRUE;
            | 45,46:    jwLowercase  := TRUE;
            | 47,48:    jwUppercase  := TRUE;
            | 49,50,51,52: jwOrgRpt  := TRUE;
            | 53:       abort(errMoreHelp,"");
            | 54:       abort(errEvenMoreHelp,"");
            | 55,56:    IF newcmd(cmd,cmdXY)=FALSE THEN abort(errCmd,"");END;
            | 57,58:    ENT32        := FALSE;
            | 59,60:    useLFN       := FALSE;
            | 61:       DEBUG        := TRUE;
            ELSE
                abort(errUnknownOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE status OF
            | waiting : Str.Copy(filespec,S);
            | gotspec : abort(errTooManyParms,S);
            END;
            INC (status);
        END;
    END;
    IF status=waiting THEN abort(errMissingSpec,"");END;

    useLFN := ( useLFN AND fileSupportLFN() );

    IF useLFN THEN
        pollevery:=polleveryWin; (* vindoze does not like frequent calls to chkEscape() *)
    ELSE
        pollevery:=polleveryDOS;
    END;

    IF cmd=cmdUndefined THEN cmd:=cmdCount;END;

    (* a few q&d sanity checks *)

    IF chkUD(filespec)=FALSE THEN abort(errBadSpec,filespec);END;
    IF same(filespec,dot) THEN Str.Copy(filespec,stardotstar); END; (* "." *)
    IF same(filespec,star) THEN Str.Copy(filespec,stardotstar); END; (* "*" *)
    IF Str.Match(filespec,"*"+antislash) THEN Str.Append(filespec,stardotstar); END; (* "*\" *)
    IF chkJoker(filespec)=FALSE THEN
        IF fileIsDirectorySpec ( useLFN,filespec) THEN
            fixDirectory(filespec);
            Str.Append(filespec,stardotstar);
        END;
    END;

    makebase(useLFN,filespec,basedir);
    lastfile := buildFileList(useLFN,DEBUG,filespec);

    CASE lastfile OF
    | firstfile-1 :  abort(errNoMatch,filespec);
    | MAX(CARDINAL): abort(errTooManyMatches,filespec);
    END;

    (* let's work now *)

    CASE cmd OF
    | cmdHiresView:
        Graph.GetVideoConfig(gvideoconfig);
        IF gvideoconfig.adapter # Graph._VGA THEN abort(errVGAneeded,""); END;
        showheader     := FALSE; (* -t *)
        showglobal     := TRUE;  (* -g  *)
        showall        := FALSE; (* -gg *)
        restricted     := FALSE; (* -a *)
        excludecontrol := FALSE; (* -c ignored *)
    END;

    CASE cmd OF
    | cmdEntropy:

        currpoll:=0;
        FOR i := firstfile TO lastfile DO
            IF i > firstfile THEN WrLn;END;
            Str.Concat(path,basedir,filename[i]);
            ok:= procEntropy (path, useLFN,ENT32,jwBitMode,jwShowTable,jwCSV,
                             jwLowercase,jwUppercase,jwOrgRpt,DEBUG);
            IF NOT(ok) THEN doProblem(jwOrgRpt,path);END;
            INC(currpoll);
            IF (currpoll MOD pollevery) = 0 THEN
                IF ChkEscape() THEN abort(errEscape,"");END;
                currpoll:=0;
            END;
        END;

    | cmdCount,cmdHiresView:
        IF showheader THEN
            WrStr(msgRem+" "+banner);WrLn;
            WrLn;
        END;

        IF showglobal THEN
            FOR i:=firstcode TO maxcode DO gcount[i]:=0; END;
            gtotal     :=0;
            gctrlchars :=0;
            gfilesize  :=0;
        END;

        currpoll:=0;
        FOR i := firstfile TO lastfile DO
            Str.Concat(path,basedir,filename[i]);
            process (path, useLFN,excludecontrol,eyecandy,showglobal,
                    ignorecase,ignoreaccents,
                    filesize,total,ctrlchars, gfilesize,gtotal,gctrlchars);
            IF showall THEN
                IF filesize # 0 THEN
                    IF i # firstfile THEN WrLn; END;
                    dump (path,restricted,percentage,excludecontrol,
                         hexval,decval,
                         eyecandy,showheader,showglobal,showbar,semigraphics,
                         ignorecase,ignoreaccents,
                         filesize,total,ctrlchars,sortmode,barwi);
                END;
            END;
            INC(currpoll);
            IF (currpoll MOD pollevery) = 0 THEN
                IF ChkEscape() THEN abort(errEscape,"");END;
                currpoll:=0;
            END;
        END;

        IF showglobal THEN
            FOR i:=firstcode TO maxcode DO  count[i]:=gcount[i]; END;
            CASE cmd OF
            | cmdHiresView :
                gvmode   := Graph._MRES16COLOR;
                S        := "320x200x4";
                IF Graph.SetVideoMode(gvmode)=FALSE THEN abort(errCannotSet,S); END;

                procView(sortmode);

                gvmode   := Graph._DEFAULTMODE;
                S        := "original";
                IF Graph.SetVideoMode(gvmode)=FALSE THEN abort(errCannotSet,S); END;
            | cmdCount:
                Str.Copy(path,"all matching files");
                WrLn;
                dump (path,restricted,percentage,excludecontrol,
                     hexval,decval,
                     eyecandy,showheader,showglobal,showbar,semigraphics,
                     ignorecase,ignoreaccents,
                     gfilesize,gtotal,gctrlchars,sortmode,barwi);
            END;
        END;
    | cmdXY:
        gvmode   := Graph._VRES16COLOR;
        S        := "640x480x4";
        IF Graph.SetVideoMode(gvmode)=FALSE THEN abort(errCannotSet,S); END;

        openScreenXY;

        alcatraz:=FALSE;
        okcoor:=TRUE;
        currpoll:=0;
        i:=firstfile-1;
        LOOP
            INC(i);
            IF i > lastfile THEN EXIT; END;
            Str.Concat(path,basedir,filename[i]);
            processxy (okcoor, useLFN,path);
            INC(currpoll);
            IF (currpoll MOD pollevery) = 0 THEN
                IF ChkEscape() THEN alcatraz:=TRUE;EXIT;END;
                currpoll:=0;
            END;
        END;

        IF NOT(alcatraz) THEN closeScreenXY; END;

        gvmode   := Graph._DEFAULTMODE;
        S        := "original";
        IF Graph.SetVideoMode(gvmode)=FALSE THEN abort(errCannotSet,S); END;

        IF alcatraz THEN abort(errEscape,"");END;
    END;

    abort(errNone,"");
END cCount.





(*

@echo off
if "%1" == "" goto help

set _ew_=c:\z\ent32.rpt
set _ed_=c:\z\entdos.rpt

set _cw_=c:\z\cc.rpt
set _cd_=c:\z\ccdos.rpt

if not exist %1 goto pb

ent      %1         > %_ew_%
ent -c   %1        >> %_ew_%
ent -b   %1        >> %_ew_%
ent -bc  %1        >> %_ew_%
ent -t   %1        >> %_ew_%
ent -ct  %1        >> %_ew_%
ent -bt  %1        >> %_ew_%
ent -bct %1        >> %_ew_%

entdos      %1      > %_ed_%
entdos -c   %1     >> %_ed_%
entdos -b   %1     >> %_ed_%
entdos -bc  %1     >> %_ed_%
entdos -t   %1     >> %_ed_%
entdos -ct  %1     >> %_ed_%
entdos -bt  %1     >> %_ed_%
entdos -bct %1     >> %_ed_%

ccount -ee              -rpt %1       > %_cw_%
ccount -ee  -ec         -rpt %1      >> %_cw_%
ccount -ee  -eb         -rpt %1      >> %_cw_%
ccount -ee  -eb -ec     -rpt %1      >> %_cw_%
ccount -ee  -et         -rpt %1      >> %_cw_%
ccount -ee  -ec -et     -rpt %1      >> %_cw_%
ccount -ee  -eb -et     -rpt %1      >> %_cw_%
ccount -ee  -eb -ec -et -rpt %1      >> %_cw_%

ccount -ee              -rpt -dos %1  > %_cd_%
ccount -ee  -ec         -rpt -dos %1 >> %_cd_%
ccount -ee  -eb         -rpt -dos %1 >> %_cd_%
ccount -ee  -eb -ec     -rpt -dos %1 >> %_cd_%
ccount -ee  -et         -rpt -dos %1 >> %_cd_%
ccount -ee  -ec -et     -rpt -dos %1 >> %_cd_%
ccount -ee  -eb -et     -rpt -dos %1 >> %_cd_%
ccount -ee  -eb -ec -et -rpt -dos %1 >> %_cd_%

call cga
call vga

filecomp %_ed_% %_cd_%
pause
filecomp %_ew_% %_cw_%
goto end

:help
echo.
echo Syntax : %0 file
echo.
echo Note ENT will crash or display nonsense if filesize is smaller than 6.
echo.
goto end

:pb
echo.
echo %1 does not exist !
echo.
goto end

:end
set _ew_=
set _ed_=
set _cw_=
set _cd_=

*)

