(* ---------------------------------------------------------------
Title         Q&D Check File Readability
Overview
Usage         see help
Notes         remember to set idNOW according to O.S. ! //DONOTFORGET
Bugs          no critical error handler
Wish List     rewrite each file to a tmp file
              before rename then del if everything went fine ?
              (handle save/restore attributes)

              better report with not found, not opened, missing ?
              show percentage of operations ?
              add companion file for exclusions ?
              handle delta in .RSM

--------------------------------------------------------------- *)

MODULE chkFread;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT DOSErr;

FROM IO IMPORT WrStr, WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    progEXEname      = "CHKFREAD";
    progTitle        = "Q&D Check File Readability";
    progVersion      = "v1.0c";
    progCopyright    = "by PhG";
    progBanner       = progTitle+" "+progVersion+" "+progCopyright;
CONST
    extLST           = ".LST";
    extRSM           = ".RSM";
    msgJoker         = "::: Joker(s) : ";
    msgNotFile       = "::: Not file : ";
    msgNotFound      = "::: No match : ";
    msgReadOK        = "+++ Success  : ";
    msgReadError     = "--- Failure  : (#) ";
CONST
    msgSkipped       = "::: Skipped  : ";
    msgFailed        = "::: Failed   : ";
    msgProcessed     = "::: Verified : ";
    msgElapsed       = "::: Elapsed time was ";
    wilc             = 10;
    msgRESUME        = '::: "~" resume file has been created.';
    msgLASTPROCESSED = '::: Last processed file was "~"';
CONST
    errNone            = 0;
    errHelp            = 1;
    errExpecting       = 2;
    errOption          = 3;
    errTooManyParms    = 4;
    errNoMatch         = 5;
    errJoker           = 6;
    errDir             = 7;
    errCaseRedefined   = 8;
    errCorrupt         = 9;
    errBadOS           = 10;
    errReserved        = 11;
    errAborted         = 255;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
msgHelp =
progBanner+nl+
nl+
"Syntax : "+progEXEname+" <list["+extLST+"]> [option]..."+nl+
nl+
"This program verifies each file entry in <list> is readable."+nl+
"Possible usefulness should be self-explanatory."+nl+
nl+
"    -r  resume from last file entry read"+nl+
"    -v  verbose mode (do not report only failures but successes too)"+nl+
"    -u  display file entries in uppercase"+nl+
"    -w  display file entries in lowercase"+nl+
"    -q  do not show list completion percentage"+nl+
"    -x  disable LFN support even if available"+nl+
nl+
"a) Each file entry in <list> is expected to be an ASCII DOS valid one."+nl+
"   Jokers are not supported."+nl+
"b) <list> should be filtered for entries locked by operating system."+nl+
"c) A critical error handler TSR might be useful."+nl+
'd) Should program be aborted with ESCape key,'+nl+
'   a resume file with "'+extRSM+'" extension will be created in current directory'+nl+
"   so program can resume from last file entry read thanks to -r option."+nl+
"e) Note even with -r option, elapsed time applies to current program run only."+nl;

VAR
    S : str1024; (* oversized for possible LFN *)
BEGIN
    CASE e OF
    | errHelp :          WrStr(msgHelp);
    | errExpecting:      S := "Expecting <list> !";
    | errOption:         S := 'Unknown "~" option !';
    | errTooManyParms:   S := '"~" is one parameter too many !';
    | errNoMatch:        S := '"~" does not exist !';
    | errJoker:          S := '"~" should not contain any joker !';
    | errDir:            S := '"~" seems to be a directory !';
    | errCaseRedefined:  S := "-u and -w options are mutually exclusive !";
    | errCorrupt:        S := '"~" seems damaged !';
    | errBadOS:          S := '"~" requires another O.S. !';
    | errReserved:       S := '"~" should not have reserved "'+extRSM+'" extension !';
    | errAborted  :      S := "Aborted by user !";
    ELSE
                         S := "This is illogical, Captain !";
    END;

    CASE e OF
    | errNone,errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (16 * 512) + FIO.BufferOverhead; (* was 8 *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferLst,ioBufferIn : ioBufferType;

CONST
    dataBufferSize      = (16 * 512);
    firstDataBuffer     = 1-1;
    lastDataBuffer      = dataBufferSize-1;
VAR
    dataBuffer : ARRAY [firstDataBuffer..lastDataBuffer] OF BYTE;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    CHKEVERY9X = 64; (* let's call chkEscape every CHKEVERY LOOP *)
    CHKEVERYDOS= 4;  (* safety *)

PROCEDURE pollEscape (VAR chkrounds:CARDINAL; useLFN,ignoreESC:BOOLEAN):BOOLEAN;
VAR
    alcatraz:BOOLEAN;
    every:CARDINAL;
BEGIN
    (*
    9x really does NOT like ChkEscape() and randomly hangs till a keypress
    Flushkey() does NOT help
    *)
    (*
    IF useLFN THEN
        hit:=FALSE;
    ELSE
        hit:=ChkEscape();
    END;
    *)
    IF ignoreESC THEN RETURN FALSE;END;
    IF useLFN THEN
        every:=CHKEVERY9X;
    ELSE
        every:=CHKEVERYDOS;
    END;
    INC(chkrounds);
    IF (chkrounds MOD every) = 0 THEN
        chkrounds:=0;
        alcatraz:=ChkEscape();
    ELSE
        alcatraz:=FALSE;
    END;
    RETURN alcatraz;
END pollEscape;

PROCEDURE preprocessLine (VAR S:pathtype);
VAR
    len:CARDINAL;
BEGIN
    LtrimBlanks(S);
    RtrimBlanks(S);
    (* we could use Str.Match(S,dquote+"*"+dquote) but... *)
    len:=Str.Length(S);
    IF len < 2 THEN RETURN; END; (* at least 1+1 for possible dquote+dquote *)
    IF S[len-1] # dquote THEN RETURN; END;
    IF S[0] # dquote THEN RETURN; END;
    S[len-1]:=charnull;
    Str.Delete(S,0,1);
    (* just in case *)
    LtrimBlanks(S);
    RtrimBlanks(S);
END preprocessLine;

PROCEDURE dmp (enquote:BOOLEAN;S1,S2:ARRAY OF CHAR   );
BEGIN
    WrStr(S1);
    IF enquote THEN WrStr(dquote);END;
    WrStr(S2);
    IF enquote THEN WrStr(dquote);END;
    WrLn;
END dmp;

PROCEDURE fmtlc (v:LONGCARD;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str16;
VAR
    R : str16;
    ok: BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    FOR i:= Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,ch);
        ELSE
            Str.Prepend(R,ch);
        END;
    END;
    IF base=16 THEN Str.Lows(R);END;
    Str.Prepend(R,prefix);
    RETURN R;
END fmtlc;

PROCEDURE fmtc (v:CARDINAL;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str16;
BEGIN
    RETURN fmtlc(LONGCARD(v),base,wi,ch,prefix);
END fmtc;

PROCEDURE buildResumeFilename (VAR R:pathtype; ext,S:pathtype);
VAR
    u,d,n,e:pathtype;
BEGIN
    qdSplitAllPath(S,u,d,n,e);
    Str.Concat(R,n,ext);
END buildResumeFilename;

(* ------------------------------------------------------------ *)

TYPE
    DTtype = RECORD
        Year      : CARDINAL;
        Month     : CARDINAL;
        Day       : CARDINAL;
        Hours     : CARDINAL;
        Minutes   : CARDINAL;
        Seconds   : CARDINAL;
        Hundredths: CARDINAL;
        DayOfWeek : Lib.DayType;
    END;

PROCEDURE GetDateTimeNow (VAR dt : DTtype);
BEGIN
    Lib.GetDate(dt.Year,dt.Month,dt.Day,dt.DayOfWeek);
    Lib.GetTime(dt.Hours,dt.Minutes,dt.Seconds,dt.Hundredths);
END GetDateTimeNow;

(*
    gregorian calendar rule : leap 366 years if divisible by 4
    but centurial years NOT divisible by 4 are common
    Leap years are divisible by 400, or by 4 and not by 100
    1900 is common, 2000 is leap
*)

PROCEDURE getDaysInFebruary (annee:CARDINAL):CARDINAL;
CONST
    common = 28;
    leap   = 29;
BEGIN
    IF (annee MOD 400) = 0 THEN RETURN leap; END;
    IF (((annee MOD 4) = 0) AND ((annee MOD 100) # 0)) THEN RETURN leap; END;
    RETURN common;
END getDaysInFebruary;

(* assume month and year are correct *)

PROCEDURE getDaysInMonth (mois,annee:CARDINAL):CARDINAL;
TYPE
    daysinmonthtype = ARRAY [1..12] OF CARDINAL;
CONST
    (*                    JanFevMarAprMayJunJulAugSepOctNovDec *)
    (*                     1  2  3  4  5  6  7  8  9 10 11 12  *)
    dpm = daysinmonthtype(31, 0,31,30,31,30,31,31,30,31,30,31);
VAR
    m : CARDINAL;
BEGIN
    m:= dpm[mois];
    IF m=0 THEN m:=getDaysInFebruary(annee);END;
    RETURN m;
END getDaysInMonth;

(* jan 1 = 1, etc. *)

PROCEDURE getCountOfDays (jour,mois,annee:CARDINAL):CARDINAL;
VAR
    n,i : CARDINAL;
BEGIN
    n:=jour;
    FOR i:= 1 TO (mois-1) DO
        INC(n, getDaysInMonth(mois,annee) );
    END;
    RETURN n;
END getCountOfDays;

(* out of lazyness, reasonably assume process won't take more than 24 hours ! *)

CONST
    strFmtTime    = "~h ~mn ~s";

PROCEDURE fmtDelta (started,ended:DTtype):str80;
CONST
    minutesPerHour   = 60;
    secondsPerMinute = 60;
    hundred          = 100;
    midnight         = 24*minutesPerHour*secondsPerMinute*hundred;
    baseyear         = 1980; (* historic ! *)
VAR
    R : str80;
    n,hmsStart,hmsEnd,jmaStart,jmaEnd : LONGCARD; (* in hundredths of seconds *)
    delta : DTtype;
    i,v:CARDINAL;
BEGIN
    n := LONGCARD(started.Hours) * minutesPerHour + LONGCARD(started.Minutes);
    n := n * secondsPerMinute + LONGCARD(started.Seconds);
    n := n * hundred + LONGCARD(started.Hundredths);
    hmsStart := n;

    n := LONGCARD(ended.Hours)   * minutesPerHour + LONGCARD(ended.Minutes);
    n := n * secondsPerMinute + LONGCARD(ended.Seconds);
    n := n * hundred + LONGCARD(ended.Hundredths);
    hmsEnd := n;

    n:=LONGCARD ( getCountOfDays(started.Day,started.Month,started.Year) );
    v:=started.Year-1;
    FOR i:= baseyear TO v DO
         INC (n, LONGCARD ( getCountOfDays(31,12,i) ) );
    END;
    jmaStart:=n;

    n:=LONGCARD ( getCountOfDays(ended.Day,ended.Month,ended.Year) );
    v:=ended.Year-1;
    FOR i:= baseyear TO v DO
         INC (n, LONGCARD ( getCountOfDays(31,12,i) ) );
    END;
    jmaEnd:=n;

    IF jmaEnd # jmaStart THEN (* assume no more than one day ! *)
        hmsEnd := hmsEnd + midnight;
    END;
    n := hmsEnd-hmsStart; (* always >= 0 *)
    delta.Hours     :=CARDINAL (n DIV (minutesPerHour*secondsPerMinute*hundred) );
    delta.Minutes   :=CARDINAL((n DIV (               secondsPerMinute*hundred) ) MOD 60 );
    delta.Seconds   :=CARDINAL((n DIV (                                hundred) ) MOD 60 );
    delta.Hundredths:=CARDINAL (n MOD hundred);

    R:=strFmtTime;
    Str.Subst(R,tilde, fmtc(delta.Hours,  10,2,"0",""));
    Str.Subst(R,tilde, fmtc(delta.Minutes,10,2,"0",""));
    Str.Subst(R,tilde, fmtc(delta.Seconds,10,2,"0",""));

    RETURN R;
END fmtDelta;

(* ------------------------------------------------------------ *)

PROCEDURE procFile (VAR errcode:CARDINAL;
                   useLFN,showpercentage:BOOLEAN;
                   total,current:LONGCARD;S:pathtype):BOOLEAN;
CONST
    tersemsg        = "::: Reading, please wait... ";
    informativemsg  = "::: Reading ~ out of ~, please wait... ";
VAR
    pb1,pb2,rc,keep:BOOLEAN;
    hin:FIO.File;
    got:CARDINAL;
    msg:str128;
BEGIN
    IF showpercentage THEN
        Str.Copy(msg,informativemsg);
        Str.Subst(msg,"~", fmtlc (current,10,1,"",""));
        Str.Subst(msg,"~", fmtlc (total,  10,1,"",""));
    ELSE
        Str.Copy(msg,tersemsg);
    END;
    keep:=FIO.EOF;
    video(msg,TRUE);
    rc:=TRUE;

    hin :=fileOpenRead(useLFN,S);
    (* we check handle now //V10A *)
    pb1:=( hin = MAX(CARDINAL) ); (* value trapped as error in FIO.MOD runtime *)
    errcode:=FIO.IOresult();
    pb2:=( errcode # DOSErr.NO_ERROR);
    IF ( pb1 OR pb2 ) THEN
        rc:=FALSE;
    ELSE
        FIO.AssignBuffer(hin,ioBufferIn);
        LOOP
            got := FIO.RdBin(hin,dataBuffer,dataBufferSize);
            errcode:= FIO.IOresult();
            IF errcode # DOSErr.NO_ERROR THEN rc:=FALSE;EXIT;END;
            IF got = 0 THEN EXIT; END;
            IF got # dataBufferSize THEN EXIT;END;
        END;
        fileClose(useLFN,hin);
    END;
    video(msg,FALSE);
    FIO.EOF := keep;
    RETURN rc;
END procFile;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE countLines ( useLFN:BOOLEAN;list:pathtype ):LONGCARD ;
CONST
    msg = "::: Counting lines, please wait... ";
VAR
    hlst:FIO.File;
    S:pathtype;
    n:LONGCARD;
BEGIN
    video(msg,TRUE);
    n:=0;
    hlst:=fileOpenRead(useLFN,list);
    FIO.AssignBuffer(hlst,ioBufferLst);
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hlst,S);
        IF FIO.EOF THEN EXIT; END;
        INC(n);
    END;
    fileClose(useLFN,hlst);
    video(msg,FALSE);
    RETURN n;
END countLines;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    idDOS   = "DOS!";
    idWin32 = "WIN!";
    idNOW   = idDOS;   (* //DONOTFORGET *)
TYPE
    statstype = RECORD (* include list filename too ? *)
        id        : ARRAY[0..3] OF CHAR;
        currfpos  : LONGCARD;
        skipped   : LONGCARD;
        failed    : LONGCARD;
        processed : LONGCARD;
        (* //V10B *)
        linecount : LONGCARD;
        currline  : LONGCARD;
        dtStart   : DTtype;
        dtEnd     : DTtype;
    END;
VAR
    i,opt,parmcount : CARDINAL;
    Z,S,R,list:pathtype;
    useLFN,verbose,resume,showpercentage : BOOLEAN;
    state:(waiting,gotlist);
    casemode:(defcase,uppercased,lowercased);
    hlst:FIO.File;
    errcode,chkrounds:CARDINAL;
    ok,alcatraz:BOOLEAN;
    stats : statstype;
    RESUMEFILE:pathtype;
BEGIN
    (* Lib.DisableBreakCheck(); *)
    FIO.IOcheck := FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone; (* very, very important ! *)
    WrLn;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    useLFN    := TRUE;
    verbose   := FALSE;
    casemode  := defcase;
    resume    := FALSE;
    chkrounds := 0;
    showpercentage:=TRUE;
    state     := waiting;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "R"+delim+"RESTART"+delim+"!"+delim+
                                  "X"+delim+"LFN"+delim+"L"+delim+
                                  "U"+delim+"UPPERCASE"+delim+
                                  "W"+delim+"LOWERCASE"+delim+
                                  "Q"+delim+"QUIET"
                              );
            CASE opt OF
            | 1,2,3   : abort(errHelp,"");
            | 4,5     : verbose := TRUE;
            | 6,7,8   : resume  := TRUE;
            | 9,10,11 : useLFN  := FALSE;
            | 12,13:    CASE casemode OF
                        | defcase,uppercased : casemode:=uppercased;
                        ELSE
                            abort(errCaseRedefined,"");
                        END;
            | 14,15:    CASE casemode OF
                        | defcase,lowercased : casemode:=lowercased;
                        ELSE
                            abort(errCaseRedefined,"");
                        END;
            | 16,17:    showpercentage:=FALSE;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            CASE state OF
            | waiting : Str.Copy(list,S);
            | gotlist:  abort(errTooManyParms,S);
            END;
            INC(state);
        END;
    END;
    CASE state OF
    | waiting: abort(errExpecting,"<list>");
    | gotlist: ;
    END;

    useLFN:=( useLFN AND fileSupportLFN() );

    IF chkJoker(list) THEN abort(errJoker,list);END;
    IF Str.Match(list,"*"+extRSM) THEN abort(errReserved,list);END;
    IF fileIsDirectorySpec(useLFN,list) THEN abort(errDir,list);END;
    IF Str.RCharPos(list,dot) = MAX(CARDINAL) THEN Str.Append(list,extLST);END;
    IF fileExists(useLFN,list)=FALSE THEN abort(errNoMatch,list);END;

    buildResumeFilename(RESUMEFILE, extRSM,list);
    IF resume THEN
        IF fileExists(useLFN,RESUMEFILE)=FALSE THEN abort(errNoMatch,RESUMEFILE);END;
        hlst:=fileOpenRead(useLFN,RESUMEFILE);
        i:=FIO.RdBin(hlst,stats,SIZE(stats));
        fileClose(useLFN,hlst);
        IF Str.Compare(stats.id,idNOW) # 0 THEN abort(errBadOS,RESUMEFILE);END;
        IF i # SIZE(stats) THEN abort(errCorrupt,RESUMEFILE);END;
    ELSE
        Str.Copy(stats.id,idNOW);
        stats.currfpos  := 0;
        stats.processed := 0;
        stats.failed    := 0;
        stats.skipped   := 0;
        stats.linecount := countLines(useLFN,list);
        stats.currline  := 0;
    END;

    GetDateTimeNow(stats.dtStart);

    FIO.EOF := FALSE; (* should not be necessary ! *)
    hlst:=fileOpenRead(useLFN,list);
    FIO.AssignBuffer(hlst,ioBufferLst);
    FIO.Seek(hlst,stats.currfpos);
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hlst,S);
        IF FIO.EOF THEN EXIT; END;
        INC(stats.currline);
        preprocessLine(S);
        CASE S[0] OF
        | charnull,semicolon,pound: ;
        ELSE
            Str.Copy(Z,S);
            CASE casemode OF
            | defcase:    ;
            | uppercased: UpperCaseAlt(Z);
            | lowercased: LowerCaseAlt(Z);
            END;
            IF chkJoker(S) THEN
                dmp(TRUE,msgJoker,Z);
                INC(stats.skipped);
            ELSE
                IF fileIsDirectorySpec(useLFN,S) THEN
                    dmp(TRUE, msgNotFile,Z);
                    INC(stats.skipped);
                ELSE
                    IF fileExists(useLFN,S) THEN
                        IF procFile (errcode,
                                    useLFN,showpercentage,
                                    stats.linecount,stats.currline,S) THEN
                            IF verbose THEN dmp(TRUE,msgReadOK,Z); END;
                            INC(stats.processed);
                        ELSE
                            R:=msgReadError;
                            Str.Subst(R,"#",fmtc(errcode,10,3," ",""));
                            dmp(TRUE,R,Z);
                            INC(stats.failed);
                        END;
                    ELSE
                        dmp(TRUE,msgNotFound,Z);
                        INC(stats.skipped);
                    END;
                END;
            END;
        END;
        alcatraz := pollEscape (chkrounds, useLFN,FALSE);
        IF alcatraz THEN EXIT; END;
    END;
    stats.currfpos:=FIO.GetPos(hlst);
    fileClose(useLFN,hlst);

    GetDateTimeNow(stats.dtEnd);

    WrLn;
    dmp(FALSE,msgSkipped  ,fmtlc(stats.skipped,  10,wilc," ",""));
    dmp(FALSE,msgFailed   ,fmtlc(stats.failed   ,10,wilc," ",""));
    dmp(FALSE,msgProcessed,fmtlc(stats.processed,10,wilc," ",""));
    WrLn;
    Str.Copy(R,fmtDelta(stats.dtStart,stats.dtEnd) );
    dmp(FALSE,msgElapsed  ,R);

    IF alcatraz THEN
        IF fileExists(useLFN,RESUMEFILE) THEN
            fileErase(useLFN,RESUMEFILE);
        END;
        hlst:=fileCreate(useLFN,RESUMEFILE);
        FIO.WrBin(hlst,stats,SIZE(stats));
        fileClose(useLFN,hlst);
        WrLn;
        R:=msgLASTPROCESSED;
        Str.Subst(R,"~",S);
        WrStr(R);WrLn;
        WrLn;
        S:=msgRESUME;
        Str.Subst(S,"~",RESUMEFILE);
        WrStr(S);WrLn;
        WrLn;
        abort(errAborted,"");
    END;

    abort(errNone,"");
END chkFread.

