(* ---------------------------------------------------------------
Title         Q&D Delete but...
Author        PhG
Overview      delete all files but specified ones
Usage         see help
Notes         under Win9X, without any reason, dirbut can freeze with 6000+ files
Wish list     recursion ? ah, we must be joking !
              showdata() with -terse and LFN could be prettier
              an ini file specifying directory to be protected
              against any action (c:\, c:\dos, c:\windows, ...)
              a DIRBAT-like option to allow for date/time and size criteria
              (del all files except dated today or smaller than # ?)
              dirbut sometimes freezes without any reason when run from 9x

Bugs          a quirk : we ask about erasing RO even when useless... so what ?

--------------------------------------------------------------- *)

MODULE DelB;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT IO;
IMPORT SYSTEM;

FROM IO IMPORT WrStr,WrLn;

FROM Storage IMPORT Available,ALLOCATE,DEALLOCATE;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XsupportLFN;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

TYPE
    casemodetype = (asis,forcelower,forceupper);
CONST
    sINFO = "::: ";
CONST
    ProgEXEname   = "DELB";
    ProgTitle     = "Q&D Delete but...";
    ProgVersion   = "v1.1j";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    errNone             = 0;
    errHelp             = 1;
    errConflict         = 2;
    errUnknownOption    = 3;
    errBadSpec          = 4;
    errTooManyFilesHere = 5;
    errTooManyMatches   = 6;
    errEmptyDirectory   = 7;
    errAborted          = 128;
    errAbortedByUser    = 255;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
    (*
     00000000011111111112222222222333333333344444444445555555555666666666677777777778
     1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
    *)
    helpmsg =
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" [file specification]... [option]..."+nl+
nl+
"  -l[n|e] list files to be deleted (DIRBUT)"+nl+
"  -x[n|e] list files NOT to be deleted (DIRONLY)"+nl+
"  -a      take hidden and system files into account"+nl+
"  -u[u]   force uppercase (-u) or original case (-uu) filenames display"+nl+
"          (default is lower case)"+nl+
"  -r      delete read-only files"+nl+
"  -d      delete without query"+nl+
"  -t[t]   terse listing (-tt = list with full path)"+nl+
"  -xp     do not show DOS name (LFN only)"+nl+
"  -n      sort by name (-l or -x only)"+nl+
"  -e      sort by extension (-l or -x only)"+nl+
"  -q      do not show program activity (-l or -x only)"+nl+
"  -o      older prompt to delete"+nl+
"  -k|p    paging (-l or -x only ; ignored if output is redirected)"+nl+
"  -!      disable ESC polling"+nl+
"  -f      disable LFN support even if available"+nl+
nl+
"a) Illogical options will be flagged, irrelevant ones will be ignored."+nl+
"b) By design, this command applies to files in current directory ONLY."+nl+
"c) Unless -! option was specified, ESCape key aborts current operation."+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp:             WrStr(helpmsg);
    | errConflict:         S:='"|" commands are mutually exclusive !';
    | errUnknownOption:    S:='Unknown "|" option !';
    | errBadSpec:          S:='Illegal "|" specification !';
    | errTooManyFilesHere :(* S := "Storage ALLOCATE() failure !"; *)
                           S:='Too many files in "|" directory !';
    | errTooManyMatches:   S:='Too many files match "|" specification !';
    | errEmptyDirectory :  S:='No files in "|" directory !';
    | errAborted :         S:="Aborted by user !";
    | errAbortedByUser:    S:="Listing aborted by user !";
    ELSE
                           S:="This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp :
        ; (* nada *)
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

PROCEDURE dbgVstr(i:CARDINAL;S:ARRAY OF CHAR);
BEGIN
     WrStr("// ");IO.WrCard(i,8);WrStr(" : ");WrStr(S);WrLn;
END dbgVstr;

PROCEDURE dbgStr(S:ARRAY OF CHAR);
BEGIN
     WrStr(":: ");WrStr(S);WrLn;
END dbgStr;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    firstEntry = 1; (* 1..count *)
TYPE
    ptrToEntry = POINTER TO entryType;
    entryType = RECORD
        next   : ptrToEntry;
        index  : CARDINAL; (* 1.. for sort *)
        slen   : CARDINAL; (* a SHORTCARD would DO fine IN almost all cases  *)
        string : CHAR;     (* here, after other data, because variable length *)
    END;

VAR
    anchorMatching   : ptrToEntry; (* files anchorMatching specifications *)
    anchorCurrent    : ptrToEntry; (* files in anchorCurrent directory *)

PROCEDURE initList (VAR anchor:ptrToEntry; VAR count:CARDINAL);
BEGIN
    count := firstEntry-1; (* complicated way to mean 0 *)
    anchor:= NIL;
END initList;

PROCEDURE getAnchorCurrent (  ):ptrToEntry;
BEGIN
    RETURN anchorCurrent;
END getAnchorCurrent;

PROCEDURE isReservedEntry (S : ARRAY OF CHAR) : BOOLEAN; (* assume uppercase *)
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE buildMatchList (VAR anchor:ptrToEntry;
                          DEBUG,useLFN,hiddensystem : BOOLEAN;spec:pathtype):CARDINAL;
VAR
    S:pathtype; (* some are oversized but safety first ! *)
    newInList : ptrToEntry;
    len,needed,count : CARDINAL;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    DOSentry     : FIO.DirEntry;
    rc,found:BOOLEAN;
    dosattr:FIO.FileAttr;
BEGIN
IF DEBUG THEN dbgStr(spec);END;
    IF hiddensystem THEN
        dosattr:=allfiles;
    ELSE
        dosattr := FIO.FileAttr{aR,aA};
    END;

    count:=0;

    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(dosattr),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,dosattr,DOSentry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(S,w9Xentry.fullfilename);
        ELSE
            Str.Copy(S,DOSentry.Name);
        END;
        IF isReservedEntry(S) THEN (* skip "." ".." *)
            ; (* silently ignore this spec *)
        ELSE
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=DOSentry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
IF DEBUG THEN dbgVstr(count," :: buildmatchlist count");END;
IF DEBUG THEN dbgVstr(count,S);END;
                IF count=MAX(CARDINAL) THEN
                    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL);
                END; (* too many files but let's fake ALLOCATE failure *)
                len:=Str.Length(S);
                needed:=SIZE(entryType)-SIZE(CHAR)+len;
                IF Available(needed)=FALSE THEN
                    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL);
                END; (* storage ALLOCATE failure *)
                IF anchor=NIL THEN
IF DEBUG THEN dbgStr("anchor was NIL");END;
                    ALLOCATE( anchor,needed);
                    newInList := anchor;
                ELSE
                    ALLOCATE(newInList^.next,needed);
                    newInList :=newInList^.next;
                END;
                Lib.FastMove( ADR(S),ADR(newInList^.string),len);
                newInList^.slen := len;
                newInList^.next := NIL;
                newInList^.index:= 0; (* we'll fill it after we've read all files *)
                INC(count);
            END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(DOSentry);
        END;
    END;
    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
    RETURN count;
END buildMatchList;

PROCEDURE freeMatchList (anchor:ptrToEntry);
VAR
    len,needed      : CARDINAL;
    firstInList,newInList   : ptrToEntry;
BEGIN
    firstInList := anchor;
    newInList := firstInList;
    WHILE newInList # NIL DO
        len         := CARDINAL(newInList^.slen);
        needed      := SIZE(entryType)-SIZE(CHAR)+len;
        firstInList := firstInList^.next;
        DEALLOCATE (newInList,needed);
        newInList := firstInList;
    END
END freeMatchList;

PROCEDURE findEntry (n:CARDINAL;  anchor:ptrToEntry  ):ptrToEntry;
VAR
    newInList:ptrToEntry;
BEGIN
    newInList := anchor;
    LOOP
        IF newInList = NIL THEN EXIT;END; (* gloups! *)
        IF newInList^.index = n THEN EXIT; END;
        newInList := newInList^.next;
    END;
    RETURN newInList;
END findEntry;

PROCEDURE getEntry (VAR R:pathtype;
                   n:CARDINAL; anchor:ptrToEntry);
VAR
    i,len:CARDINAL;
    newInList:ptrToEntry;
    S:pathtype;
BEGIN
    newInList := findEntry(n,anchor);
    IF newInList = NIL THEN RETURN;END;
    len         := newInList^.slen;
    Lib.FastMove( ADR(newInList^.string),ADR(S),len);
    S[len]      := charnull; (* REQUIRED safety ! *)
    Str.Copy(R,S); (* yep, compiler won't let us fill R directly *)
END getEntry;

PROCEDURE doMatch (S:ARRAY OF CHAR;lastmatch:CARDINAL;anchor:ptrToEntry) : BOOLEAN;
VAR
    i : CARDINAL;
    R : pathtype;
BEGIN
    FOR i := firstEntry TO lastmatch DO
        getEntry(R, i,anchor);              (* not the fastest way to scan data ! *)
        IF same(S,R) THEN RETURN TRUE; END;
    END;
    RETURN FALSE;
END doMatch;

PROCEDURE initIndexes (DEBUG:BOOLEAN;last:CARDINAL;anchor:ptrToEntry ):CARDINAL;
VAR
    newInList:ptrToEntry;
    i : CARDINAL;
    R : pathtype;
BEGIN
    i := firstEntry-1;
    newInList:=anchor;
    LOOP
        IF newInList = NIL THEN EXIT; END;
IF DEBUG THEN dbgVstr(i,"i initindexes");END;
        INC(i);
        newInList^.index := i;
        newInList := newInList^.next;
    END;
    IF DEBUG THEN
        IF i # last THEN dbgVstr(i,"ARGH !!! i !!!");dbgVstr(last,"ARGH !!! last !!!");END;
    END;
    RETURN i;
END initIndexes;


(* ------------------------------------------------------------ *)

PROCEDURE doswap (i,j:CARDINAL);
VAR
    anchor,pi,pj:ptrToEntry;
    tmp:CARDINAL;
BEGIN
    anchor:=getAnchorCurrent();
    pi:=findEntry(i,anchor);
    pj:=findEntry(j,anchor);
    tmp:=pi^.index;
    pi^.index:=pj^.index;
    pj^.index:=tmp;
END doswap;

PROCEDURE dolessname (i,j:CARDINAL ):BOOLEAN ;
VAR
    SI,SJ: pathtype ;
    anchor:ptrToEntry;
BEGIN
    anchor:=getAnchorCurrent();
    getEntry(SI,i, anchor);
    getEntry(SJ,j, anchor);
    LowerCase(SI); (* ignore accents when sorting *)
    LowerCase(SJ);
    IF Str.Compare(SI,SJ) < 0 THEN (* -1 is less *)
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END dolessname;

PROCEDURE dolessext ( i,j:CARDINAL  ):BOOLEAN ;
VAR
    e1,e2:pathtype;
    p : CARDINAL;
    anchor:ptrToEntry;
BEGIN
    anchor:=getAnchorCurrent();
    getEntry(e1,i,anchor);
    getEntry(e2,j,anchor);
    LowerCase(e1); (* ignore accents when sorting *)
    LowerCase(e2);

    p := Str.RCharPos(e1,dot);
    IF p = MAX(CARDINAL) THEN
        e1:="";
    ELSE
        Str.Delete(e1,0,p+1);
    END;
    p := Str.RCharPos(e2,dot);
    IF p = MAX(CARDINAL) THEN
        e2:="";
    ELSE
        Str.Delete(e2,0,p+1);
    END;
    CASE Str.Compare(e1,e2) OF
    | -1 : RETURN TRUE;
    |  0 : RETURN dolessname(i,j);
    |  1 : RETURN FALSE;
    END;
END dolessext;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fmtFileNameDOS ( S : ARRAY OF CHAR ) : str128;
CONST
    winame = 8;
    wiext  = 3;
VAR
    p : CARDINAL;
    R : str128;
    i : CARDINAL;
BEGIN
    IF Str.RCharPos(S,dot) = MAX(CARDINAL) THEN
        Str.Append(S,dot);
    END;
    p := Str.RCharPos(S,dot);
    Str.Slice(R,S,0,p);
    FOR i := 1 TO (winame-p) DO
        Str.Append(R,blank);
    END;

    Str.Append(R,dot);

    Str.Delete(S,0,p+1);
    p:=Str.Length(S);
    Str.Append(R,S);
    FOR i := 1 TO (wiext-p) DO
        Str.Append(R,blank);
    END;
    RETURN R;
END fmtFileNameDOS;

PROCEDURE tf (VAR S:ARRAY OF CHAR;flag:BOOLEAN;chTrue,chFalse:CHAR);
BEGIN
    IF flag THEN
        Str.Append(S,chTrue);
    ELSE
        Str.Append(S,chFalse);
    END;
END tf;

PROCEDURE fmtAttr (attr:FIO.FileAttr) : str16;
VAR
    S : str16;
BEGIN
    S:="";
    tf(S, (aD IN attr), "D","-");
    tf(S, (aR IN attr), "R","-");
    tf(S, (aH IN attr), "H","-");
    tf(S, (aS IN attr), "S","-");
    tf(S, (aA IN attr), "A","-");
    RETURN S;
END fmtAttr;

PROCEDURE using (n : CARDINAL; digits : CARDINAL; pad : CHAR) : str80;
VAR
    ok   : BOOLEAN;
    v    : LONGCARD;
    len  : CARDINAL;
    S    : str80;
BEGIN
    v := LONGCARD(n);
    Str.CardToStr(v,S,10,ok);
    len := Str.Length(S);
    LOOP
        IF Str.Length(S) >= digits THEN EXIT; END;
        Str.Prepend(S,pad);
    END;
    RETURN S;
END using;

(*
Year stored relative to 1980 (ex. 1988 stores as 8)
    year      month    day   

 F E D C B A 9 8 7 6 5 4 3 2 1 0   <-- Bit Number
*)

PROCEDURE fmtDate (datedata:CARDINAL) : str80;
CONST
    yyMask=BITSET{9..15};
    yyShft=9;
    mmMask=BITSET{5..8};
    mmShft=5;
    ddMask=BITSET{0..4};
    ddShft=0;
CONST
    separator = dash;
    paddays=blank;
    pad="0";
    baseyear = 1980;
    tmonths ="Jan Fv Mar Avr Mai Jun Jui Ao Sep Oct Nov Dc ???";
    tmonths2="Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ???";
VAR
    y,m,d : CARDINAL;
    R : str80;
BEGIN
    y := CARDINAL(BITSET(datedata) * yyMask) >> yyShft;
    m := CARDINAL(BITSET(datedata) * mmMask) >> mmShft;
    d := CARDINAL(BITSET(datedata) * ddMask) >> ddShft;

    IF ((m < 1) OR (m > 12)) THEN m := 13; END;
    Str.ItemS(R,tmonths2," ",m-1);
    Str.Prepend(R,separator);
    Str.Prepend(R,using(d,2,paddays));
    Str.Append(R,separator);
    Str.Append(R,using(baseyear+y,4,pad));
    RETURN R;
END fmtDate;

(*
Seconds are 0 to 29 -- DOS stores nearest even / 2
  hours    minutes   seconds 

 F E D C B A 9 8 7 6 5 4 3 2 1 0   <-- Bit Number
*)

PROCEDURE fmtTime (timedata:CARDINAL) : str80;
CONST
    hhMask=BITSET{11..15};
    hhShft=11;
    mmMask=BITSET{5..10};
    mmShft=5;
    ssMask=BITSET{0..4};
    ssShft=0;
CONST
    separator = colon;
    padhours = blank;
    pad="0";
VAR
    h,m,s : CARDINAL;
    R : str80;
BEGIN
    h := CARDINAL(BITSET(timedata) * hhMask) >> hhShft;
    m := CARDINAL(BITSET(timedata) * mmMask) >> mmShft;
    s := CARDINAL(BITSET(timedata) * ssMask) >> ssShft;
    s := s << 1; (* yes, yes, "* 2" works too... *)
    R := using(h,2,padhours);
    Str.Append(R,separator);
    Str.Append(R,using(m,2,pad));
    Str.Append(R,separator);
    Str.Append(R,using(s,2,pad));
    RETURN R;
END fmtTime;

PROCEDURE fmtSize (v : LONGCARD; pad:CHAR; sep:CHAR) : str80;
CONST
    field = 10+3; (* #,###,###,### *) (* 1Gb+ file okay now *)
VAR
    S,R   : str80;
    len,i : CARDINAL;
    ok  : BOOLEAN;
    ch  : CHAR;
BEGIN
    Str.CardToStr(v,S,10,ok); (* Str.FixRealToStr( LONGREAL(v),0,S,ok); was overkill *)
    len:=Str.Length(S);
    R := "";
    FOR i := 1 TO len DO
        Str.Prepend(R,S[len-i]);
        IF i < len THEN
            IF (i MOD 3) = 0 THEN
                Str.Prepend(R,sep);
            END;
        END;
    END;
    LOOP
        IF INTEGER(Str.Length(R)) >= ABS(field) THEN EXIT; END;
        IF field < 0 THEN
            Str.Append(R,pad);  (* right alignment *)
        ELSE
            Str.Prepend(R,pad);
        END;
    END;
    RETURN R;
END fmtSize;

PROCEDURE padstr (VAR R : ARRAY OF CHAR;n:INTEGER);
VAR
    i:CARDINAL;
BEGIN
    FOR i:=(Str.Length(R)+1) TO ABS(n) DO
        IF n < 0 THEN
            Str.Prepend(R,blank);
        ELSE
            Str.Append(R,blank);
        END;
    END;
END padstr;

PROCEDURE casify (VAR S:ARRAY OF CHAR;casemode:casemodetype );
BEGIN
    CASE casemode OF
    | asis: ;
    | forcelower: Str.Lows(S); (* //FIXME LowerCase(S) would fix accents *)
    | forceupper: Str.Caps(S); (* //FIXME UpperCase(S) would fix accents *)
    END;
END casify;

PROCEDURE showdata (S,here:pathtype;casemode:casemodetype;
                   DEBUG,useLFN,hiddensystem,terse,showpath,showDOSname:BOOLEAN;
                    VAR totalcount:CARDINAL;
                    VAR totalsize:LONGCARD);
CONST
    widos = 8+1+3;
VAR
    foundFile   : BOOLEAN;
    R           : str1024; (* oversized for LFN *)
    str    : str128;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    DOSentry     : FIO.DirEntry;
    rc,found:BOOLEAN;
    dosattr:FIO.FileAttr;
    u,d,n,e,Sdos,FI,PA:pathtype;
    sattr:str16;
BEGIN
    IF hiddensystem THEN
        dosattr := allfiles;
    ELSE
        dosattr := FIO.FileAttr{aR,aA};
    END;
    IF useLFN THEN
        foundFile:= w9XfindFirst (S,SHORTCARD(dosattr),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
        rc:=w9XfindClose(w9Xhandle,errcode);

        rc:=w9XlongToShort(S,errcode,Sdos);

        foundFile := FIO.ReadFirstEntry(Sdos,dosattr,DOSentry);
        dosattr:=DOSentry.attr;
        qdSplitAllPath(Sdos,u,d,n,e);
        Lib.MakeAllPath(Sdos,"","",n,e);
    ELSE
        foundFile := FIO.ReadFirstEntry(S,dosattr,DOSentry);
        dosattr:=DOSentry.attr;
        Str.Copy(Sdos,S);
    END;

    qdSplitAllPath(S,u,d,n,e);
    IF same(e,"") THEN e:=dot;END;
    Lib.MakeAllPath(FI, u,d,n,e);
    casify(FI,casemode);
    Str.Copy(PA,here);
    casify(PA,casemode);
    casify(Sdos,casemode);

    IF NOT(terse) THEN
        str:="  |  |  |  ";
        Str.Subst(str,vbar,fmtSize(DOSentry.size,blank,coma));
        Str.Subst(str,vbar,fmtDate(DOSentry.date));
        Str.Subst(str,vbar,fmtTime(DOSentry.time));
        IF hiddensystem THEN
            Str.Copy(sattr,fmtAttr(dosattr));
            casify(sattr,casemode);
            Str.Append(str,sattr);
            IF useLFN THEN Str.Append(str,"  ");END;
        END;
    END;

    IF useLFN THEN
        IF showDOSname THEN
            Str.Concat(R,fmtFileNameDOS(Sdos),"  ");
        ELSE
            R:="";
        END;
        IF terse THEN
            IF showpath THEN
                Str.Append(R,'"||"');
                Str.Subst(R,vbar,PA);
            ELSE
                Str.Append(R,'"|"');
            END;
            Str.Subst(R,vbar,FI);
        ELSE
            Str.Append(R,str);
            IF showpath THEN
                Str.Append(R,'"||"');
                Str.Subst(R,vbar,PA);
            ELSE
                Str.Append(R,'"|"');
            END;
            Str.Subst(R,vbar,FI);
        END;
    ELSE
        IF terse THEN
            IF showpath THEN
                R:="||";
                Str.Subst(R,vbar,PA);
            ELSE
                R:="|";
            END;
            Str.Subst(R,vbar,FI);
        ELSE
            R:="|  |";
            Str.Subst(R,vbar,fmtFileNameDOS(Sdos));
            Str.Subst(R,vbar,str);
        END;
    END;
    IF NOT (DEBUG) THEN WrStr(R);WrLn; END;

    INC(totalcount);
    INC(totalsize,DOSentry.size);
END showdata;

PROCEDURE canWrite (S : ARRAY OF CHAR) : BOOLEAN;
VAR
    D  : FIO.DirEntry;
    rc : BOOLEAN;
BEGIN
    IF FIO.Exists(S)=FALSE THEN RETURN TRUE; END;
    rc := FIO.ReadFirstEntry(S,allfiles,D); (* assume file exists ! *)
    IF (FIO.readonly IN D.attr) THEN RETURN FALSE; END;
    RETURN TRUE;
END canWrite;

PROCEDURE killfileOld (S:ARRAY OF CHAR;killRO:BOOLEAN;msg:ARRAY OF CHAR);
BEGIN
    IF canWrite(S) THEN
        FIO.Erase(S);
        WrStr(msg);
    ELSE
        IF killRO THEN
            setReadWrite(S);
            FIO.Erase(S);
            WrStr(msg);
            WrStr(" (was Read-only)");
        ELSE
            WrStr("Skipped (is Read-only)");
        END;
    END;
END killfileOld;

CONST
    syes            = "Y";
    sno             = "N";
    syesall         = "A";
    squit           = "Q";
    scancel         = CHR(27);
    smsgAbortedOld  = "Aborted by user !"+nl+nl;
    smsgAborted     = "ESC"+nl+nl;

PROCEDURE doDelOld (VAR ask:BOOLEAN;R:pathtype;useLFN,killRO:BOOLEAN):BOOLEAN;
VAR
    key : str2;
    rc:BOOLEAN;
    errcode:CARDINAL;
    S:str128;
    Sdos,u,d,n,e:pathtype;
BEGIN
    IF useLFN THEN
        rc:=w9XlongToShort(R,errcode,Sdos);
        qdSplitAllPath(Sdos,u,d,n,e);
        Lib.MakeAllPath(R,"","",n,e);
    END;
    S:=fmtFileNameDOS(R); (* dos name *)
    rc:=TRUE;
    CASE ask OF
    | TRUE:
        WrStr(S);  WrStr(" to be deleted (Y/N/A/Q/Esc) ? ");
        Flushkey;  key := Waitkey();  UpperCase(key);
        IF same(key,syes) THEN
            killfileOld(S,killRO,"Yes");
        ELSIF same(key,syesall) THEN
            killfileOld(S,killRO,"All");   ask := FALSE; (* modify *)
        ELSIF same(key,squit) THEN
            WrStr(smsgAbortedOld);  rc:=FALSE;
        ELSIF same(key,scancel) THEN
            WrStr(smsgAbortedOld);  rc:=FALSE;
        ELSE
            WrStr("No");
        END;
        WrLn;
    | FALSE:
        WrStr(S);WrStr(" ");
        killfileOld(S,killRO,"deleted");
        WrLn;
    END;
    RETURN rc;
END doDelOld;

(* ------------------------------------------------------------ *)

PROCEDURE wrf (useLFN:BOOLEAN;S:pathtype);
BEGIN
    IF useLFN THEN WrStr(dquote);END;
    WrStr(S);
    IF useLFN THEN WrStr(dquote);END;
END wrf;

CONST
    deletedRW    = 0;
    deletedwasRO = 1;
    skippedisRO  = 2;

PROCEDURE killfile (S:pathtype;useLFN,killRO,showmsg:BOOLEAN;orgmsg:ARRAY OF CHAR):CARDINAL;
VAR
    rc:CARDINAL;
    msg,moremsg:str16;
BEGIN
    Str.Copy(msg,orgmsg); (* we may change this string passed as a constant *)
    moremsg:="";
    IF fileIsRO(useLFN,S) THEN
        IF killRO THEN
            fileSetRW(useLFN,S);
            fileErase(useLFN,S);
            moremsg:=" (was RO)";
            rc:=deletedwasRO;
        ELSE
            Str.Copy(msg,"Skipped (is RO)"); (* 15 chars ! *)
            rc:=skippedisRO;
        END;
    ELSE
        fileErase(useLFN,S);
        rc:=deletedRW;
    END;
    IF showmsg THEN WrStr(msg);WrStr(moremsg);END;
    RETURN rc;
END killfile;

PROCEDURE doDel (VAR ask:BOOLEAN;R:pathtype; useLFN,killRO:BOOLEAN):BOOLEAN;
CONST
    msgKeys = "Y/N/A/Q/Esc : ";
    msgRW   = "Deleted     : ";
    msgRO   = "Deleted RO  : ";
    msgisRO = "Skipped RO  : ";
VAR
    key : str2;
    e:CARDINAL;
    rc:BOOLEAN;
BEGIN
    rc:=TRUE;
    IF ask THEN
        WrStr(msgKeys);wrf(useLFN,R);WrStr(" ? ");
        Flushkey;  key := Waitkey();  UpperCase(key);
        IF same(key,syes) THEN
            e:=killfile(R,useLFN,killRO,TRUE,"Yes");
        ELSIF same(key,syesall) THEN
            e:=killfile(R,useLFN,killRO,TRUE,"All");   ask := FALSE; (* modify *)
        ELSIF same(key,squit) THEN
            WrStr(smsgAborted);  rc:=FALSE;
        ELSIF same(key,scancel) THEN
            WrStr(smsgAborted);  rc:=FALSE;
        ELSE
            WrStr("No");
        END;
        WrLn;
    ELSE
        e:=killfile(R,useLFN,killRO,FALSE,"");
        CASE e OF
        | deletedRW:    WrStr(msgRW);
        | deletedwasRO: WrStr(msgRO);
        | skippedisRO:  WrStr(msgisRO);
        END;
        wrf(useLFN,R);WrLn;
    END;
    RETURN rc;
END doDel;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE doGetCurrentDirectory (useLFN:BOOLEAN; VAR R : pathtype);
VAR
    drive : SHORTCARD;
    unit  : CHAR;
    rc:BOOLEAN;
    errcode:CARDINAL;
    longform:pathtype;
BEGIN
    drive := FIO.GetDrive();
    unit  := CHR(drive + ORD("A") -1 );
    FIO.GetDir(drive,R); (* \path without trailing slash nor leading u: *)
    Str.Prepend(R,colon);
    Str.Prepend(R,unit); (* u:\path *)
    IF useLFN THEN
        rc:=w9XshortToLong(R,errcode,longform);
        IF rc THEN Str.Copy(R,longform); END;
    END;
    fixDirectory(R);     (* u:\path\ *)
END doGetCurrentDirectory;

CONST
    msgIncluding   = ", INCLUDING hidden and system files";
    msgIgnoring    = ", ignoring hidden and system files";
    msgToBeDeleted = sINFO+"Files to be deleted";
    msgBaseDir     = sINFO+"Current directory : ";

PROCEDURE precision (hiddensystem:BOOLEAN);
BEGIN
    IF hiddensystem THEN
        WrStr(msgIncluding);
    ELSE
        WrStr(msgIgnoring);
    END;
    WrLn;
END precision;

PROCEDURE active (verbose:BOOLEAN; cmd:CARDINAL  );
CONST
    msgWait = sINFO+"Working, please wait...";
BEGIN
    IF verbose THEN
        CASE cmd OF
        | cmdInit : video(msgWait,TRUE);
        | cmdStop : video(msgWait,FALSE);
        END;
        Work(cmd);
    END;
END active;

(* ------------------------------------------------------------ *)

(* //FIXME *)

CONST
    CHKEVERY9X = 128; (* let's call chkEscape every CHKEVERY LOOP *)
    CHKEVERYDOS= 8;  (* safety *)

PROCEDURE pollEscape (VAR chkrounds:CARDINAL; useLFN,ignoreESC:BOOLEAN):BOOLEAN;
VAR
    alcatraz:BOOLEAN;
    every:CARDINAL;
BEGIN
    (*
    9x really does NOT like ChkEscape() and randomely hangs till a keypress
    Flushkey() does NOT help
    *)
    (*
    IF useLFN THEN
        hit:=FALSE;
    ELSE
        hit:=ChkEscape();
    END;
    *)
    IF ignoreESC THEN RETURN FALSE;END;
    IF useLFN THEN
        every:=CHKEVERY9X;
    ELSE
        every:=CHKEVERYDOS;
    END;
    INC(chkrounds);
    IF (chkrounds MOD every) = 0 THEN
        chkrounds:=0;
        alcatraz:=ChkEscape();
    ELSE
        alcatraz:=FALSE;
    END;
    RETURN alcatraz;
END pollEscape;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    cmd : (deleteFiles,listIncluded,listExcluded);
    processAll    : BOOLEAN;
    killReadonly  : BOOLEAN;
    queryDelete   : BOOLEAN;
    terse         : BOOLEAN;
    showpath      : BOOLEAN;
    sort          : (none,byname,byextension);
    paging        : BOOLEAN;
    verbose       : BOOLEAN;
    useLFN        : BOOLEAN;
    olderdelprompt: BOOLEAN;
    ignoreESC     : BOOLEAN;
    casemode      : casemodetype;
    showDOSname   : BOOLEAN;
    DEBUG         : BOOLEAN;

PROCEDURE initDefaults (  );
BEGIN
    cmd           := deleteFiles;
    processAll    := FALSE;
    casemode      := forcelower;
    killReadonly  := FALSE;
    queryDelete   := TRUE;
    terse         := FALSE;
    showpath      := FALSE;
    sort          := none;
    paging        := FALSE;
    verbose       := TRUE ;
    useLFN        := TRUE;
    olderdelprompt:= FALSE;
    ignoreESC     := FALSE; (* was a constant pre-1.1h *)
    showDOSname   := TRUE;
    DEBUG         := FALSE;
END initDefaults;

(* ------------------------------------------------------------ *)

VAR
    parmcount : CARDINAL;
    i,opt : CARDINAL;
    S,R:pathtype;
    info:str128;
    n,lastMatching,lastCurrent,chkrounds : CARDINAL;
    totalcount : CARDINAL;
    totalsize  : LONGCARD;
    basedir    : pathtype;
    ok         : BOOLEAN;
CONST
    initpagingcounter = 1; (* account for message *)
    kbdmsg = "Hit (almost) any key to continue or Escape to abort listing : ";
    cancel = CHR(27);
VAR
    vrows  [00040H:0084H] : SHORTCARD; (* add 1 *)
    lastRow               : CARDINAL;
    rowcount              : CARDINAL;
    key                   : str2;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone; (* very, very important ! *)

    WrLn;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    initDefaults; (* ugly ! *)

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R,"?"+delim+"H"+delim+"HELP"+delim+
                                 "L"+delim+"INCLUDED"+delim+"DIRBUT"+delim+
                                 "X"+delim+"EXCLUDED"+delim+"DIRONLY"+delim+
                                 "U"+delim+"UPPERCASE"+delim+
                                 "R"+delim+"READONLY"+delim+
                                 "D"+delim+"DELETE"+delim+
                                 "T"+delim+"TERSE"+delim+
                                 "N"+delim+"NAME"+delim+
                                 "E"+delim+"EXTENSION"+delim+
                                 "K"+delim+"PAGING"+delim+
                                 "F"+delim+"LFN"+delim+
                                 "TT"+delim+
                                 "A"+delim+"ALLFILES"+delim+
                                 "Q"+delim+"QUIET"+delim+
                                 "LN"+delim+
                                 "LE"+delim+
                                 "XN"+delim+
                                 "XE"+delim+
                                 "P"+delim+
                                 "O"+delim+"OLD"+delim+"OBSOLETE"+delim+
                                 "!"+delim+"ESC"+delim+"ESCAPE"+delim+
                                 "UU"+delim+"ORIGINAL"+delim+
                                 "XP"+delim+"NODOSNAME"+delim+
                                 "DEBUG"
                              );
            CASE opt OF
            |  1,  2,  3 : abort(errHelp,"");
            |  4,  5,  6 :
                IF cmd = listExcluded THEN abort(errConflict,"-l and -x");END;
                cmd := listIncluded;
            |  7,  8,  9 :
                IF cmd = listIncluded THEN abort(errConflict,"-l and -x");END;
                cmd := listExcluded;
            | 10, 11 : casemode:=forceupper;
            | 12, 13 : killReadonly := TRUE;
            | 14, 15 : queryDelete := FALSE;
            | 16, 17 : terse := TRUE;
            | 18, 19 :
                IF sort = byextension THEN abort(errConflict,"-n and -e"); END;
                sort := byname;
            | 20, 21 :
                IF sort = byname THEN abort(errConflict,"-n and -e"); END;
                sort := byextension;
            | 22, 23 : paging := TRUE;
            | 24, 25 : useLFN:=FALSE;
            | 26     : terse:=TRUE; showpath:=TRUE;
            | 27, 28 : processAll := TRUE;
            | 29, 30 : verbose := FALSE;
            | 31 : (* -ln *)
                IF cmd = listExcluded THEN abort(errConflict,"-l and -x");END;
                cmd := listIncluded;
                IF sort = byextension THEN abort(errConflict,"-n and -e"); END;
                sort := byname;
            | 32 : (* -le *)
                IF cmd = listExcluded THEN abort(errConflict,"-l and -x");END;
                cmd := listIncluded;
                IF sort = byname THEN abort(errConflict,"-n and -e"); END;
                sort := byextension;
            | 33 : (* -xn *)
                IF cmd = listIncluded THEN abort(errConflict,"-l and -x");END;
                cmd := listExcluded;
                IF sort = byextension THEN abort(errConflict,"-n and -e"); END;
                sort := byname;
            | 34 : (* -xe *)
                IF cmd = listIncluded THEN abort(errConflict,"-l and -x");END;
                cmd := listExcluded;
                IF sort = byname THEN abort(errConflict,"-n and -e"); END;
                sort := byextension;
            | 35     : paging := TRUE;
            | 36,37,38: olderdelprompt:=TRUE;
            | 39,40,41: ignoreESC:=TRUE;
            | 42,43:    casemode:=asis;
            | 44,45:    showDOSname:=FALSE;
            | 46:       DEBUG := TRUE;
            ELSE
                abort(errUnknownOption,S);
            END;
        ELSE
            ; (* ignore specifications *)
        END;
    END;

    useLFN:=( useLFN AND w9XsupportLFN() );

    initList(anchorMatching,lastMatching);
IF DEBUG THEN dbgVstr(lastMatching,"LastMatching parsing");END;
    FOR i := 1 TO parmcount DO
IF DEBUG THEN dbgVstr(lastMatching,"LastMatching parsing loop");END;
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            ; (* ignore options *)
        ELSE
            IF Str.CharPos(R,colon) # MAX(CARDINAL) THEN abort(errBadSpec,S);END;
            IF Str.CharPos(R,antislash) # MAX(CARDINAL) THEN abort(errBadSpec,S);END;
            IF same(S,dot) THEN S:=stardotstar;END;
IF DEBUG THEN dbgStr("buildmatchlist anchormatching");END;
            n:=buildMatchList(anchorMatching, DEBUG,useLFN,processAll,S);
IF DEBUG THEN dbgVstr(n,"count");END;
            IF n=MAX(CARDINAL) THEN abort(errTooManyMatches,S); END;
            INC(lastMatching,n);
        END;
    END;

    doGetCurrentDirectory(useLFN,basedir);

    S:=stardotstar;
IF DEBUG THEN dbgStr("initList current");END;
    initList(anchorCurrent,lastCurrent);
IF DEBUG THEN dbgVstr(lastCurrent,"lastcurrent");END;
    n:= buildMatchList(anchorCurrent, DEBUG,useLFN,processAll,S);
IF DEBUG THEN dbgVstr(n,"count");END;
    IF n=MAX(CARDINAL) THEN abort(errTooManyFilesHere,basedir); END;
    INC(lastCurrent,n);

    IF lastCurrent < firstEntry THEN abort(errEmptyDirectory,basedir);END;

    casify(basedir,casemode);
IF DEBUG THEN dbgStr("GLOUPS ! init indexes anchorMatching");END;
    i:=initIndexes(DEBUG,lastMatching,anchorMatching);
IF DEBUG THEN dbgVstr(i,"indexes anchormatching");END;
IF DEBUG THEN dbgStr("init indexes anchorCurrent");END;
    i:=initIndexes(DEBUG,lastCurrent,anchorCurrent);
IF DEBUG THEN dbgVstr(i,"indexes anchorcurrent");END;

IF DEBUG THEN
    dbgVstr(lastCurrent,"lastcurrent");
    FOR i := firstEntry TO lastCurrent DO
        getEntry(S, i,anchorCurrent);
	dbgVstr(i,S);
    END;
    dbgVstr(lastMatching,"lastmatching");
    FOR i := firstEntry TO lastMatching DO
        getEntry(S, i,anchorMatching);
	dbgVstr(i,S);
    END;
END;

    CASE cmd OF
    | deleteFiles:
        WrStr(msgToBeDeleted); precision(processAll);
        WrStr(msgBaseDir);wrf(useLFN,basedir);WrLn;
        IF queryDelete THEN
            WrStr(sINFO+"(Y=yes, N=no, A=yes to delete all files, Q or Esc=abort)");
            WrLn;
        END;
    | listExcluded:
        IF NOT(terse) THEN
            WrStr(sINFO+"Files matching specification(s)"); precision(processAll);
            WrStr(msgBaseDir);wrf(useLFN,basedir);WrLn;
        END;
    | listIncluded:
        IF NOT(terse) THEN
            WrStr(sINFO+"Files NOT matching specification(s)"); precision(processAll);
            WrStr(msgBaseDir);wrf(useLFN,basedir);WrLn;
        END;
    END;
    IF NOT(terse) THEN WrLn; END;

    CASE cmd OF
    | listExcluded,listIncluded:
        CASE sort OF
        | byname      : Lib.QSort(lastCurrent,dolessname,doswap);
        | byextension : Lib.QSort(lastCurrent,dolessext ,doswap);
        END;
        IF paging THEN
            lastRow  := CARDINAL(vrows)+1; (* 25, 43, 50, etc. *)
            rowcount := initpagingcounter;
            IF IsRedirected() THEN paging:=FALSE;END;
        END;
    END;

    totalcount := 0;
    totalsize  := 0;

    CASE cmd OF
    | listExcluded, listIncluded : active(verbose,cmdInit);
    END;

    chkrounds:=0; (* try and fix nasty Win9X behavior about ChkEscape *)

IF DEBUG THEN dbgVstr(lastCurrent,"lastCurrent");END;
    FOR i := firstEntry TO lastCurrent DO
        getEntry(S, i,anchorCurrent);
IF DEBUG THEN dbgVstr(i,S);END;
        CASE cmd OF
        | deleteFiles:
            IF pollEscape(chkrounds,useLFN,ignoreESC) THEN
                abort(errAborted,"");
            END;

            ok:=doMatch(S,lastMatching,anchorMatching);
            IF ok=FALSE THEN
                IF olderdelprompt THEN
                    ok:=doDelOld(queryDelete,S,useLFN,killReadonly);
                ELSE
                    ok:=doDel   (queryDelete,S,useLFN,killReadonly);
                END;
                IF NOT(ok) THEN abort(errAborted,"");END;
            END;
        | listExcluded:
            IF pollEscape(chkrounds,useLFN,ignoreESC) THEN
                active(verbose,cmdStop);
                abort(errAbortedByUser,"");
            END;

            ok:=doMatch(S,lastMatching,anchorMatching);
            IF ok THEN
                active(verbose,cmdStop);
                showdata (S,basedir,casemode,
                         DEBUG,useLFN,processAll,terse,showpath,showDOSname,
                         totalcount,totalsize);
                IF paging THEN
                    INC(rowcount);
                    IF rowcount >= lastRow THEN (* = is enough but who knows what evil lurks in the heart of BIOS *)
                        rowcount := initpagingcounter;
                        video(kbdmsg,TRUE);
                        Flushkey;
                        key:=Waitkey();
                        video(kbdmsg,FALSE);
                        IF same(key,cancel) THEN abort(errAbortedByUser,"");END;
                    END;
                END;
                active(verbose,cmdInit);
            ELSE
                active(verbose,cmdShow);
            END;
        | listIncluded:
            IF pollEscape(chkrounds,useLFN,ignoreESC) THEN
                active(verbose,cmdStop);
                abort(errAbortedByUser,"");
            END;

            ok:= doMatch(S,lastMatching,anchorMatching);
            IF ok=FALSE THEN
                active(verbose,cmdStop);
                showdata (S,basedir,casemode,
                         DEBUG,useLFN,processAll,terse,showpath,showDOSname,
                         totalcount,totalsize);
                IF paging THEN
                    INC(rowcount);
                    IF rowcount >= lastRow THEN (* = is enough but who knows what evil lurks in the heart of BIOS *)
                        rowcount := initpagingcounter;
                        video(kbdmsg,TRUE);
                        Flushkey;
                        key:=Waitkey();
                        video(kbdmsg,FALSE);
                        IF same(key,cancel) THEN abort(errAbortedByUser,"");END;
                    END;
                END;
                active(verbose,cmdInit);
            ELSE
                active(verbose,cmdShow);
            END;
        END;
    END;

    CASE cmd OF
    | listExcluded, listIncluded : active(verbose,cmdStop);
    END;

    CASE cmd OF
    | listExcluded,listIncluded:
        IF NOT(terse) THEN
            IF totalcount # 0 THEN WrLn; END;

            info:=sINFO+"~ file~ taking ~ byte~";
            Str.Copy(R,using(totalcount,1,blank) );
            Str.Subst(info,"~",R);
            IF totalcount < 2 THEN
                R:="";
            ELSE
                R:="s";
            END;
            Str.Subst(info,"~",R);
            Str.Copy(R,fmtSize(totalsize,blank,coma) );
            LtrimBlanks(R);
            Str.Subst(info,"~",R);
            IF totalsize < 2 THEN
                R:="";
            ELSE
                R:="s";
            END;
            Str.Subst(info,"~",R);
            WrStr(info);WrLn;
        END;
    END;
    freeMatchList(anchorMatching);
    freeMatchList(anchorCurrent);

    abort(errNone,"");
END DelB.

