(* ---------------------------------------------------------------
Title         Q&D Expression Evaluator
Overview
Notes         this is how software publishers should work :
              do not provide useless examples or snippets
              but real world code !
              yes, the whole thing should have been REWRITTEN,
              NOT just ENHANCED... but who cares ?
              now, it *is* useful at last
Bugs          negative numbers are not directly supported :
              "-#" is a nono but "(0-#)" Kolossal Kludge is ok
Wish List     a complete rewrite of the parser so it handles text operators
              and other bases (at least hex and binary)
              check logic FOR <0 and >1 (see //FIXME)
              interactive (grab cli, compute, functions, save to log ?)

--------------------------------------------------------------- *)

MODULE eEval;

IMPORT Lib;
IMPORT Str;
IMPORT Storage;
IMPORT MATHLIB;

FROM IO IMPORT WrStr,WrLn;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "EEVAL";
    ProgTitle     = "Q&D Expression Evaluator";
    ProgVersion   = "v1.0e";
    ProgCopyright = "enhanced by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
    blank         = " ";
    dot           = ".";
    comma         = ",";
    charnull      = CHR(0);
    ctrlZ         = CHR(26);
    tabul         = CHR(9);
    cr            = CHR(13);
    lf            = CHR(10);
    nl            = cr+lf;
CONST
    suffixLocal   = "F";
    suffixKollabo = "E"; (* say NO to Yet Another Evil Empire : John Walker is righ here too ! *)

CONST
    (*
    yes, we know there are more clever ways of dealing with these numbers
    we'll use them when we rewrite the whole crap, eh eh...
    see //FIXME
    *)
    REALZERO      = LONGREAL(0.0);
    REALONE       = LONGREAL(1.0);

(* ------------------------------------------------------------ *)

CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParse        = 3;
    errReadToken    = 4;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" [option]... <expression>"+nl+
nl+
"a) Valid tokens in <expression> are :"+nl+
nl+
'   [0..9]             digits (only base 10 numbers are supported)'+nl+
'   "." ","            decimal separator'+nl+
'   " " "(" ")"'+nl+
'   "+" "-" "*" "/"'+nl+
'   "\"                integer division'+nl+
'   "%"                modulo'+nl+
'   "^"                power (from WinXP console please use "~" instead)'+nl+
nl+
"b) Legal suffixes for currency operations are :"+nl+
nl+
'   "'+suffixLocal  +'"                francs (default)'+nl+
'   "'+suffixKollabo+'"                EU occupation AMGOT-like currency'+nl+
nl+
"   John Walker was right about EU : it *is* Yet Another Evil Empire."+nl+
nl+
"c) Options can appear anywhere in line. They are :"+nl+
nl+
'   "="                format result as currency'+nl+
'   "!"                force American decimal format (default is French)'+nl+
'   "@"                do not beautify results'+nl+
'   "#[#[#[#]]]"       force fix 0, 4, 6 or 8 (default is fix 2)'+nl+
nl+
"d) Note this program is merely one of JPI TS M2 examples with many enhancements"+nl+
"   so it is now a small real-world tool (a command line calculator)"+nl+
"   instead of a useless small piece of source code :"+nl+
"   as such, it may still contain quirks from the original average parser"+nl+
"   which was designed for positive numbers handling and basic operations"+nl+
'   (negative numbers can still be handled if specified as "(0-number)").'+nl+
nl+
"Examples : "+ProgEXEname+" 1963*10000+2*100+1"+nl+
"           "+ProgEXEname+" = ! 1000f+2e"+nl+
"           "+ProgEXEname+" (0-4)*4+16"+nl+
"           "+ProgEXEname+" ### 355/113"+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :       WrStr(helpmsg);
    | errOption :     Str.Concat(S,"Illegal ",einfo);Str.Append(S," option !");
    | errParse:       Str.Concat(S,einfo," !");
    | errReadToken:   Str.Concat(S,einfo," !");
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    kfrancs = LONGREAL (1.0);
    kdollaros=LONGREAL (6.55957);
    kdefault =kfrancs; (* kdollaros would lead to nonsense : 10e*2= 65*13 ! *)
VAR
    what : (mult,
            suffix,
            localsuffix,
            singular,
            plural,
            localsingular,
            localplural);
CONST
    firstndx = ORD(mult);
    lastndx  = ORD(localplural);
CONST
    defaultstringsFR=
"6.55957"+delim+
"E"+delim+
"F"+delim+
"unit en monnaie d'occupation"+delim+
"units en monnaie d'occupation"+delim+
"franc"+delim+
"francs";
    defaultstringsUS=
"6.55957"+delim+
"E"+delim+
"F"+delim+
"EU occupation currency unit"+delim+
"EU occupation currency units"+delim+
"French franc"+delim+
"French francs";

(* ------------------------------------------------------------ *)

PROCEDURE fmtreal (v:LONGREAL;fix:CARDINAL;fr,beautify:BOOLEAN;
                   VAR RR:ARRAY OF CHAR);
CONST
    decimalfr = comma;
    sepfr     = dot; (* or space *)
    decimalus = dot;
    sepus     = comma;
VAR
    S,R : str128;
    ok:BOOLEAN;
    i,len:CARDINAL;
BEGIN
    Str.FixRealToStr( ABS(v),fix,S,ok);
    IF ok THEN
        (* default is FR *)
        ReplaceChar(S,decimalus,decimalfr); (* replace dot with comma or... dot ! *)
        (* "*,##" *)
        len:=Str.CharPos(S,decimalfr); (* trick ! *)
        IF len=MAX(CARDINAL) THEN len:=Str.Length(S);END; (* fix 0 ok *)
        R := "";
        FOR i := 1 TO len DO
            Str.Prepend(R,S[len-i]);
            IF i < len THEN
                IF (i MOD 3) = 0 THEN Str.Prepend(R,sepfr); END;
            END;
        END;
        Str.Delete(S,0,len);  (* ",##" *)
        Str.Prepend(S,R);     (* "###.###"+",##" *)
        IF NOT(beautify) THEN ReplaceChar(S,sepfr,"");END;
        IF NOT(fr) THEN
            Str.Subst(S,decimalfr,"|"); (* the mcleod law : there can be only one *)
            ReplaceChar(S,sepfr,sepus); (* all of them *)
            Str.Subst(S,"|",decimalus);
        END;
        IF v < REALZERO THEN Str.Prepend(S,"-");END; (* //FIXME *)
    ELSE
        S := "???";
    END;
    Str.Copy(RR,S);
END fmtreal;

PROCEDURE align (pad:CHAR;VAR S1,S2:ARRAY OF CHAR);
VAR
    l1,l2,wi,i:CARDINAL;
BEGIN
    l1:=Str.Length(S1);
    l2:=Str.Length(S2);
    IF l1 < l2 THEN
        wi:=l2;
    ELSE
        wi:=l1;
    END;
    (* ugly but functional ! *)
    FOR i:=1 TO wi DO
        IF Str.Length(S1) < wi THEN Str.Prepend(S1,pad);END;
        IF Str.Length(S2) < wi THEN Str.Prepend(S2,pad);END;
    END;
END align;

PROCEDURE fmthowmuch (v:LONGREAL;sing,plur:ARRAY OF CHAR; VAR R:ARRAY OF CHAR);
BEGIN
    IF v > REALONE THEN (* //FIXME *)
        Str.Copy(R,plur);
    ELSE
        Str.Copy(R,sing);
    END;
END fmthowmuch;

(*
PROCEDURE getCli (VAR S : ARRAY OF CHAR);
CONST
    nullchar = CHR(0);
VAR
    i : CARDINAL;
BEGIN
    i := 0;
    LOOP
        S[i] := Lib.CommandLine^[i];
        IF S[i] = nullchar THEN EXIT; END;
        INC(i);
    END;
    (* now, clean string *)
    cleantabs(S);
    LtrimBlanks(S);
    RtrimBlanks(S);
END getCli;
*)

(* ------------------------------------------------------------ *)

TYPE
    tokenType = (number,
                 add, sub, mul, divi, intdivi, power, modulo,
                 leftParen, rightParen,
                 done);
    setoftokensType = SET OF tokenType;

    treeType = POINTER TO treeRecordType;
    treeRecordType = RECORD
        CASE kind:tokenType OF
        | number :
            numberValue : LONGREAL;
        | add,sub,mul,divi,intdivi,power,modulo :
            left  : treeType;
            right : treeType;
        END;
    END;

    syntacticType = (init, expr, term, factor);

CONST
    exprTokens = setoftokensType { add,sub };
    termTokens = setoftokensType { mul,divi,intdivi,power,modulo };

(*
yes, yes, awful globerks ...
but we're merely enhancing TS code the Q&D way !
we don't care about writing a new parser...
and we admit we're not proud of it
besides, recursion almost forces this solution
*)

VAR
    cli,clicaps : str128;
    globndx : CARDINAL;
    globc : CHAR;
    token:tokenType;
    tokenNumberValue:LONGREAL;

PROCEDURE getchar (  ):CHAR ;
BEGIN
    INC(globndx);
    RETURN cli[globndx-1];
END getchar;

PROCEDURE readtoken;
VAR
    S:str128;
    i:CARDINAL;
    ok:BOOLEAN;
    oldc:CHAR;
    kmul : LONGREAL;
BEGIN
    kmul := kdefault;
    LOOP
        oldc := globc;
        globc := getchar();
        CASE oldc OF
        | ' ', tabul : ;
        | '+' : token := add; EXIT;
        | '-' : token := sub; EXIT; (* //FIXME what about negative numbers ? *)
        | '*' : token := mul; EXIT;
        | '/' : token := divi; EXIT;
        | '\' : token := intdivi; EXIT;
        | '~','^' : token := power;EXIT; (* //V10D "^" is eaten by WinXP console ! *)
        | '%' : token := modulo; EXIT;
        | '(' : token := leftParen; EXIT;
        | ')' : token := rightParen; EXIT;
        | charnull, cr, lf, ctrlZ : token := done; EXIT;
        | '0'..'9' : (* read a real number *)
            i := 1;
            S[0] := oldc;
            WHILE (globc >= '0') AND (globc <= '9') DO
                S[i] := globc;
                INC(i);
                globc := getchar();
            END;
            CASE CAP(globc) OF
            | dot, comma :
                globc:= dot; (* decimal point *)
                REPEAT (* read fraction part *)
                    S[i] := globc;
                    INC(i);
                    globc := getchar();
                UNTIL (globc < '0') OR (globc > '9');
                CASE CAP(globc) OF
                | suffixLocal:    globc:=getchar();kmul:=kfrancs;
                | suffixKollabo:  globc:=getchar();kmul:=kdollaros;
                END;
            | suffixLocal:        globc:=getchar();kmul:=kfrancs;   S[i]:=dot;INC(i);
            | suffixKollabo:      globc:=getchar();kmul:=kdollaros; S[i]:=dot;INC(i);
            ELSE
                S[i] := dot;
                INC(i);
            END;
            S[i] := CHR(0);
            tokenNumberValue := Str.StrToReal(S, ok);
            IF NOT ok THEN abort(errReadToken,"Bad number"); END;
            tokenNumberValue := tokenNumberValue * kmul;
            token := number;
            EXIT;
        ELSE
            abort(errReadToken,"Bad character");
        END;
    END;
END readtoken;

PROCEDURE parse(what:syntacticType):treeType;
VAR
    t,t1:treeType;
BEGIN
    CASE what OF
    | factor :
        IF token = leftParen THEN
            readtoken;
            t := parse(expr);
            IF token = rightParen THEN
                readtoken;
            ELSE
                abort(errParse,'Missing ")"');
            END;
        ELSIF token = number THEN
            Storage.ALLOCATE(t, SIZE(t^));
            t^.kind := number;
            t^.numberValue := tokenNumberValue;
            readtoken;
        ELSE
            abort(errParse,"Missing number");
        END;
    | term:
        t := parse(factor);
        WHILE ( token IN (termTokens) ) DO
            t1 := t;
            Storage.ALLOCATE(t, SIZE(t^));
            t^.kind := token;
            readtoken;
            t^.left := t1;
            t^.right := parse(factor);
        END;
    | expr:
        t := parse(term);
        WHILE ( token IN (exprTokens) ) DO
            t1 := t;
            Storage.ALLOCATE(t, SIZE(t^));
            t^.kind := token;
            readtoken;
            t^.left := t1;
            t^.right := parse(term);
        END;
    | init:
        globndx := 0;
        globc := getchar();
        readtoken;
        t := parse(expr);
        IF (token # done) THEN
            abort(errParse,"Missing operator");
        END;
     END;
     RETURN t;
END parse;

PROCEDURE eval(t:treeType):LONGREAL;
VAR
    vl,vr,v:LONGREAL;
BEGIN
    IF t^.kind = number THEN RETURN t^.numberValue; END;
    (* now we can eval *)
    vl:=eval(t^.left);
    vr:=eval(t^.right);
    CASE t^.kind OF
    | add :   v:= vl + vr;
    | sub :   v:= vl - vr;
    | mul :   v:= vl * vr;
    | divi:   v:= vl / vr;
    |intdivi: v:= ( vl - MATHLIB.Mod ( vl,vr ) ) / vr;
    |power  : v:= MATHLIB.Pow ( ABS(vl),vr );
              (* handle MATHLIB quirk *)
              IF vl < REALZERO THEN v:=-v;END; (* //FIXME *)
    | modulo: v:= MATHLIB.Mod ( vl,vr ); (* seems ok *)
    END;
    RETURN v;
END eval;

(* ------------------------------------------------------------ *)

VAR
    msg : ARRAY [firstndx..lastndx] OF str128;
    i,p : CARDINAL;
    S,R,S1,S2:str128;
VAR
    t : treeType;
    result,valfrancs,valdollaros : LONGREAL;
    DEBUG,fr,fmtcurrency,beautify:BOOLEAN;
    fixval : CARDINAL;
BEGIN
    WrLn;
    DEBUG       := FALSE;
    beautify    := TRUE;
    fmtcurrency := FALSE;
    fr          := TRUE;
    fixval      := 2;

    getCli(cli);
    IF same(cli,"") THEN abort(errHelp,"");END;

    Str.Copy(clicaps,cli);
    UpperCase(clicaps);

    (* kludgy but we're not creating a program of ours here, after all ! *)
    FOR i:=1 TO 11 DO
        CASE i OF
        | 1 : S := "/?";
        | 2 : S := "-?";
        | 3 : S := "?";
        | 4 : S := "=";
        | 5 : S := "!";
        | 6 : S := "####"; (* longest first *)
        | 7 : S := "###";
        | 8 : S := "##";
        | 9 : S := "#";
        | 10: S := "@";
        | 11: S := "/DEBUG";
        | 12: S := "-DEBUG";
        END;
        p:=Str.Pos(clicaps,S);     (* yes, Code Police officer, this is ugly *)
        IF p # MAX(CARDINAL) THEN
            CASE i OF
            | 1,2,3 : abort(errHelp,"");
            | 4     : fmtcurrency := TRUE;
            | 5     : fr          := FALSE;
            | 6     : fixval      := 8;
            | 7     : fixval      := 6;
            | 8     : fixval      := 4;
            | 9     : fixval      := 0;
            | 10    : beautify    := FALSE;
            | 11,12 : DEBUG       := TRUE;
            END;
            Str.Delete(clicaps,p,Str.Length(S)); (* brutal and kludgy *)
            Str.Delete(cli    ,p,Str.Length(S)); (* brutal and kludgy *)
        END;
    END;

    (* init strings *)
    IF fr THEN
        S:=defaultstringsFR;
    ELSE
        S:=defaultstringsUS;
    END;
    FOR i:=firstndx TO lastndx DO
        Str.ItemS ( msg[i], S, delim, i);
    END;

    t := parse(init);
    result := eval(t);

    IF fmtcurrency THEN
        valfrancs  := result;
        valdollaros:= result / kdollaros;

        fmtreal(valfrancs  ,fixval,fr,beautify, S);
        fmtreal(valdollaros,fixval,fr,beautify, R);
        align(blank,S,R);

        (* we could hide occasional apparent plural quirk with "Units : #" too ! *)

        fmthowmuch(valfrancs  , msg[ORD(localsingular)] , msg[ORD(localplural)] , S1);
        fmthowmuch(valdollaros, msg[ORD(singular)]      , msg[ORD(plural)]      , S2);

        WrStr(R);WrStr(" ");WrStr(S2);WrLn;
        WrStr(S);WrStr(" ");WrStr(S1);WrLn;
    ELSE
        fmtreal(result     ,fixval,fr,beautify, S);
        WrStr(S);WrLn;
    END;
    abort(errNone,"");
END eEval.





(*

// useless original is here (PROG8.MOD) as a reminder :

MODULE prog8;
(* This program implements a simple interactive calculator *)
IMPORT IO,Str,Storage;

TYPE
  TokenType = (number, add, sub, mul, div, LeftParen,
                RightParen, end);
  SyntacticType = (main, exp, term, factor);
  TreeType = POINTER TO TreeRecordType;

  TreeRecordType = RECORD
    CASE kind:TokenType OF
    | number :
      NumberValue : LONGREAL;
    | add,sub,mul,div :
      left : TreeType;
      right : TreeType;
    END;
  END;

VAR
  c:CHAR;
  token:TokenType;
  TokenNumberValue:LONGREAL;

PROCEDURE error(s:ARRAY OF CHAR);
BEGIN
  IO.WrStr(s);
  IO.WrLn;
  HALT;
END error;

PROCEDURE readtoken;
VAR
  s:ARRAY[0..99] OF CHAR;
  i:[0..99];
  done:BOOLEAN;
  oldc:CHAR;
BEGIN
  LOOP
    oldc := c;
    c := IO.RdChar();
    CASE oldc OF
    | ' ' :
    | '+' : token := add; EXIT;
    | '-' : token := sub; EXIT;
    | '*' : token := mul; EXIT;
    | '/' : token := div; EXIT;
    | '(' : token := LeftParen; EXIT;
    | ')' : token := RightParen; EXIT;
    | CHAR(10),CHAR(13),CHAR(26) : token := end; EXIT;

    | '0'..'9' : (* read a real number *)
      i := 1; s[0] := oldc;
      WHILE (c >= '0') AND (c <= '9') DO
        s[i] := c;
        INC(i);
        c := IO.RdChar();
      END;
      IF c<>'.' THEN (* add decimal point if none in input *)
        s[i] := '.';
        INC(i);
      ELSE
        REPEAT (* read fraction part *)
          s[i] := c;
          INC(i);
          c := IO.RdChar();
        UNTIL (c < '0') OR (c > '9');
      END;
      s[i] := CHAR(0);
      TokenNumberValue := Str.StrToReal(s, done);
      IF NOT done THEN
        error('Bad number?');
      END;
      token := number;
      EXIT;
    ELSE
      error('Bad character');
    END;
  END;
END readtoken;

PROCEDURE read(what:SyntacticType):TreeType;
VAR t,t1:TreeType;
BEGIN
  CASE what OF
  | factor :
    IF token = LeftParen THEN
      readtoken;
      t := read(exp);
      IF token = RightParen THEN
        readtoken;
      ELSE
        error("Missing ')'");
      END;
    ELSIF token = number THEN
      Storage.ALLOCATE(t, SIZE(t^));
      t^.kind := number;
      t^.NumberValue := TokenNumberValue;
      readtoken;
    ELSE
      error('Missing number?');
    END;
  | term:
    t := read(factor);
    WHILE (token = mul) OR (token = div) DO
      t1 := t;
      Storage.ALLOCATE(t, SIZE(t^));
      t^.kind := token;
      readtoken;
      t^.left := t1;
      t^.right := read(factor);
    END;
  | exp:
    t := read(term);
    WHILE (token = add) OR (token = sub) DO
      t1 := t;
      Storage.ALLOCATE(t, SIZE(t^));
      t^.kind := token;
      readtoken;
      t^.left := t1;
      t^.right := read(term);
    END;
  | main :
    c := IO.RdChar();
    readtoken;
    t := read(exp);
    IF (token <> end) THEN
      error('Missing operator?');
    END;
  END;
  RETURN t;
END read;

PROCEDURE eval(t:TreeType):LONGREAL;
BEGIN
  CASE t^.kind OF
  | number : RETURN t^.NumberValue;
  | add : RETURN eval(t^.left) + eval(t^.right);
  | sub : RETURN eval(t^.left) - eval(t^.right);
  | mul : RETURN eval(t^.left) * eval(t^.right);
  | div : RETURN eval(t^.left) / eval(t^.right);
  END;
END eval;

VAR
  t:TreeType;
  result:LONGREAL;
BEGIN
  LOOP
    IO.WrStr('Enter expression : ');
    t := read(main);
    result := eval(t);
    IO.WrStr(' = ');
    IO.WrLngReal(result, 4, 0);
    IO.WrLn;
  END;
END prog8.

*)


