'--------------------------------------------------------------------
' Title       : EGAclock
' Author      : PhG
' Overview    : a screen-saver clock
' Future      : increase number of possible dots in a digit shape
'               animate fixed digits with a better algorithm
' Notes       : based on MyClock v1.2e
' Usage       : EGAclock [options]
'
'--------------------------------------------------------------------

$CPU 8086

$OPTIMIZE SIZE
$COMPILE EXE

$DEBUG MAP OFF
$DEBUG PBDEBUG OFF

$LIB COM        OFF
$LIB CGA        OFF
$LIB EGA        ON
$LIB VGA        OFF
$LIB HERC       OFF
$LIB LPT        OFF
$LIB IPRINT     OFF
$LIB FULLFLOAT  ON

$ERROR BOUNDS   Off
$ERROR NUMERIC  Off
$ERROR OVERFLOW Off
$ERROR STACK    Off

$FLOAT PROCEDURE

$COM    0
$STRING 16
$STACK  8192 ' big stack for big local array in SUB
$SOUND  1

$DIM ARRAY

$DYNAMIC

$OPTION CNTLBREAK ON ' OFF in final EXE

'--------------------------------------------------------------------

DEFINT A-Z
OPTION ARRAY BASE 0
OPTION BINARY BASE 0
'
%False = 0
%True = NOT %False

%Debug = %false ' false in final exe

'--------------------------------------------------------------------
' external functions

$LINK "QDBOX.PBU"
$INCLUDE "QDBOX.DEF"

'--------------------------------------------------------------------
' error handling

%eNone = 100
%eUsage = 101
%eTooManySame = 102
%eUnknownOpt = 103
%eNeedEGA = 104
%eBadRange = 105
%eDosTooOld = 106
%eBadExeName = 107
%eVertices= 108
%ehhmm = 109
%eoneline = 110
%etwolines=111
%ePaperInk=112

'--------------------------------------------------------------------
' constants

%minX=0
%minY=0
%maxX=639
%maxY=349
%hires=9 ' EGA 640x350
' requirements for screen mode 9 ega 640x350
%PixelsPerBytePerPlane=8
%Planes=4

%page1=0
%page2=1

%text=0
%cga=2 ' 640x200 CGA

%Circle = 0
%Square = 1
%Disk = 2
%Box = 3

%Circle2 = 4
%Square2=5

%bestfont=0
%normalfont=1
%smallfont=2

%Preset = 0
%Random = 1
%User = 2

%Normal = 0
%Motion = 1
%MotionFull = 2

%soundoff=0
%soundon=1

%hhmmss = 0
%hhmm = 1

%MinDosVersion = 310 ' 3.10 or later

'--------------------------------------------------------------------
' global variables

SHARED einfo$ ' error info

SHARED vShape$,vSequence$,vOuterDot$,vInnerDot$,vPaper$,vInk$,vFont$
shared vClockMode$,vBeep$,vMotionSteps$,vDelay$
shared vUseMouse$

shared viewpage,workpage
shared flagbeep
shared etapes
shared pause

'--------------------------------------------------------------------

DIM DYNAMIC Icon(1) AS INTEGER ' defined at run time
SHARED Icon()

DIM dynamic Eraser(1) as integer ' when erasing force
shared Eraser()

%firstVertex=1
%lastVertex=512 ' should be enough!

type VertexType ' coordinate of a dot
	x as integer
    y as integer
end type

' vertex contains coordinates of a dot from (0,0)++
' i.e. gives position of each "*" in shape profile
' will have to convert them later using plotsize for relative screen coors
DIM STATIC Vertex (%FirstVertex:%LastVertex) AS VertexType ' compile time
SHARED Vertex()

%FirstShape = 1
%LastShape = 11 ' 0..9 and ":"

TYPE ShapeInfoType
	offset AS INTEGER ' offset of first dot
    count AS INTEGER ' number of dots
    maxwide AS INTEGER
    maxheight AS INTEGER
END TYPE

'shapeinfo contains infos about a shape
DIM STATIC ShapeInfo (%FirstShape:%LastShape) AS ShapeInfoType ' compile time
SHARED ShapeInfo()

DIM lasttime$(%page1:%page2)
shared lasttime$()

%Origin=0
%Target=1

' points qui se transforment pendant le morphing
DIM STATIC Path(%FirstVertex:%LastVertex,%Origin:%Target) AS VertexType ' compile time
SHARED Path()

'--------------------------------------------------------------------

ProgramName$="EGAclock"
EXEName$="EGACLOCK"
Version$="v1.0g"
Copyright$="(c) PhG"
Banner$=ProgramName$+" "+Version$+" "+Copyright$

'
ON ERROR GOTO Abort

GOTO Start: ' jump to main() ;-)

' internal functions

DECLARE SUB GetStrLoc() ' PowerBasic runtime library


SUB StdOut ( BYVAL Text AS STRING )
! push DS                    ; save DS FOR PowerBASIC
! push WORD Ptr Text         ; push STRING handle ON stack
! CALL GetStrLoc
! jcxz ExitStdOut
! mov  DS, DX
! mov  DX, AX
! mov  AH, &H40              ; DOS WRITE TO file
! mov  BX, 1                 ; file handle 1 is CONS
! INT  &H21
ExitStdOut:
! pop  DS
END SUB

SUB StdOutLn( BYVAL s as string )
CALL StdOut (s + CHR$(13, 10) )
END SUB

'--------------------------------------------------------------------

Abort:
' call textmode ' should not be necessary
IF ERR = %eUsage THEN
	wi=56
	stdoutln Banner$
	stdoutln ""
        stdoutln "Syntaxe : "+ExeName$+" [option]..."
  	stdoutln "                                                           par dfaut"
  	call Dmp(wi,"/S:n Motif (0..5)",vshape$)
  	CALL Dmp(wi,"/O:n Squence (0=prdfinie, 1=alatoire)",vSequence$)
  	CALL Dmp(wi,"/W:n Rayon extrieur du motif",vOuterDot$)
  	CALL Dmp(wi,"/R:n Rayon intrieur du motif",vInnerDot$)
  	CALL Dmp(wi,"/P:n Couleur du fond",vPaper$)
  	CALL Dmp(wi,"/I:n Couleur du motif",vInk$)
    call dmp(wi,"/F:n Police (0..2)",vFont$)
  	CALL Dmp(wi,"/M:n Animation (0=non, 1=oui, 2=OUI)",vClockMode$)
  	CALL Dmp(wi,"/B:n Beep (0=non, 1=oui ou sans effet)",vBeep$)
  	CALL Dmp(wi,"/N:n Nombre d'tapes dans une animation",vMotionSteps$)
  	CALL Dmp(wi,"/D:n Temporisation",vDelay$)
    CALL Dmp(wi,"/Z:n Souris (0=non, 1=oui)",vUseMouse$)
    END %eUsage-%eNone
END IF
SELECT CASE ERR
CASE %eUsage
    e$="Impossible (!) ;  signaler au programmeur"
CASE %eTooManySame
	e$="Au moins une option rpte sans raison"
CASE %eUnknownOpt
	e$="Au moins une option imprvue"
case %eNeedEGA
	e$="Ce programme exige une carte graphique EGA ou plus"
case %eBadRange
	e$=einfo$
case %eDosTooOld
	e$="Ce programme exige un DOS version "+einfo$+" ou plus"
case %eBadexeName
	e$=einfo$+" n'est pas le nom original du fichier excutable"
case %eVertices
	e$="Trop de points sont dfinis ;  signaler au programmeur"
CASE %ehhmm
	call textmode
    e$="L'cran n'est pas assez large pour hh:mm"
cASE %etwolines
	call textmode
    e$="L'cran n'est pas assez haut pour hh:mm et ss"
CASE %eoneline
	call textmode
    e$="L'cran n'est pas assez haut pour hh:mm:ss"
case %epaperink
	e$="Encre et papier doivent diffrer"
CASE ELSE
    e$=HEX$(ERADR) ' ERADR is a longint (7fFFffFF)
    e$=MID$("00000000",1,8-LEN(e$))+e$
    e$= "Erreur #"+MID$(STR$(ERR),2)+"  l'adresse $"+e$
END SELECT
e$=ProgramName$+" : "+e$+" !"
stdoutln e$
END ERR-%eNone

'--------------------------------------------------------------------

FUNCTION IsEGA                'test for EGA
  IF BIT (pbvScrnCard,3) = 0 THEN
  	IsEGA = %False
  ELSE
  	IsEGA = %True
  END IF
END FUNCTION

' *should* test for wide < 2 ! ;-)
SUB Dmp (BYVAL wide,BYVAL s1$,BYVAL s2$)
pad$=STRING$(wide," ")
length=LEN(s1$)
IF length > wide THEN
	s$=LEFT$(s1$,wide-2)+".."
ELSE
	s$=s1$+RIGHT$(pad$,wide-length)
END IF
s$=s$+" | "
s2$=ltrim$(rtrim$(s2$))
s2$=MID$("   ",1,3-LEN(s2$))+s2$
stdoutln s$+s2$ ' do not check if bigger than 80!
END SUB

' returns %True or %False if opt found or not, keep count of opt found
' vardest$ is not changed unless optstr$ found AND param <> ""

FUNCTION ChkOptStr (BYVAL cli$,BYVAL optstr$,BYVAL exactflag,vardest$,count)
var$=""
count=FindOptStr(cli$,optstr$,exactflag,var$)
SELECT CASE count
CASE 0
	ChkOptStr=%False
CASE 1
	if var$="" then
    	ChkOptStr=%False ' do not accept empty string as parameter
    else
		vardest$=var$
		ChkOptStr=%True
    end if
CASE ELSE
	ERROR %eTooManySame
END SELECT
END FUNCTION

SUB ChkRange (BYVAL param$,BYVAL value$,BYVAL minrange,BYVAL maxrange)
v#=VAL(value$)
IF v# < minrange OR v# > maxrange THEN
	range$="["+MID$(STR$(minrange),2)+".."+MID$(STR$(maxrange),2)+"]"
	einfo$="Le paramtre "+param$+" doit appartenir  l'intervalle "+range$
	ERROR %eBadRange
END IF
END SUB

'
SUB TextMode
SCREEN %cga ' reset width to 80
SCREEN %text
END SUB

sub ViewWork (byval vpage, byval wpage)
screen %hires,,wpage,vpage
CALL WaitVGARetrace ' must be AFTER and hopla, no longer flicker! ;-)
' at least 3/100s to avoid paradoxal flickering while refreshing
' delay 0.03
end sub

SUB WaitVGARetrace
!	mov dx,&h3DA
l1:
!   in al,dx
!   and al,&h08
!   jnz l1
l2:
!   in al,dx
!   and al,&h08
!   jz  l2
END SUB



'--------------------------------------------------------------------

SUB InitDefault
vShape$=str$(%Disk)
vSequence$=STR$(%Random)
vOuterDot$="10" ' outer half-width
vInnerDot$="8" ' inner half-width
vPaper$="0"
vInk$="10"
vFont$=str$(%bestfont)
vClockMode$=str$(%Motion) ' %MotionFull wants fewer steps (about 6 !!!)
vBeep$=str$(%False)
vMotionSteps$=str$(20) ' adequate for DX33
vDelay$="1"
vUseMouse$="1"
END SUB

sub SetHires(byval viewpage, byval workpage,byval ink,byval paper)
call viewwork(viewpage,workpage)
COLOR ink,paper
CLS
end sub

sub InitIcon (BYVAL numicon,byval ink,byval paper,byval plotsize,byval radius)
cx=%maxX \ 2
cy=%maxY \ 2
diameter=radius*2
wleft=cx-radius
wtop=cy-radius
wright=wleft+diameter-1 ' substract 1 to have just width in pixels
wbot=wtop+diameter-1 ' idem
SELECT CASE numicon
CASE %Circle
	CIRCLE (cx,cy),radius
CASE %Square
    LINE (wleft,wtop)-(wright,wbot),,B
CASE %Disk
	CIRCLE (cx,cy),radius
    PAINT (cx,cy)
CASE %Box
    LINE (wleft,wtop)-(wright,wbot),,BF
CASE %Circle2
	CIRCLE (cx,cy),radius
    circle (cx,cy),radius-0.5
    circle (cx,cy),radius-1
CASE %Square2
    LINE (wleft,wtop)-(wright,wbot),,B
    line (wleft+1,wtop+1)-(wright-1,wbot-1),,B
CASE ELSE ' if undefined, disk
	CIRCLE (cx,cy),radius
    PAINT (cx,cy)
END SELECT
dotwidthX=plotSize*2 +2 ' better safe than sorry! ;-)
dotwidthY=plotSize*2 +2 ' idem
needed=(dotwidthX\%PixelsPerBytePerPlane)+1 ' idem... and avoid CEIL function
BytesRequested=4+needed*%Planes*dotwidthY
REDIM Icon(BytesRequested\2) AS SHARED INTEGER ' more just in case...
wleft=cx-plotSize
wtop=cy-plotsize
wright=cx+plotsize -1 ' must substract in order to have just width in pixels
wbot=cy+plotsize -1 ' idem
GET (wleft,wtop)-(wright,wbot),Icon
end sub

' create a shape as an eraser with screen already blanked
sub InitEraser (BYVAL plotsize)
cx=%maxX \ 2
cy=%maxY \ 2
dotwidthX=plotSize*2 +2 ' better safe than sorry! ;-)
dotwidthY=plotSize*2 +2 ' idem
needed=(dotwidthX\%PixelsPerBytePerPlane)+1 ' idem... and avoid CEIL function
BytesRequested=4+needed*%Planes*dotwidthY
REDIM Eraser(BytesRequested\2) AS SHARED INTEGER ' more just in case...
wleft=cx-plotSize
wtop=cy-plotsize
wright=cx+plotsize -1 ' must substract in order to have just width in pixels
wbot=cy+plotsize -1 ' idem
GET (wleft,wtop)-(wright,wbot),Eraser
end sub


'--------------------------------------------------------------------
' [0..9] and ":" characters (ASCII order)

SUB InitShapes(byval sequencetype,byval fontnum)
' store vertices temporiraly here before we modify the plotting order
DIM DYNAMIC TmpVertex (%FirstVertex:%LastVertex) AS VertexType ' compile time
LOCAL TmpVertex()
select case fontnum
case %bestfont
	RESTORE BestFont
case %normalfont
	restore Normalfont
case %smallfont
	restore smallfont
end select
endOfShape$="$"
RANDOMIZE TIMER ' in case %Random is selected
currvertex=%FirstVertex
FOR currshape=%FirstShape TO %LastShape
    startofdef=currvertex
    maxWidth=0
    y=0
	DO
		READ s$
        IF s$=EndOfShape$ THEN EXIT LOOP
        x=0
        FOR ndx=1 TO len(s$)
        	IF MID$(s$,ndx,1)<>" " THEN
       			Vertex(currvertex).x=x
                Vertex(currvertex).y=y
                INCR currvertex
                IF currvertex > %LastVertex THEN ERROR %eVertices
            END IF
            INCR x
        NEXT
        IF x > maxWidth THEN maxWidth=x
        INCR y
	LOOP
    ' now we have finished reading shape definition so actualize infos
    ShapeInfo(currshape).offset=startofdef
    count=currvertex-startofdef
	ShapeInfo(currshape).count=count
    ShapeInfo(currshape).maxwide=maxWidth
    ShapeInfo(currshape).maxheight=y
    '
    ' we now have each vertex making shape stored as defined
    ' we can build new order of sequence and reset Vertex accordingly
    '
    ' 8-ung! this trick limits the number of possible dots in a shape!!!
    ' but as our digit shapes are not too complex and in [1..255] ;-)
    '
    s$=""
    FOR k = 1 TO count
    	thisdot=k-1 ' important!
    	READ position ' from 1..count
    	SELECT CASE Sequencetype
    	CASE %Preset ' let them as they have been defined
        	TmpVertex(k)=Vertex(startofdef+thisdot)
	    CASE %Random ' randomly change order of vertex
            DO
            	' int((upper-lower+1)*rnd+lower)
				newpos=INT((count-1+1)*RND+1) ' [1..num] -> chr(1)..chr$(255)
				newpos=INT((count-1+1)*RND+1)
				newpos=INT((count-1+1)*RND+1)
            LOOP UNTIL INSTR(s$,CHR$(newpos)) = 0 ' find a NEW position
            s$=s$+CHR$(newpos)
            TmpVertex(newpos)=Vertex(startofdef+thisdot)
	    CASE %User ' use data to change order of points
        	TmpVertex(position)=Vertex(startofdef+thisdot)
		END SELECT
    NEXT
    ' now we can change order of sequence in array of positions
	FOR k = 1 TO count
    	thisdot=k-1
    	Vertex(startofdef+thisdot)=TmpVertex(k)
    NEXT
NEXT
' here we should change each coorset to screen resolution using plotsize

' 0..9 should have the SAME width & height, even though routines is general
BestFont:
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA "*   * "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "      "
DATA "$"
DATA 7,6,5,4,3,2,1
DATA " ***  "
DATA "    * "
DATA "    * "
DATA " ***  "
DATA "*     "
DATA "*     "
DATA " ***  "
DATA "      "
DATA "$"
DATA 13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "    * "
DATA "    * "
DATA " ***  "
DATA "    * "
DATA "    * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "*   * "
DATA "*   * "
DATA "*   * "
DATA " **** "
DATA "    * "
DATA "    * "
DATA "    * "
DATA "      "
DATA "$"
DATA 13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*     "
DATA "*     "
DATA " ***  "
DATA "    * "
DATA "    * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*     "
DATA "*     "
DATA "****  "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "****  "
DATA "    * "
DATA "    * "
DATA "    * "
DATA "    * "
DATA "    * "
DATA "    * "
DATA "      "
DATA "$"
DATA 10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA " **** "
DATA "    * "
DATA "    * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "    "
DATA "    "
DATA " *  "
DATA "    "
DATA " *  "
DATA "    "
DATA "    "
DATA "    "
DATA "$"
DATA 2,1

' [0..9] and ":" characters (ASCII order)
normalfont:
DATA " ***  "
DATA "*   * "
DATA "*  ** "
DATA "* * * "
DATA "**  * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "  *   "
DATA " **   "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA " ***  "
DATA "      "
DATA "$"
DATA 10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*   * "
DATA "   *  "
DATA "  *   "
DATA " *    "
DATA "*     "
DATA "***** "
DATA "      "
DATA "$"
DATA 14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*   * "
DATA "    * "
DATA "  **  "
DATA "    * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "*   * "
DATA "*   * "
DATA "*   * "
DATA "***** "
DATA "    * "
DATA "    * "
DATA "    * "
DATA "      "
DATA "$"
DATA 14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "***** "
DATA "*     "
DATA "*     "
DATA "****  "
DATA "    * "
DATA "    * "
DATA "****  "
DATA "      "
DATA "$"
DATA 17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*     "
DATA "*     "
DATA "****  "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "***** "
DATA "    * "
DATA "    * "
DATA "   *  "
DATA "  *   "
DATA "  *   "
DATA "  *   "
DATA "      "
DATA "$"
DATA 11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA " ***  "
DATA "*   * "
DATA "*   * "
DATA " **** "
DATA "    * "
DATA "    * "
DATA " ***  "
DATA "      "
DATA "$"
DATA 16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "    "
DATA "    "
DATA " *  "
DATA "    "
DATA " *  "
DATA "    "
DATA "    "
DATA "    "
DATA "$"
DATA 2,1

' [0..9] and ":" characters (ASCII order)
smallfont:
DATA "*** "
DATA "* * "
DATA "* * "
DATA "* * "
DATA "*** "
DATA "$"
DATA 12,11,10,9,8,7,6,5,4,3,2,1
DATA " *  "
DATA " *  "
DATA " *  "
DATA " *  "
DATA " *  "
DATA "$"
DATA 5,4,3,2,1
DATA "*** "
DATA "  * "
DATA "*** "
DATA "*   "
DATA "*** "
DATA "$"
DATA 11,10,9,8,7,6,5,4,3,2,1
DATA "*** "
DATA "  * "
DATA "*** "
DATA "  * "
DATA "*** "
DATA "$"
DATA 11,10,9,8,7,6,5,4,3,2,1
DATA "* * "
DATA "* * "
DATA "*** "
DATA "  * "
DATA "  * "
DATA "$"
DATA 9,8,7,6,5,4,3,2,1
DATA "*** "
DATA "*   "
DATA "*** "
DATA "  * "
DATA "*** "
DATA "$"
DATA 11,10,9,8,7,6,5,4,3,2,1
DATA "*   "
DATA "*   "
DATA "*** "
DATA "* * "
DATA "*** "
DATA "$"
DATA 10,9,8,7,6,5,4,3,2,1
DATA "*** "
DATA "  * "
DATA "  * "
DATA "  * "
DATA "  * "
DATA "$"
DATA 7,6,5,4,3,2,1
DATA "*** "
DATA "* * "
DATA "*** "
DATA "* * "
DATA "*** "
DATA "$"
DATA 13,12,11,10,9,8,7,6,5,4,3,2,1
DATA "*** "
DATA "* * "
DATA "*** "
DATA "  * "
DATA "  * "
DATA "$"
DATA 10,9,8,7,6,5,4,3,2,1
DATA "    "
DATA " *  "
DATA "    "
DATA " *  "
DATA "    "
DATA "$"
DATA 2,1
end sub

'--------------------------------------------------------------------

sub ScaleShapes (byval plotsize)
scale = plotsize*2 ' radius * 2 !!!
FOR currshape=%FirstShape TO %LastShape
    startofdef=ShapeInfo(currshape).offset
	count=ShapeInfo(currshape).count
    for k = 1 to count
    	currvertex=startofdef+k-1
    	x=vertex(currvertex).x
        y=vertex(currvertex).y
		vertex(currvertex).x=x*scale
        vertex(currvertex).y=y*scale
    next
    maxwide=ShapeInfo(currshape).maxwide
	maxheight=ShapeInfo(currshape).maxheight
    shapeinfo(currshape).maxwide=maxwide*scale
    shapeinfo(currshape).maxheight=maxheight*scale
next
end sub

'--------------------------------------------------------------------
' assume all digits are SAME SIZE!

' return rightest position(s)
sub Chkdisplay(displaytype,posright,postop,posright2,postop2)
chiffre=shapeinfo(%FirstShape).maxwide ' 0..9
deuxpoints=shapeinfo(%LastShape).maxwide ' :
oneline=shapeinfo(%FirstShape).maxheight
hhmmss=chiffre*6+deuxpoints*2
hhmm=chiffre*4+deuxpoints
ss=chiffre*2
twolines=oneline*2
limx=%maxX+1 ' because width is from 1
limy=%maxY+1
IF hhmmss > limx THEN
	IF hhmm > limx THEN ERROR %ehhmm
    IF twolines > limy THEN ERROR %etwolines
	displaytype=%hhmm
	postop=(limy-twolines) \ 2
    posright=(limx-hhmm) \ 2 +hhmm
    postop2=postop+oneline
    posright2=(limx-ss) \ 2  +ss
ELSE
	IF oneline > limy THEN ERROR %eoneline
    displaytype=%hhmmss
    postop=(limy-oneline) \ 2
    posright=(limx-hhmmss) \ 2 +hhmmss
END IF
end sub

'--------------------------------------------------------------------

sub flushkeyboard
do until inkey$ = ""
loop
end sub

'--------------------------------------------------------------------

function getshapenumber (byval c$)
p=INSTR("0123456789:",c$) ' 1..lastshape
getshapenumber=p-1+%firstshape
end function

'--------------------------------------------------------------------


'--------------------------------------------------------------------

' plot digit at coordinates fromx,fromy
SUB PlotShapexor(byval digit,byval fromx,byval fromy)
origin=shapeinfo(digit).offset
count =shapeinfo(digit).count
FOR k = 1 to count
	currdot=k-1
	wleft=Vertex(origin+currdot).x+fromx
	wtop=vertex(origin+currdot).y+fromy
	PUT (wleft,wtop),Icon,XOR
NEXT
END SUB

'--------------------------------------------------------------------

sub InitFix (byval now$,byval displaytype,byval posright,byval postop,byval posright2,byval postop2)
SELECT CASE displaytype
CASE %hhmmss
   	CALL InitDrawStr(now$,1,8,posright,postop) ' hh:mm:ss
CASE %hhmm
	CALL InitDrawStr(now$,7,8,posright2,postop2) ' ss
	CALL InitDrawStr(now$,1,5,posright,postop) ' hh:mm
END SELECT
END sub

SUB InitDrawStr(byval now$,byval first, byval last,byval fromx, byval fromy)
FOR n = last TO first STEP -1 ' <<
	c$=mid$(now$,n,1)
    digit=getshapenumber(c$)
    maxwidth=shapeinfo(digit).maxwide
   	DECR fromx,maxwidth ' <<
    CALL PlotShapexor(digit,fromx,fromy) ' <-- always a first DRAW here so...
NEXT
END SUB

'--------------------------------------------------------------------
' update only normal clock (without motion)

sub updateFix (byval new$,byval old$,byval displaytype,byval posright,byval postop,byval posright2,byval postop2)
SELECT CASE displaytype
CASE %hhmmss
   	CALL DrawStr(new$,old$,1,8,posright,postop) ' hh:mm:ss
CASE %hhmm
	CALL DrawStr(new$,old$,7,8,posright2,postop2) ' ss
	CALL DrawStr(new$,old$,1,5,posright,postop) ' hh:mm
END SELECT
END sub

SUB DrawStr(byval new$,byval old$,byval first, byval last,byval fromx, byval fromy)
	FOR n = last TO first STEP -1 ' <<
		c$=mid$(old$,n,1)
        olddigit=getshapenumber(c$)
		c$=mid$(new$,n,1)
        digit=getshapenumber(c$)
    	maxwidth=shapeinfo(digit).maxwide
    	DECR fromx,maxwidth ' << update here!!!
        IF digit<>olddigit THEN
        	' erase old on workpage
            CALL Plotshapexor(olddigit,fromx,fromy)
            ' draw new on workpage
			CALL Plotshapexor(digit,fromx,fromy)
        END IF
	NEXT
END SUB

'--------------------------------------------------------------------

' build list of vertices stored in Path, while morphing, returns this #
function buildMorphState (byval sourceshape, byval targetshape)
ptrSource =   ShapeInfo(sourceshape).offset
countSource = ShapeInfo(sourceshape).count
ptrTarget=    ShapeInfo(targetshape).offset
countTarget = ShapeInfo(targetshape).count
nVertex=max%(countsource,counttarget) ' number of vertices while morphing
zx2=nVertex
zy2=    min%(countSource,countTarget)
deltax = zx2 - 1
deltay = zy2 - 1
zd = (2 * deltay) - deltax
dinc1 = deltay * 2
dinc2 = (deltay - deltax) * 2
IF countSource < countTarget THEN
' use bresenham algorithm as if from 1,1 to big,small
	FOR p = %firstvertex TO nVertex
        Path(p,%Origin)=Vertex(ptrSource)
    	Path(p,%Target)=vertex(ptrTarget)
	    IF zd < 0 THEN
       		zd = zd + dinc1
		ELSE
       		zd = zd + dinc2
      		INCR ptrsource
		END IF
        INCR ptrTarget
	NEXT
ELSE
	FOR p = %firstvertex TO nVertex
        Path(p,%Origin)=vertex(ptrsource)
   		Path(p,%Target)=vertex(ptrtarget)
	    IF zd < 0 THEN
       		zd = zd + dinc1
		ELSE
       		zd = zd + dinc2
       		INCR ptrtarget
		END IF
        INCR ptrsource
	NEXT
END IF
buildmorphstate=nvertex
end function

'--------------------------------------------------------------------
' plot digit at coordinates fromx,fromy
SUB PlotState(byval digit,byval fromx,byval fromy,byval here, byval vertices, byval operation)
FOR k = %firstvertex to vertices
	wleft1=Path(k,%Origin).x
	wtop1=Path(k,%Origin).y
	wleft2=Path(k,%Target).x
    wtop2=Path(k,%Target).y
    deltax=wleft2-wleft1
    deltay=wtop2-wtop1
	wleft=wleft1+deltax*here/etapes ' BERK algorithm !!! ;-)
	wtop=wtop1+deltay*here/etapes ' must perform 0..etapes
    wleft=wleft+fromx
    wtop=wtop+fromy
    if operation = 0 then ' erase
		PUT (wleft,wtop),Eraser,PSET
    else ' set
		PUT (wleft,wtop),Icon,OR ' avoid blocks on blocks!
    end if
NEXT
END SUB

'--------------------------------------------------------------------

' program code begins here : main() ;-)

Start:
stdoutln ""

if isEGA = %False then error %eNeedEGA
einfo$=MID$(STR$(%MinDosVersion \ 100),2)+"."+MID$(STR$(%MinDosVersion MOD 100),2)
IF DosVersion < %MinDosVersion THEN ERROR %eDosTooOld
exepath$= GetArg0$
exepath$= Upper$(exepath$)
CALL SplitPath(exepath$,currunit$,currpath$,currfile$)
CALL SplitName(currfile$,filename$,ext$)
einfo$=filename$+".EXE"
IF filename$ <> ExeName$ THEN ERROR %eBadExeName

' now check command line for help

CALL InitDefault ' do it NOW!

cli$=COMMAND$
foundSOSarg=FindArg(cli$,"?|H|HELP|SOS|AIDE",%False)
foundSOSopt=FindOpt(cli$,"?|H|HELP|SOS|AIDE",%False)
SELECT CASE (foundSOSarg+foundSOSopt)
CASE 0
	' no help call
CASE 1
	ERROR %eUsage
CASE ELSE
	ERROR %eTooManySame
END SELECT

' check command line for options

if ChkOptStr(cli$,"S|SHAPE",%False,v$,n1)=%True THEN vShape$=v$
if ChkOptStr(cli$,"O|ORDER",%False,v$,n2)=%True THEN vSequence$=v$
if ChkOptStr(cli$,"W|WIDTH",%False,v$,n3)=%True THEN vOuterDot$=v$
if ChkOptStr(cli$,"R|RADIUS",%False,v$,n4)=%True THEN vInnerDot$=v$
if ChkOptStr(cli$,"P|PAPER",%False,v$,n5)=%True THEN vPaper$=v$
if ChkOptStr(cli$,"I|INK",%False,v$,n6)=%True THEN vInk$=v$
if Chkoptstr(cli$,"F|FONT",%false,v$,n11)=%True then vFont$=v$
if ChkOptStr(cli$,"M|MOVE",%False,v$,n7)=%True THEN vClockMode$=v$
if ChkOptStr(cli$,"B|BEEP",%False,v$,n8)=%True THEN vBeep$=v$
if ChkOptStr(cli$,"N|STEPS",%False,v$,n9)=%True THEN vMotionSteps$=v$
if ChkOptStr(cli$,"D|DELAY|TEMPO",%False,v$,n10)=%True THEN vDelay$=v$
if ChkOptStr(cli$,"Z|MOUSE",%False,v$,n12)=%True THEN vUseMouse$=v$

' check for unknown commands/options on command line

optcount=n1+n2+n3+n4+n5+n6+n7+n8+n9+n10+n11+n12
IF optcount <> ArgC(cli$) THEN ERROR %eUnknownOpt

' perform some checks

CALL ChkRange ("/S",vShape$,%circle,100)
call ChkRange ("/O",vSequence$,%preset,%random) ' was %user but not useful
CALL ChkRange ("/W",vOuterDot$,2,22)
call ChkRange ("/R",vInnerDot$,2,22)
CALL ChkRange ("/P",vPaper$,0,15)
call ChkRange ("/I",vInk$,0,15)
call chkrange ("/F",vFont$,%bestfont,%smallfont)
CALL ChkRange ("/M",vClockMode$,%normal,%motionFull)
call ChkRange ("/B",vBeep$,%soundoff,%soundon)
CALL ChkRange ("/N",vMotionSteps$,2,500)
call ChkRange ("/D",vDelay$,1,999)
call ChkRange ("/Z",vUseMouse$,0,1)

ink=val(vink$)
paper=val(vpaper$)
if ink=paper then error %ePaperInk
fontnum=val(vfont$)
numicon=val (vShape$)
plotsize=VAL(vOuterDot$)
radius=VAL(vInnerDot$)
sequencetype=val(vSequence$)
clockmode=val(vClockmode$)
flagbeep=val(vbeep$)
etapes=val(vmotionsteps$)
pause=val(vdelay$) / 1000
i=VAL(vUseMouse$)
SELECT CASE i
CASE 0
	useMouse=%False
CASE 1
	useMouse=%True
END SELECT

call SetHires(%page1,%page1,ink,paper) ' clear screen 1
call SetHires(%page1,%page2,ink,paper) ' clear screen 2
call InitIcon(numicon,ink,paper,plotsize,radius)
call SetHires(%page1,%page2,ink,paper) ' clear screen 2 again
call initEraser(plotsize) ' create shape to recreate background
call InitShapes(sequencetype,fontnum)
CALL ScaleShapes(plotsize)
call Chkdisplay(displaytype,posright,postop,posright2,postop2)
call FlushKeyboard

' init pages

viewpage=%page1
workpage=%page2
call viewwork(viewpage,workpage)

' draw current time on both pages to begin in a known state

currtime$=time$
lasttime$(%page1)=currtime$
lasttime$(%page2)=currtime$
previoustime$=currtime$ ' for animation
' display on hidden page
call initfix (currtime$,displaytype,posright,postop,posright2,postop2)
swap viewpage,workpage ' toggle pages now
' show updated page
call viewwork(viewpage,workpage)
' now display on newly hidden page too
call initfix (currtime$,displaytype,posright,postop,posright2,postop2)

' check if mouse

IF useMouse = %True THEN
	flagMouse=MouseHere
END IF

' update display now

do
	select case ClockMode
    case %normal
    	currtime$=TIME$
        if currtime$ <> lasttime$(viewpage) then
        	old$=lasttime$(workpage) ' get time on hidden page
			call updatefix (currtime$,old$,displaytype,posright,postop,posright2,postop2)
            lasttime$(workpage)=currtime$ ' keep time just written on workpage
	        swap viewpage,workpage ' toggle pages now
	    	call viewwork(viewpage,workpage) ' show updated page
			IF flagBeep=%soundon THEN SOUND 444,0.02
        end if
    case %motion
    	currtime$=TIME$
        if currtime$ <> previoustime$ then
            ' perform the page flipping in routines thanks to global vars
			call updatemov (currtime$,previoustime$,displaytype,posright,postop,posright2,postop2)
			previoustime$=currtime$ ' keep time just written on workpage
            delay pause ' wait a little after animation
            call newpaths(sequencetype) ' // v1.0h
        end if

    case %motionFull
        currtime$=TIME$
        if currtime$ <> previoustime$ then
            ' perform the page flipping in routines thanks to global vars
            call updatemovFull (currtime$,previoustime$,displaytype,posright,postop,posright2,postop2)
            previoustime$=currtime$ ' keep time just written on workpage
            ' delay pause ' wait a little after animation
            call newpaths(sequencetype) ' // v1.0h
        end if
	end select
    IF instat THEN EXIT LOOP

    IF ( (useMouse=%True) AND (flagMouse = %True) ) THEN
        IF MouseButtonClicked=%True THEN EXIT LOOP
    END IF

loop
call SetHires(%page1,%page1,ink,paper) ' clear screen 1
CALL TextMode
END %eNone-%eNone

sub updateMov (byval new$,byval old$,byval displaytype,byval posright,byval postop,byval posright2,byval postop2)
' try and increase speed when simultaneous moves?
realetapes=etapes
mismatch=0
for i = 1 to len(new$)
	if mid$(new$,i,1) <> mid$(old$,i,1) then incr mismatch
next
etapes=etapes \ mismatch
for currstep=1 to etapes+1 ' take care of overhead
	SELECT CASE displaytype
	CASE %hhmmss
   		CALL DrawStrMov(new$,old$,1,8,posright,postop,currstep) ' hh:mm:ss
	CASE %hhmm
		CALL DrawStrMov(new$,old$,7,8,posright2,postop2,currstep) ' ss
		CALL DrawStrMov(new$,old$,1,5,posright,postop,currstep) ' hh:mm
	END SELECT
	swap viewpage,workpage ' toggle pages now
	call viewwork(viewpage,workpage) ' show updated page
    ' delay 0.01 to try and avoid flickering ? according to etapes ?
next
etapes=realetapes
END sub

'--------------------------------------------------------------------
'
SUB DrawStrMov(byval new$,byval old$,byval first, byval last,byval fromx, byval fromy, byval currstep)
FOR n = last TO first STEP -1 ' <<
	c$=mid$(old$,n,1)
    olddigit=getshapenumber(c$)
	c$=mid$(new$,n,1)
    digit=getshapenumber(c$)
   	maxwidth=shapeinfo(digit).maxwide
   	DECR fromx,maxwidth ' << update here!!!
    IF digit<>olddigit THEN
       	' build polygon while morphing, return # of vertices stored in Path
		vertices=buildMorphState(olddigit,digit)
        select case currstep
        case 1
	        ' we erase previous number
    	    CALL PlotState(olddigit,fromx,fromy,currstep-1,vertices,0) ' erase source
        	' and set currstep 1
	       	CALL PlotState(digit,fromx,fromy,currstep,vertices,1)
        case (etapes+1)
	        ' we erase previous number
    	    CALL PlotState(olddigit,fromx,fromy,currstep-2,vertices,0)
        	' draw new on workpage at currstep
       		CALL PlotState(digit,fromx,fromy,currstep-1,vertices,1) ' draw target
        case else
	       	' erase old on workpage
    	    ' we erase previous previous drawn page
        	CALL PlotState(olddigit,fromx,fromy,currstep-2,vertices,0)
        	' draw new on workpage at currstep
       		CALL PlotState(digit,fromx,fromy,currstep,vertices,1)
        end select
    END IF
NEXT
END SUB



'--------------------------------------------------------------------
'--------------------------------------------------------------------
' here, we succeed in morphing digits which are not changing !
' we avoid burning monitor! ;-)


' we replace drawstrmov by drawstrmovfull
sub updateMovFull (byval new$,byval old$,byval displaytype,byval posright,byval postop,byval posright2,byval postop2)
' try and increase speed when simultaneous moves?
realetapes=etapes
mismatch=0
for i = 1 to len(new$)
        if mid$(new$,i,1) <> mid$(old$,i,1) then incr mismatch
next
' etapes=etapes \ mismatch ' reduce tempo when several dots
for currstep=1 to etapes+1 ' take care of overhead
        SELECT CASE displaytype
        CASE %hhmmss
                CALL DrawStrMovFull(new$,old$,1,8,posright,postop,currstep) ' hh:mm:ss
        CASE %hhmm
                CALL DrawStrMovFull(new$,old$,7,8,posright2,postop2,currstep) ' ss
                CALL DrawStrMovFull(new$,old$,1,5,posright,postop,currstep) ' hh:mm
        END SELECT
        swap viewpage,workpage ' toggle pages now
        call viewwork(viewpage,workpage) ' show updated page
    ' delay 0.01 to try and avoid flickering ? according to etapes ?
next
etapes=realetapes
END sub

'--------------------------------------------------------------------
'
' we keep normal animation for changed digits but we use
' buildmorphstatefull for fixed digits
SUB DrawStrMovFull(byval new$,byval old$,byval first, byval last,byval fromx, byval fromy, byval currstep)
FOR n = last TO first STEP -1 ' <<
        c$=mid$(old$,n,1)
    olddigit=getshapenumber(c$)
        c$=mid$(new$,n,1)
    digit=getshapenumber(c$)
        maxwidth=shapeinfo(digit).maxwide
        DECR fromx,maxwidth ' << update here!!!
    IF digit<>olddigit THEN
        ' build polygon while morphing, return # of vertices stored in Path
                vertices=buildMorphState(olddigit,digit)
        select case currstep
        case 1
                ' we erase previous number
            CALL PlotState(olddigit,fromx,fromy,currstep-1,vertices,0) ' erase source
                ' and set currstep 1
                CALL PlotState(digit,fromx,fromy,currstep,vertices,1)
        case (etapes+1)
                ' we erase previous number
            CALL PlotState(olddigit,fromx,fromy,currstep-2,vertices,0)
                ' draw new on workpage at currstep
                CALL PlotState(digit,fromx,fromy,currstep-1,vertices,1) ' draw target
        case else
                ' erase old on workpage
            ' we erase previous previous drawn page
                CALL PlotState(olddigit,fromx,fromy,currstep-2,vertices,0)
                ' draw new on workpage at currstep
                CALL PlotState(digit,fromx,fromy,currstep,vertices,1)
        end select
    ELSE
        ' here we morph a digit with itself
        ' build polygon while morphing, return # of vertices stored in Path
                vertices=buildMorphStateFull(olddigit,digit)
        select case currstep
        case 1
                ' we erase previous number
            CALL PlotState(olddigit,fromx,fromy,currstep-1,vertices,0) ' erase source
                ' and set currstep 1
                CALL PlotState(digit,fromx,fromy,currstep,vertices,1)
        case (etapes+1)
                ' we erase previous number
            CALL PlotState(olddigit,fromx,fromy,currstep-2,vertices,0)
                ' draw new on workpage at currstep
                CALL PlotState(digit,fromx,fromy,currstep-1,vertices,1) ' draw target
        case else
                ' erase old on workpage
            ' we erase previous previous drawn page
                CALL PlotState(olddigit,fromx,fromy,currstep-2,vertices,0)
                ' draw new on workpage at currstep
                CALL PlotState(digit,fromx,fromy,currstep,vertices,1)
        end select

    END IF
NEXT
END SUB

'--------------------------------------------------------------------

' build list of vertices stored in Path, while morphing, returns this #
' same digit to same digit here
function buildMorphStateFull (byval sourceshape, byval targetshape)
ptrSource =   ShapeInfo(sourceshape).offset
countSource = ShapeInfo(sourceshape).count
nVertex=countsource ' number of vertices while morphing
FOR p = %firstvertex TO nVertex-1
    Path(p,%Origin)=Vertex(ptrSource)
        Path(p,%Target)=vertex(ptrSource+1)
    INCR ptrSource
NEXT
p=nVertex
Path(p,%Origin)=Vertex(ptrSource)
Path(p,%Target)=Vertex( ShapeInfo(sourceshape).offset ) ' 1st ptrSource
buildmorphstateFull=nvertex
end function


'--------------------------------------------------------------------

' // v1.0h

SUB newpaths(byval sequencetype)
' store vertices temporiraly here before we modify the plotting order
DIM DYNAMIC TmpVertex (%FirstVertex:%LastVertex) AS VertexType ' compile time
LOCAL TmpVertex()
if sequencetype <> %Random then exit sub

currvertex=%FirstVertex
FOR currshape=%FirstShape TO %LastShape
    startofdef = ShapeInfo(currshape).offset
	count      = ShapeInfo(currshape).count
    s$=""
    FOR k = 1 TO count
    	thisdot=k-1 ' important!
        DO
          	' int((upper-lower+1)*rnd+lower)
			newpos=INT((count-1+1)*RND+1) ' [1..num] -> chr(1)..chr$(255)
			newpos=INT((count-1+1)*RND+1)
			newpos=INT((count-1+1)*RND+1)
        LOOP UNTIL INSTR(s$,CHR$(newpos)) = 0 ' find a NEW position
        s$=s$+CHR$(newpos)
        TmpVertex(newpos)=Vertex(startofdef+thisdot)
    NEXT
    ' now we can change order of sequence in array of positions
	FOR k = 1 TO count
    	thisdot=k-1
    	Vertex(startofdef+thisdot)=TmpVertex(k)
    NEXT
NEXT
end sub
