(* ---------------------------------------------------------------
Title         Q&D time converter
Overview
Notes
Bugs
Wish List     support h:m:s ?

--------------------------------------------------------------- *)

MODULE hms;

IMPORT Lib;
IMPORT Str;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

(* ------------------------------------------------------------ *)

CONST
    progEXEname   = "HMS";
    progTitle     = "Q&D Time Converter";
    progVersion   = "v1.0";
    progCopyright = "by PhG";
    banner        = progTitle+" "+progVersion+" "+progCopyright;

(* ------------------------------------------------------------ *)

CONST
    errNone         = 0;
    errHelp         = 1;
    errSyntax       = 2;
    errBadNum       = 3;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
banner+nl+
nl+
"Syntax 1 : "+progEXEname+" [*] [[hh] mm] ss"+nl+
"Syntax 2 : "+progEXEname+" [*] value[s|m|h]"+nl+
nl+
"This program converts from HH:MM:SS to seconds (syntax 1)"+nl+
"or from specified value to HH:MM:SS (syntax 2)."+nl+
nl+
'a) Syntax 1 is either "hh mm ss", "mm ss" or "ss".'+nl+
"   By design, values are not limited to standard ranges."+nl+
'b) Syntax 2 supports "h", "m" and "s" (default) suffixes.'+nl+
'c) If first parameter is "*", padding character will be "0" (default is " ").'+nl+
nl+
"Examples : "+progEXEname+" 11 59 59"+nl+
"           "+progEXEname+" 30 00"+nl+
"           "+progEXEname+" 3600"+nl+
"           "+progEXEname+" * 60m"+nl;

VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :       WrStr(helpmsg);
    | errSyntax:      S:="Unexpected number of parameters !";
    | errBadNum:      S:='Illegal "~" number !';
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE fmtnum (v:LONGCARD;wi:CARDINAL;padchar:CHAR):str16;
VAR
    N:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(v,N,10,ok);
    IF ok THEN
        FOR i:= Str.Length(N)+1 TO wi DO
            Str.Prepend(N,padchar);
        END;
    ELSE
        N:="???";
    END;
    RETURN N;
END fmtnum;

CONST
    BADNUM = MAX(LONGCARD);

PROCEDURE getnum (S:ARRAY OF CHAR   ):LONGCARD;
VAR
    v:LONGCARD;
    ok:BOOLEAN;
BEGIN
    v:=Str.StrToCard(S,10,ok);
    IF NOT(ok) THEN v:=BADNUM;END;
    RETURN v;
END getnum;

CONST
    wi2 = 2; (* hh mm ss 00..24 00..59 *)
    wi5 = 5; (* ssss     00000..86400 *)
VAR
    shift,optcount,i,n:CARDINAL;
    S:str128;
    v,hh,mm,ss,totalseconds:LONGCARD; (* oversized *)
    units : (enHeures,enMinutes,enSecondes,rien);
    padch : CHAR;
    op    : (fromhms,fromval,pbsyntax);
BEGIN
    WrLn;
    optcount:=Lib.ParamCount();
    IF optcount = 0 THEN abort(errHelp,"");END;

    FOR i:= 1 TO optcount DO
        Lib.ParamStr(S,i);
        CASE S[0] OF
        | "-","/":
            Str.Delete(S,0,1);
            Str.Caps(S);
            n:=0;
            IF same(S,"?")    THEN INC(n);END;
            IF same(S,"H")    THEN INC(n);END;
            IF same(S,"HELP") THEN INC(n);END;
            IF n # 0 THEN abort(errHelp,"");END;
        END;
    END;

    Lib.ParamStr(S,1);
    op:=pbsyntax;
    IF same(S,"*") THEN
        padch := "0"; shift:=1;
        CASE optcount OF
        | 2:   op:=fromval;
        | 3,4: op:=fromhms;
        END;
    ELSE
        padch := " "; shift:=0;
        CASE optcount OF
        | 1:   op:=fromval;
        | 2,3: op:=fromhms;
        END;
    END;

    CASE op OF
    | fromval:
        Lib.ParamStr(S,1+shift);
        Str.Caps(S);
        units:=rien;
        i:=Str.Length(S);
        CASE S[i-1] OF
        | "H": units:=enHeures;
        | "M": units:=enMinutes;
        | "S": units:=enSecondes;
        END;
        IF units = rien THEN
            units:=enSecondes;
        ELSE
            Str.Delete(S,i-1,1);
        END;
        v:=getnum(S);
        IF v = BADNUM THEN abort(errBadNum,S);END;
        CASE units OF
        | enHeures:  v:=v*60*60;
        | enMinutes: v:=v*60;
        END;
        totalseconds:=v;
        ss:=totalseconds MOD 60;
        mm:=(totalseconds DIV 60) MOD 60;
        hh:=(totalseconds DIV 3600);
        S:="$s = $h $m $s";
        Str.Subst(S,"$",fmtnum(totalseconds,wi5,padch));
        Str.Subst(S,"$",fmtnum(hh,wi2,padch));
        Str.Subst(S,"$",fmtnum(mm,wi2,padch));
        Str.Subst(S,"$",fmtnum(ss,wi2,padch));
        WrStr(S);WrLn;
    | fromhms:
        hh:=0;
        mm:=0;
        ss:=0;
        FOR i:= 1+shift TO optcount DO
            Lib.ParamStr(S,i);
            Str.Caps(S);
            v:=getnum(S);
            IF v = BADNUM THEN abort(errBadNum,S);END;
            CASE (optcount-i) OF
            | 0: ss:=v;
            | 1: mm:=v;
            | 2: hh:=v;
            END;
        END;
        totalseconds:=hh*60*60+mm*60+ss;
        S:="$h $m $s = $s = $m $s";
        Str.Subst(S,"$",fmtnum(hh,wi2,padch));
        Str.Subst(S,"$",fmtnum(mm,wi2,padch));
        Str.Subst(S,"$",fmtnum(ss,wi2,padch));
        Str.Subst(S,"$",fmtnum(totalseconds,wi5,padch));
        ss:=totalseconds MOD 60;
        mm:=totalseconds DIV 60;
        Str.Subst(S,"$",fmtnum(mm,wi5,padch));
        Str.Subst(S,"$",fmtnum(ss,wi2,padch));
        WrStr(S);WrLn;
    | pbsyntax:
        abort(errSyntax,"");
    END;

    abort(errNone,"");
END hms.

