(* ---------------------------------------------------------------
Title         Q&D NewName
Author        PhG
Overview      remove/replace character(s) in filename(s)
Usage         see help
Notes         very, very, very quick & dirty... :-(
              minimal error messages and checking, etc.
              fix chkValid()
              available options : g i k q u w
              -y has not yet been fully tested
Bugs          Win9X would sometimes not casify as asked to !
              Win9X would sometimes not retrieve all weird chars from filename !

              fix killfile(), avoid msgAlready
Wish List     still to do in order to match older newname0 : count
              recursion ?
              allow at least "?" joker
              process all 0..9 digits at once ?

              lowercase only if all uppercase (AND vice versa)

              alternate recursive doSubst()
              (as to avoid building a string supposed to be deleted)

              remap accents from XP to 98 ?

              yes, -sss, -ppp and -iii:# are kludgy... but useful !

              we too often use UpperCase() ignoring accents

              mid$(), preserve # first or last chars when renaming ?

              yes, yes, space IS supported by DOS but it should NOT be !

--------------------------------------------------------------- *)

MODULE NewName;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT IO;
IMPORT SYSTEM;

FROM IO IMPORT WrStr, WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM Storage IMPORT Available,ALLOCATE,DEALLOCATE;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XopenFile, w9XcloseFile,
w9XsupportLFN;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    vowels        = "aeiouy";
    alphaset      = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"; (* uppercase *)
    symbols       = "_-$!(){}~#&@[]";
    legalLFN      = ",;'.+%="+blank;
    DOScharset    = alphaset+digits+symbols; (* uppercase *)
    LFNcharset    = DOScharset+legalLFN;
    UNDOBATCH     = "ROLLBACK.BAT";
    ignoredrenumprefix  = "*"; (* ignore prefix when renumbering *)
    autonumsep          = "_";
    autonummarker       = "*";
CONST
    MINFIELD      = 1;
    MAXFIELD      = 12;
    DEFFIELD      = 4;
    MINDELTA      = 1;
    MAXDELTA      = 100;
    DEFDELTA      = 1;
CONST
    ProgEXEname   = "NEWNAME";
    ProgTitle     = "Q&D Filename Modifier";
    ProgVersion   = "v1.3h";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    errNone         = 0;
    errHelp         = 1;
    errUnknownOption= 2;
    errParmOverflow = 3;
    errStorage      = 4;
    errCmdConflict  = 5;
    errSyntax       = 6;
    errBadPrefix    = 7;
    errBadSuffix    = 8;
    errBadSpec      = 9;
    errNoMatch      = 10;
    errBadChar      = 11;
    errBadRemap     = 12;
    errRemapNonsense= 13;
    errBadVal       = 14;
    errBadInfix     = 15;
    errNonsense     = 16;
    errRange        = 17;
    errSortConflict = 18;
    errFieldWidth   = 19;
    errWontProcess  = 20;
    errBadSubstring = 21;
    errBadMeta      = 22;
    errLFNonly      = 23;
    errDirectory    = 24;
    (* will become errHelp *)
    errMoreHelp     = 32;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
    cr=CHR(13);
    lf=CHR(10);
    nl=cr+lf;
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
     helpmsg=Banner+nl+
nl+
"Syntax 1  : "+ProgEXEname+" [-f] <string> <file(s)> [option]..."+nl+
"Syntax 2  : "+ProgEXEname+" [-m] <oldstring> <newstring> <file(s)> [option]..."+nl+
"Syntax 3  : "+ProgEXEname+" <-p[p[p]> <prefix|"+autonummarker+"> <file(s)> [option]..."+nl+
"Syntax 4  : "+ProgEXEname+" <-s[s[s]> <suffix|"+autonummarker+"> <file(s)> [option]..."+nl+
"Syntax 5  : "+ProgEXEname+" <-i[i[i]]:#> <insertion> <file(s)> [option]..."+nl+
"Syntax 6  : "+ProgEXEname+" <-z> <substring> <file(s)> [option]..."+nl+
"Syntax 7  : "+ProgEXEname+" <-c> <oldsubstring> <newsubstring> <file(s)> [option]..."+nl+
"Syntax 8  : "+ProgEXEname+" <-lc[c]|-uc[c]> <file(s)> [option]..."+nl+
"Syntax 9  : "+ProgEXEname+" <-n[n]> <prefix|"+ignoredrenumprefix+"> <first> <file(s)> [option]..."+nl+
"Syntax 10 : "+ProgEXEname+" <-t[#]> <file(s)> [option]..."+nl+
"Syntax 11 : "+ProgEXEname+" <-b|-j> <file(s)> [option]..."+nl+
"Syntax 12 : "+ProgEXEname+" <-trim|-blanks> <file(s)> [option]..."+nl+
nl+
"This program filters out characters from <string> (syntax 1) ;"+nl+
"remaps character(s) from <oldstring> using <newstring> (syntax 2) ;"+nl+
"processes prefixes (syntax 3), suffixes (syntax 4) and insertions (syntax 5) ;"+nl+
"removes substrings (syntax 6) ; changes <oldstring> to <newstring> (syntax 7) ;"+nl+
"changes LFN filenames case to lower or upper (syntax 8) ;"+nl+
"renumbers filenames (syntax 9) ; changes filenames to DOS form (syntax 10) ;"+nl+
"prefixes or suffixes LFN filenames with YMD_HMS datetime stamp (syntax 11) ;"+nl+
"trims spaces at right and left of filenames (syntax 12)."+nl+
nl+
"-f         filter out mode (default if two parameters were specified)"+nl+
"-m         remap mode (default if three parameters were specified)"+nl+
"-p[p[p]]   prefix mode (-pp = -d -p, -ppp = -pp deleting length of <prefix>)"+nl+
"-s[s[s]]   suffix mode (-ss = -d -s, -sss = -ss deleting length of <suffix>)"+nl+
"-i[i[i]]:# insert at specified 0-based column mode (-ii:# = -d -i:#)"+nl+
"-z         remove first substring (operation not recursive)"+nl+
"-c         change first oldsubstring to newsubstring (operation not recursive)"+nl+
"-<l|u>c[?] convert to lowercase or uppercase mode"+nl+
"           (-lc and -uc = keep accents, -lcc and -ucc = remove accents,"+nl+
"           -lcf[r] and -ucf[r] = keep French accents)"+nl+
"-az        remove accents while keeping original case"+nl+
'-n[n]      renumber mode (-nn = "-n '+ignoredrenumprefix+'" syntax = do not use <prefix>)'+nl+
'-t[#]      truncate LFN format to DOS format mode ([1..8], default method is 1)'+nl+
"-<b|j>     prefix or suffix file datetime stamp to filename"+nl+
"-trim      trim spaces at right and left"+nl+
nl+
"-r[r]      do rename file(s) (-rr = create undo "+UNDOBATCH+")"+nl+
"-l         disable LFN support even if available"+nl+
"-??        more help (to be read at least once !)"+nl;

       morehelpmsg =
"-d[?]      delete prefix or suffix (-ds = -s -d, -dp = -p -d)"+nl+
"-a         take hidden and system files into account"+nl+
"-o         rename even if new name already exists (previous content lost)"+nl+
"-i         allow renaming to same name"+nl+
"-e         process extension (only when removing, remapping or converting)"+nl+ (* chars *)
"-f:#       number field width ([1..12], default is 4)"+nl+
"-k:#       renumbering increment ([1..100], default is 1)"+nl+
"-s<?>[-]   sort when renumbering (n = name, e = extension, z = size, d = date)"+nl+
"-x         prepend renumbered value and <prefix> to <filename> (-n[n] only)"+nl+
"-<y|u>     try and make sure new filename is unique (-t[#] only)"+nl+
"-v         ignore character set validity check"+nl+
nl+
'Valid set (DOS) : "'+DOScharset+'"'+nl+
'Valid set (LFN) : "'+LFNcharset+'"'+nl+
nl+
"a) By design, this program requires <file(s)> not to be a directory alone."+nl+
"b) Any option irrelevant to specified mode command will be silently ignored."+nl+
"c) Program will handle up to 65530 files. "+UNDOBATCH+" name is reserved."+nl+
"d) Note collisions cannot be previewed : use -r only when sure of results !"+nl+
'e) Auto-numbering "-s *" and "-p *" syntaxes use "'+autonumsep+'" as a separator.'+nl+
"f) -ppp, -sss and -iii:# options delete # first, last or middle characters,"+nl+
"   where # is length of specified <prefix>, <suffix> or <insertion>."+nl+
"g) -t1 truncates characters from end of filename ;"+nl+
'   -t2 is -t1 but removes all characters preceding first "-" character if any ;'+nl+
"   -t3 is -t2 but keeps first char of filename ;"+nl+
"   -t4 is -t1 but does not try and preserve trailing digits"+nl+
"   (note -t1, -t2 and -t3 try and preserve trailing digits) ;"+nl+
"   -t4, -t5 and -t6 are identical to -t1, -t2 and -t3"+nl+
"   but they preprocess filenames by removing vowels."+nl+
"   With all -t[#] options, file extension is just truncated to fit DOS format :"+nl+
"   -e option is therefore ignored."+nl+
"   Note all -t[#] options ignore filenames already fitting DOS format."+nl+
'h) Support for escape sequences is minimalist : "'+escCh + '[' + escSet + ']" inserts symbol,' + nl +
'   while "' + escCh + "[$|x]#[" + escCh + ']" inserts (hexa)decimal character value.'+nl+
"i) Note this program does not support any string joker."+nl+
nl+
"EXamples : "+ProgEXEname+" -t2 f:\data\pictures\*.*"+nl+
"           "+ProgEXEname+" -n pic 1 -f:2 *.jpg"+nl+
"           "+ProgEXEname+" -c iso pim *.mod"+nl+
"           "+ProgEXEname+" -p m2_ src\*.mod"+nl+
"           "+ProgEXEname+" -b *.obj"+nl;

VAR
    S : str1024;
BEGIN
    CASE e OF
    | errHelp,errMoreHelp :
        WrStr(helpmsg);
        IF e = errMoreHelp THEN
            WrStr(morehelpmsg);
            e:=errHelp;
        END;
    | errUnknownOption :
        Str.Concat(S,"Unknown ",einfo);Str.Append(S," option !");
    | errParmOverflow :
        Str.Concat(S,"Uneeded ",einfo);Str.Append(S," parameter !");
    | errStorage :
        S := "Storage ALLOCATE() failure !";
    | errCmdConflict:
        S := "Mutually exclusive commands !";
    | errSyntax :
        S := "Syntax error !";
    | errBadPrefix:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" prefix !');
    | errBadSuffix:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" suffix !');
    | errBadSpec:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" specification !');
    | errNoMatch:
        Str.Concat(S,'No match for "',einfo);Str.Append(S,'" specification !');
    | errBadChar:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" string !');
    | errBadRemap:
        S:="<oldstring> and <newstring> must have the same length !";
    | errRemapNonsense:
        S:="Remapping a character to itself is a nonsense !";
    | errBadVal:
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," value !");
    | errBadInfix:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" insertion !');
    | errNonsense:
        Str.Concat(S,einfo," option is a nonsense with specified command !");
    | errRange:
        Str.Concat(S,einfo," is not in expected range !");
    | errSortConflict:
        S := "Mutually exclusive sort methods !";
    | errFieldWidth:
        S := "Field width is too small for total number of files to process !";
    | errWontProcess:
        (* S := "-rr option prevents processing of existing "+UNDOBATCH+" !"; *)
        S := "As "+UNDOBATCH+" was found, -rr will not process any file !";
    | errBadSubstring:
        Str.Concat(S,'Illegal "',einfo);Str.Append(S,'" substring !');
    | errBadMeta:
        Str.Concat(S,'Illegal escape sequence ("',einfo);Str.Append(S,'") !');
    | errLFNonly:
        S := "LFN support is required with this command !"; (* keep it short *)
    | errDirectory:
        Str.Concat(S,'Expecting <file(s)> in "',einfo);Str.Append(S,'" directory specification !');
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

PROCEDURE dbgstr (DEBUG:BOOLEAN;S1,S2:ARRAY OF CHAR   );
CONST
    wi = 20;
VAR
    i:CARDINAL;
BEGIN
    IF DEBUG THEN
        WrStr("// ");WrStr(S1);
        FOR i:=Str.Length(S1)+1 TO wi DO WrStr(" ");END;
        WrStr(S2); WrLn;
    END;
END dbgstr;

PROCEDURE dbgv (DEBUG:BOOLEAN;S1:ARRAY OF CHAR;v:CARDINAL   );
CONST
    wi = 20;
VAR
    i:CARDINAL;
BEGIN
    IF DEBUG THEN
        WrStr("// ");WrStr(S1);
        FOR i:=Str.Length(S1)+1 TO wi DO WrStr(" ");END;
        IO.WrCard(v,1); WrLn;
    END;
END dbgv;

PROCEDURE WrHexStr (hex:BOOLEAN;S:pathtype);
CONST
    shx     = "0123456789abcdef";
    gremlin = CHR(168);
VAR
    len,i,c,v:CARDINAL;
BEGIN
    len:=Str.Length(S);
    FOR i:=1 TO len DO
        c:=ORD( S[i-1] );
        IF hex THEN
            v:=(c AND 0F0H);
            v:=(v >> 4);
            WrStr(shx[v]);
            v:=(c AND 00FH);
            WrStr(shx[v]);
        ELSE
            WrStr(" ");
            IF c < ORD(blank) THEN
                WrStr(gremlin);
            ELSE
                WrStr(CHR(c));
            END;
        END;
        IF i < len THEN WrStr(" ");END;
    END;
END WrHexStr;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferOut : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    firstEntry = 1; (* 1..count *)
TYPE
    ptrToEntry = POINTER TO entryType;
    entryType = RECORD
        next   : ptrToEntry;
        index  : CARDINAL; (* for possible later sort *)
        fsize  : LONGCARD;
        fdate  : CARDINAL;
        ftime  : CARDINAL;
        slen   : CARDINAL; (* a SHORTCARD would DO fine IN almost all cases  *)
        string : CHAR;     (* here, after other data, because variable length *)
    END;

VAR
    anchorSpec  : ptrToEntry; (* globerk *)

PROCEDURE getAnchor (  ):ptrToEntry;
BEGIN
    RETURN anchorSpec;
END getAnchor;

PROCEDURE initList (VAR anchor:ptrToEntry; VAR count:CARDINAL);
BEGIN
    count := firstEntry-1; (* complicated way to mean 0 *)
    anchor:= NIL;
END initList;

PROCEDURE isReservedEntry (S : ARRAY OF CHAR) : BOOLEAN; (* assume uppercase *)
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE buildMatchList (VAR anchor:ptrToEntry;
                          useLFN,hiddensystem : BOOLEAN;spec:pathtype):CARDINAL;
VAR
    S:pathtype; (* some are oversized but safety first ! *)
    newInList : ptrToEntry;
    len,needed,count : CARDINAL;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    DOSentry     : FIO.DirEntry;
    rc,found:BOOLEAN;
    dosattr:FIO.FileAttr;
    fsize:LONGCARD;
    fdate,ftime:CARDINAL;
BEGIN
    IF hiddensystem THEN
        dosattr:=allfiles;
    ELSE
        dosattr := FIO.FileAttr{aR,aA};
    END;

    count:=0;

    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(dosattr),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,dosattr,DOSentry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(S,w9Xentry.fullfilename);
            fsize := w9Xentry.fsize.lo; (* no file should be more than 4Gb anyway ;-) *)
            (*
            QD_LFN calls win9XfindFirst requiring DOS format
            DOS date/time values in low double-word of file time QWORD
            (date is high word, time is low word of double-word)
            darn, it's not creation field but lastmod seems ok
            *)
            fdate := CARDINAL(w9Xentry.lastmod.lo >> 16);
            ftime := CARDINAL(w9Xentry.lastmod.lo AND 0FFFFH);
       ELSE
            Str.Copy(S,DOSentry.Name);
            fsize := DOSentry.size;     (* thanks, JPI, for using a reserved word here ! *)
            fdate := DOSentry.date;
            ftime := DOSentry.time;
        END;
        IF isReservedEntry(S) THEN (* skip "." ".." *)
            ; (* silently ignore this spec *)
        ELSE
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=DOSentry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                IF count=MAX(CARDINAL) THEN
                    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL);
                END; (* too many files but let's fake ALLOCATE failure *)
                len:=Str.Length(S);
                needed:=SIZE(entryType)-SIZE(CHAR)+len;
                IF Available(needed)=FALSE THEN
                    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL);
                END; (* storage ALLOCATE failure *)
                IF anchor=NIL THEN
                    ALLOCATE( anchor,needed);
                    newInList := anchor;
                ELSE
                    ALLOCATE(newInList^.next,needed);
                    newInList :=newInList^.next;
                END;
                Lib.FastMove( ADR(S),ADR(newInList^.string),len);
                newInList^.slen := len;
                newInList^.fsize:= fsize;
                newInList^.fdate:= fdate;
                newInList^.ftime:= ftime;
                newInList^.next := NIL;
                newInList^.index:= 0; (* filled later *)
                INC(count);
            END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(DOSentry);
        END;
    END;
    IF useLFN THEN rc:=w9XfindClose(w9Xhandle,errcode); END;
    RETURN count;
END buildMatchList;

PROCEDURE freeMatchList (anchor:ptrToEntry);
VAR
    len,needed      : CARDINAL;
    firstInList,newInList   : ptrToEntry;
BEGIN
    firstInList := anchor;
    newInList := firstInList;
    WHILE newInList # NIL DO
        len         := CARDINAL(newInList^.slen);
        needed      := SIZE(entryType)-SIZE(CHAR)+len;
        firstInList := firstInList^.next;
        DEALLOCATE (newInList,needed);
        newInList := firstInList;
    END
END freeMatchList;

PROCEDURE findEntry (n:CARDINAL;  anchor:ptrToEntry  ):ptrToEntry;
VAR
    newInList:ptrToEntry;
BEGIN
    newInList := anchor;
    LOOP
        IF newInList = NIL THEN EXIT;END; (* gloups! *)
        IF newInList^.index = n THEN EXIT; END;
        newInList := newInList^.next;
    END;
    RETURN newInList;
END findEntry;

PROCEDURE getEntry (VAR R:pathtype;
                   n:CARDINAL; anchor:ptrToEntry);
VAR
    len:CARDINAL;
    newInList:ptrToEntry;
    S:pathtype;
BEGIN
    newInList := findEntry(n,anchor);
    IF newInList = NIL THEN Str.Copy(R,"???"); RETURN;END;
    len         := newInList^.slen;
    Lib.FastMove( ADR(newInList^.string),ADR(S),len);
    S[len]      := charnull; (* REQUIRED safety ! *)
    Str.Copy(R,S); (* yep, compiler won't let us fill R directly *)
END getEntry;

PROCEDURE getFsizeEntry (VAR R:LONGCARD;
                        n:CARDINAL; anchor:ptrToEntry);
VAR
    newInList:ptrToEntry;
BEGIN
    newInList := findEntry(n,anchor);
    IF newInList = NIL THEN R:=MAX(LONGCARD); RETURN;END;
    R := newInList^.fsize;
END getFsizeEntry;

PROCEDURE getStampEntry (VAR D,T:CARDINAL;n:CARDINAL;anchor:ptrToEntry);
VAR
    newInList:ptrToEntry;
BEGIN
    newInList := findEntry(n,anchor);
    IF newInList = NIL THEN D:=0; T:=0; RETURN ; END;
    D := newInList^.fdate;
    T := newInList^.ftime;
END getStampEntry;

PROCEDURE initIndexes (anchor:ptrToEntry );
VAR
    newInList:ptrToEntry;
    i : CARDINAL;
BEGIN
    i := firstEntry;
    newInList:=anchor;
    LOOP
        IF newInList = NIL THEN EXIT; END;
        newInList^.index := i;
        INC(i);
        newInList := newInList^.next;
    END;
END initIndexes;

(* ------------------------------------------------------------ *)

PROCEDURE doswap (i,j:CARDINAL);
VAR
    anchor,pi,pj:ptrToEntry;
    tmp:CARDINAL;
BEGIN
    anchor:=getAnchor();
    pi:=findEntry(i,anchor);
    pj:=findEntry(j,anchor);
    tmp:=pi^.index;
    pi^.index:=pj^.index;
    pj^.index:=tmp;
END doswap;

PROCEDURE dolessname (i,j:CARDINAL ):BOOLEAN ;
VAR
    SI,SJ: pathtype ;
    anchor:ptrToEntry;
BEGIN
    anchor:=getAnchor();
    getEntry(SI,i, anchor);
    getEntry(SJ,j, anchor);
    LowerCase(SI); (* ignore accents when sorting *)
    LowerCase(SJ);
    IF Str.Compare(SI,SJ) < 0 THEN (* -1 is less *)
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END dolessname;

PROCEDURE dolessext ( i,j:CARDINAL  ):BOOLEAN ;
VAR
    e1,e2:pathtype;
    p : CARDINAL;
    anchor:ptrToEntry;
BEGIN
    anchor:=getAnchor();
    getEntry(e1,i,anchor);
    getEntry(e2,j,anchor);
    LowerCase(e1); (* ignore accents when sorting *)
    LowerCase(e2);

    p := Str.RCharPos(e1,dot);
    IF p = MAX(CARDINAL) THEN
        e1:="";
    ELSE
        Str.Delete(e1,0,p+1);
    END;
    p := Str.RCharPos(e2,dot);
    IF p = MAX(CARDINAL) THEN
        e2:="";
    ELSE
        Str.Delete(e2,0,p+1);
    END;
    CASE Str.Compare(e1,e2) OF
    | -1 : RETURN TRUE;
    |  0 : RETURN dolessname(i,j);
    |  1 : RETURN FALSE;
    END;
END dolessext;

PROCEDURE dolesssize (i,j:CARDINAL ):BOOLEAN ;
VAR
    anchor:ptrToEntry;
    fsizeI,fsizeJ:LONGCARD;
BEGIN
    anchor:=getAnchor();
    getFsizeEntry(fsizeI,i, anchor);
    getFsizeEntry(fsizeJ,j, anchor);
    IF (fsizeI = fsizeJ) THEN
        RETURN dolessname(i,j);
    ELSE
        RETURN (fsizeI < fsizeJ);
    END;
END dolesssize;

PROCEDURE dolessstamp (i,j:CARDINAL):BOOLEAN;
VAR
    anchor:ptrToEntry;
    fdateI,ftimeI,fdateJ,ftimeJ:CARDINAL;
BEGIN
    anchor:=getAnchor();
    getStampEntry(fdateI,ftimeI,i,anchor);
    getStampEntry(fdateJ,ftimeJ,j,anchor);
    IF fdateI < fdateJ THEN RETURN TRUE; END;
    IF fdateI > fdateJ THEN RETURN FALSE; END;
    RETURN (ftimeI < ftimeJ);
END dolessstamp;

PROCEDURE dolessnameINV (i,j:CARDINAL ):BOOLEAN ;
BEGIN
    RETURN NOT (dolessname(i,j));
END dolessnameINV;

PROCEDURE dolessextINV ( i,j:CARDINAL  ):BOOLEAN ;
BEGIN
    RETURN NOT(dolessext(i,j));
END dolessextINV;

PROCEDURE dolesssizeINV (i,j:CARDINAL ):BOOLEAN ;
BEGIN
    RETURN NOT(dolesssize(i,j));
END dolesssizeINV;

PROCEDURE dolessstampINV (i,j:CARDINAL):BOOLEAN;
BEGIN
    RETURN NOT(dolessstamp(i,j));
END dolessstampINV;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE WrQ (S:ARRAY OF CHAR);
BEGIN
    WrStr(dquote);WrStr(S);WrStr(dquote);
END WrQ;

(* only handle "-" alone *)

PROCEDURE isOptionForReal (S:ARRAY OF CHAR  ):BOOLEAN ;
BEGIN
    IF same(S,dash) THEN RETURN FALSE; END; (* don't be fooled by a dash alone *)
    RETURN isOption(S);
END isOptionForReal;

PROCEDURE killfile (useLFN,killRO:BOOLEAN;S:pathtype);
BEGIN
    (* //V13G *)
    IF fileIsRO(useLFN,S) THEN
        IF killRO THEN fileSetRW(useLFN,S); END;
    END;
    fileErase(useLFN,S);
END killfile;

PROCEDURE padnum (lc:LONGCARD;wi:CARDINAL   ):str80;
VAR
    R:str80;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(lc,R,10,ok);
    FOR i:=Str.Length(R)+1 TO wi DO
        Str.Prepend(R,"0");
    END;
    RETURN R;
END padnum;

(* ------------------------------------------------------------ *)

TYPE
    cmdtype = (nada,
              prefix,suffix,infix,
              prefixkill,suffixkill,infixkill,
              lenprefixkill,lensuffixkill,leninfixkill,
              remove,remap,renum,substringkill,
              substringreplace,
              strlowercase,struppercase,
              strlowercasenoaccents,struppercasenoaccents,
              strlowercasefrench,struppercasefrench,
              strasciionly,
              suffixautonum,prefixautonum,
              shorten,
              prefixstamp,suffixstamp,
              trimname);

PROCEDURE chkcmd (VAR cmd:cmdtype; wanted:cmdtype ):BOOLEAN;
BEGIN
    IF ( (cmd = nada) OR (cmd=wanted) ) THEN
        cmd:=wanted;
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END chkcmd;

TYPE
    sorttype = (nosort,
               bysize,byname,byextension,bystamp,
               bysizeINV,bynameINV,byextensionINV,bystampINV); (* bydatetime is neverware *)

PROCEDURE chksort (VAR sort:sorttype; wanted:sorttype ):BOOLEAN;
BEGIN
    IF ( (sort= nosort) OR (sort=wanted) ) THEN
        sort:=wanted;
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END chksort;

PROCEDURE chkChars (useLFN,allowdot,allowDOSspace:BOOLEAN;S:ARRAY OF CHAR):BOOLEAN;
VAR
    R,allowed:pathtype;
BEGIN
    IF useLFN THEN
        Str.Copy(allowed,LFNcharset);
        allowdot:=TRUE;
    ELSE
        IF allowDOSspace THEN
            Str.Copy(allowed,blank+DOScharset); (* only for removing chars *)
        ELSE
            Str.Copy(allowed,DOScharset);
        END;
    END;
    IF allowdot THEN
        Str.Append(allowed,dot); (* for DOS *)
    ELSE
        Str.Subst(allowed,dot,"");
    END;
    Str.Copy(R,S);
    UpperCase(R);
    RETURN verifyString (R,allowed);
END chkChars;

PROCEDURE chkValid (useLFN,isSpec,chkvalidchar:BOOLEAN;S:ARRAY OF CHAR;
                   VAR base:ARRAY OF CHAR):BOOLEAN;
VAR
    R,u,d,n,e:pathtype;
BEGIN
    qdSplitAllPath(S,u,d,n,e);
    IF isSpec THEN
        Lib.MakeAllPath(base,u,d,"","");
        IF same(base,"")=FALSE THEN fixDirectory(base);END;
    ELSE
        IF chkJoker(S) THEN RETURN FALSE; END;
        Lib.MakeAllPath(R,"","",n,e);
        IF same(S,R)=FALSE THEN RETURN FALSE;END; (* a path is not a valid fix *)
        IF chkvalidchar THEN RETURN chkChars(useLFN,TRUE,FALSE,R); END;
    END;
    RETURN TRUE;
END chkValid;

PROCEDURE chkRange (VAR v:CARDINAL;
                   lc:LONGCARD;lower,upper:CARDINAL):BOOLEAN ;
BEGIN
    IF lc > MAX(CARDINAL) THEN RETURN FALSE;END;
    v:=CARDINAL(lc); (* don't bother to check against filename len *)
    IF v < lower THEN RETURN FALSE;END;
    IF v > upper THEN RETURN FALSE;END;
    RETURN TRUE;
END chkRange;

(* ------------------------------------------------------------ *)

CONST
    placeholder    = "|"; (* does not belong to DOS or LFN charsets *)
    msgTooShort    = "--- Renamed "+placeholder+" would be empty !";
    msgTooLong     = "--- Renamed "+placeholder+" would be too long !";
    msgTooLongExt  = "--- Renamed "+placeholder+" extension would be too long !";
 (* msgAlready     = "--- Renamed "+placeholder+" already exists !"; *)
    msgAlready     = "--- "+placeholder+" would become existing "+placeholder+" !";
    msgWouldBe     = "::: "+placeholder+" would become "+placeholder;
    msgIsNow       = "+++ "+placeholder+" is now "+placeholder;
    msgFileExists  = " (file exists)";
    msgFileExisted = " (file existed)";

PROCEDURE msg (useLFN,doCR:BOOLEAN;S,F:ARRAY OF CHAR   );
VAR
    R:str1024; (* oversized *)
BEGIN
    Str.Copy(R,S);
    IF useLFN THEN Str.Subst(R,placeholder,dquote+placeholder+dquote);END;
    Str.Subst(R,placeholder,F);
    WrStr(R);IF doCR THEN WrLn; END;
END msg;

PROCEDURE msg2 (useLFN,doCR:BOOLEAN;S,F1,F2:ARRAY OF CHAR   );
VAR
    R:str1024; (* oversized *)
BEGIN
    Str.Copy(R,S);
    IF useLFN THEN Str.Subst(R,placeholder,dquote+placeholder+dquote);END;
    Str.Subst(R,placeholder,F1);
    IF useLFN THEN Str.Subst(R,placeholder,dquote+placeholder+dquote);END;
    Str.Subst(R,placeholder,F2);
    WrStr(R);IF doCR THEN WrLn; END;
END msg2;

CONST
    MINF8   = 1;        MAXF8   = 8;
    MINE3   = 0;        MAXE3   = 3+1;
    (* //V12o fix : note empty extension is still a dot here ! *)
    MINLFN  = 1;        MAXLFN  = SIZE(pathtype);
    MINLFNE = 0;        MAXLFNE = SIZE(pathtype); (* ouais... *)

PROCEDURE chknewlen (VAR messages:CARDINAL;
                    procext,useLFN,DEBUG:BOOLEAN;oldname,n,e:pathtype):BOOLEAN;
VAR
    wecandoit:BOOLEAN;
    action,mini,maxi,pass,lastpass,len : CARDINAL;
    ch:CHAR;
BEGIN
    IF procext THEN
        lastpass:=1;
    ELSE
        lastpass:=2;
    END;
    (*
    IF DEBUG THEN
        WrStr("/// n : ");WrStr(n);WrLn;
        WrStr("/// e : ");WrStr(e);WrLn;
    END;
    *)
    action:=0;
    FOR pass := 1 TO lastpass DO
        CASE pass OF
        | 1: len:=Str.Length(n);
             ch:="n";
             IF useLFN THEN
                 mini:=MINLFN;  maxi:=MAXLFN;
             ELSE
                 mini:=MINF8;   maxi:=MAXF8;
             END;
             IF len < mini THEN
                 msg(useLFN,TRUE, msgTooShort,oldname); INC(action);
             ELSIF len > maxi THEN
                 msg(useLFN,TRUE,msgTooLong,oldname);   INC(action);
             END;
        | 2: len:=Str.Length(e);
             ch:="e";
             IF useLFN THEN
                 mini:=MINLFNE; maxi:=MAXLFNE;
             ELSE
                 mini:=MINE3;   maxi:=MAXE3; (* mini is useless here for 0 is ok *)
             END;
             IF DEBUG THEN
                ;
             END;
             IF len > maxi THEN
                 msg(useLFN,TRUE,msgTooLongExt,oldname);  INC(action);
             END;
        END;
        (*
        IF DEBUG THEN
             WrStr("/// part,mini,maxi,len,action : ");WrStr(ch);
             IO.WrCard(mini,4);IO.WrCard(maxi,4);
             IO.WrCard(len,4);IO.WrCard(action,4);WrLn;
        END;
        *)
    END;
    IF action # 0 THEN INC(messages);END;
    RETURN (action = 0);
END chknewlen;

PROCEDURE doClean (VAR n,e:pathtype; procext:BOOLEAN;str:ARRAY OF CHAR):BOOLEAN ;
VAR
    filename,extension,delme:pathtype;
    action,p,i,pass,lastpass : CARDINAL;
    ch:CHAR;
BEGIN
    IF procext THEN
       lastpass:=2
    ELSE
       lastpass:=1;
    END;

    Str.Copy(filename,n);
    Str.Copy(extension,e);
    Str.Copy(delme,str);
    UpperCase(filename);
    UpperCase(extension);
    UpperCase(delme);

    action := 0;
    FOR pass:=1 TO lastpass DO
        FOR i:=1 TO Str.Length(delme) DO
            ch:=delme[i-1];
            LOOP
                CASE pass OF
                | 1 : p:=Str.CharPos(filename,ch);
                      IF p # MAX(CARDINAL) THEN
                          Str.Delete(filename,p,1);
                          Str.Delete(n,p,1);
                          INC(action);
                      END;
                | 2 : p:=Str.CharPos(extension,ch);
                      IF p # MAX(CARDINAL) THEN
                          Str.Delete(extension,p,1);
                          Str.Delete(e,p,1);
                          INC(action);
                      END;
                END;
                IF p = MAX(CARDINAL) THEN EXIT; END;
            END;
        END;
    END;
    RETURN (action # 0);
END doClean;

PROCEDURE doRemap (VAR n,e:pathtype; procext:BOOLEAN;str,mapstr:ARRAY OF CHAR):BOOLEAN ;
VAR
    filename,extension,mapme:pathtype;
    action,p,i,pass,lastpass,len : CARDINAL;
    ch:CHAR;
BEGIN
    IF procext THEN
       lastpass:=2
    ELSE
       lastpass:=1;
    END;

    Str.Copy(filename,n);
    Str.Copy(extension,e);
    Str.Copy(mapme,str);
    UpperCase(filename);
    UpperCase(extension);
    UpperCase(mapme);

    action:=0;
    FOR pass:=1 TO lastpass DO
        CASE pass OF
        | 1 : len:=Str.Length(filename);
        | 2 : len:=Str.Length(extension);
        END;
        FOR i:=1 TO len DO
            CASE pass OF
            | 1 : p:=Str.CharPos(mapme,filename[i-1]);
                  IF p # MAX(CARDINAL) THEN
                      n[i-1]:=mapstr[p];
                      INC(action);
                  END;
            | 2 : p:=Str.CharPos(mapme,extension[i-1]);
                  IF p # MAX(CARDINAL) THEN
                      e[i-1]:=mapstr[p];
                      INC(action);
                  END;
            END;
        END;
    END;
    RETURN (action # 0);
END doRemap;

PROCEDURE doCasify (VAR n,e:pathtype; op:cmdtype;procext:BOOLEAN ):BOOLEAN;
VAR
    ok:BOOLEAN;
    orgn,orge:pathtype;
BEGIN
    Str.Copy(orgn,n);
    Str.Copy(orge,e);
        CASE op OF
        | strlowercase:          LowerCaseAlt(n);
        | struppercase:          UpperCaseAlt(n);
        | strlowercasenoaccents: LowerCase(n);
        | struppercasenoaccents: UpperCase(n);
        | strlowercasefrench:    LowerCaseFR(n);
        | struppercasefrench:    UpperCaseFR(n);
        | strasciionly:          ASCIIonly(n);
        END;
        ok:=NOT ( same (orgn,n) );
    IF procext THEN
        CASE op OF
        | strlowercase:          LowerCaseAlt(e);
        | struppercase:          UpperCaseAlt(e);
        | strlowercasenoaccents: LowerCase(e);
        | struppercasenoaccents: UpperCase(e);
        | strlowercasefrench:    LowerCaseFR(e);
        | struppercasefrench:    UpperCaseFR(e);
        | strasciionly:          ASCIIonly(e);
        END;
        ok:= ( ok OR ( NOT (same(orge,e) ) ) );  (* //V13A NOT AND here ! *)
    END;
    RETURN ok;
END doCasify;

PROCEDURE doTrimName (VAR n,e:pathtype; procext:BOOLEAN ):BOOLEAN;
VAR
    ok:BOOLEAN;
    orgn,orge:pathtype;
BEGIN
    Str.Copy(orgn,n);
    Str.Copy(orge,e);
    LtrimBlanks(n);
    RtrimBlanks(n);
    ok:=NOT ( same (orgn,n) );
    IF procext THEN
        LtrimBlanks(e);
        RtrimBlanks(e);
        ok:= ( ok OR ( NOT (same(orge,e) ) ) );  (* //V13A NOT AND here ! *)
    END;
    RETURN ok;
END doTrimName;

(* ------------------------------------------------------------ *)

PROCEDURE doSubst (VAR n,e:pathtype; procext:BOOLEAN;oldstr,newstr:ARRAY OF CHAR):BOOLEAN ;
VAR
    filename,extension,ancien,nouveau,newn,newe,tmp:pathtype;
    action,p,i,pass,lastpass,len : CARDINAL;
BEGIN
    IF procext THEN
       lastpass:=2
    ELSE
       lastpass:=1;
    END;

    Str.Copy(filename,n);
    Str.Copy(extension,e);
    Str.Copy(ancien,oldstr);
    Str.Copy(nouveau,newstr);
    UpperCase(filename);
    UpperCase(extension);
    UpperCase(ancien);
    UpperCase(nouveau);

    newn:="";
    newe:="";
    len:=Str.Length(ancien);
    action := 0;
    FOR pass:=1 TO lastpass DO
        LOOP
            CASE pass OF
            | 1 : p:=Str.Pos(filename,ancien);
                  IF p # MAX(CARDINAL) THEN
                      Str.Delete(filename,0,p+len);
                      Str.Slice(tmp,n,0,p);
                      Str.Delete(n,0,p+len);
                      Str.Append(newn,tmp);
                      Str.Append(newn,newstr);
                      INC(action);
                  END;
            | 2 : p:=Str.Pos(extension,ancien);
                  IF p # MAX(CARDINAL) THEN
                      Str.Delete(extension,0,p+len);
                      Str.Slice(tmp,e,0,p);
                      Str.Delete(e,0,p+len);
                      Str.Append(newe,tmp);
                      Str.Append(newe,newstr);
                      INC(action);
                  END;
            END;
            IF p = MAX(CARDINAL) THEN EXIT; END;
        END;
    END;
    IF same(newn,"")=FALSE THEN Str.Prepend(n,newn);END;
    IF same(newe,"")=FALSE THEN Str.Prepend(e,newe);END;

    RETURN (action # 0);
END doSubst;

CONST
    shortennormal       = 01H; (* 00000001 *)
    shortendash         = 02H; (* 00000010 *)
    shorteninitial      = 04H; (* 00000100 *)
    shortenvowel        = 08H; (* 00001000 *)
    shortentruncate     = 10H; (* 00010000 *)
    shortenmachete      = 20H; (* 00100000 *)

PROCEDURE bitarmed (mask,wanted:CARDINAL):BOOLEAN;
BEGIN
    RETURN ( (mask AND wanted) = wanted);
END bitarmed;

PROCEDURE adjustlen (VAR S:pathtype; DEBUG,machete:BOOLEAN;mini,maxi:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    i:=Str.Length(S);
    IF machete THEN
        IF i # 0 THEN Str.Delete(S,i-1,1); END;
    ELSE
        LOOP
            IF i <= mini THEN EXIT;END;
            IF i <= maxi THEN EXIT;END;
            Str.Delete(S,i-1,1);
            DEC(i);
        END;
    END;
END adjustlen;

PROCEDURE novowels (VAR S:pathtype;DEBUG:BOOLEAN );
VAR
    R:pathtype;
    i:CARDINAL;
    ch:CHAR;
BEGIN
    R:="";
    FOR i:=1 TO Str.Length(S) DO
        ch:=S[i-1];
        LowerCase(ch);
        IF Belongs(vowels,ch)= FALSE THEN Str.Append(R,ch);END;
    END;
    Str.Copy(S,R);
END novowels;

PROCEDURE splitstr(VAR S,R:pathtype;DEBUG:BOOLEAN);
VAR
    i:CARDINAL;
    ch:CHAR;
BEGIN
    Str.Copy(R,"");
    i:=Str.Length(S);
    LOOP
        IF i < 1 THEN EXIT; END;
        ch:=S[i-1];
        IF Belongs(digits,ch)=FALSE THEN EXIT; END;
        Str.Prepend(R,ch);
        Str.Delete(S,i-1,1);
        DEC(i);
    END;
END splitstr;

(* [mini..maxi] *)

PROCEDURE chknamelen (mini,maxi:CARDINAL;S:pathtype):BOOLEAN;
VAR
    pb,len:CARDINAL;
BEGIN
    pb:=0;
    len:=Str.Length(S);
    IF len > maxi THEN INC(pb);END;
    IF len < mini THEN INC(pb);END;
    RETURN (pb=0);
END chknamelen;

PROCEDURE doShorten (VAR n,e:pathtype;
                    DEBUG, chkvalidchar:BOOLEAN;shortenmode:CARDINAL ):BOOLEAN;
VAR
    S,R,orgn,orge:pathtype;
    pass,i,mini,maxi,pb:CARDINAL;
    ch:CHAR;
    machete,DOSF8,DOSE3,DOSF8E3,DOSCHARSF8,DOSCHARSE3,DOSCHARS,allowDOSspace : BOOLEAN;
BEGIN
    pb:=0;

    Str.Copy(orgn,n);
    Str.Copy(orge,e);

    machete:=bitarmed(shortenmode,shortenmachete);

    FOR pass:=1 TO 2 DO
        CASE pass OF
        | 1: Str.Copy(S,n); mini:=MINF8; maxi:=MAXF8; (* 1..8 *)
        | 2: Str.Copy(S,e); mini:=MINE3; maxi:=MAXE3; (* 0..3 *)
        END;

        CASE pass OF
        | 1 : (* n *)
            IF bitarmed(shortenmode,shortenvowel) THEN novowels(S, DEBUG);END;

            IF bitarmed(shortenmode,shortendash) THEN
                i:=Str.CharPos(S,dash);
                IF i # MAX(CARDINAL) THEN                        (* Ab-c *)
                    IF bitarmed(shortenmode,shorteninitial) THEN
                        Str.Delete(S,1,i);                       (* Ac *)
                    ELSE
                        Str.Delete(S,0,i+1);                     (* c *)
                    END;
                END;
            END;

            IF bitarmed(shortenmode,shortentruncate) THEN
                adjustlen(S, DEBUG,machete,mini,maxi);
            ELSE
                splitstr(S,R,DEBUG); (* change S to trailing digits go to R *)

                adjustlen(S, DEBUG,machete,mini,maxi);
                adjustlen(R, DEBUG, FALSE ,mini,maxi); (* don't chop digits *)

                (* truncate S until S+R fits DOS [0..maxi] *)
                i:=Str.Length(S);
                LOOP
                    IF (i+Str.Length(R)) <= maxi THEN EXIT;END;
                    Str.Delete(S,i-1,1);
                    DEC(i);
                END;
                Str.Append(S,R);
            END;
        | 2 : (* e *)
            adjustlen(S, DEBUG,FALSE,mini,maxi); (* fit S to DOS [mini..maxi] *)
        END;


        R:="";
        FOR i:=1 TO Str.Length(S) DO
            ch:=S[i-1];
            UpperCase(ch);
            IF verifyString (ch,DOScharset) THEN Str.Append(R,S[i-1]);END;
        END;

        (* yes, R for both ! *)
        CASE pass OF
        | 1 : Str.Copy(n,R);
        | 2 : Str.Copy(e,R);
        END;
    END;

    (* check if all this was useless because filename was already DOS *)
    DOSF8:=chknamelen(MINF8,MAXF8,orgn);
    DOSE3:=chknamelen(MINE3,MAXE3,orge);

    allowDOSspace := NOT(chkvalidchar);
    DOSCHARSF8:=chkChars(FALSE,FALSE, allowDOSspace,orgn);
    DOSCHARSE3:=chkChars(FALSE,TRUE , allowDOSspace,orge); (* dot allowed in extension *)

    DOSF8E3 := ( DOSF8 AND DOSE3 );
    DOSCHARS:= ( DOSCHARSF8 AND DOSCHARSE3);
    IF NOT(machete) THEN
        IF ( DOSF8E3 AND DOSCHARS ) THEN INC(pb);END;
    END;
    IF ( same(orgn,n) AND same(orge,e) ) THEN INC(pb);END;
    RETURN (pb=0);
END doShorten;

(* //FIXME only way in program to set machete on *)

PROCEDURE doGuillotine (VAR n,e:pathtype;
                       DEBUG,useLFN,overwrite,chkvalidchar:BOOLEAN;shortenmode:CARDINAL;
                       base:pathtype):BOOLEAN;
CONST
    firsttry = 1;
    maxtries = 256; (* was MAXF8 *)
VAR
    rc,ok,DOSF8,DOSE3,DOSF8E3:BOOLEAN;
    newname:pathtype;
    i:CARDINAL;
BEGIN
    rc:=FALSE;

    DOSF8:=chknamelen(MINF8,MAXF8,n);
    DOSE3:=chknamelen(MINE3,MAXE3,e);
    DOSF8E3 := ( DOSF8 AND DOSE3 );
    IF NOT(DOSF8E3) THEN

        shortenmode := (shortenmode OR shortenmachete); (* //FIXME "Machete kills !" ;-) *)

        i:=firsttry;
        LOOP
            ok:=doShorten(n,e, DEBUG,chkvalidchar,shortenmode);
            IF ok THEN
                Lib.MakeAllPath(newname,"","",n,e);  (* doFix() *)
                Str.Prepend(newname,base);           (* main() *)
                IF fileExists(useLFN,newname) THEN
                    IF overwrite THEN rc:=TRUE;EXIT;END;
                ELSE
                    rc:=TRUE; (* target filename is available *)
                    EXIT;
                END;
            END;
            INC(i);
            IF i > maxtries THEN EXIT; END;
        END;
    END;
    RETURN rc;
END doGuillotine;

(* ------------------------------------------------------------ *)

(* ripped from DT v1.1d *)

PROCEDURE using (n : CARDINAL; digits : CARDINAL; pad : CHAR) : str16;
VAR
    ok   : BOOLEAN;
    v    : LONGCARD;
    len  : CARDINAL;
    S    : str16;     (* was str80 *)
BEGIN
    v := LONGCARD(n);
    Str.CardToStr(v,S,10,ok);
    IF pad="" THEN RETURN S; END;
    len := Str.Length(S);
    LOOP
        IF Str.Length(S) >= digits THEN EXIT; END;
        Str.Prepend(S,pad);
    END;
    RETURN S;
END using;

TYPE
    datetype = RECORD
        day   : CARDINAL;
        month : CARDINAL;
        year  : CARDINAL;
        dayOfWeek  : Lib.DayType;
        validDayOfWeek : BOOLEAN; (* cleaner but useless *)
    END;
    timetype = RECORD
        hours   : CARDINAL;
        minutes : CARDINAL;
        seconds : CARDINAL;
    END;


TYPE
    dttype = RECORD
        CASE : BOOLEAN OF
        | TRUE  :
            t : CARDINAL; (* hms is low  *)
            d : CARDINAL; (* ymd is high *)
        | FALSE :
            dt : LONGCARD;
        END;
    END;

(*
Year stored relative to 1980 (ex. 1988 stores as 8)
    year      month    day   

 F E D C B A 9 8 7 6 5 4 3 2 1 0   <-- Bit Number

Seconds are 0 to 29 -- DOS stores nearest even / 2
  hours    minutes   seconds 

 F E D C B A 9 8 7 6 5 4 3 2 1 0   <-- Bit Number
*)

CONST
    DOSbaseyear  = 1980;
CONST
    yyMask=BITSET{9..15};    yyShft=9;
    moMask=BITSET{5..8};     moShft=5;
    ddMask=BITSET{0..4};     ddShft=0;
CONST
    hhMask=BITSET{11..15};   hhShft=11;
    miMask=BITSET{5..10};    miShft=5;
    ssMask=BITSET{0..4};     ssShft=0;

PROCEDURE unpackDate (ymd:CARDINAL;VAR y,m,d:CARDINAL);
BEGIN
    y :=  CARDINAL(BITSET(ymd) * yyMask) >> yyShft ;
    m :=  CARDINAL(BITSET(ymd) * moMask) >> moShft ;
    d :=  CARDINAL(BITSET(ymd) * ddMask) >> ddShft ;
    INC(y,DOSbaseyear);
END unpackDate;

PROCEDURE unpackTime (hms:CARDINAL;VAR h,m,s:CARDINAL);
BEGIN
    h :=  CARDINAL(BITSET(hms) * hhMask) >> hhShft ;
    m :=  CARDINAL(BITSET(hms) * miMask) >> miShft ;
    s :=  CARDINAL(BITSET(hms) * ssMask) >> ssShft ;
    s := s << 1; (* yes, yes, "* 2" works too... *)
END unpackTime;

PROCEDURE getfilestamp (VAR dmy:datetype; VAR hms:timetype;
                       useLFN:BOOLEAN; refpath:pathtype);
VAR
    hnd:FIO.File;
    rc:BOOLEAN;     (* should be trapped but eh, we don't care *)
    stamp:dttype;
BEGIN
    hnd:=fileOpenRead(useLFN,refpath);
    IF hnd = MAX(CARDINAL) THEN (* not trapped by fileGetFileStamp() *)
        rc:=FALSE;
    ELSE
        stamp.dt:=FIO.GetFileDate(hnd);
        rc:=TRUE;
    END;
    FIO.Close(hnd);

    unpackDate(stamp.d , dmy.year  ,dmy.month  ,dmy.day );
    unpackTime(stamp.t , hms.hours ,hms.minutes,hms.seconds );
END getfilestamp;

PROCEDURE doProcStamp (VAR n:pathtype;
                       opPrefix,useLFN:BOOLEAN;oldname:pathtype);
VAR
    stamp:LONGCARD;
    dmy:datetype;
    hms:timetype;
    Z:str16;  (* "yyyymmdd_hhmmss_" *)
BEGIN
    getfilestamp(dmy,hms,useLFN,oldname);

    Z:="~~~_~~~";
    Str.Subst(Z,"~",using(dmy.year   ,4,"0"));
    Str.Subst(Z,"~",using(dmy.month  ,2,"0"));
    Str.Subst(Z,"~",using(dmy.day    ,2,"0"));
    Str.Subst(Z,"~",using(hms.hours  ,2,"0"));
    Str.Subst(Z,"~",using(hms.minutes,2,"0"));
    Str.Subst(Z,"~",using(hms.seconds,2,"0"));

    IF opPrefix THEN
        Str.Prepend(n,"_"); Str.Prepend(n,Z);
    ELSE
        Str.Append(n,"_");  Str.Append(n,Z);
    END;
END doProcStamp;

(* ------------------------------------------------------------ *)

PROCEDURE doFix (VAR messages:CARDINAL; VAR newname:pathtype;
                 oldname,base:pathtype;
                 useLFN,procext,safetybatch,prefixrenum,
                 overwrite,finduniquetarget,chkvalidchar,DEBUG : BOOLEAN;
                 op:cmdtype;colpos,shortenmode:CARDINAL;
                 str,str2:pathtype):BOOLEAN;
VAR
    u,d,n,e,S1,S2,newn,tmpn:pathtype;
    p,len,count:CARDINAL;
    wecandoit:BOOLEAN;
BEGIN
    IF DEBUG THEN
        WrStr("; OLD : ");WrHexStr(TRUE ,oldname);WrLn;
        WrStr("; OLD : ");WrHexStr(FALSE,oldname);WrLn;
    END;
    IF DEBUG THEN WrStr("; old : ");WrQ(oldname);WrLn; END;
    qdSplitAllPath(oldname,u,d,n,e);

    wecandoit:=TRUE; (* let's be optimistic *)

    CASE op OF
    | prefix,prefixautonum :
        Str.Prepend(n,str);
    | suffix,suffixautonum :
        Str.Append(n,str);
    | prefixkill:
        count:=Str.Length(str);
        Str.Slice(S1,n,0,count);
        Str.Copy(S2,str);
        UpperCase(S1);
        UpperCase(S2);
        IF same(S1,S2) THEN
            Str.Delete(n,0,count);
        ELSE
            wecandoit:=FALSE;
        END;
    | suffixkill:
        count:=Str.Length(str);
        len  :=Str.Length(n);
        IF len < count THEN
            wecandoit:=FALSE;
        ELSE
            Str.Slice(S1,n,len-count,count);
            Str.Copy(S2,str);
            UpperCase(S1);
            UpperCase(S2);
            wecandoit:= same(S1,S2);
            IF wecandoit THEN
                Str.Delete(n,len-count,count);
            END;
        END;
    | remove :
        wecandoit:=doClean(n,e, procext,str);
    | remap :
        wecandoit:=doRemap(n,e, procext,str,str2);
    | infix:
        Str.Insert(n,str,colpos); (* don't bother to check *)
    | infixkill:
        count:=Str.Length(str);
        Str.Slice(S1,n,colpos,count);
        Str.Copy(S2,str);
        UpperCase(S1);
        UpperCase(S2);
        wecandoit:=same(S1,S2);
        IF wecandoit THEN
            Str.Delete(n,colpos,count);
        END;
    | renum:
        IF prefixrenum THEN
            Str.Prepend(n,str);
        ELSE
            Str.Copy(n,str); (* str is renumbered name *)
        END;
    | strlowercase, struppercase,
      strlowercasenoaccents, struppercasenoaccents,
      strlowercasefrench, struppercasefrench,
      strasciionly:
        wecandoit:=doCasify( n,e, op,procext);
    | trimname:
        wecandoit:=doTrimName(n,e, procext);
    | substringkill:
        wecandoit:=doSubst(n,e, procext,str,"");
    | substringreplace:
        wecandoit:=doSubst(n,e, procext,str,str2);
    | shorten:
        IF finduniquetarget THEN
            wecandoit:= doGuillotine (n,e,
                                     DEBUG,useLFN,overwrite,chkvalidchar,
                                     shortenmode,base);
        ELSE
            wecandoit:=doShorten(n,e, DEBUG,chkvalidchar,shortenmode);
        END;
    | prefixstamp:
        doProcStamp(n,  TRUE, useLFN,oldname); (* useLFN is TRUE here *)
    | suffixstamp:
        doProcStamp(n,  FALSE,useLFN,oldname); (* useLFN is TRUE here *)
    | lenprefixkill:
        count:=Str.Length(str);
        Str.Delete(n,0,count);
    | lensuffixkill:
        count:=Str.Length(str);
        len  :=Str.Length(n);
        IF len < count THEN
            wecandoit:=FALSE;
        ELSE
            Str.Delete(n,len-count,count);
        END;
    | leninfixkill:
        count:=Str.Length(str);
        Str.Delete(n,colpos,count);
    END;
    IF wecandoit THEN
        wecandoit:=chknewlen(messages,procext,useLFN,DEBUG, oldname,n,e);
    END;

    IF same(e,"") THEN Str.Copy(e,dot);END; (* //V12o fix *)

    Lib.MakeAllPath(newname,"","",n,e);

    IF safetybatch THEN
        Str.Copy(S1,newname);
        UpperCase(S1);
        IF same(S1,UNDOBATCH) THEN wecandoit:=FALSE;END;
    END;
    IF DEBUG THEN WrStr("; new : ");WrQ(newname);WrLn; END;
    RETURN wecandoit;
END doFix;

PROCEDURE buildUndo (VAR R:ARRAY OF CHAR;
                    useLFN:BOOLEAN;base,oldname,newname:pathtype );
CONST
    cliren = "REN "+placeholder+" "+placeholder;
VAR
    S:str1024; (* oversized *)
    u,d,n,e,id:pathtype;
BEGIN
    Str.Copy(S,cliren);

    qdSplitAllPath(newname,u,d,n,e);
    Lib.MakeAllPath(id,"","",n,e);
    Str.Prepend(id,base);
    IF useLFN THEN Str.Subst(S,placeholder,dquote+placeholder+dquote);END;
    Str.Subst(S,placeholder,id);

    qdSplitAllPath(oldname,u,d,n,e);
    Lib.MakeAllPath(id,"","",n,e);
    IF useLFN THEN Str.Subst(S,placeholder,dquote+placeholder+dquote);END;
    Str.Subst(S,placeholder,id);
    Str.Copy(R,S);
END buildUndo;

PROCEDURE isCmdCasify (op:cmdtype):BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    CASE op OF
    | strlowercase,struppercase,
      strlowercasenoaccents,struppercasenoaccents,
      strlowercasefrench,struppercasefrench,
      strasciionly:
        rc:=TRUE;
    ELSE
        rc:=FALSE;
    END;
    RETURN rc;
END isCmdCasify;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    msgWaitSort = "Sorting data, please wait...";
CONST
    firstparm = 1;
    maxparm   = 3;
VAR
    parm : ARRAY [firstparm..maxparm] OF pathtype;
VAR
    testmode,useLFN,overwrite,kill:BOOLEAN;
    processAll,procext,safetybatch,chkvalidchar:BOOLEAN;
    userenumprefix,prefixrenum,finduniquetarget,allowRenToSame:BOOLEAN;
    shortenmode:CARDINAL;
    cmd:cmdtype;
    sort:sorttype;
VAR
    parmcount,i,opt:CARDINAL;
    S,R,spec,base,str,mapstr,oldname,newname, oldstr,newstr:pathtype;
    lastparm,lastEntry,messages,colpos,fieldwidth,v:CARDINAL;
    DEBUG,allowed,existed,wecandoit:BOOLEAN;
    lc,currnum,rendelta:LONGCARD;
    hout:FIO.File;
    hugestr:str1024; (* oversized for LFNs *)
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    Str.Copy(mapstr,"");   (* default safety *)
    colpos:=MAX(CARDINAL); (* suffix *)

    useLFN            := TRUE;
    cmd               := nada;
    sort              := nosort;
    fieldwidth        := DEFFIELD;
    rendelta          := DEFDELTA;
    currnum           := 1; (* default for "-s *" and "-p *" syntaxes *)
    testmode          := TRUE;
    safetybatch       := FALSE;
    kill              := FALSE;
    overwrite         := FALSE;
    processAll        := FALSE;
    procext           := FALSE;
    userenumprefix    := TRUE;
    prefixrenum       :=FALSE;
    chkvalidchar      := TRUE;
    shortenmode       := shortennormal;
    finduniquetarget  := FALSE;
    allowRenToSame    := FALSE;
    DEBUG             := FALSE;
    lastparm          := firstparm-1;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOptionForReal(R) THEN (* only handle "-" alone *)
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "P"+delim+"PREFIX"+delim+
                                   "S"+delim+"SUFFIX"+delim+
                                   "D"+delim+"DELETE"+delim+
                                   "O"+delim+"OVERWRITE"+delim+
                                   "R"+delim+"RENAME"+delim+
                                   "L"+delim+"LFN"+delim+
                                   "PP"+delim+
                                   "SS"+delim+
                                   "A"+delim+"ALL"+delim+
                                   "E"+delim+"EXTENSION"+delim+
                                   "I:"+delim+"INFIX:"+delim+
                                   "II:"+delim+
                                   "N"+delim+"RENUM"+delim+
                                   "F:"+delim+"FIELD:"+delim+
                                   "SN"+delim+
                                   "SE"+delim+
                                   "RR"+delim+
                                   "K:"+delim+
                                   "DP"+delim+
                                   "DS"+delim+
                                   "Z"+delim+"ZAP"+delim+
                                   "C"+delim+"CHANGE"+delim+"REPLACE"+delim+
                                   "LC"+delim+"LOWERCASE"+delim+"LOWS"+delim+
                                   "UC"+delim+"UPPERCASE"+delim+"CAPS"+delim+
                                   "F"+delim+"FILTER"+delim+
                                   "M"+delim+"REMAP"+delim+
                                   "LCC"+delim+
                                   "UCC"+delim+
                                   "SZ"+delim+
                                   "NN"+delim+
                                   "X"+delim+"PREFIXRENUM"+delim+
                                   "SD"+delim+
                                   "V"+delim+"VALIDITY"+delim+
                                   "LCF"+delim+"LCFR"+delim+
                                   "UCF"+delim+"UCFR"+delim+
                                   "AZ"+delim+"ASCIIONLY"+delim+
                                   "??"+delim+
                                   "DEBUG"+delim+

                                   "T1"+delim+"T"+delim+"TRUNCATE"+delim+
                                   "T2"+delim+"TT"+delim+
                                   "T3"+delim+"TTT"+delim+

                                   "T5"+delim+"TTTTT"+delim+
                                   "T6"+delim+"TTTTTT"+delim+
                                   "T7"+delim+"TTTTTTT"+delim+

                                   "B"+delim+"PREFIXSTAMP"+delim+
                                   "J"+delim+"SUFFIXSTAMP"+delim+

                                   "T4"+delim+"TTTT"+delim+
                                   "T8"+delim+"TTTTTTTT"+delim+
                                   "Y"+delim+"FINDUNIQUETARGET"+delim+"U"+delim+"UNIQUE"+delim+
                                   "PPP"+delim+
                                   "SSS"+delim+
                                   "III:"+delim+

                                   "SN-"+delim+
                                   "SE-"+delim+
                                   "SZ-"+delim+
                                   "SD-"+delim+
                                   "TRIM"+delim+"BLANKS"+delim+
                                   "I"+delim+"IDEM"+delim+"RENTOSAME"
                               );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5   : IF chkcmd(cmd,prefix)=FALSE THEN abort(errCmdConflict,"");END;
            | 6,7   : IF chkcmd(cmd,suffix)=FALSE THEN abort(errCmdConflict,"");END;
            | 8,9   : kill:=TRUE;
            | 10,11 : overwrite:=TRUE;
            | 12,13 : testmode:=FALSE;
            | 14,15 : useLFN:=FALSE;
            | 16    : IF chkcmd(cmd,prefixkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 17    : IF chkcmd(cmd,suffixkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 18,19 : processAll:=TRUE;
            | 20,21 : procext:=TRUE;
            | 22,23 : IF chkcmd(cmd,infix)=FALSE THEN abort(errCmdConflict,"");END;
                      IF GetLongCard(R,lc)=FALSE THEN abort(errBadVal,S);END;
                      IF chkRange(colpos,lc,0,MAX(CARDINAL))=FALSE THEN abort(errRange,S);END;
            | 24    : IF chkcmd(cmd,infixkill)=FALSE THEN abort(errCmdConflict,"");END;
                      IF GetLongCard(R,lc)=FALSE THEN abort(errBadVal,S);END;
                      IF chkRange(colpos,lc,0,MAX(CARDINAL))=FALSE THEN abort(errRange,S);END;
            | 25,26 : IF chkcmd(cmd,renum)=FALSE THEN abort(errCmdConflict,"");END;
            | 27,28 : IF GetLongCard(R,lc)=FALSE THEN abort(errBadVal,S);END;
                      IF chkRange(fieldwidth,lc,MINFIELD,MAXFIELD)=FALSE THEN abort(errRange,S);END;
            | 29    : IF chksort(sort,byname)=FALSE THEN abort(errSortConflict,"");END;
            | 30    : IF chksort(sort,byextension)=FALSE THEN abort(errSortConflict,"");END;
            | 31    : testmode:=FALSE; safetybatch:=TRUE;
            | 32    : IF GetLongCard(R,rendelta)=FALSE THEN abort(errBadVal,S);END;
                      IF chkRange(v,rendelta,MINDELTA,MAXDELTA)=FALSE THEN abort(errRange,S);END;
            | 33    : IF chkcmd(cmd,prefixkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 34    : IF chkcmd(cmd,suffixkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 35,36 : IF chkcmd(cmd,substringkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 37,38,39: IF chkcmd(cmd,substringreplace)=FALSE THEN abort(errCmdConflict,"");END;
            | 40,41,42: IF chkcmd(cmd,strlowercase)=FALSE THEN abort(errCmdConflict,"");END;
            | 43,44,45: IF chkcmd(cmd,struppercase)=FALSE THEN abort(errCmdConflict,"");END;
            | 46,47 : IF chkcmd(cmd,remove)=FALSE THEN abort(errCmdConflict,"");END;
            | 48,49 : IF chkcmd(cmd,remap)=FALSE THEN abort(errCmdConflict,"");END;
            | 50    : IF chkcmd(cmd,strlowercasenoaccents)=FALSE THEN abort(errCmdConflict,"");END;
            | 51    : IF chkcmd(cmd,struppercasenoaccents)=FALSE THEN abort(errCmdConflict,"");END;
            | 52    : IF chksort(sort,bysize)=FALSE THEN abort(errSortConflict,"");END;
            | 53    : IF chkcmd(cmd,renum)=FALSE THEN abort(errCmdConflict,"");END;
                      userenumprefix := FALSE;
            | 54,55 : prefixrenum := TRUE;
            | 56    : IF chksort(sort,bystamp)=FALSE THEN abort(errSortConflict,"");END;
            | 57,58 : chkvalidchar := FALSE;
            | 59,60 : IF chkcmd(cmd,strlowercasefrench)=FALSE THEN abort(errCmdConflict,"");END;
            | 61,62 : IF chkcmd(cmd,struppercasefrench)=FALSE THEN abort(errCmdConflict,"");END;
            | 63,64 : IF chkcmd(cmd,strasciionly)=FALSE THEN abort(errCmdConflict,"");END;
            | 65    : abort(errMoreHelp,"");
            | 66    : DEBUG := TRUE;

            | 67,68,69 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=shortennormal;
            | 70,71 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=shortendash;
            | 72,73 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=(shortendash OR shorteninitial);

            | 74,75 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=(shortennormal OR shortenvowel);
            | 76,77 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=(shortendash OR shortenvowel);
            | 78,79 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=(shortendash OR shorteninitial OR shortenvowel);

            | 80,81 : IF chkcmd(cmd,prefixstamp)=FALSE THEN abort(errCmdConflict,"");END;
            | 82,83 : IF chkcmd(cmd,suffixstamp)=FALSE THEN abort(errCmdConflict,"");END;

            | 84,85 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=shortentruncate;
            | 86,87 : IF chkcmd(cmd,shorten)=FALSE THEN abort(errCmdConflict,"");END;
                      shortenmode:=(shortentruncate OR shortenvowel);
            | 88,89,90,91 : finduniquetarget:=TRUE;
            | 92    : IF chkcmd(cmd,lenprefixkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 93    : IF chkcmd(cmd,lensuffixkill)=FALSE THEN abort(errCmdConflict,"");END;
            | 94    : IF chkcmd(cmd,leninfixkill)=FALSE THEN abort(errCmdConflict,"");END;
                      IF GetLongCard(R,lc)=FALSE THEN abort(errBadVal,S);END;
                      IF chkRange(colpos,lc,0,MAX(CARDINAL))=FALSE THEN abort(errRange,S);END;

            | 95    : IF chksort(sort,bynameINV)=FALSE THEN abort(errSortConflict,"");END;
            | 96    : IF chksort(sort,byextensionINV)=FALSE THEN abort(errSortConflict,"");END;
            | 97    : IF chksort(sort,bysizeINV)=FALSE THEN abort(errSortConflict,"");END;
            | 98    : IF chksort(sort,bystampINV)=FALSE THEN abort(errSortConflict,"");END;
            | 99,100: IF chkcmd(cmd,trimname)=FALSE THEN abort(errCmdConflict,"");END;
            |101,102,103: allowRenToSame:=TRUE;
            ELSE
                abort(errUnknownOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S); END;
	        Str.Copy(parm[lastparm],S); (* keep original case *)
        END;
    END;

    useLFN:=( useLFN AND w9XsupportLFN() );

    IF kill THEN
        CASE cmd OF
        | prefix: cmd:=prefixkill;
        | suffix: cmd:=suffixkill;
        | infix : cmd:=infixkill;
        END;
    END;

    (* we don't try too hard to protect user from himself ! ;-) *)
    (* //12M allow procext for substringreplace , substringkill , remove *)

    CASE lastparm OF
    | 0 : abort(errSyntax,"");
    | 1 : CASE cmd OF
          | strlowercase,struppercase,
            strlowercasenoaccents,struppercasenoaccents,
            strlowercasefrench,struppercasefrench,
            strasciionly,shorten,prefixstamp,suffixstamp,trimname :
              IF NOT(useLFN) THEN abort(errLFNonly,"");END;
              IF kill THEN abort(errNonsense,"-d");END;
          ELSE
              abort(errSyntax,"");
          END;
    | 2 : CASE cmd OF
          | prefix,suffix,infix,
            prefixkill,suffixkill,infixkill,
            lenprefixkill,lensuffixkill,leninfixkill:
              IF procext THEN abort(errNonsense,"-e");END;
          | substringkill: ;
          | nada,remove:
              cmd:=remove;
          ELSE
              abort(errSyntax,""); (* strlowercase struppercase remap renum substringreplace *)
          END;
    | 3 : CASE cmd OF
          | renum:
              IF kill THEN abort(errNonsense,"-d");END;
              IF procext THEN abort(errNonsense,"-e");END;
          | nada, remap:
              IF kill THEN abort(errNonsense,"-d");END;
              cmd:=remap;
          | substringreplace:
              IF kill THEN abort(errNonsense,"-d");END;
          ELSE
              abort(errSyntax,"");
          END;
    END;

    CASE cmd OF
    | prefix,prefixkill,lenprefixkill:
        Str.Copy(str    ,parm[firstparm]);
        Str.Copy(spec   ,parm[firstparm+1]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF same(str,autonummarker) THEN
            CASE cmd OF
            | prefixkill    : abort(errNonsense,"-pp");
            | lenprefixkill : abort(errNonsense,"-ppp"); (* could be allowed *)
            END;
            cmd:=prefixautonum;
        ELSE
            IF chkValid(useLFN,FALSE,chkvalidchar,str,base)=FALSE THEN abort(errBadPrefix,str);END;
        END;
    | suffix,suffixkill,lensuffixkill:
        Str.Copy(str    ,parm[firstparm]);
        Str.Copy(spec   ,parm[firstparm+1]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF same(str,autonummarker) THEN
            CASE cmd OF
            | suffixkill    : abort(errNonsense,"-ss");
            | lensuffixkill : abort(errNonsense,"-sss"); (* could be allowed *)
            END;
            cmd:=suffixautonum;
        ELSE
            IF chkValid(useLFN,FALSE,chkvalidchar,str,base)=FALSE THEN abort(errBadSuffix,str);END;
        END;
    | remove:
        Str.Copy(str    ,parm[firstparm]);
        Str.Copy(spec   ,parm[firstparm+1]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF chkChars(useLFN,FALSE,TRUE,str)=FALSE THEN abort(errBadChar,str);END;
    | remap:
        Str.Copy(str    ,parm[firstparm]);
        Str.Copy(mapstr ,parm[firstparm+1]);
        Str.Copy(spec   ,parm[firstparm+2]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF metaproc(mapstr)=FALSE THEN abort(errBadMeta,mapstr);END;
        IF Str.Length(str) # Str.Length(mapstr) THEN abort(errBadRemap,"");END;
        IF chkChars(useLFN,FALSE,FALSE,str)=FALSE THEN abort(errBadChar,str);END;
        IF chkChars(useLFN,FALSE,FALSE,mapstr)=FALSE THEN abort(errBadChar,mapstr);END;
        Str.Copy(S,str);
        Str.Copy(R,mapstr);
        UpperCase(S);
        UpperCase(R);
        FOR i:=1 TO Str.Length(S) DO
            IF str[i-1] = mapstr[i-1] THEN abort(errRemapNonsense,"");END;
        END;
    | infix,infixkill,leninfixkill:
        Str.Copy(str    ,parm[firstparm]);
        Str.Copy(spec   ,parm[firstparm+1]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF chkValid(useLFN,FALSE,chkvalidchar,str,base)=FALSE THEN abort(errBadInfix,str);END; (* could be allowed FOR leninfixkill *)
    | renum:
        Str.Copy(mapstr ,parm[firstparm]);   (* prefix *)
        Str.Copy(S      ,parm[firstparm+1]); (* first number *)
        Str.Copy(spec   ,parm[firstparm+2]);
        IF userenumprefix THEN
            IF same(mapstr,ignoredrenumprefix) THEN
                userenumprefix:=FALSE;
            ELSE
                IF metaproc(mapstr)=FALSE THEN abort(errBadMeta,mapstr);END;
                IF chkChars(useLFN,FALSE,FALSE,mapstr)=FALSE THEN abort(errBadChar,mapstr);END;
            END;
        END;
        IF NOT(userenumprefix) THEN mapstr:="";END;
        Str.Concat(R,"-FIX:",S); (* fake option *)
        IF GetLongCard(R,currnum)=FALSE THEN abort(errBadVal,S);END;
    | substringkill:
        Str.Copy(str    ,parm[firstparm]);
        Str.Copy(spec   ,parm[firstparm+1]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF chkValid(useLFN,FALSE,chkvalidchar,str,base)=FALSE THEN abort(errBadSubstring,str);END;
    | strlowercase,struppercase,
      strlowercasenoaccents,struppercasenoaccents,
      strlowercasefrench,struppercasefrench,
      strasciionly, prefixstamp, suffixstamp, shorten, trimname:
        Str.Copy(spec   ,parm[firstparm]);
    | substringreplace :
        Str.Copy(str    ,parm[firstparm]);    (* oldsubstring *)
        Str.Copy(mapstr ,parm[firstparm+1]);  (* newsubstring *)
        Str.Copy(spec   ,parm[firstparm+2]);
        IF metaproc(str)=FALSE THEN abort(errBadMeta,str);END;
        IF metaproc(mapstr)=FALSE THEN abort(errBadMeta,mapstr);END;
        IF chkValid(useLFN,FALSE,chkvalidchar,str,base)=FALSE THEN abort(errBadSubstring,str);END;
        IF chkValid(useLFN,FALSE,chkvalidchar,mapstr,base)=FALSE THEN abort(errBadSubstring,mapstr);END;
    END;

    IF same(spec,dot) THEN spec:=stardotstar;END;
    IF chkValid(useLFN,TRUE ,chkvalidchar,spec,base)=FALSE THEN abort(errBadSpec,spec);END;
dbgstr(DEBUG,"spec",spec);
    IF chkJoker(spec)=FALSE THEN
        IF fileIsDirectorySpec(useLFN,spec) THEN abort(errDirectory,spec);END;
    END;

    initList(anchorSpec,lastEntry);
    i:= buildMatchList(anchorSpec,useLFN,processAll,spec);
    IF i=MAX(CARDINAL) THEN abort(errStorage,""); END;
    INC(lastEntry,i);
    IF lastEntry < firstEntry THEN abort(errNoMatch,spec);END;
    initIndexes(anchorSpec);

    IF safetybatch THEN
        FOR i:=firstEntry TO lastEntry DO
            getEntry(S,i,anchorSpec);
            UpperCase(S);
            IF same(S,UNDOBATCH) THEN abort(errWontProcess,"");END;
        END;
        hout:=FIO.Create(UNDOBATCH);
        FIO.AssignBuffer(hout,ioBufferOut);
        FIO.WrStr(hout,"@ECHO ON");FIO.WrLn(hout);
        FIO.WrStr(hout,"PAUSE Hit Ctrl-C or Ctrl-Break to abort !");FIO.WrLn(hout);
    END;

    CASE cmd OF
    | renum, suffixautonum,prefixautonum:
        Str.Copy(S,padnum(currnum,fieldwidth)); (* currnum defaults to 1 at start *)
        lc:=currnum+rendelta*LONGCARD(lastEntry-1);
        Str.Copy(R,padnum(lc,fieldwidth));
        IF Str.Length(S) # Str.Length(R) THEN abort(errFieldWidth,"");END;

        IF sort # nosort THEN video(msgWaitSort,TRUE);END;
        CASE sort OF
        | byname      : Lib.QSort(lastEntry,dolessname, doswap);
        | byextension : Lib.QSort(lastEntry,dolessext , doswap);
        | bysize      : Lib.QSort(lastEntry,dolesssize, doswap);
        | bystamp     : Lib.QSort(lastEntry,dolessstamp,doswap);
        | bynameINV   : Lib.QSort(lastEntry,dolessnameINV, doswap);
        | byextensionINV: Lib.QSort(lastEntry,dolessextINV , doswap);
        | bysizeINV   : Lib.QSort(lastEntry,dolesssizeINV, doswap);
        | bystampINV  : Lib.QSort(lastEntry,dolessstampINV,doswap);
        ELSE
            ;
        END;
        IF sort # nosort THEN video(msgWaitSort,FALSE);END;
    END;

    (*
    WrStr(Banner);WrLn;
    WrLn;
    *)

    messages:=0;
    FOR i := firstEntry TO lastEntry DO
        getEntry(S,i,anchorSpec);
        Str.Concat(oldname,base,S);
        Str.Copy(newname,S);
        CASE cmd OF
        | renum :
            IF prefixrenum THEN
                Str.Concat(str,padnum(currnum,fieldwidth),mapstr);
            ELSE
                Str.Concat(str,mapstr,padnum(currnum,fieldwidth));
            END;
            INC(currnum,rendelta);
        | suffixautonum:
            Str.Concat(str,autonumsep,padnum(currnum,fieldwidth));
            INC(currnum,rendelta);
        | prefixautonum:
            Str.Concat(str,padnum(currnum,fieldwidth),autonumsep);
            INC(currnum,rendelta);
        END;
        wecandoit := doFix (messages,newname, oldname,base,
                           useLFN,procext,safetybatch,prefixrenum,
                           overwrite,finduniquetarget,chkvalidchar,DEBUG,
                           cmd,colpos,shortenmode,str,mapstr);
        IF wecandoit THEN
            Str.Prepend(newname,base);
            allowed:=TRUE;
            existed:=FALSE;
            IF fileExists(useLFN,newname) THEN
                IF overwrite THEN
                    IF testmode=FALSE THEN
                        (* prior TO //V13D
                        IF isCmdCasify(cmd)=FALSE THEN
                            fileSetRW(useLFN,oldname); (* safety *)
                        ELSE
                            killfile(useLFN,overwrite,oldname); (* erase readonly too *)
                            existed:=TRUE;
                        END;
                        *)
                        IF isCmdCasify(cmd)=FALSE THEN
                            killfile(useLFN,overwrite,oldname); (* erase readonly too *)
                        ELSE
                            fileSetRW(useLFN,oldname); (* safety *)
                        END;
                        existed:=TRUE;
                    END;
                ELSE
                    IF isCmdCasify(cmd)=FALSE THEN (* LFN agrees to rename changing CASE *)
                        IF NOT(allowRenToSame) THEN
                            msg2(useLFN,TRUE,msgAlready,oldname,newname);
                            allowed:=FALSE;
                            INC(messages);
                        END;
                    END;
                END;
            END;
            IF allowed THEN
                IF testmode THEN
                    msg2(useLFN,FALSE,msgWouldBe,oldname,newname);
                    IF existed THEN WrStr(msgFileExists);END;
                    WrLn;
                    INC(messages);
                ELSE
                    IF NOT (same (oldname, newname)) THEN
                        fileRename(useLFN,oldname,newname);
                    END;
                    msg2(useLFN,FALSE,msgIsNow,oldname,newname);
                    IF existed THEN WrStr(msgFileExisted);END;
                    WrLn;
                    INC(messages);

                    IF safetybatch THEN
                        buildUndo(hugestr,useLFN,base,oldname,newname);
                        FIO.WrStr(hout,hugestr);FIO.WrLn(hout);
                    END;
                END;
            END;
        END;
    END;
    IF messages=0 THEN WrStr("::: Nothing to do !");WrLn; END;

    IF safetybatch THEN
        FIO.Close(hout);
        WrLn;WrStr("::: "+UNDOBATCH+" has been created.");WrLn;
    END;

    freeMatchList(anchorSpec);

    abort(errNone,"");
END NewName.


