(* ---------------------------------------------------------------
Title         Q&D Pattern Match
Overview      self-explanatory !
Notes         Brute Force and Ignorance do rule ;-)
Bugs
Wish List     check entries are valid [A..Z]
              tokenize dictionary
              check for Escape while dumping ?

--------------------------------------------------------------- *)

MODULE patMatch;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    progEXEname     = "PATMATCH";
    progTitle       = "Q&D Pattern Match";
    progVersion     = "v1.0g";
    progCopyright   = "by PhG";
    banner          = progTitle+" "+progVersion+" "+progCopyright;
CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParmOverflow = 3;
    errExpected     = 4;
    errNotFound     = 5;
    errNotFile      = 6;
    errJoker        = 7;
    errPat          = 8;
    errNum          = 9;
    errCard         = 10;
    errNonsenseWithCode = 11;
    errLettersOnly  = 12;
CONST
    extFRA          = ".FRA";
    onechar         = "?.-=_";
    onesequence     = "*!$&";
    verbatimsymbols = "-'.";
    defaultDic      = progEXEname+extFRA;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
banner+nl+
nl+
"Syntax : "+progEXEname+" <pattern> [[minimum_]length [maximum_length]] [option]..."+nl+
nl+
"  -v[:$]   keep accents, ignore symbols, ignore case (-v:$ = -r -d:$)"+nl+
'  -e:$     alternate dictionary extension (default is "'+extFRA+'")'+nl+
"  -d:$     alternate dictionary"+nl+
"  -u       display matches in uppercase instead of default lowercase"+nl+
"  -s       display matches statistics"+nl+
"  -p       paging (-s ignored)"+nl+
"  -w|-pp   paging dividing count of rows by 2 (-s ignored)"+nl+
"  -f[f[f]] display matches according to approximate phonetics"+nl+
"  -z[z]    display matches according to vowels and consonants (-zz = packed)"+nl+
"  -x       disable LFN support even if available"+nl+
nl+
"This program scans dictionary entries as to find specified text pattern."+nl+
nl+
'a) Any character of "'+onechar+'" set matches exactly one single character,'+nl+
'   while any character of "'+onesequence+'" set matches any sequence of characters'+nl+
"   (including none at all)."+nl+
"b) Trimmed dictionary entries must use [A..Z] character set only."+nl+
"c) With any of -v or -v:$ options, dictionary must be DOS ASCII ;"+nl+
'   allowed symbols are "'+verbatimsymbols+'" ; allowed characters may be accentuated.'+nl+
"d) Unless specified with a path, dictionary is searched for"+nl+
"   first in current directory then in executable directory."+nl+
'   "'+defaultDic+'" is default dictionary.'+nl+
"e) -f = French phonetics, -ff = tweaked Soundex, -fff = original Soundex,"+nl+
"   -z = vowels and consonants, -zz = packed -z."+nl+
"   -f[f[f]] and -z[z] options require <pattern> not to contain any joker."+nl+
"   Please note these options are and will remain in beta stage,"+nl+
"   for they are very slow and worse, they are useless."+nl+
nl+
"Examples : "+progEXEname+" -p ad*tion 2 12"+nl+
"           "+progEXEname+" -u ad*t?on 14 -d:english.dic"+nl+
"           "+progEXEname+" -f etre"+nl;

VAR
    S : str1024;
BEGIN
    CASE e OF
    | errHelp :        WrStr(helpmsg);
    | errOption :      S:='Illegal "~" option !';
    | errParmOverflow: S:='"~" is one parameter too many !';
    | errExpected:     S:='"~" expected !';
    | errNotFound:     S:='"~" does not exist !';
    | errNotFile:      S:='"~" is not a file !';
    | errJoker:        S:='"~" should not contain any joker !';
    | errPat:          S:='Illegal "~" pattern !';
    | errNum:          S:='Illegal "~" number !';
    | errCard:         S:='Illegal "~" value !';
    | errNonsenseWithCode:S:="-f[f[f]] and -z[z] options require <pattern> only !";
    | errLettersOnly:  S:="<pattern> must contain only letters for -f[f[f]] and -z[z] options !";
    ELSE
        S := "This is illogical, Captain !"; (* thus spoke Spock... *)
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* as usual, buffers must be outside procedure, else program bombs ! *)

CONST
    IObufferSize      = (16 * 512) + FIO.BufferOverhead;
    firstIObufferByte = 1;
    lastIObufferByte  = IObufferSize;
TYPE
    ioBufferType      = ARRAY [firstIObufferByte..lastIObufferByte] OF BYTE; (* for private M2 IO *)
VAR
    ioBufferIn        : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE chkpat (VAR R:ARRAY OF CHAR;relaxed:BOOLEAN;S:ARRAY OF CHAR ):BOOLEAN;
CONST
    allowed = lettersUpp+onechar+onesequence;
VAR
    pb,i:CARDINAL;
BEGIN
    Str.Copy(R, S);
    (* yes, they may have different lengths *)
    FOR i:=2 TO Str.Length(onechar) DO
        ReplaceChar(R,onechar[i-1],onechar[0]);
    END;
    FOR i:=2 TO Str.Length(onesequence) DO
        ReplaceChar(R,onesequence[i-1],onesequence[0]);
    END;
    pb:=0;
    IF relaxed THEN
        LowerCaseFR(R);
    ELSE
        UpperCase(R); (* //V10F before, was done at cli parsing *)
        FOR i:=1 TO Str.Length(R) DO
            IF Str.CharPos(allowed,R[i-1])=MAX(CARDINAL) THEN INC(pb);END;
        END;
    END;
    RETURN (pb=0);
END chkpat;

PROCEDURE fmtlc (v:LONGCARD;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str80;
VAR
    R : str80;
    ok: BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    FOR i:= Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,ch);
        ELSE
            Str.Prepend(R,ch);
        END;
    END;
    IF base=16 THEN Str.Lows(R);END;
    Str.Prepend(R,prefix);
    RETURN R;
END fmtlc;

PROCEDURE fmt (v:CARDINAL;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str80;
BEGIN
    RETURN fmtlc(LONGCARD(v),base,wi,ch,prefix);
END fmt;

PROCEDURE haspath (S:pathtype  ):BOOLEAN;
VAR
    pb:CARDINAL;
BEGIN
    pb:=0;
    IF Str.CharPos(S,":") # MAX(CARDINAL) THEN INC(pb);END;
    IF Str.CharPos(S,"\") # MAX(CARDINAL) THEN INC(pb);END;
    IF Str.CharPos(S,"/") # MAX(CARDINAL) THEN INC(pb);END;
    RETURN (pb # 0);
END haspath;

PROCEDURE dmp (msg,str:ARRAY OF CHAR);
CONST
    wi=20;
    separ = " : ";
VAR
    i:CARDINAL;
BEGIN
    WrStr(msg);
    FOR i:=Str.Length(msg)+1 TO wi DO WrStr(" ");END;
    WrStr(separ);WrStr(str);WrLn;
END dmp;

PROCEDURE dmpbool (tf:BOOLEAN; msg,sTrue,sFalse:ARRAY OF CHAR);
CONST
    wi=20;
    separ = " : ";
VAR
    i:CARDINAL;
BEGIN
    WrStr(msg);
    FOR i:=Str.Length(msg)+1 TO wi DO WrStr(" ");END;
    WrStr(separ);
    IF tf THEN
        WrStr(sTrue);
    ELSE
        WrStr(sFalse);
    END;
    WrLn;
END dmpbool;

CONST
    msgWorking = "Working, please wait... ";
    msgprompt  = "Hit (almost) any key to continue or Escape to abort listing : ";
    msgaborted = "Search aborted by user !";
CONST
    initpagingcounter = 1; (* take msg into account *)

PROCEDURE waitAndSee (VAR rowcount:CARDINAL; lastrow:CARDINAL;msg:BOOLEAN):BOOLEAN;
CONST
    keyEsc = str2(CHR(27));
VAR
    zekey:str2;
BEGIN
    INC(rowcount);
    IF rowcount >= lastrow THEN (* = is enough but who knows what evil lurks in the heart of BIOS *)
        rowcount := initpagingcounter;
        IF msg THEN video(msgWorking,FALSE); END;
        video(msgprompt,TRUE);
        Flushkey;
        zekey:=Waitkey();
        video(msgprompt,FALSE);
        IF msg THEN video(msgWorking,TRUE);END;
        IF same(zekey,keyEsc) THEN RETURN FALSE; END;
    END;
    RETURN TRUE;
END waitAndSee;

(* ------------------------------------------------------------ *)

(* //FIXME to DO later ! *)

CONST
    PLAINDIC = 0;

PROCEDURE isTokenizedFile (S:pathtype):LONGCARD; (* fpos *)
BEGIN
    RETURN PLAINDIC;
END isTokenizedFile;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* assume S is lowercase *)

PROCEDURE CodeCat (packed:BOOLEAN;S:ARRAY OF CHAR;
                  VAR R:ARRAY OF CHAR);
CONST
    isV     = "v"; vowels     = "aeiouy";
    isC     = "c"; consonants = "bcdfghjklmnpqrstvwxz";
    isOther = "_";
VAR
    i:CARDINAL;
    ch:CHAR;
    Z:str2;
BEGIN
    Str.Copy(R,"");
    FOR i:=1 TO Str.Length(S) DO
        ch:=S[i-1];
        IF Str.CharPos(vowels,ch) # MAX(CARDINAL) THEN
            ch:=isV;
        ELSIF Str.CharPos(consonants,ch) # MAX(CARDINAL) THEN
            ch:=isC;
        ELSE
            ch:=isOther;
        END;
        Str.Append(R,ch);
    END;
    IF packed THEN
        FOR i:=1 TO 3 DO
            CASE i OF
            | 1 : ch:=isV;
            | 2 : ch:=isC;
            | 3 : ch:=isOther;
            END;
            Str.Concat(Z,ch,ch);
            LOOP
                IF Str.Pos(R,Z)=MAX(CARDINAL) THEN EXIT;END;
                Str.Subst(R,Z,ch);
            END;
        END;
    END;
END CodeCat;

TYPE
    pagingmode = (pagingnone,pagingfull,paginghalf);
    workmode   = (motif,verbatim,
                 fonetic,soundexalt,soundexorg,category,packedcategory);

PROCEDURE getCode (VAR R:ARRAY OF CHAR;
                  whatwork:workmode;S:ARRAY OF CHAR);
BEGIN
    LowerCase(S); (* required -- see QD_BOX source *)
    CASE whatwork OF
    | fonetic:    CodePhonetic  (S, R);
    | soundexalt: CodeSoundex   (S, R);
    | soundexorg: CodeSoundexOrg(S, R);
    | category:       CodeCat(FALSE,S, R);
    | packedcategory: CodeCat(TRUE, S, R);
    END;
END getCode;

(* completely useless ! *)

PROCEDURE samepat (S1,S2:ARRAY OF CHAR  ):BOOLEAN ;
BEGIN
    Str.Prepend(S1,"*");
    Str.Append(S1,"*");
    Str.Prepend(S2,"*");
    Str.Append(S2,"*");
    RETURN Str.Match(S1,S2);
END samepat;

PROCEDURE showmatches (useLFN,showstats,shoutmode:BOOLEAN;
                      whatwork:workmode; paging:pagingmode;
                      startfpos:LONGCARD;
                      minmatchlen,maxmatchlen:CARDINAL;
                      strpat:ARRAY OF CHAR;dicfile:pathtype  );
CONST
    msg = TRUE;
CONST
    wilen  = 3;
    wilc   = 9;
    minlen = 1;
    maxlen = 128;
CONST
    sb = 040H; (* segBiosData *)
VAR
    biosCurrentRows       [sb:084H]  : BYTE;     (* add 1 *)
VAR
    matchlen  : ARRAY[minlen..maxlen] OF LONGCARD;
    undefined,matchcount: LONGCARD;
    i,len,symlen:CARDINAL;
    tokenized,ok:BOOLEAN;
    rowcount,lastrow:CARDINAL;
    refsoundexcode:str128;
VAR
    S,R,Z:str128;
    hin:FIO.File;
BEGIN
    tokenized := (startfpos # PLAINDIC);
    IF paging # pagingnone THEN
        showstats:=FALSE;
        lastrow:=CARDINAL(biosCurrentRows) +1;
        IF paging = paginghalf THEN
            lastrow:=lastrow DIV 2;
            IF lastrow = 0 THEN INC(lastrow);END; (* probably useless *)
        END;
        rowcount:=initpagingcounter;
    END;

    CASE whatwork OF
    | motif,verbatim: ;
    ELSE
        getCode(refsoundexcode,whatwork,strpat);
    END;

    symlen:=Str.Length(verbatimsymbols);
    matchcount:=0;
    FOR i:=minlen TO maxlen DO matchlen[i]:=0;END;
    undefined:=0;

    IF msg THEN video(msgWorking,TRUE); END;
    FIO.EOF:=FALSE;
    hin:=fileOpenRead(useLFN,dicfile);
    FIO.AssignBuffer(hin,ioBufferIn);
    FIO.Seek(hin,startfpos);
    LOOP
        FIO.RdStr(hin,S);
        len:= Str.Length(S);
        IF len # 0 THEN
            CASE whatwork OF
            | motif:
                ok:=Str.Match(S,strpat);
            | verbatim:
                Str.Copy(Z,S);
                FOR i:=1 TO symlen DO
                    ReplaceChar(Z,verbatimsymbols[i-1],"");
                END;
                ok:=Str.Match(Z,strpat);
            ELSE
                getCode(R,whatwork,S);
                ok:=same(R,refsoundexcode);
                (* ok:=samepat(R,refsoundexcode); *)
            END;
            IF ok THEN
                ok:= ( (len >= minmatchlen) AND (len <= maxmatchlen) );
                IF ok THEN
                    INC(matchcount);
                    IF len <= maxlen THEN
                        INC(matchlen[len]);

                        IF paging # pagingnone THEN
                            IF waitAndSee(rowcount,lastrow,msg) = FALSE THEN
                                IF msg THEN video(msgWorking,FALSE); END;
                                WrStr(msgaborted);WrLn;
                                IF msg THEN video(msgWorking,TRUE); END;
                                EXIT;
                            END;
                        END;
                        IF shoutmode THEN
                            IF whatwork = verbatim THEN
                                UpperCaseFR(S);
                            ELSE
                                UpperCaseAlt(S);
                            END;
                        ELSE
                            IF whatwork = verbatim THEN
                                LowerCaseFR(S);
                            ELSE
                                LowerCaseAlt(S);
                            END;
                        END;
                        IF msg THEN video(msgWorking,FALSE); END;
                        WrStr(S);WrLn;
                        IF msg THEN video(msgWorking,TRUE); END;
                    ELSE
                        INC(undefined);
                    END;
                END;
            END;
        END;
        IF FIO.EOF THEN EXIT;END;
    END;
    fileClose(useLFN,hin);
    IF msg THEN video(msgWorking,FALSE); END;

    IF showstats THEN
        WrLn;
        FOR i:=minlen TO maxlen DO
            IF matchlen[i] # 0 THEN
                S:="Length = ~ -- count = ~";
                Str.Subst(S,"~",fmt(i,10,wilen," ",""));
                Str.Subst(S,"~",fmtlc(matchlen[i],10,wilc," ",""));
                WrStr(S);WrLn;
            END;
        END;
        IF undefined # 0 THEN
                S:="Length = ??? -- count = ~";
                Str.Subst(S,"~",fmtlc(undefined,10,wilc," ",""));
                WrStr(S);WrLn;
        END;
        WrLn;
        S:="Total matches         = ~";
        Str.Subst(S,"~",fmtlc(matchcount,10,wilc," ",""));
        WrStr(S);WrLn;
    END;

END showmatches;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    S,R:pathtype;
    i,opt,parmcount:CARDINAL;
    useLFN,showstats,shoutmode,DEBUG:BOOLEAN;
    paging:pagingmode;
    whatwork:workmode;
    startfpos:LONGCARD;
    dicfile:pathtype;
    state:(waiting,gotpat,gotwordlen,gotmaxwordlen);
    strpat,orgpat:str128;
    N:str16;
    v:LONGCARD;
    minmatchlen,maxmatchlen:CARDINAL;
BEGIN
    WrLn;
    FIO.IOcheck:=FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone; (* very, very important ! *)

    useLFN       := TRUE;
    dicfile      := defaultDic;
    showstats    := FALSE;
    shoutmode    := FALSE;
    whatwork     := motif;
    paging       := pagingnone;
    minmatchlen  := 0;
    maxmatchlen  := MAX(CARDINAL);
    DEBUG        := FALSE;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    state:=waiting;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R,"?"+delim+"H"+delim+"HELP"+delim+
                                  "D:"+delim+"DIC:"+delim+"DICO:"+delim+
                                  "S"+delim+"STATS"+delim+
                                  "E:"+delim+"EXT:"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "U"+delim+"UPPER"+delim+"UPPERCASE"+delim+
                                  "P"+delim+"PAGING"+delim+"PAUSE"+delim+
                                  "W"+delim+"PP"+delim+"PAGINGHALF"+delim+
                                  "F"+delim+
                                  "FF"+delim+
                                  "FFF"+delim+
                                  "V"+delim+"VERBATIM"+delim+
                                  "V:"+delim+"VERBATIM:"+delim+
                                  "Z"+delim+
                                  "ZZ"+delim+
                                  "DEBUG"
                               );
            CASE opt OF
            | 1,2,3 :   abort(errHelp,"");
            | 4,5,6 :   GetString(S,dicfile);
            | 7,8   :   showstats := TRUE;
            | 9,10  :   GetString(S,N);
                        IF Str.Match(N,dot+"*")=FALSE THEN Str.Prepend(N,dot);END;
                        Str.Concat(dicfile,progEXEname,N);
            | 11,12 :   useLFN := FALSE;
            | 13,14,15: shoutmode := TRUE;
            | 16,17,18: paging := pagingfull;
            | 19,20,21: paging := paginghalf;
            | 22    :   whatwork := fonetic;
            | 23    :   whatwork := soundexalt;
            | 24    :   whatwork := soundexorg;
            | 25,26 :   whatwork := verbatim;
            | 27,28 :   whatwork := verbatim;
                        GetString(S,dicfile);
            | 29    :   whatwork := category;
            | 30    :   whatwork := packedcategory;
            | 31    :   DEBUG:=TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting:       Str.Copy(orgpat,S);
                             (* IF chkpat(strpat,orgpat)=FALSE THEN abort(errPat,orgpat);END; *)
            | gotpat:        IF verifyString (R,digits)=FALSE THEN abort(errNum,S);END;
                             Str.Prepend(R,":"); (* fake option *)
                             IF GetLongCard(R,v)=FALSE THEN abort(errNum,S);END;
                             IF v > MAX(CARDINAL) THEN abort(errCard,S);END; (* //FIXME maxlen *)
                             minmatchlen:=CARDINAL(v);
                             maxmatchlen:=CARDINAL(v);
            | gotwordlen:    IF verifyString (R,digits)=FALSE THEN abort(errNum,S);END;
                             Str.Prepend(R,":"); (* fake option *)
                             IF GetLongCard(R,v)=FALSE THEN abort(errNum,S);END;
                             IF v > MAX(CARDINAL) THEN abort(errCard,S);END; (* //FIXME maxlen *)
                             maxmatchlen:=CARDINAL(v);
            | gotmaxwordlen: abort(errParmOverflow,S);
            END;
            INC(state);
        END;
    END;
    CASE state OF
    | waiting: abort(errExpected,"<pattern>");
    ELSE
               IF maxmatchlen < minmatchlen THEN
                   i:=minmatchlen;
                   minmatchlen:=maxmatchlen;
                   maxmatchlen:=i;
               END;
               IF chkpat(strpat,(whatwork=verbatim),orgpat)=FALSE THEN abort(errPat,orgpat);END;
    END;

    CASE whatwork OF
    | motif,verbatim:
        ;
    ELSE
        IF state # gotpat THEN abort(errNonsenseWithCode,"");END;
        IF verifyString(strpat,lettersUpp)=FALSE THEN
            abort(errLettersOnly,strpat);
        END;
    END;

    useLFN := (useLFN AND fileSupportLFN() );

    IF chkJoker(dicfile) THEN abort(errJoker,dicfile);END;
    IF fileExists(useLFN,dicfile)=FALSE THEN
        IF haspath(dicfile) THEN abort(errNotFound,dicfile); END;
        Lib.ParamStr(S,0);
        i:=Str.RCharPos(S,"\");
        IF i = MAX(CARDINAL) THEN (* never *)
            R:="";
        ELSE
            Str.Slice(R,S,0,i+1);
        END;
        Str.Prepend(dicfile,R);
        IF fileExists(useLFN,dicfile)=FALSE THEN abort(errNotFound,dicfile);END;
    END;

    IF fileIsDirectorySpec(useLFN,dicfile) THEN abort(errNotFile,dicfile);END;

    IF DEBUG THEN
        dmp    ("Dictionary",dicfile);

        isoleItemS (R,
                   "pattern,verbatim,French phonetics,tweaked Soundex,original Soundex",
                   coma,ORD(whatwork));
        dmp    ("Mode",R);

        dmp    ("Raw pattern",orgpat);
        dmp    ("Cooked pattern",strpat);

        CASE whatwork OF
        | motif,verbatim :
            R:="irrelevant";
        ELSE
            getCode(R,whatwork,strpat);
        END;
        dmp    ("Mode dependant code ",R);

        dmpbool(shoutmode,"Display","uppercase","lowercase");
        isoleItemS (R,"none,full screen,half screen",coma,ORD(paging));
        dmp    ("Paging",R);

        dmp    ("Minimum match length",fmt (minmatchlen,10,1,"",""));
        dmp    ("Maximum match length",fmt (maxmatchlen,10,1,"",""));

        WrLn;
    END;

    startfpos:=isTokenizedFile(dicfile);
    showmatches (useLFN,showstats,shoutmode,
                whatwork, paging, startfpos,
                minmatchlen,maxmatchlen,strpat,dicfile);

    abort(errNone,"");
END patMatch.

