(* ---------------------------------------------------------------
Title         Space Reserver
Notes
Bugs
Wish List     several rnd generators such as ISAAC, Mersenne, Marsaglia ?
              update Win32 with v1.2a ?

--------------------------------------------------------------- *)

MODULE Reserve;

IMPORT Lib;
IMPORT Str;
IMPORT SYSTEM;
IMPORT FIO;

FROM IO IMPORT WrLn,WrStr;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM QD_rand IMPORT InitRnd, GetRnd,
GetRndCardRange, GetRndLngCardRange,
GetRndLngRealRange, GetRndRealRange,
GetRndIntRange, GetRndLngIntRange;

(* ------------------------------------------------------------ *)

CONST
    (* we'll use hbar because DOS pipe symbol cannot belong to a filename *)
    msg           = "Filling, please wait...";
CONST
    HUGEMAXSIZE   = MAX(LONGINT);
    sMAXGB        = "2 Gb";
    sMAXBYTES     = "2147483647"; (* match HUGEMAXSIZE *)
CONST
    DECbase       = 10;
    HEXbase       = 16;
CONST
    MINFKOUNT     = 1;          sMINFKOUNT    = dot+"001";
    MAXFKOUNT     = 999;        sMAXFKOUNT    = dot+"999";
    sFkountRange  = "[1..999]";
    MINHEXFKOUNT  = 1;          sMINHEXFKOUNT = dot+"001";
    MAXHEXFKOUNT  = 0FFFH;      sMAXHEXFKOUNT = dot+"FFF";
    sHexFkountRange="[$1..$FFF]";
CONST
    ProgEXEname   = "RESERVE";
    ProgTitle     = "Q&D Space Reserver";
    ProgVersion   = "v1.2b";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone             = 0;
    errHelp             = 1;
    errOption           = 2;
    errSyntax           = 3;
    errFillerValue      = 4;
    errJoker            = 5;
    errUnit             = 6;
    errSizeValue        = 7;
    errExists           = 8;
    errGetFree          = 9;
    errCreate           = 10;
    errZeroLeft         = 11;
    errNoExtension      = 12;
    errLFNtoDOS         = 13;
    errBadVal           = 14;
    errFcountRange      = 15;
    errFcountRangeHex   = 16;
    errExhaustedDOS     = 17;
    errExhaustedDOShex  = 18;
    errNotHere          = 19;
    errNotEnoughFreeLeft =20;
    errAborted          = 21; (* or 255 ? *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msgHelp =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <size> <file> [value] [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" <"+star+"> <file> [value] [option]..."+nl+
"Syntax 3 : "+ProgEXEname+" <["+dollar+"]"+exclam+"> <file> [value] [option]..."+nl+
nl+
"  -o      overwrite existing <file>"+nl+
"  -v      verbose"+nl+
"  -ts     TopSpeed random library (default)"+nl+
"  -qd     Q&D Tools random library"+nl+
"  -n:[$]# create # files ("+sFkountRange+" or "+sHexFkountRange+"), syntax 1 only)"+nl+
"  -x      disable FAT32 support even if available"+nl+
"  -d      do no try and convert LFN path to DOS format (only with -l)"+nl+
"  -l[fn]  disable LFN support even if available (-ll = -l -d)"+nl+
nl+
"With syntax 1, <file> is created with specified <size>."+nl+
"With syntax 2, <file> is created using free space size (up to "+sMAXGB+")."+nl+
"With syntax 3, program creates as many files as needed to fill free space"+nl+
'(extension will be incremented from "'+sMINFKOUNT+'" upwards, upper limit being'+nl+
'either "'+sMAXFKOUNT+'" for "'+exclam+'" command or "'+sMAXHEXFKOUNT+'" for "'+dollar+exclam+'" command).'+nl+
nl+
"a) Optional value specifies (hexa)decimal byte value filling created file(s),"+nl+
'   "'+star+'" or "'+question+'" specifying random values.'+nl+
"b) Unit is either default unit or unit specified in <file> name."+nl+
"c) Maximum filesize is limited to "+sMAXBYTES+" bytes ("+sMAXGB+")."+nl+
"   Note <size> may have any of <Kb|Mb|Gb> suffixes (multiples of 1024),"+nl+
"   or <K|M|G> (multiples of 1000)."+nl+
'd) With -n:[$]# option, extension will be incremented from "'+sMINFKOUNT+'".'+nl+
'   Upper limit is "'+sMAXFKOUNT+'" if # is a decimal number,'+nl+
'   or "'+sMAXHEXFKOUNT+'" if # is a hexadecimal number with "$" prefix.'+nl+
"e) Should -l option be specified with a system supporting LFNs,"+nl+
"   user must specify a DOS-compatible <file> name"+nl+
"   (unless any of -d or -ll options was specified,"+nl+
"   program will try and convert <file> name to DOS f8e3 format,"+nl+
"   but user should not rely on this feature for path portion)."+nl+
"f) From Win 9X, program calls interrupt $21 with AX=$7303 ;"+nl+
"   if -x option was specified, it uses regular AH=$36 DOS call."+nl+
nl+
"Examples : "+ProgEXEname+" 2048 filler.dat ? -qd"+nl+
"           "+ProgEXEname+" $8000 filler.hex $aa"+nl+
"           "+ProgEXEname+" * filler.bin 255"+nl+
"           "+ProgEXEname+" ! filler $ff"+nl+
"           "+ProgEXEname+" 1963Mb filler.1"+nl+
"           "+ProgEXEname+" -v -o 1963M delme.2"+nl+
"           "+ProgEXEname+" 1024 delme -n:200"+nl+
"           "+ProgEXEname+" $! filler $aa"+nl;

VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :      WrStr(msgHelp);
    | errOption :    S := 'Unknown "|" option !';
    | errSyntax:     S := "Syntax error, check help screen !";
    | errFillerValue:S := 'Illegal "|" filler byte value !';
    | errJoker:      S := 'At least one unexpected joker in "|" !';
    | errUnit:       S := 'Illegal unit in "|" !';
    | errSizeValue:  S := 'Illegal "|" size value !';
    | errExists:     S := '"|" already exists !';
    | errGetFree:    S := 'Could not get |: available free space !';
    | errCreate:
        (* "name|got|wanted" *)
        Str.Subst(einfo,hbar," is ");
        Str.Subst(einfo,hbar," byte(s) long while expected size is ");
        Str.Concat(S,einfo," !");
    | errZeroLeft:   S := "No free space left on |: unit !";
    | errNoExtension:S := "No extension allowed with | !";
    | errLFNtoDOS:   S := '"|" cannot become a valid DOS name !';
    | errBadVal:     S := 'Illegal "|" value !';
    | errFcountRange:    S := "-n:# value must belong to | range !";
    | errFcountRangeHex: S := "-n:$# value must belong to | range !";
    | errExhaustedDOS   :S := "Extension has reached its "+sMAXFKOUNT+" upper limit !";
    | errExhaustedDOShex:S := "Extension has reached its "+sMAXHEXFKOUNT+" upper limit !";
    | errNotHere:    S := "-n:[$]# option and | are mutually exclusive !";
    | errNotEnoughFreeLeft:S:="Not enough free space left on |: unit !";
    | errAborted:    S := "Aborted by user !";
    ELSE
                     S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

TYPE
    HUGECARD    = LONGREAL; (* it's useless but we fake > 4Gb support *)
CONST
    HUGEZERO    = LONGREAL(0.0);
TYPE
    extendedFreeSpaceType = RECORD
        strucsize                : CARDINAL;
        version                  : CARDINAL; (* called with 0 *)
        sectorsPerCluster        : LONGCARD; (* adjusted for compression *)
        bytesPerSector           : LONGCARD;
        freeClusters             : LONGCARD; (* availableClusters *)
        totalClusters            : LONGCARD;
        physicalSectorsAvailable : LONGCARD; (* not adjusted for compression *)
        physicalTotalSectors     : LONGCARD; (* not adjusted for compression *)
        realAvailableUnits       : LONGCARD; (* not adjusted for compression *)
        totalUnits               : LONGCARD; (* not adjusted for compression *)
        reserved                 : ARRAY [1..8] OF BYTE;
    END;

(* u should be UPPERCASE but we handle it anyway *)

PROCEDURE getUnitFreeTotal (assumeFAT32:BOOLEAN; u:CHAR;
                           VAR freebytes,totalbytes:HUGECARD):BOOLEAN;
VAR
    r : SYSTEM.Registers;
    sectorsPerCluster, freeClusters, bytesPerSector, totalClusters,v : LONGCARD;
    S           : str128;
    i           : CARDINAL;
    strucFAT32  : extendedFreeSpaceType;
    unitFAT32   : str16; (* oversized *)
BEGIN
    u:=CAP(u);
    freebytes  :=HUGEZERO;
    totalbytes :=HUGEZERO;

    IF assumeFAT32 THEN
        Str.Concat(unitFAT32,u,colon+backslash);
        unitFAT32[Str.Length(unitFAT32)]:=charnull; (* safety for ASCIZ *)
        strucFAT32.version:=0;

        r.AX := 07303H;  (* FAT32 - GET EXTENDED FREE SPACE ON DRIVE *)
        r.DS := Seg( unitFAT32 );
        r.DX := Ofs( unitFAT32 );
        r.ES := Seg( strucFAT32 );
        r.DI := Ofs( strucFAT32 );
        r.CX := SIZE(strucFAT32);
        Lib.Dos(r);
        IF (SYSTEM.CarryFlag IN r.Flags) THEN RETURN FALSE; END;
        sectorsPerCluster := strucFAT32.sectorsPerCluster;
        freeClusters      := strucFAT32.freeClusters;
        bytesPerSector    := strucFAT32.bytesPerSector;
        totalClusters     := strucFAT32.totalClusters;

        freebytes := HUGECARD(bytesPerSector);
        totalbytes:= HUGECARD(bytesPerSector);
        freebytes := freebytes  * HUGECARD(strucFAT32.physicalSectorsAvailable);
        totalbytes:= totalbytes * HUGECARD(strucFAT32.physicalTotalSectors);
    ELSE
        r.AH := 036H;  (* DOS 2+ - GET FREE DISK SPACE *)
        r.DL := BYTE( ORD(u)-ORD("A")+1 ); (* $00=default, $01=A:, etc. *)
        Lib.Dos(r);
        IF r.AX = 0FFFFH THEN RETURN FALSE; END;
        sectorsPerCluster := LONGCARD(r.AX); (* IF $FFFF, invalid drive *)
        freeClusters      := LONGCARD(r.BX);
        bytesPerSector    := LONGCARD(r.CX);
        totalClusters     := LONGCARD(r.DX);

        freebytes  := HUGECARD(sectorsPerCluster) * HUGECARD(bytesPerSector);
        totalbytes := HUGECARD(sectorsPerCluster) * HUGECARD(bytesPerSector);
        freebytes  := freebytes  * HUGECARD(freeClusters);
        totalbytes := totalbytes * HUGECARD(totalClusters);
    END;
    RETURN TRUE;
END getUnitFreeTotal;

(* ------------------------------------------------------------ *)

CONST
    IObufferSize      = (8 * 512) + FIO.BufferOverhead;
    firstIObufferByte = 1;
    lastIObufferByte  = IObufferSize;
VAR
    IObuffer : ARRAY [firstIObufferByte..lastIObufferByte] OF BYTE;

CONST
    databuffersize  = 8 * 512;
    firstdatabuffer = 0;
    lastdatabuffer  = databuffersize-1;
VAR
    databuffer : ARRAY [firstdatabuffer..lastdatabuffer] OF BYTE;

(* ------------------------------------------------------------ *)

PROCEDURE strToByte (VAR filler:BYTE; S:ARRAY OF CHAR   ):BOOLEAN;
VAR
    ok : BOOLEAN;
    base,len:CARDINAL;
    v:LONGCARD;
BEGIN
    len:=Str.Length(S); (* cannot be 0 *)
    IF S[0]="$" THEN
        base:=16; Str.Delete(S,0,1);
    ELSIF ( (S[0]="0") AND (S[1]="X") ) THEN
        base:=16; Str.Delete(S,0,2);
    ELSIF S[len-1]="H" THEN
        base:=16; Str.Delete(S,len-1,1);
    ELSE
        base:=10;
    END;
    v:=Str.StrToCard(S,base,ok);
    IF ( (ok = FALSE) OR (v < MIN(BYTE)) OR (v > MAX(BYTE)) ) THEN
        ok:=FALSE;
    ELSE
        filler:=BYTE(v);
        ok:=TRUE;
    END;
    RETURN ok;
END strToByte;

PROCEDURE strToFsize (VAR fsize : LONGCARD;S:ARRAY OF CHAR ):BOOLEAN;
CONST
    bmul  = 1;
    Kbmul = 1024;
    Mbmul = Kbmul * Kbmul;
    Gbmul = Kbmul * Mbmul;
CONST
    rKmul = 1000;
    rMmul = rKmul * rKmul;
    rGmul = rKmul * rMmul;
CONST
    sMul  = "KB KO MB MO GB GO KK K MM M GG G"; (* uppercase, longest first, us and fr *)
VAR
    ok:BOOLEAN;
    i,base:CARDINAL;
    vmax,mul : LONGCARD;
    abrev:str16; (* oversized *)
BEGIN
    IF S[0]="$" THEN
        base:=16; Str.Delete(S,0,1);
    ELSE
        base:=10;
    END;
    Str.Caps(S); (* safety *)
    mul:=bmul;
    i:=0;
    LOOP
        isoleItemS(abrev, sMul,blank,i);
        IF same(abrev,"") THEN EXIT; END;
        IF Str.Pos(S,abrev) # MAX(CARDINAL) THEN
            Str.Subst(S,abrev,"");
            CASE i OF
            | 0,1:  mul:= Kbmul;
            | 2,3:  mul:= Mbmul;
            | 4,5:  mul:= Gbmul;
            | 6,7:  mul:= rKmul;
            | 8,9:  mul:= rMmul;
            |10,11: mul:= rGmul;
            END;
            EXIT;
        END;
        INC(i);
    END;
    fsize:=Str.StrToCard(S,base,ok);
    IF ok THEN
        IF mul # bmul THEN
            vmax:=LONGCARD(HUGEMAXSIZE) DIV mul;
            IF fsize > vmax THEN
                ok:=FALSE; (* too big *)
            ELSE
                fsize:=fsize * mul;
            END;
        END;
    END;
    RETURN ok;
END strToFsize;

PROCEDURE fmtval (v:LONGCARD;wi,base:CARDINAL;pad:CHAR):str16;
VAR
    ok:BOOLEAN;
    i:CARDINAL;
    R:str16;
BEGIN
    Str.CardToStr(v,R,base,ok);
    FOR i:=(Str.Length(R)+1) TO wi DO Str.Prepend(R,pad);END;
    IF base=16 THEN Str.Lows(R);END;
    RETURN R;
END fmtval;

PROCEDURE buildname (VAR R:pathtype; base:CARDINAL;v:LONGCARD;S:pathtype);
CONST
    fix = 3;
BEGIN
    Str.Concat(R,S,dot);
    Str.Append(R,fmtval(v,fix,base,"0"));
END buildname;

PROCEDURE quoteme (useLFN:BOOLEAN;S:ARRAY OF CHAR   ):str256; (* should do *)
VAR
    R:str256;
BEGIN
    Str.Copy(R,S);
    IF useLFN THEN
       Str.Append(R,dquote);
       Str.Prepend(R,dquote);
    ELSE
       Str.Caps(R);
    END;
    RETURN R;
END quoteme;

PROCEDURE LFNtoDOS (VAR renamed:BOOLEAN;VAR S:pathtype;
                   LFNos,useLFN:BOOLEAN   ):BOOLEAN;
CONST
    alphaset      = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"; (* uppercase *)
    symbols       = "_-$!(){}~#&@[]";
    legalLFN      = ",;'.+%="+blank;
    DOScharset    = alphaset+digits+symbols;
    LFNcharset    = DOScharset+legalLFN;
VAR
    org,R:pathtype;
    i,len:CARDINAL;
    ok:BOOLEAN;
    u,d:pathtype;
    n:ARRAY[0..8-1] OF CHAR;
    e:ARRAY[0..1+3-1] OF CHAR; (* ".???"  *)
BEGIN
    Str.Copy(org,S);
    renamed:=FALSE;
    ok:=TRUE;

    IF NOT(LFNos) THEN RETURN ok; END; (* vanilla DOS *)
    IF useLFN THEN RETURN ok; END;     (* Win9X *)

    UpperCase(S);

    (* filter out illegal chars *)

    R:="";
    FOR i:=1 TO Str.Length(S) DO
        CASE S[i-1] OF
        | colon:
            IF i # 2 THEN RETURN FALSE; END; (* NOT "?:*" *)
            Str.Append(R,colon);
        | dot:
            Str.Append(R,dot); (* we'll handle dot(s) later *)
        ELSE
            IF Belongs(DOScharset,S[i-1]) THEN Str.Append(R,S[i-1]);END;
        END;
    END;

    (* try and make sure we're f8e3 *)
    qdSplitAllPath(R,u,d,n,e);

    Lib.MakeAllPath(R,u,d,n,"");
    IF Str.CharPos(R,dot) # MAX(CARDINAL) THEN RETURN FALSE;END;

    Lib.MakeAllPath(R,"","",n,e);
    ok:= NOT ( same(R,"") );
    Lib.MakeAllPath(R,u,d,n,e);
    IF ok THEN
        Str.Copy(S,R);
    ELSE
        Str.Copy(S,org);
    END;
    renamed:=NOT(same(S,org));
    RETURN ok;
END LFNtoDOS;

PROCEDURE getLC (VAR v:LONGCARD; VAR numbase:CARDINAL;
                S:ARRAY OF CHAR):BOOLEAN;
VAR
    ok:BOOLEAN;
    p:CARDINAL;
BEGIN
    CASE S[0] OF
    | dollar:
        Str.Delete(S,0,1);
        numbase:=HEXbase;
    ELSE
        numbase:=DECbase;
    END;
    v := Str.StrToCard(S,numbase,ok);
    RETURN ok;
END getLC;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    sname,basename:pathtype;
    slen,sbyte :str128; (* oversized *)
    hnd:FIO.File;
    renamed,LFNos,useLFN:BOOLEAN;
    forcefix,forced,assumeFAT32,overwrite,verbose:BOOLEAN;
    fillmode:(nofill,randomfill,userfill);
    filler:BYTE;
    wantedfsize:LONGCARD;
    freespace,totalspace: HUGECARD;
    hugefsize:HUGECARD;
    unit:CHAR;
    currext:LONGCARD;
    fKount:CARDINAL;
    rndlib:(defaultrndlib,tsrndlib,qdrndlib);
    ok:BOOLEAN;
VAR
    parmcount,i,opt:CARDINAL;
    S,R:pathtype;
    state:(waiting,gotsize,gotfile,gotvalue);
    pos:LONGCARD;
    datacount:CARDINAL;
    N:str128;
    numbase:CARDINAL;
    opmode : (fixedsizeone,fixedsizemany,tryone,trymany); (* syntaxes 1, 1+n:#, 2, 3 *)
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn;

    rndlib    := defaultrndlib;
    fillmode  := nofill;
    forcefix  := TRUE;
    overwrite := FALSE;
    verbose   := FALSE;
    assumeFAT32:=TRUE;
    fKount    := MAX(CARDINAL);
    currext   := MINFKOUNT; (* 1 *)
    numbase   := DECbase;
    useLFN    := TRUE;

    state:=waiting;
    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S);
        UpperCase(R);
        cleantabs(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "X"+delim+"NOFAT32"+delim+
                                  "L"+delim+"LFN"+delim+
                                  "LL"+delim+"LLFN"+delim+
                                  "D"+delim+"DOS"+delim+
                                  "TS"+delim+"TOPSPEED"+delim+
                                  "QD"+delim+"QDTOOLS"+delim+
                                  "N:"
                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5:    verbose     := TRUE;
            | 6,7:    overwrite   := TRUE;
            | 8,9 :   assumeFAT32 := FALSE;
            | 10,11:  useLFN      := FALSE;
            | 12,13:  useLFN      := FALSE;
                      forcefix    := FALSE;
            | 14,15:  forcefix    := FALSE;
            | 16,17:  rndlib      := tsrndlib;
            | 18,19:  rndlib      := qdrndlib;
            | 20:     GetString(R, N); (* keep upper case *)
                      ok:=getLC(pos,numbase, N);
                      IF NOT(ok) THEN abort(errBadVal,S);END;
                      CASE numbase OF
                      | DECbase:
                          IF pos < MINFKOUNT THEN abort(errFcountRange,sFkountRange);END;
                          IF pos > MAXFKOUNT THEN abort(errFcountRange,sFkountRange);END;
                      | HEXbase:
                          IF pos < MINHEXFKOUNT THEN abort(errFcountRangeHex,sHexFkountRange);END;
                          IF pos > MAXHEXFKOUNT THEN abort(errFcountRangeHex,sHexFkountRange);END;
                      END;
                      fKount:=CARDINAL(pos);
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting:  Str.Copy(slen,R);
            | gotsize:  Str.Copy(sname,S);   (* keep case *)
            | gotfile:  Str.Copy(sbyte,R);
            | gotvalue: abort(errSyntax,"");
            END;
            INC(state);
        END;
    END;
    CASE state OF
    | waiting: abort(errHelp,"");
    | gotsize: abort(errSyntax,"");
    | gotfile: ; (* ok *)
    | gotvalue:
        IF same(sbyte,star) THEN
            fillmode:=randomfill;
        ELSIF same(sbyte,question) THEN
            fillmode:=randomfill;
        ELSE
            IF strToByte(filler,sbyte)=FALSE THEN abort(errFillerValue,sbyte); END;
            fillmode:=userfill;
        END;
    END;

    IF chkJoker(sname) THEN abort(errJoker,sname);END;

    LFNos       := fileSupportLFN();
    useLFN      := ( useLFN AND LFNos );
    assumeFAT32 := ( assumeFAT32 AND LFNos );

    IF forcefix THEN
        IF LFNtoDOS(renamed,sname, LFNos,useLFN)=FALSE THEN abort(errLFNtoDOS,sname);END;
    END;

    i:=Str.CharPos(sname,colon);
    CASE i OF
    | 1 :            unit:=sname[0];
    | MAX(CARDINAL): unit:=CHR(  FIO.GetDrive()-1+ORD("A") );
    ELSE
                     abort(errUnit,sname);
    END;

    IF same(slen,star) THEN
        IF fKount # MAX(CARDINAL) THEN abort(errNotHere,"syntax 2");END;
        opmode:=tryone;            (* syntax 2 *)
    ELSIF ( same(slen,exclam) OR same(slen,dollar+exclam) ) THEN
        IF fKount # MAX(CARDINAL) THEN abort(errNotHere,"syntax 3");END;
        IF Str.RCharPos(sname,dot) # MAX(CARDINAL) THEN
            abort(errNoExtension,"syntax 3");
        END;
        Str.Copy(basename,sname);
        opmode:=trymany;           (* syntax 3 *)
        CASE slen[0] OF
        | dollar:numbase:=HEXbase;
        | exclam:numbase:=DECbase;
        END;
    ELSE
        IF strToFsize(wantedfsize,slen)=FALSE THEN abort(errSizeValue,slen);END;
        IF wantedfsize = 0 THEN abort(errSizeValue,slen);END;
        IF wantedfsize > HUGEMAXSIZE THEN
            wantedfsize:=HUGEMAXSIZE; forced:=TRUE;
        END;
        IF fKount = MAX(CARDINAL) THEN
            opmode:=fixedsizeone; (* syntax 1 *)
        ELSE
            IF Str.RCharPos(sname,dot) # MAX(CARDINAL) THEN
                abort(errNoExtension,"-n:# option");
            END;
            Str.Copy(basename,sname);
            opmode:=fixedsizemany; (* syntax 1 with -n:# *)
        END;
    END;

    forced:=FALSE;
    CASE opmode OF
    | fixedsizeone:
        ;
    | tryone,trymany:
        IF getUnitFreeTotal(assumeFAT32,unit, freespace,totalspace)=FALSE THEN
            abort(errGetFree,unit);
        END;
        IF freespace = HUGEZERO THEN abort(errZeroLeft,unit);END;
        IF freespace > HUGECARD(HUGEMAXSIZE) THEN
            wantedfsize:=HUGEMAXSIZE; forced:=TRUE;
        ELSE
            wantedfsize:=LONGCARD(freespace);
        END;
    | fixedsizemany:
        IF getUnitFreeTotal(assumeFAT32,unit, freespace,totalspace)=FALSE THEN
            abort(errGetFree,unit);
        END;
        IF freespace = HUGEZERO THEN abort(errZeroLeft,unit);END;
        IF freespace < HUGECARD(wantedfsize) THEN abort(errNotEnoughFreeLeft,unit);END;
    END;

    IF fillmode = randomfill THEN
        CASE rndlib OF
        | defaultrndlib,tsrndlib: Lib.RANDOMIZE(); rndlib:=tsrndlib;
        | qdrndlib:               InitRnd();
        END;
    END;

    LOOP
        CASE opmode OF
        | trymany,fixedsizemany : buildname(sname, numbase,currext,basename);
        END;

        IF fileExists(useLFN,sname) THEN
            IF overwrite THEN             (* better safe than sorry *)
                fileSetRW(useLFN,sname);
                fileErase(useLFN,sname);
            ELSE
                abort(errExists,sname);
            END;
        END;

        IF verbose THEN
            CASE opmode OF
            | trymany,fixedsizemany : IF currext > 1 THEN WrLn;END;
            END;
            WrStr("Unit   : "); WrStr(unit);WrStr(":");WrLn;
            WrStr("File   : "); WrStr(quoteme(useLFN,sname));
                                IF renamed THEN WrStr(" (renamed)");END;
                                WrLn;
            WrStr("Size   : "); WrStr(fmtval(wantedfsize,1,10,"0"));
                                IF forced THEN WrStr(" (adjusted)");END;
                                WrLn;
            WrStr("Filler : ");
            CASE fillmode OF
            | nofill:     S:="unmodified disk data"; (* was "undefined" *)
            | randomfill: S:="random";
                CASE rndlib OF
                | tsrndlib  : Str.Append(S," (TopSpeed)");
                | qdrndlib  : Str.Append(S," (Q&D Tools)");
                END;
            | userfill:   Str.Concat(S,"$",fmtval(LONGCARD(filler),2,16,"0"));
            END;
            WrStr(S);WrLn;
            WrLn;
        END;

        hnd := fileCreate(useLFN,sname);
        FIO.Seek (hnd,1); (* try and fix DOS bug reported by intlist *)
        FIO.Truncate(hnd);
        fileClose(useLFN,hnd);

        hnd := fileOpen(useLFN,sname);
        FIO.Seek (hnd,wantedfsize);
        FIO.Truncate (hnd);
        fileClose(useLFN,hnd);

        pos:=fileGetFileSize(useLFN,sname);
        IF pos # wantedfsize THEN
            Str.Concat(S,hbar,fmtval(pos,1,10,"0"));
            Str.Prepend(S,sname);
            Str.Append(S,hbar);Str.Append(S,fmtval(wantedfsize,1,10,"0")); (* //V12A was slen ! *)
            abort(errCreate,S); (* "sname|got|wanted" *)
        END;

        Str.Concat(S,"::: ",quoteme(useLFN,sname));
        Str.Append(S," has been created with ");
        Str.Append(S,fmtval(wantedfsize,1,10,"0"));
        Str.Append(S," byte");
        IF wantedfsize > 1 THEN Str.Append(S,"s");END;
        IF forced THEN Str.Append(S," (adjusted)"); END;
        Str.Append(S," size.");
        WrStr(S);WrLn;

        IF ( (fillmode # nofill) AND (wantedfsize # 0) ) THEN
            video(msg,TRUE);
            CASE fillmode OF
            | randomfill:
                FOR i:=firstdatabuffer TO lastdatabuffer DO
                    CASE rndlib OF
                    | tsrndlib : filler:=BYTE( Lib.RANDOM(256) ); (* 0..255 *)
                    | qdrndlib : filler:=BYTE( GetRndCardRange (0,255) );
                    END;
                    databuffer[i]:=filler;
                END;
            | userfill:
                Lib.Fill(ADR(databuffer),databuffersize,filler);
            END;

            hnd := fileOpen(useLFN,sname);
            FIO.AssignBuffer(hnd,IObuffer);
            pos := 0;
            LOOP
                IF pos >= wantedfsize THEN EXIT; END;
                datacount := CARDINAL ( (wantedfsize-pos) MOD databuffersize );
                IF datacount = 0 THEN datacount := databuffersize; END;
                FIO.WrBin(hnd,databuffer,datacount);
                INC(pos,LONGCARD(datacount));
                IF fillmode=randomfill THEN
                    FOR i:=firstdatabuffer TO lastdatabuffer DO
                        CASE rndlib OF
                        | tsrndlib : filler:=BYTE( Lib.RANDOM(256) ); (* 0..255 *)
                        | qdrndlib : filler:=BYTE( GetRndCardRange (0,255) );
                        END;
                        databuffer[i]:=filler;
                    END;
                END;
            END;
            fileFlush(useLFN,hnd);
            fileClose(useLFN,hnd);
            video(msg,FALSE);
        END;

        CASE opmode OF
        | fixedsizeone,tryone:
            EXIT;
        | trymany:
            IF getUnitFreeTotal(assumeFAT32,unit, freespace,totalspace)=FALSE THEN
                abort(errGetFree,unit);
            END;
            IF freespace = HUGEZERO THEN EXIT; END; (* we're done *)
            IF freespace > HUGECARD(HUGEMAXSIZE) THEN
                wantedfsize:=HUGEMAXSIZE; forced:=TRUE;
            ELSE
                wantedfsize:=LONGCARD(freespace);
            END;
        | fixedsizemany:
            IF fKount <= MINFKOUNT THEN EXIT;END;
            DEC(fKount);

            IF getUnitFreeTotal(assumeFAT32,unit, freespace,totalspace)=FALSE THEN
                abort(errGetFree,unit);
            END;
            IF freespace = HUGEZERO THEN abort(errZeroLeft,unit);END;
            IF freespace < HUGECARD(wantedfsize) THEN abort(errNotEnoughFreeLeft,unit);END;
        END;

        CASE opmode OF
        | fixedsizemany,trymany:
            INC(currext);
            IF NOT(useLFN) THEN
                CASE numbase OF
                | DECbase: IF currext > MAXFKOUNT THEN abort(errExhaustedDOS,"");END;
                | HEXbase: IF currext > MAXHEXFKOUNT THEN abort(errExhaustedDOShex,"");END;
                END;
        END;
            IF ChkEscape() THEN abort(errAborted,"");END;
        END;
    END;

    abort(errNone,"");
END Reserve.



(*
--------D-2142-------------------------------
INT 21 - DOS 2+ - "LSEEK" - SET CURRENT FILE POSITION
        AH = 42h
        AL = origin of move
            00h start of file
            01h current file position
            02h end of file
        BX = file handle
        CX:DX = (signed) offset from origin of new file position
Return: CF clear if successful
            DX:AX = new file position in bytes from start of file
        CF set on error
            AX = error code (01h,06h) (see #01680 at AH=59h/BX=0000h)
Notes:  for origins 01h and 02h, the pointer may be positioned before the
          start of the file; no error is returned in that case (except under
          Windows NT), but subsequent attempts at I/O will produce errors
        if the new position is beyond the current end of file, the file will
          be extended by the next write (see AH=40h); for FAT32 drives, the
          file must have been opened with AX=6C00h with the "extended size"
          flag in order to expand the file beyond 2GB
BUG:    using this method to grow a file from zero bytes to a very large size
          can corrupt the FAT in some versions of DOS; the file should first
          be grown from zero to one byte and then to the desired large size
SeeAlso: AH=24h,INT 2F/AX=1228h
*)
