(* ---------------------------------------------------------------
Title         Q&D Segmented CRC
Author        PhG
Overview      compute CRC32s for split segments of a file
Usage         see help
Notes         unsegmented (-n option) is in fact a trick :
              MAX(LONGCARD)-safety (i.e. -16)
              an ini file for exclusions ?
              accept a filelist ?
Bugs
Wish List     MD5 but well... what's the use ?

--------------------------------------------------------------- *)

MODULE segCRC;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;

FROM IO IMPORT WrStr,WrLn;

FROM Storage IMPORT ALLOCATE,DEALLOCATE,Available;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XopenFile, w9XcloseFile,
w9XsupportLFN;

FROM QD_CRC IMPORT ComputeCRC32, ComputeCRCval, setSigmaUse,
SegmentFileComputeCRC32;

(* ------------------------------------------------------------ *)

CONST
    MINENTRY      = 1;
    MAXENTRY      = MAX(CARDINAL)-2; (* is a count *)
    strMAXENTRIES = "65533";
    maxDOSf8      = 8;
    maxDOSe3      = 3;
    strSIZERANGE  = "[1..1024]";
    MAXHASH       = MAX(CARDINAL)-2;
    methodCRC32   = 0;
    methodMD5     = 1;
CONST
    patLOG        = "*.LOG";
    patTMP        = "*.TMP";
    patSWAP92     = "368SPART.PAR";
    patSWAP98     = "WIN386.SWP";
    patSWAPXP     = "PAGEFILE.SYS";
    patHIBERNATE  = "HIBERFIL.SYS";
    skippedpats   = patLOG+delim+patTMP+delim+
                    patSWAP92+delim+patSWAP98+delim+patSWAPXP+delim+
                    patHIBERNATE;
CONST
    extLOG        = ".LOG";
    extEXE        = ".EXE";
    defaultSegSize= 2;
    minSegSize    = 1;
    maxSegSize    = 1024; (* if we're to check huge files, allow for one Mb safety ! *)
    bytesPerKb    = 1024;
    openmarker    = "[";
    closemarker   = "]";
    safety        = 16; (* substracted from MAX(LONGCARD) *)
CONST
    ProgEXEname   = "SEGCRC";
    ProgTitle     = "Q&D Segmented CRC32";
    ProgVersion   = "v1.1";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
    defaultlog    = ProgEXEname+extLOG;
CONST
    errNone             = 0;
    errHelp             = 1;
    errUnknownOption    = 2;
    errKierkegaard      = 3; (* aut... aut... *)
    errBadNumber        = 4;
    errBadSegSize       = 5;
    errNoMatch          = 6;
    errLogNotFound      = 7;
    errMissingSpec      = 8;
    errLogHere          = 9;
    errROlogHere        = 10;
    errNotLocalLog      = 11;
    errTooManyParms     = 12;
    errCountOverflow    = 13;
    errJoker            = 14;
    errF8E3             = 15;
    errStorage          = 16;
    errCasemode         = 17;
    errReallyNoMatch    = 18;
    errOneList          = 19;
    errJokerFileList    = 20;
    errFileListNotFound = 21;
    errEmptyFileList    = 22;
    errNotLocal         = 23;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
00000000011111111112222222222333333333344444444445555555555666666666677777777778
1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
helpmsg =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" [-b] <file>... [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" [-b] [-@] <[@]filelist> [option]..."+nl+
"Syntax 3 : "+ProgEXEname+" <-c> [option]..."+nl+
nl+
"This program builds (syntaxes 1 and 2) or checks (syntax 3) a log file"+nl+
"(default is "+defaultlog+") containing CRC32s for segmented portions of file(s)."+nl+
nl+
"  -b    build log file (default command)"+nl+
"  -c    check log file (default is "+defaultlog+")"+nl+
(* "  -m    use MD5 instead of default CRC32"+nl+ *)
'  -l:$  user-specified log file ("'+extLOG+'" highly recommended)'+nl+
"  -s:#  segment size from 1 to 1024 Kb (default is 2)"+nl+
"  -@    force list mode (thus assuming first parameter is a filelist)"+nl+
"  -n    compute normal unsegmented CRC32 for whole file (-s:# ignored)"+nl+
"  -o    overwrite existing log file (-oo = overwrite read-only)"+nl+
"  -v    verbose processing"+nl+
"  -w    display filenames in lowercase"+nl+
"  -u    display filenames in uppercase"+nl+
"  -a    audio warning"+nl+
"  -x    disable LFN support even if available"+nl+
nl+
"a) This program processes files (log and data) in current directory only."+nl+
"b) When building a log file, any file specification without match is ignored."+nl+
"c) Any option irrelevant to command (-build or -check) is ignored."+nl+
"d) With syntax 2, this program processes each entry found in <filelist>"+nl+
"   (jokers and directories not allowed, but paths are)."+nl+
"e) The following patterns are always ignored :"+nl+
"   "+skippedpats+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :         WrStr(helpmsg);
    | errUnknownOption: S := 'Unknown "~" option !';
    | errKierkegaard:   S := "-build and -check commands are mutually exclusive !";
    | errBadNumber:     S := 'Illegal number in "~" option !';
    | errBadSegSize:    S := "Specified segment size is not in legal "+strSIZERANGE+" range !";
    | errNoMatch:       S := 'No file matches "~" specification !';
    | errLogNotFound:   S := '"~" log file does not exist !';
    | errMissingSpec:   S := "Missing file specification(s) for -build command !";
    | errLogHere:       S := 'Directory already contains "~" log file !';
    | errROlogHere:     S := 'Directory already contains "~" read-only log file !';
    | errNotLocalLog:   S := 'Illegal "~" log specification !';
    | errTooManyParms:  S := '"~" is one parameter too many !';
    | errCountOverflow: S := "More than "+strMAXENTRIES+' files !';
    | errJoker:         S := '"~" should not contain any joker !';
    | errF8E3:          S := '"~" does not fit "????????.???" DOS mask !';
    | errStorage:       S := "Lib.Storage() failure !";
    | errCasemode:      S := "-w and -u options are mutually exclusive !";
    | errReallyNoMatch: S := "No file matches specifications !";
    | errOneList:       S := "-@ option requires only one filelist !";
    | errJokerFileList: S := '"~" filelist should not contain any joker !';
    | errFileListNotFound:S:='"~" filelist does not exist !';
    | errEmptyFileList: S := '"~" filelist is empty !';
    | errNotLocal:      S := 'Illegal "~" specification !';
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp :
        ; (* nada *)
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType  = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn,ioBufferOut:ioBufferType;

(* ------------------------------------------------------------ *)

PROCEDURE sound (freq,duration,pause:CARDINAL);
BEGIN
    Lib.Sound(freq);
    Lib.Delay(duration);
    Lib.NoSound();
    Lib.Delay(pause);
END sound;

PROCEDURE alarm (  );
BEGIN
    sound(55,55,100);
    sound(55,55,10);
END alarm;

PROCEDURE alert (  );
BEGIN
    sound(550,55,100);
    sound(550,55,10);
END alert;

PROCEDURE valToStr (v:LONGCARD;VAR R:ARRAY OF CHAR);
VAR
    ok:BOOLEAN;
BEGIN
    Str.CardToStr(v,R,10,ok);
END valToStr;

PROCEDURE valToHex (v:LONGCARD;VAR R:ARRAY OF CHAR);
VAR
    ok:BOOLEAN;
    len,i:CARDINAL;
BEGIN
    Str.CardToStr(v,R,16,ok);
    len:=Str.Length(R);
    FOR i:= (len+1) TO 8 DO
        Str.Prepend(R,"0");
    END;
    Str.Lows(R);
    Str.Prepend(R,"$");
END valToHex;

PROCEDURE strToVal (S:ARRAY OF CHAR):LONGCARD;
VAR
    ok:BOOLEAN;
BEGIN
    RETURN Str.StrToCard(S,10,ok);
END strToVal;

TYPE
    casifytype    = (defcase,uppercase,lowercase);

PROCEDURE casify(casemode:casifytype;VAR S:ARRAY OF CHAR);
BEGIN
    CASE casemode OF
    | defcase: ;
    | lowercase: LowerCaseAlt(S);
    | uppercase: UpperCaseAlt(S);
    END;
END casify;

PROCEDURE pad (wi:CARDINAL;doleft:BOOLEAN;padchar:CHAR;S:ARRAY OF CHAR):str80;
VAR
    R:str80;
    i:CARDINAL;
BEGIN
    Str.Copy(R,S);
    FOR i:=(Str.Length(S)+1) TO wi DO
        IF doleft THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
    RETURN R;
END pad;

PROCEDURE num2str (v,wi:CARDINAL;padme:BOOLEAN;padchar:CHAR):str16;
CONST
    base = 10;
VAR
    ok:BOOLEAN;
    R:str16;
BEGIN
    Str.CardToStr(LONGCARD(v),R,base,ok);
    IF padme THEN
        RETURN str16 ( pad(wi,FALSE,padchar,R) );
    ELSE
        RETURN R;
    END;
END num2str;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    pFname = POINTER TO fnameType;
    fnameType = RECORD
        next      : pFname;
        index     : CARDINAL; (* useless for now unless we want *)
        hashUpper : CARDINAL; (* build on uppercase filename *)
        slen      : SHORTCARD;
        str       : CHAR;
    END;

PROCEDURE initList (VAR count:CARDINAL; VAR anchor : pFname );
BEGIN
    count  := MINENTRY-1;
    anchor := NIL;
END initList;

PROCEDURE freeList (anchor : pFname);
VAR
    needed : CARDINAL;
    p      : pFname;
BEGIN
    (* p:=anchor; *)
    WHILE anchor # NIL DO
        needed := SIZE(fnameType) - SIZE(anchor^.str) + CARDINAL(anchor^.slen);
        p := anchor^.next;
        DEALLOCATE(anchor,needed);
        anchor:=p;
    END
END freeList;

PROCEDURE buildNewPtr (VAR anchor,p:pFname; index,len:CARDINAL):BOOLEAN;
VAR
    needed : CARDINAL;
BEGIN
    needed := SIZE(fnameType) - SIZE(p^.str) + len;
    IF Available(needed)=FALSE THEN RETURN FALSE; END;
    IF anchor = NIL THEN
        ALLOCATE(anchor,needed);
        p:=anchor;
    ELSE
        p:=anchor;
        WHILE p^.next # NIL DO
            p:=p^.next;
        END;
        ALLOCATE(p^.next,needed);
        p:=p^.next;
    END;
    p^.index:= index;
    p^.next := NIL;
    RETURN TRUE;
END buildNewPtr;

(* assume p is valid *)

PROCEDURE getStr (VAR S : pathtype; p:pFname);
VAR
    len:CARDINAL;
BEGIN
    len := CARDINAL(p^.slen);
    Lib.FastMove( ADR(p^.str),ADR(S),len);
    S[len] := nullchar; (* REQUIRED safety ! *)
END getStr;

PROCEDURE isunique(hash:CARDINAL;anchor:pFname;entryname:pathtype):BOOLEAN;
VAR
    p:pFname;
    rc:BOOLEAN;
    str:pathtype;
BEGIN
    rc:=TRUE;
    p:=anchor;
    LOOP
        IF p = NIL THEN EXIT;END;
        IF p^.hashUpper = hash THEN
            getStr(str,p);
            (* IF same(entryname,str) THEN rc:=FALSE;EXIT;END; *)
            IF Str.Match(entryname,str) THEN rc:=FALSE;EXIT;END;
        END;
        p := p^.next;
    END;
    RETURN rc;
END isunique;

PROCEDURE findByIndex (index:CARDINAL;anchor:pFname ):pFname;
VAR
    p:pFname;
    str:pathtype;
BEGIN
    p:=anchor;
    LOOP
        IF p = NIL THEN EXIT; END;
        IF index = p^.index THEN EXIT;END;
        p := p^.next;
    END;
    RETURN p; (* NIL means not found *)
END findByIndex;

PROCEDURE dmpList (anchor:pFname );
CONST
    wi=5; (* 65536 *)
VAR
    p:pFname;
    str:pathtype;
    i,index:CARDINAL;
BEGIN
    i:=MINENTRY-1;
    p:=anchor;
    LOOP
        IF p = NIL THEN EXIT;END;
        INC(i);
        index:=p^.index;
        str:="~ ::: ";
        Str.Subst(str,"~",num2str(index,wi,TRUE," "));
        WrStr(str);
        getStr(str,p);
        WrStr(str);WrLn;
        p:=p^.next;
    END;
END dmpList;

(* ------------------------------------------------------------ *)

PROCEDURE notskipped (S,skipthem:ARRAY OF CHAR):BOOLEAN;
VAR
    n:CARDINAL;
    rc:BOOLEAN;
    pat:pathtype;
BEGIN
    rc:=FALSE;
    n:=0;
    LOOP
        isoleItemS(pat, skipthem,delim,n);
        IF same(pat,"") THEN rc:=TRUE;EXIT;END;
        IF Str.Match(S,pat) THEN EXIT; END; (* NOT CASE sensitive *)
        INC(n);
    END;
    RETURN rc;
END notskipped;

PROCEDURE isReservedEntry (S:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

CONST
    PBSTORAGE = MAX(CARDINAL);
    PBCOUNT   = MAX(CARDINAL)-1;

PROCEDURE buildFileList (VAR count:CARDINAL; VAR anchor:pFname;
                        DEBUG,useLFN:BOOLEAN;
                        spec:pathtype;skippedpats:ARRAY OF CHAR);
VAR
    ok,found:BOOLEAN;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    entry : FIO.DirEntry;
    dosattr:FIO.FileAttr;
    entryname,upperentryname:pathtype;
    hash,len : CARDINAL;
    pp:pFname;
    includeme:BOOLEAN;
BEGIN
    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,entry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(entryname,w9Xentry.fullfilename);
        ELSE
            Str.Copy(entryname,entry.Name);
        END;

        includeme := NOT( isReservedEntry(entryname) ); (* skip "." and ".." *)
        includeme := includeme AND notskipped(entryname,skippedpats);
        IF includeme THEN
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=entry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                (* if file has no extension, add it as a marker *)
                IF Str.RCharPos(entryname,".")=MAX(CARDINAL) THEN
                    Str.Append(entryname,".");
                END;
                IF count >= MAXENTRY THEN
                    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
                    count:=PBCOUNT;
                    RETURN;
                END;

                Str.Copy(upperentryname,entryname);
                Str.Caps(upperentryname); (* FOR better duplicate detection *)
                hash:=Lib.HashString(upperentryname,MAXHASH);

                IF isunique(hash, anchor,entryname) THEN
                    len:=Str.Length(entryname);
                    IF buildNewPtr(anchor,pp,count,len)=FALSE THEN
                        IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
                        count:=PBSTORAGE;
                        RETURN;
                    END;
                    INC(count);
                    pp^.slen      := SHORTCARD(len);
                    pp^.hashUpper := hash;
                    pp^.index     := count; (* [1..#] *)
                    Lib.FastMove ( ADR(entryname),ADR(pp^.str),len );
IF DEBUG THEN WrStr('/// included : "');WrStr(entryname);WrStr('"');WrLn; END;
                ELSE
IF DEBUG THEN WrStr('=== ignored  : "');WrStr(entryname);WrStr('"');WrLn; END;
                END;
            END;
        ELSE
IF DEBUG THEN WrStr('\\\ excluded : "');WrStr(entryname);WrStr('"');WrLn; END;
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(entry);
        END;
    END;
    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
END buildFileList;

(* ------------------------------------------------------------ *)

PROCEDURE preprocessLine (VAR S:ARRAY OF CHAR);
VAR
    len:CARDINAL;
BEGIN
    LtrimBlanks(S);
    RtrimBlanks(S);
    (* we could use Str.Match(S,dquote+"*"+dquote) but... *)
    len:=Str.Length(S);
    IF len < 2 THEN RETURN; END; (* at least 1+1 for possible dquote+dquote *)
    IF S[len-1] # dquote THEN RETURN; END;
    IF S[0] # dquote THEN RETURN; END;
    S[len-1]:=charnull;
    Str.Delete(S,0,1);
    (* just in case *)
    LtrimBlanks(S);
    RtrimBlanks(S);
END preprocessLine;

PROCEDURE readline (hnd:FIO.File;VAR R:ARRAY OF CHAR):BOOLEAN;
BEGIN
    LOOP
        FIO.RdStr(hnd,R);
        preprocessLine(R);
        CASE R[0] OF
        | charnull,semicolon,pound:
            ;
        ELSE
            EXIT;
        END;
        IF FIO.EOF THEN EXIT;END;
    END;
    RETURN FIO.EOF;
END readline;

PROCEDURE fakeBuildFileList (VAR count:CARDINAL; VAR anchor:pFname;
                            DEBUG,useLFN:BOOLEAN;
                            listfile:pathtype;skippedpats:ARRAY OF CHAR);
VAR
    hash,len : CARDINAL;
    pp:pFname;
    includeme:BOOLEAN;
    hin:FIO.File;
    upperentryname,S:pathtype; (* str1024 *)
BEGIN
    hin:=fileOpenRead(useLFN,listfile);
    FIO.AssignBuffer(hin,ioBufferIn);
    FIO.EOF := FALSE; (* safety *)
    LOOP
        IF FIO.EOF THEN EXIT; END;
        IF readline(hin,S) THEN EXIT;END;
            includeme := notskipped(S,skippedpats);
            includeme := includeme AND (chkJoker(S)=FALSE);
            IF includeme THEN
                (* if file has no extension, add it as a marker *)
                IF Str.RCharPos(S,".")=MAX(CARDINAL) THEN
                    Str.Append(S,".");
                END;
                IF count >= MAXENTRY THEN
                    count:=PBCOUNT;
                    RETURN;
                END;

                Str.Copy(upperentryname,S);
                Str.Caps(upperentryname); (* FOR better duplicate detection *)
                hash:=Lib.HashString(upperentryname,MAXHASH);

                IF isunique(hash, anchor,S) THEN
                    len:=Str.Length(S);
                    IF buildNewPtr(anchor,pp,count,len)=FALSE THEN
                        count:=PBSTORAGE;
                        RETURN;
                    END;
                    INC(count);
                    pp^.slen      := SHORTCARD(len);
                    pp^.hashUpper := hash;
                    pp^.index     := count; (* [1..#] *)
                    Lib.FastMove ( ADR(S),ADR(pp^.str),len );
IF DEBUG THEN WrStr('/// INCLUDED : "');WrStr(S);WrStr('"');WrLn; END;
                ELSE
IF DEBUG THEN WrStr('=== IGNORED  : "');WrStr(S);WrStr('"');WrLn; END;
                END;
            ELSE
IF DEBUG THEN WrStr('\\\ EXCLUDED : "');WrStr(S);WrStr('"');WrLn; END;
            END;
    END;
    fileClose(useLFN,hin);
END fakeBuildFileList;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE doCheck (casemode:casifytype;
                  useLFN,verbose,audiowarn:BOOLEAN; log:pathtype):LONGCARD;
CONST
    msgFileNotFound  = '--- File not found : "~"';
    msgSizeMismatch  = '### Size mismatch  : "~"';
    msgCRCmismatch   = '### CRC32 mismatch : "~"';
    msgCRCmismatchAt = '### CRC32 mismatch : "~" in [|..|] range';
    msgOK            = '+++ File match     : "~"';
VAR
    hlog, hin:FIO.File;
    FSHOW,FORG:pathtype;
    msg:str1024; (* oversized *)
    S,N:str128; (* oversized *)
    testit:BOOLEAN;
    fsize,orgsize,chunk,crc32,remaining,count,mismatches,fpos,pb:LONGCARD;
BEGIN
    pb:=0;

    FIO.EOF := FALSE;
    hlog:=fileOpenRead(useLFN,log);
    FIO.AssignBuffer(hlog,ioBufferOut);

    IF readline(hlog,S) THEN RETURN pb;END; (* chunksize *)
    chunk:=strToVal(S);

    LOOP
        IF FIO.EOF THEN EXIT; END;
        IF readline(hlog,S) THEN EXIT;END; (* next openmarker or empty string *)

        IF readline(hlog,FORG) THEN EXIT;END; (* filename *)
        Str.Copy(FSHOW,FORG);
        casify(casemode,FSHOW);

        testit:=fileExists(useLFN,FORG);
        IF testit THEN
            fsize:=fileGetFileSize(useLFN,FORG);
        ELSE
            msg:=msgFileNotFound;
            Str.Subst(msg,"~",FSHOW);
            WrStr(msg);WrLn;
            INC(pb);
        END;

        IF readline(hlog,S) THEN EXIT;END; (* fsize *)
        IF testit THEN
            orgsize:=strToVal(S);
            IF orgsize # fsize THEN
                msg:=msgSizeMismatch;
                Str.Subst(msg,"~",FSHOW);
                WrStr(msg);WrLn;
                testit:=FALSE;
                INC(pb);
            END;
        END;
        IF testit THEN
            hin:=fileOpenRead(useLFN,FORG);
            FIO.AssignBuffer(hin,ioBufferIn);
            remaining:=fsize;
            mismatches:=0;
        END;
        LOOP
            IF readline(hlog,S) THEN EXIT;END; (* $######## for CRC32 *)
            IF same(S,closemarker) THEN EXIT; END;
            IF testit THEN
                IF remaining = 0 THEN EXIT; END;
                IF remaining < chunk THEN
                    count:=remaining;
                    remaining:=0;
                ELSE
                    count:=chunk;
                    DEC(remaining,chunk);
                END;
                fpos:=FIO.GetPos(hin);
                crc32:=SegmentFileComputeCRC32(hin,count); (* from current file position *)
                valToHex(crc32,N);
                IF same(S,N)=FALSE THEN
                    INC(mismatches);
                    IF verbose THEN
                        msg:=msgCRCmismatchAt;
                        Str.Subst(msg,"~",FSHOW);
                        valToHex(fpos,S);
                        Str.Subst(msg,"|",S);
                        valToHex(fpos+chunk-1,S);
                        Str.Subst(msg,"|",S);
                        WrStr(msg);WrLn;
                    END;
                    INC(pb);
                END;
            END;
        END;
        IF testit THEN
            fileClose(useLFN,hin);
            IF mismatches = 0 THEN
                msg:=msgOK;
                Str.Subst(msg,"~",FSHOW);
                WrStr(msg);WrLn;
            ELSE
                IF verbose=FALSE THEN
                    msg:=msgCRCmismatch;
                    Str.Subst(msg,"~",FSHOW);
                    WrStr(msg);WrLn;
                END;
            END;
        END;
    END;
    fileClose(useLFN,hlog);
    RETURN pb;
END doCheck;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*

log format :

chunksize (decimal)
[
filename
filesize
$crc32 or $md5
...
]
...

*)

PROCEDURE doBuild (anchor:pFname;useLFN,verbose,unsegmented:BOOLEAN;
                  casemode:casifytype;
                  segmentsize:LONGCARD;method:CARDINAL;log:pathtype);
VAR
    hlog, hin:FIO.File;
    i:CARDINAL;
    FSHOW,FORG:pathtype;
    N:str128;
    msg:str1024; (* oversized *)
    fsize,chunk,remaining,count,crc32:LONGCARD;
    p:pFname;
BEGIN
    hlog:=fileCreate(useLFN,log);
    FIO.AssignBuffer(hlog,ioBufferOut);

    IF unsegmented THEN
        chunk:=MAX(LONGCARD)-safety; (* -1 was enough ! *)
    ELSE
        chunk:=segmentsize * bytesPerKb;
    END;
    valToStr(chunk,N);
    FIO.WrStr(hlog,N);FIO.WrLn(hlog);

    p:=anchor;
    LOOP
        IF p = NIL THEN EXIT;END;
        FIO.WrStr(hlog,openmarker);FIO.WrLn(hlog);

        getStr(FORG,p);
        Str.Copy(FSHOW,FORG);
        casify(casemode,FSHOW);
        FIO.WrStr(hlog,dquote);FIO.WrStr(hlog,FORG);FIO.WrStr(hlog,dquote);FIO.WrLn(hlog);

        msg:='Processing "~"';
        Str.Subst(msg,"~",FSHOW);
        video(msg,TRUE);
        IF verbose THEN Work(cmdInit); END;

        fsize:=fileGetFileSize(useLFN,FORG);
        valToStr(fsize,N);
        FIO.WrStr(hlog,N);FIO.WrLn(hlog);

        FIO.EOF := FALSE; (* safety ! *)
        hin:=fileOpenRead(useLFN,FORG);
        FIO.AssignBuffer(hin,ioBufferIn);

        remaining:=fsize;
        LOOP
            IF verbose THEN Work(cmdSHOW); END;
            IF remaining=0 THEN EXIT; END;
            IF remaining < chunk THEN
                count:=remaining;
                remaining:=0;
            ELSE
                count:=chunk;
                DEC(remaining,chunk);
            END;
            CASE method OF
            | methodCRC32 :
                crc32:=SegmentFileComputeCRC32(hin,count); (* from current file position *)
                valToHex(crc32,N);
            | methodMD5:
                ;
            END;
            FIO.WrStr(hlog,N);FIO.WrLn(hlog);
        END;
        fileClose(useLFN,hin);

        FIO.WrStr(hlog,closemarker);FIO.WrLn(hlog);

        IF verbose THEN Work(cmdStop); END;
        video(msg,FALSE);
        p:=p^.next;
    END;
    fileFlush(useLFN,hlog);
    fileClose(useLFN,hlog);

    msg:='"~" log file has been created.';
    Str.Subst(msg,"~",log);
    WrStr(msg);WrLn;
END doBuild;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minParm  = 1;
    maxParm  = 16;
VAR
    parmcount,i,opt,lastparm : CARDINAL;
    S,R:pathtype;
    parm : ARRAY[minParm..maxParm] OF pathtype;
VAR
    pb,segmentsize : LONGCARD;
    DEBUG,useLFN:BOOLEAN;
    verbose,audiowarn,overwrite,killRO,unsegmented,usefilelist : BOOLEAN;
    methodtype : CARDINAL;
    cmd : (defcmd,build,check);
    casemode : casifytype;
    log      : pathtype;
    count : CARDINAL;
    anchor:pFname;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone; (* very, very important ! *)
    WrLn;                       (* must be here for pretty ulterior display ! *)

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    segmentsize := defaultSegSize;
    cmd         := defcmd;
    verbose     := FALSE;
    audiowarn   := FALSE;
    overwrite   := FALSE;
    killRO      := FALSE;
    unsegmented := FALSE;
    useLFN      := TRUE;
    Str.Copy(log,defaultlog);
    casemode    := defcase;
    methodtype  := methodCRC32;
    usefilelist := FALSE;
    DEBUG       := FALSE;
    lastparm    := minParm-1;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R,"?"+delim+"H"+delim+"HELP"+delim+
                                 "S:"+delim+"SEGMENT:"+delim+
                                 "O"+delim+"OVERWRITE"+delim+
                                 "V"+delim+"VERBOSE"+delim+
                                 "A"+delim+"AUDIO"+delim+
                                 "C"+delim+"CHECK"+delim+
                                 "OO"+delim+
                                 "L:"+delim+"LOG:"+delim+
                                 "B"+delim+"BUILD"+delim+
                                 "N"+delim+"NORMAL"+delim+"CRC"+delim+
                                 "W"+delim+"LOWERCASE"+delim+"LC"+delim+
                                 "U"+delim+"UPPERCASE"+delim+"UC"+delim+
                                 "X"+delim+"LFN"+delim+
                                 "m"+delim+"md5"+delim+ (* //maybeware *)
                                 "@"+delim+
                                 "DEBUG"
                              );
            CASE opt OF
            |  1,  2,  3: abort(errHelp,"");
            |  4,  5:
                IF GetLongCard (R, segmentsize)=FALSE THEN
                    abort(errBadNumber,S);
                END;
                IF ((segmentsize < minSegSize) OR (segmentsize > maxSegSize)) THEN
                    abort(errBadSegSize,S);
                END;
            |  6,  7: overwrite := TRUE;
            |  8,  9: verbose   := TRUE;
            | 10, 11: audiowarn := TRUE;
            | 12, 13: CASE cmd OF
                      | defcmd,check: cmd := check;
                      ELSE
                          abort(errKierkegaard,"");
                      END;
            | 14    : overwrite := TRUE; killRO := TRUE;
            | 15,16 : GetString(S,log);
                IF chkJoker(log) THEN abort(errJoker,log);END;
                IF Str.CharPos(log,colon) # MAX(CARDINAL) THEN abort(errNotLocalLog,log);END;
                IF Str.CharPos(log,antislash) # MAX(CARDINAL) THEN abort(errNotLocalLog,log);END;
                IF Str.CharPos(log,dot)=MAX(CARDINAL) THEN Str.Append(log,extLOG);END;
            | 17,18 : CASE cmd OF
                      | defcmd,build : cmd:=build;
                      ELSE
                          abort(errKierkegaard,"");
                      END;
            | 19,20,21:unsegmented:=TRUE;
            | 22,23,24 : CASE casemode OF
                         | defcase,lowercase:casemode:=lowercase;
                         ELSE
                             abort(errCasemode,"");
                         END;
            | 25,26,27 : CASE casemode OF
                         | defcase,uppercase:casemode:=uppercase;
                         ELSE
                             abort(errCasemode,"");
                         END;
            | 28,29 : useLFN    := FALSE;
            | 30,31 : methodtype:= methodMD5;
            | 32:     usefilelist:= TRUE;
            | 33 :    DEBUG     := TRUE;
            ELSE
                abort(errUnknownOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxParm THEN abort(errTooManyParms,S);END;
            IF same(S,dot) THEN Str.Copy(S,stardotstar);END;
            parm[lastparm]:=S;
        END;
    END;

    useLFN := ( useLFN AND fileSupportLFN() );

    IF NOT(useLFN) THEN
        i:=Str.RCharPos(log,dot); (* dot always here (see supra) *)
        IF (i+1) > maxDOSf8 THEN abort(errF8E3,log);END;
        IF Str.Length(log)-(i+1)-1 > maxDOSe3 THEN abort(errF8E3,log);END;
    END;

    IF cmd = defcmd THEN cmd:=build;END;

    CASE cmd OF
    | check :
        IF lastparm >= minParm THEN abort(errKierkegaard,""); END;
        IF fileExists(useLFN,log)=FALSE THEN abort(errLogNotFound,log);END;

        WrStr(Banner);WrLn;
        WrLn;
        WrStr('Processing "');WrStr(log);WrStr('" log file...');WrLn;
        WrLn;
        pb:=doCheck(casemode,useLFN,verbose,audiowarn,log);
        IF pb # 0 THEN
            IF audiowarn THEN alert; END;
        END;
    | build :
        IF lastparm < minParm THEN abort(errMissingSpec,"");END;

        IF parm[minParm][0] = "@" THEN
            Str.Delete(parm[minParm],0,1);
            usefilelist:=TRUE;
        END;

        IF usefilelist THEN
            IF lastparm # minParm THEN abort(errOneList,"");END;
            S:=parm[minParm];
            IF chkJoker(S) THEN abort(errJokerFileList,S);END;
            IF fileExists(useLFN,S)=FALSE THEN abort(errFileListNotFound,S);END;
        END;

        IF fileExists(useLFN,log) THEN
            IF overwrite THEN
                IF fileIsRO(useLFN,log) THEN
                    IF killRO THEN
                        fileSetRW(useLFN,log);
                    ELSE
                        abort(errROlogHere,log);
                    END;
                END;
            ELSE
                abort(errLogHere,log);
            END;
        END;

        FOR i:=minParm TO lastparm DO
            Str.Copy(S,parm[i]);
            IF Str.CharPos(S,colon) # MAX(CARDINAL) THEN abort(errNotLocal,S);END;
            IF Str.CharPos(S,antislash) # MAX(CARDINAL) THEN abort(errNotLocal,S);END;
        END;

        initList(count,anchor);

        (* here, if usefilelist, lastparm is minParm and filelist exists *)

        FOR i:=minParm TO lastparm DO
            Str.Copy(S,parm[i]);
            IF usefilelist THEN
                fakeBuildFileList (count,anchor,
                                  DEBUG,useLFN,S,skippedpats);
            ELSE
                buildFileList (count,anchor,
                              DEBUG,useLFN,S,skippedpats);
            END;
            CASE count OF
            | PBSTORAGE:
                freeList (anchor);
                abort(errStorage,"");
            | PBCOUNT:
                freeList (anchor);
                abort(errCountOverflow,"");
            END;
        END;
        IF count < MINENTRY THEN
            freeList(anchor);
            IF usefilelist THEN
                abort(errEmptyFileList,parm[minParm]);
            ELSE
                CASE lastparm OF
                | minParm :
                    abort(errNoMatch,parm[minParm]);
                ELSE
                    abort(errReallyNoMatch,"");
                END;
            END;
        END;
        IF DEBUG THEN dmpList(anchor); END;

        WrStr(Banner);WrLn;
        WrLn;
        doBuild (anchor,useLFN,verbose,unsegmented,
                casemode,
                segmentsize,methodtype,log);
        IF audiowarn THEN alarm; END; (* done *)

    END;
    freeList(anchor);
    abort(errNone,"");
END segCRC.

