(* -----------------------------------------------------------
Title         Q&D Unpath
Author        PhG
Overview      see help
Notes
Bugs
Wish List     LFN ? are you kidding ? ;-)

------------------------------------------------------------ *)

MODULE unpath;

IMPORT Lib;
IMPORT Str;
IMPORT FIO;
IMPORT IO;

FROM IO IMPORT WrStr, WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

(* ------------------------------------------------------------ *)

CONST
    miniwi     = 1;
    maxiwi     = 512; (* should do *)
CONST
    defSEP     = " +++ "; (* yes, we know we should take first space into account for padwi *)
    TOOLONG    = "::: ";
    REM1       = ';';
    REM2       = '#';
    AUTOWIDTH  = '*';

CONST
    ProgEXEname   = "UNPATH";
    ProgTitle     = "Q&D Unpath";
    ProgVersion   = "v1.2";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    errNone             = 0;
    errHelp             = 1;
    errWidth            = 2;
    errNotFound         = 3;
    errOption           = 4;
    errParmOverflow     = 5;
    errSyntax           = 6;
    errJoker            = 7;
    errSepTooLong       = 8;
    errBadWidth         = 9;
    errAborted          = 255;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msgHelp =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <"+AUTOWIDTH+"|width> <file> [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" <file> [option]..."+nl+
nl+
"With syntax 1, this program reformats each line in <file> :"+nl+
"it extracts filename from trailing path, pads it to <width>,"+nl+
'then prepends it to original path portion with separator (default is "'+defSEP+'").'+nl+
"With syntax 2, it rebuilds original path from <file> created with redirection."+nl+
nl+
"    -s:$ user-defined separator (up to 16 characters)"+nl+
"    -l   pad filename left (defaut is right)"+nl+
'    -k   keep remarks (lines beginning with either "'+REM1+'" or "'+REM2+'")'+nl+
'    -t   remove trailing "'+antislash+'" from path part'+nl+
"    -d   remove double quotes"+nl+
nl+
'a) Except in remarks, lines not matching "*\*" pattern are filtered out.'+nl+
'b) If first parameter is "'+AUTOWIDTH+'", program will use longest length.'+nl+
"c) If filename is longer than specified <width>,"+nl+
'   original line is dumped with "'+TOOLONG+'" prefix.'+nl+
"d) If original path was delimited with double quotes, split parts will be too."+nl+
"   Note program expects lines to be enclosed by two double quotes or none."+nl+
"e) Extensions are not beautified."+nl+
"f) Each line entries should not be longer than 1024 characters."+nl+
"g) Rebuilding requires the same separator used to reformat."+nl+
"h) LFNs are not supported : use regular 8+3 DOS filename instead."+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :      WrStr(msgHelp);
    | errWidth :     S := 'Illegal or out of [1..512] range "|" <width> !';
    | errNotFound:   S := '"|" <file> does not exist !';
    | errOption:     S := 'Unknown "|" option !';
    | errParmOverflow:S:= '"|" parameter is one too many !';
    | errSyntax:     S := "Syntax error, please check help screen !";
    | errJoker:      S := '"|" <file> should not contain any joker !';
    | errSepTooLong: S := '"|" separator cannot be more than 16 characters long !';
    | errBadWidth:   S := '"|" is not a valid <width> !';
    | errAborted:    S := "Aborted by user !";
    ELSE
                     S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp : ;
    ELSE
        Str.Subst(S,"|",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBufferIn : ARRAY[firstBufferByte..lastBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)

(* "#", "+#" *)

PROCEDURE isnum (S:ARRAY OF CHAR):BOOLEAN ;
CONST
    minus = "-";
    plus  = "+";
BEGIN
    CASE S[0] OF
    | plus : Str.Delete(S,0,1);
    END;
    RETURN verifyString (S,digits);
END isnum;

PROCEDURE evalnum (VAR v:CARDINAL;mini,maxi:CARDINAL;S:ARRAY OF CHAR):BOOLEAN;
VAR
    li:LONGCARD;
    ok:BOOLEAN;
BEGIN
    ok:=TRUE;
    li:=Str.StrToCard(S,10,ok);
    IF ok THEN
        IF li > LONGCARD( maxi ) THEN
            ok:=FALSE;
        ELSE
            v:=CARDINAL(li);
            IF v < mini THEN ok:=FALSE;END;
        END;
    END;
    RETURN ok;
END evalnum;

PROCEDURE dbgbool (S:ARRAY OF CHAR;tf:BOOLEAN);
BEGIN
    WrStr(S);
    IF tf THEN
        WrStr("yes");
    ELSE
        WrStr("no");
    END;
    WrLn;
END dbgbool;

PROCEDURE pad (VAR R:ARRAY OF CHAR;padright:BOOLEAN;
              padwi:CARDINAL;padchar:CHAR);
VAR
    i,len:CARDINAL;
BEGIN
    len:=Str.Length(R);
    FOR i:= (len+1) TO padwi DO
        IF padright THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
END pad;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    Z,RZ : str1024; (* global *)

PROCEDURE procSplit (padwi:CARDINAL;F:str128;SEP:str16;
                    DEBUG,padright,
                    keepremarks,keepfinalantislash,keepdquotes:BOOLEAN  ):BOOLEAN;
CONST
    msgFindLongest = "Looking for longest length, please wait...";
    msgProcessing  = "Processing, please wait...";
VAR
    hin:FIO.File;
    alcatraz:BOOLEAN;
    len,p,pass,firstpass:CARDINAL;
    warn:BOOLEAN;
    msg:str128;
    wasremark,wasenclosed:BOOLEAN;
BEGIN
    IF padwi = 0 THEN
        firstpass:=1;
    ELSE
        firstpass:=2;
    END;
    FOR pass:=firstpass TO 2 DO
        CASE pass OF
        | 1: msg:=msgFindLongest;
        | 2: msg:=msgProcessing;
        END;
        warn:= (IsRedirected() OR (pass=1));
        IF warn THEN video(msg,TRUE); END;
        FIO.EOF := FALSE; (* required safety -- else, pass 2 would not run *)
        hin:=FIO.OpenRead(F);
        FIO.AssignBuffer(hin,ioBufferIn);
        (* WHILE NOT (FIO.EOF) DO -- we'll need to force exit -- see infra *)
        LOOP
            IF FIO.EOF THEN EXIT; END;
            FIO.RdStr(hin,Z);
            (* better safe than sorry *)
            LtrimBlanks(Z);
            RtrimBlanks(Z);
IF DEBUG THEN WrStr("/// ");WrStr(Z);WrLn; END;
            CASE Z[0] OF
            | REM1,REM2:
                wasremark := TRUE;
            ELSE
                wasremark := FALSE;
            END;

            IF wasremark THEN
                IF pass=2 THEN
                    IF keepremarks THEN WrStr(Z);WrLn; END;
                END;
            ELSE
                p:=Str.RCharPos(Z,antislash); (* "'p\f'" *)
                IF p # MAX(CARDINAL) THEN
                    wasenclosed := Str.Match(Z,dquote+"*"+dquote);
                    IF wasenclosed THEN
                        Str.Delete(Z,0,1);
                        Z[Str.Length(Z)-1]:=charnull;
                        DEC(p);
                    END;
IF DEBUG THEN WrStr("!!! ");WrStr(Z);WrLn;END;
                    Str.Copy(RZ,Z);
                    Str.Delete(RZ,0,p+1); (* "f"  delete antislash too *)

                    IF keepfinalantislash THEN
                        Z[p+1]:=charnull;    (* "p\" keep antislash -- brutal, huh ?! ;-) *)
                    ELSE
                        Z[p]:=charnull;
                    END;

                    len:= Str.Length(RZ);

                    CASE pass OF
                    | 1:
                        IF wasenclosed THEN INC(len,2);END;
                        IF len > padwi THEN padwi:=len;END;
                    | 2:
                        IF len > padwi THEN
                            WrStr(TOOLONG);
                            IF wasenclosed THEN WrStr(dquote);END;
                            WrStr(Z);WrStr(RZ); (* "p\" "f" *)
                            IF wasenclosed THEN WrStr(dquote);END;
                            WrLn;
                        ELSE
                            IF (wasenclosed AND keepdquotes) THEN
                                (*
                                IF padwi < 0 THEN
                                    thiswi:=padwi-2;
                                ELSE
                                    thiswi:=padwi+2;
                                END;
                                *)
                                Str.Append(RZ,dquote); Str.Prepend(RZ,dquote);
                                Str.Append(Z,dquote); Str.Prepend(Z,dquote);
                            END;
                            pad(RZ,padright, padwi,blank);
                            WrStr(RZ);WrStr(SEP);WrStr(Z);WrLn;
                        END;
                    END;
                END;
            END;

            alcatraz:= ChkEscape();
            IF alcatraz THEN EXIT; END;
        END;
        FIO.Close(hin);
        IF warn THEN video(msg,FALSE); END;
    END;
    RETURN alcatraz;
END procSplit;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procRebuild (F:str128;SEP:str16;
                      DEBUG,keepremarks,keepdquotes:BOOLEAN   ):BOOLEAN;
CONST
    msgProcessing  = "Processing, please wait...";
VAR
    hin:FIO.File;
    alcatraz:BOOLEAN;
    warn:BOOLEAN;
    msg:str128;
    dquotecount,p,seplen:CARDINAL;
BEGIN
    seplen:=Str.Length(SEP);
    msg:=msgProcessing;
    warn:= IsRedirected();
    IF warn THEN video(msg,TRUE); END;
    FIO.EOF := FALSE; (* required safety -- else, pass 2 would not run *)
    hin:=FIO.OpenRead(F);
    FIO.AssignBuffer(hin,ioBufferIn);
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,Z);
        LtrimBlanks(Z);
        RtrimBlanks(Z);
IF DEBUG THEN WrStr("/// ");WrStr(Z);WrLn; END;
        p:=Str.Pos(Z,SEP);
        IF p = MAX(CARDINAL) THEN
            CASE Z[0] OF
            | REM1,REM2:
                IF keepremarks THEN WrStr(Z);END;
            END;
        ELSE
            dquotecount:=CharCount(Z,dquote); (* Q&D ! *)
            Str.Copy(RZ,Z);
            Z[p]:=charnull;             (* filename *)
            Str.Delete(RZ,0,p+seplen);  (* path *)
            LtrimBlanks(Z);
            RtrimBlanks(Z);
            LtrimBlanks(RZ);
            RtrimBlanks(RZ);

            ReplaceChar(Z,dquote,"");
            ReplaceChar(RZ,dquote,"");

            IF NOT (Str.Match(RZ,star+antislash)) THEN Str.Append(RZ,antislash);END;
            IF dquotecount = 4 THEN
                IF keepdquotes THEN
                    Str.Prepend(RZ,dquote);
                    Str.Append(Z,dquote);
                END;
            END;
            WrStr(RZ);WrStr(Z);
        END;
        WrLn;

        alcatraz:= ChkEscape();
        IF alcatraz THEN EXIT; END;
    END;
    FIO.Close(hin);
    IF warn THEN video(msg,FALSE); END;
    RETURN alcatraz;
END procRebuild;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 2;
VAR
    lastparm,i,opt,parmcount:CARDINAL;
    parm : ARRAY [minparm..maxparm] OF str128;
    S,R,F,msg:str128;
    DEBUG,keepremarks,keepfinalantislash,keepdquotes:BOOLEAN;
    SEP : str16;
    padmode : (padright,padleft);
    cmd : (cmdsplit,cmdrebuild);
    padwi:CARDINAL;
    alcatraz:BOOLEAN;
BEGIN
    FIO.IOcheck:=FALSE;
    WrLn;

    Str.Copy(SEP,defSEP);
    keepremarks         := FALSE;
    keepfinalantislash  := TRUE;
    keepdquotes         := TRUE;
    padmode             := padright;
    DEBUG               := FALSE;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    lastparm := minparm-1;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "K"+delim+"KEEP"+delim+
                                  "T"+delim+"ANTISLASH"+delim+
                                  "D"+delim+"DOUBLEQUOTES"+delim+
                                  "S:"+delim+"SEPARATOR:"+delim+
                                  "L"+delim+"PADLEFT"+delim+
                                  "DEBUG"
                                  );
            CASE opt OF
            | 1,2,3 :  abort(errHelp,"");
            | 4,5   :  keepremarks        := TRUE;
            | 6,7   :  keepfinalantislash := FALSE;
            | 8,9   :  keepdquotes        := FALSE;
            | 10,11 :  GetString(S,R);
                       IF Str.Length(R)>SIZE(SEP) THEN abort(errSepTooLong,S);END;
                       Str.Copy(SEP,R);
            | 12,13:   padmode            := padleft;
            | 14:      DEBUG              := TRUE;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy(parm[lastparm],S);
        END;
    END;

    CASE lastparm OF
    | minparm-1 :
        abort(errHelp,"");
    | minparm :
        Str.Copy(F,parm[minparm]);
        cmd:=cmdrebuild;
    | minparm+1 :
        Str.Copy(S,parm[minparm]);
        IF same(S,AUTOWIDTH) THEN
            padwi:=0;
        ELSIF isnum(S) THEN
            IF evalnum(padwi,miniwi,maxiwi,S)=FALSE THEN abort(errWidth,S);END;
        ELSE
            abort(errBadWidth,S);
        END;
        Str.Copy(F,parm[minparm+1]);
        cmd:=cmdsplit;
    END;

    IF chkJoker(F) THEN abort(errJoker,F);END;
    IF FIO.Exists(F)=FALSE THEN abort(errNotFound,F);END;

    IF DEBUG THEN
        WrStr("DOS file     : ");WrStr(F);WrLn;
        WrStr('separator    : "');WrStr(SEP);WrStr(dquote);WrLn;
        WrStr("command      : ");
        CASE cmd OF
        | cmdsplit:   S:="split";
        | cmdrebuild: S:="rebuild";
        END;
        WrStr(S);WrLn;
        WrStr("width        : ");
        IF padwi = 0 THEN
            WrStr("longest");
        ELSE
            IO.WrCard(padwi,1);
        END;
        WrLn;
        WrStr("padmode      : ");
        CASE padmode OF
        | padright:   S:="right";
        | padleft:    S:="left";
        END;
        WrStr(S);WrLn;
        dbgbool("keep remarks : ",keepremarks);
        dbgbool("keep final \ : ",keepfinalantislash);
        dbgbool("keep quotes  : ",keepdquotes);
    END;

    CASE cmd OF
    | cmdsplit:
        alcatraz:=procSplit (padwi,F,SEP,
                             DEBUG,(padmode=padright),
                             keepremarks,keepfinalantislash,keepdquotes);
    | cmdrebuild:
        alcatraz:=procRebuild(F,SEP,DEBUG,keepremarks,keepdquotes);
    END;
    IF alcatraz THEN abort(errAborted,"");END;
    abort(errNone,"");
END unpath.

(*
//FIC

 this will not be processed because of missing slash
 ; this is a remark ignored except if specified
 # this is a remark ignored except if specified
 "C:\WINDOWS\SYSTEM\DRIFTCLK.SCR"
 C:\WINDOWS\SYSTEM\Canalisations 3D.scr
 "C:\WINDOWS\SYSTEM\Courbes et couleurs.scr"
 C:\WINDOWS\SYSTEM\Skyrocket.scr
 "C:\WINDOWS\SYSTEM\Lens32.scr"
 "C:\WINDOWS\SYSTEM\Matrix.Scr"
 "C:\WINDOWS\SYSTEM\spirals.scr"
 C:\WINDOWS\SYSTEM\toiles.scr
 C:\WINDOWS\SYSTEM\Bote florissante 3D.scr
 C:\WINDOWS\SYSTEM\LIFE.SCR
 C:\WINDOWS\SYSTEM\HomePlanet.scr

*)

(*
//TEST

@echo off
unpath * fic
pause
unpath * fic /l
pause
unpath 22 fic
pause
unpath 22 fic /l
pause

unpath -s:" !___! " 33 fic
pause
unpath /k * fic
pause
unpath /t * fic
pause
unpath /d * fic
pause
unpath /t /d * fic
pause

unpath * fic > fic2
unpath fic2
unpath /d fic2

*)

