(* ---------------------------------------------------------------
Title         Q&D Life
Author        PhG
Overview      Yet Another Version Of Conway's Life
Usage         see help
Notes         COMPACT model or more to avoid 64 Kb overflow
              assume normal color palette
              at v1.2, removing a few globerks forced THE weird "isl 342" error :
              model changed from COMPACT to XLARGE so program would compile !

              minimal check for VLIFE.DAT and *.LIF files
              .LIF files grabbed from dozens of LIFE programs...
              no one had ever thought of randomness in births, eh eh !

              note -n:$ option allows use of Life Lexicon patterns
              (convert text file with LIFELEX utility)
Bugs          M2 bug when cellsize=3 for box and fillmasktype is useless
Wish List     detect static condition
              wrap at init ?
              (hash code of successive states of array ?
              even better : some kind OF unfocus on grid as FOR picture comparison)

--------------------------------------------------------------- *)

MODULE vLife;

IMPORT Lib;
IMPORT MsMouse;
IMPORT Str;
IMPORT Graph;
IMPORT IO;
IMPORT SYSTEM;
IMPORT FIO;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_rand IMPORT InitRnd, GetRnd, GetRndCardRange;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM IO IMPORT WrStr,WrLn;

(* ------------------------------------------------------------ *)

CONST
    FROZEN = TRUE; (* disable any DEBUG if TRUE *)
CONST
    gmode       = Graph._MRES256COLOR; (* 320x200 *)
    colorEmpty  = 0;
    minproba    = 1;
    maxproba    = 100;
    minsize     = 0;
    maxsize     = 32;
    mindelay    = 0;
    defaultinitialdelay = 5000;
    maxdelayinitial     = 30000;
    maxdelayrefresh     = 5000;
    minspeed    = 0;
    maxspeed    = 50;
    defaultspeed= minspeed+1;
    defaultBirthProba        = 10;
    defaultRuleRNDbirthProba = 70;
    defaultCellSize          = 2;

    nopattern   = 0;

    firstColor  = 20H; (* with standard palette, only this range is useful *)
    lastColor   = 67H;
    pausevalue  = 2000; (* in milliseconds, before restart *)
    maxWait     = 15; (* seconds *)
    sMaxWait    = "15";
    maxWaitMS   = 1000*maxWait;
    sVLPIC      = "VL_"; (* ??65535 *)
    extPCX      = ".PCX";
    defaultmaxpicndx = 100;
CONST
    cBLACK	       =0;
    cBLUE 	       =1;
    cGREEN         =2;
    cCYAN          =3;
    cRED           =4;
    cMAGENTA       =5;
    cBROWN         =6;
    cWHITE         =7;
    cGRAY          =8;
    cLIGHTBLUE     =9;
    cLIGHTGREEN    =10;
    cLIGHTCYAN     =11;
    cLIGHTRED      =12;
    cLIGHTMAGENTA  =13;
    cLIGHTYELLOW   =14;
    cBRIGHTWHITE   =15;
    minink         = cBLUE;
    maxink         = cBRIGHTWHITE;
CONST
    empty=000H;
    alive=080H;
CONST
    screenX         = 320;
    screenY         = 200;
    xmin            = 0;
    xmax            = screenX-1;
    ymin            = 0;
    ymax            = screenY-1;
    charsize        = 8;
    Htabmin         = 1;
    Htabmax         = screenX DIV charsize; (* 40 *)
    vtabmin         = 1;
    vtabmax         = screenY DIV charsize; (* 25 *)
CONST
    firstX = 0;
    firstY = 0;
(*
    maxLastX = (screenX DIV (minsize+1))-1;
    maxLastY = (screenY DIV (minsize+1))-1;
*)
    maxLastX = xmax;
    maxLastY = ymax;
VAR
    (* globerks *)
    cell         : ARRAY [firstX..maxLastX],[firstY..maxLastY] OF SHORTCARD; (* BYTE *)
VAR
    lastX            : CARDINAL;
    lastY            : CARDINAL;
    currentColor     : CARDINAL;
CONST
    escape      = CHR(27);
    special     = CHR(0);
    extEXE      = ".EXE"; (* upper case *)
    extDAT      = ".DAT"; (* idem *)
    extLIF      = ".LIF";
CONST
    defaultConwayRule = "23/3";
    LIFid             = pound+"LIFE"; (* "Life" + version *)
    LIFrem1           = pound+"D";
    LIFrem2           = pound+"C";    (* nix form *)
    LIFrule           = pound+"R";
    LIFdefrule        = pound+"N";    (* 23/3 i.e. Conway's default *)
    LIFpos            = pound+"P";
    (* LIFmaxrem         = 22; *)
CONST
    ProgEXEname   = "VLIFE";
    ProgTitle     = "Q&D Life 320x200x256";
    ProgVersion   = "v1.2c";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
    ProgBANK      = ProgEXEname+extDAT;
CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParameter    = 3;
    errConversion   = 4;
    errRange        = 5;
    errRule         = 6;
    errVGAneeded    = 7;
    errMouseNeeded  = 8;
    errCannotSet    = 9;
    errCannotReset  = 10;
    errNotFound     = 11;
    errLIFrule      = 12;
    errMoreHelp     = 13;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
00000000011111111112222222222333333333344444444445555555555666666666677777777778
1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msghelp = Banner+nl+
nl+
"Syntax : "+ProgEXEname+" [option]..."+nl+
nl+
"  m:#     mouse (0=no, 1=yes, default is 0)"+nl+
"  w:#     wrap (0=no, 1=yes, default is 1)"+nl+
"  z:#     speed ([0..50], default is 1)"+nl+
"  p:#     birth probability ([1..100], default is 10)"+nl+
"  u:#     rule random birth probability ([1..100], default is 70)"+nl+
"  c:#     cycle colors (0=yes, else [1..15], default is 0)"+nl+
"  s:#     cell size (0 or [1..16], default is 2)"+nl+
"  b:#     box (0=no, 1=yes, default is 1)"+nl+
"  s[s]    show infos while in view mode (-ss = with S/B/? reminder)"+nl+
"  t:#     initial delay in milliseconds ([0..30000], default is 5000)"+nl+
"  d:#     refresh delay in milliseconds ([0..5000], default is 0)"+nl+
"  g:#     last generation before restart (0=unlimited, default is 10000)"+nl+
"  i:#     initial pattern found in "+ProgBANK+" (0=random, default is 0)"+nl+
"  r:$     rule (Survival/Birth[/RandomBirth] : Conway's is 23/3)"+nl+
"  f:$     initial pattern found in specified file (text "+extLIF+" format)"+nl+
"  n:$     alternate databank (default is "+ProgBANK+", default extension is "+extDAT+")"+nl+
"  c       ignore "+dquote+LIFpos+dquote+" command (-f:$ and single pattern only)"+nl+
"  r       ignore any of "+dquote+LIFrule+dquote+" or "+dquote+LIFdefrule+dquote+" commands (-f:$ only)"+nl+
"  i       show current parameters and wait until keypress or "+sMaxWait+" seconds"+nl+
"  l       show pertinent infos then terminate"+nl+
"          (list of available patterns, or -f:$ specified file parameters)"+nl+
"  pcx[:#] save up to # "+sVLPIC+"* PCX steps (default is 100)"+nl+
"  x       disable LFN support even if available"+nl+
"  ?[?]    (more) help"+nl+
nl+
'a) An option can be specified with or without "/" or "-".'+nl+
"b) Only the last of repeated options is taken into account."+nl+
"c) With -l option, any other option is ignored."+nl+
"d) With -i:# option, "+ProgBANK+" is first searched for in current directory,"+nl+
"   then in executable directory."+nl+
"e) With -n:$ option, -f:$ option is ignored."+nl+
"f) With -f:$ option, -i:# option is ignored ;"+nl+
'   any of "'+LIFrule+'" or "'+LIFdefrule+'" commands override default and -r:$ rules (unless -r) ;'+nl+
'   "'+extLIF+'" file extension is assumed if omitted ;'+nl+
"   -s:0 option is recommended ; initial positions out of screen are ignored."+nl+
"g) Supported aliases are -ff[f]:$ = -f:$ -i [-s:1] ; -ii[i]:$ = -i:# -i [-s:1]."+nl+
"h) Display restarts at last generation or if no change has been detected."+nl+
"i) Here are a few interesting rules :"+nl+
"   /2  01245678/34  245/368  1358/357  125/36  23/36  238/357  235678/378"+nl+
"   125678/367  45678/3  /234  12345/3  235678/3678  5678/35678"+nl+
"j) -z:# option uses vertical retrace, while -d:# option uses a timer."+nl+
"k) Note delays may not be accurate when program is not run from real DOS."+nl+
"l) [SPACE]-step  [-+]-speed  [TAB]-color  [s]-infos  [$]-reset  [*]-init"+nl;

    msgmorehelp = nl+
"File format example 1 :"+nl+
nl+
"   #Life 1.05"+nl+
"   ; rule is specified as survival/birth, #N is an alias for #R 23/3"+nl+
"   #R /234"+nl+
"   **"+nl+
"   **"+nl+
nl+
"File format example 2 :"+nl+
nl+
"   #D This is a comment"+nl+
"   #C This is another comment"+nl+
"   #N"+nl+
"   ; x,y position is relative to screen center"+nl+
"   #P -2 -1"+nl+
"   *...*"+nl+
"   *.*.*"+nl+
"   *...*"+nl+
nl+
"File format example 3 :"+nl+
nl+
"   #R /2"+nl+
"   -11 1"+nl+
"   -10 0"+nl+
"   -13 0"+nl+
"   -8 -1"+nl+
"   -8 -2"+nl+
"   -12 -1"+nl+
"   -9 -3"+nl;

VAR
    S : str256; (* oversized in order to account for possible pathtype *)
BEGIN
    CASE e OF
    | errHelp, errMoreHelp :
        WrStr(msghelp);
        IF e = errMoreHelp THEN
            WrStr(msgmorehelp);
            e:=errHelp;
        END;
    | errOption :
        Str.Concat(S,"Unknown ",einfo); Str.Append(S," option");
    | errParameter :
        Str.Concat(S,"Unknown ",einfo); Str.Append(S," parameter");
    | errConversion :
        Str.Concat(S,"StrToCard would not convert ",einfo);
    | errRange :
        Str.Concat(S,"Illegal range for ",einfo);
    | errRule :
        Str.Concat(S,"Illegal rule in ",einfo);
    | errVGAneeded :
        S := "VGA card required";
    | errMouseNeeded :
        S := "Mouse hardware and software required";
    | errCannotSet :
        S := "Cannot set 320x200x256 video mode";
    | errCannotReset :
        S := "Cannot reset original video mode";
    | errNotFound :
        Str.Concat(S,"File "+dquote,einfo); Str.Append(S,dquote+" not found");
    | errLIFrule:
        Str.Concat(S,"Illegal "+dquote,einfo);Str.Append(S,dquote+" "+extLIF+" rule");
    ELSE
        S := "How Can Such A Thing B(i)e(rce) ???";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        WrStr(ProgEXEname+" : "); WrStr(S); WrStr(" !"); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

PROCEDURE dbg (DEBUG:BOOLEAN;S1,S2:ARRAY OF CHAR   );
BEGIN
(*%F FROZEN  *)
    IF DEBUG THEN
        WrStr("// ");WrStr(S1);WrStr(" = ");WrStr(S2);WrLn;
    END;
(*%E  *)
END dbg;

PROCEDURE dbgval (DEBUG:BOOLEAN;v:INTEGER;S1:ARRAY OF CHAR);
VAR
    S:str16;
    ok:BOOLEAN;
BEGIN
(*%F FROZEN  *)
    IF DEBUG THEN
        Str.IntToStr( LONGINT(v),S,10,ok);
        WrStr("// ");WrStr(S1);WrStr(" = ");WrStr(S);WrLn;
    END;
(*%E  *)
END dbgval;

PROCEDURE dbgbool ( DEBUG:BOOLEAN;tf:BOOLEAN;S1:ARRAY OF CHAR  );
BEGIN
(*%F FROZEN  *)
    IF DEBUG THEN
        WrStr("// ");WrStr(S1);WrStr(" = ");
        IF tf THEN
            WrStr("TRUE");
        ELSE
            WrStr("FALSE");
        END;
        WrLn;
    END;
(*%E  *)
END dbgbool;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType  = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    bufferIn : ioBufferType;
    bufferOut: ioBufferType;

(* ------------------------------------------------------------ *)

PROCEDURE WaitVGAretrace ();
BEGIN
    WHILE (SYSTEM.In(03DAH) AND 08H) # 0 DO
    END;
    WHILE (SYSTEM.In(03DAH) AND 08H) = 0 DO
    END;
END WaitVGAretrace;

PROCEDURE Set(x,y,color,cellsize : CARDINAL;box:BOOLEAN);
VAR
    xpos,ypos,r:CARDINAL;
BEGIN
    IF cellsize=0 THEN Graph.Plot(x,y,color); RETURN; END;

    IF cellsize=1 THEN (* we have to because of crash if we let graphics library handle this size ! *)
        Graph.Plot(x<<1,y<<1,color); (* supposedly faster than *2 *)
        RETURN;
    END;

    IF box THEN
        xpos := x * (cellsize +1);
        ypos := y * (cellsize +1);
        Graph.Rectangle(xpos,ypos,xpos+cellsize-1,ypos+cellsize-1,color,TRUE);
        IF cellsize=3 THEN (* fix TopSpeed bug for cellsize=3 !!! *)
            Graph.Plot(xpos+1,ypos+1,color);
        END;
    ELSE
        r := (cellsize-1) DIV 2;
        xpos := x * (cellsize +1) + r;
        ypos := y * (cellsize +1) + r;
        Graph.Disc(xpos,ypos,r,color);
        IF r=1 THEN (* another TopSpeed bug for cellsize=3 or 4 !!! *)
            Graph.Plot(xpos,ypos,color);
        END;
    END;
END Set;

PROCEDURE hiresCLS (  );
BEGIN
    Graph.ClearScreen(Graph._GCLEARSCREEN);
END hiresCLS;

(* ------------------------------------------------------------ *)

PROCEDURE MakeAlive (x,y:CARDINAL);
BEGIN
    cell[x,y]:=alive;
END MakeAlive;

PROCEDURE MakeEmpty (x,y:CARDINAL);
BEGIN
    cell[x,y]:=empty;
END MakeEmpty;

PROCEDURE isAlive (x,y:CARDINAL) : BOOLEAN;
BEGIN
    IF cell[x,y] < alive THEN
        RETURN FALSE;
    ELSE
        RETURN TRUE;
    END;
END isAlive;

PROCEDURE Add (x,y:CARDINAL);
BEGIN
    INC (cell[x,y]);
END Add;

PROCEDURE wasEmpty (x,y:CARDINAL) : BOOLEAN;
BEGIN
    IF cell[x,y] < alive THEN
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END wasEmpty;

PROCEDURE GetNeighbours (x,y:CARDINAL  ) : SHORTCARD;
VAR
    v : SHORTCARD;
BEGIN
    v := cell[x,y];
    IF cell[x,y] < alive THEN
        RETURN v;
    ELSE
        RETURN (v-alive);
    END;
END GetNeighbours;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE vpoint (x,y:CARDINAL):SHORTCARD;
BEGIN
    RETURN SHORTCARD( Graph.Point(x,y) );
END vpoint;

PROCEDURE getDAC (index:CARDINAL;VAR r,g,b:SHORTCARD);
CONST
    DACReadIndex   = 03C7H;
    DACDataRegister= 03C9H;
BEGIN
    SYSTEM.Out (DACReadIndex,SHORTCARD(index));
    r:=SYSTEM.In (DACDataRegister);
    g:=SYSTEM.In (DACDataRegister);
    b:=SYSTEM.In (DACDataRegister);
END getDAC;


CONST
    maxPlanes = 4;
    maxpixels = 1280;
    firstcode = 0;
    (* maxcode   = (1280 DIV 8) * maxPlanes -1; (* in bytes *) *)
    maxcode   = maxpixels * maxPlanes-1;
TYPE
    imgtype = RECORD
        picwidth : CARDINAL ;
        picheight: CARDINAL ;
        decodebuffer : ARRAY [firstcode..maxcode] OF BYTE;
    END;
VAR
    img : imgtype;


TYPE
    PCXrgbType = RECORD
        red,green,blue:BYTE;
    END;

    (* (almost) useless fields are indented again *)

    PCXheaderType = RECORD
     manufacturer    : BYTE; (* constant 10=$0A=ZSoft *)
     version         : BYTE; (* constant 5=PC Paintbrush v3.0+ *)
     encoding        : BYTE; (* constant 1=RLE *)
    bitsPerPixel     : BYTE; (* 1, 2, 4 or 8, i.e. mono, 4, 16 OR 256 colors  *)
	                  (* window, i.e. image dimensions : xmin, ymin, xmax, ymax *)
    leftmargin       : WORD;
    uppermargin      : WORD;
    rightmargin      : WORD; (* do not use xmax-xmin ! *)
    lowermargin      : WORD;
     hDPIresolution  : WORD; (* put 300 for instance *)
     vDPIresolution  : WORD; (* idem *)
    colormap         : ARRAY[0..15] OF PCXrgbType;
     reserved1       : BYTE; (* must be 0 *)
    NCP              : BYTE; (* number of color planes, 1 or 4 *)
    NBS              : WORD; (* number of bytes per scanline, always even *)
                            (* from here, fields may contain rubbish *)
     paletteInfo     : WORD; (* 1=color or BW, 2=grayscale *)
     hscreensize     : WORD;
     vscreensize     : WORD;
     reserved2       : ARRAY[0..53] OF BYTE; (* must be 0 *)
    END;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE buildpicname (S:ARRAY OF CHAR;VAR num:CARDINAL; VAR pic:ARRAY OF CHAR);
CONST
    F8maxlen = 8;
    wipicndx = 5; (* 65535 *)
VAR
    N:str128;
    v:LONGCARD;
    ok:BOOLEAN;
    p,len:CARDINAL;
    unit,dir,f8,e3,newf8:str128;
BEGIN
    qdSplitAllPath(S,unit,dir,f8,e3);
    LOOP
        Str.CardToStr(LONGCARD(num),N,10,ok);
        len:=Str.Length(N);
        FOR p:=len+1 TO wipicndx DO Str.Prepend(N,"0");INC(len);END;
        p:=Str.Length(f8);
        IF (p+len) > F8maxlen THEN
            Str.Delete(f8,F8maxlen-len,p+len-F8maxlen);
        END;
        Str.Concat(newf8,f8,N);
        Lib.MakeAllPath(pic,unit,dir,newf8,e3);
        IF FIO.Exists(pic)=FALSE THEN EXIT; END;
        INC(num);
    END;
END buildpicname;

TYPE
    palette16=ARRAY[1..16] OF LONGCARD;
CONST
    pal16=palette16(
    Graph._BLACK    ,
    Graph._BLUE      ,
    Graph._GREEN     ,
    Graph._CYAN      ,
    Graph._RED       ,
    Graph._MAGENTA   ,
    Graph._BROWN     ,
    Graph._WHITE     ,
    Graph._GRAY     ,
    Graph._LIGHTBLUE ,
    Graph._LIGHTGREEN,
    Graph._LIGHTCYAN ,
    Graph._LIGHTRED  ,
    Graph._LIGHTMAGENTA,
    Graph._LIGHTYELLOW ,
    Graph._BRIGHTWHITE ); (* crire la palette dans l'ordre original de dfinition des couleurs ! *)

(* remember $FC >> 2 = $3F, but $FF >> 2 = $3F too ! *)

PROCEDURE PCXpal (VAR v:BYTE);
VAR
    i : CARDINAL;
BEGIN
    i := CARDINAL(v) << 2;
    IF i >= 0FCH THEN i := 0FFH; END; (* we lose $FC but it's better here ! *)
    v := BYTE(i);
END PCXpal;

PROCEDURE buildPCXheader (xmin,ymin,xmax,ymax,colors:CARDINAL):PCXheaderType;
VAR
    h : PCXheaderType;
    i : CARDINAL;
    r,g,b:BYTE;
    bgr:LONGCARD;
BEGIN
    h.manufacturer   := BYTE(0AH);
    h.version        := BYTE(5);
    h.encoding       := BYTE(1);
    i:=8;
    h.bitsPerPixel   := BYTE(i);
    h.leftmargin     := xmin;
    h.uppermargin    := ymin;
    h.rightmargin    := xmax;
    h.lowermargin    := ymax;
    h.hDPIresolution := 300;    (* reasonable but useless default *)
    h.vDPIresolution := 300;
        FOR i := 0 TO 15 DO
            bgr := pal16[i+1];
            r := BYTE( bgr        MOD 100H);
            g := BYTE((bgr >>  8) MOD 100H);
            b := BYTE((bgr >> 16) MOD 100H);
            PCXpal(r); PCXpal(g); PCXpal(b);
            h.colormap[i].red   := r;
            h.colormap[i].green := g;
            h.colormap[i].blue  := b;
        END;
    h.reserved1      := BYTE(0);
    h.NCP            := BYTE(1);
    i := (xmax-xmin+1) ; (*  8 bits per pixel, so DIV 8 then * 8 *)
    h.NBS            := i;
    h.paletteInfo    := 1;
    h.hscreensize    := 0; (* xmax-xmin+1 *)
    h.vscreensize    := 0; (* ymax-ymin+1 *)
    FOR i := 0 TO 53 DO
        h.reserved2[i] := BYTE(0);
    END;

    RETURN h;
END buildPCXheader;

CONST
    PCXMAXCOUNT  = 03FH;
    PCXCOUNTFLAG = BYTE(0C0H); (* 11000000 *)

PROCEDURE encodeScanline256 (y,xmin,xmax:CARDINAL   ) : CARDINAL ;
VAR
    oldink,ink : BYTE;
    x,count:CARDINAL;
    p:CARDINAL; (* in img.decodebuffer[] *)
BEGIN
    oldink:= vpoint(xmin,y);
    count := 1;
    p := firstcode;
    x := xmin+1;
    LOOP
        ink := vpoint(x,y);
        IF ink=oldink THEN
            IF count = PCXMAXCOUNT THEN (* flush ! *)
                img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(count));
                INC(p);
                img.decodebuffer[p]:=oldink;
                INC(p);
                count:=1; (* write 63, write byte, restart at 1 *)
            ELSE
                INC(count);
            END;
        ELSE
            IF count > 1 THEN
                img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(count));
                INC(p);
                count:=1; (* write count, restart at 1 *)
            ELSE (* count is 1 here *)
                IF (oldink AND PCXCOUNTFLAG) = PCXCOUNTFLAG THEN
                    img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(01H));
                    INC(p); (* write 1 *)
                END;
            END;
            img.decodebuffer[p]:=oldink;
            INC(p); (* write byte *)
            oldink:=ink; (* restart at 1 with new byte *)
        END;

        INC(x);
        IF x > xmax THEN EXIT; END;
    END;

            IF count > 1 THEN
                img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(count));
                INC(p); (* write count *)
            ELSE (* count is 1 here *)
                IF (oldink AND PCXCOUNTFLAG) = PCXCOUNTFLAG THEN
                    img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(01H));
                    INC(p); (* write 1 *)
                END;
            END;
            img.decodebuffer[p]:=oldink;
            INC(p); (* write byte *)
    RETURN (p-firstcode); (* p is already +1, so useless to -1 *)
END encodeScanline256;

CONST
    palID     = BYTE(0CH); (* 12=$0C *)
    palidsize = 1;

PROCEDURE savePCXpal256 (hnd:FIO.File);
VAR
    i : CARDINAL;
    r,g,b:SHORTCARD; (* BYTE ! *)
    bgr:LONGCARD;
BEGIN
    FIO.WrBin(hnd,palID,palidsize);
        FOR i := 0 TO 255 DO

            getDAC(i,r,g,b);
            PCXpal(r); PCXpal(g); PCXpal(b);
            FIO.WrBin(hnd,r,1);
            FIO.WrBin(hnd,g,1);
            FIO.WrBin(hnd,b,1);
        END;
END savePCXpal256;

PROCEDURE savePCX (S:ARRAY OF CHAR; maxpicndx:CARDINAL; VAR num:CARDINAL);
VAR
    pic:str128;
    hout:FIO.File;
VAR
    x,y,count,colors:CARDINAL;
VAR
    h : PCXheaderType;
BEGIN
    (* we could roll, too *)
    IF num = MAX(CARDINAL) THEN RETURN;END; (* safety *)
    IF num > maxpicndx THEN RETURN;END; (* John Donothing was here *)

    buildpicname(S,num,pic);
    Str.Append(pic,extPCX);

    colors:=256;

    hout:=FIO.Create(pic);
    FIO.AssignBuffer(hout,bufferOut);
    h := buildPCXheader(xmin,ymin,xmax,ymax,colors);
    FIO.WrBin(hout,h,SIZE(PCXheaderType));
    FOR y := ymin TO ymax DO
        count:=encodeScanline256(y,xmin,xmax);
        FIO.WrBin(hout, img.decodebuffer, count);
    END;
    savePCXpal256(hout);
    FIO.Flush(hout);
    FIO.Close(hout);
    INC(num); (* next ! *)

    (* IF num > maxpicndx THEN num:=1; END; (* safety *) *)
END savePCX;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE initrandom (VAR count:LONGCARD;   proba:CARDINAL);
VAR
    x,y:CARDINAL;
BEGIN
    count:=0;
    FOR x := firstX TO lastX DO
        FOR y := firstY TO lastY DO
            IF GetRndCardRange(minproba,maxproba) <= proba THEN
                MakeAlive(x,y);
                INC(count);
            END;
        END;
    END;
END initrandom;

(* //FIXME wrap mode ? *)

PROCEDURE initfile (VAR count:LONGCARD;
                    useLFN:BOOLEAN;initialpattern:CARDINAL;datfile:pathtype);
VAR
    x,y:CARDINAL;
    hnd : FIO.File;
    S   : str128; (* should be enough *)
    i,len,rows,columns:CARDINAL;
    fpos:LONGCARD;
    state : (waiting,getting);
    ch:CHAR;
BEGIN
    count:=0;
    FIO.EOF:=FALSE;
    hnd:=fileOpenRead(useLFN,datfile);
    FIO.AssignBuffer(hnd,bufferIn);
    i := 0;
    state := waiting;
    LOOP
        IF FIO.EOF=TRUE THEN EXIT; END;
        FIO.RdStr(hnd,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE state OF
        | waiting :
            CASE S[0] OF
            | pound,semicolon,charnull :
            ELSE
                state   :=getting;
                rows    :=0;
                columns :=0;
                fpos    :=FIO.GetPos(hnd);
(* WrStr("Name : ");WrStr(S);WrLn; *)
            END;
        | getting :
            CASE S[0] OF
            | CHR(0) :
                INC(i);
                IF i = initialpattern THEN EXIT; END;
                state := waiting;
            ELSE
                len:= Str.Length(S);
                IF len > columns THEN columns:=len; END;
                INC(rows);
            END;
        END;
    END;
    (* should check here for last line without CR but who cares ? *)
    IF state=waiting THEN
        count:=0;
    ELSE
        rows:=rows DIV 2;
        columns:=columns DIV 2;
        x:=lastX DIV 2;
        y:=lastY DIV 2;
        IF rows > y THEN
            y:=firstY;
        ELSE
            y:=y - rows;
        END;
        IF columns > x THEN
            x:=firstX;
        ELSE
            x:=x - columns;
        END;
        FIO.Seek(hnd,fpos);
        LOOP
            IF FIO.EOF=TRUE THEN EXIT; END;
            FIO.RdStr(hnd,S);
            LtrimBlanks(S);
            RtrimBlanks(S);
(* WrStr("data : ");WrStr(S);WrLn; *)
            len := Str.Length(S);
            IF len = 0 THEN EXIT; END;
            FOR i := 0 TO len-1 DO
                 ch:=S[i];
                 CASE ch OF
                 | space,dot:
                 ELSE
                     IF (y<= lastY) AND ( (x+i) <= lastX) THEN
                         MakeAlive(x+i,y);
                         INC(count);
                     END;
                 END;
            END;
            INC(y);
        END;
    END;
    fileClose(useLFN,hnd);
END initfile;

(* ------------------------------------------------------------ *)

PROCEDURE strToInt (S:ARRAY OF CHAR):INTEGER;
VAR
    ok:BOOLEAN;
    v:LONGINT;
    vv:INTEGER;
BEGIN
    vv:=MAX(INTEGER) DIV 2;   (* out of 320x200 window, but should never happen *)
    v:=Str.StrToInt(S,10,ok);
    IF ok THEN
        IF v <= MAX(INTEGER) THEN
            IF v >= MIN(INTEGER) THEN vv:=INTEGER(v);END;
        END;
    END;
    RETURN vv;
END strToInt;

(* //FIXME wrap mode ? *)

PROCEDURE initLIFfile (VAR count:LONGCARD;
                       useLFN,forcecenter:BOOLEAN;datfile:pathtype );
VAR
    hnd:FIO.File;
    S,Z:str128;
    x,y,xx,yy,LIFxcenter,LIFycenter:INTEGER;
    i:CARDINAL;
BEGIN
    LIFxcenter:=INTEGER(lastX DIV 2);
    LIFycenter:=INTEGER(lastY DIV 2);

    x:=LIFxcenter;
    y:=LIFycenter;

    count:=0;
    FIO.EOF:=FALSE;
    hnd:=fileOpenRead(useLFN,datfile);
    FIO.AssignBuffer(hnd,bufferIn);
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE S[0] OF
        | charnull:
            IF FIO.EOF THEN EXIT; END;
        | pound:
            argv(Z, S,1,FALSE);
            Str.Caps(Z);
            IF same(Z,LIFpos) THEN (* upper left *)
                IF NOT(forcecenter) THEN
                    argv(Z, S,2,FALSE);  x:=LIFxcenter+strToInt(Z);
                    argv(Z, S,3,FALSE);  y:=LIFycenter+strToInt(Z);
                END;
            END;
        | "+","-","0".."9":
            ReplaceChar(S,"+",""); (* avoid parsing problem *)
            argv(Z, S,1,FALSE);  xx:=LIFxcenter+strToInt(Z);
            argv(Z, S,2,FALSE);  yy:=LIFycenter+strToInt(Z);
                      IF (yy >= INTEGER(firstY)) AND (yy <= INTEGER(lastY)) THEN
                          IF (xx >= INTEGER(firstX)) AND (xx <= INTEGER(lastX)) THEN
                              MakeAlive( CARDINAL(xx),CARDINAL(yy));
                              INC(count);
                          END;
                      END;
        | dot,star:
            yy:=y;
            FOR i:=1 TO Str.Length(S) DO
                 CASE S[i-1] OF
                 | dot: ;
                 | star:
                      xx:=x+INTEGER(i)-1;
                      IF (yy >= INTEGER(firstY)) AND (yy <= INTEGER(lastY)) THEN
                          IF (xx >= INTEGER(firstX)) AND (xx <= INTEGER(lastX)) THEN
                              MakeAlive( CARDINAL(xx),CARDINAL(yy));
                              INC(count);
                          END;
                      END;
                 END;
            END;
            INC(y);
        | semicolon:
            ;
        ELSE
            ;
        END;
    END;
    fileClose(useLFN,hnd);
END initLIFfile;

(* ------------------------------------------------------------ *)

(*
we could RETURN a BOOLEAN as a warning currentCount is 0 FOR any reason
(initialpattern out of range OR bad randomization)
but it's better TO init a random display
*)

PROCEDURE InitWorld (VAR currentCount,currgeneration:LONGCARD;
                    useLFN,useLIF,forcecenter:BOOLEAN;proba:CARDINAL;
                    initialpattern,cellsize:CARDINAL;datfile:pathtype);
VAR
    x,y : CARDINAL;
BEGIN
    InitRnd();

    lastX := (screenX DIV (cellsize+1))-1;
    lastY := (screenY DIV (cellsize+1))-1;
    currgeneration := 0;

    FOR x := firstX TO lastX DO
        FOR y := firstY TO lastY DO
           MakeEmpty(x,y);
        END;
    END;

    IF useLIF THEN
        initLIFfile(currentCount, useLFN,forcecenter,datfile);
    ELSE
        CASE initialpattern OF
        | nopattern :
            initrandom(currentCount, proba);
        ELSE
            initfile(currentCount, useLFN,initialpattern,datfile);
        END;
    END;
    IF currentCount = 0 THEN initrandom(currentCount, proba); END;
END InitWorld;

(* ------------------------------------------------------------ *)

PROCEDURE ShowWorld (cellsize:CARDINAL;box:BOOLEAN );
VAR
    i,x,y:CARDINAL;
BEGIN
    FOR x := firstX TO lastX DO
        FOR y := firstY TO lastY DO
            CASE isAlive(x,y) OF
            | FALSE :
                Set(x,y,colorEmpty,cellsize,box);
            | TRUE :
                Set(x,y,currentColor,cellsize,box);
            END;
        END;
    END;
END ShowWorld;

(* ------------------------------------------------------------ *)

PROCEDURE InitColor (cyclecolor:CARDINAL);
(*
CONST
    blockfill=Graph.FillMaskType(0FFH,0FFH,0FFH,0FFH,0FFH,0FFH,0FFH,0FFH);
*)
BEGIN
    IF cyclecolor = 0 THEN
        currentColor := firstColor;
    ELSE
        currentColor := cyclecolor;
    END;
(*
    Graph.SetFillMask(blockfill);
*)
END InitColor;

PROCEDURE CycleColor (cyclecolor:CARDINAL);
BEGIN
    IF cyclecolor = 0 THEN
        INC(currentColor);
        IF currentColor = lastColor THEN currentColor:=firstColor; END;
    END;
END CycleColor;

(* ------------------------------------------------------------ *)

PROCEDURE NextGeneration (VAR currentCount,currgeneration:LONGCARD;
                         wrap,box:BOOLEAN;proba,speed,cellsize:CARDINAL;
                         survivalrule,birthrule,rndbirthrule:str16 ) : LONGCARD;
CONST
    basechar="0";
VAR
    i,x,y:CARDINAL;
    content : SHORTCARD;
    changes : LONGCARD;
    c:CHAR;
BEGIN

    (* pass 1 : create # of neighbours for this cell *)

    FOR x := firstX+1 TO lastX-1 DO
        FOR y := firstY+1 TO lastY-1 DO
            IF isAlive(x,y) THEN
               Add (x-1,y-1);
               Add (x  ,y-1);
               Add (x+1,y-1);
               Add (x-1,y  );
               Add (x+1,y  );
               Add (x-1,y+1);
               Add (x  ,y+1);
               Add (x+1,y+1);
            END;
        END;
    END;

    (* north *)

    y := firstY;
    FOR x := firstX+1 TO lastX-1 DO
        IF isAlive(x,y) THEN
            IF wrap=TRUE THEN
                Add (x-1,lastY);
                Add (x  ,lastY);
                Add (x+1,lastY);
            END;
            Add (x-1,y  );
            Add (x+1,y  );
            Add (x-1,y+1);
            Add (x  ,y+1);
            Add (x+1,y+1);
        END;
    END;

    (* south *)

    y := lastY;
    FOR x := firstX+1 TO lastX-1 DO
        IF isAlive(x,y) THEN
            Add (x-1,y-1);
            Add (x  ,y-1);
            Add (x+1,y-1);
            Add (x-1,y  );
            Add (x+1,y  );
            IF wrap=TRUE THEN
                Add (x-1,firstY);
                Add (x  ,firstY);
                Add (x+1,firstY);
            END;
        END;
    END;

    (* west *)

    x := firstX;
    FOR y := firstY+1 TO lastY-1 DO
        IF isAlive(x,y) THEN
            IF wrap=TRUE THEN
                Add (lastX,y-1);
                Add (lastX,y  );
                Add (lastX,y+1);
            END;
            Add (x  ,y-1);
            Add (x  ,y+1);
            Add (x+1,y-1);
            Add (x+1,y  );
            Add (x+1,y+1);
        END;
    END;

    (* east *)

    x := lastX;
    FOR y := firstY+1 TO lastY-1 DO
        IF isAlive(x,y) THEN
            IF wrap=TRUE THEN
                Add (firstX,y-1);
                Add (firstX,y  );
                Add (firstX,y+1);
            END;
            Add (x  ,y-1);
            Add (x  ,y+1);
            Add (x-1,y-1);
            Add (x-1,y  );
            Add (x-1,y+1);
        END;
    END;

    (* nw *)

    x := firstX;
    y := firstY;
    IF isAlive(x,y) THEN
        IF wrap = TRUE THEN
            Add(lastX ,lastY );
            Add(x     ,lastY );
            Add(x+1   ,lastY );
            Add(lastX ,y     );
            Add(lastX ,y+1   );
        END;
        Add (x+1,y  );
        Add (x  ,y+1);
        Add (x+1,y+1);
    END;

    (* ne *)

    x := lastX;
    y := firstY;
    IF isAlive(x,y) THEN
        IF wrap = TRUE THEN
            Add(x-1   ,lastY );
            Add(x     ,lastY );
            Add(firstX,lastY );
            Add(firstX,y     );
            Add(firstX,y+1   );
        END;
        Add (x-1,y  );
        Add (x-1,y+1);
        Add (x  ,y+1);
    END;

    (* sw *)

    x := firstX;
    y := lastY;
    IF isAlive(x,y) THEN
        IF wrap = TRUE THEN
            Add(lastX ,y-1    );
            Add(lastX ,y      );
            Add(lastX ,firstY );
            Add(x     ,firstY );
            Add(x+1   ,firstY );
        END;
        Add (x  ,y-1);
        Add (x+1,y-1);
        Add (x+1,y  );
    END;

    (* se *)

    x := lastX;
    y := lastY;
    IF isAlive(x,y) THEN
        IF wrap = TRUE THEN
            Add(firstX,y-1   );
            Add(firstX,y     );
            Add(x-1   ,firstY);
            Add(x     ,firstY);
            Add(firstX,firstY);
        END;
        Add (x-1,y-1);
        Add (x  ,y-1);
        Add (x-1,y  );
    END;

    (* pass 2 : update display and reset array *)

    FOR i:=minspeed+1 TO speed DO WaitVGAretrace; END;

    changes := 0;
    FOR x := firstX TO lastX DO
        FOR y := firstY TO lastY DO
            CASE wasEmpty(x,y) OF
            | TRUE :
                c:=basechar+CHR(GetNeighbours(x,y));
                IF Str.CharPos(birthrule,c) # MAX(CARDINAL) THEN
                    MakeAlive(x,y);
                    Set(x,y,currentColor,cellsize,box);
                    INC(currentCount);
                    INC(changes);
                ELSE

                    (* // v1.1a ... *)
                    IF Str.CharPos(rndbirthrule,c) # MAX(CARDINAL) THEN
                        IF GetRndCardRange(minproba,maxproba) <= proba THEN
                            MakeAlive(x,y);
                            Set(x,y,currentColor,cellsize,box);
                            INC(currentCount);
                            INC(changes);
                        ELSE
                            MakeEmpty(x,y); (* reset neighbour counter *)
                        END;
                    ELSE
                    (* ... v1.1a // *)

                        MakeEmpty(x,y); (* reset neighbour counter *)
                    END;
                END;
            | FALSE :
                c:=basechar+CHR(GetNeighbours(x,y));
                IF Str.CharPos(survivalrule,c) # MAX(CARDINAL) THEN
                    MakeAlive(x,y); (* reset neighbour counter *)
                ELSE
                    MakeEmpty(x,y);
                    Set(x,y,colorEmpty,cellsize,box);
                    DEC(currentCount);
                    INC(changes);
                END;
            END;
        END;
    END;

    INC(currgeneration);

    RETURN changes;
END NextGeneration;

(* ------------------------------------------------------------ *)

PROCEDURE showStats (count,gen:LONGCARD;useLIF,forceuserrule,showfullinfos:BOOLEAN;
                    survivalrule,birthrule,rndbirthrule:str16);
CONST
    gencolor   = cLIGHTGREEN;
    countcolor = cLIGHTYELLOW;
    rulecolor  = cLIGHTCYAN;
    wi         = 5;
    padchar    = " ";
    DEBUG      = TRUE;
VAR
    len,i,oldcolor : CARDINAL;
    xy:Graph.TextCoords;
    S,ZAP:str16;
    R:str80; (* rule can be long *)
    erase,ok:BOOLEAN;
BEGIN
dbgbool(DEBUG,useLIF,"useLIF");
dbgbool(DEBUG,forceuserrule,"forceuserrule");
    erase := (gen= MAX( LONGCARD)) AND (count= MAX(LONGCARD));

    (*
    Str.Copy(R,rule);
    IF useLIF THEN
        IF forceuserrule=FALSE THEN
           IF same(lifrule,"")=FALSE THEN Str.Copy(R,lifrule); END;
        END;
    END;
    *)

    IF showfullinfos THEN
        R:="s~.b~.?~"; ZAP:=".?~";
    ELSE
        R:="~.~.~";    ZAP:=".~";
    END;
    Str.Subst(R,tilde,survivalrule);
    Str.Subst(R,tilde,birthrule);
    IF same(rndbirthrule,"") THEN
        Str.Subst(R,ZAP,"");
    ELSE
        Str.Subst(R,tilde,rndbirthrule);
    END;
    ReplaceChar(R,vbar,dot);

dbg(DEBUG,"R",R);

    oldcolor := Graph.SetTextColor(rulecolor);
    len := Str.Length(R);
    IF erase THEN
        FOR i:=1 TO len DO R[i-1]:=blank;END;
    END;
    xy:=Graph.SetTextPosition(vtabmin,Htabmin+(Htabmax-len) DIV 2);
    Graph.OutText(R);

    oldcolor := Graph.SetTextColor(gencolor);
    Str.CardToStr(gen,S,10,ok);
    len:=Str.Length(S);
    IF erase THEN
        FOR i:=1 TO len DO S[i-1]:=blank;END;
    END;
    xy:=Graph.SetTextPosition(vtabmin,Htabmin);
    Graph.OutText(S);

    oldcolor := Graph.SetTextColor(countcolor);
    Str.CardToStr(count,S,10,ok);
    len:=Str.Length(S);
    FOR i:=len+1 TO wi DO Str.Prepend(S,blank);END; (* should count be lower than previous... *)
    IF erase THEN
        FOR i:=1 TO len DO S[i-1]:=blank;END;
    END;
    xy:=Graph.SetTextPosition(vtabmin,Htabmin+Htabmax-wi);
    Graph.OutText(S);

END showStats;

(* ------------------------------------------------------------ *)

PROCEDURE MouseClick (  ) : BOOLEAN;
VAR
    info : MsMouse.MsData;
BEGIN
    MsMouse.GetStatus(info);
    IF (info.left_pressed=TRUE) OR (info.right_pressed=TRUE) THEN
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END MouseClick;

PROCEDURE isKeyWaiting (mouse:BOOLEAN) : BOOLEAN;
BEGIN
    IF mouse=TRUE THEN
        IF MouseClick()=TRUE THEN RETURN TRUE; END;
    END;
    RETURN IO.KeyPressed();
END isKeyWaiting;

PROCEDURE ReadNormalKey (mouse:BOOLEAN) : CHAR;
VAR
    ch : CHAR;
    c  : CHAR;
BEGIN
    IF mouse=TRUE THEN
        IF MouseClick()=TRUE THEN RETURN escape; END;
    END;
    ch := IO.RdKey();
    IF ch = special THEN c := IO.RdKey(); END;
    RETURN ch;
END ReadNormalKey;

(* ------------------------------------------------------------ *)

PROCEDURE HiresOn (  ) : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(gmode);
END HiresOn;

PROCEDURE TextOn (  ) : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(Graph._DEFAULTMODE);
END TextOn;

(* ------------------------------------------------------------ *)

PROCEDURE chkRule (VAR s:ARRAY OF CHAR) : BOOLEAN;
CONST
    legalrange="012345678";
VAR
    len,i:CARDINAL;
    p:CARDINAL;
BEGIN
    RtrimBlanks(s);
    LtrimBlanks(s);
    len:=Str.Length(s);
    IF len=0 THEN RETURN TRUE; END; (* allow empty ryle *)
    IF len > Str.Length(legalrange) THEN RETURN FALSE; END;
    FOR i := 0 TO len-1 DO
        IF Str.CharPos(legalrange,s[i])=MAX(CARDINAL) THEN RETURN FALSE; END;
    END;
    RETURN TRUE;
END chkRule;

(* should we allow S B markers ? *)

PROCEDURE ParseRule (VAR survivalrule,birthrule,rndbirthrule:str16;
                    rule : ARRAY OF CHAR) : BOOLEAN;
CONST
    separators = dash+coma+slash+colon;
VAR
    Z : str80; (* " s012345678/b012345678/012345678 " = 1+1+9+1+1+9+1+9+1 = 33 *)
    R : str16;
    i:CARDINAL;
BEGIN
    Str.Copy(Z,rule);
    IF Str.Length(Z)=0 THEN RETURN FALSE; END;
    (* trick *)
    Str.Caps(Z);
    Str.Prepend(Z,space);
    Str.Append(Z,space);

    FOR i:=1 TO 3 DO
        Str.ItemS(R,Z,separators,i-1);
        IF Str.CharPos(R,"S") # MAX(CARDINAL) THEN
            ReplaceChar(R,"S","");
            survivalrule := R;
        ELSIF Str.CharPos(R,"B") # MAX(CARDINAL) THEN
            ReplaceChar(R,"B","");
            birthrule := R;
        ELSE
            CASE i OF
            | 1: survivalrule := R;
            | 2: birthrule    := R;
            | 3: rndbirthrule := R;
            END;
        END;
    END;
    (*
    Str.ItemS(survivalrule,rule,separators,0);
    Str.ItemS(birthrule   ,rule,separators,1);
    Str.ItemS(rndbirthrule,rule,separators,2);
    *)
    IF chkRule(survivalrule)=FALSE THEN RETURN FALSE; END;
    IF chkRule(birthrule   )=FALSE THEN RETURN FALSE; END;
    IF chkRule(rndbirthrule)=FALSE THEN RETURN FALSE; END;
    RETURN TRUE;
END ParseRule;

(* ------------------------------------------------------------ *)

PROCEDURE showlong (v:LONGCARD);
VAR
    S  : str128;
    ok : BOOLEAN;
BEGIN
    Str.CardToStr(v,S,10,ok);
    Ltrim(S,space);
    WrStr(S);
END showlong;

PROCEDURE showcard (v:CARDINAL);
VAR
    S  : str128;
    ok : BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(v),S,10,ok);
    Ltrim(S,space);
    WrStr(S);
END showcard;

PROCEDURE showbool (flag:BOOLEAN);
VAR
    S : str128;
BEGIN
    IF flag=TRUE THEN
        S:="YES";
    ELSE
        S:="no";
    END;
    WrStr(S);
END showbool;

PROCEDURE displaylist (useLFN:BOOLEAN;datfile:pathtype);
CONST
    wi  = 5;
    sep = " : ";
VAR
    hnd : FIO.File;
    S   : str128; (* should be enough *)
    state : (waiting,skipping);
    currpattern:CARDINAL;
BEGIN
    WrStr( "file "+sep+dquote);WrStr(datfile);WrStr(dquote);WrLn;
    WrLn;
    WrStr( "index"+sep+"pattern");WrLn;
    WrLn;

    currpattern:=0;

    FIO.EOF:=FALSE;
    hnd:=fileOpenRead(useLFN,datfile);
    FIO.AssignBuffer(hnd,bufferIn);
    state := waiting;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE state OF
        | waiting :
            CASE S[0] OF
            | pound,semicolon,charnull :
                ;
            ELSE
                INC(currpattern);
                IO.WrCard(currpattern,wi);WrStr(sep);WrStr(S);WrLn;
                state   :=skipping;
            END;
        | skipping :
            IF S[0] = charnull THEN state := waiting; END;
        END;
    END;
    fileClose(useLFN,hnd);
END displaylist;

PROCEDURE displayLIF (VAR rule: ARRAY OF CHAR;
                     useLFN,dodmp:BOOLEAN;datfile:pathtype);
CONST
    LIFlist = LIFid+delim+
              LIFrem1+delim+
              LIFrem2+delim+
              LIFrule+delim+
              LIFdefrule;
VAR
    hnd : FIO.File;
    S,Z,msg : str128; (* more than enough *)
    i:CARDINAL;
    (* currem : CARDINAL; *)
    dmp:BOOLEAN;
BEGIN
    Str.Copy(rule,"");
    IF dodmp THEN
        WrStr("::: file : "+dquote);WrStr(datfile);WrStr(dquote);WrLn;
    END;

    (* currrem:=0; *)

    FIO.EOF:=FALSE;
    hnd:=fileOpenRead(useLFN,datfile);
    FIO.AssignBuffer(hnd,bufferIn);
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hnd,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        argv(Z, S,1,FALSE);
        IF (same(Z,"") AND FIO.EOF) THEN EXIT; END;
        Str.Caps(Z);
        i:=getStrIndex(delim, Z,LIFlist);
        dmp:=TRUE;
        CASE i OF
        | 1:   msg:="::: "+extLIF+" : v"; (* ".LIF" *)
        | 2,3: msg:="=== ";
               (*
               IF currrem < LIFmaxrem THEN
                   INC(currrem);
               ELSE
                   dmp:=FALSE;
               END;
               *)
        | 4:   msg:="::: rule : ";
        | 5:   msg:="::: rule : "+defaultConwayRule+" (Conway's original)";
        ELSE
               dmp:=FALSE;
        END;
        IF dmp THEN
            argv(Z, S,1,FALSE);
            Str.Subst(S, Z,"");
            LtrimBlanks(S);
            RtrimBlanks(S); (* useless ! *)
            IF dodmp THEN WrStr(msg);WrStr(S);WrLn; END;
            CASE i OF
            | 4: Str.Copy(rule,S);
            | 5: Str.Copy(rule,defaultConwayRule);
            END;
        END;
    END;
    fileClose(useLFN,hnd);
END displayLIF;

PROCEDURE displayparms (usemouse,wrap,box,useLIF,forcecenter,forceuserrule:BOOLEAN;
                        pBirth,cyclecolor,cellsize,
                        initialdelay,refreshdelay,
                        initialpattern,ruleRNDpBirth,speed:CARDINAL;
                        lastgeneration:LONGCARD;
                        rule,lifrule,datfile:ARRAY OF CHAR
                        );
CONST
    sFile         = "Source file             : ";
    sRule         = "Survival/Birth[/Random] : ";
    sRuleLIF      = "File-defined rule       : ";
    sForceUserRule= "Ignore #R and #N        : ";
    sPat          = "Initial pattern         : ";
    sCellSize     = "Size of cells in pixels : ";
    sPause        = "Initial pause           : ";
    sRefresh      = "Refresh pause           : ";
    sLastGen      = "Last generation         : ";
    sPbirth       = "Probability of birth    : ";
    sPrndbirth    = "Probability of RNDbirth : ";
    sShapes       = "Shapes                  : ";
    sColor        = "Color cycling           : ";
    sSpeed        = "Speed                   : ";
    sMouse        = "Abort on mouse click    : ";
    sWrap         = "Wrap coordinates        : ";
    sForceCenter  = "Ignore #P               : ";
VAR
    S : str128;
BEGIN
    WrStr(Banner); WrLn;
    WrLn;
    WrStr(sFile);WrStr(dquote);WrStr(datfile);WrStr(dquote);WrLn;

    WrStr(sRule); WrStr(rule);
    IF useLIF THEN
        IF forceuserrule THEN
            WrStr(" (forced)");
        ELSE
            IF same(lifrule,"")=FALSE THEN WrStr(" (ignored)");END;
        END;
    END;
    WrLn;

    WrStr(sRuleLIF);
    IF useLIF THEN
        IF same(lifrule,"") THEN
            WrStr("None");
        ELSE
            WrStr(lifrule);
            IF forceuserrule THEN WrStr(" (overridden)");END;
        END;
    ELSE
        WrStr("Not applicable");
    END;
    WrLn;

    WrStr(sForceUserRule);
    IF useLIF THEN
        showbool(forceuserrule);
    ELSE
        WrStr("not applicable");
    END;
    WrLn;

    WrStr(sPat);
    IF initialpattern = nopattern THEN
        WrStr("random");
    ELSE
        showcard(initialpattern);
    END;
    IF useLIF THEN WrStr(" (ignored)");END;
    WrLn;

    WrStr(sCellSize); showcard(cellsize); WrLn;
    WrStr(sPause);
    IF initialdelay=0 THEN
        WrStr("none");
    ELSE
        showcard(initialdelay);
    END;
    WrLn;
    WrStr(sRefresh);
    IF refreshdelay=0 THEN
        WrStr("none");
    ELSE
        showcard(refreshdelay);
    END;
    WrLn;
    WrStr(sLastGen);
    IF lastgeneration = 0 THEN
        WrStr("unlimited");
    ELSE
        showlong(lastgeneration);
    END;
    WrLn;

    WrStr(sPbirth); showcard(pBirth); WrLn;
    WrStr(sPrndbirth); showcard(ruleRNDpBirth); WrLn;

    WrStr   (sShapes);
    IF box THEN
        S:="boxes";
    ELSE
        S:="disks";
    END;
    WrStr(S);
    WrLn;
    WrStr   (sColor);
    IF cyclecolor=0 THEN
        WrStr("enabled");
    ELSE
        WrStr("disabled, color is "); showcard(cyclecolor);
    END;
    WrLn;
    WrStr(sSpeed); showcard(speed); WrLn;
    WrStr(sMouse); showbool(usemouse); WrLn;
    WrStr(sWrap ); showbool(wrap); WrLn;

    WrStr(sForceCenter);
    IF useLIF THEN
        showbool(forcecenter);
    ELSE
        WrStr("not applicable");
    END;
    WrLn;

END displayparms;

PROCEDURE newspeed ( VAR speed:CARDINAL; minspeed,maxspeed:CARDINAL; motion:INTEGER );
BEGIN
    IF motion < 0 THEN
        IF speed > minspeed THEN DEC(speed);END;
    ELSE
        IF (speed+1) <= maxspeed THEN INC(speed);END;
    END;
END newspeed;

PROCEDURE donothing (n:CARDINAL);
BEGIN
    Lib.Delay(n);
END donothing;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    useLFN    : BOOLEAN;
    usemouse  : BOOLEAN;
    wrap   : BOOLEAN;
    pBirth : CARDINAL;
    ruleRNDpBirth : CARDINAL;
    cyclecolor,orgcyclecolor  : CARDINAL;
    initialdelay : CARDINAL;
    refreshdelay : CARDINAL;
    speed        : CARDINAL;
    lastgeneration : LONGCARD;
    initialpattern : CARDINAL;
    rule,lifrule   : str128;
    cmd    : (show,list);
    forceuserrule,forcecenter,useLIF,showinfos,showfullinfos : BOOLEAN;
    bankfile,datfile: pathtype;
    savescr,DEBUG  : BOOLEAN;
    maxpicndx,picndx:CARDINAL; (* yes, a LONGCARD would do too *)
VAR
    survivalrule    : str16;
    birthrule       : str16;
    rndbirthrule    : str16;
    cellsize        : CARDINAL;
    box             : BOOLEAN;
    currentCount    : LONGCARD;
    currgeneration  : LONGCARD;
VAR
    singlestep : BOOLEAN;
    ch         : CHAR;
    showparms  : BOOLEAN;
    event      : BOOLEAN;
    anykey     : str2;
VAR
    parmcount,i,opt : CARDINAL;
    S,R             : pathtype;
    n         : LONGCARD;
    infovideo : Graph.VideoConfig;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn; (* must be here *)

    useLFN         := TRUE;
    usemouse       := FALSE;
    wrap           := TRUE;
    pBirth         := defaultBirthProba;
    ruleRNDpBirth  := defaultRuleRNDbirthProba;
    cyclecolor     := 0;
    cellsize       := defaultCellSize;
    box            := TRUE;
    initialdelay   := defaultinitialdelay;
    refreshdelay   := mindelay;
    lastgeneration := 10000;
    initialpattern := nopattern; (* random *)
    rule           := defaultConwayRule; (* "23/3" *)
    lifrule        := "";
    useLIF         := FALSE;
    bankfile       := "";
    speed          := defaultspeed;
    forcecenter    := FALSE;
    showinfos      := FALSE;
    showfullinfos  := FALSE;
    forceuserrule  := FALSE;
    picndx         := 1;
    maxpicndx      := defaultmaxpicndx;
    savescr        := FALSE;
    DEBUG          := FALSE;
    cmd            := show;

    singlestep := FALSE;
    showparms := FALSE;

    IF ParseRule(survivalrule,birthrule,rndbirthrule, rule)=FALSE THEN abort(errRule,rule); END; (* cannot happen here *)

    parmcount := Lib.ParamCount();
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S); cleantabs(R);
        UpperCase(R);
        IF isOption(R)=FALSE THEN Str.Prepend(R,dash); END; (* trick ! *)

            opt := GetOptIndex(R,"?"+delim+"H"+delim+"HELP"+delim+
                                 "M:"+delim+"MOUSE:"+delim+
                                 "W:"+delim+"WRAP:"+delim+
                                 "P:"+delim+"PROBA:"+delim+
                                 "C:"+delim+"CYCLE:"+delim+
                                 "S:"+delim+"SIZE:"+delim+
                                 "B:"+delim+"BOX:"+delim+
                                 "R:"+delim+"RULE:"+delim+
                                 "I"+delim+"INFO"+delim+
                                 "D:"+delim+"DELAY:"+delim+
                                 "G:"+delim+"GENERATION:"+delim+
                                 "I:"+delim+"INITIAL:"+delim+
                                 "U:"+delim+"RNDPROBA:"+delim+
                                 "L"+delim+"LIST"+delim+
                                 "T:"+delim+"TEMPO:"+delim+
                                 "Z:"+delim+"SPEED:"+delim+
                                 "F:"+delim+"FILE:"+delim+
                                 "X"+delim+"LFN"+delim+
                                 "??"+delim+

                                 "FF:"+delim+
                                 "FFF:"+delim+
                                 "II:"+delim+
                                 "III:"+delim+
                                 "C"+delim+"CENTER"+delim+"XY"+delim+
                                 "S"+delim+"STATS"+delim+
                                 "SS"+delim+"FULLSTATS"+delim+
                                 "R"+delim+"MYRULE"+delim+"USERRULE"+delim+
                                 "N:"+delim+"BANK:"+delim+
                                 "STEPS"+delim+"PCX"+delim+
                                 "STEPS:"+delim+"PCX:"+delim+
                                 "DEBUG"
                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5 :   IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | 0 : usemouse := FALSE;
                      | 1 : usemouse := TRUE;
                      ELSE
                         abort(errRange,S);
                      END;
            | 6,7 :
                      IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | 0 : wrap := FALSE;
                      | 1 : wrap := TRUE;
                      ELSE
                         abort(errRange,S);
                      END;
            | 8,9 :   IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | minproba..maxproba :  pBirth:= CARDINAL(n);
                      ELSE
                          abort(errRange,S);
                      END;
            | 10,11 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | 0 : cyclecolor := 0;
                      | minink..maxink : cyclecolor := CARDINAL(n); orgcyclecolor:=cyclecolor;
                      ELSE
                          abort(errRange,S);
                      END;
            | 12,13 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | minsize..maxsize : cellsize := CARDINAL(n);
                      ELSE
                          abort(errRange,S);
                      END;
            | 14,15 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | 0 : box := FALSE;
                      | 1 : box := TRUE;
                      ELSE
                          abort(errRange,S);
                      END;
            | 16,17 : GetString(S,rule);
                      IF ParseRule(survivalrule,birthrule,rndbirthrule, rule)=FALSE THEN abort(errRule,S); END;
            | 18,19 : showparms := TRUE;
            | 20,21 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | mindelay..maxdelayrefresh : refreshdelay := CARDINAL(n);
                      ELSE
                          abort(errRange,S);
                      END;
            | 22,23 : IF GetLongCard(S,lastgeneration)=FALSE THEN abort(errConversion,S);END;
            | 24,25 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      initialpattern := CARDINAL(n);
            | 26,27:  IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | minproba..maxproba : ruleRNDpBirth:= CARDINAL(n);
                      ELSE
                          abort(errRange,S);
                      END;
            | 28,29:  cmd:=list;
            | 30,31 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | mindelay..maxdelayinitial : initialdelay := CARDINAL(n);
                      ELSE
                          abort(errRange,S);
                      END;
            | 32,33 : IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      CASE CARDINAL(n) OF
                      | minspeed..maxspeed : speed := CARDINAL(n);
                      ELSE
                          abort(errRange,S);
                      END;
            | 34,35 : GetString(S,datfile);
                      useLIF:=TRUE;
            | 36,37:  useLFN:=FALSE;
            | 38:     abort(errMoreHelp,"");

            | 39:     GetString(S,datfile);
                      useLIF:=TRUE;
                          showparms:=TRUE;
            | 40:     GetString(S,datfile);
                      useLIF:=TRUE;
                          showparms:=TRUE; cellsize:=minsize;
            | 41:     IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      initialpattern := CARDINAL(n);
                          showparms:=TRUE;
            | 42:     IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      initialpattern := CARDINAL(n);
                          showparms:=TRUE; cellsize:=minsize;
            | 43,44,45: forcecenter:=TRUE;
            | 46,47:  showinfos:=TRUE;
            | 48,49:  showinfos:=TRUE; showfullinfos:=TRUE;
            | 50,51,52:  forceuserrule:=TRUE;
            | 53,54:  GetString(S,bankfile);
            | 55,56:  savescr:=TRUE;
            | 57,58:  IF GetLongCard(S,n)=FALSE THEN abort(errConversion,S);END;
                      maxpicndx := CARDINAL(n);
                      savescr:=TRUE;
            | 59:     DEBUG := TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
    END;

    useLFN := (useLFN AND fileSupportLFN() );

    IF same(bankfile,"")=FALSE THEN useLIF:=FALSE;END;

    IF useLIF THEN
        IF Str.RCharPos(datfile,dot)=MAX(CARDINAL) THEN Str.Append(datfile,extLIF);END;
        (* Str.Caps(datfile); (* useless ! *) *)
    ELSE
        Lib.ParamStr(datfile,0);
        Str.Caps(datfile); (* useless ! *)
        IF same(bankfile,"") THEN
            Str.Subst(datfile,extEXE,extDAT);
            Str.Copy(S,datfile);
            i:=Str.RCharPos(S,antislash);
            IF i # MAX(CARDINAL) THEN
                Str.Delete(S,0,i+1);
                IF fileExists(useLFN,S) THEN Str.Copy(datfile,S);END;
            END;
        ELSE
            IF Str.RCharPos(bankfile,dot)=MAX(CARDINAL) THEN
                Str.Append(bankfile,extDAT);
            END;
            Str.Copy(datfile,bankfile); (* //V12B *)
        END;
    END;

    (* ------------------------------------------------------------ *)

    IF cmd = list THEN
        IF fileExists(useLFN,datfile)=FALSE THEN abort (errNotFound,datfile); END;
        IF useLIF THEN
            displayLIF(lifrule, useLFN,TRUE,datfile);
        ELSE
            displaylist(useLFN,datfile);
        END;
        abort(errNone,"");
    END;

    (* ------------------------------------------------------------ *)

    (* cmd is show *)
dbg(DEBUG,"datfile",datfile);
dbg(DEBUG,"rule",rule);
dbg(DEBUG,"lifrule",lifrule);
    IF useLIF THEN
        IF fileExists(useLFN,datfile)=FALSE THEN abort (errNotFound,datfile);END;
        displayLIF(lifrule, useLFN,FALSE,datfile); (* // file rule overrides default or user-defined rule *)
        IF forceuserrule=FALSE  THEN
            IF same(lifrule,"")=FALSE THEN
                IF ParseRule(survivalrule,birthrule,rndbirthrule, lifrule)=FALSE THEN abort(errLIFrule,lifrule); END;
            END;
        END;
    ELSE
        IF initialpattern # nopattern THEN (* we'll need vlife.dat file *)
            IF fileExists(useLFN,datfile)=FALSE THEN abort (errNotFound,datfile); END;
        END;
    END;
dbg(DEBUG,"rule",rule);
dbg(DEBUG,"lifrule",lifrule);

    IF usemouse THEN
        IF MsMouse.Reset()=MAX(INTEGER) THEN abort(errMouseNeeded,""); END;
    END;

    Graph.GetVideoConfig(infovideo);
    IF infovideo.adapter # Graph._VGA THEN abort(errVGAneeded,""); END;

    IF showparms THEN
        displayparms (  usemouse,wrap,box,useLIF,forcecenter,forceuserrule,
                        pBirth,cyclecolor,cellsize,
                        initialdelay,refreshdelay,
                        initialpattern,ruleRNDpBirth,speed,
                        lastgeneration,
                        rule,lifrule,datfile);

        WrLn;
        S:="Hit ESCape to abort or almost any key to continue, or wait "+sMaxWait+" seconds... ";
        WrStr(S);
        (* anykey:=Waitkey(); (* do not use WaitkeyDelay here after all... NOT ! ;-) *) *)
        anykey := WaitkeyDelay("",maxWaitMS);
        WrLn;
        IF same(anykey,escape)=TRUE THEN abort(errNone,""); END;
    END;

    IF useLIF THEN initialpattern:=nopattern; END; (* if -f:$ then ignore -i:# *)

    IF HiresOn()=FALSE THEN abort(errCannotSet,""); END;

dbg(DEBUG,"rule",rule);
dbg(DEBUG,"lifrule",lifrule);

    InitWorld(currentCount,currgeneration, useLFN,useLIF,forcecenter,pBirth,initialpattern,cellsize,datfile);
    InitColor(cyclecolor);
    ShowWorld(cellsize,box);
    IF showinfos THEN showStats(currentCount,currgeneration,useLIF,forceuserrule,showfullinfos,survivalrule,birthrule,rndbirthrule);END;

    IF initialdelay # 0 THEN donothing(initialdelay); END; (* useless : 0 means no delay anyway *)

    LOOP

        IF savescr THEN savePCX(sVLPIC,maxpicndx,picndx);END;

        CycleColor(cyclecolor);

        IF singlestep THEN (* yes, yes, this IS kludgy *)
            WHILE isKeyWaiting(usemouse)=FALSE DO
            END;
            event:=TRUE;
        ELSE
            event:=isKeyWaiting(usemouse);
        END;
        IF event THEN
            ch := ReadNormalKey(usemouse);
            CASE ch OF
            | space :
                IF NOT(singlestep) THEN singlestep:=TRUE;END;
            | escape :
                EXIT;
            | star :
                hiresCLS;
                InitWorld(currentCount,currgeneration,useLFN,useLIF,forcecenter,pBirth,initialpattern,cellsize,datfile);
                InitColor(cyclecolor);
                ShowWorld(cellsize,box);
                IF showinfos THEN showStats(currentCount,currgeneration,useLIF,forceuserrule,showfullinfos,survivalrule,birthrule,rndbirthrule);END;
                IF initialdelay # 0 THEN donothing(initialdelay); END; (* useless : 0 means no delay anyway *)
                singlestep := FALSE;
            | "-" : newspeed(speed,minspeed,maxspeed,1);  singlestep:=FALSE;
            | "+" : newspeed(speed,minspeed,maxspeed,-1); singlestep:=FALSE;
            | tabul,dollar:
                IF ch=dollar THEN
                    cyclecolor:=orgcyclecolor;
                ELSE
                    INC(cyclecolor);
                END;
                IF cyclecolor > maxink THEN cyclecolor:=0; END;
                InitColor(cyclecolor);
                ShowWorld(cellsize,box);
            | "s","S":
                IF showinfos THEN
                    showStats( MAX( LONGCARD), MAX( LONGCARD),useLIF,forceuserrule,showfullinfos,survivalrule,birthrule,rndbirthrule);
                END;
                showinfos:=NOT(showinfos);
            ELSE
                singlestep := FALSE;
            END;
        END;

        (* WaitVGAretrace(); *)
        n:=NextGeneration(currentCount,currgeneration, wrap,box,ruleRNDpBirth,speed,cellsize,survivalrule,birthrule,rndbirthrule);

        IF (currentCount = 0) OR (n=0) THEN
            (* c := ReadNormalKey(mouse);
            EXIT; *)
            (* restart after 2 seconds ! *)
            donothing(pausevalue);
            hiresCLS;
            InitWorld(currentCount,currgeneration, useLFN,useLIF,forcecenter,pBirth,initialpattern,cellsize,datfile);
            InitColor(cyclecolor);
            ShowWorld(cellsize,box);
        END;

        IF refreshdelay # 0 THEN donothing(refreshdelay); END; (* useless : 0 means no delay anyway *)
        (* WaitVGAretrace(); *)

        IF (lastgeneration # 0) AND (lastgeneration < currgeneration) THEN
            (* restart after 2 seconds ! *)
            donothing(pausevalue);
            hiresCLS;
            InitWorld(currentCount,currgeneration, useLFN,useLIF,forcecenter,pBirth,initialpattern,cellsize,datfile);
            InitColor(cyclecolor);
            ShowWorld(cellsize,box);
        END;

        IF showinfos THEN showStats(currentCount,currgeneration,useLIF,forceuserrule,showfullinfos,survivalrule,birthrule,rndbirthrule);END;
    END;

    IF TextOn()=FALSE THEN abort(errCannotReset,""); END;
    (*
    IF ch = escape THEN
        WrStr("The Wrath of GOD has striken ! ;-)");
        WrLn;
    END;
    *)
    abort(errNone,"");

END vLife.

