(* ---------------------------------------------------------------
Title         Q&D Word/Token count
Overview
Usage
Notes
Bugs          not really a bug but a few files force a stack overflow
              thanks to deep recursion (discovered when parsing "roman 1..3999" output)
              solution is to run program twice, first with -o, then with -m

Wish List     sort ? bah, let an external program take care of that point...
              user-specified pattern target chars ?
              handle accentuated letter as special symbols ?
              what about space in quoted strings ?

--------------------------------------------------------------- *)

MODULE wCount;

IMPORT FIO;
IMPORT Str;
IMPORT Lib;
IMPORT Storage;
IMPORT IO;

FROM IO IMPORT WrStr,WrLn;

FROM Storage IMPORT Available,ALLOCATE,DEALLOCATE;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XopenFile, w9XcloseFile,
w9XsupportLFN;

(* ------------------------------------------------------------ *)

CONST
    wi               = 9; (* was 7, should do for LONGCARD width *)
    widashes         = "========="; (* wi len *)
CONST
    modulaseparators = " (),=+-*<>/\[]^:;.";
    sBlanksSet       = "[$00..$20]";
    sModulaSet       = sBlanksSet+' + "'+modulaseparators+'" + quotes';
    sNotAZset        = "^[A..Z]";
TYPE
    str5 = ARRAY [0..4] OF CHAR;
CONST
    defchUPPER    = "U";    ndxchUPPER  = 0;
    defchLOWER    = "c";    ndxchLOWER  = 1;
    defchDIGIT    = "#";    ndxchDIGIT  = 2;
    defchSYMBOL   = "!";    ndxchSYMBOL = 3;
    defchSPACE    = "_";    ndxchSPACE  = 4;
CONST
    defaultremap  = defchUPPER+defchLOWER+defchDIGIT+defchSYMBOL+defchSPACE;
    remapcount    = ndxchSPACE-ndxchUPPER+1;
    sRemapCount   = "5";
CONST
    dotstar     = ".*";
    extRPT      = ".RPT";
    extCOM      = ".COM";
    extEXE      = ".EXE";
    extDLL      = ".DLL";
    extOVR      = ".OVR";
    extOVL      = ".OVL";
    extDRV      = ".DRV";
    extVXD      = ".VXD";
    extBIN      = ".BIN";
    extZIP      = ".ZIP";
    extARJ      = ".ARJ";
    extLZH      = ".LZH";
    skippedextensions = extCOM+delim+extEXE+delim+
                        extDLL+delim+extOVR+delim+extOVL+delim+extDRV+delim+
                        extVXD+delim+extBIN+delim+
                        extZIP+delim+extARJ+delim+extLZH+delim+extRPT;
CONST
    progEXEname   = "WCOUNT";
    progTitle     = "Q&D Word/n-gram Counter";
    progVersion   = "v1.1c";
    progCopyright = "by PhG";
    banner        = progTitle+" "+progVersion+" "+progCopyright;
CONST
    errNone          = 0;
    errHelp          = 1;
    errOption        = 2;
    errTooManyParms  = 3;
    errMissingSpec   = 4;
    errTooManyMatches= 5;
    errNoMatch       = 6;
    errUpperLen      = 7;
    errLowerLen      = 8;
    errExactLen      = 9;
    errConflict      = 10;
    errNonsense      = 11;
    errAllocate      = 12;
    errMode          = 13;
    errJoker         = 14;
    errProblem       = 15;
    errBadSpec       = 16;
    errBadRemap      = 17;
    errAborted       = 255;

PROCEDURE abort (e : CARDINAL; einfo:ARRAY OF CHAR);

    MODULE message;
    IMPORT Str;
    EXPORT msg3;

    PROCEDURE msg3 (VAR R:ARRAY OF CHAR;S1,S2,S3:ARRAY OF CHAR);
    BEGIN
        Str.Concat(R,S1,S2);Str.Append(R,S3);
    END msg3;

    END message;

CONST
(*
 0        1         2         3         4         5         6         7         8
 12345678901234567890123456789012345678901234567890123456789012345678901234567890
*)
    msgHelp =
banner+nl+
nl+
"Syntax : "+progEXEname+" <file(s)> [option]..."+nl+
nl+
"This program reads tokens from specified text file(s) then builds frequencies."+nl+
'The "divide and conquer" approach is highly recommended !'+nl+
nl+
"-g    n-gram counter mode (default mode is word counter)"+nl+
"-l:#  n-gram lower length ([1..100], default is 1)"+nl+
"-u:#  n-gram upper length ([1..100], default is 2)"+nl+
"-e:#  n-gram exact length ([1..100], shortcut for -l:# -u:#)"+nl+
"-d    allow [A..Z0..9] (default is [A..Z])"+nl+
"-c    ignore case"+nl+
"-a    ignore accents"+nl+
'-f    do not expand "" to "ae" nor "" to "AE"'+nl+
"-b    set of token separators is "+sBlanksSet+nl+
"-k    set of token separators is "+sModulaSet+nl+
"-p|-w pattern analysis mode (-b forced, -g -d -c -f ignored)"+nl+
"-y    parse individual tokens in quoted strings (-p option only)"+nl+
"-n    line mode with CR ($0d) and LF ($0a) as only valid separators"+nl+
"-u:$  user-defined remapping string (upper, lower, digit, symbol, space)"+nl+
"-i    individual processing for each file (default is cumulative process)"+nl+
"-o[o] output raw data to video screen without storing to RAM"+nl+
"      (-q forced, -oo = -o -x -t -s)"+nl+
"-m    make an analysis using sorted output file (only -t taken into account)"+nl+
"-x    do not show header"+nl+
"-l    do not show list"+nl+
"-t    do not show frequency table"+nl+
"-s    do not show summary"+nl+
"-z    include D field (D=C-N)"+nl+
"-q    no processing eyecandy"+nl+
"-lfn  disable LFN support even if available"+nl+
nl+
"a) Note program may abort on a few files after creating an ERRORINF.$$$ file,"+nl+
"   for deep recursion may force a stack overflow even though stack is 64 Kb."+nl+
"   Here is a workaround : first run program with -o option and redirect output,"+nl+
"   then run program with -m option on previously redirected sorted output."+nl+
"b) L is token length, N is occurrences count, C is ( L * N ), D is C-N."+nl+
"c) Unless -b or -k options, "+sNotAZset+" is default set of token separators."+nl+
"d) Files should end with a CRLF sequence."+nl+
"e) Pattern analysis mode may be useful in order to check passwords."+nl+
"   Each token character is translated according to this rule :"+nl+
'   "'+defchUPPER+'" = uppercase letter, "'+defchLOWER+'" = lowercase letter,'+nl+
'   "'+defchDIGIT+'" = digit, "'+defchSYMBOL+'" = symbol, "'+defchSPACE+'" = space.'+nl+
"   -p option ignores leading and trailing quotes, whether double or single."+nl+
"   Without -y option, quoted strings are parsed as a single token"+nl+
"   (note leading and trailing double and single quotes are removed)."+nl+
"   Without -a option, accentuated letters are treated as symbols."+nl+
"f) -pp = -py = -yy = -p -y"+nl+
"g) If program was aborted using Escape, return code is 255."+nl+
"h) Program will not process files with any of these extensions :"+nl+
"   "+skippedextensions+" for safety."+nl;

VAR
    S : str1024; (* oversized, was str256 *)
BEGIN
    CASE e OF
    | errHelp :          WrStr(msgHelp);
    | errOption :        msg3(S,'Unknown "',einfo,'" option !');
    | errTooManyParms:   msg3(S,'Useless "',einfo,'" parameter !');
    | errMissingSpec:    S := "Missing <file(s)> specification !";
    | errTooManyMatches: msg3(S,'Too many files match "',einfo,'" specification !');
    | errNoMatch:        msg3(S,'No file matches "',einfo,'" filtered specification !');
    | errUpperLen:       msg3(S,'Illegal "',einfo,'" n-gram upper length !');
    | errLowerLen:       msg3(S,'Illegal "',einfo,'" n-gram lower length !');
    | errExactLen:       msg3(S,'Illegal "',einfo,'" n-gram exact length !');
    | errConflict:       S := "-e:# option is a nonsense with any of -u:# and -l:# options !";
    | errNonsense:       Str.Concat(S,einfo," options are mutually exclusive !");
    | errAllocate:       (* not very informative ! *)
                         msg3(S,'Storage.ALLOCATE() would not store "',einfo,'" token !');
    | errMode:           S := "-g option required with -l:#, -u:# and -e:# options !";
    | errJoker:          S := "-m option will process one file only !";
    | errProblem:        msg3(S,'Unexpected "',einfo,'" line !');
    | errBadSpec:        msg3(S,'Illegal "',einfo,'" specification !');
    | errBadRemap:       msg3(S,'"',einfo,'" user-defined string must be '+sRemapCount+' characters !');
    | errAborted:        S := "Aborted by user !";
    ELSE
                         S := "This is illogical, Captain ! ;-)";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        WrStr(progEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE dbg (DEBUG,enquote:BOOLEAN;S1,S2:ARRAY OF CHAR   );
CONST
    wi = 15;
VAR
    i:CARDINAL;
BEGIN
    IF DEBUG THEN
        WrStr("/// ");WrStr(S1);
        FOR i:=Str.Length(S1)+1 TO wi DO WrStr(" ");END;
        WrStr(" : ");
        IF enquote THEN WrStr(dquote);END;
        WrStr(S2);
        IF enquote THEN WrStr(dquote);END;
        WrLn;
    END;
END dbg;

(* taken almost verbatim from OLDNEW v1.1f code, out of weariness *)

TYPE
    pFname = POINTER TO fnameType;
    fnameType = RECORD
        next      : pFname;
        slen      : CARDINAL; (* a SHORTCARD could do but who knows ? *)
        str       : CHAR;
    END;

PROCEDURE initList (VAR anchor : pFname );
BEGIN
    anchor := NIL;
END initList;

PROCEDURE freeList (anchor : pFname);
VAR
    needed : CARDINAL;
    p      : pFname;
BEGIN
    (* p:=anchor; *)
    WHILE anchor # NIL DO
        needed := SIZE(fnameType) - SIZE(anchor^.str) + anchor^.slen;
        p := anchor^.next;
        DEALLOCATE(anchor,needed);
        anchor:=p;
    END
END freeList;

PROCEDURE buildNewPtr (VAR anchor,p:pFname; len:CARDINAL):BOOLEAN;
VAR
    needed : CARDINAL;
BEGIN
    needed := SIZE(fnameType) - SIZE(p^.str) + len;
    IF Available(needed)=FALSE THEN RETURN FALSE; END;
    IF anchor = NIL THEN
        ALLOCATE(anchor,needed);
        p:=anchor;
    ELSE
        p:=anchor;
        WHILE p^.next # NIL DO
            p:=p^.next;
        END;
        ALLOCATE(p^.next,needed);
        p:=p^.next;
    END;
    p^.next := NIL;
    RETURN TRUE;
END buildNewPtr;

(* assume p is valid *)

PROCEDURE getStr (VAR S : pathtype; p:pFname);
VAR
    len:CARDINAL;
BEGIN
    len := p^.slen;
    Lib.FastMove( ADR(p^.str),ADR(S),len);
    S[len] := nullchar; (* REQUIRED safety ! *)
END getStr;

(* ------------------------------------------------------------ *)

(* Str.Match is not case sensitive *)

PROCEDURE isReservedPattern (S,skipthem:ARRAY OF CHAR):BOOLEAN;
VAR
    e3 : str16;
    n:CARDINAL;
    rc:BOOLEAN;
BEGIN
    rc:=FALSE;
    n:=0;
    LOOP
        isoleItemS(e3, skipthem,delim,n);
        IF same(e3,"") THEN EXIT; END;
        Str.Prepend(e3,"*");
        IF Str.Match(S,e3) THEN rc:=TRUE;EXIT; END;
        INC(n);
    END;
    RETURN rc;
END isReservedPattern;

PROCEDURE isReservedEntry (S:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE buildFileList (VAR anchor:pFname;
                        noskip,useLFN,DEBUG :BOOLEAN;
                        spec:pathtype):CARDINAL;
VAR
    count:CARDINAL; (* should do ! *)
    ok,found:BOOLEAN;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    entry : FIO.DirEntry;
    dosattr:FIO.FileAttr;
    entryname:pathtype;
    len : CARDINAL;
    pp:pFname;
    excludeme1,excludeme2:BOOLEAN;
BEGIN
    count:=0;
    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,entry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(entryname,w9Xentry.fullfilename);
        ELSE
            Str.Copy(entryname,entry.Name);
        END;
        excludeme1 := isReservedEntry   (entryname);  (* skip "." and ".." *)
        IF noskip THEN
            excludeme2 := FALSE;
        ELSE
            excludeme2 := isReservedPattern (entryname,skippedextensions );
        END;
        IF NOT(excludeme1 OR excludeme2) THEN
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=entry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                (* if file has no extension, add it as a marker *)
                IF Str.RCharPos(entryname,".")=MAX(CARDINAL) THEN
                    Str.Append(entryname,".");
                END;
                dbg(DEBUG,useLFN,"Included",entryname);
                len:=Str.Length(entryname);
                IF buildNewPtr(anchor,pp,len)=FALSE THEN
                    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL); (* errStorage *)
                END;
                INC(count);
                pp^.slen      := len;
                Lib.FastMove ( ADR(entryname),ADR(pp^.str),len );
            ELSE
                dbg(DEBUG,useLFN,"Ignored",entryname);
            END;
        ELSE
            dbg(DEBUG,useLFN,"Excluded",entryname);
        END;
        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(entry);
        END;
    END;
    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
    RETURN count;
END buildFileList;

(* ------------------------------------------------------------ *)

PROCEDURE doGetCurrent (useLFN:BOOLEAN;drive:SHORTCARD;
                       VAR unit:str2; VAR current:pathtype);
VAR
    rc:CARDINAL;
    longform:pathtype;
BEGIN
    Str.Concat(unit, CHR( ORD("A")-1+ORD(drive) ),colon);

    FIO.GetDir(drive,current); (* \path without u: nor trailing \ except at root *)
    IF current[1] # colon THEN Str.Prepend(current,unit); END; (* safety *)
    IF useLFN THEN
        IF w9XshortToLong(current,rc,longform) THEN (* if error, keep DOS current *)
            Str.Copy(current,longform);
        END;
    END;
    (* LFN function seems to always return "u:\*" form except at root *)
    IF current[1] = colon THEN Str.Delete(current,0,2);END; (* safety *)
    fixDirectory(current);
END doGetCurrent;

PROCEDURE makebase (useLFN:BOOLEAN;spec:pathtype;VAR basepath:pathtype);
VAR
    u,d,n,e,current:pathtype;
    unit:str2;
    drive:SHORTCARD;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Str.Concat(basepath,u,d);
    IF same(basepath,"") THEN
        drive:=FIO.GetDrive(); (* yes we could use 0 as default drive *)
        doGetCurrent(useLFN,drive,  unit,current); (* "u:" and "\" or "\*\" *)
        Str.Concat(basepath, unit,current);
    END;
END makebase;

PROCEDURE chkUD (S:pathtype):BOOLEAN ;
VAR
    u,d,n,e:pathtype;
    pb:CARDINAL;
BEGIN
    qdSplitAllPath(S , u,d,n,e);
    pb:=0;
    IF chkJoker(u) THEN INC(pb);END;
    IF chkJoker(d) THEN INC(pb);END;
    RETURN (pb=0);
END chkUD;

PROCEDURE WrFname (useLFN:BOOLEAN;F:pathtype );
BEGIN
    IF useLFN THEN
        WrStr(dquote);WrStr(F);WrStr(dquote);
    ELSE
        WrStr(F);
    END;
END WrFname;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE sound (freq,duration,pause:CARDINAL);
BEGIN
    Lib.Sound(freq);
    Lib.Delay(duration);
    Lib.NoSound();
    Lib.Delay(pause);
END sound;

PROCEDURE alarm (  );
BEGIN
    sound(55,55,100);
    sound(55,55,10);
END alarm;

PROCEDURE alert (  );
BEGIN
    sound(550,55,100);
    sound(550,55,10);
END alert;

PROCEDURE wrQuoted (S:ARRAY OF CHAR);
BEGIN
    WrLn;WrStr(dquote);WrStr(S);WrStr(dquote);
END wrQuoted;

(* ------------------------------------------------------------ *)

PROCEDURE pad (wi:INTEGER;padchar:CHAR; VAR R:ARRAY OF CHAR);
VAR
    i:CARDINAL;
BEGIN
    FOR i:=Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
END pad;

PROCEDURE lcToStr (wi:INTEGER;padchar:CHAR; n:LONGCARD):str80;
VAR
    ok:BOOLEAN;
    R:str80;
BEGIN
    Str.CardToStr ( n, R, 10, ok);
    pad (wi,padchar,R);
    RETURN R;
END lcToStr;

(* ------------------------------------------------------------ *)

CONST
    rem = ";";
    hdr = rem+" ";
VAR
    heapsizebefore : CARDINAL;

PROCEDURE markmem (  );
BEGIN
    heapsizebefore :=Storage.HeapTotalAvail(Storage.MainHeap);
END markmem;

PROCEDURE showmem(S:ARRAY OF CHAR);
CONST
    msgBefore   = hdr+"Byte(s) free before : ";
    msgUsed     = hdr+"Byte(s) grabbed     : ";
    msgFreed    = hdr+"Byte(s) released    : ";
    msgNow      = hdr+"Byte(s) free now    : ";
    w           = 6;
VAR
    heapsize    : CARDINAL; (* in PARAGRAPHS and not in bytes ! help is wrong ! *)
    n           : LONGCARD;
BEGIN
    WrLn;
    heapsize :=Storage.HeapTotalAvail(Storage.MainHeap);

(*
    n :=16 * LONGCARD(heapsizebefore);
    WrStr(msgBefore); IO.WrLngCard(n,w);WrLn;

    IF heapsizebefore < heapsize THEN
        n := 16 * LONGCARD(heapsize-heapsizebefore);
        WrStr(msgFreed); IO.WrLngCard(n,w);WrLn;
    ELSE
        n := 16 * LONGCARD(heapsizebefore-heapsize);
        WrStr(msgUsed); IO.WrLngCard(n,w);WrLn;
    END;
*)

    n :=16 * LONGCARD(heapsize);
    WrStr(msgNow); IO.WrLngCard(n,w);WrStr(" (");WrStr(S);WrStr(")");WrLn;
END showmem;

(* ------------------------------------------------------------ *)

CONST
    ignore = " ";

    Talpha = "                                "+
             "                                "+
             " ABCDEFGHIJKLMNOPQRSTUVWXYZ     "+
             " abcdefghijklmnopqrstuvwxyz     "+
             "ueaaaaeeeiiiAAEooouuyOU     "+
             "aiounN                          "+
             "                                "+
             "                                ";

PROCEDURE map (c : CHAR ;table : ARRAY OF CHAR) : CHAR;
BEGIN
    RETURN table [ ORD (c) ];
END map;

PROCEDURE removeAccents (VAR S : ARRAY OF CHAR);
VAR
    len : CARDINAL;
    i   : CARDINAL;
    c   : CHAR;
    ch  : CHAR;
BEGIN
    len := Str.Length(S);
    IF len = 0 THEN RETURN; END;         (* this WAS a problem ! *)
    FOR i := 0 TO (len - 1) DO
        c  := S[i];
        ch := map(c,Talpha);
        IF ch # ignore THEN S[i] := ch; END;
    END;
END removeAccents;

PROCEDURE isAlpha(ch:CHAR):BOOLEAN;
BEGIN
    RETURN map(ch,Talpha) # ignore;
END isAlpha;

PROCEDURE isOK (ch:CHAR;allowdigits:BOOLEAN) : BOOLEAN;
(*
CONST
    digits = "0123456789";
*)
BEGIN
    IF isAlpha(ch) THEN RETURN TRUE; END;
    IF allowdigits THEN
        RETURN (Str.CharPos(digits,ch) # MAX(CARDINAL));
    ELSE
        RETURN FALSE;
    END;
END isOK;

PROCEDURE isBlank(linemode:BOOLEAN;ch:CHAR):BOOLEAN;
VAR
    tf:BOOLEAN;
BEGIN
    IF linemode THEN
        CASE ch OF
        | cr,lf:
            tf:=TRUE;
        ELSE
            tf:=FALSE;
        END;
    ELSE
        (*
        CASE ch OF
        | space,tabul,cr,lf:
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END;
        *)
        tf:= (ORD(ch) <= ORD(space));
    END;
    RETURN tf;
END isBlank;

PROCEDURE dofixae (VAR S : ARRAY OF CHAR):CARDINAL;
CONST
    LCae = ""; (* 145 *)
    UCae = ""; (* 146 *)
VAR
    i,p : CARDINAL;
    ch:CHAR;
    replace:str2;
BEGIN
    FOR i:=1 TO 2 DO
        CASE i OF
        | 1: ch:=LCae;replace:="ae";
        | 2: ch:=UCae;replace:="AE";
        END;
        LOOP
            p:=Str.CharPos(S,ch);
            IF p=MAX(CARDINAL) THEN EXIT; END;
            Str.Subst(S,ch,replace);
        END;
    END;
    RETURN Str.Length(S);
END dofixae;

(* ------------------------------------------------------------ *)

TYPE
    pstr = POINTER TO tokendata;
    tokendata = RECORD
        total  : LONGCARD;
        len    : SHORTCARD;
        string : CHAR;
    END;
    NODE = POINTER TO tokentype; (* pointer to a wordnode structure *)
    tokentype = RECORD
        left,right : NODE;
        data       : pstr;
    END;

VAR
    gwordtree : NODE;
CONST
    mingram = 1;
    maxgram = 100;
    undefined=MAX(CARDINAL);
VAR
    (* avoid stack problems thanks to globerks *)
    gwcount,gtwcount:LONGCARD; (* unique tokens, total tokens *)
    gdump:BOOLEAN;
    gskipped,gfilesize:LONGCARD;
TYPE
    tokinfotype = RECORD
        uniques : LONGCARD;
        freq    : LONGCARD;
    END;

PROCEDURE initCounters (counts,showlist:BOOLEAN );
BEGIN
    IF counts THEN
        gwcount   :=0;
        gtwcount  :=0;
        gdump     :=showlist;
    ELSE
        gfilesize :=0;
        gskipped  :=0;
    END;
END initCounters;

PROCEDURE initTree ();
BEGIN
    gwordtree := NIL;
END initTree;

PROCEDURE killNode ( p:NODE);
VAR
    len,used:CARDINAL;
    tmp:pstr;
    pl,pr:NODE;
BEGIN
    IF p=NIL THEN RETURN;  END;
    tmp:=p^.data;
    len:=CARDINAL(tmp^.len);
    used:=SIZE(tokendata)-1+len;
    DEALLOCATE(tmp,used);
    pl:=p^.left;
    pr:=p^.right;
    DEALLOCATE(p,SIZE(tokentype));
    killNode( pl);
    killNode( pr);
END killNode;

PROCEDURE killTree ();
BEGIN
    killNode(gwordtree) ;
END killTree;

PROCEDURE strstore (VAR ok:BOOLEAN; S:ARRAY OF CHAR ): pstr;
VAR
    len,needed:CARDINAL;
    tmp:pstr;
BEGIN
    len:=Str.Length(S);
    needed:=SIZE(tokendata)-1+len;
    ok:=Storage.Available(needed);
    IF ok=FALSE THEN RETURN NIL;END;
    ALLOCATE(tmp, needed);
    tmp^.total:=1;
    tmp^.len := SHORTCARD(len);
    Lib.FastMove( ADR(S),ADR(tmp^.string),len);
    RETURN tmp;
END strstore;

PROCEDURE newnode (VAR p:NODE; snorm:ARRAY OF CHAR):BOOLEAN ;
VAR
    nlen,clen:CARDINAL;
    ok:BOOLEAN;
BEGIN
    IF Storage.Available( SIZE(tokentype) )=FALSE THEN RETURN FALSE; END;
    ALLOCATE (p,SIZE(tokentype));
    p^.left:=NIL;
    p^.right:=NIL;
    p^.data:=strstore(ok,snorm);
    RETURN ok;
END newnode;

PROCEDURE addnode (newptr:NODE);
VAR
    p:NODE;
    s1,s2:str128; (* oversized ! *)
    len:CARDINAL;
BEGIN
    IF gwordtree = NIL THEN
        gwordtree := newptr;
        RETURN;
    END;

	(* insert into binary tree for normal search *)

    p:=gwordtree;
    WHILE p # NIL DO
        len:=CARDINAL(p^.data^.len);
        Lib.FastMove( ADR(p^.data^.string),ADR(s1),len);
        s1[len]:=charnull;
        len:=CARDINAL(newptr^.data^.len);
        Lib.FastMove( ADR(newptr^.data^.string),ADR(s2),len);
        s2[len]:=charnull;
        IF Str.Compare(s1,s2) < 0 THEN
            IF p^.left # NIL THEN
                p := p^.left;
            ELSE
                p^.left := newptr;
                p := NIL;
            END;
        ELSE
            IF p^.right # NIL THEN
                p := p^.right;
            ELSE
                p^.right := newptr;
                p := NIL;
            END;
        END;
    END;

END addnode;

PROCEDURE findMatch (ID:ARRAY OF CHAR):NODE;
VAR
    p   : NODE;
    len : CARDINAL;
    rc  : INTEGER;

    matchID : str128; (* oversized ! *)
BEGIN
    p := gwordtree;
    WHILE p # NIL DO
        len:=CARDINAL(p^.data^.len);
        Lib.FastMove(ADR(p^.data^.string)  ,ADR(matchID), len);
        matchID[len]:=charnull;
        rc:=Str.Compare(matchID,ID);
        IF rc=0 THEN
            RETURN p;
        END;
        IF rc < 0 THEN
            p:=p^.left;
        ELSE
            p:=p^.right;
        END;
    END;
    RETURN NIL;
END findMatch;

PROCEDURE getFmtStr (VAR R:ARRAY OF CHAR;
          alt,withdelta:BOOLEAN   );
CONST
    sInfo        = 'L= $,  N= $,  C= $,  S= "$"';
    sInfoAlt     = 'L= $,  N= $,  C= $,  S= $';
    sInfoFull    = 'L= $,  N= $,  C= $,  D=$,  S= "$"';
    sInfoFullAlt = 'L= $,  N= $,  C= $,  D=$,  S= $';
VAR
    S:str128;
BEGIN
    IF alt THEN
        IF withdelta THEN
            S:=sInfoFullAlt;
        ELSE
            S:=sInfoAlt;
        END;
    ELSE
        IF withdelta THEN
            S:=sInfoFull;
        ELSE
            S:=sInfo;
        END;
    END;
    Str.Copy(R,S);
END getFmtStr;


PROCEDURE dumpNode (VAR errcode:CARDINAL;
                   showdelta:BOOLEAN;p:NODE);
VAR
    len:CARDINAL;
    S:str128; (* oversized ! *)
    R:str256; (* oversized just in case ! *)
    esc:BOOLEAN;
    vN,vC:LONGCARD;
BEGIN
    IF errcode # errNone THEN RETURN; END;
    IF p=NIL THEN RETURN;  END;

    (*
    WrStr("// dumpNode() ");
    IF lnode THEN
        WrStr("LEFT");
    ELSE
        WrStr("RIGHT");
    END;
    WrLn;
    *)

    IF gdump THEN
        len:=CARDINAL(p^.data^.len);
        Lib.FastMove(ADR(p^.data^.string),ADR(S),len);
        S[len]:=charnull;

        getFmtStr(R,FALSE,showdelta);
        vN:=p^.data^.total;
        vC:=vN * LONGCARD(len);

        Str.Subst(R,"$",lcToStr(2," ", LONGCARD(len) ));
        Str.Subst(R,"$",lcToStr(wi," ",vN ));
        Str.Subst(R,"$",lcToStr(wi," ",vC ));
        IF showdelta THEN
        Str.Subst(R,"$",lcToStr(wi," ", vC-vN));
        END;
        Str.Subst(R,"$",S);
        WrStr(R);WrLn;
    END;

    INC(gwcount);
    INC(gtwcount,p^.data^.total);

    IF ChkEscape() THEN errcode:=errAborted; END;
    dumpNode(errcode, showdelta,p^.left  );
    dumpNode(errcode, showdelta,p^.right );
END dumpNode;

PROCEDURE dumpTree (VAR errcode:CARDINAL;
                   showdelta:BOOLEAN);
CONST
    msg="Working...";
BEGIN
    IF NOT(gdump) THEN video(msg,TRUE);END;
    errcode:=errNone;
    dumpNode(errcode,showdelta,gwordtree ) ;
    IF NOT(gdump) THEN video(msg,FALSE);END;
END dumpTree;

(* ------------------------------------------------------------ *)

CONST
    sBytesRead   = hdr+"Bytes read              : ";
    sFilteredOut = hdr+"Characters filtered out : ";
    sTokens      = hdr+"Tokens                  : ";
    sWords       = hdr+"Unique words            : ";
    sTotalWords  = hdr+"Total words             : ";

    sPath        = hdr+"File specification      : ";
    sUseLFN      = hdr+"LFN filesystem          : ";
    sCase        = hdr+"Ignore case             : ";
    sAccents     = hdr+"Ignore accents          : ";
    sAE          = hdr+'Expand "" and ""      : ';
    sLinemode    = hdr+"Line mode               : ";
    sSeparators  = hdr+"Token separators        : ";
    sSet         = hdr+"Accept [A..Z0..9]       : ";
    sPATmode     = hdr+"Pattern analysis mode   : ";
    sSTRasTOK    = hdr+"Parse string as a token : ";
    sREMAPSTR    = hdr+"Remapping string        : ";
    sMode        = hdr+"Word counter mode       : ";
    sRange       = hdr+"n-gram range            : ";
    sCumulative  = hdr+"Cumulative results      : ";
    sRaw         = hdr+"Raw output              : ";

PROCEDURE dumpStats (showme,wordmode,outputnow:BOOLEAN   );
VAR
    S:str128;
BEGIN
    IF NOT(showme) THEN RETURN; END;
    WrLn;
    Str.Concat(S,sBytesRead     ,lcToStr(wi," ",gfilesize));WrStr(S);WrLn;
    Str.Concat(S,sFilteredOut   ,lcToStr(wi," ",gskipped));WrStr(S);WrLn;
    IF wordmode THEN
        Str.Concat(S,sTotalWords,lcToStr(wi," ",gtwcount));WrStr(S);WrLn;
      IF NOT(outputnow) THEN
        Str.Concat(S,sWords     ,lcToStr(wi," ",gwcount));WrStr(S);WrLn;
      END;
    ELSE
      IF NOT(outputnow) THEN
        Str.Concat(S,sTokens      ,lcToStr(wi," ",gwcount));WrStr(S);WrLn;
      END;
    END;
END dumpStats;

PROCEDURE dumpFlags (showme,useLFN,
                    keepcase,keepaccents,keepae,modulamode,useblanks,allowdigits,
                    wordmode,cumulative,outputnow,PATmode,STRasTOK,linemode:BOOLEAN;
                    lowerlen,upperlen:CARDINAL;sRemap:str5;filespec:pathtype );
    MODULE message;
    IMPORT Str;
    FROM IO IMPORT WrStr,WrLn;
    EXPORT msgbool;

    PROCEDURE msgbool (flag:BOOLEAN;S:ARRAY OF CHAR);
    BEGIN
        WrStr(S);
        IF flag THEN
            WrStr("YES");
        ELSE
            WrStr("no");
        END;
        WrLn;
    END msgbool;

    END message;
VAR
    S : str128;
BEGIN
    IF NOT(showme) THEN RETURN;END;
    WrStr(sPath);WrFname(useLFN,filespec);WrLn;
    msgbool(useLFN           ,sUseLFN);

    msgbool(NOT(keepcase)    ,sCase);
    msgbool(NOT(keepaccents) ,sAccents);
    msgbool(NOT(keepae)      ,sAE);
    msgbool(linemode         ,sLinemode);

    IF modulamode THEN
        S:=sModulaSet;
    ELSIF useblanks THEN
        S:=sBlanksSet;
    ELSE
        S:=sNotAZset;
    END;
    WrStr(sSeparators);WrStr(S);WrLn;

    msgbool(allowdigits      ,sSet);
    msgbool(wordmode         ,sMode);
    msgbool(PATmode          ,sPATmode);
    msgbool(STRasTOK         ,sSTRasTOK);
    WrStr(sREMAPSTR);WrStr(dquote);WrStr(sRemap);WrStr(dquote);WrLn;
    msgbool(cumulative       ,sCumulative);
    IF NOT(wordmode) THEN
        WrStr(sRange);
        Str.Concat(S,lcToStr(1," ",LONGCARD (lowerlen)),"..");
        Str.Append(S,lcToStr(1," ",LONGCARD (upperlen)));
        Str.Prepend(S,"[");Str.Append(S,"]");
        WrStr(S);WrLn;
    END;
    msgbool(outputnow        ,sRaw);
END dumpFlags;

(* ------------------------------------------------------------ *)

PROCEDURE scanNode (VAR remaining,uniques,total:LONGCARD;
                   len:SHORTCARD;p:NODE);
BEGIN
    IF p=NIL THEN RETURN;  END;

    IF len = p^.data^.len THEN
        INC(uniques);
        INC(total,p^.data^.total);
        DEC(remaining);
    END;

    scanNode(remaining,uniques,total,len,p^.left);
    scanNode(remaining,uniques,total,len,p^.right);

END scanNode;

CONST
    sfreqInfo   = hdr+"L= $,  uniques= $,  total= $";
    sfreqFinal  = hdr+"   $            $          $";

PROCEDURE dumpFreq (showme:BOOLEAN   );
VAR
    remaining,uniques,total,granduniques,grandtotal:LONGCARD;
    len:SHORTCARD;
    R:str128;
BEGIN
    IF NOT(showme) THEN RETURN; END;
    WrLn;
    granduniques:=0;
    grandtotal:=0;
    remaining:=gwcount;
    len:=1;
    WHILE remaining > 0 DO
        uniques:=0;
        total:=0;
        scanNode(remaining,uniques,total,len,gwordtree);
        INC(granduniques,uniques);
        INC(grandtotal,total);
        Str.Copy(R,sfreqInfo);
        Str.Subst(R,"$",lcToStr(2," ", LONGCARD(len) ));
        Str.Subst(R,"$",lcToStr(wi," ",  uniques ) );
        Str.Subst(R,"$",lcToStr(wi," ",  total ) );
        WrStr(R);WrLn;

        INC(len);
    END;

    Str.Copy(R,sfreqFinal);
    Str.Subst(R,"$","  " ) ;
    Str.Subst(R,"$",widashes);
    Str.Subst(R,"$",widashes);
    WrStr(R);WrLn;

    Str.Copy(R,sfreqFinal);
    Str.Subst(R,"$","  " ) ;
    Str.Subst(R,"$",lcToStr(wi," ",  granduniques ) );
    Str.Subst(R,"$",lcToStr(wi," ",  grandtotal ) );
    WrStr(R);WrLn;

END dumpFreq;

(* ------------------------------------------------------------ *)

CONST
    IObufferSize = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = IObufferSize;
VAR
    ioBuffer : ARRAY [1..IObufferSize] OF BYTE;

(* ------------------------------------------------------------ *)

PROCEDURE strToPat (sRemap:str5;STRasTOK:BOOLEAN;len:CARDINAL;
                   VAR token:ARRAY OF CHAR);
VAR
    i:CARDINAL;
    tmptoken:str128; (* should DO *)
    ch:CHAR;
BEGIN
    tmptoken:="";
    FOR i:= 1 TO len DO
        CASE token[i-1] OF
        | "A".."Z" : ch:=sRemap[ndxchUPPER];
        | "a".."z" : ch:=sRemap[ndxchLOWER];
        | "0".."9" : ch:=sRemap[ndxchDIGIT];
        | blank:
            IF STRasTOK THEN
                     ch:=sRemap[ndxchSPACE];
            ELSE
                     ch:=sRemap[ndxchSYMBOL];
            END;
        ELSE
                     ch:=sRemap[ndxchSYMBOL];
        END;
        Str.Append(tmptoken,ch);
    END;
    Str.Copy(token,tmptoken);
END strToPat;

PROCEDURE isQuote ( ch:CHAR ):BOOLEAN;
VAR
    tf:BOOLEAN;
BEGIN
    CASE ch OF
    | dquote,singlequote:
        tf:=TRUE;
    ELSE
        tf:=FALSE;
    END;
    RETURN tf;
END isQuote;

(* //FIXME : working but overcomplicated, should be rewritten *)

PROCEDURE getToken (VAR done:BOOLEAN;VAR len:CARDINAL; VAR token:ARRAY OF CHAR;
                   DEBUG,PATmode,STRasTOK,linemode,
                   modulamode,useblanks,allowdigits:BOOLEAN;
                   hnd : FIO.File;sRemap:str5);
CONST
    m2separators = modulaseparators+dquote+quote;
VAR
    ch    : CHAR;
    state : (waiting,grabtok,grabstr);
    got   : CARDINAL;
    ok    : BOOLEAN;
    tokmode : (onlyblanks,notAZ,m2,pattern,patternstr);
    tmptoken : str128; (* should DO *)
    sepch : CHAR;
BEGIN
    done:=FALSE;
    IF PATmode THEN
        IF STRasTOK THEN
            tokmode := patternstr; (* string as a single token *)
        ELSE
            tokmode := pattern;    (* ignore quotes *)
        END;
    ELSE
        IF modulamode THEN
            tokmode:=m2;
        ELSIF useblanks THEN (* also for -p PATmode which is set earlier *)
            tokmode:=onlyblanks;
        ELSE
            tokmode:=notAZ;
        END;
    END;

    Str.Copy(token,"");
    state := waiting;
    len   := 0;
    LOOP
        got := FIO.RdBin(hnd,ch,1);
        IF got = 0 THEN done:=TRUE; EXIT; END;
        CASE state OF
        | waiting :
            CASE tokmode OF
            | m2:            ok:=NOT(isBlank(linemode,ch) OR Belongs(m2separators,ch));
            | onlyblanks:    ok:=NOT(isBlank(linemode,ch));
            | notAZ:         ok:=isOK(ch,allowdigits);
            | pattern:       ok:=NOT(isBlank(linemode,ch));
            | patternstr:
                IF isQuote(ch) THEN
                    sepch:=ch;
                    ok:=TRUE;
                ELSE
                    ok:=NOT(isBlank(linemode,ch));
                END;
            END;

            IF ok THEN
                Str.Copy(token,ch);
                len:=1;
                IF tokmode=patternstr THEN
                    state :=grabstr;
                ELSE
                    state :=grabtok;
                END;
            ELSE
                INC(gskipped);
            END;
        | grabtok :
            CASE tokmode OF
            | m2:            ok:=NOT(isBlank(linemode,ch) OR Belongs(m2separators,ch));
            | onlyblanks:    ok:=NOT(isBlank(linemode,ch));
            | notAZ:         ok:=isOK(ch,allowdigits);
            | pattern:       ok:=NOT(isBlank(linemode,ch));
            END;
            IF ok THEN
                Str.Append(token,ch);
                INC(len);
            ELSE
                INC(gskipped);
                IF tokmode = pattern THEN
                        IF isQuote(token[0]) THEN
                            Str.Delete(token,0,1);
                            DEC(len);
                            INC(gskipped);
                        END;
                    IF len # 0 THEN (* safety *)
                        IF isQuote(token[len-1]) THEN
                            token[len-1]:=charnull;
                            DEC(len);
                            INC(gskipped);
                        END;
                    END;
                END;
                EXIT; (* done *)
            END;
        | grabstr: (* -p only *)
            IF ch=sepch THEN
                Str.Append(token,ch);
                INC(len);
                ok:=FALSE;
            ELSE
                CASE ch OF
                | blank,tabul:
                    IF isQuote(sepch) THEN
                        ok:=TRUE;
                    ELSE
                        ok:=NOT(isBlank(linemode,ch));
                    END;
                | cr,lf,CHR(26): (* ctrl-Z *)
                    ok:=FALSE;
                ELSE
                    ok:=NOT(isBlank(linemode,ch));
                END;
            END;

            IF ok THEN
                Str.Append(token,ch);
                INC(len);
            ELSE
                INC(gskipped);
                ok:=Str.Match(token,dquote+star+dquote);
                ok:=ok OR Str.Match(token,singlequote+star+singlequote);
                ok:=ok AND isQuote(sepch);
                IF ok THEN
                        Str.Delete(token,0,1);
                        DEC(len);
                        INC(gskipped);
                    IF len # 0 THEN (* safety *)
                        token[len-1]:=charnull; (* fix trailing quote *)
                        DEC(len);
                        INC(gskipped);
                    END;
                END;
                EXIT; (* done *)
            END;
        END;
    END;
        IF DEBUG THEN
            tmptoken:=' ::: tok = "~"';
            Str.Subst(tmptoken,"~",token);
            IO.WrCard(len,3);WrStr(tmptoken);WrLn;
        END;
    IF PATmode THEN
        strToPat (sRemap,STRasTOK,len,token);
        IF DEBUG THEN
            tmptoken:=' ::: pat = "~"';
            Str.Subst(tmptoken,"~",token);
            IO.WrCard(len,3);WrStr(tmptoken);WrLn;
        END;
    END;
END getToken;

PROCEDURE buildGrams (VAR errcode:CARDINAL;VAR einfo:ARRAY OF CHAR;
                     lowerlen,upperlen:CARDINAL;
                     DEBUG,keepcase,keepaccents,keepae,modulamode,
                     useblanks,allowdigits,wordmode,outputnow,
                     verbose,showdelta,PATmode,STRasTOK,linemode,useLFN:BOOLEAN;
                     sRemap:str5;f:pathtype);
CONST
    msgProcessing = "Processing ";
VAR
    hin:FIO.File;
    token:str128; (* should do ! *)
    i,len:CARDINAL;
    j,lim:INTEGER;
    ptr,matchptr:NODE;
    ok:BOOLEAN;
    msg:str128;
    gram:str128; (* str16 was enough but leave room for unlikely extension ! *)
    fsize:LONGCARD;
    done:BOOLEAN;
BEGIN
    IF DEBUG THEN verbose:=FALSE;END;
    errcode:=errNone;

    FIO.EOF:=FALSE; (* silly JPI quirks/bugs ! *)
    hin:=fileOpenRead(useLFN,f);
    FIO.AssignBuffer(hin,ioBuffer);

    fsize:=FIO.Size(hin);   (* handle is ok for both DOS and LFN *)

    Str.Concat(msg,msgProcessing,f);Str.Append(msg," ");
    IF NOT(outputnow) THEN video(msg,TRUE); END;
    IF verbose THEN completed(cmdInit,fsize );END;

    INC(gfilesize,fsize);

    LOOP
        IF FIO.EOF THEN EXIT; END;
        IF verbose THEN completed(cmdShow,FIO.GetPos(hin) ); END;
        getToken (done,len,token,
                   DEBUG,PATmode,STRasTOK,linemode,modulamode,useblanks,allowdigits,
                   hin,sRemap);
        IF done THEN EXIT;END; (* //V11C was IF len=0 *)
        IF NOT(keepae) THEN len:=dofixae(token);END;
        IF keepcase THEN (* //V12C always TRUE for -p *)
            IF keepaccents = FALSE THEN removeAccents(token); END;
        ELSE
            IF keepaccents THEN
                LowerCaseAlt(token);
            ELSE
                LowerCase(token);
            END;
        END;
        IF wordmode THEN
            IF outputnow THEN
                IF PATmode THEN strToPat(sRemap,STRasTOK,Str.Length(token),token);END;
                wrQuoted(token);
                INC(gtwcount); (* update total words is ok here *)
            ELSE
(* WrStr("// findmatch() for ");WrStr(token);WrLn; *)
                matchptr:=findMatch(token);
                IF matchptr=NIL THEN
                    IF newnode(ptr, token)=FALSE THEN
                        errcode:=errAllocate;
                        Str.Concat(einfo,token," (");Str.Append(einfo,f);Str.Append(einfo,")");
                        EXIT;
                    END;
(* WrStr("//   addnode()");WrLn; *)
                    addnode(ptr);
                ELSE
                    INC( matchptr^.data^.total );
                END;
            END;
        ELSE
            FOR i:=lowerlen TO upperlen DO
                lim:=INTEGER(len)-INTEGER(i)+1;
                j:=0;
                LOOP
                    IF j>=lim THEN EXIT; END;
                    Str.Slice(gram,token, CARDINAL(j),CARDINAL(i) );
                    IF outputnow THEN
                        wrQuoted(gram); (* update gwcount unique tokens would be a nonsense here *)
                    ELSE
                        matchptr:=findMatch(gram);
                        IF matchptr=NIL THEN
                            IF newnode(ptr, gram)=FALSE THEN
                                errcode:=errAllocate;
                                Str.Concat(einfo,token," (");Str.Append(einfo,f);Str.Append(einfo,")");
                                EXIT;
                            END;
                            addnode(ptr);
                        ELSE
                            INC( matchptr^.data^.total );
                        END;
                    END;
                    INC(j);
                END;
                IF errcode # errNone THEN EXIT;END;
            END;
        END;
        IF outputnow THEN
            IF ChkEscape() THEN errcode:=errAborted;EXIT; END;
        END;
    END;
    fileClose(useLFN,hin);

    IF verbose THEN completed(completedEnd,0); END;
    IF NOT(outputnow) THEN video(msg,FALSE); END;
END buildGrams;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minlen = 1; (* 0 would be a nonsense here ! and we'll need 1-1 *)
    maxlen = 128; (* should do ! *)
TYPE
    statstype = RECORD
        uniques,total:LONGCARD;
    END;
VAR
    stats : ARRAY [minlen..maxlen] OF statstype;

PROCEDURE dumpFreqRemade (showme:BOOLEAN   );
VAR
    remaining,uniques,total,granduniques,grandtotal:LONGCARD;
    R:str128;
    len,lastlen:CARDINAL;
BEGIN
    IF NOT(showme) THEN RETURN; END;

    lastlen:=maxlen;
    LOOP
        IF stats[lastlen].uniques # 0 THEN EXIT; END;
        DEC(lastlen);
        IF lastlen < minlen THEN RETURN; END;
    END;

    WrLn;
    granduniques:=0;
    grandtotal:=0;

    FOR len:=minlen TO lastlen DO
        uniques:=stats[len].uniques;
        total:=stats[len].total;
        INC(granduniques,uniques);
        INC(grandtotal,total);
        Str.Copy(R,sfreqInfo);
        Str.Subst(R,"$",lcToStr(2," ", LONGCARD(len) ));
        Str.Subst(R,"$",lcToStr(wi," ",  uniques ) );
        Str.Subst(R,"$",lcToStr(wi," ",  total ) );
        WrStr(R);WrLn;
    END;

    Str.Copy(R,sfreqFinal);
    Str.Subst(R,"$","  " ) ;
    Str.Subst(R,"$",widashes);
    Str.Subst(R,"$",widashes);
    WrStr(R);WrLn;

    Str.Copy(R,sfreqFinal);
    Str.Subst(R,"$","  " ) ;
    Str.Subst(R,"$",lcToStr(wi," ",  granduniques ) );
    Str.Subst(R,"$",lcToStr(wi," ",  grandtotal ) );
    WrStr(R);WrLn;
END dumpFreqRemade;

PROCEDURE parseLines (VAR PB:ARRAY OF CHAR;
                     showdelta,useLFN:BOOLEAN;f:pathtype):CARDINAL;
VAR
    hin:FIO.File;
    S,sPrev,R:str128;
    rc,len:CARDINAL;
    total:LONGCARD;
    what:(ignore,valid);
    vN,vC:LONGCARD;
BEGIN
    FOR len:=minlen TO maxlen DO stats[len].uniques:=0;stats[len].total:=0;END;

    sPrev:="";
    total:=0;
    len:=0;
    rc:=errNone;
    FIO.EOF:=FALSE; (* silly JPI quirks/bugs ! *)
    hin:=fileOpenRead(useLFN,f);
    FIO.AssignBuffer(hin,ioBuffer);
    LOOP
        IF FIO.EOF THEN EXIT;END;
        FIO.RdStr(hin,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE S[0] OF
        | charnull, rem :
            what:=ignore; (* empty or remark *)
        | dquote :
            what:=valid;
            IF same(S,sPrev) THEN
                INC(total);
            ELSE
                IF len # 0 THEN
                    INC( stats[len].uniques);
                    INC( stats[len].total, total);
                    getFmtStr(R,TRUE,showdelta); (* no inner quotes ! *)
                    vN:=total;
                    vC:=vN * LONGCARD(len);
                    Str.Subst(R,"$",lcToStr( 2," ", LONGCARD( len )));
                    Str.Subst(R,"$",lcToStr(wi," ", vN ));
                    Str.Subst(R,"$",lcToStr(wi," ", vC ));
                    IF showdelta THEN
                    Str.Subst(R,"$",lcToStr(wi," ", vC-vN));
                    END;
                    Str.Subst(R,"$",sPrev);
                    WrStr(R);WrLn;
                END;
                Str.Copy(sPrev,S);
                total:=1;
                len:=Str.Length(sPrev)-2; (* no quotes ! *)
            END;
        ELSE
            Str.Copy(PB,S);
            rc:=errProblem;
            EXIT;
        END;
        IF ChkEscape() THEN
            Str.Copy(PB,"");
            rc:=errAborted;
            EXIT;
        END;
    END;
    fileClose(useLFN,hin);

    IF rc = errNone THEN
        IF total # 0 THEN
                IF len # 0 THEN
                    INC( stats[len].uniques);
                    INC( stats[len].total, total);
                    getFmtStr(R,TRUE,showdelta); (* no inner quotes ! *)
                    vN:=total;
                    vC:=total * LONGCARD(len);
                    Str.Subst(R,"$",lcToStr( 2," ", LONGCARD( len )));
                    Str.Subst(R,"$",lcToStr(wi," ", vN ));
                    Str.Subst(R,"$",lcToStr(wi," ", vC ));
                    IF showdelta THEN
                    Str.Subst(R,"$",lcToStr(wi," ", vC-vN));
                    END;
                    Str.Subst(R,"$",sPrev);
                    WrStr(R);WrLn;
                END;
        END;
    END;
    RETURN rc;
END parseLines;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE getCard (VAR v:CARDINAL;
                   lower,upper: CARDINAL;S:ARRAY OF CHAR):BOOLEAN;
VAR
    lc:LONGCARD;
    ok:BOOLEAN;
BEGIN
    IF GetLongCard(S,lc)=FALSE THEN RETURN FALSE; END;
    IF lc < LONGCARD(lower) THEN RETURN FALSE; END;
    IF lc > LONGCARD(upper) THEN RETURN FALSE; END;
    v:=CARDINAL(lc);
    RETURN TRUE;
END getCard;

(* ------------------------------------------------------------ *)

VAR
    parmcount,i,opt,lowerlen,upperlen,exactlen:CARDINAL;
    Z,S,R,filespec,basedir,fname:pathtype;
    state : (waiting,gotspec);
    keepcase,keepaccents,keepae,useblanks,allowdigits:BOOLEAN;
    showsummary,showfreq,showflags,showlist,wordmode,cumulative:BOOLEAN;
    linemode,STRasTOK,PATmode:BOOLEAN;
    modulamode,outputnow,doanalysis,verbose,noskip,showdelta:BOOLEAN;
    useLFN,DEBUG:BOOLEAN;
    errcode:CARDINAL;
    einfo:str128;
    ptr,anchor: pFname;
    count:CARDINAL;
    sRemap:str5;
BEGIN
    FIO.IOcheck := FALSE;
    WrLn; (* yes, here now for help, errors, etc. *)

    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;

    DEBUG       := FALSE;
    lowerlen    := undefined;
    upperlen    := undefined;
    exactlen    := undefined;
    keepcase    := TRUE;
    keepaccents := TRUE;
    keepae      := FALSE;
    useblanks   := FALSE;
    allowdigits := FALSE;
    wordmode    := TRUE;
    cumulative  := TRUE;
    showflags   := TRUE;
    showlist    := TRUE;
    showfreq    := TRUE;
    showsummary := TRUE;
    outputnow   := FALSE;
    doanalysis  := FALSE;
    modulamode  := FALSE;
    verbose     := TRUE;
    showdelta   := FALSE;
    PATmode     := FALSE;
    STRasTOK    := TRUE;
    linemode    := FALSE;
    Str.Copy(sRemap,defaultremap);
    useLFN      := TRUE;

    state :=waiting;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "U:"+delim+"UPPER:"+delim+
                                  "L:"+delim+"LOWER:"+delim+
                                  "E:"+delim+"EXACT:"+delim+
                                  "C"+delim+"IGNORECASE"+delim+
                                  "A"+delim+"IGNOREACCENTS"+delim+
                                  "F"+delim+"NOFIXAE"+delim+
                                  "B"+delim+"BLANKS"+delim+
                                  "D"+delim+"DIGITS"+delim+
                                  "Q"+delim+"QUIET"+delim+
                                  "G"+delim+"GRAM"+delim+
                                  "I"+delim+"INDIVIDUAL"+delim+
                                  "X"+delim+"NOFLAGS"+delim+
                                  "L"+delim+"NOLIST"+delim+
                                  "T"+delim+"NOTABLE"+delim+
                                  "S"+delim+"NOSUMMARY"+delim+
                                  "O"+delim+"OUTPUT"+delim+
                                  "M"+delim+"MAKE"+delim+"BUILD"+delim+ (* BUILD was OO *)
                                  "K"+delim+"CODE"+delim+"MODULA"+delim+
                                  "LFN"+delim+
                                  "Z"+delim+"DELTA"+delim+
                                  "P"+delim+"PATTERN"+delim+"W"+delim+"PW"+delim+"PASSWORD"+delim+
                                  "Y"+delim+"QUOTED"+delim+
                                  "PP"+delim+"PY"+delim+"YY"+delim+
                                  "R:"+delim+"USER:"+delim+"REMAP:"+delim+
                                  "OO"+delim+
                                  "N"+delim+"NL"+delim+"CRLF"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5:    IF exactlen # undefined THEN abort(errConflict,"");END;
                      IF getCard (upperlen, mingram,maxgram,R)=FALSE THEN
                          abort(errUpperLen,S);
                      END;
            | 6,7:    IF exactlen # undefined THEN abort(errConflict,"");END;
                      IF getCard (lowerlen, mingram,maxgram,R)=FALSE THEN
                          abort(errLowerLen,S);
                      END;
            | 8,9:    IF ((lowerlen # undefined) OR (upperlen # undefined)) THEN
                          abort(errConflict,"");
                      END;
                      IF getCard (exactlen, mingram,maxgram,R)=FALSE THEN
                          abort(errExactLen,S);
                      END;
            | 10,11:            keepcase    := FALSE;
            | 12,13:            keepaccents := FALSE;
            | 14,15:            keepae      := TRUE;
            | 16,17:            useblanks   := TRUE;
            | 18,19:            allowdigits := TRUE;
            | 20,21:            verbose     := FALSE;
            | 22,23:            wordmode    := FALSE;
            | 24,25:            cumulative  := FALSE;
            | 26,27:            showflags   := FALSE;  (* -x *)
            | 28,29:            showlist    := FALSE;
            | 30,31:            showfreq    := FALSE;  (* -t *)
            | 32,33:            showsummary := FALSE;  (* -s *)
            | 34,35:            outputnow   := TRUE;   (* -o *)
            | 36,37,38:         doanalysis  := TRUE;
            | 39,40,41:         modulamode  := TRUE;
            | 42:               useLFN      := FALSE;
            | 43,44:            showdelta   := TRUE;
            | 45,46,47,48,49:   PATmode     := TRUE;
            | 50,51:            STRasTOK    := FALSE;
            | 52,53,54:         PATmode     := TRUE;
                                STRasTOK    := FALSE;
            | 55,56,57:         GetString(S,Z);
                                IF Str.Length(Z) # remapcount THEN
                                    abort(errBadRemap,S);
                                END;
                                Str.Copy(sRemap,Z);
            | 58 :              outputnow   := TRUE;   (* -o *)
                                showflags   := FALSE;  (* -x *)
                                showfreq    := FALSE;  (* -t *)
                                showsummary := FALSE;  (* -s *)
            | 59,60,61 :        linemode    := TRUE;
            | 62 :              DEBUG       := TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting : Str.Copy(filespec,S);
            | gotspec : abort(errTooManyParms,S);
            END;
            INC (state);
        END;
    END;
    IF state=waiting THEN abort(errMissingSpec,"");END;

    useLFN := ( useLFN AND fileSupportLFN() );

    dbg(DEBUG,useLFN,"Filespec 1",filespec);
    IF chkUD(filespec)=FALSE THEN abort(errBadSpec,filespec);END;

    IF same(filespec,dot) THEN Str.Copy(filespec,stardotstar); END;
    IF same(filespec,star) THEN Str.Copy(filespec,stardotstar); END;
    IF Str.Match(filespec,"*"+antislash) THEN Str.Append(filespec,stardotstar); END;
    IF Str.RCharPos(filespec,dot)=MAX(CARDINAL) THEN Str.Append(filespec,dotstar);END;

    IF chkJoker(filespec)=FALSE THEN
        IF fileIsDirectorySpec ( useLFN,filespec) THEN
            fixDirectory(filespec);
            Str.Append(filespec,stardotstar);
        END;
    END;
    dbg(DEBUG,useLFN,"Filespec 2",filespec);
    makebase(useLFN,filespec,basedir);
    dbg(DEBUG,useLFN,"Basedir",basedir);

    noskip:=doanalysis;

    initList(anchor);
    count := buildFileList( anchor,noskip,useLFN,DEBUG,filespec);
    CASE count OF
    | MAX(CARDINAL): abort(errTooManyMatches,filespec);
    | 0 :            abort(errNoMatch,filespec);
    END;

(* ------------------------------------------------------------ *)

    IF PATmode THEN (* silently force and ignore *)
        (* -b forced, -g -d -c -f ignored *)
        useblanks   := TRUE;  (* -b *)

        (* restore defaults *)
        wordmode    := TRUE;  (* -g *)
        allowdigits := FALSE; (* -d *)
        keepcase    := TRUE;  (* -c *)
        keepae      := FALSE; (* -f *)
    ELSE
        STRasTOK    := TRUE;
    END;

(* ------------------------------------------------------------ *)

    IF doanalysis THEN
        IF count # 1 THEN abort(errJoker,filespec);END;
        ptr:=anchor;
        getStr(filespec,ptr);
        Str.Concat(fname,basedir,filespec);
        dbg(DEBUG,useLFN,"Fname",fname);
        i:=parseLines(R, showdelta,useLFN,filespec);
        IF i # errNone THEN abort(i,R);END;
        dumpFreqRemade(showfreq);
        abort(errNone,"");
    END;

(* ------------------------------------------------------------ *)

    IF wordmode THEN
        IF lowerlen # undefined THEN abort(errMode,"");END;
        IF upperlen # undefined THEN abort(errMode,"");END;
        IF exactlen # undefined THEN abort(errMode,"");END;
    END;

    IF lowerlen = undefined THEN lowerlen := mingram; END;
    IF upperlen = undefined THEN upperlen := mingram+1; END;
    IF exactlen = undefined THEN
        IF upperlen < lowerlen THEN
            i:=lowerlen;
            lowerlen:=upperlen;
            upperlen:=i;
        END;
    ELSE
        lowerlen:=exactlen;
        upperlen:=exactlen;
    END;

    IF (useblanks AND modulamode) THEN abort(errNonsense,"-b and -k");END;

    IF (allowdigits AND useblanks) THEN abort(errNonsense,"-b and -d");END;

    dumpFlags (showflags,useLFN,
              keepcase,keepaccents,keepae,modulamode,useblanks,allowdigits,
              wordmode,cumulative,outputnow,PATmode,STRasTOK,linemode,
              lowerlen,upperlen,sRemap,filespec);

    IF DEBUG THEN showmem( "enter");END;

    IF outputnow THEN verbose:=FALSE; END;

    IF cumulative THEN initTree();initCounters(FALSE  ,FALSE);END;

    ptr:=anchor;
    WHILE ptr # NIL DO
        getStr(filespec,ptr);
        Str.Concat(fname,basedir,filespec);
        dbg(DEBUG,useLFN,"Fname",fname);

        IF NOT(cumulative) THEN
            WrLn;WrStr(hdr+"File : ");WrFname(useLFN,fname);WrLn;
            initTree();initCounters( FALSE , FALSE);
        END;

        buildGrams (errcode,einfo,
                   lowerlen,upperlen,
                   DEBUG,keepcase,keepaccents,keepae,modulamode,
                   useblanks,allowdigits,wordmode,outputnow,
                   verbose,showdelta,PATmode,STRasTOK,linemode,useLFN,
                   sRemap,fname);
        IF errcode # errNone THEN abort(errcode,einfo);END;

        IF DEBUG THEN showmem( fname ); END;

        IF NOT(cumulative) THEN
            WrLn;
            IF NOT(outputnow) THEN
                initCounters(TRUE ,showlist);
                dumpTree(errcode,showdelta); IF errcode#errNone THEN abort(errAborted,"");END;
                dumpFreq(showfreq);
            END;
            dumpStats(showsummary,wordmode,outputnow);
            IF outputnow THEN
                initCounters(TRUE ,showlist);
            END;
            killTree();
            IF DEBUG THEN showmem( "tree killed");END;
        END;

        ptr:=ptr^.next;
    END;
    freeList(anchor);

    IF cumulative THEN
        WrLn;
        IF NOT(outputnow) THEN
            initCounters(TRUE ,showlist );
            dumpTree(errcode,showdelta); IF errcode#errNone THEN abort(errAborted,"");END;
            dumpFreq(showfreq);
        END;
        dumpStats(showsummary,wordmode,outputnow);
        killTree();
    END;
    IF DEBUG THEN showmem( "exit");END;

    abort(errNone,"");
END wCount.

