(* ---------------------------------------------------------------
Title         Q&D Expand CSV
Overview      see help
Usage         see help
Notes
Bugs
Wish List     handle quoted inner components such as
              ;"*;";"foo;""bar;""baz;" ?
              possibly later (we don't need this for now)

              file expected to be in rational format i.e. unique separators

--------------------------------------------------------------- *)

MODULE xCSV;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
    extCSV          = ".CSV";
    strMySep        = " ::: ";
    strField        = "::: ~ ::: ";
CONST
    ProgEXEname     = "XCSV";
    ProgTitle       = "Q&D Expand CSV";
    ProgVersion     = "v1.0b";
    ProgCopyright   = "by PhG";
    Banner          = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errTooManyParms = 3;
    errJoker        = 4;
    errNotFound     = 5;
    errParm         = 6;
    errConflict     = 7;
    errSep          = 8;
    errSepConflict  = 9;
    errNotWithSplit = 10;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
    placeholder = vbar;
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)

errmsg = nl+
Banner+nl+
nl+
"Syntax "+ProgEXEname+" <file["+extCSV+"]> [separator] [option]..."+nl+
nl+
"    -p    alternate field parsing"+nl+
"    -f    split each line by CSV field"+nl+
"    -s    do not insert original separator between CSV fields"+nl+
'    -m    insert "'+strMySep+'" between CSV fieds instead of original separator'+nl+
"    -a[a] display ASCII reference set then terminate (-aa = decimal)"+nl+
"    -o[o] display Windows OEM reference set then terminate (-oo = decimal)"+nl+
"    -lfn  disable LFN support even if available"+nl+
nl+
"This program either splits each <file> line by CSV field or expands it"+nl+
"so each CSV field is padded with spaces to longest relevant CSV field."+nl+
nl+
"a) A line cannot be more than 24576 characters long."+nl+
"b) Number of fields should not be greater than 100."+nl+
'c) Default field separator is "'+semicolon+'" : if specified, it must be'+nl+
"   either a single character (possibly delimited with double quotes)"+nl+
"   or a value in decimal ([0..255]) or hexadecimal ([$00..$ff])."+nl+
"   Values belonging to {0, 10, 13, $00, $0a, $0d} set are illegal."+nl+
"   Should a character force a problem ([0..9-/$]), use its value instead."+nl+
"d) Output should be redirected to a text file then carefully edited."+nl+
"e) By default, field separator is assumed to be unique ;"+nl+
"   -a option tries and parses single- and double-quoted strings."+nl+
nl+
"Examples : "+ProgEXEname+' abyss.csv ";"'+nl+
"           "+ProgEXEname+" abyss.tab $9"+nl+
"           "+ProgEXEname+" abyss.txt 9"+nl;

VAR
    S : str1024;
BEGIN
    CASE e OF
    | errHelp :        WrStr(errmsg);
    | errOption:       S:='Unknown "'+placeholder+'" option !';
    | errTooManyParms: S:='"'+placeholder+'" parameter is one too many !';
    | errJoker:        S:='No joker allowed in "'+placeholder+'" specification !';
    | errNotFound:     S:='No valid file matches "'+placeholder+'" specification !';
    | errParm:         S:="Expecting "+placeholder+" parameter !";
    | errConflict:     S:="-a[a] and -o[o] options are mutually exclusive !";
    | errSep:          S:='Illegal or ambiguous "'+placeholder+'" separator !';
    | errSepConflict:  S:="-s and -m options are mutually exclusive !";
    | errNotWithSplit: S:=placeholder+" and -f options are mutually exclusive !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp:
        ;
    ELSE
        Str.Subst(S,placeholder,einfo);
        WrLn;
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType    = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    bufferIn        : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* ripped from OLDNEW v1.1f *)

TYPE
    chardef   = RECORD
                    Code    : BYTE;
                    Flag    : CHAR;
                    MapChar : BYTE;
                END;
    CharTable = ARRAY [0..255] OF chardef;

CONST
    TwinTOpc  = CharTable(
    chardef(000H, "=", 000H),
    chardef(001H, "=", 001H),
    chardef(002H, "=", 002H),
    chardef(003H, "=", 003H),
    chardef(004H, "=", 004H),
    chardef(005H, "=", 005H),
    chardef(006H, "=", 006H),
    chardef(007H, "=", 007H),
    chardef(008H, "=", 008H),
    chardef(009H, "!", 009H),
    chardef(00AH, "=", 00AH),
    chardef(00BH, "=", 00BH),
    chardef(00CH, "=", 00CH),
    chardef(00DH, "=", 00DH),
    chardef(00EH, "=", 00EH),
    chardef(00FH, "=", 00FH),
    chardef(010H, "=", 010H),
    chardef(011H, "=", 011H),
    chardef(012H, "=", 012H),
    chardef(013H, "=", 013H),
    chardef(014H, "=", 014H),
    chardef(015H, "=", 015H),
    chardef(016H, "=", 016H),
    chardef(017H, "=", 017H),
    chardef(018H, "=", 018H),
    chardef(019H, "=", 019H),
    chardef(01AH, "=", 01AH),
    chardef(01BH, "=", 01BH),
    chardef(01CH, "=", 01CH),
    chardef(01DH, "=", 01DH),
    chardef(01EH, "=", 01EH),
    chardef(01FH, "=", 01FH),
    chardef(020H, "=",  " "),
    chardef(021H, "=",  "!"),
    chardef(022H, "=",  '"'),
    chardef(023H, "=",  "#"),
    chardef(024H, "=",  "$"),
    chardef(025H, "=",  "%"),
    chardef(026H, "=",  "&"),
    chardef(027H, "=",  "'"),
    chardef(028H, "=",  "("),
    chardef(029H, "=",  ")"),
    chardef(02AH, "=",  "*"),
    chardef(02BH, "=",  "+"),
    chardef(02CH, "=",  ","),
    chardef(02DH, "=",  "-"),
    chardef(02EH, "=",  "."),
    chardef(02FH, "=",  "/"),
    chardef(030H, "=",  "0"),
    chardef(031H, "=",  "1"),
    chardef(032H, "=",  "2"),
    chardef(033H, "=",  "3"),
    chardef(034H, "=",  "4"),
    chardef(035H, "=",  "5"),
    chardef(036H, "=",  "6"),
    chardef(037H, "=",  "7"),
    chardef(038H, "=",  "8"),
    chardef(039H, "=",  "9"),
    chardef(03AH, "=",  ":"),
    chardef(03BH, "=",  ";"),
    chardef(03CH, "=",  "<"),
    chardef(03DH, "=",  "="),
    chardef(03EH, "=",  ">"),
    chardef(03FH, "=",  "?"),
    chardef(040H, "=",  "@"),
    chardef(041H, "=",  "A"),
    chardef(042H, "=",  "B"),
    chardef(043H, "=",  "C"),
    chardef(044H, "=",  "D"),
    chardef(045H, "=",  "E"),
    chardef(046H, "=",  "F"),
    chardef(047H, "=",  "G"),
    chardef(048H, "=",  "H"),
    chardef(049H, "=",  "I"),
    chardef(04AH, "=",  "J"),
    chardef(04BH, "=",  "K"),
    chardef(04CH, "=",  "L"),
    chardef(04DH, "=",  "M"),
    chardef(04EH, "=",  "N"),
    chardef(04FH, "=",  "O"),
    chardef(050H, "=",  "P"),
    chardef(051H, "=",  "Q"),
    chardef(052H, "=",  "R"),
    chardef(053H, "=",  "S"),
    chardef(054H, "=",  "T"),
    chardef(055H, "=",  "U"),
    chardef(056H, "=",  "V"),
    chardef(057H, "=",  "W"),
    chardef(058H, "=",  "X"),
    chardef(059H, "=",  "Y"),
    chardef(05AH, "=",  "Z"),
    chardef(05BH, "=",  "["),
    chardef(05CH, "=",  "\"),
    chardef(05DH, "=",  "]"),
    chardef(05EH, "=",  "^"),
    chardef(05FH, "=",  "_"),
    chardef(060H, "=",  "`"),
    chardef(061H, "=",  "a"),
    chardef(062H, "=",  "b"),
    chardef(063H, "=",  "c"),
    chardef(064H, "=",  "d"),
    chardef(065H, "=",  "e"),
    chardef(066H, "=",  "f"),
    chardef(067H, "=",  "g"),
    chardef(068H, "=",  "h"),
    chardef(069H, "=",  "i"),
    chardef(06AH, "=",  "j"),
    chardef(06BH, "=",  "k"),
    chardef(06CH, "=",  "l"),
    chardef(06DH, "=",  "m"),
    chardef(06EH, "=",  "n"),
    chardef(06FH, "=",  "o"),
    chardef(070H, "=",  "p"),
    chardef(071H, "=",  "q"),
    chardef(072H, "=",  "r"),
    chardef(073H, "=",  "s"),
    chardef(074H, "=",  "t"),
    chardef(075H, "=",  "u"),
    chardef(076H, "=",  "v"),
    chardef(077H, "=",  "w"),
    chardef(078H, "=",  "x"),
    chardef(079H, "=",  "y"),
    chardef(07AH, "=",  "z"),
    chardef(07BH, "=",  "{"),
    chardef(07CH, "=",  "|"),
    chardef(07DH, "=",  "}"),
    chardef(07EH, "=",  "~"),
    chardef(07FH, "x",  " "),
    chardef(080H, "x",  " "),
    chardef(081H, "x",  " "),
    chardef(082H, "x",  " "),
    chardef(083H, "x",  " "),
    chardef(084H, "x",  " "),
    chardef(085H, "x",  " "),
    chardef(086H, "x",  " "),
    chardef(087H, "x",  " "),
    chardef(088H, "x",  " "),
    chardef(089H, "x",  " "),
    chardef(08AH, "x",  " "),
    chardef(08BH, "x",  " "),
    chardef(08CH, "x",  " "),
    chardef(08DH, "x",  " "),
    chardef(08EH, "x",  " "),
    chardef(08FH, "x",  " "),
    chardef(090H, "x",  " "),
    chardef(091H, "=",  "'"),
    chardef(092H, "=",  "'"),
    chardef(093H, "x",  " "),
    chardef(094H, "x",  " "),
    chardef(095H, "x",  " "),
    chardef(096H, "x",  " "),
    chardef(097H, "x",  " "),
    chardef(098H, "x",  " "),
    chardef(099H, "x",  " "),
    chardef(09AH, "x",  " "),
    chardef(09BH, "x",  " "),
    chardef(09CH, "x",  " "),
    chardef(09DH, "x",  " "),
    chardef(09EH, "x",  " "),
    chardef(09FH, "x",  " "),
    chardef(0A0H, "=",  " "),
    chardef(0A1H, "=",  ""),
    chardef(0A2H, "=",  ""),
    chardef(0A3H, "=",  ""),
    chardef(0A4H, "=",  ""),
    chardef(0A5H, "=",  ""),
    chardef(0A6H, "=",  "|"),
    chardef(0A7H, "!", 015H),
    chardef(0A8H, "?",  " "),
    chardef(0A9H, "!",  ""),
    chardef(0AAH, "=",  ""),
    chardef(0ABH, "=",  ""),
    chardef(0ACH, "=",  ""),
    chardef(0ADH, "=",  "-"),
    chardef(0AEH, "!",  ""),
    chardef(0AFH, "!",  "-"),
    chardef(0B0H, "=",  ""),
    chardef(0B1H, "=",  ""),
    chardef(0B2H, "=",  ""),
    chardef(0B3H, "?",  " "),
    chardef(0B4H, "?",  " "),
    chardef(0B5H, "=",  ""),
    chardef(0B6H, "!", 014H),
    chardef(0B7H, "?",  " "),
    chardef(0B8H, "?",  " "),
    chardef(0B9H, "?",  " "),
    chardef(0BAH, "=",  ""),
    chardef(0BBH, "=",  ""),
    chardef(0BCH, "=",  ""),
    chardef(0BDH, "=",  ""),
    chardef(0BEH, "?",  " "),
    chardef(0BFH, "=",  ""),
    chardef(0C0H, "=",  "A"),
    chardef(0C1H, "=",  "A"),
    chardef(0C2H, "=",  "A"),
    chardef(0C3H, "=",  "A"),
    chardef(0C4H, "=",  ""),
    chardef(0C5H, "=",  ""),
    chardef(0C6H, "=",  ""),
    chardef(0C7H, "=",  ""),
    chardef(0C8H, "=",  "E"),
    chardef(0C9H, "=",  ""),
    chardef(0CAH, "=",  "E"),
    chardef(0CBH, "=",  "E"),
    chardef(0CCH, "=",  "I"),
    chardef(0CDH, "=",  "I"),
    chardef(0CEH, "=",  "I"),
    chardef(0CFH, "=",  "I"),
    chardef(0D0H, "?",  " "),
    chardef(0D1H, "=",  ""),
    chardef(0D2H, "=",  "O"),
    chardef(0D3H, "=",  "O"),
    chardef(0D4H, "=",  "O"),
    chardef(0D5H, "=",  "O"),
    chardef(0D6H, "=",  ""),
    chardef(0D7H, "=",  "X"),
    chardef(0D8H, "=",  "O"),
    chardef(0D9H, "=",  "U"),
    chardef(0DAH, "=",  "U"),
    chardef(0DBH, "=",  "U"),
    chardef(0DCH, "=",  ""),
    chardef(0DDH, "=",  "Y"),
    chardef(0DEH, "?",  " "),
    chardef(0DFH, "=",  ""),
    chardef(0E0H, "=",  ""),
    chardef(0E1H, "=",  ""),
    chardef(0E2H, "=",  ""),
    chardef(0E3H, "=",  "a"),
    chardef(0E4H, "=",  ""),
    chardef(0E5H, "=",  ""),
    chardef(0E6H, "=",  ""),
    chardef(0E7H, "=",  ""),
    chardef(0E8H, "=",  ""),
    chardef(0E9H, "=",  ""),
    chardef(0EAH, "=",  ""),
    chardef(0EBH, "=",  ""),
    chardef(0ECH, "=",  ""),
    chardef(0EDH, "=",  ""),
    chardef(0EEH, "=",  ""),
    chardef(0EFH, "=",  ""),
    chardef(0F0H, "?",  " "),
    chardef(0F1H, "=",  ""),
    chardef(0F2H, "=",  ""),
    chardef(0F3H, "=",  ""),
    chardef(0F4H, "=",  ""),
    chardef(0F5H, "=",  "o"),
    chardef(0F6H, "=",  ""),
    chardef(0F7H, "=",  ""),
    chardef(0F8H, "=",  "o"),
    chardef(0F9H, "=",  ""),
    chardef(0FAH, "=",  ""),
    chardef(0FBH, "=",  ""),
    chardef(0FCH, "=",  ""),
    chardef(0FDH, "=",  "y"),
    chardef(0FEH, "?",  " "),
    chardef(0FFH, "=",  ""));

PROCEDURE num2str (n,base,wi:CARDINAL;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
BEGIN
    Str.CardToStr( LONGCARD(n),R,base,ok );
    WHILE Str.Length(R) < wi DO
        Str.Prepend(R,padchar);
    END;
    IF base=16 THEN Str.Lows(R);Str.Prepend(R,dollar);END;
    RETURN R;
END num2str;

CONST
    ASCIIset = 1;
    OEMset   = 2;

PROCEDURE showset (what,base:CARDINAL);
CONST
    default = FALSE;
    (*%T default *)
    double = doublequote;
    single = quote;
    sep    = ":";
    (*%E *)
    (*%F default  *)
    double = " ";
    single = " ";
    sep    = " "; (* ":" *)
    (*%E  *)
VAR
    i,wi,last,v:CARDINAL;
    S,H:str128;
    V:str16;
    pad:CHAR;
BEGIN
    WrLn;

    CASE base OF
    | 10: wi:=3;pad:=" "; H:="   ";
    | 16: wi:=2;pad:="0"; H:="    ";
    END;
    FOR i:=1 TO base DO
        Str.Append(H," ");
        Str.Append(H,num2str(i-1,base,wi,pad));
    END;
    WrStr(H);WrLn;

    last:= ((255 DIV base)+1)*base;DEC(last);
    FOR i:=0 TO last DO
        IF (i MOD base) = 0 THEN
            Str.Copy(V,num2str(i,base,wi,pad));
            WrStr(V);WrStr(" ");
        END;
        WrStr(sep);
        CASE what OF
        | ASCIIset:
            CASE i OF
            | 0..4,7..0AH, 0DH, 1AH..1BH :
                S:="   ";
            | ORD(doublequote):
                Str.Concat(S,single,CHR(i));Str.Append(S,single);
            ELSE
                IF i > 255 THEN
                    v:=ORD(" ");
                ELSE
                    v:=i;
                END;
                Str.Concat(S,double,CHR(v));Str.Append(S,double);
            END;
        | OEMset:
            v := ORD( CHR(TwinTOpc[i].MapChar) );
            CASE v OF
            | 0..4,7..0AH, 0DH, 1AH..1BH :
                S:="   ";
            | ORD(doublequote) :
                Str.Concat(S,single,CHR(v));Str.Append(S,single);
            ELSE
                Str.Concat(S,double,CHR(v));Str.Append(S,double);
            END;
        END;
        WrStr(S);
        IF (i MOD base) = (base-1) THEN
            WrStr(sep+" ");WrStr(V);WrLn;
        END;
    END;
    WrStr(H);WrLn;
END showset;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fmtbignum (v:LONGCARD; base:CARDINAL;wi:INTEGER; pad:CHAR ):str16;
VAR
    S:str16;
    ok:BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr( v, S, base,ok);
    IF base=16 THEN Str.Lows(S);END;
    FOR i:=Str.Length(S)+1 TO ABS(wi) DO
         IF wi < 0 THEN
             Str.Append(S,pad);
         ELSE
             Str.Prepend(S,pad);
         END;
    END;
    RETURN S;
END fmtbignum;

PROCEDURE fmtnum ( v:CARDINAL; base:CARDINAL;wi:INTEGER; pad:CHAR ):str16;
BEGIN
    RETURN fmtbignum( LONGCARD(v),base,wi,pad);
END fmtnum;

(* ------------------------------------------------------------ *)

CONST
    minfield = 1;
    maxfield = 100;
    UNDEFFIELD = MAX(CARDINAL);
VAR
    fieldmaxlen : ARRAY [minfield..maxfield] OF CARDINAL;

PROCEDURE initproc (  );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=minfield TO maxfield DO
        fieldmaxlen[i] := UNDEFFIELD;
    END;
END initproc;

PROCEDURE showsep ( DEBUG:BOOLEAN;sep:CHAR );
VAR
    i:CARDINAL;
BEGIN
    IF DEBUG THEN
        i:=ORD(sep);
        WrLn;
        WrStr("Separator : $");
        WrStr( fmtnum ( i,16,2,"0"));
        WrStr(" -- ");
        WrStr( fmtnum ( i,10,3," "));
        WrStr(" -- ");
        CASE i OF
        | 0..4,7..0AH, 0DH, 1AH..1BH, 0FFH :
            WrStr(".");
        ELSE
            WrStr(sep);
        END;
        WrLn;
        abort(errNone,"");
    END;
END showsep;

(* ------------------------------------------------------------ *)

VAR (* globerks because moved out OF procMe() *)
    SSS,ZZZ : ARRAY [0..24576-1] OF CHAR;

(* //FIXME FOR now we won't handle sep in delimited string *)
(* assume separators are unique *)

PROCEDURE countFields (ALTPARSE:BOOLEAN;sep:CHAR   ):CARDINAL;
VAR
    p,n,strlen:CARDINAL;
    ch,strdelim:CHAR;
    state:(waiting,inquoted);
BEGIN
    IF ALTPARSE THEN
        strlen:=Str.Length(SSS);
        p:=0;
        n:=0;
        state:=waiting;
        LOOP
            IF p >= strlen THEN EXIT;END;
            ch:=SSS[p];
            CASE state OF
            | waiting:
                CASE ch OF
                | singlequote,doublequote:
                    strdelim:=ch;
                    state:=inquoted;
                ELSE
                    IF ch=sep THEN
                        INC(n);
                    END;
                END;
            | inquoted:
                IF ch = strdelim THEN state:=waiting;END;
            END;
            INC(p);
        END;
        IF strlen # 0 THEN
            INC(n);
        END;
    ELSE
        n:= CharCount(SSS,sep)+1; (* example : *;*;* = 3 fields *)
    END;
    RETURN n;
END countFields;

PROCEDURE getFieldStr (ALTPARSE:BOOLEAN;sep:CHAR;i:CARDINAL);
VAR
    p,n,strlen:CARDINAL;
    ch,strdelim:CHAR;
    state:(waiting,inquoted);
BEGIN
    IF ALTPARSE THEN
        Str.Copy(ZZZ,"");
        strlen:=Str.Length(SSS);
        p:=0;
        n:=0;
        state:=waiting;
        LOOP
            IF p >= strlen THEN EXIT;END;
            ch:=SSS[p];
            CASE state OF
            | waiting:
                CASE ch OF
                | singlequote,doublequote:
                    strdelim:=ch;
                    state:=inquoted;
                    Str.Append(ZZZ,ch);
                ELSE
                    IF ch=sep THEN
                        INC(n);
                        IF n=i THEN EXIT;END;
                        Str.Copy(ZZZ,"");
                    ELSE
                        Str.Append(ZZZ,ch);
                    END;
                END;
            | inquoted:
                Str.Append(ZZZ,ch);
                IF ch = strdelim THEN state:=waiting;END;
            END;
            INC(p);
        END;
    ELSE
        isoleItemS(ZZZ,SSS,sep,i-1);
    END;
END getFieldStr;

(* ------------------------------------------------------------ *)

TYPE
    sepmodetype = (orgsep,nosep,mysep);

PROCEDURE procMe (DEBUG,useLFN,BYFIELD,ALTPARSE:BOOLEAN;
                 sepmode:sepmodetype;sep:CHAR;src:pathtype );
CONST
    wifield = 3; (* ### *)
VAR
    hin:FIO.File;
    lastpass,len,nfields,pass,i,j:CARDINAL;
    dmpsep:str16;
BEGIN
    CASE sepmode OF
    | orgsep: Str.Copy(dmpsep,sep);
    | nosep : Str.Copy(dmpsep,"");
    | mysep : Str.Copy(dmpsep,strMySep);
    END;

    initproc;
    IF BYFIELD THEN
        lastpass := 1;
    ELSE
        lastpass := 2;
    END;
    FOR pass := 1 TO lastpass DO
        hin := fileOpenRead(useLFN,src);
        FIO.AssignBuffer(hin,bufferIn);

        hin := fileOpenRead(useLFN,src);
        FIO.AssignBuffer(hin,bufferIn);
        FIO.EOF := FALSE;
        LOOP
            IF FIO.EOF THEN EXIT;END;
            FIO.RdStr(hin,SSS);
            IF ((SSS[0]=0C) AND FIO.EOF) THEN EXIT; END;

            nfields:=countFields(ALTPARSE,sep);
            IF nfields > maxfield THEN nfields:=maxfield;END; (* brutal *)

            FOR i:=1 TO nfields DO
                getFieldStr(ALTPARSE,sep,i);
                IF BYFIELD THEN
                    dmpsep:=strField;
                    Str.Subst(dmpsep,"~", fmtnum ( i,10,wifield,"0"));
                    WrStr(dmpsep);
                    WrStr(ZZZ);WrLn;
                ELSE
                    len:=Str.Length(ZZZ);
                    CASE pass OF
                    | 1:
                        IF fieldmaxlen[i]=UNDEFFIELD THEN
                            fieldmaxlen[i]:=0;
                        END;
                        IF len > fieldmaxlen[i] THEN fieldmaxlen[i]:=len;END;
                    | 2:
                        WrStr(ZZZ);
                        FOR j:=Str.Length(ZZZ)+1 TO fieldmaxlen[i] DO
                            WrStr(" ");
                        END;
                        IF i < nfields THEN WrStr(dmpsep);END;
                    END;
                END;
            END;
            IF NOT(BYFIELD) THEN
                IF pass=2 THEN WrLn;END;
            END;
        END;
        fileClose(useLFN,hin);
    END;
END procMe;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* ? or $# or 0X# *)

PROCEDURE getsep (VAR sep:CHAR;S:ARRAY OF CHAR   ):BOOLEAN;
VAR
    v:LONGCARD;
    ok:BOOLEAN;
    base:CARDINAL;
    R:str16;
BEGIN
    Str.Copy(R,S);
    Str.Caps(R);
    IF Str.Match(R,"$*") THEN
        Str.Delete(R,0,1);
        base:=16;
    ELSIF Str.Match(R,"0X*") THEN (* a concession to C-believers *)
        Str.Delete(R,0,2);
        base:=16;
    ELSE
        base:=10;
    END;
    v:=Str.StrToCard(R,base,ok);
    IF ok THEN
        IF v > MAX(SHORTCARD) THEN
            ok:=FALSE;
        ELSE
            sep:=CHR( CARDINAL(v) );
        END;
    ELSE
        IF Str.Length(S)=1 THEN (* conversion failed, try char *)
            sep:=S[0]; (* keep case *)
            ok:=TRUE;
        END;
    END;
    IF ok THEN
        CASE ORD(sep) OF
        | 0,10,13 :
            ok:=FALSE;
        ELSE
            ;
        END;
    END;
    RETURN ok;
END getsep;

(* ------------------------------------------------------------ *)

TYPE
    cmdtype = (nocmd,doexpand,dosplit,
              showASCIIhex,showASCIIdec,
              showOEMhex,showOEMdec);

PROCEDURE newcmd (VAR cmd:cmdtype;wanted:cmdtype  ):BOOLEAN;
VAR
    ok:BOOLEAN;
BEGIN
    IF ( (cmd = nocmd) OR (cmd = wanted) ) THEN
        cmd:=wanted;
        ok:=TRUE;
    ELSE
        ok:=FALSE;
    END;
    RETURN ok;
END newcmd;

(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 2;
VAR
    parm : ARRAY[minparm..maxparm] OF pathtype;
    lastparm,parmcount,i,opt : CARDINAL;
    S,R,src:pathtype;
    useLFN,DEBUG : BOOLEAN;
    separator : CHAR;
    cmd:cmdtype;
    ALTPARSE,EATSEP,USEMYSEP,BYFIELD:BOOLEAN;
    sepmode:sepmodetype;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;

    (* WrLn; *)

    cmd         := nocmd;
    EATSEP      := FALSE;
    USEMYSEP    := FALSE;
    BYFIELD     := FALSE;
    ALTPARSE    := FALSE;
    useLFN      := TRUE;
    DEBUG       := FALSE;

    lastparm    := minparm-1;
    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "DEBUG"+delim+
                                   "S"+delim+"SEP"+delim+
                                   "M"+delim+"MYSEP"+delim+
                                   "A"+delim+
                                   "AA"+delim+
                                   "O"+delim+
                                   "OO"+delim+
                                   "X"+delim+"LFN"+delim+
                                   "F"+delim+"FIELD"+delim+
                                       "R"+delim+"RECORD"+delim+
                                   "P"+delim+"PARSE"+delim+
                                       "Q"+delim+"QUOTED"
                               );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4     : DEBUG:=TRUE;
            | 5,6   : EATSEP:=TRUE;
            | 7,8   : USEMYSEP :=TRUE;
            | 9     : IF newcmd(cmd,showASCIIhex)=FALSE THEN abort(errConflict,"");END;
            | 10    : IF newcmd(cmd,showASCIIdec)=FALSE THEN abort(errConflict,"");END;
            | 11    : IF newcmd(cmd,showOEMhex)  =FALSE THEN abort(errConflict,"");END;
            | 12    : IF newcmd(cmd,showOEMdec)  =FALSE THEN abort(errConflict,"");END;
            | 13,14 : useLFN := FALSE;
            | 15,16,17,18: BYFIELD := TRUE;
            | 19,20,21,22: ALTPARSE:= TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errTooManyParms,S); END;
            Str.Copy(parm[lastparm],S); (* keep case *)
        END;
    END;

    CASE cmd OF
    | showASCIIhex:  showset(ASCIIset,16);
    | showOEMhex:    showset(OEMset,16);
    | showASCIIdec:  showset(ASCIIset,10);
    | showOEMdec:    showset(OEMset,10);
    END;
    IF cmd # nocmd THEN abort(errNone,"");END;
    IF BYFIELD THEN
        cmd :=dosplit;
        IF EATSEP THEN abort(errNotWithSplit,"-s");END;
        IF USEMYSEP THEN abort(errNotWithSplit,"-m");END;
    ELSE
        cmd :=doexpand;
    END;


    useLFN := ( useLFN AND fileSupportLFN() );

    CASE lastparm OF
    | minparm-1 : abort(errParm,"<file>");
    | minparm:    Str.Copy(src,parm[minparm]);
                  separator:=semicolon;
    | minparm+1 : Str.Copy(src,parm[minparm]);
                  Str.Copy(S,parm[minparm+1]); (* ?, "?", [$]# *)
                  IF getsep(separator, S) = FALSE THEN abort(errSep,S);END;
    ELSE
        ; (* handled supra *)
    END;

    IF EATSEP THEN
        IF USEMYSEP THEN
            abort(errSepConflict,"");
        ELSE
            sepmode:=nosep;
        END;
    ELSE
        IF USEMYSEP THEN
            sepmode:=mysep;
        ELSE
            sepmode:=orgsep;
        END;
    END;

    IF same(src,dot) THEN Str.Copy(src,star+extCSV); END;
    IF same(src,star) THEN Str.Copy(src,star+extCSV); END;
    IF Str.Match(src,"*"+antislash) THEN Str.Append(src,star+extCSV); END;
    IF Str.RCharPos(src,dot)=MAX(CARDINAL) THEN Str.Append(src,extCSV);END;

    IF chkJoker(src) THEN abort(errJoker,src);END;

    IF fileExists(useLFN,src)=FALSE THEN abort(errNotFound,src);END;

    showsep(DEBUG,separator);
    procMe(DEBUG,useLFN,BYFIELD,ALTPARSE,sepmode,separator,src);

    abort(errNone,"");
END xCSV.

