(* ---------------------------------------------------------------
Title         Q&D Block Dump
Author        PhG
Overview      show plain data found in saved file (MBR)
Notes         why do we have to force CARDINAL for ABS() ?
              fmt subs() have been slightly revised
Bugs          let's say "quirk" :
              JPI TS does not like appending to an empty string constant
              while it's ok from Str lib : go figure...
Wish List     show plain primary or logical partition boot record
              show CMOS data according to BIOS manufacturer ?
              duplicating VITAL is not clever, but...
              we did not think about grabbing infos from files at the time :-(
              (well, we did NOT need this either)

--------------------------------------------------------------- *)

MODULE BlockDmp;

IMPORT SYSTEM;
IMPORT Lib;
IMPORT Str;
IMPORT FIO;

FROM FIO IMPORT FIXEDLIBS;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    progTitle     = "Q&D Block Dump";
    progVersion   = "v1.0";
    progCopyright = "by PhG";
    Banner        = progTitle+" "+progVersion+" "+progCopyright;
    progEXEname   = "BLOCKDMP";
CONST
    extBIN        = ".BIN";
    extSAV        = ".SAV";
    extHD0        = ".HD0";
    extHD1        = ".HD1";
    extMBR        = ".MBR";
    extBLK        = ".BLK";
    defaultExtensions = extBIN+delim+extSAV+delim+
                        extHD0+delim+extHD1+delim+
                        extMBR+delim+
                        extBLK;
    sSectorSize   = "512";      (* see infra *)
    sMagic        = "$aa55";    (* see infra *)
    sMagicPos     = "$01fe";    (* see infra *)
CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParameter    = 3;
    errExpected     = 4;
    errOnlyOne      = 5;
    errJoker        = 6;
    errNotFound     = 7;
    errExhausted    = 8;
    errTooSmall     = 9;
    errAA55         = 10;
    errMoreHelp     = 11;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msgHelp =
Banner+nl+
nl+
"Syntax : "+progEXEname+" <file> [option]..."+nl+
nl+
"This program dumps data found in a block saved to a file."+nl+
nl+
"    -m    dump <file> as a master boot record (default)"+nl+
"    -a    alternate unbeautified format"+nl+
"    -g    partition size in gigabytes (default is megabytes)"+nl+
"    -s    check "+sMagic+" signature"+nl+
"    -b[b] dump <file> as hexadecimal too (-b = by 16, -bb = by 8)"+nl+
"    -x    disable LFN support even if available"+nl+
"    -??   technical memento"+nl+
nl+
"a) Without extension, <file> is tested against these default ones :"+nl+
"   "+defaultExtensions+nl+
"b) <file> length must be at least "+sSectorSize+" long for data to be parsed ;"+nl+
"   "+sMagic+" signature MUST exist at "+sMagicPos+" file position."+nl+
"c) Being irrelevant, cylinders, heads and sectors will not be displayed."+nl;

VAR
    S:str128;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(msgHelp);
    | errOption :
        S:='Unknown "~" option';
    | errParameter :
        S:='Unexpected "~" parameter';
    | errExpected:
        S:="<file> parameter expected";
    | errOnlyOne:
        S:="-m, -p and -l options are mutually exclusive";
    | errJoker:
        S:='"~" cannot contain any joker';
    | errNotFound:
        S:='"~" does not exist';
    | errExhausted:
        S:="All supported default extensions were tried without match";
    | errTooSmall:
        S:='"~" should be at least '+sSectorSize+" bytes long";
    | errAA55:
        S:='Expected '+sMagic+' signature was not found in "~"';
    | errMoreHelp:
        S:=""; (* never displayed *)
        e:=errHelp;
    ELSE
        S := "How did you get THERE ? Amazing";
    END;
    CASE e OF
    | errNone, errHelp, errMoreHelp : (* yes, errMoreHelp cannot happpen ! *)
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : ");WrStr(S);WrStr(" !");WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize      = (8 * 512) + FIO.BufferOverhead; (* 8 or less is ok *)
    firstBufferByte   = 1;
    lastBufferByte    = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn   : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
    about these formatting subs :
    - note "wi" only applies to numbers : it does NOT include prefix and suffix
    - yes, yes, str80 is oversized but str16 could be too small at times
    - if pad is " ", prefix is inserted at first non blank, not before blank(s)
*)

PROCEDURE padme (VAR R:str80;
                wi:INTEGER;pad:CHAR;prefix,suffix:ARRAY OF CHAR );
VAR
    len,i,lp,ls,added:CARDINAL;
BEGIN
    len:=Str.Length(R);
    lp:=Str.Length(prefix);
    ls:=Str.Length(suffix);
    added:=0;
    FOR i:=(len+1) TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,pad);      (* "???......." "???" *)
        ELSE
            Str.Prepend(R,pad);     (* ".......???" "???" *)
        END;
        INC(added);
    END;
    IF wi < 0 THEN
        Str.Prepend(R,prefix);
        Str.Insert(R,suffix,lp+len);
    ELSE
        Str.Append(R,suffix);
        IF pad=blank THEN   (* prefix before blanks would be ugly *)
            Str.Insert(R,prefix,added);
        ELSE
            Str.Prepend(R,prefix); (* insert at 0 *)
        END;
    END;

END padme;

CONST
    strErrConv = "???";

PROCEDURE fmtlc (v:LONGCARD;base:CARDINAL;wi:INTEGER;
                pad:CHAR;prefix,suffix:ARRAY OF CHAR ) : str80;
VAR
    R : str80;
    ok: BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    IF NOT(ok) THEN RETURN strErrConv; END;
    IF base=16 THEN Str.Lows(R);END;
    padme(R,wi,pad,prefix,suffix);
    RETURN R;
END fmtlc;

PROCEDURE fmtcard (v:CARDINAL;base:CARDINAL;
                  wi:INTEGER;pad:CHAR;prefix,suffix:ARRAY OF CHAR) : str80;
BEGIN
    RETURN fmtlc(LONGCARD(v),base,wi,pad,prefix,suffix);
END fmtcard;

PROCEDURE fmtbyte (v:SHORTCARD;base:CARDINAL;wi:INTEGER;pad,prefix,suffix:CHAR) : str80;
BEGIN
    RETURN fmtlc(LONGCARD(v),base,wi,pad,prefix,suffix);
END fmtbyte;

PROCEDURE fmtchar (enclose:BOOLEAN;c : BYTE) : str16;
VAR
    S : str16;
    enclosechar:CHAR;
BEGIN
    CASE ORD(c) OF
    | 0..ORD(blank)-1 : S:=".";
    | 255 :             S:=".";
    ELSE
                        Str.Copy(S,CHR(c));
    END;
    IF enclose THEN
        CASE ORD(c) OF
        | ORD(dquote) :
            enclosechar := singlequote;
        ELSE
            enclosechar := dquote;
        END;
        Str.Prepend(S,enclosechar);Str.Append(S,enclosechar);
    END;
    RETURN S;
END fmtchar;

PROCEDURE beautifiedlc (v : LONGCARD;pad:CHAR; sep:CHAR; wi:INTEGER) : str80;
VAR
    S,R   : str80;
    len,i : CARDINAL;
    ok  : BOOLEAN;
    ch  : CHAR;
BEGIN
    Str.CardToStr(v,S,10,ok);
    len:=Str.Length(S);
    R := "";
    FOR i := 1 TO len DO
        Str.Prepend(R,S[len-i]);
        IF i < len THEN
            IF (i MOD 3) = 0 THEN
                Str.Prepend(R,sep);
            END;
        END;
    END;
    LOOP
        IF INTEGER(Str.Length(R)) >= ABS(wi) THEN EXIT; END;
        IF wi < 0 THEN
            Str.Append(R,pad);  (* right alignment *)
        ELSE
            Str.Prepend(R,pad);
        END;
    END;
    RETURN R;
END beautifiedlc;

TYPE VERYLONGCARD = RECORD (* 64 bits *)
    lo : LONGCARD;
    hi : LONGCARD;
END;

PROCEDURE fmtvlc (v:VERYLONGCARD;base:CARDINAL;wi:INTEGER;prefix,suffix:CHAR) : str80;
VAR
    S:str80;
    pad:CHAR;
BEGIN
    base  :=16; (* //FIXME v1.3p force hex until we handle decimal *)
    wi    :=8;
    pad   :="0";
    prefix:="$";

    Str.Concat(S,fmtlc(v.hi,base,wi,pad,prefix,""), fmtlc(v.lo,base,wi,pad,"","") );
    RETURN S;
END fmtvlc;

PROCEDURE roundnum (VAR mb,frac:LONGCARD; total,divisor:LONGCARD;decimales:CARDINAL);
VAR
    v,limit:LONGCARD;
    i:CARDINAL;
BEGIN
    IF divisor = 0 THEN divisor:=1;END; (* berk *)
    mb   := total DIV divisor;  (* integer part *)
    v    := total - mb*divisor; (* remainder *)

    limit:= 1;
    FOR i:=1 TO (decimales+1) DO v:=v*10; limit:=limit*10; END;
    DEC(limit);

    frac:=v DIV divisor;

    IF (frac MOD 10) >= 5 THEN
        INC(frac,10);
        IF frac > limit THEN frac:=0; INC(mb);END; (* not pretty but close enough *)
    END;
    frac := frac DIV 10;
END roundnum;

CONST
    useMega     = 0;
    useGiga     = 1;
    sMegaSuffix = " Mb";
    sGigaSuffix = " Gb";
    fix1        = 1;
    fix2        = 2;

PROCEDURE fmtrounded (partblocks:LONGCARD;ga,decimales:CARDINAL;fracsep:CHAR):str80;
CONST
    (* assume 512 bytes per sector *)
    oneMega = LONGCARD(2048);  (* expressed in blocks : 1024 * 1024 / 512 *)
    oneGiga = LONGCARD(1024*oneMega);  (* idem : 1024 * 1024 * 1024 / 512 *)
VAR
    frac,mb,whatGA:LONGCARD;
    R:str80;
    suffix : str16;
BEGIN
    CASE ga OF
    | useMega : whatGA := oneMega;
    | useGiga : whatGA := oneGiga;
    ELSE
                whatGA := oneMega;
    END;
    CASE decimales OF
    | fix1,fix2 : ;
    ELSE
                decimales:=fix1;
    END;

    roundnum(mb,frac, partblocks,whatGA,decimales);
    Str.Copy  (R, fmtlc(mb,10,0,"","","") );
    Str.Append(R,fracsep);
    Str.Append(R, fmtlc(frac,10,decimales,"0","","")); (* bugfix for stupid ,0,"" ! *)
    RETURN R;
END fmtrounded;

PROCEDURE prettySize (ga:CARDINAL;wi:INTEGER;decimales:CARDINAL; v:LONGCARD;
                     pad,fracsep:CHAR):str80;
CONST
    suffixMb = " Mb";
    suffixGb = " Gb";
VAR
    suffix : str16;
    R:str80;
    units,i,p:CARDINAL;
BEGIN
    CASE ga OF
    | useMega: suffix:=suffixMb; units:=useMega;
    | useGiga: suffix:=suffixGb; units:=useGiga;
    END;
    R:=fmtrounded(v,units,decimales,fracsep);
    padme(R,wi,pad,"",suffix);
    RETURN R;
END prettySize;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    sNoLongID = "Unknown partition type";

PROCEDURE getIDpartStr (IDpart:BYTE;VAR sHex:ARRAY OF CHAR;VAR sID:str256);
VAR
    S:str256;
    ok:BOOLEAN;
BEGIN
    Str.Copy(sHex,fmtcard( CARDINAL(IDpart),16,2,"0","$","") );
    CASE IDpart OF
    | 000H : S:="Empty partition-table entry"; (* // *)
    | 001H : S:="DOS FAT12";
    | 002H : S:="XENIX root file system";
    | 003H : S:="XENIX /usr file system (obsolete)";
    | 004H : S:="DOS FAT16 (< 32M)";
    | 005H : S:="DOS 3.3+ extended partition";         (* // *)
    | 006H : S:="DOS 3.31+ Large File System (FAT16, >= 32M)";
    | 007H : S:="8-ung ! QNX, OS/2 HPFS, Windows NT NTFS, Advanced Unix, or unknown";
    | 008H : S:="8-ung ! OS/2 (v1.0-1.3 only), AIX bootable partition, SplitDrive, Commodore DOS, or DELL partition spanning multiple drives";
    | 009H : S:="8-ung ! AIX data partition, or Coherent filesystem";
    | 00AH : S:="8-ung ! OS/2 Boot Manager, OPUS, or Coherent swap partition";
    | 00BH : S:="Windows95 FAT32";
    | 00CH : S:="Windows95 FAT32 (using LBA-mode INT $13 extensions)";
    | 00EH : S:="logical-block-addressable VFAT (same as 06h but using LBA-mode INT $13)";
    | 00FH : S:="logical-block-addressable VFAT (same as 05h but using LBA-mode INT $13)"; (* // *)
    | 010H : S:="OPUS";
    | 011H : S:="OS/2 Boot Manager hidden FAT12 partition";
    | 012H : S:="Compaq Diagnostics partition";
    | 014H : S:="8-ung ! (resulted from using Novell DOS 7.0 FDISK to delete Linux Native part), "+
                "or OS/2 Boot Manager hidden < 32M FAT16 partition";
    | 015H : S:="hidden DOS 3.3+ extended partition"; (* // added by me, probably used by XOSL *)
    | 016H : S:="OS/2 Boot Manager hidden >= 32M FAT16 partition";
    | 017H : S:="8-ung ! OS/2 Boot Manager hidden HPFS partition, or hidden NTFS partition";
    | 018H : S:='AST special Windows swap file ("Zero-Volt Suspend" partition)';
    | 019H : S:="Willowtech Photon coS";
    | 01BH : S:="hidden Windows95 FAT32 partition";
    | 01CH : S:="hidden Windows95 FAT32 partition (using LBA-mode INT $13 extensions)";
    | 01EH : S:="hidden LBA VFAT partition (PowerQuest specific ?)";
    | 01FH : S:="hidden LBA VFAT extended partition"; (* // added by me, used by XOSL *)
    | 020H : S:="Willowsoft Overture File System (OFS1)";
    | 021H : S:="8-ung ! officially listed as reserved, or Oxygen FSo2";
    | 022H : S:="Oxygen Extended";
    | 023H : S:="officially listed as reserved";
    | 024H : S:="NEC MS-DOS 3.x";
    | 026H : S:="officially listed as reserved";
    | 031H : S:="officially listed as reserved";
    | 033H : S:="officially listed as reserved";
    | 034H : S:="officially listed as reserved";
    | 036H : S:="officially listed as reserved";
    | 038H : S:="Theos";
    | 03CH : S:="PowerQuest PartitionMagic recovery partition";
    | 03DH : S:="Hidden Netware";
    | 040H : S:="VENIX 80286";
    | 041H : S:="8-ung ! Personal RISC Boot, PowerPC boot partition, or PTS-DOS Linux / DR DOS";
    | 042H : S:="8-ung ! SFS (Secure File System) by Peter Gutmann, or PTS-DOS Linux Swap / DR DOS";
    | 043H : S:="8-ung ! Alternative Linux EXT2fs, or PTS-DOS DR DOS";
    | 045H : S:="EUMEL/Elan";
    | 046H : S:="EUMEL/Elan";
    | 047H : S:="EUMEL/Elan";
    | 048H : S:="EUMEL/Elan";
    | 04AH : S:="ALFS/THIN filesystem for DOS";
    | 04DH : S:="QNX";
    | 04EH : S:="QNX";
    | 04FH : S:="8-ung ! Oberon boot/data partition, or QNX";
    | 050H : S:="OnTrack Disk Manager (read-only partition)";
    | 051H : S:="8-ung ! OnTrack Disk Manager (read/write partition) or NOVELL";
    | 052H : S:="8-ung ! CP/M, or Microport System V/386";
    | 053H : S:="OnTrack Disk Manager (write-only partition ?)";
    | 054H : S:="OnTrack Disk Manager (DDO)";
    | 055H : S:="EZ-Drive";
    | 056H : S:="GoldenBow VFeature";
    | 05CH : S:="Priam EDISK";
    | 061H : S:="SpeedStor";
    | 063H : S:="8-ung ! Unix SysV/386 (386/ix), Mach (MtXinu BSD 4.3 on Mach), or GNU HURD";
    | 064H : S:="8-ung ! Novell NetWare 286, or SpeedStore";
    | 065H : S:="Novell NetWare (3.11)";
    | 066H : S:="Novell Netware 386";
    | 067H : S:="Novell";
    | 068H : S:="Novell";
    | 069H : S:="Novell";
    | 070H : S:="DiskSecure Multi-Boot";
    | 071H : S:="officially listed as reserved";
    | 073H : S:="officially listed as reserved";
    | 074H : S:="officially listed as reserved";
    | 075H : S:="PC/IX";
    | 076H : S:="officially listed as reserved";
    | 078H : S:="XOSL";
    | 07EH : S:="F.I.X.";
    | 080H : S:="Minix v1.1-1.4a";
    | 081H : S:="8-ung ! Minix v1.4b+, Linux, or Mitac Advanced Disk Manager";
    | 082H : S:="8-ung ! Linux Swap partition, Prime, or Solaris (Unix)";
    | 083H : S:="Linux native file system (ext2fs/xiafs)";
    | 084H : S:="OS/2-renumbered type 04h partition (related to hiding DOS C: drive)";
    | 085H : S:="Linux EXT";
    | 086H : S:="FAT16 volume/stripe set (Windows NT)";
    | 087H : S:="8-ung ! HPFS Fault-Tolerant mirrored partition, or NTFS volume/stripe set";
    | 093H : S:="Amoeba file system";
    | 094H : S:="Amoeba bad block table";
    | 098H : S:="Datalight ROM-DOS SuperBoot";
    | 099H : S:="Mylex EISA SCSI";
    | 09FH : S:="BDSI";
    | 0A0H : S:='Phoenix NoteBIOS Power Management "Save-to-Disk" partition';
    | 0A1H : S:="officially listed as reserved";
    | 0A3H : S:="officially listed as reserved";
    | 0A4H : S:="officially listed as reserved";
    | 0A5H : S:="FreeBSD, BSD/386";
    | 0A6H : S:="OpenBSD";
    | 0A7H : S:="NextStep";
    | 0A9H : S:="NetBSD";
    | 0AAH : S:="Olivetti DOS FAT 12";
    | 0B0H : S:="BootStar";
    | 0B1H : S:="officially listed as reserved";
    | 0B3H : S:="officially listed as reserved";
    | 0B4H : S:="officially listed as reserved";
    | 0B6H : S:="8-ung ! officially listed as reserved, but Windows NT mirror set (master) FAT16 file system";
    | 0B7H : S:="8-ung ! BSDI file system (secondarily swap), or Windows NT mirror set (master) NTFS file system";
    | 0B8H : S:="BSDI swap partition (secondarily file system)";
    | 0BBH : S:="8-ung ! Hidden PTS BootWizard, or Acronis";
    | 0BCH : S:="Acronis";
    | 0BEH : S:="Solaris boot partition";
    | 0C0H : S:="8-ung ! DR-DOS/Novell DOS secured partition, or CTOS";
    | 0C1H : S:="DR DOS 6.0 LOGIN.EXE-secured FAT12 partition";
    | 0C2H : S:="DR DOS 7";
    | 0C3H : S:="DR DOS 7";
    | 0C4H : S:="DR DOS 6.0 LOGIN.EXE-secured FAT16 partition";
    | 0C6H : S:="8-ung ! DR DOS 6.0 LOGIN.EXE-secured Huge partition, corrupted FAT16 volume/stripe set (Windows NT), or Windows NT mirror set (slave) FAT16 file system";
    | 0C7H : S:="8-ung ! Syrinx Boot, corrupted NTFS volume/stripe set, or Windows NT mirror set (slave) NTFS file system";
    | 0C8H : S:="DR DOS 7";
    | 0C9H : S:="DR DOS 7";
    | 0CAH : S:="DR DOS 7";
    | 0CBH : S:="Reserved for DR-DOS secured FAT32";
    | 0CCH : S:="Reserved for DR-DOS secured FAT32 (LBA)";
    | 0CDH : S:="DR DOS 7";
    | 0CEH : S:="Reserved for DR-DOS secured FAT16 (LBA)";
    | 0CFH : S:="Reserved for DR-DOS extended secured FAT16 (LBA)";
    | 0D0H : S:="Multiuser DOS secured FAT12";
    | 0D1H : S:="Old Multiuser DOS secured FAT12";
    | 0D4H : S:="Old Multiuser DOS secured FAT16 (< 32M)";
    | 0D5H : S:="Old Multiuser DOS secured extended partition";
    | 0D6H : S:="Old Multiuser DOS secured FAT16 (>= 32M)";
    | 0D8H : S:="CP/M-86";
    | 0DBH : S:="8-ung ! CP/M, Concurrent CP/M, Concurrent DOS, or CTOS (Convergent Technologies OS)";
    | 0DEH : S:="Dell FAT16";
    | 0E1H : S:="SpeedStor FAT12 extended partition";
    | 0E2H : S:="DOS read-only (Florian Painke's XFDISK 1.0.4)";
    | 0E3H : S:="8-ung ! DOS read-only, or Storage Dimensions";
    | 0E4H : S:="SpeedStor FAT16 extended partition";
    | 0E5H : S:="officially listed as reserved";
    | 0E6H : S:="officially listed as reserved";
    | 0EBH : S:="BeOS BFS (BFS1)";
    | 0EDH : S:="Spryt*x";
    | 0EEH : S:="GPT protected MBR (faking whole disk partition)"; (* I 8 m$ *)
    | 0EFH : S:="UEFI system partition (ESP)";
    | 0F1H : S:="Storage Dimensions";
    | 0F2H : S:="DOS 3.3+ secondary partition";
    | 0F3H : S:="officially listed as reserved";
    | 0F4H : S:="8-ung ! SpeedStor, or Storage Dimensions";
    | 0F5H : S:="Prologue";
    | 0F6H : S:="officially listed as reserved";
    | 0FDH : S:="reserved for FreeDOS";
    | 0FEH : S:="8-ung ! LANstep, or IBM PS/2 IML (Initial Microcode Load) partition";
    | 0FFH : S:="Xenix bad block table";
    ELSE
        S:=sNoLongID;
    END;
    Str.Copy(sID,S);
END getIDpartStr;

(* most useful/used only ! should match PartInfo *)

CONST
    sNoShortID = "...";

PROCEDURE getIDshortDesc (IDpart:BYTE;isPrimary,pad:BOOLEAN):str16;
CONST
    widesc = 16; (* change strHeader accordingly *)
VAR
    S : str16;
    i : CARDINAL;
BEGIN
    CASE IDpart OF
    | 000H : S:="Empty";
    | 001H : S:="FAT12";
    | 004H : S:="FAT16";
    | 005H : IF isPrimary THEN
             S:="Extended";
             ELSE
             S:="EBPR";
             END;
    | 006H : S:="FAT16B";
    | 007H : S:="NT NTFS"; (* v1.3p *)
    | 00BH : S:="FAT32";
    | 00CH : S:="FAT32X";
    | 00EH : S:="FAT16X";
    | 00FH : IF isPrimary THEN
             S:="ExtendedX"
             ELSE
             S:="EBPRX";
             END;
    | 011H : S:="Hidden FAT12";
    | 014H : S:="Hidden FAT16";
    | 015H : IF isPrimary THEN
             S:="Hidden Extended";
             ELSE
             S:="Hidden EBPR";
             END;
    | 016H : S:="Hidden FAT16B";
    | 017H : S:="Hidden NT NTFS"; (* v1.3p *)
    | 01BH : S:="Hidden FAT32";
    | 01CH : S:="Hidden FAT32X";
    | 01EH : S:="Hidden FAT16X";
    | 01FH : IF isPrimary THEN
             S:="Hidden ExtendedX";
             ELSE
             S:="Hidden EBPRX";
             END;
             (* "1234567890123456" *)
    | 078H : S:="XOSL"; (* this one deserves its entry here too *)
    ELSE
        S:=sNoShortID;
    END;
    IF pad THEN
        WHILE Str.Length(S) < widesc DO
           Str.Append(S, " ");
        END;
    END;
    RETURN S;
END getIDshortDesc;

(* ------------------------------------------------------------ *)

(* data taken from Ralf Brown's Interrupt List 60, with minor tweaking in descriptions *)

CONST
    IDemptyPart             = BYTE(00H);
CONST
    IDextendedPart          = BYTE(05H);
    IDextendedPartLBA       = BYTE(0FH); (* id 05H but using LBA-mode *)
    IDextendedPartHidden    = BYTE(15H);
    IDextendedPartLBAhidden = BYTE(1FH); (* id 05H but using LBA-mode *)

(* don't care about extended partitions, whether primary or logical, which haven't a boot record *)

PROCEDURE isFAT (ID:BYTE):BOOLEAN;
BEGIN
    CASE ID OF
    | 001H, 004H, 006H, 00EH :
        RETURN TRUE;
    | 011H, 014H, 016H, 01EH :
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END isFAT;

PROCEDURE isFAT32 (ID:BYTE):BOOLEAN;
BEGIN
    CASE ID OF
    | 00BH, 00CH :
         RETURN TRUE;
    | 01BH, 01CH :
         RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END isFAT32;

PROCEDURE isNTFS (ID:BYTE):BOOLEAN;
BEGIN
    CASE ID OF
    | 007H, 017H: (* yes, yes, there are others but let's stick to these for now *)
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END isNTFS;

PROCEDURE isEmptyPart (ID:BYTE):BOOLEAN;
BEGIN
    CASE ID OF
    | IDemptyPart :
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END isEmptyPart;

PROCEDURE isExtendedPart (ID:BYTE):BOOLEAN;
BEGIN
    CASE ID OF
    | IDextendedPart, IDextendedPartLBA :
        RETURN TRUE;
    | IDextendedPartHidden, IDextendedPartLBAhidden :
        RETURN TRUE;
    ELSE
        RETURN FALSE;
    END;
END isExtendedPart;

PROCEDURE getHideUnhideStatus (VAR R:ARRAY OF CHAR; IDpart:BYTE):BOOLEAN;
VAR
    rc: BOOLEAN;
    S:str80;
BEGIN
    rc:=TRUE;
    CASE IDpart OF
    | 001H : S:="FAT12";
    | 004H : S:="FAT16";
    | 005H : S:="Extended/EBPR";
    | 006H : S:="FAT16B";
    | 007H : S:="NT NTFS";
    | 00BH : S:="FAT32";
    | 00CH : S:="FAT32X";
    | 00EH : S:="FAT16X";
    | 00FH : S:="ExtendedX/EBPRX";
    | 011H : S:="Hidden FAT12";
    | 014H : S:="Hidden FAT16";
    | 015H : S:="Hidden Extended/EBPR";
    | 016H : S:="Hidden FAT16B";
    | 017H : S:="Hidden NT NTFS";
    | 01BH : S:="Hidden FAT32";
    | 01CH : S:="Hidden FAT32X";
    | 01EH : S:="Hidden FAT16X";
    | 01FH : S:="Hidden ExtendedX/EBPRX";
             (* "12345678901234567890123456789" *)
    ELSE
             S:="unsupported";
             rc:=FALSE;
    END;
    Str.Copy(R,S);
    RETURN rc;
END getHideUnhideStatus;

(* ------------------------------------------------------------ *)

PROCEDURE dmpDoc (  );
CONST
    memento =
"(* MBR *)"+nl+
nl+
"$1b8     4 BYTEs    Windows XP disk ID serial number"+nl+
nl+
"$1be    16 BYTEs    partition record for partition 1"+nl+
"$1ce    16 BYTEs    partition record for partition 2"+nl+
"$1de    16 BYTEs    partition record for partition 3"+nl+
"$1ee    16 BYTEs    partition record for partition 4"+nl+
"$1fe    WORD        signature, AA55h indicates valid boot block"+nl+
nl+
"(* partition record *)"+nl+
nl+
"$00     BYTE        boot indicator (80h = active partition)"+nl+
"$01     BYTE        partition start head"+nl+
"$02     BYTE        partition start sector (bits 0-5)"+nl+
"$03     BYTE        partition start track (bits 8,9 in bits 6,7 of sector)"+nl+
"$04     BYTE        operating system indicator (* partition code *)"+nl+
"$05     BYTE        partition end head"+nl+
"$06     BYTE        partition end sector (bits 0-5)"+nl+
"$07     BYTE        partition end track (bits 8,9 in bits 6,7 of sector)"+nl+
"$08     DWORD       sectors preceding partition"+nl+
"$0C     DWORD       length of partition in sectors"+nl+
nl+
"(* DOS boot record *)"+nl+
nl+
"$000    3 BYTEs     jmpCode               (* eb ?? 90 or e9 ?? ?? *)"+nl+
"$003    8 CHARs     OEMname               (* $00 terminated *)"+nl+
"$00b    CARDINAL    bytesPerSector        (* 512 *)"+nl+
"$00d    SHORTCARD   sectorsPerCluster     (* power of 2 : 1,2,4,...,128 *)"+nl+
"$00e    CARDINAL    reservedSectors       (* 1, starting at 0 *)"+nl+
"$010    SHORTCARD   numberOfFATs          (* 2 *)"+nl+
"$011    CARDINAL    numberOfRootEntries   (* 512 *)"+nl+
"$013    CARDINAL    numberOfTotalSectors  (* 0 if BIGDOS i.e. > 32Mb *)"+nl+
"$015    SHORTCARD   mediaDescriptor       (* $f8 *)"+nl+
"$016    CARDINAL    sectorsPerFAT"+nl+
"$018    CARDINAL    sectorsPerTrack"+nl+
"$01a    CARDINAL    numberOfHeads"+nl+
"$01c    LONGCARD    hiddenSectors         (* ignore high if dos < 4.0 - is first sector of partition *)"+nl+
"$020    LONGCARD    bigTotal              (* if numberOfTotalSectors was 0 *)"+nl+
"$024    SHORTCARD   physicalDriveNumber   (* $80 or $81 *)"+nl+
"$025    BYTE        dirtyflag             (* reserved *)"+nl+
"$026    BYTE        extBRsignature        (* $29 *)"+nl+
"$027    LONGCARD    volumeSerialNumber"+nl+
"$02b    11 CHARs    volumeLabel           (* $00 terminated *)"+nl+
"$036    8 CHARs     fileSystemID          (* padded with blanks *)"+nl+
"$03e    448 BYTEs   xcode"+nl+
"$1fe    CARDINAL    magic                 (* $aa55 *)"+nl+
nl+
"(* Win boot record *)"+nl+
nl+
"$000    3 BYTEs     jmpCode               (* eb 58 90 *)"+nl+
"$003    8 CHARs     OEMname               (* $00 terminated *)"+nl+
"$00b    CARDINAL    bytesPerSector        (* 512 *)"+nl+
"$00d    SHORTCARD   sectorsPerCluster     (* power of 2 : 1,2,4,...,128 *)"+nl+
"$00e    CARDINAL    reservedSectors       (* 33, starting at 0 *)"+nl+
"$010    SHORTCARD   numberOfFATs          (* 2 *)"+nl+
"$011    4 BYTEs     rsvd"+nl+
"$015    SHORTCARD   mediaDescriptor       (* $f8 *)"+nl+
"$016    CARDINAL    sectorsPerFAT"+nl+
"$018    CARDINAL    sectorsPerTrack"+nl+
"$01a    CARDINAL    numberOfHeads"+nl+
"$01c    LONGCARD    hiddenSectors         (* ignore high if dos < 4.0 - is first sector of partition *)"+nl+
"$020    LONGCARD    bigTotal              (* if numberOfTotalSectors was 0 *)"+nl+
"$024    LONGCARD    bigSectorsPerFAT"+nl+
"$028    SHORTCARD   activeFATs            (* attrExt *)"+nl+
"$029    SHORTCARD   FSversionMajor"+nl+
"$02a    CARDINAL    FSversionMinor"+nl+
"$02c    LONGCARD    firstClusterOfRoot    (* 2 *)"+nl+
"$030    CARDINAL    FSinfoSector          (* 1 *)"+nl+
"$032    CARDINAL    FSbackupBootSector    (* 6 *)"+nl+
"$034    12 BYTEs    rsvd2"+nl+
"$040    SHORTCARD   physicalDriveNumber   (* $80 or $81 *)"+nl+
"$041    BYTE        rsvd3                 (* for NT *)"+nl+
"$042    BYTE        extBRsignature        (* $29 *)"+nl+
"$043    LONGCARD    volumeSerialNumber"+nl+
"$047    11 CHARs    volumeLabel           (* $00 terminated *)"+nl+
"$052    8 CHARs     fileSystemID          (* padded with blanks *)"+nl+
"$05a    418 BYTEs   xcode"+nl+
"$1fc    LONGCARD    magic                 (* $aa550000*)"+nl;

(* //FIXME add NTFS boot record too ? *)

CONST
    sep = "     ";
    isPrimary = TRUE;
    isLogical = FALSE;
VAR
    i,j:CARDINAL;
    S1,R1,R2:str16;
    S2:str256;
    ok1,ok2:BOOLEAN;
BEGIN
    FOR j:=1 TO 2 DO
        CASE j OF
        | 1 : S2:="(* long O.S. indicator codes *)";
        | 2 : S2:="(* most common short O.S. indicator codes *)";
        END;
        WrStr(S2);WrLn;
        WrLn;
        FOR i:=MIN(BYTE) TO MAX(BYTE) DO
            getIDpartStr( BYTE(i), S1,S2);
            CASE j OF
            | 1 :
                IF same(S2,sNoLongID)=FALSE THEN
                    WrStr(S1);WrStr(sep);WrStr(S2);WrLn;
                END;
            | 2 :
                R1:=getIDshortDesc ( BYTE(i), isPrimary,FALSE);
                R2:=getIDshortDesc ( BYTE(i), isLogical,FALSE);

                ok1:=( same(R1,sNoShortID)=FALSE );
                ok2:=( same(R2,sNoShortID)=FALSE );
                IF (ok1 OR ok2) THEN
                    WrStr(S1);WrStr(sep);
                    IF same(R1,R2) THEN
                        WrStr(R1);
                    ELSE
                        IF ok1 THEN WrStr(R1);WrStr(" (primary)");END;
                        IF ok2 THEN
                            IF ok1 THEN WrStr(" --- ");END;
                            WrStr(R2);WrStr(" (logical)");
                        END;
                    END;
                    WrLn;
                END;
            END;
        END;
        WrLn;
    END;
    WrStr(memento);
END dmpDoc;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
Format of hard disk master boot sector:
Offset	Size	Description	(Table 00650)
 00h 446 BYTEs	Master bootstrap loader code

(1B8h  4 BYTEs  Windows XP disk ID serial number)

1BEh 16 BYTEs	partition record for partition 1 (see #00651)
1CEh 16 BYTEs	partition record for partition 2
1DEh 16 BYTEs	partition record for partition 3
1EEh 16 BYTEs	partition record for partition 4
1FEh	WORD	signature, AA55h indicates valid boot block

Format of partition record:
Offset	Size	Description	(Table 00651)
 00h	BYTE	boot indicator (80h = active partition)
 01h	BYTE	partition start head
 02h	BYTE	partition start sector (bits 0-5)
 03h	BYTE	partition start track (bits 8,9 in bits 6,7 of sector)
 04h	BYTE	operating system indicator (see #00652)
 05h	BYTE	partition end head
 06h	BYTE	partition end sector (bits 0-5)
 07h	BYTE	partition end track (bits 8,9 in bits 6,7 of sector)
 08h	DWORD	sectors preceding partition
 0Ch	DWORD	length of partition in sectors
SeeAlso: #00650
*)

CONST
    sectorSize        = 512;
    firstByteInSector = 0;
    lastByteInSector  = sectorSize-1;
    firstPartNdx      = 1;
    lastPartNdx       = 4;
TYPE
    SectorType        = ARRAY [firstByteInSector..lastByteInSector] OF BYTE;
CONST
    firstLoaderCodeByte     = 0;
    winXPuniqueIDoffset     = 01B8H; (* LONGCARD : M$ bastards taking liberties with standards ! *)
    lastLoaderCodeByte      = 01BDH;                (* I love hard-coded constants... sometimes ;-) *)
    firstPartitionTableByte = lastLoaderCodeByte+1; (* $01bd+1=$01be til $1fd *)
    lastPartitionTableByte  = 01FFH-1-1;            (* ignore $AA55 at $1fe..$1ff *)
    validSignature          = 0AA55H;
TYPE
    PartitionRecord = RECORD
        BootIndicator    : BYTE; (* 80h is active *)
        StartHead        : BYTE;
        StartSector      : BYTE; (* bits 0-5 *)
        StartTrack       : BYTE; (* bits 8-9 in bits 6-7 of StartSector *)
        OSindicator      : BYTE;
        EndHead          : BYTE;
        EndSector        : BYTE; (* bits 0-5 *)
        EndTrack         : BYTE; (* bits 8-9 in bits 6-7 of EndSector *)
        SectorsPreceding : LONGCARD;  (* DWORD *)
        LengthInSectors  : LONGCARD;  (* DWORD *)
    END;
    MBRrecordType = RECORD
        CASE : BOOLEAN OF
        | TRUE:
            LoaderCode : ARRAY [firstLoaderCodeByte..lastLoaderCodeByte] OF BYTE; (* $000..$1bd *)
            PartitionX : ARRAY [firstPartNdx..lastPartNdx] OF PartitionRecord;    (* $1be..$1fd *)
            Signature  : CARDINAL; (* $aa55 *)                                    (* $1fe..$1ff *)
        | FALSE :
            bytearray  : SectorType;
        END;
    END;

(* ------------------------------------------------------------ *)

(*
good old Ralf is of no use here, or very, very partially :
INT 21 - DOS 2+ internal - TRANSLATE BIOS PARAMETER BLOCK TO DRIVE PARAM BLOCK
INT 21 - Windows95 - FAT32 - "Get_ExtDPB" - GET EXTENDED DPB

source code from FIPS, PART and XOSL were a little more useful for boot structures
though they disagree on FAT32 boot record parsing
*)

TYPE DOSbrType = RECORD
    jmpCode             : ARRAY [1..3] OF BYTE; (* "eb ?? 90" or  "e9 ?? ??" *)
    OEMname             : ARRAY [0..7] OF CHAR; (* $00 terminated *)
    bytesPerSector      : CARDINAL;             (* 512 *)
    sectorsPerCluster   : SHORTCARD;            (* power of 2 : 1,2,4,...,128 *)
    reservedSectors     : CARDINAL;             (* 1, starting at 0 *)
    numberOfFATs        : SHORTCARD;            (* 2 *)
    numberOfRootEntries : CARDINAL;             (* 512 *)
    numberOfTotalSectors: CARDINAL;             (* 0 if BIGDOS i.e. > 32Mb *)
    mediaDescriptor     : SHORTCARD;            (* $f8 *)
    sectorsPerFAT       : CARDINAL;
    sectorsPerTrack     : CARDINAL;
    numberOfHeads       : CARDINAL;
    hiddenSectors       : LONGCARD;             (* ignore high if dos < 4.0 - is first sector of partition *)
    bigTotal            : LONGCARD;             (* if numberOfTotalSectors was 0 *)
    physicalDriveNumber : SHORTCARD;            (* $80 or $81 *)
    dirtyflag           : BYTE;                 (* reserved *)
    extBRsignature      : BYTE;                 (* $29 *)
    volumeSerialNumber  : LONGCARD;
    volumeLabel         : ARRAY [0..10]  OF CHAR; (* $00 terminated *)
    fileSystemID        : ARRAY [0..7]   OF CHAR; (* padded with blanks *)
    xcode               : ARRAY [1..448] OF BYTE;

    magic               : CARDINAL;             (* $aa55 *)
END;

TYPE WINbrType = RECORD
    jmpCode             : ARRAY [1..3] OF BYTE; (* "eb 58 90" *)
    OEMname             : ARRAY [0..7] OF CHAR; (* $00 terminated *)
    bytesPerSector      : CARDINAL;             (* 512 *)
    sectorsPerCluster   : SHORTCARD;            (* power of 2 : 1,2,4,...,128 *)
    reservedSectors     : CARDINAL;             (* 33, starting at 0 *)
    numberOfFATs        : SHORTCARD;            (* 2 *)
    rsvd                : ARRAY [1..4] OF BYTE;
    mediaDescriptor     : SHORTCARD;            (* $f8 *)
    sectorsPerFAT       : CARDINAL;
    sectorsPerTrack     : CARDINAL;
    numberOfHeads       : CARDINAL;
    hiddenSectors       : LONGCARD;             (* ignore high if dos < 4.0 - is first sector of partition *)
    bigTotal            : LONGCARD;             (* if numberOfTotalSectors was 0 *)
    bigSectorsPerFAT    : LONGCARD;
    activeFATs          : SHORTCARD;            (* attrExt *)
    FSversionMajor      : SHORTCARD;
    FSversionMinor      : CARDINAL;
    firstClusterOfRoot  : LONGCARD;             (* 2 *)
    FSinfoSector        : CARDINAL;             (* 1 *)
    FSbackupBootSector  : CARDINAL;             (* 6 *)
    rsvd2               : ARRAY [1..12] OF BYTE;
    physicalDriveNumber : SHORTCARD;            (* $80 or $81 *)
    rsvd3               : BYTE;                 (* for NT *)
    extBRsignature      : BYTE;                 (* $29 *)
    volumeSerialNumber  : LONGCARD;
    volumeLabel         : ARRAY [0..10]  OF CHAR; (* $00 terminated *)
    fileSystemID        : ARRAY [0..7]   OF CHAR; (* padded with blanks *)
    xcode               : ARRAY [1..418] OF BYTE;
    magic               : LONGCARD;             (* $aa550000*)
END;

(* many conflicting sources : visopsys, WinHex, LinuxNFTS, testdisk, etc. *)
(* this crap filesystem may use several blocks here ! *)

(* defined supra
TYPE VERYLONGCARD = RECORD (* 64 bits *)
    lo : LONGCARD;
    hi : LONGCARD;
END;
*)

TYPE NTFSbrType = RECORD
    jmpCode             : ARRAY [1..3] OF BYTE; (* 00: $eb ?? 90 *)
    OEMname             : ARRAY [0..7] OF CHAR; (* 03: "NTFS    " *)
    bytesPerSector      : CARDINAL;             (* 0b: *)
    sectorsPerCluster   : SHORTCARD;            (* 0d: power of 2 *)
    reservedSectors     : CARDINAL;             (* 0e: *)
    numberOfFATs        : SHORTCARD;            (* 10: *)
    rootentries         : CARDINAL;             (* 11: *)
    numberOfSectors     : CARDINAL;             (* 13: *)
    mediaDescriptor     : SHORTCARD;            (* 15: *)
    FATlength           : CARDINAL;             (* 16: *)
    sectorsPerTrack     : CARDINAL;             (* 18: *)
    numberOfHeads       : CARDINAL;             (* 1a: *)
    hiddenSectors       : LONGCARD;             (* 1c: *)
    totalSectors        : LONGCARD;             (* 20: *)
    physicalDriveNumber : SHORTCARD;            (* 24: *)
    currhead            : SHORTCARD;            (* 25: *)
    xtdbootsig          : SHORTCARD;            (* 26: *)
    rsvd1               : SHORTCARD;            (* 27: *)
    sectorsPerVolume    : VERYLONGCARD;         (* 28: *)
    mftStart            : VERYLONGCARD;         (* 30: LCN of VCN 0 of the $MFT *)
    mftMirrorStart      : VERYLONGCARD;         (* 38: same for $MFTmirror *)
    clustersPerMFT      : SHORTCARD;            (* 40: *)
    rsvd2               : ARRAY[1..3] OF BYTE;  (* 41: *)
    clustersPerIndex    : SHORTCARD;            (* 44: *)
    rsvd3               : ARRAY[1..3] OF BYTE;  (* 45: *)
    volumeSerialNumber  : VERYLONGCARD;         (* 48: *)
    checksum            : LONGCARD;             (* 50: *)
    xcode               : ARRAY [1..426] OF BYTE; (* 54: *)
    magic               : CARDINAL;             (* 1fe: $aa55 *)
END;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE dmp (S:ARRAY OF CHAR   );
BEGIN
    WrStr(S);
END dmp;

PROCEDURE dmpwi (padwi:CARDINAL;S:ARRAY OF CHAR);
VAR
    i:CARDINAL;
BEGIN
    dmp(S);
    FOR i:=Str.Length(S)+1 TO padwi DO dmp(" ");END;
END dmpwi;

CONST
    dash5  = "=====";
    dash10 = dash5+dash5;
    dashes = dash10+dash10+dash10+dash10+dash10+dash10+dash10+dash5; (* was 70 pre-v1.3u *)

PROCEDURE chapter (S:ARRAY OF CHAR);
BEGIN
    dmp(dashes+nl);
    dmp(S);
    dmp(nl);
    dmp(dashes+nl+nl);
END chapter;

PROCEDURE chapterF ( S1:ARRAY OF CHAR;F:pathtype );
VAR
    S:str256;
BEGIN
    S:='~ : "~"';
    Str.Subst(S,"~",S1);
    Str.Subst(S,"~",F);
    chapter(S);
END chapterF;

PROCEDURE dmp1 (S:ARRAY OF CHAR   );
BEGIN
    dmp(S);
    dmp(nl);
END dmp1;

PROCEDURE dmp2 (S1,S2:ARRAY OF CHAR);
BEGIN
    dmpwi(1,S1);
    dmp(" : ");
    dmp(S2);
    dmp(nl);
END dmp2;

PROCEDURE warn (S:ARRAY OF CHAR   );
BEGIN
    dmp("WARNING ! ");
    dmp(S);
    dmp(nl);
END warn;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR (* globerks *)
    buffSector                  : SectorType;
    buffMBR                     : MBRrecordType;

PROCEDURE dmpHexa (hcount:CARDINAL   );
CONST
    sepaddr = " :";
    sepascii= " | ";
VAR
    i,j,v:CARDINAL;
    SH,SA:str128;
    alcatraz:BOOLEAN;
BEGIN
    i:=firstByteInSector;
    alcatraz:=FALSE;
    LOOP
        Str.Concat(SH,fmtcard(i,16,4,"0","",""),sepaddr);
        Str.Copy(SA,sepascii);
        j:=1;
        LOOP
            IF alcatraz THEN
                Str.Append(SH, " "); Str.Append(SH, "  "); (* ## *)
                Str.Append(SA," ");
            ELSE
                v:=CARDINAL(buffSector[i]);
                Str.Append(SH, " "); Str.Append(SH, fmtcard(v,16,2,"0","",""));
                Str.Append(SA,fmtchar(FALSE, BYTE(v) ));
            END;
            INC(i);
            IF i > lastByteInSector THEN alcatraz:=TRUE; END;
            INC(j);
            IF j > hcount THEN EXIT; END;
        END;
        Str.Append(SH,SA);Str.Append(SH,nl);
        dmp(SH);
        IF alcatraz THEN EXIT; END;
    END;
END dmpHexa;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE dmpMBR ( hcount,whatsizeunit:CARDINAL; altformat,showraw:BOOLEAN; F:pathtype );
CONST
    wiblock  = 10+3;  (* #.###.###.### *)
    wipretty = 6+1+2; (* ######.## *)
    sepgroup = comma; (* us format *)
    sepfrac  = dot;   (* us format *)
CONST
    strPri      = "Pri";
    strExt      = "Ext";
    strEmpty    = "...";
    strBootable = ",Boot";
    strData     = "     ";
    mbrheader =
   "#  Type Status    Terse ID          Start          Sectors count  Size";
(* "?  $??  ???,Boot  1234567890123456  #,###,###,###  #,###,###,###  #" *)

VAR
    id : LONGCARD;
    i  : CARDINAL;
    bootflag,osid:SHORTCARD;
    preceding,partlen:LONGCARD;
    S:str128;
    partcount,bootablecount,badflagcount,warnings:CARDINAL;
    isPrimary,isBootable:BOOLEAN;
    R,RR:str80; (* "#,###,###,###" = 13 but "Who knows what evil..." etc. *)
BEGIN
    IF hcount # 0 THEN
       chapterF("dumped as raw data",F);
       dmpHexa(hcount);
       dmp(nl);
    END;
    chapterF("dumped as master boot record (MBR)",F);

    Lib.Move( ADR(buffMBR.bytearray[winXPuniqueIDoffset]), ADR(id), SIZE(id));
    dmp2("Windows XP disk ID",fmtlc(id,16,8,"0","$","") );
    dmp(nl);

    chapter(mbrheader);

    partcount:=0;
    bootablecount:=0;
    badflagcount:=0;

    FOR i:= firstPartNdx TO lastPartNdx DO
        bootflag := buffMBR.PartitionX[i].BootIndicator; (* $80=boot *)
        osid     := buffMBR.PartitionX[i].OSindicator;
        preceding:= buffMBR.PartitionX[i].SectorsPreceding;
        partlen  := buffMBR.PartitionX[i].LengthInSectors;

        S:="~  ~  ~~  ~  ~  ~  ~";
        Str.Subst(S,"~", fmtcard(i,10,1,"","","") );
        Str.Subst(S,"~", fmtbyte(osid,16,2,"0","$","") );

        isPrimary := NOT (isExtendedPart(osid) );
        IF isEmptyPart(osid) THEN
            R:=strEmpty;
        ELSE
            INC(partcount);
            IF isPrimary THEN
                R:=strPri;
            ELSE
                R:=strExt;
            END;
        END;
        Str.Subst(S,"~",R);

        isBootable:=FALSE;
        CASE CARDINAL(bootflag) OF
        | 0 : ;
        | CARDINAL(80H) : INC(bootablecount); isBootable:=TRUE;
        ELSE   INC(badflagcount);
        END;
        IF isBootable THEN
            R:=strBootable;
        ELSE
            R:=strData;
        END;
        Str.Subst(S,"~",R);

        Str.Subst(S,"~", getIDshortDesc(osid,isPrimary,TRUE) );

        IF altformat THEN
            Str.Copy(R, fmtlc(preceding,10,wiblock,blank,"","") );
            Str.Copy(RR,fmtlc(partlen  ,10,wiblock,blank,"","") );
        ELSE
            Str.Copy(R, beautifiedlc(preceding,blank,sepgroup,wiblock) );
            Str.Copy(RR,beautifiedlc(partlen,blank,sepgroup,wiblock) );
        END;
        Str.Subst(S,"~",R);
        Str.Subst(S,"~",RR);

        R:= prettySize (whatsizeunit,wipretty,fix2,partlen," ",sepfrac);
        Str.Subst(S,"~",R);

        dmp1(S);
    END;

    warnings:=0;
    IF partcount = 0 THEN INC(warnings);END;
    IF bootablecount # 1 THEN INC(warnings);END;
    IF badflagcount # 0 THEN INC(warnings);END;
    IF warnings # 0 THEN dmp(nl);END;

    IF partcount = 0 THEN warn("No partition defined !");END;
    CASE bootablecount OF
    | 0 : warn("Make sure there is one active bootable partition !");
    | 1 : ;
    ELSE  warn("Make sure there is only one active bootable partition !");
    END;
    IF badflagcount # 0 THEN warn("Make sure there are no illegal values for boot flags !");END;

END dmpMBR;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    blocktype = (undefined,mbr,primary,logical);
CONST
    firstparm = 1;
    maxparm = 1;
VAR
    parm : ARRAY [firstparm..maxparm] OF pathtype;
    lastparm:CARDINAL;
    i,opt,parmcount:CARDINAL;
    F,S,R:pathtype;
    sigchk,showraw,altformat,useLFN,DEBUG:BOOLEAN;
    whatblock : blocktype;
    whatsizeunit,hcount,got,wanted:CARDINAL;
    hin:FIO.File;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck:=FALSE;       (* always ! *)
    WrLn;

    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;

    whatblock     := undefined;
    altformat     := FALSE;
    showraw       := FALSE;
    hcount        := 0;
    whatsizeunit  := useMega;
    sigchk        := FALSE;
    useLFN        := TRUE;
    DEBUG         := FALSE;
    lastparm      := firstparm-1;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "M"+delim+"MBR"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "B"+delim+"H16"+delim+"HEX"+delim+
                                  "BB"+delim+"H8"+delim+
                                  "??"+delim+"MOREHELP"+delim+
                                  "A"+delim+"ALT"+delim+"ALTFORMAT"+delim+
                                  "G"+delim+"GIGA"+delim+
                                  "S"+delim+"AA55"+delim+"SIG"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3 :   abort(errHelp,"");
            | 4,5:      CASE whatblock OF
                        | undefined,mbr:whatblock:=mbr;
                        ELSE abort(errOnlyOne,"");
                        END;
            | 6,7:      useLFN := FALSE;
            | 8,9,10:   hcount:=16;
            | 11,12:    hcount:=8;
            | 13,14:    dmpDoc();
                        abort(errMoreHelp,"");
            | 15,16,17: altformat:=TRUE;
            | 18,19:    whatsizeunit:=useGiga;
            | 20,21,22: sigchk:=TRUE;
            | 23:       DEBUG:=TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParameter,S);END;
            Str.Copy( parm[lastparm], R);
        END;
    END;
    useLFN := ( useLFN AND fileSupportLFN() );

    CASE lastparm OF
    | firstparm-1 : abort(errExpected,"");
    | firstparm :
        F:=parm[firstparm];
        IF chkJoker(F) THEN abort(errJoker,F);END;

        IF Str.RCharPos(F,dot)=MAX(CARDINAL) THEN
            i:=0;
            LOOP
                isoleItemS(R, defaultExtensions,delim,i);
                IF same(R,"") THEN abort(errExhausted,parm[firstparm]); END;
                Str.Concat(F,parm[firstparm],R);
                IF fileExists(useLFN,F) THEN EXIT; END;
                INC(i);
            END;
        END;
        IF fileExists(useLFN,F)=FALSE THEN abort(errNotFound,F);END;
    ELSE
        ; (* cannot happen *)
    END;
    useLFN:=(useLFN AND fileSupportLFN() );

    wanted:=sectorSize;
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);
    got:=FIO.RdBin(hin,buffSector,wanted);
    fileClose(useLFN,hin);
    IF got # wanted THEN abort(errTooSmall,F);END;

    (* copy buffer data TO MBR buffer too FOR ease OF parsing *)
    Lib.Move( ADR(buffSector), ADR(buffMBR), sectorSize);

    IF sigchk THEN
        IF buffMBR.Signature # validSignature THEN abort(errAA55,F);END;
    END;

    CASE whatblock OF
    | undefined,mbr : dmpMBR    (hcount,whatsizeunit,altformat,showraw,F);
    END;

    abort(errNone,"");
END BlockDmp.


