(* --------------------------------------------------------------
Title         Q&D Cells
Author        PhG
Overview
Usage
Notes
Bugs

Wish List     to do : a prettier display !
              for each quadrant, we should :
              - build a group
              - find or build a lead
              - focus each cell on lead
              - if group too big, repulse instead of attract

              use move to target, move from target if group becomes larger

--------------------------------------------------------------- *)

MODULE Cells;

IMPORT Lib;
IMPORT Str;
IMPORT MATHLIB;
IMPORT SYSTEM;
IMPORT BiosIO;
IMPORT MsMouse;

FROM IO IMPORT WrStr,WrLn;

FROM Lib IMPORT FarWordMove;

FROM Storage IMPORT ALLOCATE;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_rand IMPORT InitRnd, GetRnd,
GetRndCardRange, GetRndLngCardRange,
GetRndLngRealRange, GetRndRealRange,
GetRndIntRange, GetRndLngIntRange;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "CELLS";
    ProgTitle     = "Q&D Cells demo";
    ProgVersion   = "v1.0";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone           = 0;
    errHelp           = 1;
    errOpt            = 2;
    errParm           = 3;
    errRange          = 4;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" [option]..."+nl+
nl+
"  -f:#  number of cells ([1..1400], default is 200)"+nl+
"  -a    show palette until keypress or 10 seconds"+nl+
"  -z    end on mouseclick too"+nl+
"  -p:#  color palette (0..8, default is 0)"+nl+
"  -d:#  number of waits for vertical retrace ([0..20], default is 0)"+nl+
"  -s    small dots"+nl+
(* "  -m    alternate motion mode"+nl+ *)
nl+
"  Here are the keys to interact with program while it is running :"+nl+
"  [Ss]-toggle small dots  [Home|End]-speed"+nl+ (*  [Mm]-toggle motion mode"+nl+ *)
"  [rRGB1230!]-palette  [Space]-pause  [Escape|Enter]-exit"+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp:    WrStr(helpmsg);
    | errOpt:     S:="Illegal ~ option !";
    | errParm:    S:="Unexpected ~ parameter !";
    | errRange :  S:="~ value is out of expected range !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp :
        ; (* nada *)
    ELSE
        Str.Subst(S,"~",'"~"');
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    PI : LONGREAL; (* globerk in CASE value OF PI would change ;-) *)

PROCEDURE initTrigo (  );
BEGIN
    PI := 4.0 * MATHLIB.ATan(1.0);
END initTrigo;

PROCEDURE initRAND (  );
BEGIN
    InitRnd();
END initRAND;

PROCEDURE getrndrange (mini,maxi:CARDINAL ):CARDINAL;
BEGIN
    RETURN GetRndCardRange(mini,maxi);
END getrndrange;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    keyEscape  = 01B00H;
    keySpace   = 02000H;
    keyCR      = 00D00H;
    keyPageUp  = 00049H;
    keyPageDn  = 00051H;
    keyEnd     = 0004FH;
    keyHome    = 00047H;
    keyDivide  = ORD("/") << 8 ;
    keyStar    = ORD("*") << 8 ;
    upperR     = ORD("R") << 8 ;
    lowerR     = ORD("r") << 8 ;
    upperG     = ORD("G") << 8 ;
    lowerG     = ORD("g") << 8 ;
    upperB     = ORD("B") << 8 ;
    lowerB     = ORD("b") << 8 ;
    upperS     = ORD("S") << 8 ;
    lowerS     = ORD("s") << 8 ;
    upperM     = ORD("M") << 8 ;
    lowerM     = ORD("m") << 8 ;
    keyLeft    = 0004BH;
    keyRight   = 0004DH;
    key1       = ORD("1") << 8 ;
    key2       = ORD("2") << 8 ;
    key3       = ORD("3") << 8 ;
    key0       = ORD("0") << 8 ;
    keyBang    = ORD("!") << 8 ;

PROCEDURE flushKeyboard (  );
VAR
    c : CHAR;
BEGIN
    LOOP
        IF BiosIO.KeyPressed()=FALSE THEN EXIT; END;
        c := BiosIO.RdKey();
        IF c = CHR(0) THEN c := BiosIO.RdKey(); END;
    END;
END flushKeyboard;

PROCEDURE getKeyboardCode (VAR keycode:CARDINAL):BOOLEAN;
VAR
    c1,c2:CHAR;
BEGIN
    IF BiosIO.KeyPressed()=FALSE THEN RETURN FALSE; END;
    c1 := BiosIO.RdKey();
    IF c1 = CHR(0) THEN
        c2 := BiosIO.RdKey();
    ELSE
        c2 := CHR(0);
    END;
    keycode := (ORD(c1) << 8) + ORD(c2);
    RETURN TRUE;
END getKeyboardCode;

PROCEDURE input (maxWait:LONGINT); (* in seconds *)
VAR
    keycode:CARDINAL;
    h,m,s,ss:CARDINAL;
    start : LONGINT;
    now   : LONGINT;
BEGIN
    Lib.GetTime(h,m,s,ss);
    start := (LONGINT(h)*60+LONGINT(m))*60+LONGINT(s);
    LOOP
        IF getKeyboardCode(keycode) THEN EXIT; END;
        Lib.GetTime(h,m,s,ss);
        now := (LONGINT(h)*60+LONGINT(m))*60+LONGINT(s);
        IF ABS(now-start) > maxWait THEN EXIT; END;
    END;
END input;

PROCEDURE mouseclick (  ):BOOLEAN;
VAR
    msdata:MsMouse.MsData;
BEGIN
    MsMouse.GetStatus(msdata);
    IF msdata.left_pressed THEN RETURN TRUE; END;
    IF msdata.right_pressed THEN RETURN TRUE; END;
    RETURN msdata.middle_pressed;
END mouseclick;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    hiresmode   = 13H; (* 320x200x8 *)
    textmode    = 03H; (* 80x25 *)
CONST
    xcount      = 320; xmin = 0; xmax = xcount-1;
    ycount      = 200; ymin = 0; ymax = ycount-1;
    screensize  = xcount * ycount;
    screensizeW =(screensize DIV 2);
    screenwidth = xcount;
    screenheight= ycount;
TYPE
    screenType = ARRAY [0..screensize-1 +1+xcount] OF BYTE;   (* 320x200=64000 with security *)
VAR
    WorkScreen    : POINTER TO screenType;
    SmoothScreen  : POINTER TO screenType;
    NextWorkScreen: POINTER TO screenType;
VAR
    videoscreen [0A000H:0000H] : ARRAY [0..screensize-1] OF BYTE;
    Ybase                      : ARRAY [ymin..ymax] OF CARDINAL;

(* ------------------------------------------------------------ *)

PROCEDURE initYbase ();
VAR
    i,p:CARDINAL;
BEGIN
    p := 0;
    FOR i := ymin TO ymax DO
        Ybase[i]:=p;
        INC(p,xcount);
    END;
END initYbase;

PROCEDURE setVideoMode (mode:CARDINAL);
VAR
    R : SYSTEM.Registers;
BEGIN
    R.AH := 00H;
    R.AL := BYTE(mode);
    Lib.Intr(R,10H);
END setVideoMode;

PROCEDURE waitVGAretrace ();
BEGIN
    WHILE (SYSTEM.In(03DAH) AND 08H) # 0 DO
    END;
    WHILE (SYSTEM.In(03DAH) AND 08H) = 0 DO
    END;
END waitVGAretrace;

PROCEDURE clearvideoscreen (inkndx:CARDINAL);
BEGIN
    Lib.FarWordFill(FarADR(videoscreen),screensizeW,(inkndx << 8 + inkndx));
END clearvideoscreen;

PROCEDURE clearworkscreen(inkndx:CARDINAL );
BEGIN
    Lib.FarWordFill(FarADR(WorkScreen^),screensizeW,(inkndx << 8 + inkndx));
END clearworkscreen;

PROCEDURE clearsmoothscreen(inkndx:CARDINAL);
BEGIN
    Lib.FarWordFill(FarADR(SmoothScreen^),screensizeW,(inkndx << 8 + inkndx));
END clearsmoothscreen;

PROCEDURE clearnextworkscreen(inkndx:CARDINAL);
BEGIN
    Lib.FarWordFill(FarADR(NextWorkScreen^),screensizeW,(inkndx << 8 + inkndx));
END clearnextworkscreen;

PROCEDURE nextworkscreen2workscreen ();
BEGIN
    FarWordMove(FarADR(NextWorkScreen^),FarADR(WorkScreen^),screensizeW);
END nextworkscreen2workscreen;

PROCEDURE smooth2video();
BEGIN
    waitVGAretrace;
    FarWordMove(FarADR(SmoothScreen^),FarADR(videoscreen),screensizeW);
END smooth2video;

PROCEDURE work2video();
BEGIN
    waitVGAretrace;
    FarWordMove(FarADR(WorkScreen^),FarADR(videoscreen),screensizeW);
END work2video;

PROCEDURE wplot (x,y:CARDINAL;ink:BYTE);
BEGIN
    WorkScreen^[Ybase[y]+x]:=ink;
END wplot;

PROCEDURE splot (x,y:CARDINAL;ink:BYTE);
BEGIN
    SmoothScreen^[Ybase[y]+x]:=ink;
END splot;

(* ------------------------------------------------------------ *)

CONST
    black       = LONGCARD(0000000H);
    white       = LONGCARD(03F3F3FH); (* ega/vga range is $00..$3F *)
    red         = LONGCARD(03F0000H);
    green       = LONGCARD(0003F00H);
    darkgreen   = LONGCARD(0002000H);
    blue        = LONGCARD(000003FH);
    darkblue    = LONGCARD(0000020H);
    cyan        = LONGCARD(0002F3FH);
    yellow      = LONGCARD(03F3F00H);
    darkred     = LONGCARD(0200000H);
    orange      = LONGCARD(02F2F00H);
CONST
    egarange    = 40H;
    mininkindex = 0;
    maxinkindex = 256-1;
    ndxblack    = mininkindex;
    maxcolorindex=egarange-1; (* we use 64 colors palette here *)

PROCEDURE setDAC(index:CARDINAL;red,green,blue:BYTE);
CONST
    DACWriteIndex  = 03C8H;
    DACDataRegister= 03C9H;
BEGIN
    SYSTEM.Out (DACWriteIndex,SHORTCARD(index));
    SYSTEM.Out (DACDataRegister, red);
    SYSTEM.Out (DACDataRegister, green);
    SYSTEM.Out (DACDataRegister, blue);
END setDAC;

PROCEDURE readDACrgb (index:CARDINAL):LONGCARD;
CONST
    DACReadIndex   = 03C7H;
    DACDataRegister= 03C9H;
VAR
    r,g,b:SHORTCARD;
    rgb : LONGCARD;
BEGIN
    SYSTEM.Out (DACReadIndex,SHORTCARD(index));
    r:=SYSTEM.In (DACDataRegister);
    g:=SYSTEM.In (DACDataRegister);
    b:=SYSTEM.In (DACDataRegister);
    rgb := (LONGCARD(r) << 16) + (LONGCARD(g) << 8) + LONGCARD(b);
    RETURN rgb;
END readDACrgb;

PROCEDURE resetdac (index,red,green,blue:CARDINAL);
BEGIN
    setDAC(index,BYTE(red),BYTE(green),BYTE(blue));
END resetdac;

PROCEDURE maxmin (VAR bigger,smaller:INTEGER);
VAR
    tmp:INTEGER;
BEGIN
    IF smaller <= bigger THEN RETURN; END;
    tmp:=bigger;
    bigger:=smaller;
    smaller:=tmp;
END maxmin;

PROCEDURE blend (ndx,count:CARDINAL; startink, endink:LONGCARD);
CONST
    rshift = LONGCARD(16);
    gshift = LONGCARD(8);
VAR
    r1,g1,b1:INTEGER;
    r2,g2,b2:INTEGER;
    r,g,b:INTEGER;
    i : CARDINAL;
BEGIN
    IF count=0 THEN count := 1; END; (* safety to avoid division by 0 ! *)

    r1 := INTEGER (startink >> rshift) MOD egarange;
    g1 := INTEGER (startink >> gshift) MOD egarange;
    b1 := INTEGER (startink          ) MOD egarange;
    r2 := INTEGER (endink   >> rshift) MOD egarange;
    g2 := INTEGER (endink   >> gshift) MOD egarange;
    b2 := INTEGER (endink            ) MOD egarange;

    (* certainly useless but... *)
(*
    maxmin(r2,r1);
    maxmin(g2,g1);
    maxmin(b2,b1);
*)
    FOR i := 1 TO count DO
        r := r1 + ((r2 - r1) * INTEGER(i) ) DIV INTEGER(count);
        g := g1 + ((g2 - g1) * INTEGER(i) ) DIV INTEGER(count);
        b := b1 + ((b2 - b1) * INTEGER(i) ) DIV INTEGER(count);
        IF r < 0 THEN r:=0; END;
        IF g < 0 THEN g:=0; END;
        IF b < 0 THEN b:=0; END;
        IF r >= egarange THEN r:=egarange-1; END;
        IF g >= egarange THEN g:=egarange-1; END;
        IF b >= egarange THEN b:=egarange-1; END;
        setDAC(ndx+i-1,BYTE(r),BYTE(g),BYTE(b));
    END;
END blend;

PROCEDURE normalize (VAR lc:LONGCARD);
VAR
    r,g,b:CARDINAL;
BEGIN
    r := CARDINAL(lc MOD 100H);
    g := CARDINAL(lc DIV 100H) MOD 100H;
    b := CARDINAL(lc DIV 10000H);
    r := r >> 2; (* MOD egarange; *)
    g := g >> 2;
    b := b >> 2;
    lc := LONGCARD(b) << 16 + LONGCARD(g) << 8 + LONGCARD(r);
END normalize;

PROCEDURE genRGB (  ):LONGCARD;
CONST
    dimmest   = 000H;
    brightest = 03FH;
VAR
    r,g,b : CARDINAL;
    bgr : LONGCARD;
BEGIN
    r := getrndrange(dimmest,brightest);
    g := getrndrange(dimmest,brightest);
    b := getrndrange(dimmest,brightest);
    bgr := LONGCARD(b) << 16 + LONGCARD(g) << 8 + LONGCARD(r);
    RETURN bgr;
END genRGB;

TYPE
    fixedpal64type = ARRAY [0..64*3-1] OF BYTE;
CONST
    fixedfirepal = fixedpal64type(
     0, 0, 0,
     1, 0, 0,
     2, 0, 0,
    10, 0, 0,
    18, 0, 0,
    26, 0, 0,
    32, 0, 0,
    36, 0, 0,
    39, 0, 0,
    42, 0, 0,
    46, 1, 0,
    50, 3, 0,
    54, 5, 0,
    58, 7, 0,
    63, 9, 0,
    63,11, 0,
    63,13, 0,
    63,15, 0,
    63,17, 0,
    63,19, 0,
    63,21, 0,
    63,23, 0,
    63,25, 0,
    63,27, 0,
    63,29, 0,
    63,31, 0,
    63,33, 0,
    63,35, 0,
    63,38, 0,
    63,40, 0,
    63,42, 0,
    63,44, 0,
    63,46, 0,
    63,48, 0,
    63,50, 0,
    63,52, 0,
    63,53, 0,
    63,54, 0,
    63,54, 0,
    63,55, 0,
    63,56, 0,
    63,57, 0,
    63,58, 0,
    63,59, 0,
    63,60, 0,
    63,61, 0,
    63,62, 0,
    63,63, 1,
    63,63, 5,
    63,63, 9,
    63,63,12,
    63,63,16,
    63,63,20,
    63,63,23,
    63,63,27,
    63,63,31,
    63,63,34,
    63,63,38,
    63,63,42,
    63,63,45,
    63,63,49,
    63,63,52,
    63,63,56,
    63,63,60);

TYPE
    palettetype = (firepal,
                   rpal,gpal,bpal,
                   palrouge,palvert,palbleu,
                   palorg,
                   palgris
                   );

PROCEDURE newPalette64 (pal:palettetype);
VAR
    ndx : CARDINAL;
    n   : CARDINAL;
    i   : CARDINAL;
    r,g,b:BYTE;
    ir,ig,ib:CARDINAL;
BEGIN
    waitVGAretrace(); (* reduce noise on screen *)
    CASE pal OF
    | rpal:
        ndx := 0;         n := 2;
        FOR i := 1 TO n DO resetdac(ndx+i-1,00H,00H,00H);END;
        INC(ndx,n);       n := 8;
        blend (ndx,n,black,darkred);
        INC(ndx,n);       n := 19;
        blend (ndx,n,darkred,red);
        INC(ndx,n);       n := 19;
        blend (ndx,n,red,yellow);
        INC(ndx,n);       n := 16;
        blend (ndx,n,yellow,white);
    | gpal:
        ndx := 0;         n := 2;
        FOR i := 1 TO n DO resetdac(ndx+i-1,00H,00H,00H);END;
        INC(ndx,n);       n := 16;
        blend (ndx,n,black,darkgreen);
        INC(ndx,n);       n := 15;
        blend (ndx,n,darkgreen,green);
        INC(ndx,n);       n := 15;
        blend (ndx,n,green,yellow);
        INC(ndx,n);       n := 16;
        blend (ndx,n,yellow,white);
    | bpal:                           (* 2 18 14 30 *)
        ndx := 0;         n := 2;
        FOR i := 1 TO n DO resetdac(ndx+i-1,00H,00H,00H);END;
        INC(ndx,n);       n := 8;
        blend (ndx,n,black,darkblue);
        INC(ndx,n);       n := 8;
        blend (ndx,n,darkblue,blue);
        INC(ndx,n);       n := 20;
        blend (ndx,n,blue,cyan);
        INC(ndx,n);       n := 26;
        blend (ndx,n,cyan,white);
    | firepal:
        FOR ndx:= 0 TO egarange-1 DO
            n := ndx*3;
            setDAC(ndx,fixedfirepal[n],fixedfirepal[n+1],fixedfirepal[n+2]);
        END;
    | palrouge:
        n := 0;
	    FOR i:=32 TO 64 DO
            r:= BYTE(i) ; g:= 0; b:= 0;
            setDAC(n,r,g,b);setDAC(n+1,r,g,b);INC(n,2);
        END;
    | palvert:
        n := 0;
	    FOR i:=32 TO 64 DO
		    r:= 0; g:= BYTE(i); b:= 0;
            setDAC(n,r,g,b);setDAC(n+1,r,g,b);INC(n,2);
        END;
    | palbleu:
        n := 0;
	    FOR i:=32 TO 64 DO
		    r:= 0; g:= 0; b:= BYTE(i);
            setDAC(n,r,g,b);setDAC(n+1,r,g,b);INC(n,2);
        END;
    | palgris:
        n := 0;
	    FOR i:=32 TO 64 DO
		    r:= BYTE(i); g:= BYTE(i); b:= BYTE(i);
            setDAC(n,r,g,b);setDAC(n+1,r,g,b);INC(n,2);
        END;
    | palorg:
        n:=0;
        FOR i :=0 TO 64 DO
            r := BYTE(i); g := BYTE(i); b := BYTE(i DIV 4);
            setDAC(n,r,g,b);setDAC(n+1,r,g,b);INC(n);
        END;
    END;
    setDAC(ndxblack,00H,00H,00H);    (* safety 0=black *)
    n := maxinkindex-egarange+1;
    blend (egarange,n,black,black);  (* 64..255=black *)
    (* setDAC(maxinkindex,3FH,3FH,3FH); *)
END newPalette64;

PROCEDURE showPalette (maxWaitInSeconds:LONGINT );
VAR
    i    : CARDINAL;
    xpos : CARDINAL;
    ypos : CARDINAL;
    p    : CARDINAL;
BEGIN
    clearvideoscreen(ndxblack); (* index of sky, so black *)
    FOR i := 0 TO 255 DO
        FOR ypos := ymin TO ymax DO
            p := Ybase[ypos] + i; (* was ypos * xcount + i *)
            videoscreen[p]:=BYTE(i);
        END;
    END;
    input (maxWaitInSeconds);
END showPalette;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    DOTRADIUS = 1;
    xminlim   = xmin+DOTRADIUS;
    xmaxlim   = xmax-DOTRADIUS;
    yminlim   = ymin+DOTRADIUS;
    ymaxlim   = ymax-DOTRADIUS;

PROCEDURE dotplot(x,y:INTEGER;smalldots:BOOLEAN );
CONST
    ink = maxcolorindex;
BEGIN
    IF x < xminlim THEN
        x:= xminlim;
    ELSIF x > xmaxlim THEN
        x:= xmaxlim;
    END;
    IF y < yminlim THEN
        y:= yminlim;
    ELSIF y > ymaxlim THEN
        y:= ymaxlim;
    END;

    IF smalldots THEN
    wplot(x,           y,           ink);
    ELSE
    wplot(x-DOTRADIUS, y-DOTRADIUS, ink);
    wplot(x          , y-DOTRADIUS, ink);
    wplot(x+DOTRADIUS, y-DOTRADIUS, ink);
    wplot(x-DOTRADIUS, y,           ink);
    wplot(x,           y,           ink);
    wplot(x+DOTRADIUS, y,           ink);
    wplot(x-DOTRADIUS, y+DOTRADIUS, ink);
    wplot(x,           y+DOTRADIUS, ink);
    wplot(x+DOTRADIUS, y+DOTRADIUS, ink);
    END;
END dotplot;

PROCEDURE zline(x1, y1, x2, y2 : INTEGER; ink:BYTE);
VAR
    i, deltax, deltay, numpixels,
    d, dinc1, dinc2,
    x, xinc1, xinc2,
    y, yinc1, yinc2 : INTEGER;
BEGIN
    (* Calculate deltax and deltay for initialisation *)
    deltax := ABS( x2 - x1 );
    deltay := ABS( y2 - y1 );

    (* Initialize all vars based on which is the independent variable *)
    IF deltax >= deltay THEN
        (* x is independent variable *)
        numpixels := deltax + 1;
        d := (2 * deltay) - deltax;
        dinc1 := deltay << 1;
        dinc2 := (deltay - deltax) << 1;
        xinc1 := 1;
        xinc2 := 1;
        yinc1 := 0;
        yinc2 := 1;
    ELSE
        (* y is independent variable *)
        numpixels := deltay + 1;
        d := (2 * deltax) - deltay;
        dinc1 := deltax << 1;
        dinc2 := (deltax - deltay) << 1;
        xinc1 := 0;
        xinc2 := 1;
        yinc1 := 1;
        yinc2 := 1;
    END;

    (* Make sure x and y move in the right directions *)
    IF x1 > x2 THEN
        xinc1 := - xinc1;
        xinc2 := - xinc2;
    END;
    IF y1 > y2 THEN
        yinc1 := - yinc1;
        yinc2 := - yinc2;
    END;

    (* Start drawing at <x1, y1> *)
    x := CARDINAL(x1);
    y := CARDINAL(y1);

    (* Draw the pixels *)
    FOR i := 1 TO numpixels DO
        wplot(x,y,ink);
        IF d < 0 THEN
            d := d + dinc1;
            x := x + xinc1;
            y := y + yinc1;
        ELSE
            d := d + dinc2;
            x := x + xinc2;
            y := y + yinc2;
        END;
    END;
END zline;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE realtoint (v:LONGREAL   ):INTEGER ;
BEGIN
    IF v < 0.0 THEN
        v:=v-0.5;
    ELSE
        v:=v+0.5;
    END;
    RETURN INTEGER(v);
END realtoint;

CONST
    NOLEADER = MAX(CARDINAL);
CONST
    QUADRANTCOUNTX = 4;
    QUADRANTCOUNTY = 4;
    QUADRANTCOUNT  = QUADRANTCOUNTX * QUADRANTCOUNTY;
CONST
    MINRETRACECOUNT = 0;   MAXRETRACECOUNT  = 20;   defaultRETRACECOUNT = 0;
    MINCELLCOUNT    = 1;   MAXCELLCOUNT     = 1400; defaultNUM_CELLS    = 200;
CONST
    MINSPEED        = 0.1; MAXSPEED         = 10.0;
CONST
    COLLISION_RADIUS = 2.5;
TYPE
    celltype = RECORD
        x         : LONGREAL;
        y         : LONGREAL;
        xpos      : INTEGER;
        ypos      : INTEGER;
        direction : LONGREAL;
        speed     : LONGREAL;
        sectorx   : INTEGER;
        sectory   : INTEGER;
        isleader  : BOOLEAN;
    END;
VAR
    cell  : ARRAY[MINCELLCOUNT..MAXCELLCOUNT] OF celltype;

(* ------------------------------------------------------------ *)

PROCEDURE update_screen (countcells,retracecount:CARDINAL;smalldots,DEBUG:BOOLEAN);
VAR
    i,x,y,px,p,sum,ink:CARDINAL;
    cx,cy,ndx,j:CARDINAL;
BEGIN
    clearworkscreen(ndxblack);

    FOR i:=MINCELLCOUNT TO countcells DO
        dotplot(cell[i].xpos, cell[i].ypos,smalldots);
    END;

    FOR y := yminlim TO ymaxlim DO
        p := Ybase[y];
        FOR x := xminlim TO xmaxlim DO
            px := p+x;
            sum:=0;
            INC (sum, CARDINAL(WorkScreen^[px-1-xcount]) ); (* x-1 ,y-1 *)
            INC (sum, CARDINAL(WorkScreen^[px  -xcount]) ); (* x   ,y-1 *)
            INC (sum, CARDINAL(WorkScreen^[px+1-xcount]) ); (* x+1 ,y-1 *)

            INC (sum, CARDINAL(WorkScreen^[px-1]) );        (* x-1 ,y   *)
            INC (sum, CARDINAL(WorkScreen^[px  ]) << 1 );   (* x   ,y   *)
            INC (sum, CARDINAL(WorkScreen^[px+1]) );        (* x+1 ,y   *)

            INC (sum, CARDINAL(WorkScreen^[px-1+xcount]) ); (* x-1 ,y+1 *)
            INC (sum, CARDINAL(WorkScreen^[px  +xcount]) ); (* x   ,y+1 *)
            INC (sum, CARDINAL(WorkScreen^[px+1+xcount]) ); (* x+1 ,y+1 *)

            ink:=sum DIV (10);

            SmoothScreen^[px] :=BYTE(ink);
        END;
    END;
    FOR i := 1 TO retracecount DO waitVGAretrace; END;
    smooth2video;
END update_screen;

(* ------------------------------------------------------------ *)

PROCEDURE init_cells (countcells:CARDINAL );
VAR
    i,ndx,n:CARDINAL;
    sectorx,sectory:INTEGER;
BEGIN
    n:=0;
    FOR i:=MINCELLCOUNT TO countcells DO
        cell[i].x          := LONGREAL( GetRndCardRange(xmin,xmax));
        cell[i].y          := LONGREAL( GetRndCardRange(ymin,ymax));
        cell[i].xpos       := realtoint(cell[i].x);
        cell[i].ypos       := realtoint(cell[i].y);
        cell[i].direction  := GetRndLngRealRange(0.0,PI+PI);
        cell[i].speed      := GetRndLngRealRange(0.0,COLLISION_RADIUS); (* /2 *)
        IF getrndrange(1,4)=1 THEN
            cell[i].isleader:=TRUE;
            INC(n);
        ELSE
            cell[i].isleader:=FALSE;
        END;
    END;
    IF n = 0 THEN
        ndx:=getrndrange(MINCELLCOUNT,countcells);
        cell[ndx].isleader:=TRUE;
    END;
END init_cells;

CONST
     NOISESPEED     = LONGREAL(0.5);
     NOISEDIRECTION = LONGREAL(0.5);

PROCEDURE movecell (i:CARDINAL);
VAR
    sin,cos,nx,ny,speed:LONGREAL;
    newx,newy:INTEGER;
    recalcspeed:BOOLEAN;
BEGIN
    sin := MATHLIB.Sin( cell[i].direction);
    cos := MATHLIB.Cos( cell[i].direction);
    nx  := cell[i].x + sin * cell[i].speed;
    ny  := cell[i].y + cos * cell[i].speed;
    newx:= realtoint(nx);
    newy:= realtoint(ny);
    recalcspeed:=FALSE;
    IF    newx < xminlim THEN
        nx:=LONGREAL(xminlim);  sin:=-sin;  recalcspeed:=TRUE;
    ELSIF newx > xmaxlim THEN
        nx:=LONGREAL(xmaxlim);  sin:=-sin;  recalcspeed:=TRUE;
    END;
    IF    newy < yminlim THEN
        ny:=LONGREAL(yminlim);  cos:=-cos;  recalcspeed:=TRUE;
    ELSIF newy > ymaxlim THEN
        ny:=LONGREAL(ymaxlim);  cos:=-cos;  recalcspeed:=TRUE;
    END;

    IF recalcspeed THEN
        speed:=cell[i].speed+GetRndLngRealRange(-NOISESPEED,NOISESPEED);
        IF    speed < MINSPEED THEN
            speed := MINSPEED;
        ELSIF speed > MAXSPEED THEN
            speed := MAXSPEED;
        END;
        cell[i].speed := speed;
    END;

    cell[i].direction := MATHLIB.ATan2(cos,sin); (* y,x *)
    cell[i].direction := cell[i].direction + GetRndLngRealRange(-NOISEDIRECTION,NOISEDIRECTION);

    cell[i].x       := nx;
    cell[i].y       := ny;

    cell[i].xpos    := realtoint(nx);
    cell[i].ypos    := realtoint(ny);

END movecell;

(* ------------------------------------------------------------ *)

PROCEDURE update_cells (countcells:CARDINAL;altmotion:BOOLEAN);
CONST
    EPSILON        = 0.01;
    ONEFLOAT       = 1.01;
VAR
    i,zeleader: CARDINAL;
    dx,dy,delta,dist,direction,speed:LONGREAL;
BEGIN
    IF NOT( altmotion ) THEN
        FOR i:=MINCELLCOUNT TO countcells DO
            movecell(i);
        END;
    END;
END update_cells;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE getValue (VAR v: CARDINAL;
                mini,maxi:CARDINAL;S:ARRAY OF CHAR):BOOLEAN;
VAR
    lc : LONGCARD;
BEGIN
    IF GetLongCard(S,lc)=FALSE THEN RETURN FALSE; END;
    IF lc < LONGCARD(mini) THEN RETURN FALSE; END;
    IF lc > LONGCARD(maxi) THEN RETURN FALSE; END;
    v := CARDINAL(lc);
    RETURN TRUE;
END getValue;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    maxWaitInSeconds = 10;
VAR
    parmcount,i,opt:CARDINAL;
    S,R:str128;
VAR
    DEBUG,singlestep,chk:BOOLEAN;
    keycode:CARDINAL;
    showpal,stopmouse,smalldots,altmotion : BOOLEAN;
    palette:palettetype;
    NUM_CELLS,retracecount,v:CARDINAL;
BEGIN
    Lib.DisableBreakCheck();

    WrLn;

    DEBUG        := FALSE;
    showpal      := FALSE;
    stopmouse    := FALSE;
    smalldots    := FALSE;
    altmotion    := FALSE;
    palette      := firepal;
    retracecount := defaultRETRACECOUNT;
    NUM_CELLS   := defaultNUM_CELLS;

    parmcount := Lib.ParamCount();

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S); cleantabs(R);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "A"+delim+"SHOWPAL"+delim+
                                   "Z"+delim+"MOUSE"+delim+
                                   "P:"+delim+"PALETTE:"+delim+
                                   "F:"+delim+"CELLS:"+delim+
                                   "D:"+delim+"DELAY:"+delim+
                                   "S"+delim+"SMALL"+delim+
                                   "m"+delim+"motion"+delim+ (* //FIXME *)
                                   "DEBUG"
                               );
            CASE opt OF
            | 1,2,3 :   abort(errHelp,"");
            | 4,5 :     showpal     := TRUE;
            | 6,7:      stopmouse   := TRUE;
            | 8,9:      IF getValue(v,ORD(firepal),ORD(palgris),R)=FALSE THEN
                            abort(errRange,"palette");
                        END;
                        palette     := palettetype(v);
            | 10,11:    IF getValue(NUM_CELLS,MINCELLCOUNT,MAXCELLCOUNT,R)=FALSE THEN
                            abort(errRange,"number of cells");
                        END;
            | 12,13:    IF getValue(retracecount,MINRETRACECOUNT,MAXRETRACECOUNT,R)=FALSE THEN
                            abort(errRange,"delay");
                        END;
            | 14,15:    smalldots := TRUE;
            | 16,17:    (* altmotion := TRUE; *) ;
            | 18:       DEBUG     := TRUE;
            ELSE
                abort(errOpt,S);
            END;
        ELSE
            abort(errParm,S);
        END;
    END;

    initRAND;
    initTrigo;

    initYbase;
    NEW(WorkScreen);
    NEW(SmoothScreen);
    NEW(NextWorkScreen);

    setVideoMode(hiresmode);
    newPalette64(palette);
    clearvideoscreen(ndxblack);

    IF showpal THEN
        showPalette(maxWaitInSeconds);
        clearvideoscreen(ndxblack);
    END;

    IF stopmouse THEN
        IF MsMouse.Reset()=MAX(INTEGER) THEN
            stopmouse := FALSE;
        END;
    END;

    flushKeyboard;

    clearworkscreen(ndxblack);
    clearnextworkscreen(ndxblack);
    clearsmoothscreen(ndxblack);

    singlestep := FALSE;

    init_cells(NUM_CELLS);

    LOOP
        update_cells(NUM_CELLS,altmotion);
        update_screen(NUM_CELLS,retracecount,smalldots,DEBUG);
        (* read and process key *)
        IF singlestep THEN
            WHILE getKeyboardCode(keycode)=FALSE DO
            END;
            chk:=(keycode # keySpace);
            singlestep:=NOT(chk);
        ELSE
            chk:=getKeyboardCode(keycode);
        END;
        IF chk THEN
            CASE keycode OF
            | lowerR        : newPalette64(firepal);
            | upperR        : newPalette64(rpal);
            | upperG,lowerG : newPalette64(gpal);
            | upperB,lowerB : newPalette64(bpal);
            | key1          : newPalette64(palrouge);
            | key2          : newPalette64(palvert);
            | key3          : newPalette64(palbleu);
            | key0          : newPalette64(palgris);
            | keyBang       : newPalette64(palorg);
            | upperS,lowerS : smalldots := NOT(smalldots);
            (* | upperM,lowerM : altmotion := NOT(altmotion); *)
            | keyHome       : IF retracecount > MINRETRACECOUNT THEN DEC(retracecount);END;
            | keyEnd        : IF retracecount < MAXRETRACECOUNT THEN INC(retracecount);END;
            | keySpace      : singlestep:=NOT (singlestep);
            | keyEscape     : EXIT;
            | keyCR         : EXIT;
            END;
        END;
        IF stopmouse THEN
            IF mouseclick() THEN EXIT; END;
        END;
    END;

    setVideoMode(textmode);

    abort(errNone,"");
END Cells.

