(* ---------------------------------------------------------------
Title         Q&D Check Date and Time Stamp
Overview      see help
Usage         see help
Notes
Bugs
Wish List

--------------------------------------------------------------- *)

MODULE chkStamp;

IMPORT Lib;
IMPORT Str;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    placeholder1    = CHR(1); (* was "$" *)
    placeholder2    = CHR(2); (* was "^" *)
    opolder         = 0;  msgOlder        = placeholder1+" is older than "+placeholder2;
    opolderorsame   = 1;  msgOlderOrSame  = placeholder1+" is older than or same stamp as "+placeholder2;
    opnewer         = 2;  msgNewer        = placeholder1+" is newer than "+placeholder2;
    opnewerorsame   = 3;  msgNewerOrSame  = placeholder1+" is newer than or same stamp as "+placeholder2;
    opsame          = 4;  msgSame         = placeholder1+" is same as "+placeholder2;

CONST
    errHelp         = 1;
    errOption       = 2;
    errExpecting    = 3;
    errParmOverflow = 4;
    errJoker        = 5;
    errDir          = 6;
    errCondition    = 7;
    errNotFound     = 8;
    errEitherOr     = 9;
    errFalse        = 128; srcfalse = "128";
    errTrue         = 255; srctrue  = "255";
CONST
    exe         = "CHKSTAMP";
    title       = "Q&D Check Date and Time stamp";
    version     = "1.0b";
    author      = "by PhG";

PROCEDURE abort (rc:CARDINAL;einfo:ARRAY OF CHAR);
CONST
    msgHelp     =
title+" "+version+" "+author+nl+
nl+
"Syntax : "+exe+" <spec1> <condition> <spec2> [option]..."+nl+
nl+
"This program checks <spec1> stamp against <spec2> stamp."+nl+
nl+
"    -l show filename if condition is matched"+nl+
"    -v verbose (show all data)"+nl+
"    -q quiet (show nothing)"+nl+
"    -x disable LFN support even if available"+nl+
nl+
"a) <condition> is any of the following forms :"+nl+
nl+
"       - OLDER"+nl+
"       - OLDERORSAME"+nl+
"       - NEWER"+nl+
"       - NEWERORSAME"+nl+
"       - SAME"+nl+
nl+
"b) Program returns "+srctrue+" if <condition> is true, "+srcfalse+" if <condition> is false."+nl+
"   Note all other codes are below "+srcfalse+".";

VAR
    S:str1024;
BEGIN
    CASE rc OF
    | errHelp:         WrStr(msgHelp);WrLn;
                       S := "";
    | errOption:       S := 'Unknown "|" option !';
    | errExpecting:    S := "Expecting | !";
    | errParmOverflow: S := '"|" is one parameter too many !';
    | errJoker:        S := '"|" cannot contain any joker !';
    | errDir:          S := '"|" must be a file!';
    | errCondition:    S := 'Illegal "|" <condition> !';
    | errNotFound:     S := '"|" does not exist !';
    | errEitherOr:     S := "| options are mutually exclusive !";
    | errFalse:        S := "";
    | errTrue:         S := "";
    ELSE
                       S := "This is illogical, Captain !";
    END;
    IF same(S,"")=FALSE THEN
        Str.Subst(S,"|",einfo);
        WrStr(exe+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode( SHORTCARD(rc) );
    HALT;
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE using (n : CARDINAL; digits : CARDINAL; pad : CHAR) : str80;
VAR
    ok   : BOOLEAN;
    v    : LONGCARD;
    len  : CARDINAL;
    S    : str80;
BEGIN
    v := LONGCARD(n);
    Str.CardToStr(v,S,10,ok);
    len := Str.Length(S);
    LOOP
        IF Str.Length(S) >= digits THEN EXIT; END;
        Str.Prepend(S,pad);
    END;
    RETURN S;
END using;

PROCEDURE booltostr ( tf:BOOLEAN ):str16;
VAR
    R:str16;
BEGIN
    IF tf THEN
        R:="TRUE"
    ELSE
        R:="FALSE";
    END;
    RETURN R;
END booltostr;

PROCEDURE enquote (VAR S:ARRAY OF CHAR );
BEGIN
    Str.Prepend(S,dquote);
    Str.Append(S,dquote);
END enquote;

(* ------------------------------------------------------------ *)

(*
Year stored relative to 1980 (ex. 1988 stores as 8)
    year      month    day   

 F E D C B A 9 8 7 6 5 4 3 2 1 0   <-- Bit Number
*)

CONST
    yyMask=BITSET{9..15};
    yyShft=9;
    mmMask=BITSET{5..8};
    mmShft=5;
    ddMask=BITSET{0..4};
    ddShft=0;
    mindd=1;
    maxdd=31;
    minmm=1;
    maxmm=12;
    minyy=1980; (* base year for messdos *)
    maxyy=minyy+127; (* was 2099 *)
    baseyear=minyy;  (* 1980 *)

PROCEDURE UnpackDMY (dmy:CARDINAL;VAR d,m,y:CARDINAL);
BEGIN
    y := CARDINAL(BITSET(dmy) * yyMask) >> yyShft;
    m := CARDINAL(BITSET(dmy) * mmMask) >> mmShft;
    d := CARDINAL(BITSET(dmy) * ddMask) >> ddShft;
    INC(y,baseyear);
END UnpackDMY;

CONST
    dateplain = 0; (* dd-MMM-yyyy *)
    datestd   = 1; (* dd-mm-yyyy *)
    daterev   = 2; (* yyyy-mm-dd *)

PROCEDURE fmtDate (dmy:CARDINAL;datefmt:CARDINAL) : str16;
CONST
    tokday    = "~";
    tokmonth  = "@";
    tokmois   = "$";
    tokyear   = "!";
    separator = dash;
    pad       = "0";
    baseyear  = 1900;
    tmonths   = "Jan Fv Mar Avr Mai Jun Jui Ao Sep Oct Nov Dc ???";
    tmonths2  = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ???";
VAR
    R,S: str16;
    d,m,y:CARDINAL;
BEGIN
    UnpackDMY(dmy,d,m,y);
    IF ((m < minmm) OR (m > maxmm)) THEN m := 13; END;
    S:="";
    CASE datefmt OF
    | dateplain: R:=tokday +separator+tokmois +separator+tokyear;
                 Str.ItemS(S,tmonths2," ",m-1);
    | datestd:   R:=tokday +separator+tokmonth+separator+tokyear;
    | daterev:   R:=tokyear+separator+tokmonth+separator+tokday;
    ELSE
                 R:="??"+separator+"???"+separator+"????";
    END;
    Str.Subst(R,tokday   ,   using(d,2,pad));
    Str.Subst(R,tokmonth ,   using(m,2,pad));
    Str.Subst(R,tokyear  ,   using(y,4,pad));
    Str.Subst(R,tokmois  ,   S);

    RETURN R;
END fmtDate;

(*
Seconds are 0 to 29 -- DOS stores nearest even / 2
  hours    minutes   seconds 

 F E D C B A 9 8 7 6 5 4 3 2 1 0   <-- Bit Number
*)

CONST
    timelong     = 0;
    timestd      = 1;
    timelongzero = 2;
    timestdzero  = 3;

PROCEDURE fmtTime (timedata:CARDINAL;timefmt:CARDINAL) : str16;
CONST
    hhMask=BITSET{11..15};
    hhShft=11;
    mmMask=BITSET{5..10};
    mmShft=5;
    ssMask=BITSET{0..4};
    ssShft=0;
CONST
    tokhh     = "~";
    tokmm     = "@";
    tokss     = "$";
    tokHZ     = "!";
    separator = colon;
    padhours  = blank;
    pad       = "0";
VAR
    h,m,s : CARDINAL;
    R : str16;
BEGIN
    h := CARDINAL(BITSET(timedata) * hhMask) >> hhShft;
    m := CARDINAL(BITSET(timedata) * mmMask) >> mmShft;
    s := CARDINAL(BITSET(timedata) * ssMask) >> ssShft;
    s := s << 1; (* FIXED ! yes, yes, "* 2" works too... *)

    CASE timefmt OF
    | timelong    : R:=tokhh+separator+tokmm+separator+tokss;
    | timestd     : R:=tokhh+separator+tokmm;
    | timelongzero: R:=tokHZ+separator+tokmm+separator+tokss;
    | timestdzero : R:=tokHZ+separator+tokmm;
    ELSE
                 R:="??"+separator+"??"+separator+"??";
    END;

    Str.Subst(R, tokhh, using(h,2,padhours) );
    Str.Subst(R, tokmm, using(m,2,pad));
    Str.Subst(R, tokss, using(s,2,pad));
    Str.Subst(R, tokHZ, using(h,2,pad) );

    RETURN R;
END fmtTime;

(* ------------------------------------------------------------ *)

PROCEDURE fmtjma (jma:CARDINAL):str16;
BEGIN
    RETURN fmtDate(jma,dateplain);
END fmtjma;

PROCEDURE fmthms (hms:CARDINAL):str16;
BEGIN
    RETURN fmtTime(hms,timelongzero);
END fmthms;

PROCEDURE fmtdt (stamp:LONGCARD):str80;
VAR
    jma,hms:CARDINAL;
    R:str80;
BEGIN
    R:="jj-mm-aaaa hh:mm:ss";
    jma:=CARDINAL ( stamp >> 16 );
    hms:=CARDINAL ( stamp AND 0000FFFFH );
    Str.Concat(R,fmtjma(jma)," -- ");
    Str.Append(R,fmthms(hms));
    RETURN R;
END fmtdt;

(* ------------------------------------------------------------ *)

(* from DD *)

PROCEDURE nicefmtlc (v : LONGCARD; pad:CHAR; sep:CHAR; field:INTEGER) : str80;
VAR
    S,R   : str80;
    len,i : CARDINAL;
    ok  : BOOLEAN;
    ch  : CHAR;
BEGIN
    Str.CardToStr(v,S,10,ok);
    len:=Str.Length(S);
    R := "";
    FOR i := 1 TO len DO
        Str.Prepend(R,S[len-i]);
        IF i < len THEN
            IF (i MOD 3) = 0 THEN
                Str.Prepend(R,sep);
            END;
        END;
    END;
    LOOP
        IF INTEGER(Str.Length(R)) >= ABS(field) THEN EXIT; END;
        IF field < 0 THEN
            Str.Append(R,pad);  (* right alignment *)
        ELSE
            Str.Prepend(R,pad);
        END;
    END;
    RETURN R;
END nicefmtlc;

PROCEDURE fmtlc (v:LONGCARD;base:CARDINAL;wi:INTEGER;ch,prefix:CHAR) : str80;
VAR
    R : str80;
    ok: BOOLEAN;
    i : CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    FOR i:= Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,ch);
        ELSE
            Str.Prepend(R,ch);
        END;
    END;
    IF base=16 THEN Str.Lows(R);END;
    Str.Prepend(R,prefix);
    RETURN R;
END fmtlc;

(* ------------------------------------------------------------ *)

PROCEDURE eval (useLFN,verbose,quiet,showmatch:BOOLEAN;
               opcode:CARDINAL;spec1,spec2:pathtype):CARDINAL;
CONST
    base = 10;
    wilc = 10;
VAR
    isdir1,isdir2,tf:BOOLEAN;
    stamp1,stamp2,fsize1,fsize2:LONGCARD;
    msg:str1024; (* safety *)
    rc:CARDINAL;
    sval,sbool:str16;
BEGIN
    IF showmatch THEN
        verbose:=FALSE; (* already *)
        quiet  :=TRUE;
    END;
    (* isdir1 and isdir2 will never be true here *)
    isdir1:=fileIsDirectorySpec(useLFN,spec1);
    IF isdir1 THEN unfixDirectory(spec1);END;
    isdir2:=fileIsDirectorySpec(useLFN,spec2);
    IF isdir2 THEN unfixDirectory(spec2);END;

    stamp1:=fileGetFileStamp(useLFN,spec1);
    stamp2:=fileGetFileStamp(useLFN,spec2);

    fsize1:=fileGetFileSize(useLFN,spec1);
    fsize2:=fileGetFileSize(useLFN,spec2);

    CASE opcode OF
    | opolder:        tf:=(stamp1 <  stamp2); msg:=msgOlder;
    | opolderorsame:  tf:=(stamp1 <= stamp2); msg:=msgOlderOrSame;
    | opnewer:        tf:=(stamp1 >  stamp2); msg:=msgNewer;
    | opnewerorsame:  tf:=(stamp1 >= stamp2); msg:=msgNewerOrSame;
    | opsame:         tf:=(stamp1 =  stamp2); msg:=msgSame;
    END;
    IF tf THEN
        rc:=errTrue;
    ELSE
        rc:=errFalse;
    END;

    IF isdir1 THEN fixDirectory(spec1);END;
    IF isdir2 THEN fixDirectory(spec2);END;

    IF useLFN THEN
        enquote(spec1);
        enquote(spec2);
    END;

    Str.Subst(msg,placeholder1,spec1);
    Str.Subst(msg,placeholder2,spec2);

    IF verbose THEN
        WrStr("spec1       : ");WrStr(spec1);WrLn;
        WrStr("spec2       : ");WrStr(spec2);WrLn;

        WrStr("stamp1      : ");WrStr(fmtdt(stamp1));WrLn;
        WrStr("stamp2      : ");WrStr(fmtdt(stamp2));WrLn;

        WrStr("filesize1   : ");WrStr(fmtlc(fsize1,base, wilc,blank,""));WrLn;
        WrStr("filesize2   : ");WrStr(fmtlc(fsize2,base, wilc,blank,""));WrLn;

        WrStr("condition   : ");WrStr(msg);WrLn;
        WrStr("evaluation  : ");WrStr(booltostr(tf));WrLn;
        WrStr("return code : ");WrStr(using(rc,1,""));WrLn;
    ELSE
        IF NOT(quiet) THEN
            msg:="::: Returning | (condition is |)";
            CASE rc OF
            | errFalse: sval:=srcfalse;sbool:="FALSE";
            | errTrue:  sval:=srctrue ;sbool:="TRUE";
            END;
            Str.Subst(msg,"|",sval);
            Str.Subst(msg,"|",sbool);
            WrStr(msg);WrLn;
        END;
    END;

    IF showmatch THEN
        IF rc=errTrue THEN WrStr(spec1);WrLn;END;
    END;

    RETURN rc;
END eval;

(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 3;
VAR
    parm : ARRAY[minparm..maxparm] OF pathtype;
    S,R,spec1,spec2:pathtype;
    quiet,verbose,showmatch,useLFN:BOOLEAN;
    lastparm,opt,i,parmcount:CARDINAL;
    rc,opcode :CARDINAL;
BEGIN
    FIO.IOcheck:=FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone;

    WrLn;

    verbose   := FALSE;
    quiet     := FALSE;
    showmatch := FALSE;
    useLFN    := TRUE;
    lastparm  := minparm-1;

    parmcount := Lib.ParamCount();
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "Q"+delim+"QUIET"+delim+
                                  "L"+delim+"LIST"+delim+"SHOWMATCH"
                              );
            CASE opt OF
            | 1,2,3    : abort(errHelp,"");
            | 4,5      : verbose  := TRUE;
            | 6,7      : useLFN   := FALSE;
            | 8,9      : quiet    := TRUE;
            | 10,11,12 : showmatch:= TRUE;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy( parm[lastparm] , S);
        END;
    END;
    CASE lastparm OF
    | 0 : abort(errHelp,"");
    | 1 : abort(errExpecting,"<condition> <file2>");
    | 2 : abort(errExpecting,"<file2>");
    | 3 : ; (* ok *)
    ELSE
        ; (* trapped supra *)
    END;

    IF (verbose AND quiet) THEN abort(errEitherOr,"-v and -q");END;

    IF showmatch THEN
        IF verbose THEN abort(errEitherOr,"-l and -v");END;
        IF quiet   THEN abort(errEitherOr,"-l and -q");END;
    END;

    spec1 :=parm[1];
    IF chkJoker(spec1) THEN abort(errJoker,spec1);END;
    IF fileExists(useLFN,spec1)=FALSE THEN abort(errNotFound,spec1);END;
    IF fileIsDirectorySpec(useLFN,spec1) THEN abort(errDir,spec1);END;
    spec2 :=parm[3];
    IF chkJoker(spec2) THEN abort(errJoker,spec2); END;
    IF fileExists(useLFN,spec2)=FALSE THEN abort(errNotFound,spec2);END;
    IF fileIsDirectorySpec(useLFN,spec2) THEN abort(errDir,spec2);END;

    Str.Concat(R,dash,parm[2]);
    UpperCase(R);
    i :=GetOptIndex(R, "OLDER"           +delim+"OLDERSTAMP"+delim+
                       "OLDERORSAME"     +delim+"SAMEOROLDER"+delim+
                       "NEWER"           +delim+"NEWERSTAMP"+delim+
                       "NEWERORSAME"     +delim+"SAMEORNEWER"+delim+
                       "SAME"            +delim+"SAMESTAMP"+delim+
                       "OLDERTHAN"       +delim+
                       "OLDERTHANORSAMEAS"+delim+
                       "NEWERTHAN"       +delim+
                       "NEWERTHANORSAMEAS"+delim+
                       "SAMEAS");
    CASE i OF
    | 1,2   : opcode := opolder;
    | 3,4   : opcode := opolderorsame;
    | 5,6   : opcode := opnewer;
    | 7,8   : opcode := opnewerorsame;
    | 9,10  : opcode := opsame;
    | 11    : opcode := opolder;
    | 12    : opcode := opolderorsame;
    | 13    : opcode := opnewer;
    | 14    : opcode := opnewerorsame;
    | 15    : opcode := opsame;
    ELSE
        abort(errCondition,parm[2]);
    END;

    useLFN:=( useLFN AND fileSupportLFN() );

    rc := eval(useLFN,verbose,quiet,showmatch,opcode,spec1,spec2);
    abort(rc,"");
END chkStamp.
