(* ---------------------------------------------------------------
Title         Q&D Chrono
Author        PhG
Overview
Usage
Notes         assume date won't change while we're initializing...
              weird : remove ".bat" if specified as not to abort EMM386
              weird : it won't run w386zero.com !
              we don't use any lib as to keep program as simple as possible
              Pourquoi faire simple quand on peut rinventer la roue ? ;-)
Wish list

Bugs          str128 everywhere... even when oversized :
              le's be as stupid as M$ using 32bits vars where 8bits are enough
              (and yes, Virginia, I know about CPU optimization...
              which is almost irrelevant now)

--------------------------------------------------------------- *)

MODULE Chrono;

IMPORT Lib;
IMPORT Str;

FROM IO IMPORT WrStr,WrLn;

TYPE
    str128 = ARRAY [0..127] OF CHAR;

(* ------------------------------------------------------------ *)

CONST
    exe      = "CHRONO";
    version  = "v1.0c";
CONST
    errNone  = 0;
    errHelp  = 1;

PROCEDURE abort (e : CARDINAL);
BEGIN
    Lib.SetReturnCode( SHORTCARD(e) );
    HALT;
END abort;

CONST
    nullchar = CHR(0);
    dash     = "-";
    colon    = ":";
    blank    = " ";
    dquote   = '"';
    dot      = ".";
    tabul    = CHR(9);
    nl       = CHR(13)+CHR(10);
    extBAT   = dot+"BAT";

(* ------------------------------------------------------------ *)

PROCEDURE getCli (VAR S : ARRAY OF CHAR);
VAR
    i : CARDINAL;
    Z : str128;
    ch: CHAR;
BEGIN
    Z := "";
    i := 0;
    LOOP
        ch := Lib.CommandLine^[i];
        CASE ch OF
        | nullchar : Str.Append(Z,ch);    EXIT;
        | tabul    : Str.Append(Z,blank);
        ELSE
                     Str.Append(Z,ch);
        END;
        INC(i);
        IF i > HIGH(Z) THEN EXIT; END; (* useless safety *)
    END;
    Str.Copy(S,Z);
END getCli;

PROCEDURE splitCli(cli:ARRAY OF CHAR;VAR cmd,cliparms : ARRAY OF CHAR);
VAR
    len   : CARDINAL;
    i     : CARDINAL;
    ch    : CHAR;
BEGIN
    len   := Str.Length(cli);
    i     := 1-1;
    LOOP
        IF i >= len THEN EXIT; END;
        ch := cli[i];
        IF ORD(ch) > ORD(blank) THEN EXIT; END;
        INC(i);
    END;
    LOOP
        IF i >= len THEN EXIT; END;
        ch := cli[i];
        IF ORD(ch) <= ORD(blank) THEN EXIT; END;
        INC(i);
    END;
    Str.Slice(cmd,cli,0,i);
    Str.Delete(cli,0,i);
    Str.Copy(cliparms,cli);
END splitCli;

(* ------------------------------------------------------------ *)

TYPE
    DTtype = RECORD
        year      : CARDINAL;
        month     : CARDINAL;
        day       : CARDINAL;
        hours     : CARDINAL;
        minutes   : CARDINAL;
        seconds   : CARDINAL;
        hundredth : CARDINAL;
        weekday   : Lib.DayType;
    END;

PROCEDURE GetDateTimeNow (VAR dt : DTtype);
BEGIN
    Lib.GetDate(dt.year,dt.month,dt.day,dt.weekday);
    Lib.GetTime(dt.hours,dt.minutes,dt.seconds,dt.hundredth);
END GetDateTimeNow;

(* ------------------------------------------------------------ *)

PROCEDURE card2str (n : CARDINAL; digits : CARDINAL; pad : CHAR) : str128;
VAR
    ok   : BOOLEAN;
    v    : LONGCARD;
    len  : CARDINAL;
    S    : str128;
BEGIN
    v := LONGCARD(n);
    Str.CardToStr(v,S,10,ok);
    len := Str.Length(S);
    LOOP
        IF Str.Length(S) >= digits THEN EXIT; END;
        Str.Prepend(S,pad);
    END;
    RETURN S;
END card2str;

PROCEDURE fmtDate (dt : DTtype; pad : CHAR) : str128;
CONST
    separator = dash;
    tmonths   = "Jan Fv Mar Avr Mai Jun Jui Ao Sep Oct Nov Dc ";
    tmonths2  = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
VAR
    R : str128;
BEGIN
    Str.ItemS(R,tmonths2," ",dt.month-1);
    Str.Prepend(R,separator);
    Str.Prepend(R,card2str(dt.day,2,pad));
    Str.Append(R,separator);
    Str.Append(R,card2str(dt.year,4,pad));
    RETURN R;
END fmtDate;

PROCEDURE fmtTime (dt : DTtype; pad : CHAR) : str128;
CONST
    separator = colon;
VAR
    R : str128;
BEGIN
    R := card2str(dt.hours,2,pad);
    Str.Append(R,separator);
    Str.Append(R,card2str(dt.minutes,2,pad));
    Str.Append(R,separator);
    Str.Append(R,card2str(dt.seconds,2,pad));
    Str.Append(R,".");
    Str.Append(R,card2str(dt.hundredth,2,pad));
    RETURN R;
END fmtTime;

CONST
    (*              "dd-mmm-yyyy at " *)
    msgDateFiller = "               ";

PROCEDURE fmtDateTime ( dt : DTtype) : str128;
VAR
    dmy : str128;
    hms : str128;
    R   : str128;
BEGIN
    dmy:=fmtDate(dt," ");
    hms:=fmtTime(dt,"0");
    R := " at ";
    Str.Prepend(R,dmy);
    Str.Append(R,hms);
    RETURN R;
END fmtDateTime;

(* by lazyness, assume program won't be run for more than one day ! *)

PROCEDURE fmtDelta (start,done : DTtype) : str128;
CONST
    minutesPerHour   = 60;
    secondsPerMinute = 60;
    hundred          = 100;
    midnight         = 24*minutesPerHour*secondsPerMinute*hundred;
VAR
    R        : str128;
    n        : LONGCARD; (* in hundredths of seconds *)
    hmsStart : LONGCARD;
    hmsEnd   : LONGCARD;
    delta    : DTtype;
BEGIN
    n := LONGCARD(start.hours) * minutesPerHour + LONGCARD(start.minutes);
    n := n * secondsPerMinute + LONGCARD(start.seconds);
    n := n * hundred + LONGCARD(start.hundredth);
    hmsStart := n;
    n := LONGCARD(done.hours)   * minutesPerHour + LONGCARD(done.minutes);
    n := n * secondsPerMinute + LONGCARD(done.seconds);
    n := n * hundred + LONGCARD(done.hundredth);
    hmsEnd := n;

    IF start.day # done.day THEN (* assume no more than one day ! *)
        hmsEnd := hmsEnd + midnight;
    END;
    n := hmsEnd-hmsStart; (* always >= 0 *)
    delta.hours    :=CARDINAL (n DIV (minutesPerHour*secondsPerMinute*hundred) );
    delta.minutes  :=CARDINAL((n DIV (               secondsPerMinute*hundred) ) MOD 60 );
    delta.seconds  :=CARDINAL((n DIV (                                hundred) ) MOD 60 );
    delta.hundredth:=CARDINAL (n MOD hundred);
    R := fmtTime(delta,"0");
    RETURN R;
END fmtDelta;

(* ------------------------------------------------------------ *)

(* fix weird problem if .bat specified : EMM386 bombs ! *)

PROCEDURE fixBatch (VAR cmd : ARRAY OF CHAR):BOOLEAN ;
VAR
    S : str128;
    len,p : CARDINAL;
    fixed:BOOLEAN;
BEGIN
    fixed:=FALSE;
    Str.Copy(S,cmd);
    Str.Caps(S);  (* required here ! *)
    p := Str.RCharPos(S,dot);
    IF p # MAX(CARDINAL) THEN
        Str.Delete(S,0,p);
        IF Str.Compare(S,extBAT)=0 THEN
            len:=Str.Length(extBAT);
            Str.Delete(cmd,p,len);
            fixed:=TRUE;
        END;
    END;
    RETURN fixed;
END fixBatch;

PROCEDURE WrQuotedStr (S:ARRAY OF CHAR);
BEGIN
    WrStr(dquote);WrStr(S);WrStr(dquote);
END WrQuotedStr;

(* ------------------------------------------------------------ *)

CONST
    msgHelp = nl+
"Q&D Chrono "+version+" by PhG"+nl+
nl+
"Syntax : "+exe+" <command> [parameter]..."+nl+
nl+
"a) Note only DOS format (8+3) filenames are supported."+nl+
"b) <command> is first run as an external command, then as an internal one."+nl+
"c) Program returns <command> error code."+nl;

CONST
    msgFixed    = '".BAT" removal    : ';
    msgInternal = "Batch or internal : ";
    msgExternal = "External command  : ";
    msgParms    = "Command line      : ";
    msgExitCode = "Exit code         : ";
    msgStarted  = "Start             : ";
    msgEnded    = "End               : ";
    msgElapsed  = "Elapsed           : "+msgDateFiller;
VAR
    S,cli,cmd,cliparms      : str128;
    start,done              : DTtype;
    parmcount,rc            : CARDINAL;
    showhelp,external,fixed : BOOLEAN;
BEGIN
    parmcount:=Lib.ParamCount();
    showhelp:=FALSE;
    CASE parmcount OF
    | 0 : showhelp:=TRUE;
    | 1 : Lib.ParamStr(S,1);
          showhelp:= ( (Str.Compare(S,"-?")=0) OR (Str.Compare(S,"/?")=0) );
    END;
    IF showhelp THEN
        WrStr(msgHelp);
        abort(errHelp);
    END;
    getCli(cli);
    splitCli(cli,cmd,cliparms);

    fixed:=fixBatch(cmd);

    GetDateTimeNow(start);
    rc := Lib.Exec (cmd,cliparms,NIL);
    GetDateTimeNow(done);
    external:=( rc # MAX(CARDINAL) );
    IF NOT(external) THEN
        (* external failed : try internal command or batch *)
        GetDateTimeNow(start);
        rc := Lib.ExecCmd (cli); (* useless to check FOR MAX(CARDINAL) here *)
        GetDateTimeNow(done);
    END;

    WrLn;

    WrStr(msgFixed)   ;
    IF fixed THEN
                          S:="yes";
    ELSE
                          S:="irrelevant";
    END;
                                                    WrStr(S);WrLn;
    IF external THEN
        WrStr(msgExternal);
    ELSE
        WrStr(msgInternal);
    END;
                        WrQuotedStr(cmd);WrLn;
        WrStr(msgParms);WrQuotedStr(cliparms);WrLn;

    WrStr(msgExitCode);   S:=card2str(rc,1,"");     WrStr(S);WrLn;
    WrStr(msgStarted);    S:=fmtDateTime(start);    WrStr(S);WrLn;
    WrStr(msgEnded);      S:=fmtDateTime(done);     WrStr(S);WrLn;
    WrStr(msgElapsed);    S:=fmtDelta(start,done);  WrStr(S);WrLn;
    (* rc:=errNone; *)
    abort(rc);
END Chrono.
