(* ---------------------------------------------------------------
Title         Q&D Fibonacci numbers
Overview
Notes         compact model, stack 65534 in project
              note HUGECARD are LONGREALs because LONGCARD(47) could be max

              with P4, normal 0..40 is 0.06s but recursion is 6mn36s !

Bugs
Wish List

--------------------------------------------------------------- *)

MODULE Fibo;

IMPORT Str;
IMPORT Lib;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    HUGECARD = LONGREAL;

CONST
    ProgEXEname   = "FIBO";
    ProgTitle     = "Q&D Fibonacci function";
    ProgVersion   = "v1.0b";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    defaultPolling     = 16; (* prevent Win9X FROM freezing when calling ChkEscape() "too often" ! how weird... *)
    strDefaultPolling  = "16";
    minPolling         = 1;
    maxPolling         = 32;
    strPollingRange    = "[1..32]";
CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParmOverflow = 3;
    errBadLC        = 4;
    errBadRange     = 5;
    errOutOfRange   = 6;
    errAborted      = 7;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <number>... [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" <first..last> [option]..."+nl+
nl+
"  -r   force recursion algorithm (default is linear algorithm)"+nl+
"  -u   use American digits separator (defaut is to use French digits separator)"+nl+
"  -a   alternate display"+nl+
"  -p:# ESCape polling interval ("+strPollingRange+", default is "+strDefaultPolling+")"+nl+
nl+
"With syntax 1, this program computes Fibonacci function for <number> ;"+nl+
"with syntax 2, it does so for a range of numbers."+nl+
nl+
"a) With LONGCARDs, valid range was [0..47] ; with LONGREALs, it is [0..87]."+nl+
"b) Recursion may look more elegant, but it is very slow."+nl+
"c) -p:# option may be useful, should Win9X freeze while recursing (-r option)."+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :        WrStr(helpmsg);
    | errOption :      S:='Illegal "~" option !';
    | errParmOverflow: S:='"~" is one parameter too many !';
    | errBadLC:        S:='Illegal "~" number !';
    | errBadRange:     S:='Illegal "~" range !';
    | errOutOfRange:   S:='"~" value out of expected range !';
    | errAborted:      S:="Aborted by user !";
    ELSE
                       S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp : ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE padstr (VAR R:ARRAY OF CHAR;wi:INTEGER;padchar:CHAR   );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
END padstr;

PROCEDURE fmthc(v : HUGECARD; pad:CHAR; sep:CHAR; field:INTEGER) : str80;
VAR
    S,R   : str80;
    len,i : CARDINAL;
    ok  : BOOLEAN;
    ch  : CHAR;
BEGIN
    Str.FixRealToStr( LONGREAL(v),0,S,ok);
    len:=Str.Length(S);
    R := "";
    FOR i := 1 TO len DO
        Str.Prepend(R,S[len-i]);
        IF i < len THEN
            IF (i MOD 3) = 0 THEN
                Str.Prepend(R,sep);
            END;
        END;
    END;
    LOOP
        IF INTEGER(Str.Length(R)) >= ABS(field) THEN EXIT; END;
        IF field < 0 THEN
            Str.Append(R,pad);  (* right alignment *)
        ELSE
            Str.Prepend(R,pad);
        END;
    END;
    RETURN R;
END fmthc;

PROCEDURE isValidDecimalNumber (VAR v:HUGECARD; S:ARRAY OF CHAR ):BOOLEAN;
CONST
    decdigits = digits;
VAR
    rc:BOOLEAN;
BEGIN
    rc:=FALSE;
    IF verifyString(S,decdigits) THEN
        v:=Str.StrToReal( S,rc);
    END;
    RETURN rc;
END isValidDecimalNumber;

PROCEDURE isValidRange (VAR mini,maxi:HUGECARD;  S:ARRAY OF CHAR  ):BOOLEAN;
CONST
    seprange = "..";
VAR
    rc:BOOLEAN;
    p:CARDINAL;
    N:str128; (* oversized for safety *)
BEGIN
    rc:=FALSE;
    p:=Str.Pos(S,seprange);
    IF p # MAX(CARDINAL) THEN
        Str.Copy(N,S);
        N[p]:=charnull;
        IF isValidDecimalNumber(mini,N) THEN
            Str.Copy(N,S);
            Str.Delete(N,0,p+2); (* delete ".." too *)
            rc:=isValidDecimalNumber(maxi,N);
        END;
    END;
    RETURN rc;
END isValidRange;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    HUGEZERO = HUGECARD(0);
    HUGEONE  = HUGECARD(1);
    epsilon  = HUGECARD(0.0001); (* should do *)

PROCEDURE isHUGEZERO (v:HUGECARD ):BOOLEAN;
CONST
    seuil = HUGEZERO + epsilon;
BEGIN
    (* RETURN ( v = HUGEZERO ); *)
    RETURN ( ABS(v) < seuil );
END isHUGEZERO;

PROCEDURE isHUGEONE (v:HUGECARD ):BOOLEAN;
CONST
    loseuil = HUGEONE - epsilon;
    hiseuil = HUGEONE + epsilon;
BEGIN
    (* RETURN ( v = HUGEONE ); *)
    IF ( (v < loseuil) OR (v > hiseuil) ) THEN RETURN FALSE;END;
    RETURN TRUE;
END isHUGEONE;

PROCEDURE HUGEINC (VAR r:HUGECARD;n:HUGECARD );
BEGIN
    r := r + n;
END HUGEINC;

PROCEDURE HUGEDEC (VAR r:HUGECARD;n:HUGECARD );
BEGIN
    r := r - n;
END HUGEDEC;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    pollinginterval  : CARDINAL;
    currpollinground : CARDINAL;

PROCEDURE fiboRec (VAR esc:BOOLEAN;v:HUGECARD):HUGECARD;
VAR
    nminus1,nminus2:HUGECARD;
BEGIN
    IF esc THEN RETURN HUGEZERO;END;

    INC(currpollinground);
    currpollinground := currpollinground MOD pollinginterval;
    IF currpollinground = 0 THEN
        IF ChkEscape() THEN esc:=TRUE; RETURN HUGEZERO;END;
    END;
    IF isHUGEZERO(v) THEN RETURN HUGEZERO;END;
    IF isHUGEONE(v)  THEN RETURN HUGEONE; END;
    nminus1 := v - HUGEONE;
    nminus2 := nminus1 - HUGEONE;
    RETURN fiboRec(esc,nminus1) + fiboRec(esc,nminus2);
END fiboRec;

PROCEDURE fiboLinear (v:HUGECARD):HUGECARD;
VAR
    nminus1,nminus2,sum,currval:HUGECARD;
BEGIN
    IF isHUGEZERO(v) THEN RETURN HUGEZERO;END;
    IF isHUGEONE(v)  THEN RETURN HUGEONE; END;
    (* v >= 2 here *)
    nminus2:=HUGEZERO;
    nminus1:=HUGEONE;
    currval:=HUGEONE;
    LOOP
        sum:=nminus1+nminus2;
        HUGEINC(currval,HUGEONE);
        IF currval >= v THEN EXIT; END;
        nminus2 := nminus1;
        nminus1 := sum;
    END;
    RETURN sum;
END fiboLinear;

(* v cannot be < 0 here *)

PROCEDURE funcFibo (VAR ok,esc:BOOLEAN;
                   doRecurse:BOOLEAN;v:HUGECARD): HUGECARD;
VAR
    f:HUGECARD;
BEGIN
    ok:=TRUE;
    esc:=FALSE;
    (*
    IF isHUGEZERO(v) THEN RETURN HUGEZERO;END;
    IF isHUGEONE(v)  THEN RETURN HUGEONE; END;
    *)
    IF doRecurse THEN
        f:=fiboRec(esc,v);
    ELSE
        f:=fiboLinear(v);
    END;
    RETURN f;
END funcFibo;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    padwindx = 4;
    padwi    = 24;
    spat     = "Fibo ( ~ ) ==> ~";
    spatalt  = "~ : ~";
    frsep    = dot;
    ussep    = coma;
    msgWait  = "Computing, please wait... ";
CONST
    minparm = 1;
    maxparm = 32;
VAR
    parm : ARRAY [minparm..maxparm] OF str128;
    pass,lastparm,parmcount,i,opt:CARDINAL;
    S,R:str128;
    mode : (convert,showrange);
    v,currval,mini,maxi:HUGECARD;
    altdisplay,doRecurse:BOOLEAN;
    ok,esc:BOOLEAN;
    sep:CHAR;
    lc:LONGCARD;
BEGIN
    WrLn;

    sep             := frsep;
    doRecurse       := FALSE;
    altdisplay      := FALSE;

    currpollinground:= 0;
    pollinginterval := defaultPolling;

    parmcount := Lib.ParamCount();
    (* IF parmcount=0 THEN abort(errHelp,"");END; *)

    lastparm  := minparm-1;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "R"+delim+"RECURSE"+delim+
                                   "U"+delim+"US"+delim+"UK"+delim+
                                   "P:"+delim+"POLL:"+delim+
                                   "A"+delim+"DISPLAY"
                               );
            CASE opt OF
            | 1,2,3   :    abort(errHelp,"");
            | 4,5     :    doRecurse := TRUE;
            | 6,7,8   :    sep:=ussep;
            | 9,10    :    IF GetLongCard(R,lc)=FALSE THEN abort(errBadLC,S);END;
                           IF lc > MAX(CARDINAL) THEN abort(errOutOfRange,S);END;
                           pollinginterval := CARDINAL(lc);
                           IF pollinginterval > maxPolling THEN abort(errOutOfRange,S);END;
                           IF pollinginterval < minPolling THEN abort(errOutOfRange,S);END;
            | 11,12   :    altdisplay := TRUE;
            ELSE
                           abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy(parm[lastparm],S); (* keep case *)
        END;
    END;

    mode:=convert;
    CASE lastparm OF
    | minparm-1 :
        abort(errHelp,"");
    | minparm :
        S:=parm[minparm];
        IF isValidRange(mini,maxi,S) THEN
           mode:=showrange;
        END;
    ELSE
        ;
    END;
    CASE mode OF
    | convert:
        FOR pass:=1 TO 2 DO
            FOR i:=minparm TO lastparm DO
                S:=parm[i];
                ok:=isValidDecimalNumber (currval, S);
                CASE pass OF
                | 1 :
                     IF NOT(ok) THEN abort(errBadLC,S);END;
                | 2 :
                     IF altdisplay THEN
                         R:=spatalt;
                     ELSE
                         R:=spat;
                     END;
                     Str.Subst( R, "~", fmthc(currval,blank,sep,padwindx) );

                     video(msgWait,TRUE);
                     v:=funcFibo(ok,esc, doRecurse,currval);
                     video(msgWait,FALSE);

                     IF esc THEN abort(errAborted,"");END;
                     Str.Subst( R, "~", fmthc(v,blank,sep,padwi) );
                     WrStr(R);WrLn;
                END;
            END;
        END;
    | showrange:
        S:=parm[minparm];
        IF isValidRange(mini,maxi,S)=FALSE THEN abort(errBadRange,S);END;

        currval:=mini;
        LOOP
            IF altdisplay THEN
                R:=spatalt;
            ELSE
                R:=spat;
            END;
            Str.Subst( R, "~", fmthc(currval,blank,sep,padwindx) );

            video(msgWait,TRUE);
            v:=funcFibo(ok,esc, doRecurse,currval);
            video(msgWait,FALSE);

            IF esc THEN abort(errAborted,"");END;
            Str.Subst( R, "~", fmthc(v,blank,sep,padwi) );
            WrStr(R);WrLn;

            IF maxi < mini THEN
                IF isHUGEZERO(currval) THEN EXIT; END;
                HUGEDEC(currval,HUGEONE);
                IF currval < maxi THEN EXIT; END;
            ELSE
                HUGEINC(currval,HUGEONE);
                IF currval > maxi THEN EXIT; END;
            END;
        END;
    END;

    abort(errNone,"");
END Fibo.

