(* ---------------------------------------------------------------
Title         Q&D Reveal "hidden" words
Overview
Usage         see help
Notes         no checks for processor, ram etc.
              LastEntry const is currently 5000 (8000 max, 3000 okay)
              yes, syntax could have been more clever
              but who cares ? and it's NOT essential software !
Bugs          vindoze does not like "frequent" calls to ChkEscape() !
Wish List     XMS support using pragma for stacksize at least 256 bytes
              (M2 default stack size is 16 Kb anyway!). But maybe XMS
              is overkill for such a thing... ;-)
              escape key to abort, audio beep when one name found ?
--------------------------------------------------------------- *)

MODULE findHIDN;

IMPORT IO;
IMPORT Lib;
IMPORT Str;
IMPORT FIO;
IMPORT LIM;
IMPORT SYSTEM;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM IO IMPORT WrStr,WrLn;

(* ------------------------------------------------------------ *)

CONST
    progPREPROC   = "PREPHIDN";
    progREVEAL    = "FINDHIDN";
    progSHOW      = "SHOWHIDN";
CONST
    ProgEXEname   = progREVEAL;
    ProgTitle     = 'Q&D Search "hidden" words';
    ProgVersion   = "v1.0e";
    ProgCopyright = "by PhG";

CONST
    errNone             = 0;
    errHelp             = 1;
    errUnknown          = 2;
    errTooMany          = 3;
    errNotFound         = 4;
    errConversion       = 5;
    errNeeded           = 6;
    errMinCount         = 7;
    errMaxCount         = 8;
    errTooLong          = 9;
    errTooShort         = 10;
    errNeedRoom         = 11;
    errNeverWare        = 12;
    errParmOverflow     = 13;
    errAborted          = 255;

PROCEDURE abort (e : CARDINAL;einfo:ARRAY OF CHAR );
CONST
    hlp =ProgTitle+" "+ProgVersion+" "+ProgCopyright+nl+
nl+
"Syntax : "+ProgEXEname+" <textfile> <-w:word> [-l:#] [-u:#] [-r]"+nl+
nl+
'This program *slowly* searches "hidden" words in a DOS ASCII text file,'+nl+
"provided file has *already* been processed with "+progPREPROC+" (un)utility,"+nl+
"another "+progREVEAL+" companion (un)utility with "+progSHOW+"."+nl+
"The three programs have been written in order to confirm"+nl+
'the nonsense of "hidden" words in the Bible or in any long text.'+nl+
nl+
"    -w:$ word to find"+nl+
"    -l:# lower count of characters to skip (default is 1)"+nl+   (* see infra *)
"    -u:# upper count of characters to skip (default is 50)"+nl+
"    -lim use EMS/LIM memory if present"+nl+
"    -a   audio warning for each entry found"+nl+
"    -v   verbose eyecandy"+nl+
"    -z   show position as 0-based (default is 1-based)"+nl+
"    -sa  do not sort filenames alphabetically"+nl+
nl+
"a) Please remember <textfile> *must* have been preprocessed with "+progPREPROC+" !"+nl+
"b) Please remember this program is *slow*."+nl+
"c) Note position of any match found in text file is 1-based"+nl+
"   unless -z option has been specified."+nl+
"d) Note word is first searched verbatim then reversed."+nl+
"e) EMS/LIM is about five times SLOWER than conventional memory"+nl+
"   (at least with a true DOS). It SHOULD be safer anyway."+nl+
"   Besides, this option gives a sorted output."+nl+
"f) Using conventional memory, output will have to be sorted later."+nl+
"   Moreover, buffer overflow is likely to happen for big intervals."+nl+
"   Use the divide and conquer approach by searching small intervals."+nl+
nl+
"Examples : "+ProgEXEname+" *.t -w:PHIL -u:666"+nl+
"           "+ProgEXEname+" test -w:HELL"+nl;

VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(hlp);
    | errUnknown :
        S := "Unknown ~ option !";
    | errTooMany :
        S := "Too many files match ~ specification !";
    | errNotFound :
        S := "No file matches ~ specification !";
    | errConversion :
        S := "~ parameter is not a LONGCARD !";
    | errNeeded :
        S := "Word needed !";
    | errMinCount :
        S := "~ lower count to skip cannot be 0 !";
    | errMaxCount :
        S := "~ upper count to skip cannot be less than lower count !";
    | errTooLong :
        S := "~ word length cannot be more than 256 characters !";
    | errTooShort :
        S := "~ word length cannot be more less than 1 character !";
    | errNeedRoom :
        S := "Pending positions array overflow : please retry with a smaller interval !";
    | errNeverWare :
        S := "This MaybeWare is NeverWare : XMS will NOT be implemented !";
    | errParmOverflow:
        S := "~ is one parameter too many !";
    | errAborted :
        S := "Search aborted by user !";
    ELSE
        S := "How did you get THERE ?";
    END;
    CASE e OF
    | errNone,errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

PROCEDURE alert (  );
BEGIN
    Lib.Sound(111);
    Lib.Delay(22);
    Lib.NoSound();
END alert;

PROCEDURE candy (verbose:BOOLEAN;cmd:CARDINAL   );
BEGIN
    IF verbose THEN Work(cmd); END;
END candy;

(* prevent freeze from Win9X hating fast kbd polling ! *)
CONST
    CHKEVERY = 64; (* 16 seems okay but let's play it safe *)
VAR
    globpollval  : CARDINAL; (* fix Win9X weird problem *)

PROCEDURE pollEscape (  ):BOOLEAN ;
VAR
    hit : BOOLEAN;
BEGIN
    hit := FALSE;
    INC(globpollval);
    IF (globpollval MOD CHKEVERY) = 0 THEN
        globpollval := 0;
        hit:=ChkEscape();
    END;
    RETURN hit;
END pollEscape;

(* ------------------------------------------------------------ *)

CONST
    IObufferSize = (8 * 512) + FIO.BufferOverhead; (* 64 i.e. 32Kb was STUPID ! *)
    firstIObufferByte = 1;
    lastIObufferByte  = IObufferSize;
VAR
    IObufferIn : ARRAY [firstIObufferByte..lastIObufferByte] OF BYTE; (* for private M2 IO *)

(* ------------------------------------------------------------ *)

CONST
    sizePage            = 16384; (* for LIM *)
    wordsPerPage        = 8192; (* for lib.wordmove *)
    firstInPage         = 1;
    lastInPage          = sizePage;
    firstpage           = 0; (* 0..pages-1 *)
    firstPhysicalPage   = 0; (* 0..3 *)
VAR
    pages : CARDINAL; (* global for LIM pages *)
    available : LONGCARD; (* global *)
    LIMhandle : CARDINAL;

(* ------------------------------------------------------------ *)

PROCEDURE ChkLIMstatus ( );
VAR
    rc : SHORTCARD;
BEGIN
    rc := LIM.GetStatus();
    IF rc <> 00H THEN
        WrStr(ProgEXEname+": LIM error code ");
        IO.WrShtHex(rc,2);
        WrStr("H");
        Lib.SetReturnCode(SHORTCARD(rc));
        HALT;
    END;
END ChkLIMstatus;

(* ------------------------------------------------------------ *)

PROCEDURE dmpFound (VAR matchesfound:LONGCARD;
                   fpos:LONGCARD;interval:CARDINAL;
                   flagShowZeroBased,warn,flagReverse:BOOLEAN);
BEGIN
    IF flagShowZeroBased THEN DEC(fpos);END;
    IO.WrLngCard(fpos,9);
    WrStr(tabul);
    WrStr("(");
    IO.WrCard(interval,5);
    WrStr(")");
    IF flagReverse THEN WrStr(" R");END;
    WrLn;
    INC(matchesfound);
    IF warn THEN alert;END;
END dmpFound;

(* ------------------------------------------------------------ *)

(* buffers must be outside procedure, else bomb! *)

CONST
    bufferSize = 64 * 512 ; (* is AT LEAST 16 Kb, and here 32 Kb *)
    firstBufferChar = 1;
    lastBufferChar = bufferSize;
VAR
    Buffer : ARRAY [firstBufferChar..lastBufferChar] OF CHAR;

(* berk... should use linked list but... *)

CONST
    firstEntry = 1;
    lastEntry = 5000; (* maximum practical is 8000 here, 3000 seems enough *)
    freeEntry = 0; (* flag as interval 0 is a nonsense *)
TYPE
    entry = RECORD
                addr : LONGCARD;
                interval : CARDINAL;
                matchcount : SHORTCARD;
            END;
VAR
    pending : ARRAY [firstEntry..lastEntry] OF entry;

(* ------------------------------------------------------------ *)

(* yes, yes, we should have returned an error code instead of aborting inproc *)

PROCEDURE doSearchMain (VAR matchesfound:LONGCARD;
                       source : ARRAY OF CHAR;
                       wantedword : str128;
                       mincount,maxcount : CARDINAL;
                       flagReverse,warn,verbose,flagShowZeroBased : BOOLEAN);
VAR
    wordlen : CARDINAL;
    filesize : LONGCARD;
    posinfile : LONGCARD;
    got      : CARDINAL;
    ibuffer : CARDINAL;
    interval : CARDINAL;
    iword : CARDINAL;
    p : LONGCARD;
    i : CARDINAL;
    pbase,here : LONGCARD;
    offset : CARDINAL;
    liminf,limsup : LONGCARD;
    IOhndIn:FIO.File;
BEGIN
    FOR i := firstEntry TO lastEntry DO
        pending[i].interval := freeEntry ; (* used as a flag for free or not *)
    END;
  candy(verbose,cmdInit);
    wordlen := Str.Length(wantedword);
    IOhndIn:=FIO.OpenRead(source);
    FIO.AssignBuffer(IOhndIn,IObufferIn);
    filesize := FIO.Size(IOhndIn);
    posinfile := 0;
    LOOP
        got := FIO.RdBin(IOhndIn,Buffer,bufferSize);
        IF got = 0 THEN EXIT; END; (* filesize was a multiple of buffersize *)
        (* scan pending positions just in case *)
        liminf := posinfile + firstBufferChar; (* position of current start of buffer *)
        limsup := liminf + LONGCARD(got);
        FOR i := firstEntry TO lastEntry DO
  candy(verbose,cmdShow);
            IF pending[i].interval <> freeEntry THEN
                pbase := pending[i].addr;
                interval := pending[i].interval;
                iword := CARDINAL( pending[i].matchcount );
                (* here check and keep if necessary or show then clear *)
                LOOP
                    p := pbase + LONGCARD(interval) * LONGCARD(iword);
                    IF p >= limsup THEN (* past last byte of file *)
                        EXIT;
                    END;
                    IF p >= liminf  THEN (* p < limsup *)
                        offset := CARDINAL(p-liminf)+firstBufferChar;
                        IF wantedword[iword] # Buffer[offset] THEN
                            pending[i].interval:=freeEntry; (* clear ! *)
                            EXIT; (* no match so we can abort search here *)
                        END;
                        INC(iword); (* advance *)
                        IF iword = wordlen THEN
  candy(verbose,cmdStop);
                            here:=pbase;
                            dmpFound (matchesfound,
                                     here,interval,flagShowZeroBased,warn,flagReverse);
  candy(verbose,cmdInit);
                            pending[i].interval:=freeEntry; (* clear ! *)
                            EXIT; (* abort now, for iword is here >= wordlen *)
                        END;
                        INC(pending[i].matchcount);
                    END;
                END;
            END;
        END;

        (* scan normally *)
        FOR ibuffer := firstBufferChar TO got DO
  candy(verbose,cmdShow);
            FOR interval := mincount TO maxcount DO
                iword := 0;
                LOOP
                    (* first iword is always 0 *)
                    p := LONGCARD(ibuffer) + LONGCARD(iword) * LONGCARD(interval);
                    IF p > LONGCARD(got) THEN
                        IF got = lastBufferChar THEN (* past buffer so keep info for later *)
                            i := firstEntry;
                            LOOP
                                IF pending[i].interval = freeEntry THEN
                                    pending[i].addr := posinfile+LONGCARD(ibuffer);
                                    pending[i].interval := interval;
                                    pending[i].matchcount := SHORTCARD(iword);
                                    EXIT;
                                END;
                                INC(i);
                                IF i > lastEntry THEN
  candy(verbose,cmdStop);
                                    abort(errNeedRoom,"");
                                END;
                            END;
                        END;
                        EXIT; (* exit in all case : past buffer or past filesize ! *)
                    END;
                    IF wantedword[iword] # Buffer[CARDINAL(p)] THEN
                        EXIT; (* no match so we can abort search here *)
                    END;
                    INC(iword); (* advance *)
                    IF iword = wordlen THEN (* match for all chars of wantedword *)
  candy(verbose,cmdStop);
                        here:=posinfile+LONGCARD(ibuffer);
                        dmpFound (matchesfound,
                                 here,interval,flagShowZeroBased,warn,flagReverse);
  candy(verbose,cmdInit);
                        EXIT; (* abort now, for iword is here >= wordlen *)
                    END;
                END;
                (* check for escape key *)
                IF pollEscape () = TRUE THEN
                   FIO.Close(IOhndIn);
  candy(verbose,cmdStop);
                   abort(errAborted,"");
                END;
            END;
        END;
        INC (posinfile, LONGCARD (got) );
        IF got < bufferSize THEN EXIT; END;
    END;
    FIO.Close(IOhndIn);
  candy(verbose,cmdStop);
END doSearchMain;

(* ------------------------------------------------------------ *)

CONST
    useMain = 0;
    useLIM  = 1;

(* ------------------------------------------------------------ *)

(* yes, yes, we should have returned an error code instead of aborting inproc *)

PROCEDURE doSearchLIM (VAR matchesfound:LONGCARD;
                       source : ARRAY OF CHAR;
                       wantedword : str128;
                       mincount,maxcount : CARDINAL;
                       flagReverse,warn,verbose,flagShowZeroBased : BOOLEAN);
VAR
    filesize : LONGCARD;
    got : CARDINAL;
    logicalpage : CARDINAL; (* logical page in EMS memory [0..pages-1] *)
    physicalpage: CARDINAL; (* physical page in memory [0..3] at segment frame *)
    currpage : CARDINAL;
    PageSegment:CARDINAL;
    PageOffset:CARDINAL;
    lastpage : CARDINAL;
    bytescount:LONGCARD; (* useless but as a check *)
    pbase,here : LONGCARD;
    wordlen : CARDINAL;
    interval : CARDINAL;
    iword : CARDINAL;
    offset : CARDINAL;
    p : LONGCARD;
    oldpage : CARDINAL;
    ch : CHAR;
    IOhndIn:FIO.File;
BEGIN
    IOhndIn:=FIO.OpenRead(source);
    FIO.AssignBuffer(IOhndIn,IObufferIn);
    filesize := FIO.Size(IOhndIn);
    IF filesize >= available THEN
        FIO.Close(IOhndIn);
        WrStr("WARNING ! File too big for available LIM/EMS expanded memory ! Skipped !");
        WrLn;
        RETURN;
    END;
  candy(verbose,cmdInit);
    PageSegment := LIM.PageFr; (* has been initialized by LIM library *)
    currpage := firstpage;
    bytescount := 0;
    LOOP
  candy(verbose,cmdShow);
        got := FIO.RdBin(IOhndIn,Buffer,sizePage);
        IF got = 0 THEN EXIT; END; (* filesize was a multiple of sizePage *)

        PageOffset:=LIM.MapPage(firstPhysicalPage,currpage,LIMhandle); (* retourne physpage<<14 *)
        ChkLIMstatus();
        Lib.WordMove (ADR(Buffer),[PageSegment:PageOffset],wordsPerPage);

        INC (bytescount,LONGCARD(got));
        IF got < sizePage THEN EXIT; END; (* all was loaded now *)
        INC(currpage);
    END;
    FIO.Close(IOhndIn);
    lastpage := currpage;
    oldpage := MAX(CARDINAL); (* force change *)
    wordlen := Str.Length(wantedword);

    (* now scan LIM memory *)

    FOR pbase := 0 TO (filesize-1) DO
  candy(verbose,cmdShow);
        FOR interval := mincount TO maxcount DO
            iword := 0;
            LOOP
                p := pbase + LONGCARD(iword) * LONGCARD(interval);
                IF p >= filesize THEN EXIT; END;
                currpage := CARDINAL (p DIV sizePage);
                offset := CARDINAL (p MOD sizePage);
                IF currpage # oldpage THEN
                    PageOffset:=LIM.MapPage(firstPhysicalPage,currpage,LIMhandle); (* retourne physpage<<14 *)
                    ChkLIMstatus();
                    oldpage := currpage;
                END;
                (* no way to directly access this memory zone!!! *)
                Lib.Move ([PageSegment:PageOffset+offset],ADR(ch),1);
                IF wantedword[iword] # ch THEN
                    EXIT; (* no match so we can abort search here *)
                END;
                INC(iword); (* advance *)
                IF iword = wordlen THEN (* match for all chars of wantedword *)
  candy(verbose,cmdStop);
                    here:=pbase+1;
                    dmpFound (matchesfound,
                             here,interval,flagShowZeroBased,warn,flagReverse);
  candy(verbose,cmdInit);
                    EXIT; (* abort now, for iword is here >= wordlen *)
                END;
            END;
            (* check for escape key *)
            IF pollEscape () = TRUE THEN
  candy(verbose,cmdStop);
                LIM.DeAllocatePages(LIMhandle); (* else no more EMS _AND_ XMS !!! *)
                ChkLIMstatus();
                abort(errAborted,"");
            END;
        END;
    END;
  candy(verbose,cmdStop);
END doSearchLIM;

(* ------------------------------------------------------------ *)

PROCEDURE doReverse (VAR S : ARRAY OF CHAR );
VAR
    i : CARDINAL;
    p : CARDINAL;
    N : str128;
BEGIN
    N := str128(S);
    p := Str.Length(S) -1;
    FOR i := 0 TO (Str.Length(S)-1) DO
        S[i] := N[p];
        DEC(p);
    END;
END doReverse;

(* ------------------------------------------------------------ *)

TYPE
    f8e3 = ARRAY [0..8+1+3-1] OF CHAR;
CONST
    firstFile = 1;
    lastFile  = 2000;
    maxFile   = lastFile-firstFile+1;
VAR
    fileArray : ARRAY [firstFile..lastFile] OF f8e3;

PROCEDURE buildList (VAR basepath:ARRAY OF CHAR;
                    spec:ARRAY OF CHAR):CARDINAL;
VAR
    countFile : CARDINAL;
    found     : BOOLEAN;
    entry     : FIO.DirEntry;
    u,path,f8,e3    : str128; (* oversized just in case *)
BEGIN
    qdSplitAllPath(spec,u,path,f8,e3);
    Str.Concat(basepath,u,path);

    countFile := 0;
    found := FIO.ReadFirstEntry(spec,allfiles,entry);
    WHILE found DO
        IF countFile = maxFile THEN RETURN MAX(CARDINAL); END;
        fileArray[firstFile+countFile]:=f8e3(entry.Name);
        Str.Caps(fileArray[firstFile+countFile]); (* useless ! *)
        INC (countFile);
        found :=FIO.ReadNextEntry(entry);
    END;
    RETURN countFile;
END buildList;

PROCEDURE isLessByName (i,j:CARDINAL):BOOLEAN;
VAR
    si,sj:f8e3;
BEGIN
    si:=fileArray[i];
    sj:=fileArray[j];
    RETURN ( Str.Compare(si,sj) < 0 );
END isLessByName;

PROCEDURE doSwapNames (i,j:CARDINAL);
VAR
    si,sj,tmp:f8e3;
BEGIN
    si:=fileArray[i];
    sj:=fileArray[j];
    tmp:=si;
    si:=sj;
    sj:=tmp;
    fileArray[i]:=si;
    fileArray[j]:=sj;
END doSwapNames;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minLen = 2; (* trick because 1 char gives all intervals... and I'm too lazy to put limit test at the beginning! :-) *)
    maxLen = 48; (* should be enough! *)
VAR
    mincount : LONGCARD; (* 1.. *)
    maxcount : LONGCARD;
    matchesfound   : LONGCARD;
    DEBUG,flagReverse,flagShowZeroBased,audiowarn,verbose,alphasort : BOOLEAN;
    wantedword,reversedwantedword,wanted : str128;

    flagMemType : CARDINAL;

    parms : CARDINAL;
    i : CARDINAL;
    S : str128;
    opt : CARDINAL;
    memtype : CARDINAL;

    source,basepath:str128;
    filecount:CARDINAL;
    state:(waiting,gotspec);
BEGIN
    FIO.IOcheck := FALSE;
    WrLn;

    audiowarn       := FALSE;
    mincount        := 1;
    maxcount        := 50;
    wantedword      := "";
    memtype         := useMain;
    alphasort       := TRUE;
    verbose         := FALSE;
    flagShowZeroBased  := FALSE;
    DEBUG           := FALSE;

    state           := waiting;

    parms := Lib.ParamCount();
    IF parms = 0 THEN abort (errHelp,""); END;
    FOR i := 1 TO parms DO
        Lib.ParamStr(S,i);cleantabs(S);
        IF isOption(S)=TRUE THEN
            UpperCase (S);
            opt := GetOptIndex (S,"W:"+delim+"WORD:"+delim+
                                  "L:"+delim+"LOWER:"+delim+
                                  "U:"+delim+"UPPER:"+delim+
                                  "Z"+delim+"ZEROBASED"+delim+
                                  "LIM"+delim+"EMS"+delim+
                                  "A"+delim+"AUDIOWARN"+delim+
                                  "?"+delim+"H"+delim+"HELP"+delim+
                                  "SA"+delim+"NOALPHASORT"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "DEBUG"
                                );

            CASE opt OF
            | 1,2 :
                 GetString(S,wantedword);
                 LowerCase(wantedword);
                 FixAE(wantedword); (* unlikely but... *)
                 ReplaceChar(wantedword,"","c"); (* french safety *)
            | 3,4 :
                 IF GetLongCard(S,mincount)=FALSE THEN abort(errConversion,S);END;
            | 5,6 :
                 IF GetLongCard(S,maxcount)=FALSE THEN abort(errConversion,S);END;
            | 7,8:
                flagShowZeroBased:=TRUE;

            | 9,10 :
                 IF memtype # useMain THEN abort(errHelp,"");END; (* Q&D conflict resolution *)
                 memtype := useLIM;
            | 11,12:
                 audiowarn := TRUE;
            | 13,14,15 :
                abort(errHelp,"");
            | 16,17:
                alphasort:=FALSE;
            | 18,19:
                verbose:=TRUE;
            | 20:
                DEBUG := TRUE;
            ELSE
                abort(errUnknown,S);
            END;
        ELSE
            CASE state OF
            | waiting: UpperCase (S); source:=S;
            | gotspec: abort(errParmOverflow,S);
            END;
            INC(state);
        END;
    END;

    IF state = waiting THEN abort(errHelp,"");END;

    IF same(wantedword,"") THEN abort(errNeeded,""); END;

    IF mincount < 1 THEN abort (errMinCount,"-l:#"); END;
    IF maxcount < mincount THEN abort (errMaxCount,"-u:#"); END;
    IF Str.Length(wantedword) > maxLen THEN abort(errTooLong,wantedword); END;
    IF Str.Length(wantedword) < minLen THEN abort(errTooShort,wantedword); END;


    CASE memtype OF
    | useMain : ;
    | useLIM :
        IF LIM.LIMPresent = TRUE THEN
            pages := LIM.FreePages();
            available := sizePage * LONGCARD(pages);
            (* grab all LIM pages for simplicity! Q&D! *)
            LIMhandle := LIM.AllocatePages( pages );
            ChkLIMstatus();
        ELSE
            WrStr("-LIM option ignored, for LIM/EMS memory was not found !");
            WrLn;

            WrLn;
            memtype := useMain;
        END;
    END;

    IF same(source,dot) THEN source:=stardotstar;END;

    filecount := buildList(basepath, source);
    CASE filecount OF
    | 0 : abort(errNotFound,source);
    | MAX(CARDINAL): abort(errTooMany,source);
    ELSE
        IF alphasort THEN Lib.QSort(filecount,isLessByName,doSwapNames);END;
    END;

    (*
    WrStr(ProgTitle+" "+ProgVersion+" "+ProgCopyright);
    WrLn;
    *)

    WrLn;

    CASE memtype OF
    | useMain :
         WrStr("Using conventional memory");
    | useLIM :
         WrStr("Using LIM/EMS expanded memory (free bytes : ");
         IO.WrLngCard(available,1);
         WrStr(")");
    END;
    WrLn;

    WrLn;

    reversedwantedword:=wantedword;
    doReverse(reversedwantedword);

    FOR i := 1 TO filecount DO
        Str.Concat(source,basepath,fileArray[firstFile+i-1]);

        (* no longer in do*search as TO avoid isl342 ! *)

        WrStr("Text file        : "); WrStr(source);WrLn;
        WrStr('Searched string  : "');WrStr(wantedword);WrStr('"');WrLn;
        WrStr('Reversed string  : "');WrStr(reversedwantedword);WrStr('"');WrLn;
        WrStr("Minimum interval : ");IO.WrLngCard(mincount,1);WrLn;
        WrStr("Maximum interval : ");IO.WrLngCard(maxcount,1);WrLn;
        WrStr("Position mode    : ");
        IF flagShowZeroBased THEN
            WrStr("0");
        ELSE
            WrStr("1");
        END;
        WrStr("-based");WrLn;
        WrLn;

        globpollval := 0; (* init Win9X fix *)

        matchesfound := 0;

        FOR opt := 1 TO 2 DO
            CASE opt OF
            | 1: flagReverse := FALSE; wanted:=wantedword;
            | 2: flagReverse := TRUE;  wanted:=reversedwantedword;
            END;
            CASE memtype OF
            | useMain :
                doSearchMain ( matchesfound,
                               source, wanted,
                               CARDINAL(mincount),CARDINAL(maxcount),
                               flagReverse,audiowarn,verbose,flagShowZeroBased);
            | useLIM :
                doSearchLIM  ( matchesfound,
                               source, wanted,
                               CARDINAL(mincount),CARDINAL(maxcount),
                               flagReverse,audiowarn,verbose,flagShowZeroBased);
            END;
        END;

        IF matchesfound # 0 THEN WrLn;END;
        WrStr("Entries found    : ");
        IO.WrLngCard(matchesfound,1);WrLn;
        WrLn;
    END;

    CASE memtype OF
    | useLIM :
         LIM.DeAllocatePages(LIMhandle);
         ChkLIMstatus();
    END;
    abort(errNone,"");
END findHIDN.



(*
24 bytes of test file : convert it to LOWERCASE ! ;-)
WHELLHHEELXLHHXEHELLXLYZ
2,1
6,2
17,1
13,3
*)



(*
    (* do not use buffer here for writing: its whole 32Kb are flushed in
    binary mode, even though the last byte is at 16 Kb !!!! *)

    IOhndIn:=FIO.Create("EMS.bin"); (* does NOT like buffers here for writing !!! *)
    currpage := firstpage;
    LOOP
        PageOffset:=LIM.MapPage(firstPhysicalPage,currpage,LIMhandle); (* retourne physpage<<14 *)
        ChkLIMstatus();
        Lib.WordMove ([PageSegment:PageOffset],ADR(Buffer),wordsPerPage);
        FIO.WrBin(IOhndIn,Buffer,sizePage);
        INC(currpage);
        IF currpage = lastpage THEN
            IF lastgot <> 0 THEN
                PageOffset:=LIM.MapPage(firstPhysicalPage,currpage,LIMhandle); (* retourne physpage<<14 *)
                ChkLIMstatus();
                Lib.WordMove ([PageSegment:PageOffset],ADR(Buffer),wordsPerPage);
                FIO.WrBin(IOhndIn,Buffer,lastgot);
            END;
            EXIT;
        END;
    END;
    FIO.Close(IOhndIn);

*)







