(* ---------------------------------------------------------------
Title         Q&D FON <-> TXT Font Tool
Overview      self-explanatory !
Notes         very, very, very quick & dirty... :-(
Bugs
Wish List     useless -p[#] could be removed (see //FIXME)
              create a video buffer without setting 320x200x8 ? bah...

--------------------------------------------------------------- *)

MODULE FontTool;

IMPORT Lib;
IMPORT Str;
IMPORT IO;
IMPORT FIO;
IMPORT SYSTEM;
IMPORT Graph;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "FONTTOOL";
    ProgTitle     = "Q&D FON <-> TXT Font Tool";
    ProgVersion   = "v1.0e";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;
CONST
    extFON        = ".FON";
    extDAT        = ".DAT";
    extPCX        = ".PCX";

(* ------------------------------------------------------------ *)

CONST
    errNone               = 0;
    errHelp               = 1;
    errTooManyParms       = 2;
    errOption             = 3;
    errNonsenseCmd        = 4;
    errSyntax             = 5;
    errNotFound           = 6;
    errTooMany            = 7;
    errBadFont            = 8;
    errExists             = 9;
    errBadExt             = 10;
    errEither             = 11;
    errNotSpecific        = 12;
    errBuildOnly          = 13;
    errIrrelevant         = 14;
    errJokers             = 15;
    errInk                = 16;
    errPaper              = 17;
    errSameInkPaper       = 18;
    errGridColor          = 19;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <-b> <binary_font["+extFON+"]> [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" <-r> <text_definition["+extDAT+"]> [option]..."+nl+
"Syntax 3 : "+ProgEXEname+" <-s> <binary_font["+extFON+"]> <target> [option]..."+nl+
"Syntax 4 : "+ProgEXEname+" <-g> <binary_font["+extFON+"]> [option]..."+nl+
nl+
"This program builds a text definition file from a binary font (syntax 1),"+nl+
"rebuilds a binary font from a text definition file (syntax 2),"+nl+
"creates a squashed version of specified font (syntax 3)"+nl+
"or saves character set to a PCX file (syntax 4)."+nl+
nl+
"    -p[#] allow partial set of characters (-p2 = 25 lines, -p5 = 50 lines)"+nl+
"    -a    automatically adjust font height if different from 8 or 16"+nl+
"    -m#   squash method [1..6]"+nl+
"    -i:#  ink ([0..255], default is 10)"+nl+
"    -p:#  paper ([0..255], default is 0)"+nl+
"    -g:#  grid color ([0..255], default is same as paper)"+nl+
"    -x[x] increase character width by 1 or 2 (-xy = -x -y)"+nl+
"    -y[y] increase character height by 1 or 2 (-xy = -x -y)"+nl+
"    -o    overwrite"+nl+
"    -v    verbose"+nl+
nl+
"a) Options irrelevant to one syntax are silently ignored."+nl+
"b) Only DOS names are supported."+nl+
"c) -i:#, -p:#, -g:#, -x[x] and -y[y] options require -g option."+nl+
"   Note ink, paper and grid values are default 320x200x8 palette colors."+nl+
"d) -g option briefly sets 320x200x8 video mode, hence the brief black screen."+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(helpmsg);
    | errTooManyParms :
        Str.Concat(S,"Uneeded ",einfo);Str.Append(S," parameter !");
    | errOption :
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," option !");
    | errNonsenseCmd:
        S := "-b, -r, -s and -g options are mutually exclusive !";
    | errSyntax:
        S := "Syntax error in command line !";
    | errNotFound :
        Str.Concat(S,"No file matches ",einfo);Str.Append(S," specification !");
    | errTooMany:
        Str.Concat(S,"Too many files match ",einfo);Str.Append(S," specification !");
    | errBadFont :
        Str.Concat(S,einfo," font file should be 4096 or 2048 bytes long !");
    | errExists :
        Str.Concat(S,einfo," target already exists !");
    | errBadExt:
        Str.Concat(S,"Reserved target extension in ",einfo);
        Str.Append(S," source specification !");
    | errEither:
        S := "-p2 and -p5 options are mutually exclusive !";
    | errNotSpecific:
        S := "-p option is not specific enough here !";
    | errBuildOnly:
        S := "-p2 and -p5 options apply to -build command only !";
    | errIrrelevant:
        S := "-p option is irrelevant here !";
    | errJokers:
        Str.Concat(S,'Illegal joker(s) in "',einfo);Str.Append(S,'" !');
    | errInk:
        Str.Concat(S,'Illegal or out of range "',einfo);Str.Append(S,'" ink value !');
    | errPaper:
        Str.Concat(S,'Illegal or out of range "',einfo);Str.Append(S,'" paper value !');
    | errSameInkPaper:
        S := "Ink and paper must be different !";
    | errGridColor:
        Str.Concat(S,'Illegal or out of range "',einfo);Str.Append(S,'" grid color value !');
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp : ;
    ELSE
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
VAR
    ioBuffer : ARRAY [firstBufferByte..lastBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)

CONST
    charwidth        = 8;
    charheight50     = 8;
    charheight25     = 16;
    fontsize50       = 2048; (* 8 *)
    fontsize25       = 4096; (* 16 *)
    fontsizemax      = fontsize25;
    firstdata        = 0;
    lastdata         = fontsize25-1; (* biggest *)
VAR
    fontdata : ARRAY [firstdata..lastdata] OF BYTE;

PROCEDURE initFontBuffer (  );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=firstdata TO lastdata DO
        fontdata[i]:=BYTE(00H);
    END;
END initFontBuffer;

PROCEDURE chkFont (relaxed,autoheight:BOOLEAN; S:ARRAY OF CHAR):BOOLEAN;
VAR
    fsize:LONGCARD;
BEGIN
    fsize:=getFileSize(S);
    IF (relaxed OR autoheight) THEN RETURN (fsize <= fontsizemax); END;
    IF fsize=fontsize25 THEN RETURN TRUE; END;
    IF fsize=fontsize50 THEN RETURN TRUE; END;
    RETURN FALSE;
END chkFont;

(* ------------------------------------------------------------ *)

TYPE
    f8e3 = ARRAY [0..8+1+3-1] OF CHAR;
CONST
    firstfile = 1;
    maxfile   = 1000;
VAR
    fileArray : ARRAY [firstfile..maxfile] OF f8e3;

PROCEDURE buildFileList (S:ARRAY OF CHAR):CARDINAL;
VAR
    found     : BOOLEAN;
    direntry  : FIO.DirEntry;
    countFile : CARDINAL;
BEGIN
    countFile := 0;
    found := FIO.ReadFirstEntry(S,allfiles,direntry);
    WHILE found DO
        IF countFile = (maxfile-firstfile+1) THEN RETURN MAX(CARDINAL); END;
        fileArray[firstfile+countFile]:=f8e3(direntry.Name);
        Str.Caps(fileArray[firstfile+countFile]); (* useless ! *)
        INC (countFile);
        found :=FIO.ReadNextEntry(direntry);
    END;
    RETURN countFile;
END buildFileList;

PROCEDURE makebase (spec:ARRAY OF CHAR;VAR basepath:ARRAY OF CHAR );
CONST
    defaultdrive = SHORTCARD(0);
VAR
    u,d,n,e:str128;
BEGIN
    qdSplitAllPath(spec,u,d,n,e);
    Str.Concat(basepath,u,d);
    IF same(basepath,"") THEN
        Str.Concat(basepath, CHAR( ORD("A")-1+ FIO.GetDrive()), ":");
        FIO.GetDir(defaultdrive,d);
        Str.Append(basepath,d);
        fixDirectory(basepath); (* in case *)
    END;
END makebase;

PROCEDURE defaultExtension (ext:ARRAY OF CHAR;VAR R:ARRAY OF CHAR);
BEGIN
    IF Str.CharPos(R,dot)=MAX(CARDINAL) THEN
        Str.Append(R,ext);
    END;
END defaultExtension;

PROCEDURE makeTarget (S,ext:ARRAY OF CHAR;VAR R:ARRAY OF CHAR);
VAR
    u,d,n,e:str128;
BEGIN
    qdSplitAllPath(S,u,d,n,e);
    Lib.MakeAllPath(R,u,d,n,ext);
END makeTarget;

PROCEDURE padval (v : CARDINAL; digits,base : CARDINAL;padchar:CHAR) : str16;
VAR
    S,R,padStr : str16;
    ok : BOOLEAN;
    len,delta,i : CARDINAL;
BEGIN
    FOR i:=0 TO 15 DO
        padStr[i]:=padchar;
    END;
    digits := digits MOD 16; (* better safe than sorry! *)
    Str.CardToStr (LONGCARD(v),S,base,ok);
    len := Str.Length(S);
    IF len > digits THEN digits := len; END;
    delta := digits - len;
    Str.Slice (R,padStr,0,delta);
    Str.Append (R,S);
    RETURN R;
END padval;

PROCEDURE Filter (c : CHAR) : str16;
VAR
    R : str16;
BEGIN
    CASE ORD(c) OF
    | 0..ORD(blank)-1  : R:="'?'";
    | ORD(doublequote) : R:=singlequote+doublequote+singlequote;
    | 255              : R:="'?'";
    ELSE
                         Str.Concat(R,doublequote,c);Str.Append(R,doublequote);
    END;
    RETURN R;
END Filter;

PROCEDURE binstrToVal (S:ARRAY OF CHAR):BYTE;
VAR
    i:CARDINAL;
    ok:BOOLEAN;
BEGIN
    FOR i:=1 TO Str.Length(S) DO
        CASE S[i-1] OF
        | blank :
            S[i-1]:="0";
        ELSE
            S[i-1]:="1";
        END;
    END;
    RETURN BYTE( Str.StrToCard(S,2,ok) );
END binstrToVal;

(* ------------------------------------------------------------ *)

(* //FIXME *)

PROCEDURE fixFontDims (VAR fsize:CARDINAL;VAR targetcharheight:CARDINAL;
                      force25,force50,autoheight:BOOLEAN);
VAR
    charsize:CARDINAL;
BEGIN
    IF force25 THEN
        fsize := fontsize25;
    ELSIF force50 THEN
        fsize := fontsize50;
    END;

    charsize := (fsize DIV 256);

    IF autoheight THEN
        IF charsize > charheight50 THEN
            targetcharheight := charheight25;
        ELSE
            targetcharheight := charheight50;
        END;
    ELSE
        targetcharheight := charsize;
    END;
END fixFontDims;

PROCEDURE doBuild (force25,force50,autoheight,verbose:BOOLEAN;
                   infile:ARRAY OF CHAR;
                   VAR outfile:ARRAY OF CHAR):BOOLEAN;
VAR
    S:str128;
    hnd:FIO.File;
    fsize,got,i,j,p,v,targetcharheight:CARDINAL;
BEGIN
    IF verbose THEN Work(cmdInit);END;

    fsize := CARDINAL( getFileSize(infile) ); (* buffer/legal size already checked *)
    hnd:=FIO.OpenRead(infile);
    FIO.AssignBuffer(hnd,ioBuffer);
    got:= FIO.RdBin(hnd,fontdata,fsize);
    FIO.Close(hnd);

    fixFontDims(fsize,targetcharheight,  force25,force50,autoheight);

    makeTarget(infile,extDAT,outfile);
    hnd:=FIO.Create(outfile);
    FIO.AssignBuffer(hnd,ioBuffer);

    Str.Concat(S,"; Font file : ",infile);
    FIO.WrStr(hnd,S);FIO.WrLn(hnd);
    FOR i:=0 TO 255 DO
        IF verbose THEN Work(cmdSHOW);END;
        FIO.WrLn(hnd);
        FIO.WrStr(hnd,";");FIO.WrLn(hnd);

        Str.Concat (S,"; ::: ",padval(i,3,10," "));
        Str.Append (S,"  $");Str.Append(S,padval(i,2,16,"0"));
        Str.Append (S,"  ");Str.Append(S,Filter(CHR(i)));
        FIO.WrStr(hnd,S);FIO.WrLn(hnd);

        FIO.WrStr(hnd,";");FIO.WrLn(hnd);
        FIO.WrLn(hnd);

        p := i * targetcharheight;                  (* // *)
        FOR j:=1 TO targetcharheight DO             (* // *)
            v := CARDINAL( fontdata[p+j-1] );
            Str.Concat(S,doublequote,padval(v,8,2,"0"));Str.Append(S,doublequote);
            ReplaceChar (S,"0"," ");
            ReplaceChar (S,"1","*");
            FIO.WrStr(hnd,S);FIO.WrLn(hnd);
        END;
        (*
        (* ugly, ugly : call the Code Police ! *)
        FOR j:=(charsize+1) TO targetcharheight DO
            v := 0;
            Str.Concat(S,doublequote,padval(v,8,2,"0"));Str.Append(S,doublequote);
            ReplaceChar (S,"0"," ");
            ReplaceChar (S,"1","*");
            FIO.WrStr(hnd,S);FIO.WrLn(hnd);
        END;
        *)
    END;
    FIO.Flush(hnd);
    FIO.Close(hnd);
    IF verbose THEN Work(cmdStop);END;
    RETURN TRUE;
END doBuild;

(* ------------------------------------------------------------ *)

PROCEDURE doRebuild (relaxed,verbose:BOOLEAN; infile:ARRAY OF CHAR;
                     VAR outfile:ARRAY OF CHAR):BOOLEAN;
VAR
    S:str128;
    p,fsize,currline,len:CARDINAL;
    hnd:FIO.File;
    ok:BOOLEAN;
BEGIN
    IF verbose THEN Work(cmdInit);END;

    hnd:=FIO.OpenRead(infile);
    FIO.AssignBuffer(hnd,ioBuffer);
    currline:=0;
    p:=0;
    LOOP
        IF verbose THEN Work(cmdSHOW);END;
        FIO.RdStr(hnd,S);
        INC(currline);
        LtrimBlanks(S);
        RtrimBlanks(S);
        ok:=FALSE;
        CASE S[0] OF
        | CHR(0)      : ok:=TRUE;   (* empty string *)
        | semicolon   : ok:=TRUE;   (* comment *)
        | doublequote :
            IF CharCount(S,doublequote) = 2 THEN
                len:=Str.Length(S);
                IF S[len-1]=S[0] THEN
                    S[len-1]:=0C;
                    Str.Delete(S,0,1);
                    IF Str.Length(S) = 8 THEN
                        fontdata[p]:=binstrToVal(S);
                        ok:=TRUE;
                        INC(p);
                        IF p > lastdata THEN EXIT; END; (* force here *)
                    END;
                END;
            END;
        END;
        IF FIO.EOF THEN ok:=TRUE; EXIT; END; (* legitimate exit ! *)
        IF ok=FALSE THEN EXIT; END;
    END;
    FIO.Close(hnd);
    IF verbose THEN Work(cmdStop);END;

    IF ok THEN
        CASE p OF
        | fontsize25,fontsize50:
            fsize:=p;
        ELSE
            IF relaxed THEN
                fsize:=p;
                ok:=TRUE;
            ELSE
                ok:=FALSE;
            END;
        END;
    END;
    IF ok=FALSE THEN
        Str.Copy(outfile,padval(currline,4,10,"0"));
        RETURN FALSE;
    END;

    makeTarget(infile,extFON,outfile);
    hnd:=FIO.Create(outfile);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.WrBin(hnd,fontdata,fsize);
    FIO.Flush(hnd);
    FIO.Close(hnd);
    RETURN TRUE;
END doRebuild;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    method1 = 1;
    method2 = 2;
    method3 = 3;
    method4 = 4;
    method5 = 5;
    method6 = 6;
TYPE
    FontDef      = ARRAY [firstdata..lastdata] OF BYTE;
VAR
    gFont25 : FontDef;

(* see 25to50.bas and no, won't bother to write a cleaner bit-manipulating version ! *)

PROCEDURE squash (squashmethod:CARDINAL; VAR fontbuff:FontDef);
TYPE
    ktype     = ARRAY [1..3] OF CARDINAL;
    matrixtype= ARRAY [1..4] OF ktype;
CONST
    transfo1 = matrixtype ( ktype(0,1,0),
                            ktype(1,2,1),
                            ktype(1,2,1),
                            ktype(0,1,0));
    divisor1 = 2+2+1+1+1+1+1+1-1;
    transfo2 = matrixtype ( ktype(0,0,0),
                            ktype(0,2,0),
                            ktype(0,2,0),
                            ktype(0,0,0));
    divisor2 = 2+2-1;
    transfo3 = matrixtype ( ktype(0,0,0),
                            ktype(0,2,0),
                            ktype(0,2,0),
                            ktype(0,0,0));
    divisor3 = 2+2-1-1;
    transfo4 = matrixtype ( ktype(1,0,1),
                            ktype(0,2,0),
                            ktype(0,2,0),
                            ktype(1,0,1));
    divisor4 = 1+1+2+2+1+1-1;
TYPE
    str8 = ARRAY [1..8] OF CHAR;
CONST
    blanc = "00000000"; (* 8 blanks *)
    combi1= "0"+delim+"1";
    combi2= "00"+delim+"10"+delim+"01"+delim+"11";
VAR
    old : ARRAY [1..16] OF str8;
    now : ARRAY [1..8] OF str8;
    R,a,b : str8;
    ca,cb,c:CHAR;
    sa,sb,sc:str2;
    psrc,pto,n,i,w,j,y,x : CARDINAL;
    v:BYTE;
    lv:LONGCARD;
    ok:BOOLEAN;
    ii,jj,sigma,k,bval,divisor:CARDINAL;
    transfo:matrixtype;
BEGIN
    psrc := firstdata;
    pto  := firstdata;
    FOR n := 0 TO 255 DO
        FOR i:=1 TO 16 DO
            v:=fontbuff[psrc];
            Str.CardToStr( LONGCARD(v),R,2,ok);
            FOR w:=Str.Length(R)+1 TO 8 DO Str.Prepend(R,"0");END;
            old[i]:=R;
            INC(psrc);
        END;
        FOR i:=1 TO 8 DO
            now[i]:=blanc;
        END;

        CASE squashmethod OF
        | method1: (* raw squashing considering 2 dots (x,y x,y+1) *)
            j:=1;
            FOR y:=1 TO 16 BY 2 DO
		        a:=old[y];
		        b:=old[y+1];
                R:="";
		        FOR x:=1 TO 8 DO
        	        ca:=a[x];
                    cb:=b[x];
                    CASE getStrIndex(delim,ca,  combi1) OF
                    | 1:
                        CASE getStrIndex(delim,cb, combi1) OF
                        | 1: c:="0";
                        | 2: c:="1";
                        END;
                    | 2:
                         CASE getStrIndex(delim,cb, combi1) OF
                        | 1: c:="1";
                        | 2: c:="1";
                        END;
                    END;
                    Str.Append(R,c);
                END;
                now[j]:=R;
                INC(j);
            END;
        | method2: (* raw squashing considering 4 dots (x,y x,y+1 x+1,y x+1,y+1) *)
            j:=1;
            FOR y:=1 TO 16 BY 2 DO
		        a:=old[y];
                b:=old[y+1];
		        R:="";
		        FOR x:=1 TO 8 BY 2 DO
        	        Str.Slice (sa, a, x,2);
                    Str.Slice (sb, b, x,2);
                    CASE getStrIndex(delim,sa,  combi2) OF
                    | 1 :
                        CASE getStrIndex(delim,sb,  combi2) OF
                        | 1 : sc:="00";
                        | 2 : sc:="10";sc:="00";
                        | 3 : sc:="01";sc:="00";
                        | 4 : sc:="11";
                        END;
                    | 2 :
                        CASE getStrIndex(delim,sb,  combi2) OF
                        | 1 : sc:="10";sc:="00";
                        | 2 : sc:="10";
                        | 3 : sc:="01";
                        | 4 : sc:="11";
                        END;
                    | 3 :
                        CASE getStrIndex(delim,sb,  combi2) OF
                        | 1 : sc:="01"; sc:="00";
                        | 2 : sc:="10";
                        | 3 : sc:="01";
                        | 4 : sc:="11";
                        END;
                    | 4 :
                        CASE getStrIndex(delim,sb,  combi2) OF
                        | 1 : sc:="11";
                        | 2 : sc:="11";
                        | 3 : sc:="11";
                        | 4 : sc:="11";
                        END;
                    END;
                    Str.Append(R,sc);
                END;
                now[j]:=R;
                INC(j);
            END;
        | method3,method4,method5,method6:
            CASE squashmethod OF
            | method3 : transfo:=transfo1; divisor:=divisor1;
            | method4 : transfo:=transfo2; divisor:=divisor2;
            | method5 : transfo:=transfo3; divisor:=divisor3;
            | method6 : transfo:=transfo4; divisor:=divisor4;
            END;
            j:=1;
            FOR y:=1+1 TO 16-1 BY 2 DO
                a:=old[y];
                R:="";
		        FOR x:=1 TO 8 DO
		            CASE a[x] OF
		            | "0": bval:=0;
		            | "1": bval:=1;
		            END;
		            sigma:=0;
		            FOR jj:=1 TO 4 DO
		                CASE (y+jj-1-1) OF
		                | 1..16:
		                    FOR ii:=1 TO 3 DO
		                        CASE (x+ii-1-1) OF
		                        | 1..8:
		                            k:=transfo[jj][ii];
		                            INC(sigma,bval*k);
		                        END;
		                    END;
		                END;
		            END;
		            sigma:=sigma DIV divisor;
		            IF sigma=0 THEN
		                Str.Append(R,"0");
		            ELSE
		                Str.Append(R,"1");
		            END;
		        END;
		        now[j]:=R;
		        INC(j);
            END;
        END;

        (* update *)
        FOR i:=1 TO 8 DO
            R:=now[i];
            lv:=Str.StrToCard(R,2,ok);
            fontbuff[pto]:=BYTE(lv);
            INC(pto);
        END;
    END;
END squash;

PROCEDURE doSquash (method:CARDINAL;infile,outfile:ARRAY OF CHAR):BOOLEAN;
VAR
    hnd:FIO.File;
    got:CARDINAL;
BEGIN
    Lib.Fill(ADR(gFont25),fontsize25, 0);
    hnd:=FIO.OpenRead(infile);
    FIO.AssignBuffer(hnd,ioBuffer);
    got:=FIO.RdBin(hnd,gFont25,SIZE(gFont25));
    FIO.Close(hnd);

    squash (method, gFont25);

    hnd:=FIO.Create(outfile);
    FIO.AssignBuffer(hnd,ioBuffer);
    FIO.WrBin(hnd,gFont25,fontsize50);
    FIO.Flush(hnd);
    FIO.Close(hnd);

    RETURN TRUE;
END doSquash;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    egarange     = 40H;
    mininkindex  = 0;
    maxinkindex  = 256-1;
    ndxblack     = mininkindex;
    maxcolorindex= egarange-1; (* we use 64 colors palette here *)
    ndxwhite     = maxinkindex;

TYPE
    triplet = RECORD
        r,g,b:SHORTCARD;
    END;

PROCEDURE setDAC(index:CARDINAL;red,green,blue:BYTE);
CONST
    DACWriteIndex  = 03C8H;
    DACDataRegister= 03C9H;
BEGIN
    SYSTEM.Out (DACWriteIndex,SHORTCARD(index));
    SYSTEM.Out (DACDataRegister, red);
    SYSTEM.Out (DACDataRegister, green);
    SYSTEM.Out (DACDataRegister, blue);
END setDAC;

PROCEDURE resetdac(slow:BOOLEAN;index,red,green,blue:CARDINAL);
BEGIN
    setDAC(index,BYTE(red),BYTE(green),BYTE(blue));
END resetdac;

PROCEDURE getDAC (index:CARDINAL;VAR r,g,b:SHORTCARD);
CONST
    DACReadIndex   = 03C7H;
    DACDataRegister= 03C9H;
BEGIN
    SYSTEM.Out (DACReadIndex,SHORTCARD(index));
    r:=SYSTEM.In (DACDataRegister);
    g:=SYSTEM.In (DACDataRegister);
    b:=SYSTEM.In (DACDataRegister);
END getDAC;

PROCEDURE readDACrgb (index:CARDINAL):LONGCARD;
VAR
    r,g,b:SHORTCARD;
    rgb : LONGCARD;
BEGIN
    getDAC(index,r,g,b);
    rgb := (LONGCARD(r) << 16) + (LONGCARD(g) << 8) + LONGCARD(b);
    RETURN rgb;
END readDACrgb;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    hires       = 13H; (* 320x200x256 *)
    text        = 03H; (* 80x25 *)
    xcount      = 320;
    ycount      = 200;
    xmin        = 0;
    xmax        = xcount-1;
    ymin        = 0;
    ymax        = ycount-1;
    cx          = xmax DIV 2;
    cy          = ymax DIV 2;
    screensize  = xcount * ycount;
    screensizeW =(screensize DIV 2);
TYPE
    screenType = ARRAY [0..screensize-1 +xcount] OF BYTE;   (* 320x200=64000 with security *)
VAR
    videoscreen [0A000H:0000H] : screenType;
    Ybase                      : ARRAY [ymin..ymax] OF CARDINAL;

PROCEDURE initYbase ();
VAR
    i,p:CARDINAL;
BEGIN
    p := 0;
    FOR i := ymin TO ymax DO
        Ybase[i]:=p;
        INC(p,xcount);
    END;
END initYbase;

PROCEDURE vplot (x,y:CARDINAL;ink:BYTE);
BEGIN
    videoscreen[Ybase[y]+x]:=ink;
END vplot;

PROCEDURE vpoint (x,y:CARDINAL):BYTE;
BEGIN
    RETURN videoscreen[Ybase[y]+x];
END vpoint;

(* ------------------------------------------------------------ *)

TYPE
    PCXrgbType = RECORD
        red,green,blue:BYTE;
    END;

    (* (almost) useless fields are indented again *)

    PCXheaderType = RECORD
     manufacturer    : BYTE; (* constant 10=$0A=ZSoft *)
     version         : BYTE; (* constant 5=PC Paintbrush v3.0+ *)
     encoding        : BYTE; (* constant 1=RLE *)
    bitsPerPixel     : BYTE; (* 1, 2, 4 or 8, i.e. mono, 4, 16 OR 256 colors  *)
	                  (* window, i.e. image dimensions : xmin, ymin, xmax, ymax *)
    leftmargin       : WORD;
    uppermargin      : WORD;
    rightmargin      : WORD; (* do not use xmax-xmin ! *)
    lowermargin      : WORD;
     hDPIresolution  : WORD; (* put 300 for instance *)
     vDPIresolution  : WORD; (* idem *)
    colormap         : ARRAY[0..15] OF PCXrgbType;
     reserved1       : BYTE; (* must be 0 *)
    NCP              : BYTE; (* number of color planes, 1 or 4 *)
    NBS              : WORD; (* number of bytes per scanline, always even *)
                            (* from here, fields may contain rubbish *)
     paletteInfo     : WORD; (* 1=color or BW, 2=grayscale *)
     hscreensize     : WORD;
     vscreensize     : WORD;
     reserved2       : ARRAY[0..53] OF BYTE; (* must be 0 *)
    END;


(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    palette16=ARRAY[1..16] OF LONGCARD;
CONST
    pal16=palette16(
    Graph._BLACK    ,
    Graph._BLUE      ,
    Graph._GREEN     ,
    Graph._CYAN      ,
    Graph._RED       ,
    Graph._MAGENTA   ,
    Graph._BROWN     ,
    Graph._WHITE     ,
    Graph._GRAY     ,
    Graph._LIGHTBLUE ,
    Graph._LIGHTGREEN,
    Graph._LIGHTCYAN ,
    Graph._LIGHTRED  ,
    Graph._LIGHTMAGENTA,
    Graph._LIGHTYELLOW ,
    Graph._BRIGHTWHITE ); (* crire la palette dans l'ordre original de dfinition des couleurs ! *)

(* remember $FC >> 2 = $3F, but $FF >> 2 = $3F too ! *)

PROCEDURE PCXpal (VAR v:BYTE);
VAR
    i : CARDINAL;
BEGIN
    i := CARDINAL(v) << 2;
    IF i >= 0FCH THEN i := 0FFH; END; (* we lose $FC but it's better here ! *)
    v := BYTE(i);
END PCXpal;

PROCEDURE buildPCXheader (xorg,yorg,xdest,ydest,colors:CARDINAL):PCXheaderType;
VAR
    h : PCXheaderType;
    i : CARDINAL;
    r,g,b:BYTE;
    bgr:LONGCARD;
BEGIN
    h.manufacturer   := BYTE(0AH);
    h.version        := BYTE(5);
    h.encoding       := BYTE(1);
    i:=8;
    h.bitsPerPixel   := BYTE(i);
    h.leftmargin     := xorg;
    h.uppermargin    := yorg;
    h.rightmargin    := xdest;
    h.lowermargin    := ydest;
    h.hDPIresolution := 300;    (* reasonable but useless default *)
    h.vDPIresolution := 300;
        FOR i := 0 TO 15 DO
            bgr := pal16[i+1];
            r := BYTE( bgr        MOD 100H);
            g := BYTE((bgr >>  8) MOD 100H);
            b := BYTE((bgr >> 16) MOD 100H);
            PCXpal(r); PCXpal(g); PCXpal(b);
            h.colormap[i].red   := r;
            h.colormap[i].green := g;
            h.colormap[i].blue  := b;
        END;
    h.reserved1      := BYTE(0);
    h.NCP            := BYTE(1);

    i                := (xdest-xorg+1) ; (*  8 bits per pixel, so DIV 8 then * 8 *)

    h.NBS            := i;
    h.paletteInfo    := 1;
    h.hscreensize    := 0; (* xmaxi-xmini+1 *)
    h.vscreensize    := 0; (* ymaxi-ymini+1 *)
    FOR i := 0 TO 53 DO
        h.reserved2[i] := BYTE(0);
    END;

    RETURN h;
END buildPCXheader;

CONST
    PCXMAXCOUNT  = 03FH;
    PCXCOUNTFLAG = BYTE(0C0H); (* 11000000 *)
CONST
    maxPlanes = 4;
    maxpixels = 1280;
    firstcode = 0;
    (* maxcode   = (1280 DIV 8) * maxPlanes -1; (* in bytes *) *)
    maxcode   = maxpixels * maxPlanes-1;
TYPE
    imgtype = RECORD
        decodebuffer : ARRAY [firstcode..maxcode] OF BYTE;
    END;
VAR
    img : imgtype;

PROCEDURE encodeScanline256 (y,xorg,xdest:CARDINAL   ) : CARDINAL ;
VAR
    oldink,ink : BYTE;
    x,count:CARDINAL;
    p:CARDINAL; (* in img.decodebuffer[] *)
BEGIN
    x:=xorg;
    oldink:= vpoint(x,y);
    count := 1;
    p := firstcode;
    INC(x);
    LOOP
        ink := vpoint(x,y);
        IF ink=oldink THEN
            IF count = PCXMAXCOUNT THEN (* flush ! *)
                img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(count));
                INC(p);
                img.decodebuffer[p]:=oldink;
                INC(p);
                count:=1; (* write 63, write byte, restart at 1 *)
            ELSE
                INC(count);
            END;
        ELSE
            IF count > 1 THEN
                img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(count));
                INC(p);
                count:=1; (* write count, restart at 1 *)
            ELSE (* count is 1 here *)
                IF (oldink AND PCXCOUNTFLAG) = PCXCOUNTFLAG THEN
                    img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(01H));
                    INC(p); (* write 1 *)
                END;
            END;
            img.decodebuffer[p]:=oldink;
            INC(p); (* write byte *)
            oldink:=ink; (* restart at 1 with new byte *)
        END;

        INC(x);
        IF x > xdest THEN EXIT; END;
    END;

            IF count > 1 THEN
                img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(count));
                INC(p); (* write count *)
            ELSE (* count is 1 here *)
                IF (oldink AND PCXCOUNTFLAG) = PCXCOUNTFLAG THEN
                    img.decodebuffer[p]:=(PCXCOUNTFLAG OR BYTE(01H));
                    INC(p); (* write 1 *)
                END;
            END;
            img.decodebuffer[p]:=oldink;
            INC(p); (* write byte *)
    RETURN (p-firstcode); (* p is already +1, so useless to -1 *)
END encodeScanline256;

CONST
    palID     = BYTE(0CH); (* 12=$0C *)
    palidsize = 1;

PROCEDURE savePCXpal256 (hnd:FIO.File);
VAR
    i : CARDINAL;
    r,g,b:SHORTCARD; (* BYTE ! *)
    bgr:LONGCARD;
BEGIN
    FIO.WrBin(hnd,palID,palidsize);
        FOR i := 0 TO 255 DO
            getDAC(i,r,g,b);
            PCXpal(r); PCXpal(g); PCXpal(b);
            FIO.WrBin(hnd,r,1);
            FIO.WrBin(hnd,g,1);
            FIO.WrBin(hnd,b,1);
        END;
END savePCXpal256;

PROCEDURE savePCX (xorg,yorg,xdest,ydest:CARDINAL; pic:ARRAY OF CHAR);
VAR
    hnd:FIO.File;
    x,y,count,colors:CARDINAL;
    h : PCXheaderType;
BEGIN
    colors:=256;

    hnd:=FIO.Create(pic);
    FIO.AssignBuffer(hnd,ioBuffer);
    h := buildPCXheader(xorg,yorg,xdest,ydest,colors);
    FIO.WrBin(hnd,h,SIZE(PCXheaderType));
    FOR y := yorg TO ydest DO
        count:=encodeScanline256(y,xorg,xdest);
        FIO.WrBin(hnd, img.decodebuffer, count);
    END;
    savePCXpal256(hnd);
    FIO.Flush(hnd);
    FIO.Close(hnd);
END savePCX;

PROCEDURE setVideoMode (mode:CARDINAL);
VAR
    R : SYSTEM.Registers;
BEGIN
    R.AH := 00H;
    R.AL := BYTE(mode);
    Lib.Intr(R,10H);
END setVideoMode;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE doGenPic (force25,force50,autoheight:BOOLEAN;
                   xfix,yfix,columnscount,ink,paper,gridcolor:CARDINAL;
                   infile:ARRAY OF CHAR;
                   VAR outfile:ARRAY OF CHAR) : BOOLEAN ;
VAR
    hnd:FIO.File;
    bitmask,mask,bitstate,i,j,k,rows,columns,p,x,y:CARDINAL;
    targetcharheight,fsize,got:CARDINAL;
    grille,encre,papier:BYTE;
    xorg,yorg,xdest,ydest:CARDINAL;
BEGIN
    encre:=BYTE(ink);
    papier:=BYTE(paper);
    grille:=BYTE(gridcolor);

    fsize := CARDINAL( getFileSize(infile) ); (* buffer/legal size already checked *)
    hnd:=FIO.OpenRead(infile);
    FIO.AssignBuffer(hnd,ioBuffer);
    got:= FIO.RdBin(hnd,fontdata,fsize);
    FIO.Close(hnd);

    fixFontDims(fsize,targetcharheight,  force25,force50,autoheight);

    setVideoMode(hires);

    i:=0;
    FOR rows := 1 TO (256 DIV columnscount) DO
        y:= ymin + (rows-1) * (targetcharheight+yfix);   (* // *)
        FOR columns := 1 TO columnscount DO

            x := xmin + (columns-1) * (charwidth+xfix);
            p := i * targetcharheight;                   (* // *)
            FOR j:=1 TO targetcharheight DO              (* // *)
                mask := CARDINAL( fontdata[p+j-1] );
                bitmask := 080H; (* %10000000 *)
                FOR k:=1 TO charwidth DO
                    bitstate := (mask AND bitmask ); (* 256>>k would do too *)
                    IF bitstate # 0 THEN
                        vplot(x+k-1,y+j-1,encre);
                    ELSE
                        vplot(x+k-1,y+j-1,papier);
                    END;
                    bitmask:=(bitmask >> 1);
                END;
                (* grid beyond charwidth *)
                FOR k:=(charwidth+1) TO (charwidth+xfix) DO
                    vplot(x+k-1,y+j-1,grille);
                END;
            END;
            (* grid beyond charheight *)
            FOR j:=(targetcharheight+1) TO (targetcharheight+yfix) DO
                FOR k:=1 TO (charwidth+xfix) DO
                vplot(x+k-1,y+j-1,grille);
                END;
            END;

            INC(i); (* advance data pointer... er, index *)
        END;
    END;

    xorg := xmin;
    yorg := ymin;
    xdest:= xmin + columnscount * (charwidth+xfix);
    DEC(xdest);
    ydest:= ymin + rows         * (targetcharheight+yfix);
    DEC(ydest);

    makeTarget(infile,extPCX,outfile);
    savePCX(xorg,yorg,xdest,ydest, outfile);

    setVideoMode(text);
    RETURN TRUE;
END doGenPic;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE getcard (VAR v:CARDINAL;
                  mini,maxi:CARDINAL; S:ARRAY OF CHAR  ):BOOLEAN;
VAR
    ok:BOOLEAN;
    lc:LONGCARD;
BEGIN
    ok:= GetLongCard (S, lc);
    IF ok THEN
        IF lc > MAX(CARDINAL) THEN
            ok:=FALSE;
        ELSE
            v:=CARDINAL(lc);
            IF ( ( v < mini ) OR ( v > maxi ) ) THEN
                ok:=FALSE;
            ELSE
                ok:=TRUE; (* useless *)
            END;
        END;
    END;
    RETURN ok;
END getcard;

(* ------------------------------------------------------------ *)

CONST
    defaultink    = 10;
    defaultpaper  = 0;
CONST
    prefixok      = "+++ ";
    prefixerr     = "--- ";
    msgProcessing = "Processing ";
    msgCreated    = " created !";
    msgProblemAt  = "Problem at line ";
    msgInFile     = " while processing ";
VAR
    parmcount,i,opt : CARDINAL;
    S,R             : str128;
    infile,outfile,base : str128;
    state           : (waiting,gotparm1,gotparm2);
    cmd             : (nocmd,build,rebuild,squashfont,genpic);
    overwrite,verbose,usergridcolor:BOOLEAN;
    xfix,yfix:CARDINAL;
    relaxed,relaxed25,relaxed50,autoheight : BOOLEAN;
    columnscount,method,ink,paper,gridcolor  : CARDINAL;
    countFile       : CARDINAL;
    ok              : BOOLEAN;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn;

    cmd       := nocmd;
    overwrite := FALSE;
    verbose   := FALSE;
    relaxed   := FALSE;
    relaxed25 := FALSE;
    relaxed50 := FALSE;
    autoheight:= FALSE;
    method    := method1;
    columnscount := 32;    (* will fit 320x200 screen, 16 would required 640x480 *)
    ink       := defaultink;
    paper     := defaultpaper;
    gridcolor := paper; (* useless here *)
    usergridcolor := FALSE;
    xfix      := 0;
    yfix      := 0;
    state     := waiting;

    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;

    FOR i := 1 TO parmcount DO (* for future extension ! *)
        Lib.ParamStr(S,i);
        RtrimBlanks(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "B"+delim+"BUILD"+delim+
                                  "R"+delim+"REBUILD"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "P"+delim+"PARTIAL"+delim+
                                  "P2"+delim+"25"+delim+
                                  "P5"+delim+"50"+delim+
                                  "A"+delim+"AUTO"+delim+
                                  "S"+delim+"SQUASH"+delim+
                                  "M1"+delim+
                                  "M2"+delim+
                                  "M3"+delim+
                                  "M4"+delim+
                                  "M5"+delim+
                                  "M6"+delim+
                                  "G"+delim+"PCX"+delim+
                                  "I:"+delim+"INK:"+delim+
                                  "P:"+delim+"PAPER:"+delim+
                                  "X"+delim+"XBY1"+delim+
                                  "Y"+delim+"YBY1"+delim+
                                  "XY"+delim+"YX"+delim+"XYBY1"+delim+
                                  "G:"+delim+"GRID:"+delim+
                                  "XX"+delim+"XBY2"+delim+
                                  "YY"+delim+"YBY2"+delim+
                                  "XXYY"+delim+"YYXX"+delim+"XYBY2"
                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5   : CASE cmd OF
                      | nocmd,build:cmd:=build;
                      ELSE abort(errNonsenseCmd,"");END;
            | 6,7   : CASE cmd OF
                      | nocmd,rebuild:cmd:=rebuild;
                      ELSE abort(errNonsenseCmd,"");END;
            | 8,9   : overwrite := TRUE;
            | 10,11 : verbose   := TRUE;
            | 12,13 : relaxed   := TRUE;
            | 14,15 : relaxed25 := TRUE;
            | 16,17 : relaxed50 := TRUE;
            | 18,19 : autoheight:= TRUE;
            | 20,21 : CASE cmd OF
                      | nocmd,squashfont:cmd:=squashfont;
                      ELSE abort(errNonsenseCmd,"");END;
            | 22:     method    :=method1;
            | 23:     method    :=method2;
            | 24:     method    :=method3;
            | 25:     method    :=method4;
            | 26:     method    :=method5;
            | 27:     method    :=method6;
            | 28,29:  CASE cmd OF
                      | nocmd,genpic:cmd:=genpic;
                      ELSE abort(errNonsenseCmd,"");END;
            | 30,31:  IF getcard(ink,  mininkindex,maxinkindex,R)=FALSE THEN
                          abort(errInk,S);
                      END;
            | 32,33:  IF getcard(paper,mininkindex,maxinkindex,R)=FALSE THEN
                          abort(errPaper,S);
                      END;
            | 34,35:   xfix      := 1;
            | 36,37:                   yfix      := 1;
            | 38,39,40:xfix      := 1; yfix      := 1;
            | 41,42:  IF getcard(gridcolor,mininkindex,maxinkindex,R)=FALSE THEN
                          abort(errGridColor,S);
                      END;
                      usergridcolor:=TRUE;
            | 43,44:   xfix      := 2;
            | 45,46:                   yfix      := 2;
            | 47,48,49:xfix      := 2; yfix      := 2;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting:  Str.Copy(infile,R);
            | gotparm1: Str.Copy(outfile,R);
            | gotparm2: abort(errTooManyParms,S); (* errSyntax would do *)
            END;
            INC(state);
        END;
    END;
    CASE state OF
    | waiting: abort(errSyntax,"");
    | gotparm1:
        CASE cmd OF
        | nocmd:
            abort(errSyntax,"");
        | build:
            defaultExtension(extFON,infile);
            IF Str.Pos(infile,extDAT) # MAX(CARDINAL) THEN
                abort(errBadExt,infile);
            END;
            IF (relaxed25 AND relaxed50) THEN abort(errEither,"");END;
            IF relaxed THEN abort(errNotSpecific,"");END;
            IF (relaxed25 OR relaxed50) THEN relaxed:=TRUE;END; (* trick, eh eh *)
        | rebuild:
            defaultExtension(extDAT,infile);
            IF Str.Pos(infile,extFON) # MAX(CARDINAL) THEN
                abort(errBadExt,infile);
            END;
            IF (relaxed25 OR relaxed50) THEN abort(errBuildOnly,"");END;
        | squashfont:
            abort(errSyntax,"");
        | genpic:
            defaultExtension(extFON,infile);
            IF Str.Pos(infile,extPCX) # MAX(CARDINAL) THEN
                abort(errBadExt,infile);
            END;
            IF (relaxed25 AND relaxed50) THEN abort(errEither,"");END;
            IF relaxed THEN abort(errNotSpecific,"");END;
            IF (relaxed25 OR relaxed50) THEN relaxed:=TRUE;END; (* trick, eh eh *)

            IF ink = paper THEN abort(errSameInkPaper,"");END;
            IF NOT(usergridcolor) THEN gridcolor:=paper;END;

            initYbase();

        END;
    | gotparm2:
        CASE cmd OF
        | nocmd,build,rebuild,genpic :
            abort(errSyntax,"");
        | squashfont :
            defaultExtension(extFON,infile);
            IF Str.Pos(infile,extDAT) # MAX(CARDINAL) THEN
                abort(errBadExt,infile);
            END;
            IF chkJoker(infile) THEN abort(errJokers,infile);END;
            IF chkJoker(outfile) THEN abort(errJokers,outfile);END;
            IF relaxed THEN abort(errIrrelevant,"");END;
            IF (relaxed25 OR relaxed50) THEN abort(errBuildOnly,"");END;
        END;
    END;

    makebase(infile,base);
    countFile:=buildFileList(infile);
    CASE countFile OF
    | 0 : abort(errNotFound,infile);
    | MAX(CARDINAL):abort(errTooMany,infile);
    END;

    (* first, we check for possible problems *)

    FOR i := 1 TO countFile DO
        Str.Copy(S,fileArray[firstfile+i-1]);
        Str.Concat(infile,base,S);
        CASE cmd OF
        | build:
            IF chkFont(relaxed,autoheight,infile)=FALSE THEN abort(errBadFont,infile);END;
            IF NOT(overwrite) THEN
                makeTarget(infile,extDAT,S);
                IF FIO.Exists(S) THEN abort(errExists,S);END;
            END;
        | rebuild:
            IF NOT(overwrite) THEN
                makeTarget(infile,extFON,S);
                IF FIO.Exists(S) THEN abort(errExists,S);END;
            END;
        | squashfont:
            (* infile does not contain any joker *)
            IF NOT(overwrite) THEN
                IF FIO.Exists(outfile) THEN abort(errExists,outfile);END;
            END;
        | genpic:
            IF chkFont(relaxed,autoheight,infile)=FALSE THEN abort(errBadFont,infile);END;
            IF NOT(overwrite) THEN
                makeTarget(infile,extPCX,S);
                IF FIO.Exists(S) THEN abort(errExists,S);END;
            END;
        END;
    END;

    (* now we can proceed *)

    WrStr(Banner);WrLn;
    WrLn;

    FOR i := 1 TO countFile DO
        Str.Copy(S,fileArray[firstfile+i-1]);
        Str.Concat(infile,base,S);
        Str.Concat(S,msgProcessing,infile);
        video(S,TRUE);
        initFontBuffer; (* in case we are in relaxed mode *)
        CASE cmd OF
        | build:
            ok:=doBuild(relaxed25,relaxed50,autoheight,verbose,infile,R);
        | rebuild:
            ok:=doRebuild(relaxed,verbose,infile,R);
        | squashfont:
            ok:=doSquash(method,infile,outfile);
            Str.Copy(R,outfile);
        | genpic:
            ok:=doGenPic (relaxed25,relaxed50,autoheight,
                         xfix,yfix,columnscount,ink,paper,gridcolor,infile,R);
        END;
        video(S,FALSE);
        IF ok THEN
            Str.Concat(S,prefixok,R);Str.Append(S,msgCreated);
        ELSE
            Str.Concat(S,prefixerr,msgProblemAt);Str.Append(S,R);
            Str.Append(S,msgInFile);Str.Append(S,infile);
        END;
        WrStr(S);WrLn;
    END;

    abort(errNone,"");
END FontTool.


