(* ---------------------------------------------------------------
Title         Q&D Missing
Author        PhG
Overview      see help
Usage         see help
Notes

              handle any file with one valid decimal number
              specifying its position (prepended, appended)
              find lowest, find highest, find missing
Bugs          not so good an idea to allow several number positions
Wish List     write code for unused -b and -u
              inner value ?
              show duplicate numbers

--------------------------------------------------------------- *)

MODULE Missing;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;

FROM IO IMPORT WrStr, WrLn;

FROM Storage IMPORT ALLOCATE,DEALLOCATE,Available;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype,w9XnothingRequired;

FROM QD_LFN IMPORT path9X, huge9X, findDataRecordType,
unicodeConversionFlagType, w9XchangeDir,
w9XgetDOSversion, w9XgetTrueDOSversion, w9XisWindowsEnh, w9XisMSDOS7,
w9XfindFirst, w9XfindNext, w9XfindClose, w9XgetCurrentDirectory,
w9XlongToShort, w9XshortToLong, w9XtrueName, w9XchangeDir,
w9XmakeDir, w9XrmDir, w9Xrename, w9XopenFile, w9XcloseFile,
w9XsupportLFN;

(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "MISSING";
    ProgTitle     = "Q&D missing numbered entries detector";
    ProgVersion   = "v1.0f";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errTooManyParms = 3;
    errMissingSpec  = 4;
    errTooManyFiles = 5;
    errNoMatch      = 6;
    errExpecting    = 7;
    errExpectingCmd = 8;
    errVal          = 9;
    errValRange     = 10;
    errPadTooSmall  = 11;
    errRequired     = 12;
    errIllogical    = 13;
    errHighlanderTheSeries = 14; (* errThereCanBeOnlyOne *)
    errAbortedByUser= 255;

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
CONST
    msghelp=
Banner+nl+
nl+
"Syntax : "+ProgEXEname+" <specification> [option]..."+nl+
nl+
"This program detects missing entries in numbered files."+nl+
nl+
"-w[w]:# build command to rename filenames with index padded to specified width"+nl+
'-p      process prepended number ("#*" pattern)'+nl+
'-a      process appended number ("*#" pattern)'+nl+
'-i      process first inner number ("*#*" pattern)'+nl+
'-d      process number alone ("#" pattern)'+nl+
nl+
"-m      show lower index, upper index and maximum index width then terminate"+nl+
"-k      show duplicate values"+nl+
"-v      verbose display (individual index instead of range)"+nl+
"-t      do not left- and right- trim filename (default is to trim filename)"+nl+
"-c      use COPY instead of RENAME (-w[w]:# only)"+nl+
"-n      do not show header (-w[w]:# only, -ww:# = -w:# -n)"+nl+
"-z      warn if target name does not fit DOS n8.e3 pattern (-w[w]:# only)"+nl+
"-!      disable ESC polling"+nl+
"-x      disable LFN support even if available"+nl+
nl+
"a) Only LONGCARD decimal numbers are supported."+nl+
"b) Number formatting does not matter : value does"+nl+
'   ("1" = "01" = "001" etc.).'+nl+
'c) Please note -k option lists values and not strings ("1" = "01" = "001" etc.)'+nl+
"d) -m option requires any of -p, -a, -i or -d options."+nl+
"e) -w[w]:# option requires only one of -p, -a, -i or -d options."+nl+
'   It ignores -k and -v options. Indexes are padded with "0".'+nl+
"   Without -n option, redirected output will require editing."+nl+
"   Note filenames will not have a path."+nl+
"f) Filename extension is ignored."+nl+
"g) Directories are ignored."+nl+
"h) -z option is forced is LFN support is either unavailable or disabled."+nl+
"i) Unless -! option was specified, ESCape key aborts listing."+nl+
"j) -debug option shows list of processed files ; -list omits excluded files."+nl;

VAR
    S : str1024;
BEGIN
    CASE e OF
    | errHelp :         WrStr(msghelp);
    | errOption :       S := 'Illegal "~" option !';
    | errTooManyParms : S := '"~" is one parameter too many !';
    | errMissingSpec :  S := "Missing file specification !";
    | errTooManyFiles : S := 'Too many numbered files match "~" specification !';
    | errNoMatch :      S := 'No numbered file matches "~" specification !';
    | errExpecting:     S := "Expecting ~ !";
    | errExpectingCmd:  S := "Expecting at least one of -p, -a, -i or -d options !";
    | errVal:           S := 'Illegal "~" value !';
    | errValRange:      S := '"~" value cannot be more than 16 !';
    | errPadTooSmall:   S := "User-specified width is smaller than maximum index width !";
    | errRequired:      S := "-n option requires -w[w]:# option !";
    | errIllogical:     S := "-n and -m options are mutually exclusive !";
    | errHighlanderTheSeries: S:="There can be only one ! ;-)";
                        S := "Expecting only one of -p, -a, -i or -d options !";
    | errAbortedByUser: S := "Listing aborted by user !";
    ELSE
                        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp:
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE pad (wi:CARDINAL;doleft:BOOLEAN;padchar:CHAR;S:ARRAY OF CHAR):str80;
VAR
    R:str80;
    i:CARDINAL;
BEGIN
    Str.Copy(R,S);
    FOR i:=(Str.Length(S)+1) TO wi DO
        IF doleft THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
    RETURN R;
END pad;

(* //TODO beautify and french are maybeware *)

CONST
    wiglobal = 10;

PROCEDURE num2str (v:LONGCARD;wi:CARDINAL;
                  padme,beautify,french:BOOLEAN;padchar:CHAR):str16;
CONST
    base = 10;
VAR
    ok:BOOLEAN;
    R:str16;
BEGIN
    Str.CardToStr(v,R,base,ok);
    IF padme THEN
        RETURN str16 ( pad(wi,FALSE,padchar,R) );
    ELSE
        RETURN R;
    END;
END num2str;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* //FIXME *)

CONST
    CHKEVERY9X = 128; (* let's call chkEscape every CHKEVERY LOOP *)
    CHKEVERYDOS= 8;  (* safety *)

PROCEDURE pollEscape (VAR chkrounds:CARDINAL; useLFN,ignoreESC:BOOLEAN):BOOLEAN;
VAR
    alcatraz:BOOLEAN;
    every:CARDINAL;
BEGIN
    (*
    9x really does NOT like ChkEscape() and randomely hangs till a keypress
    Flushkey() does NOT help
    *)
    (*
    IF useLFN THEN
        hit:=FALSE;
    ELSE
        hit:=ChkEscape();
    END;
    *)
    IF ignoreESC THEN RETURN FALSE;END;
    IF useLFN THEN
        every:=CHKEVERY9X;
    ELSE
        every:=CHKEVERYDOS;
    END;
    INC(chkrounds);
    IF (chkrounds MOD every) = 0 THEN
        chkrounds:=0;
        alcatraz:=ChkEscape();
    ELSE
        alcatraz:=FALSE;
    END;
    RETURN alcatraz;
END pollEscape;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    numpostype = (startpos,endpos,innerpos,numberalone);
    setofnumpos = SET OF numpostype;

PROCEDURE newpos (VAR numpos:setofnumpos;wanted:numpostype );
BEGIN
    INCL(numpos,wanted);
END newpos;

PROCEDURE chkoneifhere(numpos:setofnumpos;cmd:numpostype):BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    IF (cmd IN numpos) THEN
        EXCL(numpos,cmd);
        rc:=(numpos = setofnumpos{});
    ELSE
        rc:=TRUE;
    END;
    RETURN rc;
END chkoneifhere;

(* ------------------------------------------------------------ *)

(* we assume S is mere "n.e" *)

PROCEDURE chkMatch (VAR num:LONGCARD;VAR startnum,winum:CARDINAL;
                   numpos:setofnumpos;S:pathtype ):BOOLEAN ;
CONST
    base = 10;
VAR
    ok,rc:BOOLEAN;
    Z:str16;
    headerhere,trailerhere:CARDINAL;
    i,last:CARDINAL;
BEGIN
    i:=Str.RCharPos(S,dot);
    IF i # MAX(CARDINAL) THEN S[i]:=charnull;END;
    rc:=TRUE;
    last:=Str.Length(S);
    IF last = 0 THEN RETURN FALSE;END; (* cannot happen anyway *)
    DEC(last); (* 0-based *)

    Z:="";

    IF startpos IN numpos THEN (* "#*" ok *)
        startnum:=MAX(CARDINAL);
        winum:=0;
        i:=0;
        LOOP
            IF i > last THEN EXIT;END;
            IF Str.CharPos(digits,S[i]) = MAX(CARDINAL) THEN EXIT;END;
            IF startnum = MAX(CARDINAL) THEN startnum:=i;END;
            Str.Append(Z,S[i]);
            INC(i);
            INC(winum);
        END;
        IF verifyString(S,digits) THEN Z:="";END;
    END;

    ok := (Str.Length(Z) = 0);
    IF ( ok AND (endpos IN numpos) ) THEN (* "*#" *)
        startnum:=last+1;
        winum:=0;
        i:=last;
        LOOP
            IF Str.CharPos(digits,S[i]) = MAX(CARDINAL) THEN EXIT;END;
            Str.Prepend(Z,S[i]);
            IF i = 0 THEN EXIT;END;
            DEC(i);
            DEC(startnum);
            INC(winum);
        END;
        IF verifyString(S,digits) THEN Z:="";END;
    END;

    ok := (Str.Length(Z) = 0);
    IF ( ok AND (innerpos IN numpos) ) THEN (* "*#*" ok *)
        headerhere:=0;
        trailerhere:=0;
        i:=0;
        LOOP
            IF i > last THEN EXIT; END;
            IF Str.CharPos(digits,S[i]) # MAX(CARDINAL) THEN EXIT;END;
            INC(headerhere);
            INC(i);
        END;
        startnum:=i;
        winum:=0;
        LOOP
            IF i > last THEN EXIT; END;
            IF Str.CharPos(digits,S[i]) = MAX(CARDINAL) THEN
                INC(trailerhere);
                EXIT;
            END;
            Str.Append(Z,S[i]);
            INC(i);
            INC(winum);
        END;
        IF NOT ( (headerhere # 0) AND (trailerhere # 0 ) ) THEN Z:="";END;
    END;

    ok := (Str.Length(Z) = 0);
    IF ( ok AND (numberalone IN numpos) ) THEN (* "*" ok *)
        startnum:=0;
        winum:=0;
        i:=0;
        LOOP
            IF i > last THEN EXIT;END;
            IF Str.CharPos(digits,S[i]) = MAX(CARDINAL) THEN EXIT;END;
            Str.Append(Z,S[i]);
            INC(i);
            INC(winum);
        END;
        IF verifyString(S,digits)=FALSE THEN Z:="";END;
    END;

    rc:=Str.Length(Z) # 0;
    IF rc THEN num:=Str.StrToCard(Z,base,ok); END;
    RETURN rc;
END chkMatch;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    pFname = POINTER TO fnameType; (* yes, bad name *)
    fnameType = RECORD
        next      : pFname;
        startnum  : SHORTCARD;
        winum     : SHORTCARD;
        value     : LONGCARD;
        slen      : SHORTCARD;
        str       : CHAR;
    END;

PROCEDURE initList (VAR anchor : pFname );
BEGIN
    anchor := NIL;
END initList;

PROCEDURE freeList (anchor : pFname);
VAR
    needed : CARDINAL;
    p      : pFname;
BEGIN
    (* p:=anchor; *)
    WHILE anchor # NIL DO
        needed := SIZE(fnameType) - SIZE(anchor^.str) + CARDINAL(anchor^.slen);
        p := anchor^.next;
        DEALLOCATE(anchor,needed);
        anchor:=p;
    END
END freeList;

PROCEDURE buildNewPtr (VAR anchor,p:pFname; len:CARDINAL):BOOLEAN;
VAR
    needed : CARDINAL;
BEGIN
    needed := SIZE(fnameType) - SIZE(p^.str) + len;
    IF Available(needed)=FALSE THEN RETURN FALSE; END;
    IF anchor = NIL THEN
        ALLOCATE(anchor,needed);
        p:=anchor;
    ELSE
        p:=anchor;
        WHILE p^.next # NIL DO
            p:=p^.next;
        END;
        ALLOCATE(p^.next,needed);
        p:=p^.next;
    END;
    p^.next := NIL;
    RETURN TRUE;
END buildNewPtr;

(* assume p is valid *)

PROCEDURE getStr (VAR S : pathtype; p:pFname);
VAR
    len:CARDINAL;
BEGIN
    len := CARDINAL(p^.slen);
    Lib.FastMove( ADR(p^.str),ADR(S),len);
    S[len] := nullchar; (* REQUIRED safety ! *)
END getStr;

PROCEDURE foundvalue (v:LONGCARD; anchor:pFname):BOOLEAN;
VAR
    ptr:pFname;
BEGIN
    ptr:=anchor;
    LOOP
        IF ptr = NIL THEN EXIT;END;
        IF ptr^.value = v THEN RETURN TRUE;END;
        ptr:=ptr^.next;
    END;
    RETURN FALSE;
END foundvalue;

PROCEDURE countme (v:LONGCARD; anchor:pFname):LONGCARD;
VAR
    ptr:pFname;
    n:LONGCARD;
BEGIN
    n:=0;
    ptr:=anchor;
    LOOP
        IF ptr = NIL THEN EXIT;END;
        IF ptr^.value = v THEN INC(n);END;
        ptr:=ptr^.next;
    END;
    RETURN n;
END countme;

(* ------------------------------------------------------------ *)

PROCEDURE isReservedEntry (S:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    IF same(S,dot) THEN RETURN TRUE; END;
    RETURN same(S,dotdot);
END isReservedEntry;

PROCEDURE buildMatchList (VAR anchor:pFname;
                         DEBUG,showexcluded,useLFN,trim:BOOLEAN;
                         numpos:setofnumpos;spec:pathtype):CARDINAL;
VAR
    count:CARDINAL; (* should do ! *)
    ok,found:BOOLEAN;
    unicodeconversion:unicodeConversionFlagType;
    w9Xentry : findDataRecordType;
    w9Xhandle,errcode:CARDINAL;
    entry : FIO.DirEntry;
    dosattr:FIO.FileAttr;
    entryname:pathtype;
    len : CARDINAL;
    pp:pFname;
    includeme:BOOLEAN;
    lc:LONGCARD;
    startnum,winum:CARDINAL;
    S:str80;
BEGIN
    count:=0;
    IF useLFN THEN
        found := w9XfindFirst (spec,SHORTCARD(everything),SHORTCARD(w9XnothingRequired),
                              unicodeconversion,w9Xentry,w9Xhandle,errcode);
    ELSE
        found := FIO.ReadFirstEntry(spec,everything,entry);
    END;
    WHILE found DO
        IF useLFN THEN
            Str.Copy(entryname,w9Xentry.fullfilename);
        ELSE
            Str.Copy(entryname,entry.Name);
        END;
        IF trim THEN
            LtrimBlanks(entryname);
            RtrimBlanks(entryname);
        END;
        includeme := NOT( isReservedEntry(entryname) ); (* skip "." and ".." *)
        includeme := ( includeme AND chkMatch(lc,startnum,winum, numpos,entryname) );
        IF includeme THEN
            IF useLFN THEN
                dosattr:=FIO.FileAttr(w9Xentry.attr AND 0FFH);
            ELSE
                dosattr:=entry.attr;
            END;
            IF NOT (aD IN dosattr) THEN
                (* if file has no extension, add it as a marker *)
                IF Str.RCharPos(entryname,".")=MAX(CARDINAL) THEN
                    Str.Append(entryname,".");
                END;
                len:=Str.Length(entryname);
                IF buildNewPtr(anchor,pp,len)=FALSE THEN
                    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
                    RETURN MAX(CARDINAL); (* errStorage *)
                END;
                INC(count);
                pp^.startnum:=SHORTCARD(startnum);
                pp^.winum:=SHORTCARD(winum);
                pp^.value := lc;
                pp^.slen      := SHORTCARD(len);
                Lib.FastMove ( ADR(entryname),ADR(pp^.str),len );
            ELSE
                includeme:=FALSE;
            END;
        END;

        IF DEBUG THEN
            IF includeme THEN
                S:="+++ ";
                ok:=TRUE;
            ELSE
                S:="--- ";
                ok:=showexcluded;
            END;
            IF ok THEN WrStr(S);WrStr(entryname);WrLn; END;
            IF includeme THEN
                S:="::: pos,count = ~,~";
                Str.Subst(S,"~", num2str ( LONGCARD(startnum),wiglobal,
                                           TRUE,TRUE,TRUE,blank) );
                Str.Subst(S,"~", num2str ( LONGCARD(winum   ),wiglobal,
                                           TRUE,TRUE,TRUE,blank) );
                WrStr(S);WrLn;
            END;
        END;

        IF useLFN THEN
            found :=w9XfindNext(w9Xhandle, unicodeconversion,w9Xentry,errcode);
        ELSE
            found :=FIO.ReadNextEntry(entry);
        END;
    END;
    IF useLFN THEN ok:=w9XfindClose(w9Xhandle,errcode); END;
    IF DEBUG THEN WrLn;END;
    RETURN count;
END buildMatchList;

PROCEDURE buildPath (VAR path:pathtype; spec:pathtype);
VAR
    u,d,n,e:pathtype;
BEGIN
    qdSplitAllPath(spec, u,d,n,e);
    Str.Concat(path, u,d);
    fixDirectory(path); (* safety *)
END buildPath;

(* "u:" + "\" OR "\*" *)

PROCEDURE GetCurrent (useLFN:BOOLEAN;
                     VAR  current:pathtype);
VAR
    rc:CARDINAL;
    longform:pathtype;
    drive:SHORTCARD;
    unit:str2;
BEGIN
    drive:=FIO.GetDrive();

    Str.Concat(unit, CHR( ORD("A")-1+ORD(drive) ),colon);

    FIO.GetDir(drive,current); (* \path without u: nor trailing \ except at root *)
    IF current[1] # colon THEN Str.Prepend(current,unit); END; (* safety *)
    IF useLFN THEN
        IF w9XshortToLong(current,rc,longform) THEN (* if error, keep DOS current *)
            Str.Copy(current,longform);
        END;
    END;
    (* LFN function seems to always return "u:\*" form except at root *)
    IF current[1] = colon THEN Str.Delete(current,0,2);END; (* safety *)
    fixDirectory(current);
    Str.Prepend(current,unit);
END GetCurrent;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE msg (S1,S2:ARRAY OF CHAR );
CONST
    wi = 1+19; (* was 16 before msgMaxDigits *)
BEGIN
    WrStr( pad(wi,TRUE,blank,S1)); WrStr(": ");WrStr(S2);WrLn;
END msg;

CONST
                 (* 1234567890123456789 *)
    msgBase      = "Base path";
    msgPos       = "Number position";
    msgStart     = ", prepended";
    msgEnd       = ", appended";
    msgInner     = ", first inner";
    msgAlone     = ", alone";
    msgLower     = "Lower index";
    msgUpper     = "Upper index";
    msgMaxDigits = "Maximum index width";
    msgMissing   = "Missing index";
    msgRange     = "Missing range";
    msgDuplicates= "Repeated index";
    strRange     = "[~..~]";
    strIndex     = " ~";
CONST
    msgNoMissing = "::: No missing index was found.";
    msgJustLimits= "::: Missing indexes will not be search for.";

PROCEDURE dmprange (VAR alcatraz:BOOLEAN; VAR chkrounds:CARDINAL;
                   lowerval,upperval:LONGCARD;
                   verbose,beautify,french,
                   useLFN,ignoreESC:BOOLEAN);
VAR
    i:LONGCARD;
    S:str80;
BEGIN
    alcatraz:=FALSE;
    IF ( (lowerval = upperval) OR verbose) THEN
        i:=lowerval;
        LOOP
            IF i > upperval THEN EXIT;END;
            S:=strIndex;
            Str.Subst(S,"~",num2str(i,wiglobal,TRUE,beautify,french,blank));
            msg(msgMissing,S);
            INC(i);
            alcatraz:= pollEscape(chkrounds,useLFN,ignoreESC);
            IF alcatraz THEN EXIT;END;
        END;
    ELSE
        S:=strRange;
        Str.Subst(S,"~",num2str(lowerval,wiglobal,TRUE,beautify,french,blank));
        Str.Subst(S,"~",num2str(upperval,wiglobal,TRUE,beautify,french,blank));
        msg(msgRange,S);
    END;
END dmprange;

PROCEDURE dmpduplicates (v,n:LONGCARD;beautify,french:BOOLEAN);
VAR
    S:str80;
BEGIN
    IF n > 1 THEN
        S:="Value ~ was found ~ times";
        Str.Subst(S,"~",num2str(v,wiglobal,FALSE,beautify,french,blank));
        Str.Subst(S,"~",num2str(n,wiglobal,FALSE,beautify,french,blank));
        msg(msgDuplicates,S);
    END;
END dmpduplicates;

(* ------------------------------------------------------------ *)

(* warn for DOS chars too ? bah... *)

PROCEDURE procPad (userwi:CARDINAL;
                  dorena,beautify,french,warnDOS:BOOLEAN;p:pFname);
CONST
    wiDOSname = 8;
    wiDOSext  = 1+3; (* dot is always here even WITH an empty extension *)
    sRemOld   = 'REM ::: OLD ::: "~"';
    sRemNew   = 'REM ::: NEW ::: "~"';
    sRemDOS   = 'REM --- DOS ! "~"';
    sCmd      = ' "~" "|"';
    sRename   = "ren";
    sCopy     = "copy";
VAR
    oldname,newname:pathtype;
    pb,posi,count:CARDINAL;
    v:LONGCARD;
    Z:str16;
    S:str1024; (* oversized *)
    u,d,n,e:pathtype;
BEGIN
    count:=CARDINAL(p^.winum);
    IF count < userwi THEN
        posi:=CARDINAL(p^.startnum);
        v:=p^.value;
        getStr(oldname,p);
        Str.Copy(newname,oldname);
        Str.Delete(newname,posi,count);
        Str.Insert(newname, num2str ( v,userwi,TRUE,beautify,french,"0"), posi);
        (*
        S:=sRemOld;
        Str.Subst(S,"~",oldname);
        WrStr(S);WrLn;
        S:=sRemNew;
        Str.Subst(S,"~",newname);
        WrStr(S);WrLn;
        *)

        IF warnDOS THEN
            qdSplitAllPath(newname, u,d,n,e);
            pb:=0;
            IF Str.Length(n) > wiDOSname THEN INC(pb);END;
            IF Str.Length(e) > wiDOSext THEN INC(pb);END;
            IF pb # 0 THEN
                S:=sRemDOS;
                Str.Subst(S,"~",newname);
                WrLn;
                WrStr(S);WrLn;
            END;
        END;

        S:=sCmd;
        Str.Subst(S,"~",oldname);
        Str.Subst(S,"|",newname);
        IF dorena THEN
            WrStr(sRename);
        ELSE
            WrStr(sCopy);
        END;
        WrStr(S);WrLn;
    END;
END procPad;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    UNDEFINED = MAX(CARDINAL);
    MAXIPAD   = SIZE(str16);
VAR
    i,opt,parmcount:CARDINAL;
    S,R,spec,basedir:pathtype;
    useLFN,DEBUG,showexcluded,ignoreESC,alcatraz,warnDOS:BOOLEAN;
    showheader,verbose,beautify,french,trim,warndups,justlimits,dorena:BOOLEAN;
    state:(waiting,gotspec);
    searchstate:(attente,inrange);
    numpos:setofnumpos;
    anchor,ptr:pFname;
    countFile:CARDINAL;
    v,vv,minval,maxval,lowerval,upperval,dumped:LONGCARD;
    chkrounds,maxdigits,userdigits:CARDINAL;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn;

    useLFN       := TRUE;
    numpos       := setofnumpos{};
    verbose      := FALSE;
    beautify     := FALSE;  (* //TODO *)
    french       := TRUE;   (* //TODO *)
    trim         := TRUE;
    warndups     := FALSE;
    state        := waiting;
    justlimits   := FALSE;
    userdigits   := UNDEFINED;
    dorena       := TRUE;
    showexcluded := TRUE;
    showheader   := TRUE;
    warnDOS      := FALSE;
    chkrounds    := 0; (* try and fix nasty Win9X behavior about ChkEscape *)
    ignoreESC    := FALSE;

    parmcount := Lib.ParamCount();
    IF parmcount=0 THEN abort(errHelp,"");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "DEBUG"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "P"+delim+"PREPENDED"+delim+
                                  "A"+delim+"APPENDED"+delim+
                                  "B"+delim+"BEAUTIFY"+delim+
                                  "U"+delim+"US"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "PP"+delim+"PV"+delim+"VP"+delim+
                                  "AA"+delim+"AV"+delim+"VA"+delim+
                                  "I"+delim+"INNER"+delim+
                                  "II"+delim+"IV"+delim+"VI"+delim+
                                  "T"+delim+"NOTRIM"+delim+
                                  "D"+delim+"DIGITS"+delim+"NUMBER"+delim+
                                  "DD"+delim+"DV"+delim+"VD"+delim+
                                  "K"+delim+"DUPS"+delim+
                                  "M"+delim+"LIMITS"+delim+"MINMAX"+delim+
                                  "W:"+delim+"WIDTH:"+delim+"PAD:"+delim+
                                  "WP:"+delim+
                                  "WA:"+delim+
                                  "WI:"+delim+
                                  "WD:"+delim+
                                  "C"+delim+"COPY"+delim+
                                  "LIST"+delim+
                                  "N"+delim+"HEADER"+delim+
                                  "WW:"+delim+
                                  "!"+delim+"ESC"+delim+"ESCAPE"+delim+
                                  "Z"+delim+"DOS"+delim+"DOSWARN"
                              );
            CASE opt OF
            | 1,2,3 :  abort(errHelp,"");
            | 4 :      DEBUG:=TRUE;
            | 5,6 :    useLFN := FALSE;
            | 7,8 :    newpos(numpos,startpos);
            | 9,10 :   newpos(numpos,endpos);
            | 11,12:   beautify:=TRUE;
            | 13,14:   french := FALSE;
            | 15,16:   verbose := TRUE;
            | 17,18,19:newpos(numpos,startpos);     verbose := TRUE;
            | 20,21,22:newpos(numpos,endpos);       verbose := TRUE;
            | 23,24 :  newpos(numpos,innerpos);
            | 25,26,27:newpos(numpos,innerpos);     verbose := TRUE;
            | 28,29:   trim:=FALSE;
            | 30,31,32:newpos(numpos,numberalone);
            | 33,34,35:newpos(numpos,numberalone);  verbose:=TRUE;
            | 36,37   :warndups := TRUE;
            | 38,39,40:justlimits:=TRUE;
            | 41,42,43:IF GetLongCard(S,v)=FALSE THEN abort(errVal,S);END;
                       IF v > MAXIPAD THEN abort(errValRange,S);END;
                       userdigits := CARDINAL(v);
            | 44:      IF GetLongCard(S,v)=FALSE THEN abort(errVal,S);END;
                       IF v > MAXIPAD THEN abort(errValRange,S);END;
                       userdigits := CARDINAL(v);
                       newpos(numpos,startpos);
            | 45:      IF GetLongCard(S,v)=FALSE THEN abort(errVal,S);END;
                       IF v > MAXIPAD THEN abort(errValRange,S);END;
                       userdigits := CARDINAL(v);
                       newpos(numpos,endpos);
            | 46:      IF GetLongCard(S,v)=FALSE THEN abort(errVal,S);END;
                       IF v > MAXIPAD THEN abort(errValRange,S);END;
                       userdigits := CARDINAL(v);
                       newpos(numpos,innerpos);
            | 47:      IF GetLongCard(S,v)=FALSE THEN abort(errVal,S);END;
                       IF v > MAXIPAD THEN abort(errValRange,S);END;
                       userdigits := CARDINAL(v);
                       newpos(numpos,numberalone);
            | 48,49:   dorena:=FALSE;
            | 50:      DEBUG := TRUE;
                       showexcluded := FALSE;
            | 51,52:   showheader := FALSE;
            | 53:      IF GetLongCard(S,v)=FALSE THEN abort(errVal,S);END;
                       IF v > MAXIPAD THEN abort(errValRange,S);END;
                       userdigits := CARDINAL(v);
                       showheader := FALSE;
            | 54,55,56:ignoreESC  := TRUE;
            | 57,58,59:warnDOS    := TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            CASE state OF
            | waiting : Str.Copy(spec,S);
            | gotspec : abort(errTooManyParms,S);
            END;
            INC(state);
        END;
    END;

    (* syntax checks *)

    IF state = waiting THEN abort(errExpecting,"<specification>");END;
    IF numpos = setofnumpos{} THEN abort(errExpectingCmd,"");END; (* use a default ? *)

    IF NOT(showheader) THEN
        IF justlimits THEN abort(errIllogical,"");END; (* hide header and show data is a nono *)
        IF userdigits = UNDEFINED THEN abort(errRequired,"");END; (* hide header requires -w[w]:# *)
    END;
    IF userdigits # UNDEFINED THEN
        IF chkoneifhere(numpos,startpos)=FALSE THEN abort(errHighlanderTheSeries,"");END;
        IF chkoneifhere(numpos,endpos)=FALSE THEN abort(errHighlanderTheSeries,"");END;
        IF chkoneifhere(numpos,innerpos)=FALSE THEN abort(errHighlanderTheSeries,"");END;
        IF chkoneifhere(numpos,numberalone)=FALSE THEN abort(errHighlanderTheSeries,"");END;
    END;

    useLFN := ( useLFN AND w9XsupportLFN() ); (* here ! *)

    IF NOT(useLFN) THEN warnDOS := TRUE; END; (* force it on a DOS filesystem *)

    IF same(spec,".") THEN spec := stardotstar;END;
    IF Str.Match(spec,"*\") THEN Str.Append(spec,stardotstar); END;
    IF Str.Match(spec,"*\.") THEN
        i:=Str.Length(spec);
        spec[i-1]:=nullchar;
        Str.Append(spec,stardotstar);
    END;
    IF Str.RCharPos(spec,dot)=MAX(CARDINAL) THEN Str.Append(spec,dot+star);END;

    buildPath(basedir,  spec);
    IF same(basedir,"") THEN
        GetCurrent(useLFN, basedir);
    END;

    initList(anchor);
    countFile:=buildMatchList (anchor,DEBUG,showexcluded,useLFN,trim,
                              numpos,spec);
    IF countFile=MAX(CARDINAL) THEN abort(errTooManyFiles,spec);END; (* errStorage *)
    IF countFile=0 THEN abort(errNoMatch,spec); END;

    (* find smallest AND highest *)

    FOR i:=1 TO 2 DO
        CASE i OF
        | 1: v:=MAX(LONGCARD);
        | 2: v:=MIN(LONGCARD);
        END;
        ptr:=anchor;
        WHILE ptr # NIL DO
            vv:=ptr^.value;
            CASE i OF
            | 1: IF vv < v THEN v:= vv; minval:=v;END;
            | 2: IF vv > v THEN v:= vv; maxval:=v;END;
            END;
            ptr:=ptr^.next;
        END;
    END;

    (* find maxdigits width *)

    maxdigits:=MIN(CARDINAL);
    ptr:=anchor;
    WHILE ptr # NIL DO
        i:=CARDINAL(ptr^.winum);
        IF i > maxdigits THEN maxdigits:=i;END;
        ptr:=ptr^.next;
    END;

    (* header *)

    IF showheader THEN

        msg(msgBase,basedir);

        S:="";
        IF startpos    IN numpos THEN Str.Append(S,msgStart);END;
        IF endpos      IN numpos THEN Str.Append(S,msgEnd);END;
        IF innerpos    IN numpos THEN Str.Append(S,msgInner);END;
        IF numberalone IN numpos THEN Str.Append(S,msgAlone);END;
        Str.Subst(S, ", " , "");
        i:=Str.RCharPos(S,",");
        IF i # MAX(CARDINAL) THEN
            Str.Delete(S,i,1);
            Str.Insert(S," or",i);
        END;
        msg(msgPos,S);
        WrLn;

        IF justlimits THEN
            WrStr(msgJustLimits);WrLn;
            WrLn;
        END;

        S:=strIndex;
        Str.Subst(S,"~",num2str(minval,wiglobal,TRUE,beautify,french,blank));
        msg(msgLower,S);
        S:=strIndex;
        Str.Subst(S,"~",num2str(maxval,wiglobal,TRUE,beautify,french,blank));
        msg(msgUpper,S);
        WrLn;
        S:=strIndex;
        Str.Subst(S,"~",num2str(LONGCARD(maxdigits),wiglobal,TRUE,beautify,french,blank));
        msg(msgMaxDigits,S);
    END;

    (* ------------------------------------------------------------ *)
    (* -m *)

    IF justlimits THEN
        freeList(anchor);
        abort(errNone,"");
    END;

    WrLn;

    (* ------------------------------------------------------------ *)
    (* -w[w]:# *)

    IF userdigits # UNDEFINED THEN
        IF userdigits < maxdigits THEN abort(errPadTooSmall,"");END;

        ptr:=anchor;
        WHILE ptr # NIL DO
            procPad(userdigits,dorena,beautify,french,warnDOS,ptr);
            ptr:=ptr^.next;
            IF pollEscape(chkrounds,useLFN,ignoreESC) THEN
                freeList(anchor);
                abort(errAbortedByUser,"");
            END;
        END;

        freeList(anchor);
        abort(errNone,"");
    END;

    (* ------------------------------------------------------------ *)
    (* list *)

    dumped:=0;
    searchstate:=attente;
    v:=minval;
    LOOP
        IF v > maxval THEN EXIT;END;
        IF foundvalue(v,anchor) THEN
            IF warndups THEN
                vv:=countme(v,anchor);
                dmpduplicates(v,vv,beautify,french);
            END;
            CASE searchstate OF
            | attente:
                ;
            | inrange:
                dmprange (alcatraz,chkrounds,
                         lowerval,upperval,
                         verbose,beautify,french,
                         useLFN,ignoreESC);
                IF alcatraz THEN
                    freeList(anchor);
                    abort(errAbortedByUser,"");
                END;
                INC(dumped);
                searchstate:=attente;
            END;
        ELSE
            CASE searchstate OF
            | attente:
                lowerval:=v;
                upperval:=v;
                searchstate:=inrange;
            | inrange:
                upperval:=v;
            END;
        END;
        INC(v);
        IF pollEscape(chkrounds,useLFN,ignoreESC) THEN
            freeList(anchor);
            abort(errAbortedByUser,"");
        END;
    END;
    CASE searchstate OF
    | attente:
        ;
    | inrange:
        dmprange (alcatraz,chkrounds,
                 lowerval,upperval,
                 verbose,beautify,french,
                 useLFN,ignoreESC);
        IF alcatraz THEN
            freeList(anchor);
            abort(errAbortedByUser,"");
        END;
        INC(dumped);
        searchstate:=attente; (* useless ! *)
    END;

    IF dumped = 0 THEN
        WrStr(msgNoMissing);WrLn;
    END;

    freeList(anchor);

    abort(errNone,"");
END Missing.



