(* ---------------------------------------------------------------
Title         Q&D Now Plus
Overview      see help
Notes         see help
Bugs
Wish List     -us could assume MDY date format instead of current DMY

--------------------------------------------------------------- *)

MODULE NowPlus;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetString, CharCount, same,
aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything,str16, cleantabs;

IMPORT SYSTEM;
IMPORT Lib;
IMPORT Str;
IMPORT IO;
IMPORT MATHLIB;

FROM IO IMPORT WrStr,WrLn;

(* ------------------------------------------------------------ *)

TYPE
    languagetype = (french,english);
CONST
    joke          = "Vero Biddeuse";
    defaultNowFmtFR  = "  $dd !d $mm $yyyy  %hh %mmn";
    defaultNowFmtUS  = "  $dd, $mm $d, $yyyy at %0h:%0m";
    defaultPlusFR    = "+ $d, $h et $m";
    defaultPlusUS    = "+ $d, $h and $m";
    defaultNxtFmtFR  = "= $dd !d $mm $yyyy  %hh %mmn";
    defaultNxtFmtUS  = "= $dd, $mm $d, $yyyy at %0h:%0m";
    nl      = CHR(13) + CHR(10);
    tab     = CHR(9);
    dash    = "-";
    slash   = "/";
    colon   = ":";
CONST
    ProgEXEname   = "NOWPLUS";
    ProgTitle     = "Q&D Now Plus";
    ProgVersion   = "v1.0b";
    ProgCopyright = "by PhG";
CONST
    errNone         = 0;
    errHelp         = 1;
    errUnknownOpt   = 2;
    errTooManyParms = 3;
    errDD           = 4;
    errHH           = 5;
    errMM           = 6;
    errMissing      = 7;
    errDosVersion   = 8;
    errSyntax       = 9;
    errDate         = 10;
    errTime         = 11;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
(*
 0        1         2         3         4         5         6         7         8
 12345678901234567890123456789012345678901234567890123456789012345678901234567890
*)
CONST
    helpmsg =
ProgTitle+" "+ProgVersion+" "+ProgCopyright+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <days> <hours> <minutes> [-us]"+nl+
nl+
"    Add specified days, hours and minutes to current date and time."+nl+
nl+
"Syntax 2 : "+ProgEXEname+" <date> <time> <days> <hours> <minutes> [-us]"+nl+
nl+
"    Add specified days, hours and minutes to specified date and time."+nl+
"    Date separator is '"+dash+"' or '"+slash+"', and 'yyyy' can be shortened to 'yy'."+nl+
"    Time separator is '"+colon+"', and seconds can be omitted."+nl+
nl+
"This program was written for eBayWatcher "+joke+" :"+nl+
"it may be useful for other eBay customers."+nl+
nl+
"a) Please note U.T. (Universal Time) is assumed."+nl+
"b) Syntax 2 assumes French for D/M/Y format date parsing and messages :"+nl+
"   -us option forces American M/D/Y format date parsing and messages."+nl+
"   1900 is added to Year if specified without century (i.e. [0..99])."+nl+
nl+
"Examples : "+ProgEXEname+" 4 1 0"+nl+
"           "+ProgEXEname+" 1/2/63 19h45 12053 0 0"+nl+
"           "+ProgEXEname+" -us feb-1-1963 19:45:00 12053 0 0"+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(helpmsg);
    | errUnknownOpt :
        Str.Concat(S,"Illegal ",einfo); Str.Append(S," option !");
    | errTooManyParms :
        Str.Concat(S,einfo," is just one parameter too many !");
    | errDD:
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," <days> parameter !");
    | errHH:
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," <hours> parameter !");
    | errMM:
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," <minutes> parameter !");
    | errMissing:
        Str.Concat(S,"Missing ",einfo);Str.Append(S," !");
    | errDosVersion:
        Str.Concat(S,"DOS ",einfo);Str.Append(S," or later is required !");
    | errSyntax:
        S := "Illegal number of parameters !";
    | errDate:
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," date !");
    | errTime:
        Str.Concat(S,"Illegal ",einfo);Str.Append(S," time !");
    ELSE
        S := "This is illogical, Captain !!!";
    END;
    CASE  e OF
    | errNone, errHelp : ;
    ELSE
        WrStr(ProgEXEname+" : ");
        WrStr(S);
        WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

PROCEDURE AtLeastDosVersion (minmajor,minminor:CARDINAL) : BOOLEAN;
VAR
    R             : SYSTEM.Registers;
    minDosVersion : CARDINAL;
    major         : CARDINAL;
    minor         : CARDINAL;
    thisDosVersion: CARDINAL;
BEGIN
    minDosVersion := (minmajor << 8) + minminor;
    R.AX := 3000H;
    Lib.Dos(R);
    major := CARDINAL(R.AL);
    minor := CARDINAL(R.AH);
    thisDosVersion := (major << 8) + minor;
    IF thisDosVersion < minDosVersion THEN RETURN FALSE; END;
    RETURN TRUE;
END AtLeastDosVersion;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* ---------------------------------------------------------------
   input  : value
   output : int(value) ie int(0.95) gives 0 (no rounding)

   TRUNC(n) gives only a CARDINAL without rounding either
--------------------------------------------------------------- *)

PROCEDURE int (v : LONGREAL) : LONGREAL;
VAR
    tmp : LONGINT; (* this should be enough! *)
BEGIN
    tmp := VAL(LONGINT,v); (* as intended, does NOT perform any rounding *)
    RETURN VAL(LONGREAL,tmp);
END int;

(* ---------------------------------------------------------------
   input  : value
   output : frac(value)

   frac keeps sign
--------------------------------------------------------------- *)

PROCEDURE frac (v : LONGREAL) : LONGREAL;
BEGIN
    RETURN (v - int (v));
END frac;

(* ---------------------------------------------------------------
   input  : value1, value2
   output : value1 mod value2

   mod keeps sign
--------------------------------------------------------------- *)

PROCEDURE mod (v1, v2 : LONGREAL) : LONGREAL;
BEGIN
    RETURN ( MATHLIB.Mod(v1, v2) );
END mod;

(* ---------------------------------------------------------------
   assume day with decimal, and whole date is ok
   valid only for JD >= 0 i.e. NOT before 1.5 Jan -4712 !
   input  : day (with decimals), month, year
   output : JD
--------------------------------------------------------------- *)

PROCEDURE DateToJD (Day,Month,Year:LONGREAL) : LONGREAL;
VAR
    YYYYMMDD  : LONGREAL;
    A,gregorian : LONGREAL; (* gregorian correction *)
    JD        : LONGREAL;
BEGIN
    YYYYMMDD := Year * 10000.0 + Month * 100.0 + Day;
    IF Month < 2.5 THEN
        Year  := Year - 1.0;
        Month := Month + 12.0;
    END;
    IF YYYYMMDD < 15821015.0 THEN (* before 15 october 1582, julian calendar *)
        gregorian := 0.0;
    ELSE
        A         := int (Year / 100.0);
        gregorian := 2.0 - A + int (A / 4.0);
    END;
    JD := int (365.25 * (Year + 4716.0)) + int (30.6001 * (Month + 1.0));
    JD := JD + Day + gregorian - 1524.5;
    RETURN JD;
END DateToJD;

(* ---------------------------------------------------------------
   assume JD >= 0
   input  : JD
   output : day (with decimals), month, year
--------------------------------------------------------------- *)

PROCEDURE JDtoDate (JD : LONGREAL; VAR Day,Month,Year : LONGREAL);
VAR
    Z,F,A,alf,B,C,D,E:LONGREAL;
BEGIN
    JD := JD + 0.5;
    Z := int(JD);
    F := frac(JD);
    IF Z < 2299161.0 THEN
        A := Z;
    ELSE
        alf := int( (Z - 1867216.25) / 36524.25);
        A := Z + 1.0 + alf - int( alf / 4.0 );
    END;
    B := A + 1524.0;
    C := int ( (B - 122.1) / 365.25);
    D := int ( 365.25 * C );
    E := int ( (B - D) / 30.6001 );
    Day := B - D - int( 30.6001 * E) + F;
    IF E < 13.5 THEN
        Month := E - 1.0;
    ELSE
        Month := E - 13.0;
    END;
    IF Month > 2.5 THEN
        Year := C - 4716.0;
    ELSE
        Year := C - 4715.0;
    END;
END JDtoDate;

(* ---------------------------------------------------------------
   find JD at 0h
   input  : JD
   output : JD for this day at 0H
--------------------------------------------------------------- *)

PROCEDURE JDtoJD0H (JD : LONGREAL) : LONGREAL;
VAR
    JD0H : LONGREAL;
BEGIN
    IF frac(JD) < 0.5 THEN
       JD0H := int(JD) - 0.5; (* was < 23H59 so take previous 0H *)
    ELSE
       JD0H := int(JD) + 0.5; (* was 0H or later so take current 0H *)
    END;
    RETURN JD0H;
END JDtoJD0H;

(* ---------------------------------------------------------------
   0=sunday, 1=monday... 6=saturday
   input  : JD
   output : day of week
--------------------------------------------------------------- *)

PROCEDURE JDtoWeekday (JD : LONGREAL) : CARDINAL;
VAR
    v : LONGREAL;
BEGIN
    v := mod ( (JDtoJD0H (JD) + 1.5), 7.0 );
    RETURN ( VAL(CARDINAL, v ) );
END JDtoWeekday;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    hoursPerDay = 24.0;
    minutesPerHour = 60.0;
    secondsPerMinute=60.0;

PROCEDURE makeMyDay (dd,hh,mm,ss:CARDINAL ) : LONGREAL;
VAR
    k : LONGREAL;
BEGIN
    k := LONGREAL(mm) + LONGREAL(ss) / secondsPerMinute;
    k := LONGREAL(hh) + LONGREAL(k) / minutesPerHour;
    k := LONGREAL(dd) + LONGREAL(k) / hoursPerDay;
    RETURN k;
END makeMyDay;

PROCEDURE unmakeMyDay(fracday:LONGREAL; VAR dd,hh,mm,ss : CARDINAL);
VAR
    day : LONGREAL;
    seconds : LONGCARD; (* 24*60*60=86400 so CARDINAL is not enough *)
BEGIN
    day := int(fracday);
    dd  := CARDINAL(day);
    day := frac(fracday);
    seconds := VAL(LONGCARD, (day * hoursPerDay * minutesPerHour * secondsPerMinute + 0.5) );
    hh := CARDINAL(seconds DIV 3600); (* 60*60 *)
    mm := CARDINAL( (seconds MOD 3600) DIV 60);
    ss := CARDINAL(seconds MOD 60);
    IF hh > 23 THEN (* if this happens... do not bother too much ! *)
        hh := 23;
        mm := 59;
        ss := 59;
    END;
END unmakeMyDay;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    datetype = RECORD
        day   : CARDINAL;
        month : CARDINAL;
        year  : CARDINAL;
        dayOfWeek  : Lib.DayType;
    END;
    timetype = RECORD
        hours   : CARDINAL;
        minutes : CARDINAL;
        seconds : CARDINAL;
    END;

(* ------------------------------------------------------------ *)

CONST
    alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    digits = "0123456789";

CONST
    century = 1900;

PROCEDURE parseDateFull (language:languagetype; S : ARRAY OF CHAR;
                         VAR date : datetype) : BOOLEAN;
CONST
    separator=dash;
    legaldateset = digits+separator+alphabet;
    mindd=1;
    maxdd=31;
    minmm=1;
    maxmm=12;
    minyy=1600;    (* ephemeris range is 1600..2400 *)
    maxyy=2400-1;
VAR
    i : CARDINAL;
    R : str80;
    v : LONGCARD;
    ok: BOOLEAN;
    jd: LONGREAL;
    dow:Lib.DayType;
    ndxday,ndxmonth:CARDINAL;
BEGIN
    UpperCase(S); (* in case months would be letters *)
    ReplaceChar(S,slash,separator);
    FOR i := 0 TO (Str.Length(S)-1) DO
        IF Str.CharPos(legaldateset,S[i])=MAX(CARDINAL) THEN RETURN FALSE; END;
    END;
    IF CharCount(S,separator) # 2 THEN RETURN FALSE; END;

    CASE language OF
    | french: ndxday := 0; ndxmonth := 1;
    | english:ndxday := 1; ndxmonth := 0;
    END;

    Str.ItemS(R,S,separator,ndxday);
    v := Str.StrToCard(R,10,ok);
    IF ok=FALSE THEN RETURN FALSE; END;
    IF (v < mindd) OR (v > maxdd) THEN RETURN FALSE; END;
    date.day := CARDINAL(v);

    Str.ItemS(R,S,separator,ndxmonth);
    v := Str.StrToCard(R,10,ok);
    IF ok=FALSE THEN
        Str.Prepend(R,"-"); (* fake command line parameter ! *)
        i := GetOptIndex(R,"JAN"+delim+"JAN"+delim+
                           "FEB"+delim+"FEV"+delim+
                           "MAR"+delim+"MAR"+delim+
                           "APR"+delim+"AVR"+delim+
                           "MAY"+delim+"MAI"+delim+
                           "JUN"+delim+"JUN"+delim+
                           "JUL"+delim+"JUI"+delim+
                           "AUG"+delim+"AOU"+delim+
                           "SEP"+delim+"SEP"+delim+
                           "OCT"+delim+"OCT"+delim+
                           "NOV"+delim+"NOV"+delim+
                           "DEC"+delim+"DEC");
        CASE i OF
        | 1..24 :
            v := LONGCARD(i+1) DIV 2;
        ELSE
            RETURN FALSE;
        END;
    END;
    IF (v < minmm) OR (v > maxmm) THEN RETURN FALSE; END;
    date.month := CARDINAL(v);

    Str.ItemS(R,S,separator,2); (* Y here *)
    v := Str.StrToCard(R,10,ok);
    IF ok=FALSE THEN RETURN FALSE; END;
    IF v < 100 THEN INC(v,century); END;
    IF (v < minyy) OR (v > maxyy) THEN RETURN FALSE; END;
    date.year := CARDINAL(v);

    (* do not forget to fill day of week ! *)

    jd:=DateToJD (LONGREAL(date.day),LONGREAL(date.month),LONGREAL(date.year));
    CASE JDtoWeekday (jd) OF (* 0=sunday, 1=monday... 6=saturday *)
    | 0: dow:=Lib.Sunday;
    | 1: dow:=Lib.Monday;
    | 2: dow:=Lib.Tuesday;
    | 3: dow:=Lib.Wednesday;
    | 4: dow:=Lib.Thursday;
    | 5: dow:=Lib.Friday;
    | 6: dow:=Lib.Saturday;
    END;
    date.dayOfWeek:=dow;

    RETURN TRUE;
END parseDateFull;

PROCEDURE parseTimeHMS (S : ARRAY OF CHAR;
                        VAR time : timetype) : BOOLEAN;
CONST
    separator=colon;
    separatoralt="H";
    legaltimeset = digits+separator+separatoralt;
    minhh=0;
    maxhh=23;
    minmm=0;
    maxmm=59;
    minss=0;
    maxss=59;
VAR
    i : CARDINAL;
    R : str80;
    v : LONGCARD;
    ok: BOOLEAN;
BEGIN
    FOR i := 0 TO (Str.Length(S)-1) DO
        IF Str.CharPos(legaltimeset,S[i])=MAX(CARDINAL) THEN RETURN FALSE; END;
    END;
    LOOP
        IF Str.CharPos(S,separatoralt) = MAX(CARDINAL) THEN EXIT; END;
        Str.Subst(S,separatoralt,separator);
    END;
    i := CharCount(S,separator);
    IF (i # 1) AND (i # 2) THEN RETURN FALSE; END;

    Str.ItemS(R,S,separator,0);
    v := Str.StrToCard(R,10,ok);
    IF ok=FALSE THEN RETURN FALSE; END;
    IF (v < minhh) OR (v > maxhh) THEN RETURN FALSE; END;
    time.hours := CARDINAL(v);

    Str.ItemS(R,S,separator,1);
    v := Str.StrToCard(R,10,ok);
    IF ok=FALSE THEN RETURN FALSE; END;
    IF (v < minmm) OR (v > maxmm) THEN RETURN FALSE; END;
    time.minutes := CARDINAL(v);

    CASE i OF
    | 1 :
        time.seconds := minss;
    | 2 :
        Str.ItemS(R,S,separator,2);
        v := Str.StrToCard(R,10,ok);
        IF ok=FALSE THEN RETURN FALSE; END;
        IF (v < minss) OR (v > maxss) THEN RETURN FALSE; END;
        time.seconds := CARDINAL(v);
    END;

    RETURN TRUE;
END parseTimeHMS;

(* ------------------------------------------------------------ *)

PROCEDURE getDateNow (VAR d : datetype);
BEGIN
    Lib.GetDate(d.year,d.month,d.day,d.dayOfWeek);
END getDateNow;

PROCEDURE getTimeNow (VAR t : timetype);
VAR
    hundredth : CARDINAL;
BEGIN
    Lib.GetTime(t.hours,t.minutes,t.seconds,hundredth);
END getTimeNow;

PROCEDURE substring (S1,S2:ARRAY OF CHAR) : BOOLEAN;
BEGIN
    Str.Caps(S2);
    IF Str.Pos(S1,S2) # MAX(CARDINAL) THEN RETURN TRUE; END;
    Str.Lows(S2);
    IF Str.Pos(S1,S2) # MAX(CARDINAL) THEN RETURN TRUE; END;
    RETURN FALSE;
END substring;

PROCEDURE newstring (VAR R:ARRAY OF CHAR; old,new:ARRAY OF CHAR);
BEGIN
    Str.Caps(old);
    IF Str.Pos(R,old) # MAX(CARDINAL) THEN
        Str.Subst(R,old,new);
    ELSE
        Str.Lows(old);
        IF Str.Pos(R,old) = MAX(CARDINAL) THEN RETURN; END;
        Str.Subst(R,old,new);
    END;
END newstring;

PROCEDURE oldnew (sOld,sNew:ARRAY OF CHAR;VAR R:ARRAY OF CHAR);
BEGIN
    LOOP
        IF substring(R,sOld)=FALSE THEN EXIT;END;
        newstring(R,sOld,sNew);
    END;
END oldnew;

PROCEDURE using (n : CARDINAL; digits : CARDINAL; pad : CHAR) : str80;
VAR
    ok   : BOOLEAN;
    v    : LONGCARD;
    len  : CARDINAL;
    S    : str80;
BEGIN
    v := LONGCARD(n);
    Str.CardToStr(v,S,10,ok);
    IF pad="" THEN RETURN S; END;
    len := Str.Length(S);
    LOOP
        IF Str.Length(S) >= digits THEN EXIT; END;
        Str.Prepend(S,pad);
    END;
    RETURN S;
END using;

(*
    $dd     lundi..dimanche / monday..sunday
    $d      1..31
    $0d     01..31
    $mm     janvier..dcembre / january..december
    $m      1..12
    $0m     01..12

    $yyyy   ####
    $yy     00..99

    !d      1er..31
    !0d     1er..31

    %h      0..23
    %0h     00..23
    %m      0..59
    %0m     00..59
    %s      0..59
    %0s     00..59
*)

PROCEDURE formatDT(format:ARRAY OF CHAR;language:languagetype;
                   dmy:datetype;hms:timetype;
                   VAR R : ARRAY OF CHAR  );
CONST
    jours = "dimanche lundi mardi mercredi jeudi vendredi samedi";
    mois  = "janvier fvrier mars avril mai juin "+
            "juillet aot septembre octobre novembre dcembre";
    joursUS = "Sunday Monday Tuesday Wednesday Thursday Friday Saturday";
    moisUS  = "January February March April May June "+
              "July August September October November December";
VAR
    S,T    : str16;
    pad    : CHAR;
    v      : CARDINAL;
BEGIN
    Str.Copy(R,format);

    v := ORD(dmy.dayOfWeek);
    CASE language OF
    | french:   Str.ItemS(T,jours," ",v);
    | english:  Str.ItemS(T,joursUS," ",v);
    END;
    oldnew("$dd",T,R);

    v:=dmy.day;
    pad := "";   oldnew("$d",    using(v,2,pad), R);
    pad := "0";  oldnew("$0d",   using(v,2,pad), R);

    IF v=1 THEN
    pad := "";   oldnew("!d",    "1er"         , R);
    pad := "0";  oldnew("!0d",   "1er"         , R);
    ELSE
    pad := "";   oldnew("!d",    using(v,2,pad), R);
    pad := "0";  oldnew("!0d",   using(v,2,pad), R);
    END;

    v := dmy.month-1;
    CASE language OF
    | french:   Str.ItemS(T,mois," ",v);
    | english:  Str.ItemS(T,moisUS," ",v);
    END;
    oldnew("$mm",T,R);

    v := dmy.month;
    pad := "";  oldnew("$m",    using(v,2,pad), R);
    pad := "0"; oldnew("$0m",   using(v,2,pad), R);

    v := dmy.year;
    pad := "";  oldnew("$yyyy", using(v,4,pad), R);
    pad := "0";
    IF v >= 1900 THEN (* always here *)
        v  := dmy.year - 1900;IF v > 100 THEN DEC(v,100);END;
    END;
                oldnew("$yy",   using(v,2,pad), R);

    v := hms.hours;
    pad := "";   oldnew("%h",    using(v,2,pad), R);
    pad := "0";  oldnew("%0h",   using(v,2,pad), R);

    v := hms.minutes;
    pad := "";   oldnew("%m",    using(v,2,pad), R);
    pad := "0";  oldnew("%0m",   using(v,2,pad), R);

    v := hms.seconds;
    pad := "";   oldnew("%s",    using(v,2,pad), R);
    pad := "0";  oldnew("%0s",   using(v,2,pad), R);

END formatDT;

(*
    $d      jour(s)
    $h      heure(s)
    $m      minute(s)
*)

PROCEDURE formatDeltaDHM (format:ARRAY OF CHAR;language:languagetype;
                     dd,hh,mm:CARDINAL; VAR R : ARRAY OF CHAR  );
VAR
    S      : str80;
    pad    : CHAR;
    v      : CARDINAL;
    sD,sH,sM:str16;
BEGIN
    CASE language OF
    |french:
        Str.Copy(sD," jour");
        Str.Copy(sH," heure");
        Str.Copy(sM," minute");
    |english:
        Str.Copy(sD," day");
        Str.Copy(sH," hour");
        Str.Copy(sM," minute");
    END;
    Str.Copy(R,format);

    v := dd;
    pad := "";
    Str.Concat(S, using(v,2,pad), sD); IF v>1 THEN Str.Append(S,"s");END;
    oldnew("$d",S,R);
    v := hh;
    pad := "";
    Str.Concat(S, using(v,2,pad), sH); IF v>1 THEN Str.Append(S,"s");END;
    oldnew("$h",S,R);
    v := mm;
    pad := "";
    Str.Concat(S, using(v,2,pad), sM); IF v>1 THEN Str.Append(S,"s");END;
    oldnew("$m",S,R);
END formatDeltaDHM;

PROCEDURE plusnow (VAR dmy:datetype;VAR hms:timetype;
                   dd,hh,mm:CARDINAL );
VAR
    jdorg,plusdays,newjd,d,m,y:LONGREAL;
BEGIN
    d := makeMyDay (CARDINAL(dmy.day),
                    CARDINAL(hms.hours),
                    CARDINAL(hms.minutes),
                    0);
    m := LONGREAL(dmy.month);
    y := LONGREAL(dmy.year);
    jdorg := DateToJD(d,m,y);
    plusdays:= makeMyDay(dd,hh,mm,0);
    newjd := jdorg + plusdays;
    JDtoDate (newjd, d,m,y);
    dmy.month := CARDINAL(m);
    dmy.year  := CARDINAL(y);
    unmakeMyDay(d, dmy.day, hms.hours, hms.minutes, hms.seconds);
    dmy.dayOfWeek := Lib.DayType ( JDtoWeekday (newjd) );
END plusnow;

PROCEDURE getNumber (S:ARRAY OF CHAR;upper:CARDINAL;VAR n:CARDINAL):BOOLEAN ;
VAR
    v:LONGCARD;
    ok:BOOLEAN;
BEGIN
    v:=Str.StrToCard(S,10,ok);
    IF ok=FALSE THEN RETURN FALSE;END;
    IF v > MAX(CARDINAL) THEN RETURN FALSE;END;
    n:=CARDINAL(v);
    RETURN (n <= upper);
END getNumber;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    maxdd = MAX(CARDINAL)-1; (* was 366, should do ! *)
    maxhh = 24-1;
    maxmm = 60-1;
    firststr = 1; (* eh, we'll init with -1 *)
    maxstr   = 5;
VAR
    str : ARRAY [firststr..maxstr] OF str128;
    laststr:CARDINAL;
    i,opt,parmcount:CARDINAL;
    S,R,format:str128;
    language:languagetype;
    dd,hh,mm:CARDINAL;
    dmy,odmy : datetype;
    hms,ohms : timetype;
BEGIN
    Lib.DisableBreakCheck();
    WrLn;
    language := french;
    dd := 0;
    hh := 0;
    mm := 0;
    laststr:=firststr-1;
    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,""); END;
    IF AtLeastDosVersion(3,20)=FALSE THEN abort(errDosVersion,"3.20");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "US"+delim+"UK"+delim+"ENGLISH"+delim+"E"
                               );
            CASE opt OF
            | 1,2,3   : abort(errHelp,"");
            | 4,5,6,7 : language := english;
            ELSE
                abort(errUnknownOpt,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            IF same(R,"?") THEN abort(errHelp,"");END;
            INC(laststr);
            IF laststr > maxstr THEN abort(errTooManyParms,S);END;
            Str.Copy( str[laststr], R);
        END;
    END;
    i:=firststr;
    CASE laststr OF
    | 3:
        IF getNumber(str[i  ],maxdd,dd)=FALSE THEN abort(errDD,str[i  ]);END;
        IF getNumber(str[i+1],maxhh,hh)=FALSE THEN abort(errHH,str[i+1]);END;
        IF getNumber(str[i+2],maxmm,mm)=FALSE THEN abort(errMM,str[i+2]);END;
    | 5:
        IF parseDateFull(language,
                         str[i  ],dmy)=FALSE THEN abort(errDate,str[i  ]);END;
        IF parseTimeHMS (str[i+1],hms)=FALSE THEN abort(errTime,str[i+1]);END;

        IF getNumber(str[i+2],maxdd,dd)=FALSE THEN abort(errDD,str[i+2]);END;
        IF getNumber(str[i+3],maxhh,hh)=FALSE THEN abort(errHH,str[i+3]);END;
        IF getNumber(str[i+4],maxmm,mm)=FALSE THEN abort(errMM,str[i+4]);END;
    ELSE
        abort(errSyntax,"");
    END;

    CASE language OF
    | french  : Str.Copy(format,defaultNowFmtFR);
    | english : Str.Copy(format,defaultNowFmtUS);
    END;

    CASE laststr OF
    | 3:
        getDateNow(dmy);
        getTimeNow(hms); (* let's hope we don't cross midnight ! ;-) maybe we should reread date ! *)
        odmy:=dmy;
        ohms:=hms;
    END;

    formatDT(format,language,dmy,hms,S);
    WrStr(S);WrLn;
    WrLn;

    CASE language OF
    | french  : Str.Copy(format,defaultPlusFR);
    | english : Str.Copy(format,defaultPlusUS);
    END;
    formatDeltaDHM(format,language,dd,hh,mm,S);
    WrStr(S);WrLn;
    WrLn;

    plusnow(dmy,hms,dd,hh,mm);
    CASE language OF
    | french  : Str.Copy(format,defaultNxtFmtFR);
    | english : Str.Copy(format,defaultNxtFmtUS);
    END;
    formatDT(format,language,dmy,hms,S);
    WrStr(S);WrLn;

    abort(errNone,"");
END NowPlus.

