(* -----------------------------------------------------------

Notes     : small model cannot pass str1024 strings
Wish list :

------------------------------------------------------------ *)

MODULE opGroup;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;
IMPORT IO;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    prefix  = "FCOMP"+blank;
    prefixbatch= "FCOMP -i"+blank;
    testbatch  = "if not errorlevel 255 MPAUSE";

    rembase = "REM"+blank;
    remark  = "rem"+blank;
TYPE
    hugestr = str1024;
CONST
    errNone             = 0;
    errHelp             = 1;
    errOpt              = 2;
    errParm             = 3;
    errExpecting        = 4;
    errNotFound         = 5;
    errJoker            = 6;
    errOp               = 7;
    errUnexpected       = 8;

CONST
    exe     = "OPGROUP";
    version = "v1.0h";
    msgHelp =
"Q&D Operation on groups of lines "+version+" by PhG"+nl+
nl+
"Syntax 1 : "+exe+" <source> [-g] [option]..."+nl+
"Syntax 2 : "+exe+" <source> <-n> [-k]"+nl+
"Syntax 3 : "+exe+" <source> <string> [-z] [-m] [-x]"+nl+
"Syntax 4 : "+exe+" <source> <-p[p]> [-i] [-c[c]] [-u] [-f]"+nl+
"Syntax 5 : "+exe+" <source> <-s>"+nl+
nl+
"This program glues groups of lines for later processing with FCOMP (syntax 1),"+nl+
"inserts an empty line in order to separate groups of lines (syntax 2),"+nl+
"filters out groups not containing <string> (syntax 3),"+nl+
"packs trimmed identical lines keeping one entry prefixed with count"+nl+
"provided <source> has already been sorted (syntax 4),"+nl+
"or swaps first and last lines in DUPLINES generated groups (syntax 5)."+nl+
nl+
"    -g    glue groups of lines"+nl+
"    -n    seperate groups of lines"+nl+
"    -z    filter out lines"+nl+
"    -p[p] pack previously sorted identical lines (-pp = -p -f)"+nl+
"    -s    swap first and last lines in DUPLINES generated groups"+nl+
nl+
"    -t    glue components with tabulation (default is space)"+nl+
"    -e    enclose components with double quotes"+nl+
"    -r    keep components as remarks"+nl+
"    -b    batch mode (check error code forcing a pause if mismatch)"+nl+
"    -n[n] insert en empty line between groups of lines (-nn = -n -k)"+nl+
"    -k    convert text to lowercase before comparing arguments"+nl+
"    -m    process only groups with exact (case sensitive) <string> matches"+nl+
"    -x    reverse filter (-z only)"+nl+
"    -i    pack identical lines whatever case"+nl+
'    -c[c] pad prefixed count (-cc = pad with " " instead of "0")'+nl+
"    -u    do not add prefix if count equals 1"+nl+
"    -f    alternate prefix format"+nl+
nl+
"a) A group is two or more successive lines without an empty line."+nl+
"   Single lines are ignored. A line can be up to 1024 characters long."+nl+
"b) Syntax 1 output assumes FCOMP exists (-b option assumes MPAUSE exists too)."+nl+
"c) Syntax 2 creates groups based upon first argument in each <source> line,"+nl+
"   from first char to first space or tab (existing empty lines are preserved)."+nl+
'   This function is best used on lists created with "NEWLINE -u -- -o:1..33".'+nl+
"d) Syntax 5 requires DUPLINES generated groups."+nl+
"e) Any option irrelevant to syntax is silently ignored."+nl+
"f) Output should be redirected to a file (whether it is a batch or not)."+nl+
nl+
"Examples : "+exe+" dups.f -e -r"+nl+
"           "+exe+" dups.f -b"+nl;

CONST
    msgOp = "-g, -n, -z, -p[p] and -d options are mutually exclusive !";

PROCEDURE abort (rc:CARDINAL;S:ARRAY OF CHAR);
BEGIN
    CASE rc OF
    | errNone: ;
    | errHelp:
        WrLn; (* not done at program start *)
        WrStr(msgHelp);;
    ELSE
        IF Str.Length(S) # 0 THEN
            WrLn; (* not done at program start *)
            WrStr(exe+" : ");WrStr(S);WrLn;
        END;
    END;
    Lib.SetReturnCode( SHORTCARD(rc) );
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType    = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn : ioBufferType;

(* ------------------------------------------------------------ *)

PROCEDURE doquote (enquote:BOOLEAN;VAR S:ARRAY OF CHAR   );
BEGIN
    IF enquote THEN
        Str.Prepend(S,dquote);
        Str.Append(S,dquote);
    END;
END doquote;

PROCEDURE dmp ( keeporg:BOOLEAN;rem,S:ARRAY OF CHAR  );
BEGIN
    IF keeporg THEN
        WrStr(rem);WrStr(S);WrLn;
    END;
END dmp;

PROCEDURE dbg (ok:BOOLEAN;S:ARRAY OF CHAR   );
BEGIN
    IF ok THEN
        WrStr("/// ");WrStr(S);WrLn;
    END;
END dbg;

(* ------------------------------------------------------------ *)

PROCEDURE procJoin ( DEBUG,useLFN,enquote,keeporg,batchmode:BOOLEAN;sepa:CHAR;
                     prefix,src:ARRAY OF CHAR  );
VAR
    F:pathtype;
    hin:FIO.File;
    S,BASE,Z:hugestr;
    state:(waiting,ingroup);
BEGIN
    Str.Copy(F,src);
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);

    IF batchmode THEN
        WrStr("@ECHO OFF");WrLn;
        WrLn;
    END;

    state:=waiting; (* //V10B FIXED *)

    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        dbg(DEBUG,S);
        IF ( S[0] = charnull ) AND FIO.EOF THEN EXIT; END;
        CASE state OF
        | waiting:
dbg(DEBUG,"waiting");
            CASE S[0] OF
            | charnull : ;
            ELSE
                dmp(keeporg,rembase,S);
                doquote(enquote,S);
                BASE:=S;
                state:=ingroup;
            END;
        | ingroup:
dbg(DEBUG,"ingroup");
            CASE S[0] OF
            | charnull :
                WrLn;
                state:=waiting;
            ELSE
                dmp(keeporg,remark,S);
                doquote(enquote,S);
                Z:="||||";
                IF batchmode THEN
                Str.Subst(Z,hbar,prefixbatch);
                ELSE
                Str.Subst(Z,hbar,prefix);
                END;
                Str.Subst(Z,hbar,BASE);
                Str.Subst(Z,hbar,sepa);
                Str.Subst(Z,hbar,S);
                WrStr(Z);WrLn;
                IF batchmode THEN
                WrStr(testbatch);WrLn;
                END;
            END;
        END;
    END;

    CASE state OF
    | waiting : ;
    | ingroup:
            CASE S[0] OF
            | charnull :
                WrLn;
            ELSE
                dmp(keeporg,remark,S);
                doquote(enquote,S);
                Z:="||||";
                Str.Subst(Z,hbar,prefix);
                Str.Subst(Z,hbar,BASE);
                Str.Subst(Z,hbar,sepa);
                Str.Subst(Z,hbar,S);
                WrStr(Z);WrLn;
            END;
    END;
    fileClose(useLFN,hin);
END procJoin;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procSplit ( DEBUG,useLFN,lower:BOOLEAN;src:ARRAY OF CHAR  );
VAR
    F:pathtype;
    hin:FIO.File;
    S,REF,NOW:hugestr; (* both oversized just in case *)
    p:CARDINAL;
    currline:LONGCARD;
    ok:BOOLEAN;
BEGIN
    Str.Copy(F,src);
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);
    currline:=0;
    REF := "";
    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,S);
        (*
        LtrimBlanks(S);
        RtrimBlanks(S);
        *)
        IF ( S[0] = charnull ) AND FIO.EOF THEN EXIT; END;
        IF same(S,"") THEN
            WrLn;
            REF := "";
        ELSE
            p:=Str.CharPos(S,blank);
            IF p = MAX(CARDINAL) THEN p:=Str.CharPos(S,tabul);END;
            IF p = MAX(CARDINAL) THEN
                Str.Copy(NOW,S);
            ELSE
                Str.Slice(NOW,S,0,p);
            END;
            IF lower THEN LowerCase(NOW);END;
            IF same(REF,NOW) THEN
                WrStr(S);WrLn;
            ELSE
                ok:=(currline # 0);
                ok:=(ok AND NOT(same(REF,"")) );
                IF ok THEN WrLn;END;
                WrStr(S);WrLn;
                Str.Copy(REF,NOW);
            END;
        END;
        INC(currline);
    END;
    fileClose(useLFN,hin);
END procSplit;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procKeep ( DEBUG,useLFN,keepcase,filterout:BOOLEAN;src,str:ARRAY OF CHAR  );
VAR
    F:pathtype;
    hin:FIO.File;
    ORGS,S,what:hugestr;
    state:(waiting,ingroup);
    fpos,anchor:LONGCARD;
    nfound:CARDINAL;
    dmpme:BOOLEAN;
BEGIN
    Str.Copy(what,str);
    IF NOT(keepcase) THEN LowerCase(what);END;

    Str.Copy(F,src);
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);

    state:=waiting;

    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        fpos:=FIO.GetPos(hin);
        FIO.RdStr(hin,ORGS);
        Str.Copy(S,ORGS);
        LtrimBlanks(S);
        RtrimBlanks(S);
        dbg(DEBUG,ORGS);
        IF NOT(keepcase) THEN LowerCase(S);END;
        IF ( S[0] = charnull ) AND FIO.EOF THEN EXIT; END;
        CASE state OF
        | waiting:
dbg(DEBUG,"waiting");
            CASE S[0] OF
            | charnull: ;
            ELSE
                nfound:=0;
                IF Str.Pos(S,what) # MAX(CARDINAL) THEN INC (nfound);END;
                anchor:=fpos;
                state:=ingroup;
            END;
        | ingroup:
dbg(DEBUG,"ingroup");
            CASE S[0] OF
            | charnull:
                IF filterout THEN
                    dmpme := ( nfound # 0 );
                ELSE
                    dmpme := ( nfound = 0 );
                END;
                IF dmpme THEN
                    FIO.Seek(hin,anchor);
                    LOOP
                        FIO.RdStr(hin,S);
                        WrStr(S);WrLn;
                        IF FIO.GetPos(hin) > fpos THEN EXIT; END;
                    END;
                END;
                state:=waiting;
            ELSE
                IF Str.Pos(S,what) # MAX(CARDINAL) THEN INC(nfound);END;
            END;
        END;
    END;

    CASE state OF
    | waiting : ;
    | ingroup:
                IF filterout THEN
                    dmpme := ( nfound # 0 );
                ELSE
                    dmpme := ( nfound = 0 );
                END;
                IF dmpme THEN
                    FIO.Seek(hin,anchor);
                    LOOP
                        FIO.RdStr(hin,S);
                        WrStr(S);WrLn;
                        IF FIO.GetPos(hin) > fpos THEN EXIT; END;
                    END;
                END;
    END;
    fileClose(useLFN,hin);
END procKeep;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE num2str (n:LONGCARD;base:CARDINAL;wi:INTEGER;prefix,padchar:CHAR):str16;
VAR
    ok:BOOLEAN;
    R,Z:str16;
    i:CARDINAL;
BEGIN
    Str.CardToStr(n,R,base,ok);
    IF NOT(ok) THEN RETURN "--- ERROR !"; END;
    FOR i:=Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,blank);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
    IF prefix # "" THEN Str.Prepend(R,prefix);END;
    IF base=16 THEN Str.Lows(R);END;
    RETURN R;
END num2str;

PROCEDURE procPack ( DEBUG,useLFN,
                   packwhatevercase,padcount,padspace,showallcounts,altpackfmt:BOOLEAN;
                   src:ARRAY OF CHAR  );
CONST
    placeholder = tilde;
    wilc = 9; (* wi for LONGCARD *)
VAR
    hin:FIO.File;
    R,Z:str16;
    prev,curr,ORGprev,ORGcurr:hugestr;
    count:LONGCARD;
    wi:INTEGER;
    emptystr,dmp:BOOLEAN;
    padchar:CHAR;
    F:pathtype;
    state:(waiting,working);
BEGIN
    IF padspace THEN
        padchar:=blank;
    ELSE
        padchar:="0";
    END;
    IF padcount THEN
        wi:=wilc;
    ELSE
        wi:=1;
    END;
    Str.Copy(F,src);
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);

    state:=waiting;
    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,ORGcurr);
dbg(DEBUG,ORGcurr);
        Str.Copy(curr,ORGcurr);
        LtrimBlanks(curr);
        RtrimBlanks(curr);
        emptystr:=( curr[0] = charnull);
        IF emptystr THEN
            IF FIO.EOF THEN EXIT; END;
        ELSE
            IF packwhatevercase THEN LowerCase(curr);END;
dbg(DEBUG,curr);
            CASE state OF
            | waiting:
                count:=1;
                prev:=curr;
                ORGprev:=ORGcurr;
                state:=working;
            | working:
                IF same(curr,prev) THEN
                    INC(count);
                ELSE
                    R:=num2str(count,10,wi,"",padchar);
                    IF NOT(altpackfmt) THEN
                        Z:="~ ::: ";
                    ELSE
                        Z:="(~) ";
                    END;
                    Str.Subst(Z,placeholder,R);
                    IF showallcounts THEN
                        dmp:=TRUE;
                    ELSE
                        dmp:=(count # 1);
                    END;
                    IF dmp THEN WrStr(Z); END;
                    IF packwhatevercase THEN
                        WrStr(prev);
                    ELSE
                        WrStr(ORGprev);
                    END;
                    WrLn;
                    count  :=1;
                    prev   :=curr;
                    ORGprev:=ORGcurr;
                END;
            END;
        END;
    END;

                    R:=num2str(count,10,wi,"",padchar);
                    IF NOT(altpackfmt) THEN
                        Z:="~ ::: ";
                    ELSE
                        Z:="(~) ";
                    END;
                    Str.Subst(Z,placeholder,R);
                    IF showallcounts THEN
                        dmp:=TRUE;
                    ELSE
                        dmp:=(count # 1);
                    END;
                    IF dmp THEN WrStr(Z); END;
                    IF packwhatevercase THEN
                        WrStr(prev);
                    ELSE
                        WrStr(ORGprev);
                    END;
                    WrLn;



    fileClose(useLFN,hin);
END procPack;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
# 899D9CCC6DEB8FD9D595F9AC1F5A8D6E d:*ARC\032\ADDED\TOOLS\MEMOREX\KEYBOARD\SETUP.EXE
; 899D9CCC6DEB8FD9D595F9AC1F5A8D6E d:*ARC\051\UNITD\ORGDISKS\USING\MEMOREX\KEYBOARD\SETUP.EXE
; 899D9CCC6DEB8FD9D595F9AC1F5A8D6E f:*ue\ORGDISKS\USING\MEMOREX\KEYBOARD\SETUP.EXE
*)

PROCEDURE procSwap ( DEBUG,useLFN:BOOLEAN;src:ARRAY OF CHAR  );
VAR
    F:pathtype;
    hin:FIO.File;
    fpos,anchor,prevfpos:LONGCARD;
    nfound:CARDINAL; (* should do ! *)
    state:(waiting,ingroup);
    ORGS,S:hugestr;
BEGIN
    Str.Copy(F,src);
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);

    state:=waiting;

    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        prevfpos:=fpos;
        fpos:=FIO.GetPos(hin);
        FIO.RdStr(hin,ORGS);
        Str.Copy(S,ORGS);
        LtrimBlanks(S);
        RtrimBlanks(S);
        dbg(DEBUG,ORGS);
        IF ( S[0] = charnull ) AND FIO.EOF THEN EXIT; END;
        CASE state OF
        | waiting:
dbg(DEBUG,"waiting");
            CASE S[0] OF
            | charnull: ;
            ELSE
                nfound:=1;
                anchor:=fpos;     (* "# first line" *)
                state:=ingroup;
            END;
        | ingroup:
dbg(DEBUG,"ingroup");
            CASE S[0] OF
            | charnull:
                IF nfound > 1 THEN (* one single line is NOT a group ! *)

                    FIO.Seek(hin,prevfpos); (* "; last line" *)
                    FIO.RdStr(hin,ORGS);
                    ORGS[0]:=pound;
                    WrStr(ORGS);WrLn;

                    FIO.Seek(hin,anchor);
                    FIO.RdStr(hin,ORGS);
                    ORGS[0]:=semicolon;

                    LOOP
                        IF FIO.GetPos(hin) < prevfpos THEN
                            FIO.RdStr(hin,S);
                            WrStr(S);WrLn;
                        ELSE
                            EXIT;
                        END;
                    END;
                    WrStr(ORGS);WrLn;
                    WrLn;
                END;
                state:=waiting;
            ELSE
                INC(nfound);
            END;
        END;
    END;

    CASE state OF
    | waiting : ;
    | ingroup:
                IF nfound > 1 THEN (* one single line is NOT a group ! *)

                    FIO.Seek(hin,prevfpos); (* "; last line" *)
                    FIO.RdStr(hin,ORGS);
                    ORGS[0]:=pound;
                    WrStr(ORGS);WrLn;

                    FIO.Seek(hin,anchor);
                    FIO.RdStr(hin,ORGS);
                    ORGS[0]:=semicolon;

                    LOOP
                        IF FIO.GetPos(hin) < prevfpos THEN
                            FIO.RdStr(hin,S);
                            WrStr(S);WrLn;
                        ELSE
                            EXIT;
                        END;
                    END;
                    WrStr(ORGS);WrLn;
                    WrLn;
                END;

    END;
    fileClose(useLFN,hin);
END procSwap;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    firstparm = 1;
    maxparm   = 2;
VAR
    parm : ARRAY [firstparm..maxparm] OF pathtype;
    i,opt,parmcount,lastparm:CARDINAL;
    S,R,src:pathtype;
    useLFN,DEBUG:BOOLEAN;
    batchmode,enquote,keeporg,lower,ignorecase,keepcase:BOOLEAN;
    packwhatevercase,padcount,padspace,showallcounts,altpackfmt,filterout:BOOLEAN;
    sepa : CHAR;
    op : (opNone,opJoin,opSplit,opKeep,opPack,opSwap);
BEGIN
    FIO.IOcheck := FALSE; (* don't let topspeed handle problems *)
    FIO.ShareMode:=FIO.ShareDenyNone; (* very, very important ! *)
    (* WrLn;  not here because of redirection *)

    op              := opNone;
    enquote         := FALSE;
    keeporg         := FALSE;
    batchmode       := FALSE;
    sepa            := blank;
    lower           := FALSE;
    keepcase        := FALSE;
    packwhatevercase:= FALSE;
    padcount        := FALSE;
    padspace        := FALSE;
    showallcounts   := TRUE;
    altpackfmt      := FALSE;
    filterout       := TRUE;
    DEBUG           := FALSE;

    lastparm        := firstparm-1;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,msgHelp);END;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "T"+delim+"TAB"+delim+
                                  "E"+delim+"ENQUOTE"+delim+
                                  "R"+delim+"REM"+delim+
                                  "B"+delim+"BATCH"+delim+

                                  "N"+delim+"NEWLINE"+delim+
                                  "K"+delim+"LOWERCASE"+delim+
                                  "NN"+delim+"NK"+delim+"KN"+delim+
                                  "M"+delim+"CASE"+delim+

                                  "G"+delim+"GLUE"+delim+
                                  "Z"+delim+"KEEP"+delim+"ONLY"+delim+
                                  "P"+delim+"PACK"+delim+
                                  "I"+delim+"IGNORECASE"+delim+

                                  "C"+delim+"PADCOUNT"+delim+
                                  "PP"+delim+"PACKALT"+delim+"PF"+delim+"FP"+delim+
                                  "U"+delim+"UNIQUE"+delim+
                                  "F"+delim+"FORMAT"+delim+
                                  "X"+delim+"INVERSE"+delim+"SWAP"+delim+
                                  "S"+delim+"SWAP"+delim+
                                  "CC"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3   : abort(errHelp,msgHelp);
            | 4,5     : sepa:=tabul;
            | 6,7     : enquote:=TRUE;
            | 8,9     : keeporg:=TRUE;
            | 10,11   : batchmode:=TRUE;

            | 12,13   : CASE op OF
                        | opNone,opSplit:op:=opSplit;
                        ELSE
                            abort(errOp,msgOp);
                        END;
            | 14,15   : lower  := TRUE;
            | 16,17,18: CASE op OF
                        | opNone,opSplit:op:=opSplit;
                        ELSE
                            abort(errOp,msgOp);
                        END;
                        lower:=TRUE;
            | 19,20   : keepcase:=TRUE;

            | 21,22   : CASE op OF
                        | opNone,opJoin:op:=opJoin;
                        ELSE
                            abort(errOp,msgOp);
                        END;
            | 23,24,25: CASE op OF
                        | opNone,opKeep:op:=opKeep;
                        ELSE
                            abort(errOp,msgOp);
                        END;
            | 26,27:    CASE op OF
                        | opNone,opPack:op:=opPack;
                        ELSE
                            abort(errOp,msgOp);
                        END;
            | 28,29:    packwhatevercase:=TRUE;

            | 30,31:    padcount:=TRUE;
            | 32,33,34,35:
                        CASE op OF
                        | opNone,opPack:op:=opPack;
                        ELSE
                            abort(errOp,msgOp);
                        END;
                        altpackfmt:=TRUE;
            | 36,37:    showallcounts:=FALSE;
            | 38,39:    altpackfmt:=TRUE;
            | 40,41,42: filterout :=FALSE;
            | 43,44:
                        CASE op OF
                        | opNone,opSwap:op:=opSwap;
                        ELSE
                            abort(errOp,msgOp);
                        END;
            | 45:       padcount:=TRUE; padspace:=TRUE;
            | 46:       DEBUG  :=TRUE;
            ELSE
                abort(errOpt,"Unknown option !");
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParm,"Too many parameters !");END;
            Str.Copy(parm[lastparm],S); (* keep case *)
        END;
    END;
    CASE lastparm OF
    | firstparm-1 : abort(errExpecting,"<source> parameter expected !");
    | firstparm   : CASE op OF
                    | opKeep : abort(errExpecting,"<string> parameter expected !");
                    ELSE
                        ;
                    END;
    | firstparm+1 : CASE op OF
                    | opKeep : ;
                    | opNone : op:=opKeep; (* 2 parms force -z *)
                    ELSE
                        abort(errUnexpected,"Unexpected <string> parameter !");
                    END;
    ELSE                (* already trapped *)
                    ;
    END;
    IF op=opNone THEN op:=opJoin;END; (* default is -g *)

    Str.Copy(src,parm[firstparm]);
    IF chkJoker(src) THEN abort(errJoker,"Illegal joker(s) in <source> !");END;
    useLFN:=fileSupportLFN();
    IF fileExists(useLFN,src)=FALSE THEN abort(errNotFound,"<source> does not exist !");END;

    CASE op OF
    | opSplit:
        procSplit(DEBUG,useLFN,lower,src);
    | opJoin:
        procJoin(DEBUG,useLFN,enquote,keeporg,batchmode,sepa,prefix,src);
    | opKeep:
        procKeep(DEBUG,useLFN,keepcase,filterout,src,parm[firstparm+1]);
    | opPack:
        procPack(DEBUG,useLFN,packwhatevercase,padcount,padspace,showallcounts,altpackfmt,src);
    | opSwap:
        procSwap(DEBUG,useLFN,src);
    END;

    abort(errNone,"");
END opGroup.




