(* ---------------------------------------------------------------
Title         Q&D Random Number Generator
Author        PhG
Overview
Notes         minimal error messages and checking, etc.
              we know, each rndgen should have had its own private MODULE
              because of featuritis, we're far from original small utility !
Bugs          check JSF8 rescaling !
Wish List     are you kidding ?
              show elapsed time or number of numbers for screen to be blank
              add RC4 ?

--------------------------------------------------------------- *)

MODULE rndgen;

IMPORT Lib;
IMPORT Str;
IMPORT IO;
IMPORT FIO;
IMPORT MATHLIB;

IMPORT Graph;
IMPORT SYSTEM;

FROM IO IMPORT WrStr, WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_rand IMPORT InitRnd, GetRnd,
GetRndCardRange, GetRndLngCardRange,
GetRndLngRealRange, GetRndRealRange,
GetRndIntRange, GetRndLngIntRange;

(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "RNDGEN";
    ProgTitle     = "Q&D Random Number Generator";
    ProgVersion   = "v1.1f";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone       = 0;
    errHelp       = 1;
    errOption     = 2;
    errParameter  = 3;
    errBinRange   = 4;
    errInterval   = 5;
    errBadNum     = 6;
    errSyntax     = 7;
    errJoker      = 8;
    errExists     = 9;
    errZeroLength = 10;
    errRangePi    = 11;
    errOddPi      = 12;
    err2gigas     = 13;
    errBadSyntax  = 14;
    errDisplay    = 15;
    errGenerator  = 16;
    errPalette    = 17;
    errCardRange  = 18;
    errRangeForMissing=19;
    errMissing    = 20;
    errByte       = 21;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
00000000011111111112222222222333333333344444444445555555555666666666677777777778
1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    msgHelp =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <lower> <upper> [-g:#] [-v[v]]"+nl+
"Syntax 2 : "+ProgEXEname+" <lower> <upper> <length> <file> [-g:#] [-o] [-pi] [-m]"+nl+
"Syntax 3 : "+ProgEXEname+" <-t[t]|-y[y]|-u[u]> [-g:#] [-x]"+nl+
nl+
"This program returns a random number in errorlevel (syntax 1),"+nl+
"creates a DOS <file> of specified <length> filled with random bytes (syntax 2),"+nl+
"or shows an unscientific randomgram for specified random generator."+nl+
"Generated random number will belong to [<lower>..<upper>] range."+nl+
nl+
"  -a    create an ASCII file (-pi ignored, default file is binary)"+nl+
"  -t[t] 320x200x4 display (-t = immediate color, -tt = progressive color)"+nl+
"  -y[y] 640x480x4 display (-y = immediate color, -yy = progressive color)"+nl+
"  -u[u] 320x200x8 display (-u = immediate color, -uu = progressive color)"+nl+
"  -g:#  random generator ([1..6] range, default is 1)"+nl+
"  -v[v] (very) verbose (-v forced with syntax 2)"+nl+
"  -o    overwrite existing <file> if any"+nl+
"  -pi   compute PI value"+nl+
"  -m    show missing byte values"+nl+
"  -fix  force xmax and ymax to 255 if larger"+nl+
"  -p:#  palette ([0..7], default is 0)"+nl+
nl+
"a) Please note program always returns 0 after an error or after help screen."+nl+
"b) Without -a option, <lower> and <upper> values must belong to [0..255] range."+nl+
"   Note with -a option, <length> specifies the count of entries to create."+nl+
'c) "." and "," characters, if any, are removed from <length> parameter.'+nl+
"d) Random generators : 1 = TopSpeed, 2 = shuffled TopSpeed,"+nl+
"   3 = Knuth (additive congruential), 4 = Marsaglia,"+nl+
"   5 = Lehmer (linear congruential), 6 = Jenkins (I.S.A.A.C.),"+nl+
"   7 = JSF32, 8 = JSF16, 9 = JSF8."+nl+
"   Note code for 3..4 and 6..9 seems correct but has not been fully checked."+nl+
"e) If conflicting options are specified (random generator, display, palette),"+nl+
"   program will use the last specified one."+nl+
"f) -t[t], -y[y] and -u[u] displays respectively use 16, 16 and 256 colors."+nl+
"   Any key exists -t[t], -y[y] and -u[u] displays."+nl+
"   Note -fix option is required for 8 bit random generators."+nl+
"g) One useless option exists for compatibility with "+ProgEXEname+" prior to v1.0b :"+nl+
"   -w uses older value for random generator warmup."+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :       WrLn; WrStr(msgHelp);
    | errOption :     S := 'Unknown "~" option !';
    | errParameter :  S := '"~" is one parameter too many !';
    | errBinRange :   S := '"~" is out of valid [0..255] range !';
    | errInterval :   S := "<lower> and <upper> should be different !";
    | errBadNum :     S := 'Illegal "~" number !';
    | errSyntax:      S := "Syntax error !";
    | errJoker:       S := '"~" contains at least one joker !';
    | errExists:      S := '"~" already exists !';
    | errZeroLength:  S := "Creating an empty file is a nonsense !";
    | errRangePi:     S := "-pi option requires [<lower>..<upper>] range to be [0..255] !";
    | errOddPi:       S := " -pi option requires <length> to be an even number !";
    | err2gigas:      S := "<length> cannot be more than 2 Gb !";
    | errBadSyntax :  S := "Syntax ~ command line error !";
    | errDisplay:     S := "Requested video mode would not initialize !";
    | errGenerator :  S := '"~" generator is out of expected range !';
    | errPalette:     S := '"~" palette is out of expected range !';
    | errCardRange :  S := '"~" is out of valid [0..65535] range !';
    | errRangeForMissing:S:="-m option requires [0.255] range for <lower> and <upper> !";
    | errMissing:     S := "-m option requires syntax ~ !";
    | errByte:        S := "-x option requires syntax ~ !";
    ELSE
                      S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrLn;
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;

    (* Lib.SetReturnCode(SHORTCARD(e)); *)
    Lib.SetReturnCode(errNone); (* force 0 in any case ! *)
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    firstiobuffer = 1;
    lastiobuffer  = 16*512+FIO.BufferOverhead;
VAR
    iobuffer : ARRAY [firstiobuffer..lastiobuffer] OF BYTE;

(* ------------------------------------------------------------ *)

CONST
    sizevalue = 256;
    minvalue  = 0;
    maxvalue  = sizevalue-1;
VAR
    value : ARRAY[minvalue..maxvalue] OF LONGCARD;
    valuetotal :LONGCARD;
    valueoverflow : BOOLEAN;

PROCEDURE initvalues (  );
VAR
    i:CARDINAL ;
BEGIN
    FOR i:=minvalue TO maxvalue DO value[i]:=0;END;
    valueoverflow:=FALSE;
END initvalues;

PROCEDURE chkvalues (lower,upper:CARDINAL ):BOOLEAN;
BEGIN
    RETURN ( (lower<=maxvalue) AND (upper<=maxvalue) );
END chkvalues;

PROCEDURE updatevalue ( v:LONGCARD);
VAR
    ndx:CARDINAL;
BEGIN
    ndx:= CARDINAL(v MOD sizevalue);
    INC(value[ndx]);
    IF valuetotal = MAX(LONGCARD) THEN
        valueoverflow:=TRUE;
        (* valuetotal := 0; *) (* let value as is for division *)
    ELSE
        INC(valuetotal);
    END;
END updatevalue;

PROCEDURE showmissing ( );
CONST
    wib = 3;
    sOVERFLOW  = "::: More than MAX(LONGCARD) values !";
    sNOMISSING = "::: No missing value.";
    sMISSING   = "Missing value : ";
VAR
    n,i : CARDINAL;
BEGIN
    IF valueoverflow THEN
        WrStr(sOVERFLOW);WrLn;
        RETURN;
    END;
    n:=0;
    FOR i:=minvalue TO maxvalue DO
        IF value[i] = 0 THEN
            INC(n);
            WrStr(sMISSING);
            IO.WrCard(i,wib);WrLn;
        END;
    END;
    IF n = 0 THEN WrStr(sNOMISSING);WrLn;END;
END showmissing;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* //FIXME should be ok *)

CONST
    kC  = 54;
    fR  = LONGREAL(65536.0);
VAR
    aDK      : ARRAY [0 .. kC] OF CARDINAL;
    jDK, kDK : [0 .. kC];

PROCEDURE getNextKNUTH() : CARDINAL;
BEGIN
    aDK [kDK] := aDK [kDK] + aDK [jDK];

    IF kDK = 0 THEN
        kDK := kC;
    ELSE
        DEC(kDK);
    END;
    IF jDK = 0 THEN
        jDK := kC;
    ELSE
        DEC(jDK);
    END;
    RETURN aDK [kDK];
END getNextKNUTH;

PROCEDURE RandomInitKNUTH( seed : CARDINAL );
VAR
    i, dummy : CARDINAL;
BEGIN
    jDK := 24;
    kDK := 0;
    FOR i := 0 TO kC DO
        aDK [i] := 0;
    END;
    aDK [kDK] := 31415 + seed;
    IF aDK [kDK] = 0 THEN aDK [kDK] := 31415; END;
    FOR i := 1 TO 2000 DO dummy := getNextKNUTH (); END;
END RandomInitKNUTH;

PROCEDURE RandomCardKNUTH( bound : CARDINAL ) : CARDINAL;
BEGIN
    IF bound = 0 THEN
        RETURN getNextKNUTH ();
    ELSE
        RETURN TRUNC (LONGREAL (bound) * LONGREAL (getNextKNUTH ()) / fR);
    END;
END RandomCardKNUTH;

PROCEDURE RandomIntKNUTH( bound : INTEGER ) : INTEGER;
BEGIN
      RETURN INTEGER (RandomCardKNUTH (CARDINAL (ABS (bound))));
END RandomIntKNUTH;

PROCEDURE RandomRealKNUTH() : LONGREAL;
BEGIN
      RETURN LONGREAL (RandomCardKNUTH (10000)) * 1.0E-16 +
             LONGREAL (RandomCardKNUTH (10000)) * 1.0E-12 +
             LONGREAL (RandomCardKNUTH (10000)) * 1.0E-08 +
             LONGREAL (RandomCardKNUTH (10000)) * 1.0E-04;
END RandomRealKNUTH;

PROCEDURE RandomizeKNUTH();
VAR
    i,dummy : CARDINAL;
    h,m,s,ss: CARDINAL;
BEGIN
    Lib.GetTime(h,m,s,ss);
    RandomInitKNUTH (m*60+s*60+ss); (* was ss but PC timer is not precise *)
    jDK := m * s;
    FOR i := 0 TO jDK DO
        dummy := getNextKNUTH ();
    END;
END RandomizeKNUTH;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* //FIXME should be ok *)

VAR
    m_w, m_z : LONGCARD;

PROCEDURE getMarsa32():LONGCARD;
BEGIN
    m_z := LONGCARD(36969) * (m_z AND LONGCARD(65535)) + (m_z >> 16);
    m_w := LONGCARD(18000) * (m_w AND LONGCARD(65535)) + (m_w >> 16);
    RETURN (m_z << 16) + m_w;
END getMarsa32;

PROCEDURE getRndMarsa (lower,upper:CARDINAL  ):LONGREAL ;
VAR
    range:CARDINAL;
    lc:LONGCARD;
    v:LONGREAL;
BEGIN
    lc:=getMarsa32();
    range:=upper-lower+1;
    v:= ( LONGREAL(lc) / ( LONGREAL(MAX(LONGCARD))+LONGREAL(1.0) ) );
    RETURN LONGREAL(lower)+(LONGREAL(range) * v);
END getRndMarsa;

PROCEDURE initMarsa (  );
CONST
    avoidMW1 = 0;
    avoidMW2 = LONGCARD (0464FFFFFH);
    avoidMZ1 = 0;
    avoidMZ2 = LONGCARD (09068FFFFH);
VAR
    h,m,s,ss,pb,i:CARDINAL;
    a,b,avoid1,avoid2,v:LONGCARD;
BEGIN
    FOR i:=1 TO 2 DO
        CASE i OF
        | 1: avoid1:=avoidMW1; avoid2:=avoidMW2;
        | 2: avoid1:=avoidMZ1; avoid2:=avoidMZ2;
        END;
        LOOP
            Lib.GetTime(h,m,s,ss);
            a:=LONGCARD(m * s)  ;
            b:=LONGCARD(h * ss) << 16;
            v:=(a OR b);
            pb:=0;
            IF v = avoid1 THEN INC(pb); END;
            IF v = avoid2 THEN INC(pb); END;
            IF pb = 0 THEN EXIT; END;
        END;
        CASE i OF
        | 1: m_w:=v;
        | 2: m_z:=v;
        END;
    END;
END initMarsa;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
    seems rather good a random generator, can't remember where it comes from
    //V11D
    though we're sure we had not seen it there,
    this linear congruential generator appears in
    "Programming in Oberon" by Wirth and Reiser,
    where it is attruibuted to a D. H. Lehmer
*)

VAR
    ZZ: LONGINT;

PROCEDURE getSimple(): LONGREAL;
CONST
    a = LONGINT (16807);	
    m = LONGINT (2147483647); 	(* 2^31-1  MAX(LONGINT)  *)
    q = LONGINT (m DIV a);	
    r = LONGINT (m MOD a);
VAR
    g: LONGINT;
BEGIN
	g := a*(ZZ MOD q) - r*(ZZ DIV q);
	IF g > 0 THEN
	   ZZ := g;
	ELSE
	   ZZ := g + m;
	END;
	RETURN LONGREAL(ZZ)*(1.0/LONGREAL(m))
END getSimple;

PROCEDURE initSimple();
VAR
    seed : LONGINT;
    h,m,s,ss:CARDINAL;
BEGIN
    Lib.GetTime(h,m,s,ss);
    seed:=LONGINT(s*3600)+LONGINT(m*60)+LONGINT(h)+LONGINT(ss);
	ZZ := seed;
END initSimple;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE lcXOR (b1,b2 : LONGCARD) : LONGCARD;
TYPE
    lcBITSET = SET OF [0..31];
BEGIN
    RETURN LONGCARD ( (lcBITSET (b1) / lcBITSET (b2)) );
END lcXOR;

(* ------------------------------------------------------------ *)

CONST
    ISAACSTATESIZE = 256;
    minisaacstate  = 0;
    maxisaacstate  = ISAACSTATESIZE-1;

TYPE
    isaacCTX = RECORD
         (* internal state *)
         isaacmem     : ARRAY [minisaacstate..maxisaacstate] OF LONGCARD;
         isaacAA,isaacBB,isaacCC:LONGCARD;
         (* external *)
         isaacpool    : ARRAY [minisaacstate..maxisaacstate] OF LONGCARD; (* ub4 in readable.c *)
         isaacpoolndx : CARDINAL; (* an index, in fact *)
    END;
VAR
    (* local globerks ! *)
    ctx        : isaacCTX;

(* ------------------------------------------------------------ *)

PROCEDURE updateisaac();
VAR
    i,ndx:CARDINAL;
    x,y:LONGCARD;
BEGIN
    INC(ctx.isaacCC);        (* isaacCC just gets incremented once per 256 results *)
    INC(ctx.isaacBB,ctx.isaacCC);     (* then combined with isaacBB *)

    FOR i:=minisaacstate TO maxisaacstate DO
        x := ctx.isaacmem[i];
        CASE (i MOD 4) OF
        | 0: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA << 13) );
        | 1: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA >>  6) );
        | 2: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA <<  2) );
        | 3: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA >> 16) );
        END;
        ndx         := (i + 128) MOD ISAACSTATESIZE;
        ctx.isaacAA      := ctx.isaacmem[ ndx] + ctx.isaacAA;

        ndx         := CARDINAL ( (x >> 2) MOD ISAACSTATESIZE) ;
        y           := ctx.isaacmem[ ndx] + ctx.isaacAA + ctx.isaacBB;
        ctx.isaacmem[i]   := y;

        ndx         := CARDINAL ( (y >> 10) MOD ISAACSTATESIZE) ;
        ctx.isaacBB      := ctx.isaacmem[ ndx ] + x;
        ctx.isaacpool[i]  := ctx.isaacBB;
    END;
    ctx.isaacpoolndx := minisaacstate; (* reset index *)
END updateisaac;

(* ------------------------------------------------------------ *)

CONST
    DOzeroPOOL = TRUE;
    DONOTzeroPOOL= FALSE;

PROCEDURE initisaac(isaacseedfrompool,zeropool:BOOLEAN);
CONST
    magic = 9E3779B9H; (* the golden ratio *)
VAR
    i:CARDINAL;
    a,b,c,d,e,f,g,h:LONGCARD;
BEGIN
    ctx.isaacAA:=0;
    ctx.isaacBB:=0;
    ctx.isaacCC:=0;
    FOR i:=minisaacstate TO maxisaacstate DO
        ctx.isaacmem[i]  := 0;
        IF zeropool THEN (* DOzeroPOOL *)
            ctx.isaacpool[i]:=0;
        END;
    END;

    a:=magic;
    b:=magic;
    c:=magic;
    d:=magic;
    e:=magic;
    f:=magic;
    g:=magic;
    h:=magic;

    FOR i:=0 TO 3 DO (* scramble it *)
        (* mix *)
        a:=lcXOR(a,b<<11);   INC(d,a);   INC(b,c);
        b:=lcXOR(b,c>> 2);   INC(e,b);   INC(c,d);
        c:=lcXOR(c,d<< 8);   INC(f,c);   INC(d,e);
        d:=lcXOR(d,e>>16);   INC(g,d);   INC(e,f);
        e:=lcXOR(e,f<<10);   INC(h,e);   INC(f,g);
        f:=lcXOR(f,g>> 4);   INC(a,f);   INC(g,h);
        g:=lcXOR(g,h<< 8);   INC(b,g);   INC(h,a);
        h:=lcXOR(h,a>> 9);   INC(c,h);   INC(a,b);
    END;

    FOR i:=minisaacstate TO maxisaacstate BY 8 DO (* fill in isaacmem[] with messy stuff *)
        IF isaacseedfrompool THEN (* use all the information in the seed *)
            INC(a,ctx.isaacpool[i  ]);
            INC(b,ctx.isaacpool[i+1]);
            INC(c,ctx.isaacpool[i+2]);
            INC(d,ctx.isaacpool[i+3]);
            INC(e,ctx.isaacpool[i+4]);
            INC(f,ctx.isaacpool[i+5]);
            INC(g,ctx.isaacpool[i+6]);
            INC(h,ctx.isaacpool[i+7]);
        END;
        (* mix *)
        a:=lcXOR(a,b<<11);   INC(d,a);   INC(b,c);
        b:=lcXOR(b,c>> 2);   INC(e,b);   INC(c,d);
        c:=lcXOR(c,d<< 8);   INC(f,c);   INC(d,e);
        d:=lcXOR(d,e>>16);   INC(g,d);   INC(e,f);
        e:=lcXOR(e,f<<10);   INC(h,e);   INC(f,g);
        f:=lcXOR(f,g>> 4);   INC(a,f);   INC(g,h);
        g:=lcXOR(g,h<< 8);   INC(b,g);   INC(h,a);
        h:=lcXOR(h,a>> 9);   INC(c,h);   INC(a,b);

        ctx.isaacmem[i  ]:=a;
        ctx.isaacmem[i+1]:=b;
        ctx.isaacmem[i+2]:=c;
        ctx.isaacmem[i+3]:=d;
        ctx.isaacmem[i+4]:=e;
        ctx.isaacmem[i+5]:=f;
        ctx.isaacmem[i+6]:=g;
        ctx.isaacmem[i+7]:=h;
    END;

    IF isaacseedfrompool THEN
        (* do a second pass to make all of the seed affect all of isaacmem *)
        FOR i:=minisaacstate TO maxisaacstate BY 8 DO
            INC(a,ctx.isaacmem[i  ]);
            INC(b,ctx.isaacmem[i+1]);
            INC(c,ctx.isaacmem[i+2]);
            INC(d,ctx.isaacmem[i+3]);
            INC(e,ctx.isaacmem[i+4]);
            INC(f,ctx.isaacmem[i+5]);
            INC(g,ctx.isaacmem[i+6]);
            INC(h,ctx.isaacmem[i+7]);

            (* mix *)
            a:=lcXOR(a,b<<11);   INC(d,a);   INC(b,c);
            b:=lcXOR(b,c>> 2);   INC(e,b);   INC(c,d);
            c:=lcXOR(c,d<< 8);   INC(f,c);   INC(d,e);
            d:=lcXOR(d,e>>16);   INC(g,d);   INC(e,f);
            e:=lcXOR(e,f<<10);   INC(h,e);   INC(f,g);
            f:=lcXOR(f,g>> 4);   INC(a,f);   INC(g,h);
            g:=lcXOR(g,h<< 8);   INC(b,g);   INC(h,a);
            h:=lcXOR(h,a>> 9);   INC(c,h);   INC(a,b);

            ctx.isaacmem[i  ]:=a;
            ctx.isaacmem[i+1]:=b;
            ctx.isaacmem[i+2]:=c;
            ctx.isaacmem[i+3]:=d;
            ctx.isaacmem[i+4]:=e;
            ctx.isaacmem[i+5]:=f;
            ctx.isaacmem[i+6]:=g;
            ctx.isaacmem[i+7]:=h;
        END;
    END;

    updateisaac(); (* isaacpoolndx autohandled *)
END initisaac;

PROCEDURE getLCisaacrnd():LONGCARD;
VAR
    v:LONGCARD;
BEGIN
    v:= ctx.isaacpool[ctx.isaacpoolndx];
    INC(ctx.isaacpoolndx);
    IF ctx.isaacpoolndx > maxisaacstate THEN updateisaac(); END; (* isaacpoolndx autohandled *)
    RETURN v;
END getLCisaacrnd;

(* ------------------------------------------------------------ *)

PROCEDURE initisaacseedval (isaacseedfrompool:BOOLEAN; PP:ARRAY OF LONGCARD);
VAR
    i,p,len:CARDINAL;
    v:LONGCARD;
BEGIN
    initisaac(isaacseedfrompool,DOzeroPOOL);

    FOR i:= minisaacstate TO maxisaacstate DO
        v:=PP[i];
        ctx.isaacpool[p]:=v;
        (* WrStr(fmtlc(v,16,8,"0",FALSE)); *)
    END;

    initisaac(isaacseedfrompool,DONOTzeroPOOL);
END initisaacseedval;

PROCEDURE initISAACrng ();
CONST
    isaacseedfrompool = TRUE;
VAR
    i,lo,hi: CARDINAL;
    v:LONGCARD;
    vlc:ARRAY[minisaacstate..maxisaacstate] OF LONGCARD;
BEGIN
    (* we could use time and a few random locations too *)
    Lib.RANDOMIZE;
    FOR i:=minisaacstate TO maxisaacstate DO
        lo := Lib.RANDOM(MAX(CARDINAL) );
        hi := Lib.RANDOM(MAX(CARDINAL) );
        v:=LONGCARD(hi) << 16 + LONGCARD(lo);
        vlc[i]:=v;
    END;
    initisaacseedval(isaacseedfrompool,vlc);
END initISAACrng;

PROCEDURE getISAAC (lower,upper:CARDINAL):LONGREAL;
CONST
    MAXlongCARDRANGE = LONGREAL(MAX(LONGCARD))+1.0;
VAR
    abcd:LONGCARD;
    v:LONGREAL;
BEGIN
    abcd:= getLCisaacrnd();
    v:=LONGREAL(abcd) / MAXlongCARDRANGE; (* [0..(2^32)-1] TO [0..1[ *)
    v:=((LONGREAL(upper)-LONGREAL(lower)+1.0) * v) + LONGREAL(lower);
    RETURN v;
END getISAAC;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    ranctx32type = RECORD
        a,b,c,d:LONGCARD;
    END;
VAR
    globranctx32 : ranctx32type;

PROCEDURE rot32 (x,k:LONGCARD):LONGCARD;
CONST
    bits=32;
BEGIN
    RETURN ( (x << k) OR (x >> (bits-k)) );
END rot32;

PROCEDURE ranval32( VAR x:ranctx32type ):LONGCARD;
VAR
    e:LONGCARD;
BEGIN
    e   := x.a - rot32(x.b, 27);
    x.a := lcXOR( x.b , rot32(x.c, 17) );
    x.b := x.c + x.d;
    x.c := x.d + e;
    x.d := e + x.a;
    RETURN x.d;
END ranval32;

PROCEDURE raninit32( VAR x:ranctx32type;seed:LONGCARD );
VAR
    i:CARDINAL;
    v:LONGCARD;
BEGIN
    x.a := 0F1EA5EEDH;
    x.b := seed;
    x.c := seed;
    x.d := seed;
    FOR i:=1 TO 20 DO v:=ranval32(x);END;
END raninit32;

CONST
    MAXlongCARDRANGE = LONGREAL(MAX(LONGCARD))+1.0;

PROCEDURE getrandJSF32 (lower,upper:CARDINAL):LONGREAL;
VAR
    v:LONGREAL;
BEGIN
    v:=LONGREAL(ranval32(globranctx32)) / MAXlongCARDRANGE;
    v := (LONGREAL(upper)-LONGREAL(lower)+1.0) * v + LONGREAL(lower);
    RETURN v;
END getrandJSF32;

PROCEDURE initJSF32 ( );
VAR
    seed:LONGCARD;
BEGIN
    Lib.RANDOMIZE;
    seed:=LONGCARD( Lib.RAND() * MAXlongCARDRANGE);
    raninit32(globranctx32,seed);
END initJSF32;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE wXOR (b1,b2 : CARDINAL) : CARDINAL;
TYPE
    lcBITSET = SET OF [0..15];
BEGIN
    RETURN CARDINAL ( (lcBITSET (b1) / lcBITSET (b2)) );
END wXOR;

TYPE
    ranctx16type = RECORD
        a,b,c,d:CARDINAL;
    END;
VAR
    globranctx16 : ranctx16type;

PROCEDURE rot16 (x,k:CARDINAL):CARDINAL;
CONST
    bits=16;
BEGIN
    RETURN ( (x << k) OR (x >> (bits-k)) );
END rot16;

PROCEDURE ranval16( VAR x:ranctx16type ):CARDINAL;
VAR
    e:CARDINAL;
BEGIN
    e   := x.a - rot16(x.b, 13);
    x.a := wXOR( x.b , rot16(x.c, 8) );
    x.b := x.c + x.d;
    x.c := x.d + e;
    x.d := e + x.a;
    RETURN x.d;
END ranval16;

PROCEDURE raninit16( VAR x:ranctx16type;seed:CARDINAL );
VAR
    i:CARDINAL;
    v:CARDINAL;
BEGIN
    x.a := 05EEDH;
    x.b := seed;
    x.c := seed;
    x.d := seed;
    FOR i:=1 TO 20 DO v:=ranval16(x);END;
END raninit16;

CONST
    MAXcardRANGE = LONGREAL(MAX(CARDINAL))+1.0;

PROCEDURE getrandJSF16 (lower,upper:CARDINAL):LONGREAL;
VAR
    v:LONGREAL;
BEGIN
    v:=LONGREAL(ranval16(globranctx16)) / MAXcardRANGE;
    v := (LONGREAL(upper)-LONGREAL(lower)+1.0) * v + LONGREAL(lower);
    RETURN v;
END getrandJSF16;

PROCEDURE initJSF16 ( );
VAR
    seed:CARDINAL;
BEGIN
    Lib.RANDOMIZE;
    seed:=CARDINAL( Lib.RAND() * MAXcardRANGE);
    raninit16(globranctx16,seed);
END initJSF16;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE bXOR (b1,b2 : SHORTCARD) : SHORTCARD;
TYPE
    lcBITSET = SET OF [0..7];
BEGIN
    RETURN SHORTCARD ( (lcBITSET (b1) / lcBITSET (b2)) );
END bXOR;

TYPE
    ranctx8type = RECORD
        a,b,c,d:SHORTCARD;
    END;
VAR
    globranctx8 : ranctx8type;

PROCEDURE rot8 (x,k:SHORTCARD):SHORTCARD;
CONST
    bits=8;
BEGIN
    RETURN ( (x << k) OR (x >> (bits-k)) );
END rot8;

PROCEDURE ranval8( VAR x:ranctx8type ):SHORTCARD;
VAR
    e:SHORTCARD;
BEGIN
    e   := x.a - rot8(x.b, 1);
    x.a := bXOR( x.b , rot8(x.c, 4) );
    x.b := x.c + x.d;
    x.c := x.d + e;
    x.d := e + x.a;
    RETURN x.d;
END ranval8;

PROCEDURE raninit8( VAR x:ranctx8type;seed:SHORTCARD );
VAR
    i:CARDINAL;
    v:SHORTCARD;
BEGIN
    x.a := 0EDH;
    x.b := seed;
    x.c := seed;
    x.d := seed;
    FOR i:=1 TO 20 DO v:=ranval8(x);END;
END raninit8;

CONST
    MAXshortCARDRANGE = LONGREAL(MAX(SHORTCARD))+1.0;

PROCEDURE getrandJSF8 (lower,upper:CARDINAL):LONGREAL;
VAR
    v:LONGREAL;
BEGIN
    v:=LONGREAL(ranval8(globranctx8)) / MAXshortCARDRANGE;
    v := (LONGREAL(upper)-LONGREAL(lower)+1.0) * v + LONGREAL(lower);
    RETURN v;
END getrandJSF8;

PROCEDURE initJSF8 ( );
VAR
    seed:SHORTCARD;
BEGIN
    Lib.RANDOMIZE;
    seed:=SHORTCARD( Lib.RAND() * MAXshortCARDRANGE);
    raninit8(globranctx8,seed);
END initJSF8;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE chkRange (val:LONGCARD;min,max:CARDINAL):BOOLEAN;
BEGIN
    RETURN ( (val >= LONGCARD(min) ) AND (val <= LONGCARD(max) ) );
END chkRange;

PROCEDURE swapCards (VAR a,b:CARDINAL);
VAR
    tmp:CARDINAL;
BEGIN
    tmp := a;
    a   := b;
    b   := tmp;
END swapCards;

CONST
    TSRAND          = 1;
    SHUFFLEDRAND    = 2;
    KNUTHRAND       = 3;
    MARSARAND       = 4;
    LEHMERRAND      = 5;
    ISAACRAND       = 6;
    JSF32RAND       = 7;
    JSF16RAND       = 8;
    JSF8RAND        = 9;
    minRAND         = TSRAND;
    maxRAND         = JSF8RAND;

(* assume lower AND upper are ok *)

PROCEDURE genRAND (lower,upper,generator:CARDINAL): LONGREAL ;
VAR
    v: LONGREAL;
BEGIN
    CASE generator OF
    | TSRAND:
        v := (LONGREAL(upper)-LONGREAL(lower)+1.0) * LONGREAL(Lib.RAND()) + LONGREAL(lower);
    | SHUFFLEDRAND:
        v := LONGREAL( GetRndCardRange( lower,upper ) );
    | KNUTHRAND :
        v := LONGREAL( lower+ RandomCardKNUTH( upper-lower+1 ) );
    | MARSARAND :
        v := getRndMarsa(lower,upper);
    | LEHMERRAND :
        v := (LONGREAL(upper)-LONGREAL(lower)+1.0) * getSimple() + LONGREAL(lower);
    | ISAACRAND:
        v := getISAAC(lower,upper);
    | JSF32RAND:
        v := getrandJSF32(lower,upper);
    | JSF16RAND:
        v := getrandJSF16(lower,upper);
    | JSF8RAND:
        v := getrandJSF8(lower,upper);
    ELSE
        v := LONGREAL(0.0);
    END;
    RETURN v;
END genRAND;

PROCEDURE initRAND (generator:CARDINAL);
BEGIN
    CASE generator OF
    | TSRAND :          Lib.RANDOMIZE;
    | SHUFFLEDRAND :    InitRnd;
    | KNUTHRAND:        RandomizeKNUTH();
    | MARSARAND:        initMarsa;
    | LEHMERRAND:       initSimple;
    | ISAACRAND:        initISAACrng;
    | JSF32RAND:        initJSF32;
    | JSF16RAND:        initJSF16;
    | JSF8RAND:         initJSF8;
    END;
END initRAND;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST (* see showParms() strings *)
    sTRUEPI      = "Exact PI      : "; (* yes, we know it's NOT an EXACT value ! *)
    sCOMPUTEDPI  = "Estimated PI  : ";
    sERRORPI     = "PI error      : ";

PROCEDURE showParms (lower,upper,generator:CARDINAL;
                    verbose,veryverbose,tofile,binmode:BOOLEAN;
                    rndval:LONGREAL;fsize:LONGCARD;filename:ARRAY OF CHAR);
CONST
    sGenerator   = "Generator     : ";
    sRange       = "Range         : ";
    sValue       = "Random value  : ";

    sFilename    = "Filename      : ";
    sFilesize    = "Filesize      : ";

    sEntries     = "Entries       : ";
    sMode        = "Mode          : ";
    wishort      = 1;
    wilong       = 1;
VAR
    S:str128;
    prenl,shownum,showgenparms,shownumverbose:BOOLEAN;
BEGIN
    IF tofile THEN
        shownum       := FALSE;
        showgenparms  := (verbose OR veryverbose);
        shownumverbose:= FALSE;
        prenl         := TRUE;
    ELSE
        shownum       := verbose;
        showgenparms  := veryverbose;
        shownumverbose:= veryverbose;
        prenl         := (verbose OR veryverbose);
    END;

    IF prenl THEN WrLn;END;

    IF shownum THEN IO.WrCard( CARDINAL(rndval),wishort);WrLn; END;

    IF showgenparms THEN
        CASE generator OF
        | TSRAND       : S:="TopSpeed";
        | SHUFFLEDRAND : S:="QD_Rand (shuffled TopSpeed)";
        | KNUTHRAND    : S:="Knuth (additive congruential)";
        | MARSARAND    : S:="Marsaglia";
        | LEHMERRAND   : S:="Lehmer (linear congruential)";
        | ISAACRAND    : S:="Jenkins (I.S.A.A.C.)";
        | JSF32RAND    : S:="Jenkins 32";
        | JSF16RAND    : S:="Jenkins 16";
        | JSF8RAND     : S:="Jenkins 8";
        ELSE
                         S:="?";
        END;
        WrStr(sGenerator);WrStr(S);WrLn;

        WrStr(sRange); WrStr("[");
        IO.WrCard(lower,wishort);WrStr("..");
        IO.WrCard(upper,wishort);WrStr("]");WrLn;

        IF shownumverbose THEN
            WrStr(sValue);IO.WrShtCard(SHORTCARD(rndval),3);WrLn;
        END;
        IF tofile THEN
           WrStr(sFilename);WrStr(filename);WrLn;
           WrStr(sMode);
           IF binmode THEN
               WrStr("binary");WrLn;
               WrStr(sFilesize);
           ELSE
               WrStr("text");WrLn;
               WrStr(sEntries);
           END;
           IO.WrLngCard(fsize,wilong);WrLn;
        END;
    END;
END showParms;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE flushkbd (  );
VAR
    ch:CHAR;
BEGIN
    ch:=IO.RdKey();
    IF ch = CHR(0) THEN
       ch:=IO.RdKey();
    END;
END flushkbd;

PROCEDURE setmode (mode:CARDINAL) : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(mode);
END setmode;

PROCEDURE blackpaper (  );
VAR
    oldink:LONGCARD;
BEGIN
    oldink:=Graph.SetBkColor(Graph._BLACK);
    Graph.ClearScreen(Graph._GWINDOW);
END blackpaper;

PROCEDURE WaitVGAretrace ();
BEGIN
    WHILE (SYSTEM.In(03DAH) AND 08H) # 0 DO
    END;
    WHILE (SYSTEM.In(03DAH) AND 08H) = 0 DO
    END;
END WaitVGAretrace;

PROCEDURE setDAC(index:CARDINAL;r,g,b:BYTE);
CONST
    DACWriteIndex  = 03C8H;
    DACDataRegister= 03C9H;
BEGIN
    SYSTEM.Out (DACWriteIndex,SHORTCARD(index));
    SYSTEM.Out (DACDataRegister, r);
    SYSTEM.Out (DACDataRegister, g);
    SYSTEM.Out (DACDataRegister, b);
END setDAC;

CONST
    black       = LONGCARD(0000000H);
    white       = LONGCARD(03F3F3FH); (* ega/vga range is $00..$3F *)
    red         = LONGCARD(03F0000H);
    green       = LONGCARD(0003F00H);
    darkgreen   = LONGCARD(0002000H);
    blue        = LONGCARD(000003FH);
    darkblue    = LONGCARD(0000020H);
    cyan        = LONGCARD(0002F3FH);
    yellow      = LONGCARD(03F3F00H);
    darkred     = LONGCARD(0200000H);
    orange      = LONGCARD(02F2F00H);

CONST
    mininkindex = 0;
    maxinkindex = 128-1;

PROCEDURE blend (ndx,count:CARDINAL; startink, endink:LONGCARD);
CONST
    egarange = 40H;
    rshift   = LONGCARD(16);
    gshift   = LONGCARD(8);
VAR
    r1,g1,b1:INTEGER;
    r2,g2,b2:INTEGER;
    r,g,b:INTEGER;
    i : CARDINAL;
BEGIN
    r1 := INTEGER (startink >> rshift) MOD egarange;
    g1 := INTEGER (startink >> gshift) MOD egarange;
    b1 := INTEGER (startink          ) MOD egarange;
    r2 := INTEGER (endink   >> rshift) MOD egarange;
    g2 := INTEGER (endink   >> gshift) MOD egarange;
    b2 := INTEGER (endink            ) MOD egarange;

    FOR i := 1 TO count DO
        r := r1 + ((r2 - r1) * INTEGER(i) ) DIV INTEGER(count);
        g := g1 + ((g2 - g1) * INTEGER(i) ) DIV INTEGER(count);
        b := b1 + ((b2 - b1) * INTEGER(i) ) DIV INTEGER(count);
        IF r < 0 THEN r:=0; END;
        IF g < 0 THEN g:=0; END;
        IF b < 0 THEN b:=0; END;
        IF r >= egarange THEN r:=egarange-1; END;
        IF g >= egarange THEN g:=egarange-1; END;
        IF b >= egarange THEN b:=egarange-1; END;
        setDAC(ndx+i-1,BYTE(r),BYTE(g),BYTE(b));
    END;
END blend;

TYPE
    palettetype = (redpal,greenpal,
                  rpal,gpal,
                  rougepal,vertpal,
                  grispal,defaultpal);
CONST
    minpal = ORD(redpal);
    maxpal = ORD(defaultpal);
    minink = 0;

PROCEDURE newPalette (maxink:CARDINAL;pal:palettetype);
VAR
    ndx : CARDINAL;
    n   : CARDINAL;
    i,j : CARDINAL;
    r,g,b:BYTE;
    ir,ig,ib,di:CARDINAL;
BEGIN
    CASE maxink OF
    | 16-1:  di:=16;
    | 256-1: di:=1;
    END;
    WaitVGAretrace(); (* reduce noise on screen *)
    CASE pal OF
    | redpal:
        ndx := 0;         n := 8 DIV di;
        blend (ndx,n,black   ,02F0804H);
        INC(ndx,n);       n := 120 DIV di;
        blend (ndx,n,02F0804H,03F2F2FH);
        INC(ndx,n);       n := 128 DIV di;
        blend (ndx,n,03F2F2FH,white);
    | greenpal:
        ndx := 0;         n := 8 DIV di;
        blend (ndx,n,black   ,0082F04H);
        INC(ndx,n);       n := 128 DIV di;
        blend (ndx,n,0082F04H,01F3F1FH);
        INC(ndx,n);       n := 120 DIV di;
        blend (ndx,n,01F3F1FH,white);
    | rpal:
        ndx := 0;         n := 8 DIV di;
        blend (ndx,n,black   ,02F0804H);
        INC(ndx,n);       n := 128 DIV di;
        blend (ndx,n,02F0804H,yellow);
        INC(ndx,n);       n := 120 DIV di;
        blend (ndx,n,yellow,white);
    | gpal:
        ndx := 0;         n := 8 DIV di;
        blend (ndx,n,black ,0082F04H);
        INC(ndx,n);       n := 128 DIV di;
        blend (ndx,n,0082F04H,yellow);
        INC(ndx,n);       n := 120 DIV di;
        blend (ndx,n,yellow,white);
    | rougepal:
        n := 0;
	    FOR i:= 0 TO (255 DIV di) DO
	        FOR j:= 1 TO 4 DO
                r:= BYTE( (32+i >> 2) MOD 255) ; g:= 0; b:= 0;
                setDAC(n,r,g,b);
                IF n < maxink THEN INC(n); END;
            END;
        END;
    | vertpal:
        n := 0;
	    FOR i:= 0 TO (255 DIV di) DO
	        FOR j:= 1 TO 4 DO
                r:=0; g:= BYTE( (32+i >> 2) MOD 255) ; b:= 0;
                setDAC(n,r,g,b);
                IF n < maxink THEN INC(n); END;
            END;
        END;
    | grispal:
        n := 0;
	    FOR i:= 0 TO (255 DIV di) DO
	        FOR j:= 1 TO 4 DO
                r:= BYTE( (32+i >> 2) MOD 255) ; g:= r; b:= r;
                setDAC(n,r,g,b);
                IF n < maxink THEN INC(n); END;
            END;
        END;
    | defaultpal:
        ; (* bad idea : this palette is ugly *)
    END;
    setDAC(minink  ,00H,00H,00H);
    setDAC(maxink  ,3FH,3FH,3FH);
END newPalette;

(* ------------------------------------------------------------ *)

(* unscientific but obvious visual check of randomness *)

PROCEDURE buildscreen (generator:CARDINAL;immediate,fixForByte,DEBUG:BOOLEAN;
                      pal:palettetype):LONGCARD;
CONST
    pauseInSeconds = 10;
VAR
    x,y,ink,i,k:CARDINAL;
    xmin,xmax,ymin,ymax,maxink:CARDINAL;
    v:Graph.VideoConfig;
    pairs:LONGCARD;
BEGIN
    Graph.GetVideoConfig(v);
    xmin:=0;
    xmax:=v.numxpixels-1;
    ymin:=0;
    ymax:=v.numypixels-1;
    maxink:=v.numcolors-1;
    CASE maxink OF
    | 16-1:  k:=1;
    | 256-1: k:=4;
    END;

    newPalette(maxink,pal);

    IF DEBUG THEN
        blackpaper();
        FOR x:=minink TO maxink DO
            Graph.Line(x,ymin,x,ymax,x);
        END;
        Lib.Delay(pauseInSeconds * 1000);
    END;

    IF fixForByte THEN
        IF ymax > 255 THEN ymax:=255; END;
        xmax := ymax;
    END;

    pairs:=0;
    blackpaper();
    LOOP
        x:=CARDINAL ( genRAND(xmin,xmax,generator) );
        y:=CARDINAL ( genRAND(ymin,ymax,generator) );

        INC(pairs);
        IF immediate THEN
           Graph.Plot(x,y,maxink);
        ELSE
           FOR i:=1 TO k DO
               ink:=Graph.Point(x,y);
               IF ink < maxink THEN
                   Graph.Plot(x,y,ink+1);
               END;
           END;
        END;
        IF IO.KeyPressed() THEN EXIT; END;
    END;
    flushkbd();
    RETURN pairs;
END buildscreen;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    vmode320    = Graph._MRES16COLOR;
    vmode640    = Graph._VRES16COLOR;
    vmode320alt = Graph._MRES256COLOR;
    txtmode     = Graph._DEFAULTMODE;
CONST
    minNum    = MIN(SHORTCARD);
    maxNum    = MAX(SHORTCARD);
    maxCardNum= MAX(CARDINAL);
    oldwarmup = 32;
    warmup    = 128;
    msgWait   = "Please wait... ";
CONST
    minparm = 1;
    maxparm = 4;
VAR
    lastparm:CARDINAL;
    parm:ARRAY[minparm..maxparm] OF str128;
VAR
    parmcount,i,opt,j   : CARDINAL;
    S,R           : str128;
    warmupvalue,generator,lower,upper : CARDINAL;
    DEBUG,ok:BOOLEAN;
    fixForByte,flagMissing,tofile:BOOLEAN;
    verbose,veryverbose,overwrite,computePI,binmode : BOOLEAN;
    hitswithinradius,lc,fsize : LONGCARD;
    rndfile:str128;
    hout : FIO.File;
    x,y,delta,rndval,PI,RNDPI:LONGREAL;
    ch:CHAR;
    showmode:(NOSHOW,IMMEDIATESHOW,PROGRESSIVESHOW);
    vmode:CARDINAL;
    pal:palettetype;
    pairs:LONGCARD;
    maxi,e:CARDINAL;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck:=FALSE;

    binmode   := TRUE;
    verbose   := FALSE;
    veryverbose:=FALSE;
    generator := TSRAND;
    overwrite := FALSE;
    computePI := FALSE;
    warmupvalue:=warmup;
    (* infra init is useless *)
    tofile    := FALSE;
    flagMissing:=FALSE;
    fixForByte := FALSE;
    lower     := minNum;
    upper     := maxNum;
    fsize     := 0;
    rndfile   := "";
    showmode  := NOSHOW;
    pal       := palettetype(minpal);
    DEBUG     := FALSE;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    lastparm:=minparm-1;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "V"+delim+"VERBOSE"+delim+
                                  "VV"+delim+"VERYVERBOSE"+delim+
                                  "A"+delim+"ASCII"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "G:"+delim+"GENERATOR:"+delim+
                                  "PI"+delim+"MONTECARLO"+delim+
                                  "W"+delim+"WARMUP"+delim+
                                  "T"+delim+
                                  "TT"+delim+
                                  "Y"+delim+
                                  "YY"+delim+
                                  "U"+delim+
                                  "UU"+delim+
                                  "P:"+delim+"PALETTE:"+delim+
                                  "M"+delim+"MISSING"+delim+
                                  "XY"+delim+"FIX"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3 :  abort(errHelp,"");
            | 4,5 :    verbose     := TRUE;
            | 6,7 :    veryverbose := TRUE;
            | 8,9 :    binmode     := FALSE; (* generator := SHUFFLEDRAND; obsolete *)
            | 10,11:   overwrite   := TRUE;
            | 12,13:   generator   := TSRAND;
                       IF GetLongCard(R,lc)=FALSE THEN abort(errBadNum,S);END;
                       IF lc > MAX(CARDINAL) THEN abort(errBadNum,S);END;
                       generator   := CARDINAL(lc);
                       IF (generator < minRAND) OR (generator > maxRAND) THEN abort(errGenerator,S);END;
            | 14,15:   computePI   := TRUE;
            | 16,17:   warmupvalue := oldwarmup;    (* obsolete *)
            | 18:      showmode    := IMMEDIATESHOW;   vmode:=vmode320;
            | 19:      showmode    := PROGRESSIVESHOW; vmode:=vmode320;
            | 20:      showmode    := IMMEDIATESHOW;   vmode:=vmode640;
            | 21:      showmode    := PROGRESSIVESHOW; vmode:=vmode640;
            | 22:      showmode    := IMMEDIATESHOW;   vmode:=vmode320alt;
            | 23:      showmode    := PROGRESSIVESHOW; vmode:=vmode320alt;
            | 24,25:   IF GetLongCard(R,lc)=FALSE THEN abort(errBadNum,S);END;
                       IF lc > MAX(CARDINAL) THEN abort(errBadNum,S);END;
                       j:=CARDINAL(lc);
                       IF ( j < minpal ) OR ( j > maxpal) THEN abort(errPalette,S);END;
                       pal := palettetype (j);
            | 26,27:   flagMissing := TRUE;
            | 28,29:   fixForByte   := TRUE;
            | 30:      DEBUG       := TRUE;
            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParameter,S);END;
            IF lastparm = 4 THEN
                parm[lastparm]:=S; (* keep filename CASE *)
            ELSE
                parm[lastparm]:=R;
            END;
        END;
    END;

    IF NOT(binmode) THEN computePI:=FALSE;END; (* better safe than sorry ! *)

    CASE lastparm OF
    | 0 :
        IF flagMissing THEN abort(errMissing,"2");END;
        IF showmode = NOSHOW THEN abort(errBadSyntax,"3");END;
    | 2 :
        IF flagMissing THEN abort(errMissing,"2");END;
        IF fixForByte THEN abort(errByte,"3");END;

        R:=parm[minparm];
        IF GetLongCard(R,lc)=FALSE THEN abort(errBadNum,R);END;
        IF chkRange(lc,minNum,maxNum)=FALSE THEN abort(errBinRange,R);END;
        lower := CARDINAL(lc);

        R:=parm[minparm+1];
        IF GetLongCard(R,lc)=FALSE THEN abort(errBadNum,R);END;
        IF chkRange(lc,minNum,maxNum)=FALSE THEN abort(errBinRange,R);END;
        upper := CARDINAL(lc);

        IF showmode # NOSHOW THEN abort(errBadSyntax,"1");END;
        (* IF lower=upper THEN abort(errInterval,""); END; //V11A was overzealous *)
        IF upper < lower THEN swapCards(lower,upper);END;
        tofile :=FALSE;
    | 4 :
        IF fixForByte THEN abort(errByte,"3");END;

        IF binmode THEN
            maxi:=maxNum;
            e:=errBinRange;
        ELSE
            maxi:=maxCardNum;
            e:=errCardRange;
        END;
        R:=parm[minparm];
        IF GetLongCard(R,lc)=FALSE THEN abort(errBadNum,R);END;
        IF chkRange(lc,minNum,maxi)=FALSE THEN abort(e,R);END;
        lower := CARDINAL(lc);

        R:=parm[minparm+1];
        IF GetLongCard(R,lc)=FALSE THEN abort(errBadNum,R);END;
        IF chkRange(lc,minNum,maxi)=FALSE THEN abort(e,R);END;
        upper := CARDINAL(lc);

        R:=parm[minparm+2];
        IF same(R,"*") THEN
            fsize:=MAX(LONGINT);
        ELSE
            FOR j:=1 TO 2 DO
                CASE j OF
                | 1: ch:= ",";
                | 2: ch:= ".";
                END;
                LOOP
                    IF Str.CharPos(R,ch)=MAX(CARDINAL) THEN EXIT;END;
                    Str.Subst(R,ch,"");
                END;
            END;
            IF GetLongCard(R,fsize)=FALSE THEN abort(errBadNum,R);END;
            IF fsize > MAX(LONGINT) THEN abort(err2gigas,R);END;
            IF fsize = 0 THEN abort(errZeroLength,"");END;
        END;

        R:=parm[minparm+3];
        IF chkJoker(R) THEN abort(errJoker,R);END;
        Str.Copy(rndfile,R);

        IF showmode # NOSHOW THEN abort(errBadSyntax,"2");END;
        IF NOT(overwrite) THEN
            IF FIO.Exists(rndfile) THEN abort(errExists,rndfile);END;
        END;
        IF upper < lower THEN swapCards(lower,upper);END;
        IF computePI THEN
            IF ( (lower # minNum) OR (upper # maxNum) ) THEN abort(errRangePi,"");END;
            IF ODD(fsize) THEN abort(errOddPi,"");END;
        END;
        IF fsize = MAX(LONGINT) THEN
            IF ODD(fsize) THEN DEC(fsize);END;
        END;
        tofile := TRUE;
        hitswithinradius := 0;

        IF flagMissing THEN
            IF chkvalues(lower,upper)=FALSE THEN abort(errRangeForMissing,"");END;
        END;

    ELSE
        abort(errSyntax,"");
    END;

    (* ------------------------------------------------------------ *)

    initRAND(generator);

    FOR i := 1 TO warmupvalue DO
        rndval := LONGREAL( Lib.RAND() );
        rndval := LONGREAL( GetRnd() );
    END;

    PI := LONGREAL(4.0) *  MATHLIB.ATan(1.0) ;

    (* ------------------------------------------------------------ *)

    IF showmode # NOSHOW THEN
        IF setmode(vmode) THEN
            pairs := buildscreen (generator,
                                 (showmode=IMMEDIATESHOW),fixForByte,DEBUG,
                                 pal);
            i:=errNone;
        ELSE
            i:=errDisplay;
        END;
        ok:=setmode(txtmode);
        WrStr("Random pairs generated : ");IO.WrLngCard(pairs,10);WrLn;
        abort(i,"");
    END;

    (* ------------------------------------------------------------ *)

    IF NOT(tofile) THEN
        rndval:=genRAND(lower,upper,generator);

        showParms (lower,upper,generator,verbose,veryverbose,tofile,binmode,
                  rndval,fsize,rndfile);
        (* no abort() here ! *)
        Lib.SetReturnCode(SHORTCARD(rndval));
        HALT;
    ELSE
        IF flagMissing THEN initvalues; END;

        verbose:=TRUE;
        showParms (lower,upper,generator,verbose,veryverbose,tofile,binmode,
                  rndval,fsize,rndfile);

        video(msgWait,TRUE);

        hout:=FIO.Create(rndfile);
        FIO.AssignBuffer(hout,iobuffer);
        lc:=0;
        LOOP
            INC(lc);
            IF lc > fsize THEN EXIT; END;
            rndval:=genRAND(lower,upper,generator);
            IF flagMissing THEN updatevalue( LONGCARD(rndval) );END;
            IF binmode THEN
                FIO.WrBin(hout,SHORTCARD(rndval),SIZE(SHORTCARD));
            ELSE
                FIO.WrCard(hout,CARDINAL(rndval),5);FIO.WrLn(hout);
            END;
            IF computePI THEN
                IF ODD(lc) THEN (* note lc is incremented ! *)
                    x:=LONGREAL(rndval) / LONGREAL(256.0);
                ELSE
                    y:=LONGREAL(rndval) / LONGREAL(256.0);
                    delta := (x * x) + (y * y)  ;
                    IF delta <= LONGREAL(1.0) THEN INC (hitswithinradius);END;
                END;
            END;
        END;
        FIO.Close(hout);

        video(msgWait,FALSE);

        IF computePI THEN
            WrLn;
            WrStr(sTRUEPI);
            Str.FixRealToStr( LONGREAL(PI)   ,9,S,ok);
            WrStr(S);WrLn;

            RNDPI:=LONGREAL(4.0) * LONGREAL(hitswithinradius) / LONGREAL(fsize DIV 2); (* we got PAIRS of rnd numbers *)
            WrStr(sCOMPUTEDPI);
            Str.FixRealToStr( LONGREAL(RNDPI),9,S,ok);
            WrStr(S);WrLn;

            WrStr(sERRORPI);
            Str.FixRealToStr(  (100.0 * (ABS(PI-RNDPI) / PI)) ,9,S,ok);
            WrStr(S);WrStr(" percent");WrLn;
        END;

        IF flagMissing THEN WrLn;showmissing;END;
    END;

    abort(errNone,"");
END rndgen.


