(* ---------------------------------------------------------------
Title         Q&D Roman Numerals Converter
Overview
Notes         now, everyone will be able to check MGM's movies dates ;-)
Bugs
Wish List

--------------------------------------------------------------- *)

MODULE Roman;

IMPORT Str;
IMPORT Lib;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "ROMAN";
    ProgTitle     = "Q&D Roman Numerals Converter";
    ProgVersion   = "v1.0";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParmOverflow = 3;
    errBadNum       = 4;
    errRomanRange   = 5;
    errBadRomanNum  = 6;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <number>... [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" <first..last> [option]..."+nl+
nl+
"With syntax 1, this program converts a number from decimal to roman numerals"+nl+
"and from roman numerals to decimal. With syntax 2, it converts a range,"+nl+
"provided it is a valid one."+nl+
nl+
"   -l show output roman as lowercase roman (default is uppercase)"+nl+
"   -u show input roman as uppercase (default is lowercase)"+nl+
nl+
"a) Note valid range for roman numerals is [1..3999]."+nl+
nl+
"Examples : "+ProgEXEname+" -l 1963 3888"+nl+
"           "+ProgEXEname+" -u mcmlxiii mmmdccclxxxviii"+nl+
"           "+ProgEXEname+" 1..3999"+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(helpmsg);
    | errOption :
        S:='Illegal "~" option !';
    | errParmOverflow:
        S:='"~" is one parameter too many !';
    | errBadNum:
        S:='Illegal "~" number !';
    | errRomanRange:
        S:='"~" does not belong to valid roman numerals range !';
    | errBadRomanNum:
        S:='"~" does not look like a valid roman number !';
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp : ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    strERROR = "??? ERROR ???";

PROCEDURE padstr (VAR R:ARRAY OF CHAR;wi:INTEGER;padchar:CHAR   );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
END padstr;

PROCEDURE bignum2str (v:LONGCARD;base:CARDINAL;wi:INTEGER;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
BEGIN
    Str.CardToStr( v,R,base,ok);
    IF ok THEN
        padstr(R,wi,padchar);
    ELSE
        R:=strERROR;
    END;
    RETURN R;
END bignum2str;

PROCEDURE num2str (v:CARDINAL;base:CARDINAL;wi:INTEGER;padchar:CHAR):str16;
BEGIN
    RETURN bignum2str( LONGCARD(v),base,wi,padchar);
END num2str;

PROCEDURE isValidDecimalNumber (VAR v:LONGCARD; S:ARRAY OF CHAR ):BOOLEAN;
CONST
    decdigits = digits;
VAR
    rc,ok:BOOLEAN;
BEGIN
    rc:=FALSE;
    IF verifyString(S,decdigits) THEN
        v:=Str.StrToCard( S, 10, ok);
        rc:=ok;
    END;
    RETURN rc;
END isValidDecimalNumber;

PROCEDURE isValidRomanNumber (S:ARRAY OF CHAR   ):BOOLEAN ;
CONST
    romandigits = "IVXLCDMivxlcdm"; (* how clever ! :-) *)
BEGIN
    RETURN verifyString(S,romandigits);
END isValidRomanNumber;

PROCEDURE isValidRange (VAR mini,maxi:LONGCARD;  S:ARRAY OF CHAR  ):BOOLEAN;
CONST
    seprange = "..";
VAR
    rc:BOOLEAN;
    p:CARDINAL;
    N:str128; (* oversized for safety *)
BEGIN
    rc:=FALSE;
    p:=Str.Pos(S,seprange);
    IF p # MAX(CARDINAL) THEN
        Str.Copy(N,S);
        N[p]:=charnull;
        IF isValidDecimalNumber(mini,N) THEN
            Str.Copy(N,S);
            Str.Delete(N,0,p+2); (* delete ".." too *)
            rc:=isValidDecimalNumber(maxi,N);
        END;
    END;
    RETURN rc;
END isValidRange;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minRomanNumber = 1;
    maxRomanNumber = 4000-1;

PROCEDURE isInRomanRange (v:LONGCARD):BOOLEAN;
BEGIN
    RETURN ( (v >= minRomanNumber) AND (v <= maxRomanNumber) );
END isInRomanRange;

(* adapted from my old 6502 PhilPrinter code, 1981 *)

CONST
    maxromanwidth    = 15; (*  *)
    maxdecromanwidth = 4;  (* 3999 *)

(* yes, we know there are a few inscriptions with faulty "iiii" instead of "iv" *)

PROCEDURE dec2rom (v:LONGCARD;wi:CARDINAL;flagLOWER:BOOLEAN):str16;
CONST
    numeral = "MCXI";
    special = " DLV";
VAR
    N,R:str16;
    i,j,k:CARDINAL;
BEGIN
    IF isInRomanRange(v) THEN
        Str.Copy(N,bignum2str(v,10,4,"0"));
        R:="";
        FOR i:=(1-1) TO (maxdecromanwidth-1) DO
            k:=ORD( N[i] ) - ORD( "0" );
            CASE k OF
            | 0:     ;
            | 1,2,3: FOR j:=1 TO k DO Str.Append(R,numeral[i]); END;
            | 4:     Str.Append(R,numeral[i]);
                     Str.Append(R,special[i]);
            | 5:     Str.Append(R,special[i]);
            | 6,7,8: Str.Append(R,special[i]);
                     FOR j:=1 TO (k-5) DO Str.Append(R,numeral[i]); END;
            | 9:     Str.Append(R,numeral[i]);
                     Str.Append(R,numeral[i-1]);
            END;
        END;
        padstr(R,wi,space);
        IF flagLOWER THEN Str.Lows(R);END;
    ELSE
        R:=strERROR;
    END;
    RETURN R;
END dec2rom;

(* yes, we could write a more clever sub but this ugly code WORKS ! :-) *)

PROCEDURE rom2dec (S:ARRAY OF CHAR ):LONGCARD;
VAR
    N:str16;
    i:CARDINAL;
BEGIN
    Str.Lows(S);
    FOR i := minRomanNumber TO maxRomanNumber DO
        Str.Copy(N,dec2rom( LONGCARD(i),1,TRUE));
        IF same(S,N) THEN RETURN LONGCARD(i);END;
    END;
    RETURN MAX(LONGCARD);
END rom2dec;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 32;
VAR
    parm : ARRAY [minparm..maxparm] OF str128;
    lastparm,parmcount,i,opt:CARDINAL;
    S,R:str128;
    flagLOWERdst,flagLOWERsrc:BOOLEAN;
    mode : (convert,showrange);
    v,mini,maxi:LONGCARD;
    okdec,okrom:BOOLEAN;
    pass:CARDINAL;
    currval:LONGCARD;
    maxlen : CARDINAL;
BEGIN
    WrLn;
    parmcount := Lib.ParamCount();
    (* IF parmcount=0 THEN abort(errHelp,"");END; *)

    flagLOWERdst := FALSE;
    flagLOWERsrc := TRUE;

    lastparm  := minparm-1;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "L"+delim+"LOWER"+delim+"LOWERCASE"+delim+"LOWS"+delim+
                                   "U"+delim+"UPPER"+delim+"UPPERCASE"+delim+"CAPS"
                               );
            CASE opt OF
            | 1,2,3   :    abort(errHelp,"");
            | 4,5,6,7 :    flagLOWERdst := TRUE;
            | 8,9,10,11:   flagLOWERsrc := FALSE;
            ELSE           abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy(parm[lastparm],S); (* keep case *)
        END;
    END;

    mode:=convert;
    CASE lastparm OF
    | minparm-1 :
        abort(errHelp,"");
    | minparm :
        S:=parm[minparm];
        IF isValidRange(mini,maxi,S) THEN
           mode:=showrange;
        END;
    ELSE
        ;
    END;
    CASE mode OF
    | convert:
        FOR pass:=1 TO 2 DO
            FOR i:=minparm TO lastparm DO
                S:=parm[i];
                CASE pass OF
                | 1:
                    okdec:=isValidDecimalNumber(v,S);
                    IF okdec THEN
                        IF isInRomanRange(v)=FALSE THEN abort(errRomanRange,S);END;
                    END;
                    okrom:=isValidRomanNumber(S);
                    IF okrom THEN
                        IF rom2dec(S)=MAX(LONGCARD) THEN abort(errBadRomanNum,S);END;
                    END;
                    IF ( (okdec = FALSE) AND (okrom = FALSE) ) THEN abort(errBadNum,S);END;
                | 2:
                    (* both can't be true at the same time, eh eh *)
                    IF isValidDecimalNumber(v,S) THEN
                        okdec:=isValidDecimalNumber(v,S);
                        S:="~  ==>  ~";
                        Str.Subst(S,"~",bignum2str(v,10,maxdecromanwidth,space) );
                        Str.Subst(S,"~",dec2rom(v,1,flagLOWERdst) );
                        WrStr(S);WrLn;
                    END;
                    IF isValidRomanNumber(S) THEN
                        IF flagLOWERsrc THEN
                            Str.Lows(S);
                        ELSE
                            Str.Caps(S);
                        END;
                        v:=rom2dec(S);
                        padstr(S,maxromanwidth,space);
                        R:="~  ==>  ~";
                        Str.Subst(R,"~",S);
                        Str.Subst(R,"~",bignum2str(v,10,maxdecromanwidth,space));
                        WrStr(R);WrLn;
                    END;
                END;
            END;
        END;
    | showrange:
        S:=parm[minparm];
        IF isInRomanRange(mini)=FALSE THEN abort(errRomanRange,S);END;
        IF isInRomanRange(maxi)=FALSE THEN abort(errRomanRange,S);END;

        (* maxlen:=0; *)
        currval:=mini;
        LOOP
            (*
            Str.Copy(S,dec2rom(currval,1,flagLOWER));
            i:=Str.Length(S);
            IF i > maxlen THEN maxlen:=i;END;
            *)

            (*
            S:="~ ~ ~";
            Str.Subst(S,"~",bignum2str(currval,10,maxdecromanwidth,space));
            Str.Subst(S,"~",dec2rom(currval, maxromanwidth,TRUE));
            Str.Subst(S,"~",dec2rom(currval,-maxromanwidth,FALSE));
            WrStr(S);WrLn;
            *)

            S:="~  ==>  ~";
            Str.Subst(S,"~",bignum2str(currval,10,maxdecromanwidth,space));
            Str.Subst(S,"~",dec2rom(currval, 1,flagLOWERdst));
            WrStr(S);WrLn;

            IF maxi < mini THEN
                DEC(currval);
                IF currval < maxi THEN EXIT; END;
            ELSE
                INC(currval);
                IF currval > maxi THEN EXIT; END;
            END;
        END;
        (* WrStr("maxlen = ");WrStr(num2str(maxlen,10,1,""));WrLn; *)
    END;

    abort(errNone,"");
END Roman.