(* ---------------------------------------------------------------
Title         Q&D Show "hidden" words
Overview
Usage         see help
Notes
Bugs
Wish List     transformer a:# pour limiter l'affichage

--------------------------------------------------------------- *)

MODULE ShowHIDN;

IMPORT IO;
IMPORT Lib;
IMPORT Str;
IMPORT FIO;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory, qdSplitAllPath,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM IO IMPORT WrStr,WrLn;

(* ------------------------------------------------------------ *)

CONST
    progPREPROC   = "PREPHIDN";
    progREVEAL    = "FINDHIDN";
    progSHOW      = "SHOWHIDN";
CONST
    ProgEXEname   = progSHOW;
    ProgTitle     = 'Q&D Show "hidden" words';
    ProgVersion   = "v1.0c";
    ProgCopyright = "by PhG";

CONST
    errNone         = 0;
    errHelp         = 1;
    errUnknown      = 2;
    errTooMany      = 3;
    errNotFound     = 4;
    errConversion   = 5;
    errStart        = 6;
    errCount        = 7;
    errOffsetTooBig = 8;
    errLines        = 9;
    errParmOverflow = 10;
    errAborted      = 255;

PROCEDURE abort (e : CARDINAL; einfo:ARRAY OF CHAR );
CONST
    msg = ProgTitle+" "+ProgVersion+" "+ProgCopyright+nl+
nl+
"Syntax : "+ProgEXEname+" <textfile> [-a:lines] [-s:offset] [-c:count] [-u]"+nl+
nl+
'This program shows "hidden" words in a DOS ASCII text file'+nl+
"provided file has *already* been processed with "+progPREPROC+" (un)utility."+nl+
"Required values should have been found using "+progREVEAL+" (un)utility."+nl+
"The three programs have been written in order to confirm"+nl+
'the nonsense of "hidden" words in the Bible or in any long text.'+nl+
nl+
"    -a:# number of lines to align display on"+nl+
"    -s:# starting offset (1-based)"+nl+
"    -c:# count of characters to skip"+nl+
"    -u   uppercase"+nl+
"    -sa  do not sort filenames alphabetically"+nl+
nl+
"a) Note offset is 1-based : position 1 is thus first byte in file."+nl+
"b) -a:# should be searched word length."+nl+
nl+
"Example : "+ProgEXEname+" TEST -s:13 -c:3"+nl;

VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(msg);
    | errUnknown :
        S := "Unknown ~ option !";
    | errTooMany :
        S := "Too many files match ~ specification !";
    | errNotFound :
        S := "No file matches ~ specification !";
    | errConversion :
        S := "~ parameter is not a LONGCARD !";
    | errStart :
        S := "~ starting offset cannot be 0 !";
    | errCount :
        S := "~ count of characters to skip cannot be 0 !";
    | errOffsetTooBig :
        S := "~ starting offset is bigger than file size !";
    | errLines :
        S := "~ number of lines cannot be 0 !";
    | errParmOverflow:
        S := "~ is one parameter too many !";
    | errAborted:
        S := "Aborted by user !";
    ELSE
        S := "How did you get THERE ?";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : "); WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    IObufferSize = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = IObufferSize;
VAR
    IObufferIn : ARRAY [1..IObufferSize] OF BYTE;

(* ------------------------------------------------------------ *)

PROCEDURE WriteChar ( c : CHAR );
VAR
    ok:BOOLEAN;
    hx:str16; (* oversized *)
BEGIN
    IF ( ORD(c) < ORD(blank)) OR ( ORD(c) = 255) THEN
       Str.CardToStr ( LONGCARD(ORD(c)),hx,16,ok);
       IF Str.Length(hx) < 2 THEN Str.Prepend(hx,"0");END;
       Str.Lows(hx);
       WrStr("($"); WrStr(hx);WrStr(")");
    ELSE
       IO.WrChar( c );
    END;
END WriteChar;

(* ------------------------------------------------------------ *)

(* yes, yes, we should have returned an error code instead of aborting inproc *)

PROCEDURE doShow (source : ARRAY OF CHAR;
                  startoffset : LONGCARD; countwidth : LONGCARD;
                  flagAlign,flagUpper : BOOLEAN ; nLines : LONGCARD);
VAR
    c   : CHAR;
    got : CARDINAL;
    filesize : LONGCARD;
    i : LONGCARD;
    k : LONGCARD;
    IOhndIn    : FIO.File;	
BEGIN
    IOhndIn := FIO.OpenRead(source);
    FIO.AssignBuffer(IOhndIn,IObufferIn);
    filesize := FIO.Size(IOhndIn);
    DEC(startoffset); (* was in bytes from 1.. *)
    IF startoffset >= filesize THEN
        FIO.Close(IOhndIn);
        abort(errOffsetTooBig,"-s:#");
    END;

    FIO.Seek(IOhndIn,startoffset);
    CASE flagAlign OF
    | FALSE : (* use only chars *)
        LOOP
            got := FIO.RdBin(IOhndIn,c,1);
            WriteChar( c );
            INC (startoffset,countwidth);
            IF startoffset >= filesize THEN EXIT; END;
            FIO.Seek(IOhndIn,startoffset);
        END;
    | TRUE : (* show all chars *)
        k := 0;
        LOOP
            FOR i := 1 TO countwidth DO
                got := FIO.RdBin(IOhndIn,c,1);
                IF got = 0 THEN EXIT; END;
                IF flagUpper THEN UpperCase(c);END;
                WriteChar( c );
            END;
            WrLn;
            INC (k);
            IF k = nLines THEN EXIT; END;
        END;
    END;
    FIO.Close(IOhndIn);
    WrLn;
END doShow;

(* ------------------------------------------------------------ *)

TYPE
    f8e3 = ARRAY [0..8+1+3-1] OF CHAR;
CONST
    firstFile = 1;
    lastFile  = 2000;
    maxFile   = lastFile-firstFile+1;
VAR
    fileArray : ARRAY [firstFile..lastFile] OF f8e3;

PROCEDURE buildList (VAR basepath:ARRAY OF CHAR;
                    spec:ARRAY OF CHAR):CARDINAL;
VAR
    countFile : CARDINAL;
    found     : BOOLEAN;
    entry     : FIO.DirEntry;
    u,path,f8,e3    : str128; (* oversized just in case *)
BEGIN
    qdSplitAllPath(spec,u,path,f8,e3);
    Str.Concat(basepath,u,path);

    countFile := 0;
    found := FIO.ReadFirstEntry(spec,allfiles,entry);
    WHILE found DO
        IF countFile = maxFile THEN RETURN MAX(CARDINAL); END;
        fileArray[firstFile+countFile]:=f8e3(entry.Name);
        Str.Caps(fileArray[firstFile+countFile]); (* useless ! *)
        INC (countFile);
        found :=FIO.ReadNextEntry(entry);
    END;
    RETURN countFile;
END buildList;

PROCEDURE isLessByName (i,j:CARDINAL):BOOLEAN;
VAR
    si,sj:f8e3;
BEGIN
    si:=fileArray[i];
    sj:=fileArray[j];
    RETURN ( Str.Compare(si,sj) < 0 );
END isLessByName;

PROCEDURE doSwapNames (i,j:CARDINAL);
VAR
    si,sj,tmp:f8e3;
BEGIN
    si:=fileArray[i];
    sj:=fileArray[j];
    tmp:=si;
    si:=sj;
    sj:=tmp;
    fileArray[i]:=si;
    fileArray[j]:=sj;
END doSwapNames;

(* ------------------------------------------------------------ *)

VAR
    DEBUG,flagAlign,flagUpper,alphasort : BOOLEAN;
    startoffset : LONGCARD;
    countskip : LONGCARD;
    nLines : LONGCARD;

    parms : CARDINAL;
    i : CARDINAL;
    S : str128;
    source,basepath : str128;

    filecount:CARDINAL;
    path : str128;
    opt : CARDINAL;
    state : (waiting,gotsource);
BEGIN
    FIO.IOcheck:=FALSE;
    WrLn;

    flagAlign   := FALSE;
    flagUpper   :=FALSE;
    startoffset := 1;
    countskip   := 1;
    nLines      := 1;
    alphasort   := TRUE;
    DEBUG       := FALSE;

    state       := waiting;

    parms := Lib.ParamCount();
    IF parms = 0 THEN abort (errHelp,""); END;
    FOR i := 1 TO parms DO
        Lib.ParamStr(S,i); cleantabs(S);
        IF isOption(S) THEN
            UpperCase (S);
            opt := GetOptIndex (S,"A:"+delim+"ALIGN:"+delim+
                                  "S:"+delim+"START:"+delim+
                                  "C:"+delim+"COUNT:"+delim+
                                  "U"+delim+"UPPERCASE"+delim+
                                  "?"+delim+"H"+delim+"HELP"+delim+
                                  "SA"+delim+"NOALPHASORT"+delim+
                                  "DEBUG"
                               );
            CASE opt OF
            | 1,2 :
                flagAlign := TRUE;
                IF GetLongCard(S,nLines)=FALSE THEN abort(errConversion,S);END;
            | 3,4 :
                IF GetLongCard(S,startoffset)=FALSE THEN abort(errConversion,S);END;
            | 5,6 :
                IF GetLongCard(S,countskip)=FALSE THEN abort(errConversion,S);END;
            | 7,8:
                flagUpper:=TRUE;
            | 9,10,11:
                abort(errHelp,"");
            | 12,13:
                alphasort:=FALSE;
            | 14:
                DEBUG := TRUE;
            ELSE
                abort(errUnknown,S);
            END;
        ELSE
            CASE state OF
            |waiting :    UpperCase (S); source:=S;
            | gotsource:  abort(errParmOverflow,S);
            END;
            INC(state);
        END;
    END;
    IF state = waiting THEN abort(errHelp,"");END;

    IF startoffset < 1 THEN abort(errStart,"-s:#"); END;
    IF countskip < 1 THEN abort(errCount,"-c:#"); END;
    IF nLines < 1 THEN abort(errLines,"-a:#"); END;

    IF same(source,dot) THEN source:=stardotstar;END;

    filecount := buildList(basepath, source);
    CASE filecount OF
    | 0 : abort(errNotFound,source);
    | MAX(CARDINAL): abort(errTooMany,source);
    ELSE
        IF alphasort THEN Lib.QSort(filecount,isLessByName,doSwapNames);END;
    END;

    (*
    WrStr(ProgTitle+" "+ProgVersion+" "+ProgCopyright);WrLn;
    WrLn;
    *)
    FOR i := 1 TO filecount DO
        Str.Concat(source,basepath,fileArray[firstFile+i-1]);


        WrStr("Text file       : "); WrStr(source); WrLn;
        WrStr("Starting offset : ");IO.WrLngCard(startoffset,1);WrLn;
        WrStr("Skipped chars   : ");IO.WrLngCard(countskip,1);WrLn;
        WrStr("Alignment       : ");
        IF flagAlign THEN
            WrStr("yes");
        ELSE
            WrStr("no");
        END;
        WrLn;
        WrStr("Number of lines : ");
        IF flagAlign THEN
            IO.WrLngCard(nLines,1);
        ELSE
            WrStr("not applicable");
        END;
        WrLn;
        WrLn;

        doShow ( source,startoffset,countskip,flagAlign,flagUpper,nLines);
        IF ChkEscape() THEN abort(errAborted,"");END;
    END;

    abort(errNone,"");
END ShowHIDN.


