(* ---------------------------------------------------------------
Title         Q&D Text Tool
Overview      see help
Usage         see help
Notes         very, very, very quick & dirty... :-( but effective :-)
Bugs
Wish List     -t could preprocess files with hashes and binary search tables
              handle remarks (";", "#") in tokens and source ?
              add eyecandy while processing ?
              when creating exploded tokens, we could sort tokens again

              line length when exploding or tokenizing is limited to [1..96] :
              this is by design (created for dictionaries !)
              (could be more general though and/or allow hex ext but who cares ?)

              allow ESC to abort program ?

              "e pluribus spaghetticode" : this looks like a can OF worms now :-(

              //TODO : syntax 9 (-zz#) then remove errMaybeWare
              //keep previous packed len

              //TODO syntax 8 : <source> <tokens> <target>
              create tokenized, excluding cr,lf ($00=esc)
              file format :
              *
              ordered tokens
              *
              tokenized words

--------------------------------------------------------------- *)

MODULE textTool;

IMPORT IO;
IMPORT FIO;
IMPORT Lib;
IMPORT Str;
IMPORT Storage;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    MAYBEWARE = TRUE;
TYPE
    str4 = ARRAY[0..3] OF CHAR;
    str8 = ARRAY[0..7] OF CHAR;

CONST
	errNone				= 0;
	errHelp 			= 1;
	errNotFound 	    = 2;
	errTooManyWords 	= 3;
    errParmOverflow     = 4;
    errOption           = 5;
    errCmd              = 6;
    errExpecting        = 7;
    errJoker            = 8;
    errEmpty            = 9;
    errTooManyStrings   = 10;
    errBadExt           = 11;
    errSyntax           = 12;
    errEsuffix          = 13;
    errBadChar          = 14;
    errPbTokenFile      = 15;
    errPbTokenized      = 16;
    errBadTag           = 17;
    errMoreHelp         = 18;
    errValOrRange       = 19;
    errEvenMoreHelp     = 20;
    errTooManyInitials  = 21;
    errMaybeWare        = 22;
CONST
    DEFINCLUDETAG = "$*!";
    INCLUDETAGLEN = 3;
CONST
    BADESUFFIX    = "|<>+";
CONST
    strMAXINITIALS   = "16000";
    strMAXLINES      = "1024";
    extDIC           = ".DIC";
    extNDX           = ".NDX";
    extDIGITSplus    = ".##";
    extDIGITSstart   = "_";
    extHASHplus      = "."+extDIGITSstart+"##";
    extBAK           = ".BK!";
    extTOK           = ".00"; (* not expected .TOK, eh eh *)
    extTOKplus       = extTOK+"?";
    EOLSYMBOL        = "*";
CONST
    msgProcessing    = '::: Processing "~", please wait...';
    msgDone          = " OK";
    msgDoneNothing   = " OK (no marker found)";
    INCLUDEDREMINDER = '(* ::: //INCLUDED from "~" file *)';
CONST
    exe         = "TEXTTOOL";
    version     = "1.1j";
    culprit     = "by PhG";
    title       = "Q&D Text Tool v"+version+" "+culprit;

PROCEDURE abort (errcode : CARDINAL; einfo : ARRAY OF CHAR );
CONST
    help = title+nl+
nl+
"Syntax 1 : "+exe+" <source> <-l[l]> [-r]"+nl+
"Syntax 2 : "+exe+" <source> <-b> [-u]"+nl+
"Syntax 3 : "+exe+" <source> <-x>"+nl+
"Syntax 4 : "+exe+" <source> <-t> <tokens> [extension_suffix] [-e[e]] [-c] [-a]"+nl+
"Syntax 5 : "+exe+" <source> <-i> <include_file> [-g:$] [-!]"+nl+
"Syntax 6 : "+exe+" <source> <-d:#[..#]|-k:#[..#]> [-m]"+nl+
"Syntax 7 : "+exe+" <source> <-y[#]> [-a]"+nl+
(*%F MAYBEWARE *)
"Syntax 8 : "+exe+" <source> <-> <tokens> <target>"+nl+
"Syntax 9 : "+exe+" <source> <-z[z][#]> <tokens> [-c] [-a]"+nl+
(*%E *)
nl+
"This program sorts <source> lines by length (syntax 1) ;"+nl+
"converts a raw list of reserved words to Boxer 7.0a format"+nl+
"so result can be merged into Boxer DEFAULT.CFG file (syntax 2) ;"+nl+
"explodes a file by line length (syntax 3) ;"+nl+
"explodes a file by tokenized length (syntax 4) ;"+nl+
"includes a file into <source> at specified marker position (syntax 5) ;"+nl+
"removes or keeps all lines of specified length or range (syntax 6) ;"+nl+
"indexes sorted <source> using # first letters (syntax 7)."+nl+
(*%F MAYBEWARE *)
nl+
"creates tokenized dictionary and its index using # first letters (syntax 9)."+nl+
(*%E  *)
nl+
"  -l[l]     sort by line length (-ll = -l -r)"+nl+
"  -b[b]     sort according to Boxer 7.0a for DOS format (-bb = -b -u)"+nl+
"  -x        explode by line length"+nl+
"  -t[t[t]]  explode by tokenized length (-tt = -t -e, -ttt = -t -ee)"+nl+
"  -i        include <include_file> into <source> at specified marker position"+nl+
"  -d:#[..#] delete all lines of specified length or range"+nl+
"  -k:#[..#] keep only lines of specified length or range"+nl+
"  -y[#]     index dictionary ([1..4] first letters, default is 2)"+nl+
(*%F MAYBEWARE *)
"  -z[z][#]  tokenize and index dictionary ([1..4] first letters, default is 2)"+nl+
(*%E *)
nl+
"  -r        reverse order (-l only)"+nl+
"  -u        convert each keyword to uppercase (-b only)"+nl+
"  -e[e]     prepend binary header (-t only)"+nl+
"  -c        create hash table (-t only)"+nl+
"  -a        alternate fixed-strings format (-t and -y[#] only)"+nl+
'  -g:$      redefine marker string (3 symbols required, default is "'+DEFINCLUDETAG+'")'+nl+ (* -i only *)
'  -!        enclose <include_file> with "'+INCLUDEDREMINDER+'"'+nl+ (* (-i only) *)
"  -m        trim line before processing (-d:#[..#] and -k:#[..#] only)"+nl+
"  -lfn      disable LFN support even if available"+nl+
"  -??[?]    (even) more help (to be read at least once !)"+nl+
nl+
"a) Whatever the syntax, maximum string line or keyword length is 256 chars."+nl+
"   The following limit apply to syntaxes 1 and 2 :"+nl+
"   <source> cannot contain more than "+strMAXLINES+" lines or keywords."+nl+
"b) For safety, original <source> is kept with "+extBAK+" extension."+nl;
CONST
    morehelp =
"c) With syntax 2, <source> must contain either one keyword per line,"+nl+
'   or several keywords separated by "'+coma+'" delimiters.'+nl+
"   Result is in format expected by Boxer 7.0a for DOS's .CFG file."+nl+
"d) With syntaxes 3 and 4, target files have "+extDIGITSplus+" extension"+nl+
"   where ## is line or tokenized length, both in [1..96] range."+nl+
"e) With syntax 4, <source> is a list of words (only one word per line)."+nl+
"   <tokens> file contains up to 255 strings sorted by decreasing length"+nl+
"   (token length should belong to [1..5] range)."+nl+
"   Tokens and words should be in lower case."+nl+
"   Optional <extension_suffix> is a single char appended to "+extDIGITSplus+" extension."+nl+
"   Enclosing double quotes will be removed from <source> and <tokens> files."+nl+
"   Optional -e header is SHORTCARD(tokenized_length) repeated ## times"+nl+
"   while -ee form writes only a CARDINAL(tokenized_length) header."+nl+
"   Optional -c hash tables will have "+extHASHplus+" extension ;"+nl+
"   they are built applying TopSpeed Lib.HashString() to each original entry,"+nl+
"   range being (MAX(CARDINAL)-1)."+nl+
"f) With syntax 5, program expects marker to be alone on its own line."+nl+
"   File to be included is merged verbatim : trailing Ctrl-Z is no filtered."+nl+
"g) With syntax 6, should range be illogical, values will be swapped."+nl+
'   Note range maximum value can be "'+EOLSYMBOL+'" for "end of line".'+nl+
"   Lines are dumped verbatim, even if trimming was on for length test."+nl+
"h) With syntax 7, indexing uses # first letters ([1..4], default is 2)."+nl+
"   Entries should be in lower case. Enclosing double quotes will be removed."+nl+
"   Up to "+strMAXINITIALS+" sequences are allowed. Index will have "+extNDX+" extension."+nl+
(*%F MAYBEWARE *)
"i) With syntax 8,"+nl+
"j) With syntax 9, program creates "+extDIC+" tokenized dictionary and its index."+nl+
"   Note -zz[#] truncates first # characters of each entry prior to tokenizing."+nl+
(*%E  *)
nl+
"Examples : "+exe+" -t lex.fra c.fra C"+nl+
"           "+exe+" -i -! source.mod disclaimer.txt -g:$*!"+nl+
"           "+exe+" -d:78..* dico.txt"+nl+
"           "+exe+" -y2 lex.fra"+nl;
CONST
    evenmorehelp = nl+
"(* "+extTOKplus+" file format *)"+nl+
nl+
"List of <tokens> is built from a dictionary using WCOUNT, QSORT and a brain."+nl+
"Once best tokenization list is found, the list must be sorted by length"+nl+
"(from longest to shortest, for obvious reasons)."+nl+
nl+
"A token code is a byte belonging to [0..254] ([$00..$fe]) range."+nl+
"Token code 0 ($00) is reserved : it means next byte will be used verbatim."+nl+
nl+
"Syntax 4 creates a binary file with "+extTOKplus+" extension containing table of tokens."+nl+
"First entry is a SHORTCARD containing last token index (from 1 to 255),"+nl+
"followed by an array of tokenEntryType according to format :"+nl+
nl+
"    tokenEntryType = RECORD"+nl+
"        tokenLen : SHORTCARD;"+nl+
"        tokenStr : ARRAY OF CHAR;            (* variable length *)"+nl+
"    END;"+nl+
nl+
"    tokenEntryType = RECORD"+nl+
"        tokenStr : ARRAY [0..7] OF CHAR;"+nl+
"    END;"+nl+
nl+
"(* "+extNDX+" file format *)"+nl+
nl+
"Syntax 7 creates a binary index with "+extNDX+" extension."+nl+
"First entry is a CARDINAL containing last initial sequence,"+nl+
"followed by an array of initialEntryType according to format :"+nl+
nl+
"    initialEntryType = RECORD"+nl+
"        initialFpos  : LONGCARD;             (* position in text dictionary *)"+nl+
"        initialCount : LONGCARD;             (* number of entries *)"+nl+
"        initialLen   : SHORTCARD;"+nl+
"        initialStr   : ARRAY OF CHAR;        (* variable length *)"+nl+
"    END;"+nl+
nl+
"    initialEntryType = RECORD"+nl+
"        initialFpos  : LONGCARD;             (* position in text dictionary *)"+nl+
"        initialCount : LONGCARD;             (* number of entries *)"+nl+
"        initialStr   : ARRAY [0..3] OF CHAR;"+nl+
"    END;"+nl+
nl+
"(* "+extDIC+" file format *)"+nl+
nl+
"Syntax 8 creates a binary file."+nl;

VAR
    S : str1024;
BEGIN
	CASE errcode OF
	| errHelp,errMoreHelp,errEvenMoreHelp:
	    WrStr(help);
	    IF errcode # errHelp THEN
	        WrStr(morehelp);
	        IF errcode # errMoreHelp THEN
	            WrStr(evenmorehelp);
	        END;
	        errcode:=errHelp;
	    END;
	| errNotFound :        S:='"~" does not exist !';
	| errTooManyWords :    S:='Too many words in "~" !';
    | errParmOverflow:     S:='"~" is one parameter too many !';
	| errOption:           S:='Unknown "~" option !';
	| errCmd:              S:="-l, -b, -x, -t, -i, -d:#, -k:#, -y[#] and -z[z][#] options are mutually exclusive !";
                           S:="One of -l, -b, -x, -t, -i, -d:#, -k:#, -y[#] or -z[z][#] options !";
	| errExpecting:        S:="Expecting ~ !";
	| errJoker:            S:='At least one illegal joker in "~" !';
	| errEmpty:            S:='"~" is empty !';
	| errTooManyStrings:   S:='Too many lines in "~" !';
	| errBadExt:           S:='"~" extension cannot be made of digits !';
	| errSyntax:           S:='Unexpected "~" option for specified syntax !';
	| errEsuffix:          S:='"~" extension suffix must be a single character !';
	| errBadChar:          S:='"~" extension suffix cannot be one of "'+BADESUFFIX+'" !';
    | errPbTokenFile:      S:='"~" contains too few, too many or illegal tokens !';
    | errPbTokenized:      S:="Tokenization failure at line ~ !";
    | errBadTag:           S:='Illegal "~" marker !';
    | errValOrRange:       S:='Illegal "~" value or range !';
    | errTooManyInitials:  S:='"~" contains too many different initial entries !';
    | errMaybeWare:        S:="-z[z][#] option is MaybeWare and possibly NeverWare ! :-(";
	ELSE
		                   S:="How did you get there ???";
    END;
    CASE errcode OF
    | errNone,errHelp  :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(exe);WrStr(" : ");WrStr(S);WrLn;
    END;
	Lib.SetReturnCode(SHORTCARD(errcode));
	HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize= (8 * 512) + FIO.BufferOverhead;
TYPE
    ioBufferType = ARRAY [1..ioBufferSize] OF BYTE;
VAR
    ioBuffer,ioBufferOut : ARRAY [1..ioBufferSize] OF BYTE;

CONST
    databuffersize = 49152;
    mindata        = 0;
    maxdata        = databuffersize-1;
VAR
    databuffer : ARRAY [mindata..maxdata] OF BYTE;

(* ------------------------------------------------------------ *)

PROCEDURE dbg (DEBUG:BOOLEAN;S0,S1,S2:ARRAY OF CHAR  );
CONST
    wi=20;
VAR
    i:CARDINAL;
BEGIN
    IF NOT(DEBUG) THEN RETURN ; END;
    WrStr("/// ");
    WrStr(S0);WrStr(S1);
    FOR i:=( Str.Length(S0)+Str.Length(S1)+1 ) TO wi DO WrStr(" ");END;
    WrStr(': "');
    WrStr(S2);
    WrStr('"');WrLn;
END dbg;

PROCEDURE procBackup (VAR SRC,DST : pathtype;
                     DEBUG,useLFN:BOOLEAN;newExt : ARRAY OF CHAR);
VAR
	u,d,n,e:pathtype;
BEGIN
dbg(DEBUG,"SRC","",SRC);
dbg(DEBUG,"DST","",DST);
    Str.Copy(DST,SRC); (* DST is SRC as future target to be created *)
	qdSplitAllPath(SRC,u,d,n,e);
	Lib.MakeAllPath(SRC,u,d,n,newExt); (* have backup for source *)
    IF fileExists(useLFN,SRC) THEN (* erase existing previous backup *)
        fileSetRW(useLFN,SRC);
        fileErase(useLFN,SRC);
    END;
	fileRename(useLFN,DST,SRC);
dbg(DEBUG,"new SRC","",SRC);
dbg(DEBUG,"new DST","",DST);
END procBackup;

PROCEDURE GetCard (VAR v : CARDINAL; S : ARRAY OF CHAR) : BOOLEAN;
VAR
    lc:LONGCARD;
    rc:BOOLEAN;
BEGIN
    rc:=GetLongCard(S,lc);
    IF rc THEN
        IF lc > MAX(CARDINAL) THEN
            rc:=FALSE;
        ELSE
            v:=CARDINAL(lc);
        END;
    END;
    RETURN rc;
END GetCard;

(*
supported :
    #
    #..#
    #-#
    #,#
*)

PROCEDURE getValOrRange (VAR minival,maxival:CARDINAL;
                        S:ARRAY OF CHAR ):BOOLEAN;
CONST
    trycount      = 1;   delimcount    = ",";
    tryrange      = 2;   delimrange    = "..";
    tryrangealt   = 3;   delimrangealt = "-";
    firsttry      = trycount;
    lasttry       = tryrangealt;
VAR
    p,j:CARDINAL;
    rc:BOOLEAN;
    pattern,snum:str128; (* could be str16 *)
    lc:LONGCARD;
    sep:str2;
BEGIN
    rc:=FALSE;
    j:=firsttry-1;
    LOOP
        INC(j);
        IF j > lasttry THEN (* no separator found *)
            Str.Concat(snum,"=",S); (* ugly trick *)
            rc:= GetCard(minival,snum);
            IF rc THEN maxival:=minival; END;
            EXIT;
        END;
        CASE j OF
        | trycount    : sep:=delimcount;
        | tryrange    : sep:=delimrange;
        | tryrangealt : sep:=delimrangealt;
        END;
        p:=Str.Pos(S,sep);
        IF p # MAX(CARDINAL) THEN
            Str.Slice(snum,S,0,p);
            Str.Prepend(snum,"="); (* ugly trick *)
            rc:=GetCard(minival,snum);
            IF rc THEN
                Str.Delete(S,0,p+Str.Length(sep));
                IF same(S,EOLSYMBOL) THEN
                    maxival:=MAX(CARDINAL);
                ELSE
                    Str.Concat(snum,"=",S);
                    rc:=GetCard(maxival,snum);
                    IF rc THEN
                        CASE j OF
                        | trycount :
                            IF maxival = 0 THEN
                                rc:=FALSE;
                            ELSE
                                DEC(maxival);
                            END;
                            INC(maxival,minival);
                        | tryrange,tryrangealt :
                            IF minival > maxival THEN
                                p      :=minival;
                                minival:=maxival;
                                maxival:=p;
                            END;
                        END;
                    END;
                END;
            END;
            EXIT;
        END;
    END;
    RETURN rc;
END getValOrRange;

PROCEDURE bignum2str (n:LONGCARD;base,wi:CARDINAL;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
BEGIN
    Str.CardToStr( n,R,base,ok );
    WHILE Str.Length(R) < wi DO
        Str.Prepend(R,padchar);
    END;
    IF base=16 THEN Str.Lows(R);Str.Prepend(R,dollar);END;
    RETURN R;
END bignum2str;

PROCEDURE num2str (n,base,wi:CARDINAL;padchar:CHAR):str16;
BEGIN
    RETURN bignum2str(LONGCARD(n),base,wi,padchar);
END num2str;

PROCEDURE dbgv (DEBUG:BOOLEAN;S0,S1:ARRAY OF CHAR;v:LONGCARD);
BEGIN
    dbg(DEBUG,S0,S1,bignum2str(v,10,9," "));
END dbgv;

PROCEDURE makebasename (VAR base:pathtype;
                       addDot:BOOLEAN; source:pathtype );
VAR
    p:CARDINAL;
BEGIN
    p:=Str.RCharPos(source,dot);
    IF p = MAX(CARDINAL) THEN
        Str.Copy(base,source);
    ELSE
        Str.Slice(base, source,0,p);
    END;
    IF addDot THEN Str.Append(base,dot); END;
END makebasename;

PROCEDURE goodExt (source:pathtype ):BOOLEAN;
VAR
    ext : str16;
    p:CARDINAL;
BEGIN
    p:= Str.RCharPos(source,dot);
    IF p = MAX(CARDINAL) THEN RETURN TRUE;END;
    Str.Delete(source,0,p+1);
    RETURN NOT( verifyString(source,digits) );
END goodExt;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
	stringPtr	= POINTER TO str256;

CONST
	firstEntry 	= 1;
	lastEntry  	= 1024; (* should be more than enough! *)

VAR
	entry 		: ARRAY [firstEntry..lastEntry] OF stringPtr;
	sortedEntry : ARRAY [firstEntry..lastEntry] OF stringPtr;

(* ------------------------------------------------------------ *)

(* s1 < s2, now check if s1 is included in s2 in sortedEntry array *)

PROCEDURE Matching (i : CARDINAL; j : CARDINAL) : BOOLEAN;
VAR
    rc : BOOLEAN;
BEGIN
    RETURN ( Str.Pos( sortedEntry[j]^ , sortedEntry[i]^ )=0 );
END Matching;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* compare two strings from entry array *)

PROCEDURE LessBoxer (i,j : CARDINAL) : BOOLEAN;
VAR
	isLess : BOOLEAN;
BEGIN
	RETURN ( Str.Compare(entry[i]^, entry[j]^)  < 0 ) ;
END LessBoxer;

PROCEDURE SwapBoxer (i,j : CARDINAL);
VAR
	tmp		: stringPtr;
BEGIN
	tmp 	:= entry[i];
	entry[i]:= entry[j];
	entry[j]:= tmp;
END SwapBoxer;

PROCEDURE SwapSortedBoxer (i, j : CARDINAL);
VAR
	tmp		: stringPtr;
BEGIN
	tmp 	:= sortedEntry[i];
	sortedEntry[i]:= sortedEntry[j];
	sortedEntry[j]:= tmp;
END SwapSortedBoxer;

PROCEDURE LessByASCII ( i,j:CARDINAL  ):BOOLEAN;
BEGIN
    RETURN ( Str.Compare(entry[i]^, entry[j]^)  < 0 );
END LessByASCII;

PROCEDURE LessByLen (i , j : CARDINAL) : BOOLEAN;
VAR
    ilen,jlen:CARDINAL;
BEGIN
    ilen := Str.Length( entry[i]^ );
    jlen := Str.Length( entry[j]^ );
    RETURN  (ilen < jlen);
END LessByLen;

PROCEDURE SwapByLen (i , j : CARDINAL);
VAR
	tmp		: stringPtr;
BEGIN
	tmp 	:= entry[i];
	entry[i]:= entry[j];
	entry[j]:= tmp;
END SwapByLen;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* check for empty string, and string with many keywords
separated with spaces or comas
note Boxer supports inner spaces for keywords : now, we only allow comas
*)

PROCEDURE ParseString (S : ARRAY OF CHAR; VAR index : CARDINAL):BOOLEAN;
CONST
	separators	= ","; (* was " ," *)
VAR
	i,len : CARDINAL;
	ok:BOOLEAN;
	keyword 	: str256;
BEGIN
	RtrimBlanks(S);
	LtrimBlanks(S);
	ok:=NOT (same(S,"") );
    ok:=( ok AND (Str.CharPos(S,equal)=MAX(CARDINAL)) );
    IF ok THEN
		i := 0;
		LOOP
			Str.ItemS ( keyword , S , separators , i);
			IF same(keyword,"") THEN EXIT; END;
			IF index >= lastEntry THEN (* = is enough but > is a safety *)
				RETURN FALSE;
			END;
			INC (i);
			INC (index);
    		len := Str.Length(keyword);
		    Storage.ALLOCATE(entry[index], len + 1); (* include final $00 *)
    		Lib.Move(ADR(keyword), ADR(entry[index]^), len + 1);
		END;
	END;
	RETURN TRUE;
END ParseString;

PROCEDURE procBoxer (VAR DST:pathtype;
                    DEBUG,useLFN,flagUPPER:BOOLEAN;SRC:pathtype):BOOLEAN;
CONST
	MaxWidth  = 72;
	sReserved = "RESERVED=";
VAR
	wordCount		: CARDINAL;
	i,len,sortedWordCount	: CARDINAL;
    hnd:FIO.File;
    ok,done:BOOLEAN;
    S:str256;
    Z:str1024;
BEGIN
    procBackup(SRC,DST, DEBUG,useLFN,extBAK);
    Z:=msgProcessing;
    Str.Subst(Z,"~",DST);
    WrStr(Z);

	hnd := fileOpenRead(useLFN,SRC);
	FIO.AssignBuffer(hnd,ioBuffer);
	wordCount := firstEntry-1;
	LOOP
		FIO.RdStr(hnd, S);
		IF flagUPPER THEN Str.Caps(S); END;
		IF ParseString(S,wordCount) = FALSE THEN
		    fileClose(useLFN,hnd);
            WrLn;
		    RETURN FALSE;
		END;
		IF FIO.EOF THEN EXIT; END;
	END;
	fileClose(useLFN,hnd);

	Lib.QSort(wordCount, LessBoxer, SwapBoxer);

	sortedWordCount := firstEntry-1;
  	FOR i := firstEntry TO wordCount DO
    	IF (i = firstEntry) OR (same (entry[i]^, entry[i-1]^) = FALSE) THEN
			INC (sortedWordCount);
      		sortedEntry[sortedWordCount] := entry[i];
    	END;
  	END;

	LOOP
		done := TRUE;
		FOR i := firstEntry TO wordCount-1 DO
 		   	IF Str.Compare(sortedEntry[i]^, sortedEntry[i+1]^)  < 0 THEN
				IF Matching(i,i+1) = TRUE THEN
					SwapSortedBoxer(i,i+1);
					done := FALSE;
				END;
           	END;
		END;
		IF done THEN EXIT; END;
	END;

	hnd := fileCreate(useLFN,DST);
  	FIO.AssignBuffer(hnd,ioBuffer);
	S := "";
	i := firstEntry;
	LOOP
        CASE ( same(S,"") ) OF
        | TRUE :
			S := sReserved;
			Str.Append(S,sortedEntry[i]^);
			INC (i);
		| FALSE :
			len := Str.Length(S);
			len := len+Str.Length(coma);
			len := len+Str.Length(sortedEntry[i]^);
			IF len < MaxWidth THEN
				Str.Append(S,coma);
				Str.Append(S,sortedEntry[i]^);
				INC (i);
			ELSE
			    ok:= NOT( same(S,sReserved) );
			    IF ok THEN
		   		    FIO.WrStr(hnd, S); FIO.WrLn(hnd);
   				END;
				S := "";
			END;
		END;
        IF i > sortedWordCount THEN
            EXIT;
        END;
	END;
	IF same(S,"")=FALSE THEN
	    ok:=NOT( same(S,sReserved) );
	    IF ok THEN
   		    FIO.WrStr(hnd, S); FIO.WrLn(hnd);
   		END;
	END;
  	fileClose(useLFN,hnd);
    WrStr(msgDone);WrLn;
    RETURN TRUE;
END procBoxer;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procByLen (VAR DST : pathtype;
                    DEBUG,useLFN,flagREVERSE:BOOLEAN;SRC:pathtype ):BOOLEAN;
VAR
    hnd:FIO.File;
    S:str256;
    Z:str1024;
    ndx,i,lineCount,len:CARDINAL;
BEGIN
    procBackup(SRC,DST, DEBUG,useLFN,extBAK);
    Z:=msgProcessing;
    Str.Subst(Z,"~",DST);
    WrStr(Z);

	hnd := fileOpenRead(useLFN,SRC);
	FIO.AssignBuffer(hnd,ioBuffer);
	lineCount := firstEntry-1;
	LOOP
		FIO.RdStr(hnd, S);
		LtrimBlanks(S);
		RtrimBlanks(S);
		IF same(S,"") = FALSE THEN
			IF lineCount >= lastEntry THEN (* = is enough but > is a safety *)
                fileClose(useLFN,hnd); WrLn;
                fileRename(useLFN,SRC,DST); (* //V11J *)
				RETURN FALSE;
			END;
			INC (lineCount);
    		len := Str.Length(S);
		    Storage.ALLOCATE(entry[lineCount], len + 1); (* include final $00 *)
    		Lib.Move(ADR(S), ADR(entry[lineCount]^), len + 1);
		END;
		IF FIO.EOF THEN EXIT; END;
	END;
	fileClose(useLFN,hnd);

	Lib.QSort(lineCount, LessByASCII, SwapByLen);
	Lib.QSort(lineCount, LessByLen, SwapByLen);

	hnd := fileCreate(useLFN,DST);
  	FIO.AssignBuffer(hnd,ioBuffer);
	FOR i:= firstEntry TO lineCount DO
	    IF flagREVERSE THEN
	        ndx:=lineCount-i+1;
        ELSE
	        ndx:=i;
        END;
        S:= entry[ndx]^;
        FIO.WrStr(hnd,S);FIO.WrLn(hnd);
	END;
  	fileClose(useLFN,hnd);
    WrStr(msgDone);WrLn;
    RETURN TRUE;
END procByLen;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* tag is already uppercased because we except symbols *)

PROCEDURE validTag (S:ARRAY OF CHAR):BOOLEAN;
VAR
    i,pb,len:CARDINAL;
BEGIN
    len:=Str.Length(S);
    IF len # INCLUDETAGLEN THEN RETURN FALSE; END;
    pb:=0;
    FOR i:=1 TO len DO
        CASE S[i-1] OF
        | "A".."Z" : INC(pb);
        | "0".."9" : INC(pb);
        | '"'      : INC(pb);
        END;
    END;
    RETURN (pb=0);
END validTag;

PROCEDURE getStrREMINDER (isSTART:BOOLEAN;F:pathtype ):str1024;
CONST
    strOPEN = " :::((( ";
    strCLOSE= " :::))) ";
VAR
    R:str1024;
    Z:str16;
BEGIN
    Str.Copy(R,nl+INCLUDEDREMINDER+nl+nl);
    IF isSTART THEN
        Z:=strOPEN;
    ELSE
        Z:=strCLOSE;
    END;
    Str.Subst(R," ::: ",Z); (* assume it exists ! *)
    Str.Subst(R,"~",F);
    RETURN R;
END getStrREMINDER;

CONST
    opCREATE = 0;
    opAPPEND = 1;

PROCEDURE procXtract (useLFN,flagREMINDER:BOOLEAN;op:CARDINAL;fpos,count:LONGCARD;
                     SRC,DST:pathtype);
VAR
    hin,hout:FIO.File;
    remaining : LONGCARD;
    wanted,got : CARDINAL;
BEGIN
    hin:=fileOpenRead(useLFN,SRC);
    FIO.AssignBuffer(hin,ioBuffer);
    CASE op OF
    | opCREATE: hout:=fileCreate(useLFN,DST);
    | opAPPEND: hout:=fileOpen(useLFN,DST);
                FIO.Seek(hout, FIO.Size(hout) );
    END;
    FIO.AssignBuffer(hout,ioBufferOut);

    FIO.Seek(hin,fpos);

    IF flagREMINDER THEN FIO.WrStr(hout,getStrREMINDER(TRUE,SRC) );END; (* kludge *)

    remaining:=count;
    LOOP
        IF remaining = 0 THEN EXIT; END;
        IF remaining > databuffersize THEN
            wanted := CARDINAL ( databuffersize );
            DEC(remaining,databuffersize);
        ELSE
            wanted := CARDINAL(remaining);
            remaining := 0;
        END;
        got := FIO.RdBin(hin,databuffer,wanted);
        FIO.WrBin(hout,databuffer,got);
    END;

    IF flagREMINDER THEN FIO.WrStr(hout,getStrREMINDER(FALSE,SRC) );END; (* kludge *)

    fileClose(useLFN,hout);
    fileClose(useLFN,hin);
END procXtract;

PROCEDURE procInclude (DEBUG,useLFN,flagREMINDER:BOOLEAN;
                      SRC,DST,INCLUDEFILE:pathtype;INCLUDETAG:ARRAY OF CHAR);
VAR
    S : str1024; (* should DO *)
    hin:FIO.File;
    fpos,fposrestart,fcount,slen:LONGCARD;
    found:BOOLEAN;
BEGIN
    S:=msgProcessing;
    Str.Subst(S,"~",SRC);
    WrStr(S);

    found:=FALSE;
    hin  :=fileOpenRead(useLFN,SRC);
    FIO.AssignBuffer(hin,ioBuffer);
    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        fpos:=FIO.GetPos(hin);
        FIO.RdStr(hin,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        IF Str.Length(S)=INCLUDETAGLEN THEN
            found:=same(S,INCLUDETAG);
            IF found THEN
                fposrestart:=FIO.GetPos(hin);
                EXIT;
            END;
        END;
    END;
    fileClose(useLFN,hin);
    IF found THEN
        procBackup(SRC,DST, DEBUG,useLFN,extBAK); (* original source renamed as backup *)

        (*
        IF fileExists(useLFN,DST) THEN (* is useless in fact for we just renamed it ! *)
            fileSetRW(useLFN,DST);
            fileErase(useLFN,DST);
        END;
        *)

        procXtract(useLFN,FALSE,opCREATE,0,fpos,SRC,DST);

        fcount:=fileGetFileSize(useLFN,INCLUDEFILE);
        procXtract(useLFN,flagREMINDER,opAPPEND,0,fcount,INCLUDEFILE,DST);

        fcount:=fileGetFileSize(useLFN,SRC);
        fcount:=fcount-fposrestart;
        procXtract(useLFN,FALSE,opAPPEND,fposrestart,fcount,SRC,DST);
        S:=msgDone;
    ELSE
        S:=msgDoneNothing;
    END;
    WrStr(S);WrLn;
END procInclude;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procLines (DEBUG,useLFN,dodel,dotrim:BOOLEAN;minilen,maxilen:CARDINAL;SRC:pathtype);
VAR
    DST:pathtype;
    hin,hout:FIO.File;
    S,Z:str1024;
    len:CARDINAL;
    dmp:BOOLEAN;
BEGIN
    procBackup(SRC,DST, DEBUG,useLFN,extBAK);
    S:=msgProcessing;
    Str.Subst(S,"~",DST);
    WrStr(S);

    hin  :=fileOpenRead(useLFN,SRC);
    FIO.AssignBuffer(hin,ioBuffer);
    hout :=fileCreate(useLFN,DST);
    FIO.AssignBuffer(hout,ioBufferOut);

    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,Z);
        Str.Copy(S,Z);
        IF dotrim THEN
            LtrimBlanks(S);
            RtrimBlanks(S);
        END;
        len:=Str.Length(S);
        IF dodel THEN
            dmp := ( (len < minilen) OR (len > maxilen) );
        ELSE
            dmp := ( (len >= minilen) AND (len <= maxilen) );
        END;
        IF dmp THEN
            FIO.WrStr(hout,Z);FIO.WrLn(hout);
dbg(DEBUG,"keeping","",Z);
        END;
    END;
    fileClose(useLFN,hout);
    fileClose(useLFN,hin);
    WrStr(msgDone);WrLn;
END procLines;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    NOPROBLEMOLINE = MAX(LONGCARD);
CONST
    minpackedlen   = 1;  (* allow -1 *)
    maxpackedlen   = 96; (* 64+32 should do *)
    MAXIpackedlen  = maxpackedlen+1;
VAR
    packedcount : ARRAY[minpackedlen..MAXIpackedlen] OF LONGCARD;

TYPE
    tokenizedtype = RECORD
        kod : ARRAY[0..maxpackedlen-1] OF SHORTCARD; (* oversized, was 32, but who knows what evil... *)
    END;
CONST
    mintoken     = 1;
    maxtoken     = 255;
    ESCVERBATIM  = mintoken-1;
    realmintoken = mintoken-1;
    UNDEFINEDTOKENLEN = MAX(CARDINAL); (* always bigger than any token length *)
    blank4       = "    ";
    blanktoken   = blank4+blank4; (* 16 blanks *)

TYPE
    tokentrytype = RECORD
        toklen : SHORTCARD;
        tokstr : str8;
    END;

CONST
    ndxmain = 0;
    ndxalt  = 1;
VAR
    zetoken     : ARRAY[realmintoken..maxtoken] OF str8; (* was str16 tokenstr 0..4 *)
    tokenlen    : ARRAY[realmintoken..maxtoken] OF SHORTCARD;
    firstaccess : ARRAY[minpackedlen..maxpackedlen],[ndxmain..ndxalt] OF BOOLEAN;

PROCEDURE initAccessed (  );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=minpackedlen TO maxpackedlen DO
        firstaccess[i][ndxmain]:=TRUE;
        firstaccess[i][ndxalt]:=TRUE;
    END;
END initAccessed;

PROCEDURE alreadyAccessed ( i,j:CARDINAL  ):BOOLEAN;
BEGIN
    RETURN NOT( firstaccess[i][j] );
END alreadyAccessed;

PROCEDURE markAsAccessed ( i,j:CARDINAL  );
BEGIN
    firstaccess[i][j]:=FALSE;
END markAsAccessed;

(* ------------------------------------------------------------ *)

CONST
    initcounts = 0;
    showcounts = 1;

PROCEDURE procPackedCounts (what:CARDINAL   );
CONST
    strStats = "::: Packed length = ~ ::: Entries = ~ ::: Bytes = ~";
    strTotal = "::: ~ total bytes (not counting optional headers)";
    msgTokenizedProblem = "--- Tokenization failure !";
    wismall = 2;
    wilong  = 9;
VAR
    i:CARDINAL;
    entrycount,bytecount,grandtotal:LONGCARD;
    S:str128;
BEGIN
    CASE what OF
    | initcounts:
        FOR i:=minpackedlen TO MAXIpackedlen DO packedcount[i]:=0;END;
    | showcounts:
        IF packedcount[MAXIpackedlen] # 0 THEN
            WrLn;
            WrStr(msgTokenizedProblem);WrLn;
            WrLn;
            RETURN;
        END;
        grandtotal := 0;
        WrLn;
        FOR i:=minpackedlen TO maxpackedlen DO
            entrycount:=packedcount[i];
            IF entrycount # 0 THEN
                bytecount := entrycount * LONGCARD(i);
                INC(grandtotal,bytecount);
                S:=strStats;
                Str.Subst(S,"~",num2str(i,10,wismall," "));
                Str.Subst(S,"~",bignum2str(entrycount,10,wilong," "));
                Str.Subst(S,"~",bignum2str(bytecount,10,wilong," "));
                WrStr(S);WrLn;
            END;
        END;
        WrLn;
        S:=strTotal;
        Str.Subst(S,"~",bignum2str(grandtotal,10,1," "));
        WrStr(S);WrLn;
    END;
END procPackedCounts;

(* ------------------------------------------------------------ *)

PROCEDURE loadTokens (DEBUG,useLFN:BOOLEAN;F:pathtype):CARDINAL;
VAR
    lasttoken,i : CARDINAL;
    hnd:FIO.File;
    S:str256;
BEGIN
    FOR i:=realmintoken TO maxtoken DO
        Str.Copy(zetoken[i],blanktoken);
        Str.Copy(zetoken[i],"");
        tokenlen[i]:=UNDEFINEDTOKENLEN;
    END;

	hnd := fileOpenRead(useLFN,F);
	FIO.AssignBuffer(hnd,ioBuffer);
    lasttoken:=mintoken;
	LOOP
		FIO.RdStr(hnd, S);
		LtrimBlanks(S);
		RtrimBlanks(S);
		IF Str.Match(S,dquote+star+dquote) THEN
            Str.Delete(S,0,1);
            Str.Delete(S,Str.Length(S)-1,1);
		END;
		IF same(S,"") = FALSE THEN
			IF lasttoken > maxtoken THEN
                fileClose(useLFN,hnd);
				RETURN MAX(CARDINAL);
			END;
			Str.Copy(zetoken[lasttoken],S);
			tokenlen[lasttoken]:= SHORTCARD( Str.Length(S) ) ;
dbg(DEBUG," lasttoken","",num2str(lasttoken,10,3," "));
dbg(DEBUG," token","",zetoken[lasttoken]);
dbg(DEBUG," tokenlen","",num2str( CARDINAL ( tokenlen[lasttoken] ),10,2," "));
			INC (lasttoken);
		END;
		IF FIO.EOF THEN EXIT; END;
	END;
	fileClose(useLFN,hnd);
	IF lasttoken = mintoken THEN RETURN MAX(CARDINAL);END;
    RETURN lasttoken-1; (* convert count to index *)
END loadTokens;

(* assume lower case *)

PROCEDURE tokenize (VAR packedlen:CARDINAL;VAR packed:tokenizedtype;
                   lasttoken:CARDINAL;S:ARRAY OF CHAR):BOOLEAN ;
VAR
    pb,i,wordlen,toklen:CARDINAL;
    R:str16;
BEGIN
    pb:=0;
    packedlen:=0;
    WHILE Str.Length(S) # 0 DO
        wordlen := Str.Length(S);
        i:=mintoken;
        LOOP
            toklen := CARDINAL( tokenlen[i] );
            IF toklen <= wordlen THEN
                Str.Slice(R, S,0,toklen);
                IF Str.Compare(zetoken[i],R)=0 THEN
                    packed.kod[packedlen]:=SHORTCARD(i);
                    IF packedlen > maxpackedlen THEN
                        INC(pb);
                    ELSE
                        INC(packedlen);
                    END;
                    Str.Delete(S,0,toklen);
                    EXIT;
                END;
            END;
            INC(i);
            IF i > maxtoken THEN (* assume it's a single CHAR : will become ESCVERBATIM THEN char *)
                packed.kod[packedlen]:=SHORTCARD(ESCVERBATIM);
                IF packedlen > maxpackedlen THEN
                    INC(pb);
                ELSE
                    INC(packedlen);
                END;
                packed.kod[packedlen]:=SHORTCARD( ORD(S[0]) );
                IF packedlen > maxpackedlen THEN
                    INC(pb);
                ELSE
                    INC(packedlen);
                END;
                Str.Delete(S,0,toklen);
                EXIT;
            END;
        END;
    END;
    RETURN (pb=0);
END tokenize;

PROCEDURE appendtoken ( useLFN,flagHEADER,flagHEADERALT,flagHASHTABLE:BOOLEAN;
                      lasttoken:CARDINAL;ESUFFIX:CHAR;
                      base:pathtype;S:ARRAY OF CHAR):BOOLEAN;
CONST
    HASHRANGE = MAX(CARDINAL)-1;
VAR
    ext:str16;
    file:pathtype;
    packed:tokenizedtype;
    hout:FIO.File;
    hashvalue,packedlen,i:CARDINAL;
    pb:BOOLEAN;
BEGIN
    pb:=TRUE;
    IF tokenize(packedlen,packed,lasttoken,S)=FALSE THEN
        INC( packedcount[MAXIpackedlen] );
        RETURN pb;
    END;
    pb:=FALSE;
    (*
        empty string is already trapped by caller but who knows ?
        0 can happen with an empty string but would be a nonsense
    *)
    IF packedlen=0 THEN RETURN pb;END;

    INC( packedcount[packedlen]);

    ext:= num2str (packedlen,10,2,"0");
    Str.Append(ext,ESUFFIX);
    Str.Concat(file,base,ext);
    IF alreadyAccessed(packedlen,ndxmain) THEN (* // was fileExists() *)
        hout:=fileAppend(useLFN,file);
    ELSE
        hout:=fileCreate(useLFN,file);
        IF flagHEADER THEN
            IF flagHEADERALT THEN
                FIO.WrBin(hout,packedlen,SIZE(packedlen));
            ELSE
                FOR i:=1 TO packedlen DO
                    FIO.WrBin(hout,SHORTCARD(packedlen),SIZE(SHORTCARD));
                END;
            END;
        END;
        markAsAccessed(packedlen,ndxmain);
    END;
    FIO.WrBin(hout,packed,packedlen);
    fileClose(useLFN,hout);

    IF flagHASHTABLE THEN
        Str.Concat(ext,extDIGITSstart,num2str (packedlen,10,2,"0")); (* no dot *)
        Str.Concat(file,base,ext);
        IF alreadyAccessed(packedlen,ndxalt) THEN (* // was fileExists() *)
            hout:=fileAppend(useLFN,file);
        ELSE
            hout:=fileCreate(useLFN,file);
            markAsAccessed(packedlen,ndxalt);
        END;
        hashvalue := Lib.HashString(S,HASHRANGE);
        FIO.WrBin(hout,hashvalue,SIZE(hashvalue));
        fileClose(useLFN,hout);
    END;

    RETURN pb;
END appendtoken;

PROCEDURE wrTokens (flagFIXED:BOOLEAN;lasttoken:CARDINAL;hout:FIO.File );
VAR
    total,i,ndx:CARDINAL;
BEGIN
    FIO.WrBin(hout,SHORTCARD(lasttoken),SIZE(SHORTCARD(lasttoken)));

    FOR i:=mintoken TO lasttoken DO
        IF flagFIXED THEN
            FIO.WrBin( hout,zetoken[i],SIZE(zetoken[i]));
        ELSE
            FIO.WrBin( hout,tokenlen[i],SIZE(tokenlen[i]));
            FOR ndx:=1 TO Str.Length(zetoken[i]) DO
                FIO.WrBin(hout,zetoken[i][ndx-1],SIZE(CHAR));
            END;
        END;
    END;
END wrTokens;

PROCEDURE procTokenize (useLFN,flagFIXED,flagHEADER,flagHEADERALT,flagHASHTABLE:BOOLEAN;
                       lasttoken:CARDINAL;ESUFFIX:CHAR;SRC:pathtype):LONGCARD;
VAR
    S:str1024; (* should do *)
    base,tokensfile:pathtype;
    hin,hout:FIO.File;
    len:CARDINAL;
    currline:LONGCARD;
    alcatraz:BOOLEAN;
BEGIN
    S:=msgProcessing;
    Str.Subst(S,"~",SRC);
    WrStr(S);

    initAccessed();

    makebasename(base,FALSE,SRC); (* remember extTOK already starts WITH a dot *)

    Str.Concat(tokensfile,base,extTOK);
    Str.Append(tokensfile,ESUFFIX);
    hout:=fileCreate(useLFN,tokensfile);
    wrTokens(flagFIXED,lasttoken,hout);
    fileClose(useLFN,hout);

    makebasename(base,TRUE,SRC);

    procPackedCounts(initcounts);

    currline:=0;
    alcatraz:=FALSE;
    hin  :=fileOpenRead(useLFN,SRC);
    FIO.AssignBuffer(hin,ioBuffer);
    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,S);
        INC(currline);
        LtrimBlanks(S);
        RtrimBlanks(S);
		IF Str.Match(S,dquote+star+dquote) THEN
            Str.Delete(S,0,1);
            Str.Delete(S,Str.Length(S)-1,1);
		END;
        len:=Str.Length(S);
        IF len > 0 THEN
            alcatraz:=appendtoken (useLFN,flagHEADER,flagHEADERALT,flagHASHTABLE,
                                  lasttoken,ESUFFIX,base,S);
            IF alcatraz THEN EXIT;END;
        END;
    END;
    fileClose(useLFN,hin);
    WrStr(msgDone);WrLn;

    procPackedCounts(showcounts);

    IF NOT(alcatraz) THEN currline:=NOPROBLEMOLINE;END;
    RETURN currline;
END procTokenize;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)


CONST
    mininitial    = 1;
    maxinitial    = 16000; (* oversized *)
    INITIALLENDEF = 2;
    UNDEFINEDFPOS = MAX(LONGCARD);
CONST
    TOOMANYINITIALS = MAX(CARDINAL);

(*
   yes, we'd rather use an ARRAY OF records,
   but this is NOT protected mode DOS alas !
*)

VAR
    initialstr       : ARRAY [mininitial..maxinitial] OF str4;
    initialcount     : ARRAY [mininitial..maxinitial] OF LONGCARD; (* mere CARDINAL could probably do too *)
    initialfpos      : ARRAY [mininitial..maxinitial] OF LONGCARD;
    initialpackedfpos: ARRAY [mininitial..maxinitial] OF LONGCARD;

PROCEDURE buildInitials (DEBUG,useLFN,flagPACK,flagTRUNCATE:BOOLEAN;
                        initiallen,lasttoken:CARDINAL;SRC:pathtype):CARDINAL;
VAR
    hin:FIO.File;
    last,len,i,pb:CARDINAL;
    S,Z:str256;
    prev,curr:str4;
    count,fpos,anchor,packedfpos,packedanchor:LONGCARD;
    tokenized:tokenizedtype;
    packedlen:CARDINAL;
    ok:BOOLEAN;
BEGIN
    FOR i:=mininitial TO maxinitial DO
        initialstr[i]        := blank4;
        initialcount[i]      := 0;
        initialfpos[i]       := UNDEFINEDFPOS;
        initialpackedfpos[i] := UNDEFINEDFPOS; (* always init *)
        initialstr[i]        := "";
    END;

    prev :="";
    count:=0;
    last := mininitial-1;

    packedfpos:=0;
    packedanchor:=0;

    pb:=0;
    hin  :=fileOpenRead(useLFN,SRC);
    FIO.AssignBuffer(hin,ioBuffer);
    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        fpos:=FIO.GetPos(hin);
        FIO.RdStr(hin,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
		IF Str.Match(S,dquote+star+dquote) THEN
            Str.Delete(S,0,1);
            Str.Delete(S,Str.Length(S)-1,1);
		END;
        len:=Str.Length(S);
        IF len > 0 THEN
            IF flagPACK THEN
                Str.Copy(Z,S);
                IF flagTRUNCATE THEN Str.Delete(Z,0,initiallen); END;
                ok:=tokenize (packedlen,tokenized,  lasttoken,Z);
                IF NOT(ok) THEN
                    INC(pb);
                    EXIT;
                END;
            END;
            Str.Slice(curr,S,0,initiallen);
            IF same(curr,prev) THEN
                INC(count);
            ELSE
                    IF last < maxinitial THEN
                        IF count # 0 THEN (* Q&D brutal way to ignore first "" *)
                            INC(last);
                            Str.Copy(initialstr[last],prev);
                            initialcount[last]:=count;
                            initialfpos[last]:=anchor;
                            IF flagPACK THEN initialpackedfpos[last]:=packedanchor;END;
dbgv(DEBUG,prev,"[last]",LONGCARD(last));
dbg (DEBUG,prev,"[anchor]",bignum2str(anchor,16,8,"0"));
dbgv(DEBUG,prev,"[count]",count);
IF flagPACK THEN
dbgv(DEBUG,Z,"[packedlen]",LONGCARD(packedlen));
dbg (DEBUG,prev,"[packedanchor]",bignum2str(packedanchor,16,8,"0"));
dbg (DEBUG,prev,"[packedfpos]",bignum2str(packedfpos,16,8,"0"));
END;
                        END;
                        Str.Copy(prev,curr);
                        count:=1;
                        anchor:=fpos;
                        IF flagPACK THEN
                            packedanchor:=packedfpos;
                            INC(packedfpos, LONGCARD( SIZE(SHORTCARD)+packedlen) );
                        END;
                    ELSE
                        INC(pb);
                        EXIT;
                    END;
            END;

        END;
    END;
    (* handle pending *)
                    IF last < maxinitial THEN
                        INC(last);
                        initialfpos[last]:=anchor;
                        Str.Copy(initialstr[last],prev);
                        initialcount[last]:=count;
dbgv(DEBUG,prev,"[last]",LONGCARD(last));
dbg (DEBUG,prev,"[anchor]",bignum2str(anchor,16,8,"0"));
dbgv(DEBUG,prev,"[count]",count);
                    ELSE
                        INC(pb);
                    END;

    fileClose(useLFN,hin);
    IF pb # 0 THEN last:=TOOMANYINITIALS;END;
    RETURN last;
END buildInitials;

PROCEDURE wrInitials (flagFIXED:BOOLEAN;lastinitial:CARDINAL; hout:FIO.File);
VAR
    ndx,i,len : CARDINAL;
BEGIN
    FIO.WrBin(hout,lastinitial,SIZE(lastinitial));

    FOR i:=mininitial TO lastinitial DO
        FIO.WrBin( hout,initialfpos[i],SIZE(LONGCARD));
        FIO.WrBin( hout,initialcount[i],SIZE(LONGCARD));
        len:=Str.Length(initialstr[i] );
        IF flagFIXED THEN
            FIO.WrBin( hout,initialstr[i],SIZE(initialstr[i]));
        ELSE
            FIO.WrBin( hout,SHORTCARD(len),SIZE(SHORTCARD));
            FOR ndx:=1 TO len DO
                FIO.WrBin(hout,initialstr[i][ndx-1],SIZE(CHAR));
            END;
        END;
    END;
END wrInitials;

PROCEDURE procIndex (DEBUG,useLFN,flagFIXED:BOOLEAN;
                    initiallen:CARDINAL;SRC:pathtype):BOOLEAN;
VAR
    ok:BOOLEAN;
    fpos,currline:LONGCARD;
    S:str1024; (* should do *)
    hin,hout:FIO.File;
    base,ndxfile:pathtype;
    tk:tokentrytype;
    i,lastinitial,lasttoken:CARDINAL;
BEGIN
    S:=msgProcessing;
    Str.Subst(S,"~",SRC);
    IF NOT(DEBUG) THEN WrStr(S); END;

    makebasename(base,FALSE,SRC); (* remember extDIC already starts with a dot *)

    Str.Concat(ndxfile,base,extNDX);

    (* note FALSE,FALSE,mintoken-1 i.e. 0 *)
    lasttoken:=mintoken-1; (* note FALSE,FALSE for flagPACK,flagTRUNCATE*)
    lastinitial := buildInitials (DEBUG,useLFN,FALSE,FALSE,
                                 initiallen,lasttoken,SRC);

    ok:=( lastinitial # TOOMANYINITIALS );
    IF ok THEN
        hout:=fileCreate(useLFN,ndxfile);
        FIO.AssignBuffer(hout,ioBufferOut);

        wrInitials(flagFIXED,lastinitial,hout);

        fileClose(useLFN,hout);
    END;

    IF NOT(DEBUG) THEN WrStr(msgDone);WrLn; END;
    RETURN ok;
END procIndex;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE appendit ( useLFN:BOOLEAN; packedlen:CARDINAL;
                   file:pathtype; ext,entry:ARRAY OF CHAR  );
VAR
    hout:FIO.File;
BEGIN
    IF alreadyAccessed(packedlen,ndxmain) THEN (* // was fileExists() *)
        hout:=fileAppend(useLFN,file);
    ELSE
        hout:=fileCreate(useLFN,file);
        FIO.WrLn(hout);
        FIO.WrStr(hout,"; ");
        FIO.WrStr(hout,ext);
        FIO.WrLn(hout);
        FIO.WrLn(hout);
        markAsAccessed(packedlen,ndxmain);
    END;
    FIO.WrStr(hout,entry);
    FIO.WrLn(hout);
    fileClose(useLFN,hout);
END appendit;

PROCEDURE procExplode(useLFN:BOOLEAN;source: pathtype);
VAR
    hin:FIO.File;
    S:str1024;
    Z:str80;
    ext:str16;
    file,base:pathtype;
    len:CARDINAL;
BEGIN
    S:=msgProcessing;
    Str.Subst(S,"~",source);
    WrStr(S);

    initAccessed();

    makebasename(base,TRUE,source);
    hin  :=fileOpenRead(useLFN,source);
    FIO.AssignBuffer(hin,ioBuffer);
    FIO.EOF:=FALSE;
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,S);
        LtrimBlanks(S);
        RtrimBlanks(S);
        len:=Str.Length(S);
        IF len > 0 THEN
            ext:= num2str (len,10,2,"0");
            Str.Concat(file,base,ext);      (* [1..96] *)
            appendit(useLFN,len,file,ext,S);
        END;
    END;
    fileClose(useLFN,hin);
    WrStr(msgDone);WrLn;
END procExplode;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE procBuild (DEBUG,useLFN,flagFIXED,flagHEADER,flagHEADERALT,
                    flagHASHTABLE,flagTRUNCATE:BOOLEAN;
                    initiallen,lasttoken:CARDINAL;SRC:pathtype):LONGCARD;
VAR
    S:str1024; (* should do *)
    currline:LONGCARD;
    alcatraz:BOOLEAN;
    base,ndxfile:pathtype;
    lastinitial:CARDINAL;
BEGIN
    S:=msgProcessing;
    Str.Subst(S,"~",SRC);
    IF NOT(DEBUG) THEN WrStr(S); END;

    makebasename(base,FALSE,SRC); (* remember extDIC already starts with a dot *)

    Str.Concat(ndxfile,base,extNDX);

    lastinitial := buildInitials (DEBUG,useLFN,TRUE,flagTRUNCATE,
                                 initiallen,lasttoken,SRC);


    currline := 0;
    alcatraz := FALSE;


    IF NOT(DEBUG) THEN WrStr(msgDone);WrLn; END;

    IF NOT(alcatraz) THEN currline:=NOPROBLEMOLINE;END;
    RETURN currline;
END procBuild;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    cmdtype = (cmdnone,
              cmdsortbylen,cmdboxer,cmdexplodebylen,cmdtokenize,cmdinclude,
              cmddelrange,cmdkeeprange,cmdndx,cmdbuild);

PROCEDURE newcmd (VAR cmd:cmdtype;wanted:cmdtype  ):BOOLEAN;
BEGIN
    IF ( (cmd = cmdnone) OR (cmd=wanted) ) THEN
        cmd:=wanted;
        RETURN TRUE;
    END;
    RETURN FALSE;
END newcmd;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 3;
VAR
    S,R,SRC,DST,TOKENFILE,INCLUDEFILE : pathtype;
    ESUFFIX : CHAR;
	DEBUG,useLFN,rc:BOOLEAN;
	flagUPPER,flagREVERSE,flagHEADER,flagHEADERALT,flagREMINDER:BOOLEAN;
	flagTRIM,flagFIXED,flagHASHTABLE,flagTRUNCATE: BOOLEAN;
	initiallen,minilen,maxilen,i,opt,parmcount,lastparm : CARDINAL;
    cmd             : cmdtype;
    parm : ARRAY[minparm..maxparm] OF pathtype;
    v:LONGCARD;
    INCLUDETAG : str16; (* oversized *)
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn;

    cmd          := cmdnone;
    flagREVERSE  := FALSE;
    flagUPPER    := FALSE;
    flagHEADER   := FALSE;
    flagHEADERALT:= FALSE;
    flagHASHTABLE:= FALSE;
    flagREMINDER := FALSE;
    flagTRIM     := FALSE;
    flagFIXED    := FALSE;
    flagTRUNCATE := FALSE;
    initiallen   := INITIALLENDEF;
    useLFN       := TRUE;
    Str.Copy(INCLUDETAG,DEFINCLUDETAG);
    DEBUG        := FALSE;
    lastparm     := minparm-1;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S);
        UpperCase(R);
        RtrimBlanks(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R,"?"+delim+"H"+delim+"HELP"+delim+
                                 "U"+delim+"UPPERCASE"+delim+
                                 "R"+delim+"REVERSE"+delim+
                                 "L"+delim+"BYLINELENGTH"+delim+
                                 "B"+delim+"BOXER"+delim+
                                 "LL"+delim+"LR"+delim+
                                 "BB"+delim+"BU"+delim+
                                 "X"+delim+"EXPLODE"+delim+
                                 "LFN"+delim+
                                 "DEBUG"+delim+
                                 "T"+delim+"TOKENIZE"+delim+
                                 "E"+delim+"HEADER"+delim+
                                 "TT"+delim+"TE"+delim+
                                 "I"+delim+"INCLUDE"+delim+
                                 "G:"+delim+"TAG:"+delim+
                                 "!"+delim+"REMINDER"+delim+
                                 "??"+delim+
                                 "Y1"+delim+
                                 "Y2"+delim+
                                 "Y3"+delim+
                                 "D:"+delim+"DEL:"+delim+"DELETE:"+delim+
                                 "K:"+delim+"KEEP:"+delim+
                                 "M"+delim+"TRIM"+delim+
                                 "Y"+delim+
                                 "???"+delim+
                                 "Y4"+delim+
                                 "A"+delim+"FIXED"+delim+
                                 "C"+delim+"HASH"+delim+

                                 "Z"+delim+
                                 "Z1"+delim+
                                 "Z2"+delim+
                                 "Z3"+delim+
                                 "Z4"+delim+
                                 "ZZ"+delim+
                                 "ZZ1"+delim+
                                 "ZZ2"+delim+
                                 "ZZ3"+delim+
                                 "ZZ4"+delim+
                                 "EE"+delim+
                                 "TTT"+delim+"TEE"

                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5 :   flagUPPER   := TRUE;
            | 6,7 :   flagREVERSE := TRUE;
            | 8,9 :   IF newcmd(cmd,cmdsortbylen)=FALSE THEN abort(errCmd,S);END;
            |10,11:   IF newcmd(cmd,cmdboxer)=FALSE THEN abort(errCmd,S);END;
            |12,13:   IF newcmd(cmd,cmdsortbylen)=FALSE THEN abort(errCmd,S);END;
                      flagREVERSE := TRUE;
            |14,15:   IF newcmd(cmd,cmdboxer)=FALSE THEN abort(errCmd,S);END;
                      flagUPPER := TRUE;
            |16,17:   IF newcmd(cmd,cmdexplodebylen)=FALSE THEN abort(errCmd,S);END;
            |18:      useLFN:=FALSE;
            |19:      DEBUG:=TRUE;
            |20,21:   IF newcmd(cmd,cmdtokenize)=FALSE THEN abort(errCmd,S);END;
            |22,23:   flagHEADER:=TRUE;
            |24,25:   IF newcmd(cmd,cmdtokenize)=FALSE THEN abort(errCmd,S);END;
                      flagHEADER:=TRUE;
            |26,27:   IF newcmd(cmd,cmdinclude)=FALSE THEN abort(errCmd,S);END;
            |28,29:   GetString(R,INCLUDETAG); (* use uppercased tag *)
                      IF validTag(INCLUDETAG)=FALSE THEN abort(errBadTag,INCLUDETAG);END;
            |30,31:   flagREMINDER := TRUE;
            |32:      abort(errMoreHelp,"");
            |33:      IF newcmd(cmd,cmdndx)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 1;
            |34:      IF newcmd(cmd,cmdndx)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 2;
            |35:      IF newcmd(cmd,cmdndx)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 3;
            |36,37,38:IF newcmd(cmd,cmddelrange)=FALSE THEN abort(errCmd,S);END;
                      GetString(R,SRC);
                      IF getValOrRange(minilen,maxilen,SRC)=FALSE THEN abort(errValOrRange,S);END;
            |39,40:   IF newcmd(cmd,cmdkeeprange)=FALSE THEN abort(errCmd,S);END;
                      GetString(R,SRC);
                      IF getValOrRange(minilen,maxilen,SRC)=FALSE THEN abort(errValOrRange,S);END;
            |41,42:   flagTRIM := TRUE;
            |43:      IF newcmd(cmd,cmdndx)=FALSE THEN abort(errCmd,S);END;
                      initiallen := INITIALLENDEF;
            |44:      abort(errEvenMoreHelp,"");
            |45:      IF newcmd(cmd,cmdndx)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 4;
            |46,47:   flagFIXED := TRUE;
            |48,49:   flagHASHTABLE := TRUE;

            |50:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := INITIALLENDEF;
            |51:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 1;
            |52:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 2;
            |53:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 3;
            |54:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 4;
            |55:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := INITIALLENDEF; flagTRUNCATE:=TRUE;
            |56:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 1;             flagTRUNCATE:=TRUE;
            |57:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 2;             flagTRUNCATE:=TRUE;
            |58:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 3;             flagTRUNCATE:=TRUE;
            |59:      IF newcmd(cmd,cmdbuild)=FALSE THEN abort(errCmd,S);END;
                      initiallen := 4;
            |60:      flagHEADER:=TRUE; flagHEADERALT:=TRUE;
            |61,62:   IF newcmd(cmd,cmdtokenize)=FALSE THEN abort(errCmd,S);END;
                      flagHEADER:=TRUE; flagHEADERALT:=TRUE;

            ELSE
                abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy( parm[lastparm], S);
        END;
    END;
    CASE lastparm OF
    | minparm-1 : abort(errExpecting,"<source> parameter");
    | minparm   : CASE cmd OF
                  | cmdtokenize : abort(errSyntax,"-t");
                  | cmdinclude :  abort(errSyntax,"-i");
                  | cmdbuild :    abort(errSyntax,"-z[z][#]");
                  END;
                  Str.Copy(SRC,parm[minparm]);
    | minparm+1 : CASE cmd OF
                  |cmdtokenize:
                      Str.Copy(SRC,parm[minparm]);
                      Str.Copy(TOKENFILE,parm[minparm+1]);
                      ESUFFIX := "";
                  |cmdinclude:
                      Str.Copy(SRC,parm[minparm]);
                      Str.Copy(INCLUDEFILE,parm[minparm+1]);
                  |cmdbuild:
                      Str.Copy(SRC,parm[minparm]);
                      Str.Copy(TOKENFILE,parm[minparm+1]);
                  ELSE
                      abort(errExpecting,"any of -t, -i or -z[z][#] options");
                  END;
    | minparm+2 : IF newcmd(cmd,cmdtokenize)=FALSE THEN abort(errCmd,"-t");END;
                  Str.Copy(SRC,parm[minparm]);
                  Str.Copy(TOKENFILE,parm[minparm+1]);
                  Str.Copy(S,parm[minparm+2]);
                  IF Str.Length(S) # 1 THEN abort(errEsuffix,S);END;
                  ESUFFIX := S[0];
                  IF Str.CharPos(BADESUFFIX,ESUFFIX) # MAX(CARDINAL) THEN
                      abort(errBadChar,ESUFFIX);
                  END;
    END;
    IF chkJoker(SRC) THEN abort(errJoker,SRC);END;

    useLFN := ( useLFN AND fileSupportLFN() );

    CASE cmd OF
    | cmdnone :
        abort(errExpecting,"any of -l, -b, -x, -t, -d:#, -k:# or -y[#] options");
    | cmdboxer:
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        IF procBoxer(DST, DEBUG,useLFN,flagUPPER,SRC)=FALSE THEN abort(errTooManyWords,DST);END;
    | cmdsortbylen:
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        IF procByLen(DST, DEBUG,useLFN,flagREVERSE,SRC)=FALSE THEN abort(errTooManyStrings,DST);END;
    | cmdexplodebylen:
        IF goodExt(SRC)=FALSE THEN abort(errBadExt,SRC);END;
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        procExplode(useLFN,SRC);
    | cmdtokenize:
        IF chkJoker(TOKENFILE) THEN abort(errJoker,TOKENFILE);END;

        IF goodExt(SRC)=FALSE THEN abort(errBadExt,SRC);END;
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        IF fileExists(useLFN,TOKENFILE)=FALSE THEN abort(errNotFound,TOKENFILE);END;

        i:=loadTokens(DEBUG,useLFN,TOKENFILE);
        IF i = MAX(CARDINAL) THEN abort(errPbTokenFile,TOKENFILE);END;

        v:=procTokenize (useLFN,flagFIXED,flagHEADER,flagHEADERALT,flagHASHTABLE,
                        i,ESUFFIX,SRC);
        IF v # NOPROBLEMOLINE THEN
            Str.Copy(S,bignum2str(v,10,1,""));
            abort(errPbTokenized,S);
        END;
    | cmdinclude:
        IF chkJoker(INCLUDEFILE) THEN abort(errJoker,INCLUDEFILE);END;

        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        IF fileExists(useLFN,INCLUDEFILE)=FALSE THEN abort(errNotFound,INCLUDEFILE);END;
        procInclude(DEBUG,useLFN,flagREMINDER,SRC,DST,INCLUDEFILE,INCLUDETAG);
    | cmddelrange,cmdkeeprange:
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        procLines(DEBUG,useLFN,(cmd=cmddelrange),flagTRIM,minilen,maxilen,SRC);
    | cmdndx:
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        rc:= procIndex(DEBUG,useLFN,flagFIXED,initiallen,SRC);
        IF rc=FALSE THEN abort(errTooManyInitials,SRC);END;
    | cmdbuild:
        abort(errMaybeWare,""); (* //FIXME //TODO //MAYBEWARE *)

        IF chkJoker(TOKENFILE) THEN abort(errJoker,TOKENFILE);END;

        IF goodExt(SRC)=FALSE THEN abort(errBadExt,SRC);END;
        IF fileExists(useLFN,SRC)=FALSE THEN abort(errNotFound,SRC);END;
        IF fileGetFileSize(useLFN,SRC)=0 THEN abort(errEmpty,SRC);END;
        IF fileExists(useLFN,TOKENFILE)=FALSE THEN abort(errNotFound,TOKENFILE);END;

        i:=loadTokens(DEBUG,useLFN,TOKENFILE);
        IF i = MAX(CARDINAL) THEN abort(errPbTokenFile,TOKENFILE);END;

        (*
            //FIXME //TODO :
            dic header :
            magic
            version
            offset token table
            offset initials
            offset dic
            offset hashes
            value TO fix offset IN dico

            while tokenizing dico truncating OR NOT,
            build token table
            build index with count and offsets,
            build hashes
            fix offsets if concatenating dic
        *)

        v:=procBuild (DEBUG,useLFN,flagFIXED,flagHEADER,flagHEADERALT,
                     flagHASHTABLE,flagTRUNCATE,
                     initiallen,i,SRC);
        IF v # NOPROBLEMOLINE THEN
            Str.Copy(S,bignum2str(v,10,1,""));
            abort(errPbTokenized,S);
        END;

    END;
    (* let DOS DEALLOCATE strings *)
    abort(errNone,"");
END textTool.




(*

@echo off
del rpt
dd *.?? /ky /n
del lex

rem ============================ n3.fra

set _n_=3
set _f_=n3
set _e_=fra

copy /b lex.%_e_% lex
copy /b lex.%_e_% str%_n_%.%_e_%

newline -a "     " str%_n_%.%_e_%
columns -k 1 %_n_% str%_n_%.%_e_%
trim -e            str%_n_%.%_e_%

columns -d 1 %_n_% lex
c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
newline -u str%_n_%.%_e_%
c:\bat\tools\lcount -e str%_n_%.%_e_% >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ n3.eng

set _n_=3
set _f_=n3
set _e_=eng

copy /b lex.%_e_% lex
copy /b lex.%_e_% str%_n_%.%_e_%

newline -a "     " str%_n_%.%_e_%
columns -k 1 %_n_% str%_n_%.%_e_%
trim -e            str%_n_%.%_e_%

columns -d 1 %_n_% lex
c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
newline -u str%_n_%.%_e_%
c:\bat\tools\lcount -e str%_n_%.%_e_% >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ c2.fra

set _n_=2
set _f_=c2
set _e_=fra

copy /b lex.%_e_% lex
copy /b lex.%_e_% str%_n_%.%_e_%

newline -a "     " str%_n_%.%_e_%
columns -k 1 %_n_% str%_n_%.%_e_%
trim -e            str%_n_%.%_e_%

columns -d 1 %_n_% lex
c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
newline -u str%_n_%.%_e_%
c:\bat\tools\lcount -e str%_n_%.%_e_% >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ n2.eng

set _n_=2
set _f_=n2
set _e_=eng

copy /b lex.%_e_% lex
copy /b lex.%_e_% str%_n_%.%_e_%

newline -a "     " str%_n_%.%_e_%
columns -k 1 %_n_% str%_n_%.%_e_%
trim -e            str%_n_%.%_e_%

columns -d 1 %_n_% lex
c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
newline -u str%_n_%.%_e_%
c:\bat\tools\lcount -e str%_n_%.%_e_% >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ n1.fra

set _n_=1
set _f_=n1
set _e_=fra

copy /b lex.%_e_% lex
copy /b lex.%_e_% str%_n_%.%_e_%

newline -a "     " str%_n_%.%_e_%
columns -k 1 %_n_% str%_n_%.%_e_%
trim -e            str%_n_%.%_e_%

columns -d 1 %_n_% lex
c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
newline -u str%_n_%.%_e_%
c:\bat\tools\lcount -e str%_n_%.%_e_% >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ n1.eng

set _n_=1
set _f_=n1
set _e_=eng

copy /b lex.%_e_% lex
copy /b lex.%_e_% str%_n_%.%_e_%

newline -a "     " str%_n_%.%_e_%
columns -k 1 %_n_% str%_n_%.%_e_%
trim -e            str%_n_%.%_e_%

columns -d 1 %_n_% lex
c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
newline -u str%_n_%.%_e_%
c:\bat\tools\lcount -e str%_n_%.%_e_% >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ c0.fra

set _f_=c0
set _e_=fra

copy /b lex.%_e_% lex

c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
call clean
dd *.?? /ky /n
del lex

rem ============================ N0.eng

set _f_=n0
set _e_=eng

copy /b lex.%_e_% lex

c:\bat\tools\texttool -t lex %_f_%.%_e_%
dd *.00 /ky /n
dd *.?? /c >> rpt
call clean
dd *.?? /ky /n
del lex


set _n_=
set _f_=
set _e_=

*)

