(* ---------------------------------------------------------------
Title         Q&D Test ISAAC
Overview      see help
Usage         see help
Notes         a port of READABLE.C by Bob Jenkins
              with a bugfix and a few enhancements for testing this PRNG

              CCOUNT can help checking binary data

              never had this (not very informative) error before :
              "make error : more than one source for filename"
              in fact, it happens if compiler directory contains
              a file matching exe with .a extension !
              (probably temporary files)
Bugs
Wish List     find a clever method to scale down to range [0..255]
              without losing randomness
              bresenham mapping ?

--------------------------------------------------------------- *)

MODULE tstIsaac;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT IO;

FROM IO IMPORT WrStr, WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    methodABCD           = 1;
    methodDCBA           = 2;
    methodMOD255         = 3;
    methodMOD256         = 4;
    methodDIV1000000H    = 5;
    methodSHIFT24        = 6;
    methodDIVMAXLC256    = 7;
    methodDIVRANGE256    = 8;
    methodDIVMAXLC255    = 9;
    method32TO8          = 10;
CONST
    minMethod         = methodABCD;
    maxMethod         = method32TO8;
CONST
    roundsize         = 256*SIZE(LONGCARD); (* ISAACSTATESIZE*4 *)
CONST
    FMAXLC            = LONGREAL(MAX(LONGCARD));
    FMAXLCPLUS1       = LONGREAL(MAX(LONGCARD))+1.0;
    FMAXB             = LONGREAL(MAX(BYTE));
    FMAXBPLUS1        = LONGREAL(MAX(BYTE))+1.0;
    DIV32TO8          = MAX(LONGCARD) DIV MAX(SHORTCARD);
CONST
    progEXEname      = "TSTISAAC";
    progTitle        = "Q&D Test Bob Jenkins' I.S.A.A.C.";
    progVersion      = "v0.9";
    progCopyright    = "by PhG";
    progBanner       = progTitle+" "+progVersion+" "+progCopyright;
CONST
    errNone            = 0;
    errHelp            = 1;
    errOption          = 2;
    errTooManyParms    = 3;
    errBadCount        = 4;
    errJoker           = 5;
    errBase            = 6;
    errAdjusted        = 7;
    errExpecting       = 8;
    errSpecifiedPP     = 9;
    errPadNotApplicable= 10;
    errAborted         = 32;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
msgHelp =
progBanner+nl+
nl+
"Syntax 1 : "+progEXEname+nl+
"Syntax 2 : "+progEXEname+" <count> <basename> [passphrase] [option]..."+nl+
nl+
"This program is a slightly enhanced port of READABLE.C source code"+nl+
'created by Bob "I.S.A.A.C." Jenkins who generously made his PRNG public domain.'+nl+
"Syntax 1 should output the same data as RANDVECT.TXT and Rosetta Stone example."+nl+
'Syntax 2 creates target files filled with <count> random values'+nl+
"(either LONGCARDs or BYTEs) so they can be compared using randomness tests"+nl+
"(each file is created with a different export method)."+nl+
nl+
"  -s do not seed generator from pool"+nl+
"  -p pad seed pool with repeated passphrase (default is to pad with $00 value)"+nl+
"  -x init seed pool with an array of LONGCARDs instead of a string"+nl+
"  -t create text files (default is to create binary files)"+nl+
"  -d show dice results"+nl+
"  -m show missing byte values"+nl+
nl+
"a) Syntax 1 ignores all options."+nl+
"b) With syntax 2, program can be aborted by pressing almost any key."+nl+
"   Note <count> will be adjusted to the nearest multiple of 1024."+nl+
"   Existing target files will be overwritten. <basename> cannot be a LFN."+nl+
"   Passphrase, if not specified, is a random string created by program ;"+nl+
"   -x option creates an array of LONGCARDs instead of a string."+nl+
"c) Here are syntax 2 experimental methods (here, abcd is 32 bit ISAAC random) :"+nl+
"     A = abcd"+nl+
"     B = dcba"+nl+
"     C = abcd MOD 255"+nl+
"     D = abcd MOD 256"+nl+
"     E = abcd DIV $1000000"+nl+
"     F = abcd right-shifted by 24"+nl+
"     G = LONGREAL(abcd) / LONGREAL(MAX(LONGCARD))*256"+nl+
"     H = LONGREAL(abcd) / LONGREAL(MAX(LONGCARD)+1)*256"+nl+
"     I = LONGREAL(abcd) / LONGREAL(MAX(LONGCARD))*255"+nl+
"     J = abcd DIV (MAX(LONGCARD) DIV MAX(SHORTCARD))"+nl+
"   Most methods are bad (C, I, J) and/or give same results (E=F=G=H, I=J)."+nl+
"   Check "+progEXEname+".MOD source code for more details."+nl+
"d) Note CCOUNT utility can be of some help analyzing binary data."+nl;

VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :          WrStr(msgHelp);
    | errOption:         S := 'Unknown "~" option !';
    | errTooManyParms:   S := '"~" is one parameter too many !';
    | errBadCount:       S := 'Illegal "~" <count> !';
    | errJoker:          S := '"~" should not contain any joker !';
    | errBase:           S := '"~" should not contain an extension !';
    | errAdjusted:       S := 'Ajusted "~" <count> would become 0 !';
    | errExpecting:      S := "Expecting ~ !";
    | errSpecifiedPP:    S := "Passphrase and -x option are mutually exclusive !";
    | errPadNotApplicable:S:= "-p option is a nonsense with -x option !";
    | errAborted:        S := "Aborted by user !";
    ELSE
                         S := "This is illogical, Captain !";
    END;

    CASE e OF
    | errNone,errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (8 * 512) + FIO.BufferOverhead;
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferOut : ioBufferType;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE lcXOR (b1,b2 : LONGCARD) : LONGCARD;
TYPE
    lcBITSET = SET OF [0..31];
BEGIN
    RETURN LONGCARD ( (lcBITSET (b1) / lcBITSET (b2)) );
END lcXOR;

PROCEDURE wXOR (b1,b2 : CARDINAL) : CARDINAL;
TYPE
    wBITSET = SET OF [0..15];
BEGIN
    RETURN CARDINAL ( (wBITSET (b1) / wBITSET (b2)) );
END wXOR;

PROCEDURE fmtlc (v:LONGCARD;base,wi:CARDINAL;pad:CHAR;upper:BOOLEAN  ):str16;
VAR
    R:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(v,R,base,ok);
    FOR i:=Str.Length(R)+1 TO wi DO Str.Prepend(R,pad);END;
    IF NOT(upper) THEN Str.Lows(R);END;
    RETURN R;
END fmtlc;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

MODULE isaac;

IMPORT lcXOR;
IMPORT Str;
(* IMPORT fmtlc,WrStr; *)

EXPORT minisaacstate,maxisaacstate,
       initisaac,updateisaac,initisaacseedstr,initisaacseedval,
       getLCisaacrnd,getisaacpool,procisaacstate,
       DOzeroPOOL,DONOTzeroPOOL;

CONST
    ISAACSTATESIZE = 256;
    minisaacstate  = 0;
    maxisaacstate  = ISAACSTATESIZE-1;

TYPE
    isaacCTX = RECORD
         (* internal state *)
         isaacmem     : ARRAY [minisaacstate..maxisaacstate] OF LONGCARD;
         isaacAA,isaacBB,isaacCC:LONGCARD;
         (* external *)
         isaacpool    : ARRAY [minisaacstate..maxisaacstate] OF LONGCARD; (* ub4 in readable.c *)
         isaacpoolndx : CARDINAL; (* an index, in fact *)
    END;
VAR
    (* local globerks ! *)
    ctx        : isaacCTX;
    ctxbak     : isaacCTX;
    isaacsaved : BOOLEAN;

(* ------------------------------------------------------------ *)

PROCEDURE updateisaac();
VAR
    i,ndx:CARDINAL;
    x,y:LONGCARD;
BEGIN
    INC(ctx.isaacCC);        (* isaacCC just gets incremented once per 256 results *)
    INC(ctx.isaacBB,ctx.isaacCC);     (* then combined with isaacBB *)

    FOR i:=minisaacstate TO maxisaacstate DO
        x := ctx.isaacmem[i];
        CASE (i MOD 4) OF
        | 0: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA << 13) );
        | 1: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA >>  6) );
        | 2: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA <<  2) );
        | 3: ctx.isaacAA := lcXOR(ctx.isaacAA, (ctx.isaacAA >> 16) );
        END;
        ndx         := (i + 128) MOD ISAACSTATESIZE;
        ctx.isaacAA      := ctx.isaacmem[ ndx] + ctx.isaacAA;

        ndx         := CARDINAL ( (x >> 2) MOD ISAACSTATESIZE) ;
        y           := ctx.isaacmem[ ndx] + ctx.isaacAA + ctx.isaacBB;
        ctx.isaacmem[i]   := y;

        ndx         := CARDINAL ( (y >> 10) MOD ISAACSTATESIZE) ;
        ctx.isaacBB      := ctx.isaacmem[ ndx ] + x;
        ctx.isaacpool[i]  := ctx.isaacBB;
    END;
    ctx.isaacpoolndx := minisaacstate; (* reset index *)
END updateisaac;

(* ------------------------------------------------------------ *)

CONST
    DOzeroPOOL = TRUE;
    DONOTzeroPOOL= FALSE;

PROCEDURE initisaac(isaacseedfrompool,zeropool:BOOLEAN);
CONST
    magic = 9E3779B9H; (* the golden ratio *)
VAR
    i:CARDINAL;
    a,b,c,d,e,f,g,h:LONGCARD;
BEGIN
    ctx.isaacAA:=0;
    ctx.isaacBB:=0;
    ctx.isaacCC:=0;
    FOR i:=minisaacstate TO maxisaacstate DO
        ctx.isaacmem[i]  := 0;
        IF zeropool THEN (* DOzeroPOOL *)
            ctx.isaacpool[i]:=0;
        END;
    END;

    a:=magic;
    b:=magic;
    c:=magic;
    d:=magic;
    e:=magic;
    f:=magic;
    g:=magic;
    h:=magic;

    FOR i:=0 TO 3 DO (* scramble it *)
        (* mix *)
        a:=lcXOR(a,b<<11);   INC(d,a);   INC(b,c);
        b:=lcXOR(b,c>> 2);   INC(e,b);   INC(c,d);
        c:=lcXOR(c,d<< 8);   INC(f,c);   INC(d,e);
        d:=lcXOR(d,e>>16);   INC(g,d);   INC(e,f);
        e:=lcXOR(e,f<<10);   INC(h,e);   INC(f,g);
        f:=lcXOR(f,g>> 4);   INC(a,f);   INC(g,h);
        g:=lcXOR(g,h<< 8);   INC(b,g);   INC(h,a);
        h:=lcXOR(h,a>> 9);   INC(c,h);   INC(a,b);
    END;

    FOR i:=minisaacstate TO maxisaacstate BY 8 DO (* fill in isaacmem[] with messy stuff *)
        IF isaacseedfrompool THEN (* use all the information in the seed *)
            INC(a,ctx.isaacpool[i  ]);
            INC(b,ctx.isaacpool[i+1]);
            INC(c,ctx.isaacpool[i+2]);
            INC(d,ctx.isaacpool[i+3]);
            INC(e,ctx.isaacpool[i+4]);
            INC(f,ctx.isaacpool[i+5]);
            INC(g,ctx.isaacpool[i+6]);
            INC(h,ctx.isaacpool[i+7]);
        END;
        (* mix *)
        a:=lcXOR(a,b<<11);   INC(d,a);   INC(b,c);
        b:=lcXOR(b,c>> 2);   INC(e,b);   INC(c,d);
        c:=lcXOR(c,d<< 8);   INC(f,c);   INC(d,e);
        d:=lcXOR(d,e>>16);   INC(g,d);   INC(e,f);
        e:=lcXOR(e,f<<10);   INC(h,e);   INC(f,g);
        f:=lcXOR(f,g>> 4);   INC(a,f);   INC(g,h);
        g:=lcXOR(g,h<< 8);   INC(b,g);   INC(h,a);
        h:=lcXOR(h,a>> 9);   INC(c,h);   INC(a,b);

        ctx.isaacmem[i  ]:=a;
        ctx.isaacmem[i+1]:=b;
        ctx.isaacmem[i+2]:=c;
        ctx.isaacmem[i+3]:=d;
        ctx.isaacmem[i+4]:=e;
        ctx.isaacmem[i+5]:=f;
        ctx.isaacmem[i+6]:=g;
        ctx.isaacmem[i+7]:=h;
    END;

    IF isaacseedfrompool THEN
        (* do a second pass to make all of the seed affect all of isaacmem *)
        FOR i:=minisaacstate TO maxisaacstate BY 8 DO
            INC(a,ctx.isaacmem[i  ]);
            INC(b,ctx.isaacmem[i+1]);
            INC(c,ctx.isaacmem[i+2]);
            INC(d,ctx.isaacmem[i+3]);
            INC(e,ctx.isaacmem[i+4]);
            INC(f,ctx.isaacmem[i+5]);
            INC(g,ctx.isaacmem[i+6]);
            INC(h,ctx.isaacmem[i+7]);

            (* mix *)
            a:=lcXOR(a,b<<11);   INC(d,a);   INC(b,c);
            b:=lcXOR(b,c>> 2);   INC(e,b);   INC(c,d);
            c:=lcXOR(c,d<< 8);   INC(f,c);   INC(d,e);
            d:=lcXOR(d,e>>16);   INC(g,d);   INC(e,f);
            e:=lcXOR(e,f<<10);   INC(h,e);   INC(f,g);
            f:=lcXOR(f,g>> 4);   INC(a,f);   INC(g,h);
            g:=lcXOR(g,h<< 8);   INC(b,g);   INC(h,a);
            h:=lcXOR(h,a>> 9);   INC(c,h);   INC(a,b);

            ctx.isaacmem[i  ]:=a;
            ctx.isaacmem[i+1]:=b;
            ctx.isaacmem[i+2]:=c;
            ctx.isaacmem[i+3]:=d;
            ctx.isaacmem[i+4]:=e;
            ctx.isaacmem[i+5]:=f;
            ctx.isaacmem[i+6]:=g;
            ctx.isaacmem[i+7]:=h;
        END;
    END;

    updateisaac(); (* isaacpoolndx autohandled *)
END initisaac;

PROCEDURE getLCisaacrnd():LONGCARD;
VAR
    v:LONGCARD;
BEGIN
    v:= ctx.isaacpool[ctx.isaacpoolndx];
    INC(ctx.isaacpoolndx);
    IF ctx.isaacpoolndx > maxisaacstate THEN updateisaac(); END; (* isaacpoolndx autohandled *)
    RETURN v;
END getLCisaacrnd;

PROCEDURE getisaacpool (i:CARDINAL):LONGCARD;
BEGIN
    RETURN ctx.isaacpool[i];
END getisaacpool;

PROCEDURE initisaacseedstr (isaacseedfrompool,padwithPP:BOOLEAN; PP:ARRAY OF CHAR);
VAR
    i,p,len:CARDINAL;
    v:LONGCARD;
BEGIN
    initisaac(isaacseedfrompool,DOzeroPOOL);

    (* we could pack string 4 chars BY 4 chars too faking 32 bit values *)

    len:=Str.Length(PP);
    p:=minisaacstate;
    i:=1-1;
    LOOP
        IF i < len THEN
            v:=LONGCARD( ORD(PP[i]) );
            INC(i);
        ELSE
            IF padwithPP THEN
                i:=1-1;
                v:=LONGCARD( ORD(PP[i] ) );
                INC(i);
            ELSE
                v:=0; (* Rosetta code does so *)
            END;
        END;
        ctx.isaacpool[p]:=v;
        (* WrStr(fmtlc(v,16,8,"0",FALSE)); *)
        INC(p);
        IF p > maxisaacstate THEN EXIT; END;
    END;

    initisaac(isaacseedfrompool,DONOTzeroPOOL);
END initisaacseedstr;

PROCEDURE initisaacseedval (isaacseedfrompool:BOOLEAN; PP:ARRAY OF LONGCARD);
VAR
    i,p,len:CARDINAL;
    v:LONGCARD;
BEGIN
    initisaac(isaacseedfrompool,DOzeroPOOL);

    (* we could pack string 4 chars BY 4 chars too faking 32 bit values *)

    FOR i:= minisaacstate TO maxisaacstate DO
        v:=PP[i];
        ctx.isaacpool[p]:=v;
        (* WrStr(fmtlc(v,16,8,"0",FALSE)); *)
    END;

    initisaac(isaacseedfrompool,DONOTzeroPOOL);
END initisaacseedval;

PROCEDURE procisaacstate ( saveit:BOOLEAN  );
BEGIN
    IF saveit THEN
        ctxbak := ctx;
        isaacsaved := TRUE;
    ELSE
        IF isaacsaved THEN ctx := ctxbak; END;
    END;
END procisaacstate;

BEGIN
     isaacsaved:=FALSE;
END isaac;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE fwrLC (h:FIO.File;binexport:BOOLEAN;v:LONGCARD );
CONST
    wilc = 10;
BEGIN
    IF binexport THEN
        FIO.WrBin(h,v,SIZE(LONGCARD));
    ELSE
        FIO.WrLngCard(h,v,wilc);FIO.WrLn(h);
        (* FIO.WrLngHex(h,v,wilc);FIO.WrLn(h); *)
    END;
END fwrLC;

PROCEDURE fwrB (h:FIO.File;binexport:BOOLEAN;v:LONGCARD );
CONST
    wib = 3;
BEGIN
    IF binexport THEN
        FIO.WrBin(h,SHORTCARD(v),SIZE(SHORTCARD));
    ELSE
        FIO.WrShtCard(h,SHORTCARD(v),wib);FIO.WrLn(h);
        (* FIO.WrShtHex(h,SHORTCARD(v),wib);FIO.WrLn(h); *)
    END;
END fwrB;

PROCEDURE getrandchar():CARDINAL;
BEGIN
    RETURN CARDINAL( getLCisaacrnd() MOD 95 + ORD(" ") ); (* from space TO underscore *)
END getrandchar;

PROCEDURE vernam(VAR R:ARRAY OF CHAR;S:ARRAY OF CHAR);
VAR
    i,code,v: CARDINAL;
BEGIN
    Str.Copy(R,"");
    FOR i := 1 TO Str.Length(S) DO
        code:=ORD(S[i-1]);
        v:=wXOR(code,getrandchar());
        Str.Append(R,CHR(v));
    END;
END vernam;

PROCEDURE dmphex (S:ARRAY OF CHAR );
VAR
    i,v:CARDINAL ;
BEGIN
    FOR i:=1 TO Str.Length(S) DO
        v:=ORD(S[i-1]);
        WrStr( fmtlc (LONGCARD(v),16,2,"0",FALSE));
    END;
END dmphex;

PROCEDURE wr (quoted:BOOLEAN; S:ARRAY OF CHAR);
BEGIN
    IF quoted THEN WrStr(dquote);END;
    WrStr(S);
    IF quoted THEN WrStr(dquote);END;
END wr;

PROCEDURE chkdmp (DEBUG:BOOLEAN;lines:CARDINAL):BOOLEAN;
VAR
    rc:BOOLEAN;
BEGIN
    rc:=FALSE;
    IF DEBUG THEN
        rc:=TRUE;
    ELSE
        CASE lines OF
        | 1,2   : rc:=TRUE; (* 2*256 entries, 8 entries per line *)
        | 63,64 : rc:=TRUE; (* 512 DIV 8 = 64 *)
        END;
    END;
    RETURN rc;
END chkdmp;

PROCEDURE dmpbool (tf:BOOLEAN;S,strYes,strNo:ARRAY OF CHAR);
BEGIN
    WrStr(S);
    IF tf THEN
        WrStr(strYes);
    ELSE
        WrStr(strNo);
    END;
    WrLn;
END dmpbool;

PROCEDURE dmpcard (v:LONGCARD;S:ARRAY OF CHAR );
BEGIN
    WrStr(S);WrStr(fmtlc(v,10,10," ",FALSE));WrLn;
END dmpcard;

(* ------------------------------------------------------------ *)

PROCEDURE genPP (VAR S : str256 );
VAR
    i: CARDINAL;
BEGIN
    Lib.RANDOMIZE;
    Str.Copy(S,"");
    FOR i:=minisaacstate TO maxisaacstate DO
        (* avoid chr(0) : 1+ [0..254] i= [1..255] *)
        Str.Append(S,CHR( 1+Lib.RANDOM(255-1) ) );
    END;
END genPP;

PROCEDURE genPPLC (VAR PP : ARRAY OF LONGCARD);
VAR
    i,lo,hi: CARDINAL;
    v:LONGCARD;
BEGIN
    Lib.RANDOMIZE;
    FOR i:=minisaacstate TO maxisaacstate DO
        lo := Lib.RANDOM(MAX(CARDINAL) );
        hi := Lib.RANDOM(MAX(CARDINAL) );
        v:=LONGCARD(hi) << 16 + LONGCARD(lo);
        PP[i]:=v;
    END;
END genPPLC;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    sizevalue = 256;
    minvalue  = 0;
    maxvalue  = sizevalue-1;
VAR
    value : ARRAY[minvalue..maxvalue] OF LONGCARD;
    valuetotal :LONGCARD;
    valueoverflow : BOOLEAN;

PROCEDURE initvalues (  );
VAR
    i:CARDINAL ;
BEGIN
    FOR i:=minvalue TO maxvalue DO value[i]:=0;END;
    valueoverflow:=FALSE;
END initvalues;

PROCEDURE updatevalue ( v:LONGCARD);
VAR
    ndx:CARDINAL;
BEGIN
    ndx:= CARDINAL(v MOD sizevalue);
    INC(value[ndx]);
    IF valuetotal = MAX(LONGCARD) THEN
        valueoverflow:=TRUE;
        (* valuetotal := 0; *) (* let value as is for division *)
    ELSE
        INC(valuetotal);
    END;
END updatevalue;

PROCEDURE showmissing ( );
CONST
    wib = 3;
VAR
    i : CARDINAL;
BEGIN
    IF valueoverflow THEN
        (* WrStr("::: More than MAX(LONGCARD) values !");WrLn; *)
        RETURN;
    END;
    FOR i:=minvalue TO maxvalue DO
        IF value[i] = 0 THEN
            WrStr("         Missing value  : ");
            IO.WrCard(i,wib);WrLn;
        END;
    END;
END showmissing;

(* ------------------------------------------------------------ *)

CONST
    sizedice = 6;
    mindice  = 0;
    maxdice  = sizedice-1;
VAR
    dice : ARRAY[mindice..maxdice] OF LONGCARD; (* yes, yes, LONGREALs would do too *)
    dicetotal : LONGCARD;
    diceoverflow : BOOLEAN;

PROCEDURE initdice();
VAR
    i : CARDINAL;
BEGIN
    FOR i:=mindice TO maxdice DO dice[i]:=0;END;
    dicetotal := 0;
    diceoverflow := FALSE;
END initdice;

PROCEDURE updatedice(v:LONGCARD);
VAR
    ndx:CARDINAL;
BEGIN
    ndx:= CARDINAL(v MOD sizedice);
    INC(dice[ndx]);
    IF dicetotal = MAX(LONGCARD) THEN
        diceoverflow:=TRUE;
        (* dicetotal := 0; *) (* let value as is for division *)
    ELSE
        INC(dicetotal);
    END;
END updatedice;

PROCEDURE showdice ( );
CONST
    wilc = 10;
VAR
    i,j,p : CARDINAL;
    fl,sum: LONGREAL;
    S:str128;
    ok:BOOLEAN;
BEGIN
    IF diceoverflow THEN
        (* WrStr("::: More than MAX(LONGCARD) dices !");WrLn; *)
        RETURN;
    END;
    sum:=0.0;
    FOR i:=mindice TO maxdice DO
        WrStr("         Side ");
        IO.WrCard(i+1,1);WrStr(" of dice : ");IO.WrLngCard(dice[i],wilc);
        fl:=LONGREAL(dice[i]) / LONGREAL(dicetotal) * 100.0;
        sum:=sum+fl; (* utterly useless ! *)
        Str.FixRealToStr(fl,6,S,ok);
        p:=Str.CharPos(S,dot);
        IF p # MAX(CARDINAL) THEN
            INC(p);
            FOR j:=p TO 3 DO Str.Prepend(S," ");END;
        END;
        WrStr(" (");WrStr(S);WrStr("%)");WrLn;
    END;

    WrStr("         Total          : ");
    IO.WrLngCard(dicetotal,wilc);
    Str.FixRealToStr(sum,6,S,ok);
    WrStr(" (");WrStr(S);WrStr("%)");WrLn;
END showdice;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    rosettamsg = "a Top Secret secret";
    rosettakey = "this is my secret key";
    expectedhex= "1c0636190b1260233b35125f1e1d0e2f4c5422";
    expectedvect="f650e4c8e448e96d98db2fb4f5fad54f433f1afbedec154ad837048746ca4f9a"+nl+
                 "5de3743e88381097f1d444eb823cedb66a83e1e04a5f6355c744243325890e2e"+nl+
                 "..."+nl+
                 "94d401258aba3ffbd32be0608c17eff021e2547e07cffad905340e15f3310c92"+nl+
                 "9d8d190886ba527ff943f672ef73fbf046d95ca5c54cd95b9d855e894bb5af29";
    msgWait = "Working, please wait... ";
VAR
    opt,parmcount:CARDINAL;
    i,j,k,lines:CARDINAL;
    S,R:str128;
    state:(waiting,gotcount,gotbase,gotpassphrase);
    hout:FIO.File;
    flagmissing,PPstring,padwithPP,flagdice,binexport,isaacseedfrompool : BOOLEAN;
    DEBUG,alcatraz,dmpme:BOOLEAN;
    x,abcd,count,localcount:LONGCARD;
    basename:str128;
    fl:LONGREAL;
    PP : str256;
    vlc : ARRAY[minisaacstate..maxisaacstate] OF LONGCARD;
BEGIN
    (* Lib.DisableBreakCheck(); *)
    FIO.IOcheck := FALSE;
    FIO.ShareMode:=FIO.ShareDenyNone; (* very, very important ! *)
    WrLn;

    basename          := progEXEname; (* useless *)
    isaacseedfrompool := TRUE;
    binexport         := TRUE;
    flagdice          := FALSE;
    padwithPP         := FALSE;
    PPstring          := TRUE;
    flagmissing       := FALSE;
    DEBUG             := FALSE;
    Str.Copy(PP,"");

    parmcount := Lib.ParamCount();
    state     := waiting;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "S"+delim+"STRONGER"+delim+"F"+delim+"FLAG"+delim+
                                  "T"+delim+"TEXT"+delim+"A"+delim+"ASCII"+delim+
                                  "D"+delim+"DICE"+delim+
                                  "P"+delim+"PAD"+delim+
                                  "X"+delim+"LONGCARD"+delim+
                                  "M"+delim+"MISSING"+delim+
                                  "DEBUG"
                              );
            CASE opt OF
            | 1,2,3   :   abort(errHelp,"");
            | 4,5,6,7 :   isaacseedfrompool := FALSE;
            | 8,9,10,11 : binexport         := FALSE;
            | 12,13:      flagdice          := TRUE;
            | 14,15:      padwithPP         := TRUE;
            | 16,17:      PPstring          := FALSE;
            | 18,19 :     flagmissing       := TRUE;
            | 20:         DEBUG             := TRUE;
            ELSE
                abort(errOption,S);
            END;
        ELSE
            CASE state OF
            | waiting :
                ReplaceChar(R,dot,"");
                ReplaceChar(R,coma,"");
                Str.Prepend(R,"-?:");
                IF GetLongCard(R,count)=FALSE THEN abort(errBadCount,S);END;
                IF count=0 THEN abort(errAdjusted,S);END;
                IF ( count - (count MOD roundsize) ) = 0 THEN abort(errAdjusted,S);END;
            | gotcount :
                IF chkJoker(S) THEN abort(errJoker,S);END;
                IF Str.CharPos(S,dot) # MAX(CARDINAL) THEN abort(errBase,S);END;
                Str.Copy(basename,S);
            | gotbase  :
                Str.Copy(PP,S);
            | gotpassphrase:
                abort(errTooManyParms,S);
            END;
            INC(state);
        END;
    END;
    CASE state OF
    | waiting:
        IF NOT(DEBUG) THEN
            WrStr("::: Computed first and last lines :");WrLn;
        END;

        isaacseedfrompool := TRUE;  (* forced *)
        padwithPP         := FALSE; (* forced *)

        lines:=1;

        initisaac(isaacseedfrompool,DOzeroPOOL);

        FOR i:=1 TO 2 DO (* //1 *)
            updateisaac();
            FOR j:=minisaacstate TO maxisaacstate DO
                dmpme:=chkdmp(DEBUG,lines);
                IF dmpme THEN WrStr(fmtlc(getisaacpool(j),16,8,"0",FALSE)); END;
                CASE (j AND 7) OF
                | 0..6 :
                    ; (* WrStr(" "); *)
                | 7 :
                    IF dmpme THEN WrLn; END;
                    INC(lines);
                    IF (NOT(DEBUG) AND (lines=3)) THEN WrStr("...");WrLn;END;
                END;
            END;
        END;
        WrLn;

        WrStr("::: Expected first and last lines from RANDVECT.TXT :");WrLn;
        WrStr(expectedvect);WrLn;
        WrLn;
        WrStr("::: message  : ");wr(TRUE,rosettamsg);WrLn;
        WrStr("::: key      : ");wr(TRUE,rosettakey);WrLn;

        initisaacseedstr (isaacseedfrompool,padwithPP,rosettakey);
        vernam(R, rosettamsg);

        WrStr("::: encoded  : ");dmphex(R);WrLn;
        WrStr("::: expected : ");WrStr(expectedhex);WrLn;

        initisaacseedstr (isaacseedfrompool,padwithPP,rosettakey);
        vernam(S, R);

        WrStr("::: decoded  : ");wr(TRUE,S);WrLn;
    | gotcount:
        abort(errExpecting,"<basename> [passphrase]");
    | gotbase,gotpassphrase:
        IF same(PP,"") THEN (* state if gotbase *)
            IF PPstring THEN
                genPP(PP);
                initisaacseedstr (isaacseedfrompool,padwithPP,PP);
            ELSE
                IF padwithPP THEN abort(errPadNotApplicable,"");END;
                genPPLC(vlc);
                initisaacseedval(isaacseedfrompool,vlc);
            END;
        ELSE
            IF PPstring=FALSE THEN abort(errSpecifiedPP,"");END;
            initisaacseedstr (isaacseedfrompool,padwithPP,PP);
        END;

        (* initisaac(isaacseedfrompool,DOzeroPOOL); (* gives same sequence as in test *) *)
        procisaacstate(TRUE);

        dmpcard(count,            "Specified count : ");
        count := count - (count MOD roundsize);
        dmpcard(count,            "Adjusted count  : ");
        WrStr("Base name       : ");wr(TRUE,basename);WrLn;
        dmpbool(binexport,        "Export format   : ","binary","text");
        dmpbool(isaacseedfrompool,"Seed from pool  : ","yes","no");
        dmpbool(padwithPP,        "Pool padding    : ","repeated passphrase","$00");
        dmpbool(PPstring,         "Auto-passphrase : ","string","array of LONGCARDs");

        i:=minMethod;
        LOOP
            localcount := count;

            procisaacstate(FALSE);
            initdice();
            initvalues();

            video(msgWait,TRUE);
            Str.Concat(R,basename,dot);Str.Append(R,CHR(ORD("A")+i-1));
            (* IF FIO.Exists(R) THEN FIO.Erase(R);END; *)
            hout:=FIO.Create(R);
            FIO.AssignBuffer(hout,ioBufferOut);

            alcatraz := FALSE;
            LOOP
                IF localcount = 0 THEN EXIT;END;
                updateisaac();
                FOR j:=minisaacstate TO maxisaacstate DO
                     abcd:=getisaacpool(j);
                     CASE i OF
                     | methodABCD:
                         x:=abcd;
                         fwrLC(hout,binexport,x);
                         updatedice(x); (* one 32-bit value *)
                         FOR k:=1 TO 4 DO
                             CASE k OF
                             | 4: x:=(abcd AND 0FF000000H) >> 24;
                             | 3: x:=(abcd AND 0FF0000H  ) >> 16;
                             | 2: x:=(abcd AND 0FF00H    ) >> 8;
                             | 1: x:=(abcd AND 0FFH      );
                             END;
                             updatevalue(x);
                         END;
                     | methodDCBA:
                         FOR k:=1 TO 4 DO
                             CASE k OF
                             | 1: x:=(abcd AND 0FF000000H) >> 24;
                             | 2: x:=(abcd AND 0FF0000H  ) >> 16;
                             | 3: x:=(abcd AND 0FF00H    ) >> 8;
                             | 4: x:=(abcd AND 0FFH      );
                             END;
                             fwrB(hout,binexport,x);
                             (* updatedice(x); four 8-bit values *)
                             updatevalue(x);
                         END;
                         updatedice(x); (* one 32-bit value *)
                     | methodMOD255:
                         x := abcd MOD 0FFH;
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | methodMOD256:
                         x := abcd MOD 100H;
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | methodDIV1000000H:
                         x := abcd DIV 1000000H;
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | methodSHIFT24:
                         x := abcd >> 24;
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | methodDIVMAXLC256:
                         fl:=LONGREAL(abcd) / FMAXLC;
                         fl:=fl*FMAXBPLUS1;
                         x:=LONGCARD(fl);
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | methodDIVRANGE256:
                         fl:=LONGREAL(abcd) / FMAXLCPLUS1;
                         fl:=fl*FMAXBPLUS1;
                         x:=LONGCARD(fl);
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | methodDIVMAXLC255:
                         fl:=LONGREAL(abcd) / FMAXLC;
                         fl:=fl*FMAXB;
                         x:=LONGCARD(fl);
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     | method32TO8:
                         x := abcd DIV DIV32TO8;
                         fwrB(hout,binexport, x);
                         updatedice(x);
                         updatevalue(x);
                     END;
                     DEC(localcount);
                END;
                alcatraz:=IO.KeyPressed();
                IF alcatraz THEN EXIT;END;
            END;
            FIO.Close(hout);
            video(msgWait,FALSE);
            WrStr("Created         : ");wr(TRUE,R);WrLn;
            IF alcatraz THEN EXIT; END;
            IF flagdice THEN
                showdice();
                IF i < maxMethod THEN
                    IF NOT(flagmissing) THEN WrLn;END;
                END;
            END;
            IF flagmissing THEN
                showmissing();
            END;
            INC(i);
            IF i > maxMethod THEN EXIT; END;
        END;
        IF alcatraz THEN abort(errAborted,"");END;
    END;
    abort(errNone,"");
END tstIsaac.
