Option Explicit
DefInt A-Z

Function DateToJD# (ByVal zDay#, ByVal zMonth#, ByVal zYear#)
'   assume day with decimal, and whole date is ok
'   valid only for JD >= 0 i.e. NOT before 1.5 Jan -4712 !
'   input  : day (with decimals), month, year
'   output : JD
'   do not forget to add deltaTU if necessary!
Dim YYYYMMDD#
Dim A#
Dim gregorian#
Dim JD#

YYYYMMDD# = zYear# * 10000# + zMonth# * 100# + zDay#
If zMonth# < 2.5 Then
    zYear# = zYear# - 1#
    zMonth# = zMonth# + 12#
End If
If YYYYMMDD# < 15821015# Then  ' before 15 october 1582, julian calendar
    gregorian# = 0#
Else
    A# = vint#(zYear# / 100#)
    gregorian# = 2# - A# + vint#(A# / 4#)
End If
JD# = vint#(365.25 * (zYear# + 4716#)) + vint#(30.6001 * (zMonth# + 1#))
JD# = JD# + zDay# + gregorian# - 1524.5
DateToJD# = JD#
End Function

Function frac# (ByVal v#)
'   input  : value
'   output : frac(value)
'
'   frac keeps sign
frac# = v# - vint#(v#)
End Function

Sub JDtoDate (ByVal JD#, zDay#, zMonth#, zYear#)
'   assume JD >= 0
'   input  : JD
'   output : day (with decimals), month, year
Dim Z#, F#, A#, alf#, B#, C#, D#, E#

JD# = JD# + .5
Z# = vint#(JD#)
F# = frac(JD#)
If Z# < 2299161# Then
    A# = Z#
Else
    alf# = vint((Z# - 1867216.25) / 36524.25)
    A# = Z# + 1# + alf# - vint#(alf# / 4#)
End If
B# = A# + 1524#
C# = vint#((B# - 122.1) / 365.25)
D# = vint#(365.25 * C)
E# = vint#((B# - D#) / 30.6001)
zDay# = B# - D# - vint#(30.6001 * E#) + F#
If E# < 13.5 Then
    zMonth# = E# - 1#
Else
    zMonth# = E# - 13#
End If
If zMonth# > 2.5 Then
    zYear# = C# - 4716#
Else
    zYear# = C# - 4715#
End If
End Sub

Function JDtoJD0H# (ByVal JD#)
'   find JD at 0h
'   input  : JD
'   output : JD for this day at 0H
Dim JD0H#

If frac(JD#) < .5 Then
    JD0H# = vint#(JD#) - .5  ' (* was < 23H59 so take previous 0H *)
Else
    JD0H# = vint#(JD#) + .5  ' (* was 0H or later so take current 0H *)
End If
JDtoJD0H = JD0H#
End Function

Function JDtoWeekday% (ByVal JD#)
'   0=sunday, 1=monday... 6=saturday
'   input  : JD
'   output : day of week
Dim v#
v# = vmod#((JDtoJD0H#(JD#) + 1.5), 7#)
JDtoWeekday% = v#
End Function

Function vint# (ByVal v#)
'   input  : value
'   output : int(value) ie int(0.95) gives 0 (no rounding)
'
'   TRUNC(n) gives only a CARDINAL without rounding either
vint# = Int(v#)
End Function

Function vmod# (ByVal v1#, ByVal v2#)
'   input  : value1, value2
'   output : value1 mod value2
'
'   mod keeps sign
vmod# = v1# Mod v2#
End Function

