(*

a Q&D visual Wolfram rules tester

guess my original typo on name had something TO do WITH Wolfram & Hart ;-)
or with keyboard as well...

notes :
- we don't try and be clever for this small hack
  (we just wanted to quickly experiment CA rules,
  and at first it was merely designed TO check rule 30)
- safety checks are limited

to do :
- dump CA state as to check its randomness ?
- dump screen to PCX ?
- handle wider than 3 bits patterns ? 5 bits could do
- toggle infos on screen while running ?
- allow "?" bit state from cli ?

default rule is rule 30
   1. If the cell and its right-hand neighbor are both 0,
      then the new value of the cell is the value of its left-hand neighbor.
   2. Otherwise, the new value of the cell is the opposite
      of the value of its left-hand neighbor.

   3. Another description : x(n+1,i) = x(n,i-1) xor [x(n,i) or x(n,i+1)]

   4. Using patterns :
      current pattern           111 110 101 100 011 010 001 000
      new state for center cell  0   0   0   1   1   1   1   0

*)

MODULE wolfRule;

IMPORT Graph;
IMPORT IO;
IMPORT Lib;
IMPORT Str;
IMPORT SYSTEM;
IMPORT FIO;

FROM IO IMPORT WrStr,WrLn;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileRename, fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

(* ------------------------------------------------------------ *)

CONST
    DBG = FALSE;

(*%T DBG  *)
PROCEDURE dbgv (S:ARRAY OF CHAR;v:CARDINAL );
CONST
    F = "WOLFRULE.DBG";
VAR
    hout:FIO.File;
BEGIN
    IF FIO.Exists(F) THEN
        hout:=FIO.Append(F);
    ELSE
        hout:=FIO.Create(F);
    END;
    FIO.WrStr(hout,S);FIO.WrStr(hout," = ");
    FIO.WrCard(hout,v,0);FIO.WrLn(hout);
    FIO.Close(hout);
END dbgv;

CONST
    DBGFILE = "WOLFRULE.DBG";

PROCEDURE dbg (S,R:ARRAY OF CHAR   );
VAR
    hout:FIO.File;
BEGIN
    IF FIO.Exists(DBGFILE) THEN
        hout:=FIO.Append(DBGFILE);
    ELSE
        hout:=FIO.Create(DBGFILE);
    END;
    FIO.WrStr(hout,S); FIO.WrStr(hout," = ");
    FIO.WrStr(hout,R);FIO.WrLn(hout);
    FIO.Close(hout);
END dbg;
(*%E  *)

(* ------------------------------------------------------------ *)

CONST
    exe        = "WOLFRULE";
    allowed    = "01 .X";
    allowedrnd = allowed+"?"; (* "?" means random "0" OR "1" *)
    ver        = "v1.0b";
    extTXT     = ".TXT";
    extWR      = ".RUL";
CONST
    msghelp= "Q&D Wolfram Rule "+ver+" by PhG"+nl+
nl+
"Syntax : "+exe+" [file["+extWR+"|"+extTXT+"]] [option]..."+nl+
nl+
"-r:#   8 bits bitmask"+nl+
"-w:#   initial cellular automaton width (default is 1)"+nl+
"-i:$   initial binary pattern"+nl+
"-n     do not force at least one bit on in initial pattern"+nl+
"-b     black and white mode (default is color mode)"+nl+
"-v     wait for screen vertical retrace"+nl+
"-t[:#] testmode (-t:# = -t -r:#)"+nl+
nl+
"a) If no bit is on in initial pattern, background will be red."+nl+
"b) Note wrapping is enabled for first and last cellular automaton cells."+nl+
"c) -r:# bitmask assumes 111 110 101 100 011 010 001 000 order ;"+nl+
"   it can be expressed as a $hexadecimal, decimal or %binary number."+nl+
"d) While program is running, [Space] = pause, [Return] = quit."+nl+
'   In test mode, "+" = increment rule, "-" = decrement rule.'+nl+
"e) -t option requires -r:# option."+nl+
"f) -r:# less uninteresting displays seem to be :"+nl+
"   18 22 26 30 45 60 75 82 86 89 90"+nl+
"   101 102 105 110 124 126 129 135 137 146 149"+nl+
"   150 153 154 161 165 167 169 181 182 193 195 210 218 225."+nl+
"g) Here is an example of rule file (Wolframm's Rule 30) :"+nl+
nl+
"   ; line 1 : each of the 8 possible patterns"+nl+
"   ; line 2 : new center cell state"+nl+
"   ; optional line 3 : initial pattern (-w:# ignored)"+nl+
"   ; without optional line 3, initial pattern is randomly created"+nl+
"   111 110 101 100 011 010 001 000"+nl+
"    0   0   0   1   1   1   1   0"+nl+
"   1"+nl+
'   ; alternate characters are "." for 0 and "x" for 1'+nl+
'   ; "?" will randomly return either "0" or "1"'+nl+
'   ; undefined patterns will force "0" for new center cell state'+nl
(*%F DBG  *)
;
(*%E  *)
(*%T DBG  *)
+
nl+
"Debug data dumped to "+DBGFILE+" text file."+nl;
(*%E  *)

CONST
    errNone         = 0;
    errHelp         = 1;
    errBadWidth     = 2;
    errWidthRange   = 3;
    errOption       = 4;
    errParm         = 5;
    errJoker        = 6;
    errNotFound     = 7;
    errGraphMode    = 8;
    errCharSet      = 9;
    errThreeLines   = 10;
    errMaxPatterns  = 11;
    errMaxInitPattern=12;
    errCount        = 13;
    errBadMask      = 14;
    errRuleOrFile   = 15;
    errBadInitial   = 16;
    errInitialOrFile= 17;
    errTestMode     = 18;
    errRuleRequired = 19;

PROCEDURE abort (e:CARDINAL);
VAR
   S : str128;
BEGIN
   CASE e OF
   | errNone:   ;
   | errHelp:        WrStr(msghelp);
   | errBadWidth :   S := "Illegal -w:#";
   | errWidthRange:  S := "Out of [1..640] range -w:#";
   | errOption :     S := "Illegal option";
   | errParm   :     S := "At least one parameter too many";
   | errJoker  :     S := "Jokers are not allowed";
   | errNotFound:    S := "File does not exist";
   | errGraphMode:   S := "Cannot set graphics mode";
   | errCharSet:     S := "Only characters from ["+allowed+"][?] are allowed in rule file entries";
   | errThreeLines:  S := "Rule file cannot contain more than 3 lines";
   | errMaxPatterns :S := "More than 8 patterns or new states";
   | errMaxInitPattern:S:="More than 1 initial pattern";
   | errCount      : S := "Patterns and new states count mismatch";
   | errBadMask    : S := "Illegal -r:# value";
   | errRuleOrFile : S := "-r:# option and rule file specification are mutually exclusive";
   | errBadInitial : S := "Illegal -i:$ pattern";
   | errInitialOrFile:S:= "-i:$ option and rule file specification are mutually exclusive";
   | errTestMode:    S := "-t option and rule file specification are mutually exclusive";
   | errRuleRequired:S := "-t option requires -r:# option";
   ELSE
                     S := "This is impossible, Captain";
   END;
   CASE e OF
   | errNone,errHelp:
       e:=errNone;
   ELSE
       WrStr(exe+" : ");WrStr(S);WrStr(" !");WrLn;
   END;
   Lib.SetReturnCode( SHORTCARD(e) );
   HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    gmode  = Graph._VRES16COLOR; (* 640x480x4 *)
    tmode  = Graph._DEFAULTMODE;
CONST
    black  = 0;
    white  = 15;
    yellow = 14;
    mauve  = 13;
    red    = 12;
    blue   = 11;
    green  = 10;
    minink = black+1;
    maxink = white;

(* ------------------------------------------------------------ *)

VAR
    xmin,xmax,ymin,ymax : CARDINAL;

PROCEDURE retrace (n:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=1 TO n DO
        WHILE (SYSTEM.In(03DAH) AND 08H) # 0 DO
        END;
        WHILE (SYSTEM.In(03DAH) AND 08H) = 0 DO
        END;
    END;
END retrace;

PROCEDURE HiresON () : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(gmode);
END HiresON;

PROCEDURE HiresOFF() : BOOLEAN;
BEGIN
    RETURN Graph.SetVideoMode(tmode);
END HiresOFF;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE XOR (b1,b2 : CARDINAL) : CARDINAL;
BEGIN
    RETURN CARDINAL ( (BITSET (b1) / BITSET (b2)) );
END XOR;

CONST
    maxpatterns     = 8;
    maxinitpattern  = 1;
    defaultPat      = "111 110 101 100 011 010 001 000"; (* all bitmasks from 7 to 0 *)
    defaultNewState = " 0   0   0   1   1   1   1   0";
    defaultInit     = ""; (* empty = random, could be "1" *)

PROCEDURE readRule (VAR sPat,sNewState,sInit:ARRAY OF CHAR;
                   useLFN:BOOLEAN;rulefile:pathtype  ):CARDINAL;
VAR
    hin:FIO.File;
    S:str1024;
    rc:CARDINAL;
    grabmode:(waiting,gotpat,gotnewstate,gotinit);
BEGIN
    rc:=errNone;
    Str.Copy(sPat,"");
    Str.Copy(sNewState,"");
    Str.Copy(sInit,"");
    grabmode:=waiting;
    FIO.EOF:=FALSE;
    hin:=fileOpenRead(useLFN,rulefile);
    LOOP
        FIO.RdStr(hin,S);
        (*%T DBG  *) dbg("S",S); (*%E *)
        LtrimBlanks(S);
        RtrimBlanks(S);
        CASE S[0] OF
        | charnull,semicolon,pound : ;
        ELSE
            ReplaceChar(S,tabul,space);
            CASE grabmode OF
            | waiting:     IF verifyString(S,allowed)=FALSE THEN
                               rc:=errCharSet;
                               EXIT;
                           END;
                           Str.Copy(sPat,S);
            | gotpat:      IF verifyString(S,allowedrnd)=FALSE THEN
                               rc:=errCharSet;
                               EXIT;
                           END;
                           Str.Copy(sNewState,S);
            | gotnewstate: IF verifyString(S,allowed)=FALSE THEN
                               rc:=errCharSet;
                               EXIT;
                           END;
                           Str.Copy(sInit,S);
            | gotinit:     rc:=errThreeLines; EXIT;
            END;
            INC(grabmode);
        END;
        IF FIO.EOF THEN EXIT;END;
    END;
    fileClose(useLFN,hin);
    IF rc=errNone THEN
        (* any error is an error is an error is an error... *)
        IF argc(sPat,FALSE) > maxpatterns THEN rc:=errMaxPatterns;END;
        IF argc(sNewState,FALSE) > maxpatterns THEN rc:=errMaxPatterns;END;
        IF argc(sInit,FALSE) > maxinitpattern THEN rc:=errMaxInitPattern;END;
        IF rc=errNone THEN
            IF argc(sPat,FALSE) # argc(sNewState,FALSE) THEN rc:=errCount;END;
        END;
        (* //FIXME yes, yes, we should check lmr's and newcenter's length *)
    END;
    RETURN rc;
END readRule;

(* ------------------------------------------------------------ *)

CONST
    MAXHASH = MAX(CARDINAL)-1;
    CAmin = 0;
    CAmax = 640-1;
    firstclock = 1;
    lastclock  = 2;
    minbitpattern = 1;
    maxbitpattern = 8;
TYPE
    str3 = ARRAY [0..2] OF CHAR;
    transformtype = RECORD
        hash : CARDINAL;
        lmr : str3;
        newcenter : CHAR;
    END;

TYPE
    CAtype = ARRAY[CAmin..CAmax] OF CHAR; (* bit state i.e. BOOLEAN *)
VAR
    CA : ARRAY [firstclock..lastclock] OF CAtype;
    bitpattern : ARRAY[minbitpattern..maxbitpattern] OF transformtype;

PROCEDURE chToVal ( ch:CHAR ):CARDINAL ;
VAR
    v:CARDINAL;
BEGIN
    CASE ch OF
    | "0", dot : v:=0;
    | "1", "X" : v:=1;
    ELSE
        v:=MAX(CARDINAL); (* should never happen *)
        v:=0; (* //HACK *)
    END;
    RETURN v;
END chToVal;

PROCEDURE valToCh ( v:CARDINAL ):CHAR;
VAR
    ch:CHAR;
BEGIN
    CASE v OF
    | 0 : ch:="0";
    | 1 : ch:="1";
    ELSE
        ch:=charnull;
        ch:="0";       (* //HACK *)
    END;
    RETURN ch;
END valToCh;

PROCEDURE initCA (VAR ndx,inkbk:CARDINAL;
                 initwi,last:CARDINAL; forceonelit:BOOLEAN;
                 sPat,sNewState,sInit:ARRAY OF CHAR);
VAR
     hash,i,p,v,cx,x,lit:CARDINAL;
     rnd:BOOLEAN;
     lmr:str3;
     newcenter:CHAR;
BEGIN
     (*%T DBG *) dbgv("initwi at entry",initwi); (*%E  *)
     (*%T DBG *) dbg("sPat",sPat); (*%E  *)
     (*%T DBG *) dbg("sNewState",sNewState); (*%E  *)
     (*%T DBG *) dbg("sInit",sInit); (*%E  *)
     rnd:=same(sInit,"");
     IF NOT(rnd) THEN
         initwi := Str.Length(sInit);
     END;
     (*%T DBG *) dbgv("initwi",initwi); (*%E  *)
     DEC(initwi);

     FOR i:=CAmin TO last DO
         CA[firstclock,i]:="0";
         CA[lastclock ,i]:="0";
     END;

     lit:=0;
     cx:=(last-CAmin) DIV 2;   (* 639-0 => 319 *)
     x:=cx - (initwi DIV 2);     (* initwi=0 => 319-0 , initwi=8 => 319-4 *)
     FOR i:=CAmin TO initwi DO
         p:=x+i;
         (*%T DBG *) dbgv("screen index",p); (*%E *)
         IF rnd THEN
             v:=Lib.RANDOM(2); (* 0..1 *)
         ELSE
             v:=chToVal( sInit[i] );
         END;
         IF v = 1 THEN INC(lit);END;
         CA[firstclock,p]:=valToCh(v);
     END;
     (*%T DBG   *)dbgv("lit cells",lit); (*%E *)
     IF forceonelit THEN
         IF lit = 0 THEN
             p:=cx - Lib.RANDOM(initwi+1) DIV 2;  (* +1 because of previous DEC() *)
             CA[firstclock,p]:="1";
             lit:=1;
             (*%T DBG *) dbgv("force lit cell",p); (*%E  *)
         END;
     END;

     IF lit=0 THEN
         inkbk:=red;
     ELSE
         inkbk:=black;
     END;

     (* now parse sPat and sNewState *)

     lmr:="";     newcenter:="";
     lmr:="000";  newcenter:="0";  (* //HACK *)

     hash:=Lib.HashString(lmr,MAXHASH);
     FOR i:=minbitpattern TO maxbitpattern DO
         hash:=hash;
         bitpattern[i].lmr:=lmr;
         bitpattern[i].newcenter:=newcenter;
     END;

     FOR i:=minbitpattern TO argc(sPat,FALSE) DO
         argv(lmr, sPat,i,FALSE);
         argv(newcenter, sNewState,i,FALSE);
         hash:=Lib.HashString(lmr,MAXHASH);
         bitpattern[i].hash := hash;
         bitpattern[i].lmr := lmr;
         bitpattern[i].newcenter := newcenter;
     END;

     ndx:=firstclock;
END initCA;

(* for now, limited to 3 bits patterns *)

PROCEDURE processline (ndx,last:CARDINAL ):CARDINAL ;
VAR
    j,hash,i,nextndx:CARDINAL;
    l,m,r:CHAR;
    newcenter:CHAR;
    lmr:str3;
BEGIN
    CASE ndx OF
    | firstclock : nextndx:=lastclock;
    | lastclock  : nextndx:=firstclock;
    END;

    FOR i:= CAmin TO last DO
        m := CA[ ndx,i ];
        IF i = CAmin THEN
            l := CA [ndx,last];
            r := CA [ndx,i+1 ];
        ELSIF i= last THEN
            l := CA [ndx,last-1];
            r := CA [ndx,CAmin];
        ELSE
            l := CA [ndx,i-1];
            r := CA [ndx,i+1];
        END;

        (* yes, yes, storing to string would DO too *)
        Str.Concat(lmr,l,m);
        Str.Append(lmr,r);
        hash:=Lib.HashString(lmr,MAXHASH);

        j:=minbitpattern-1;
        LOOP
            INC(j);
            IF j > maxbitpattern THEN
                newcenter:="";
                newcenter:="0"; (* //HACK *)
                EXIT;
            END;
            IF hash = bitpattern[j].hash THEN
                IF lmr = bitpattern[j].lmr THEN (* collision may happen ! ;-) *)
                    newcenter:=bitpattern[j].newcenter;
                    IF newcenter = "?" THEN
                        newcenter := valToCh( Lib.RANDOM(2) ); (* 0..1 *)
                    END;
                    EXIT;
                END;
            END;
        END;
        CA[nextndx,i]:=newcenter;
    END;
    RETURN nextndx;
END processline;

PROCEDURE showstate (ndx,last,ycurr,ink,inkbk,vsync:CARDINAL):CARDINAL;
VAR
    myink,x:CARDINAL;
BEGIN
    retrace(vsync);
    FOR x:=CAmin TO xmax DO
         IF CA[ndx,x]="1" THEN
             myink:=ink;
         ELSE
             myink:=inkbk;
         END;
         Graph.Plot(x,ycurr,myink);
    END;
    INC(ycurr);
    IF ycurr > ymax THEN ycurr:=ymin;END;
    RETURN ycurr;
END showstate;

PROCEDURE nextink ( ink,ycurr:CARDINAL;cycle:BOOLEAN  ):CARDINAL;
BEGIN
    IF cycle THEN
        IF ycurr = ymin THEN
            (* ugly
            INC(ink);
            IF ink > maxink THEN ink:=minink;END; (* black avoided *)
            *)
            CASE ink OF
            | white : ink:=green;
            | green : ink:=white;
            END;
        END;
    END;
    RETURN ink;
END nextink;

PROCEDURE gotkey (VAR ch:CHAR):BOOLEAN;
VAR
    got:BOOLEAN;
BEGIN
    ch:="";
    got:=IO.KeyPressed();
    IF got THEN
        ch:=IO.RdKey();
        IF ch = CHR(0) THEN ch:=IO.RdKey();ch:="";END;
    END;
    RETURN got;
END gotkey;

PROCEDURE showparms (testmode:BOOLEAN;rulemask:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    WrLn;
    WrStr("Original pattern : ");
    FOR i:=minbitpattern TO maxbitpattern DO
        WrStr( bitpattern[i].lmr);
        IF i < maxbitpattern THEN WrStr(" ");END;
    END;
    WrLn;
    WrStr("New center cell  : ");
    FOR i:=minbitpattern TO maxbitpattern DO
        WrStr(" ");
        WrStr( bitpattern[i].newcenter);
        IF i < maxbitpattern THEN WrStr("  ");END;
    END;
    WrLn;
    IF testmode THEN
        WrStr("Rule             : ");IO.WrCard(rulemask,0);WrLn;
    END;
END showparms;

PROCEDURE grabrulemask (VAR rule:CARDINAL;S:ARRAY OF CHAR    ):BOOLEAN;
VAR
    ok:BOOLEAN;
    R:str16;
    allowed:str80;
    base:CARDINAL;
    lc:LONGCARD;
BEGIN
    GetString(S, R);
    CASE R[0] OF
    | "$": base:=16; allowed:=digits+"ABCDEF"; Str.Delete(R,0,1);
    | "%": base:=2;  allowed:="01";            Str.Delete(R,0,1);
    ELSE
           base:=10; allowed:=digits;
    END;
    Str.Caps(R); (* already uppercase but who knows ? *)
    ok:=verifyString(R,allowed);
    IF ok THEN
        lc:=Str.StrToCard(R,base,ok);
        IF ok THEN
            ok:= ( lc <= MAX(SHORTCARD) );
            IF ok THEN rule:=CARDINAL(lc);END;
        END;
    END;
    RETURN ok;
END grabrulemask;

PROCEDURE grabinitial ( VAR R:ARRAY OF CHAR; S:ARRAY OF CHAR  ):BOOLEAN;
VAR
    ok:BOOLEAN;
BEGIN
    GetString(S, R);
    IF R[0] = "%" THEN Str.Delete(R,0,1);END;
    ok:=verifyString(R,"01");
    RETURN ok;
END grabinitial;

PROCEDURE testbit(Target, BitNum : CARDINAL) : BOOLEAN;
BEGIN
    RETURN ( (BitNum MOD 16) IN BITSET(Target) );
END testbit;

PROCEDURE byteToMask (VAR R:ARRAY OF CHAR;v:CARDINAL );
VAR
    i:CARDINAL;
BEGIN
    Str.Copy(R,"");
    FOR i:=maxbitpattern TO minbitpattern BY -1 DO (* 8..1 *)
        IF testbit(v,i-1) THEN (* 0-based *)
            Str.Append(R," 1");
        ELSE
            Str.Append(R," 0");
        END;
    END;
END byteToMask;

PROCEDURE maskToByte ( S:ARRAY OF CHAR  ):CARDINAL;
VAR
    lc:LONGCARD;
    base:CARDINAL;
    ok:BOOLEAN;
BEGIN
    ReplaceChar(S,blank,"");
    lc:=Str.StrToCard(S,2,ok);
    RETURN CARDINAL(lc);
END maskToByte;

PROCEDURE newvalue (k:INTEGER;v,mini,maxi:CARDINAL ):CARDINAL ;
VAR
    n:INTEGER;
BEGIN
    n:=INTEGER(v);
    INC(n,k);
    IF n < INTEGER(mini) THEN
        n:=maxi;
    ELSIF n > INTEGER(maxi) THEN
        n:=mini;
    END;
    RETURN CARDINAL(n);
END newvalue;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

VAR
    parmcount,i,opt : CARDINAL;
    S,R,rulefile    : pathtype;
    lc              : LONGCARD;
    ok              : BOOLEAN;
    ch              : CHAR;
    rulemask,ink,inkbk,vsync,ycurr,ndx,initwi            : CARDINAL;
    V:Graph.VideoConfig;
    useLFN,cycle,forceonelit,testmode:BOOLEAN;
    state : (waiting,gotfile);
    sPat,sNewState:str128;
    sInit:str1024;
    sInitCLI:str128;
    k:INTEGER;
BEGIN
    FIO.IOcheck:=FALSE;
    WrLn;

    initwi := 1;
    cycle  := TRUE;
    forceonelit := TRUE;
    vsync  := 0;
    rulemask := MAX(CARDINAL);
    sInitCLI:="";
    testmode:=FALSE;
    useLFN := TRUE;
    state  := waiting;

    parmcount := Lib.ParamCount();

    (* IF parmcount = 0 THEN abort(errHelp);END; *)

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "W:"+delim+"WIDTH:"+delim+
                                  "B"+delim+"BW"+delim+
                                  "V"+delim+"VSYNC"+delim+
                                  "N"+delim+"DONOTFORCE"+delim+
                                  "R:"+delim+"RULE:"+delim+
                                  "I:"+delim+"INIT:"+delim+
                                  "T"+delim+"TEST"+delim+
                                  "T:"+delim+"TEST:"
                              );
            CASE opt OF
            | 1,2,3  : abort(errHelp);
            | 4,5    : ok:=GetLongCard(S,lc);
                       IF NOT(ok) THEN abort(errBadWidth);END;
                       initwi:=CARDINAL(lc);
                       IF ( ( initwi < (CAmin+1) ) OR (initwi > (CAmax+1)) ) THEN
                           abort(errWidthRange);
                       END;
            | 6,7    : cycle := FALSE;
            | 8,9    : vsync:=1;
            |10,11:    forceonelit:=FALSE;
            |12,13:    IF grabrulemask(rulemask,R)=FALSE THEN abort(errBadMask);END;
            |14,15:    IF grabinitial(sInitCLI,R)=FALSE THEN abort(errBadInitial);END;
            |16,17:    testmode:=TRUE;
            |18,19:    IF grabrulemask(rulemask,R)=FALSE THEN abort(errBadMask);END;
                       testmode:=TRUE;
            ELSE
                abort(errOption);
            END;
        ELSE
            CASE state OF
            | waiting:
                IF same(S,"?") THEN abort(errHelp);END;
                Str.Copy(rulefile,S);
            ELSE
                abort(errParm);
            END;
            INC(state);
        END;
    END;

    useLFN := (useLFN AND fileSupportLFN() );

    (*%T DBG *) dbg("//","NEW RUN !"); (*%E  *)

    Lib.RANDOMIZE;

    CASE state OF
    | waiting:
        (*%T DBG *) dbg("rulefile","none"); (*%E  *)
        sPat      := defaultPat;
        sNewState := defaultNewState;
        sInit     := defaultInit;
        IF rulemask = MAX(CARDINAL) THEN
            IF testmode THEN abort(errRuleRequired);END;
        ELSE
            byteToMask(sNewState,rulemask);
        END;
        IF same(sInitCLI,"")=FALSE THEN Str.Copy(sInit,sInitCLI);END;
    | gotfile:
        IF testmode THEN abort(errTestMode);END;
        IF rulemask # MAX(CARDINAL) THEN abort(errRuleOrFile);END;
        IF same(sInitCLI,"")=FALSE THEN abort(errInitialOrFile);END;

        (*%T DBG *) dbg("rulefile",rulefile); (*%E  *)
        IF chkJoker(rulefile) THEN abort(errJoker);END;
        Str.Copy(S,rulefile);
        IF Str.CharPos(rulefile,dot)=MAX(CARDINAL) THEN Str.Append(rulefile,extWR);END;
        IF fileExists(useLFN,rulefile)=FALSE THEN
            Str.Concat(rulefile,S,extTXT);
            IF fileExists(useLFN,rulefile)=FALSE THEN abort(errNotFound);END;
        END;
        i:=readRule(sPat,sNewState,sInit, useLFN,rulefile);
        IF i # errNone THEN abort(i);END;
    END;

    IF HiresON() = FALSE THEN ok:=HiresOFF();abort(errGraphMode); END;


    Graph.GetVideoConfig(V);
    xmin:=0; xmax:=V.numxpixels-1;
    ymin:=0; ymax:=V.numypixels-1;

    Graph.ClearScreen(Graph._GCLEARSCREEN);
    initCA(ndx,inkbk,  initwi,xmax,forceonelit,sPat,sNewState,sInit);
    ink:=white;
    ycurr:=ymin;

    LOOP
        ycurr:=showstate(ndx,xmax,ycurr,ink,inkbk,vsync);
        ndx:=processline(ndx,xmax);
        ink:=nextink(ink,ycurr,cycle);

        IF gotkey(ch) THEN
            CASE ch OF
            | space:
                LOOP
                    IF gotkey(ch) THEN EXIT; END;
                END;
            | "+", "-":
                IF testmode THEN
                    IF ch="+" THEN
                        k:=1;
                    ELSE
                        k:=-1;
                    END;
                    rulemask:=maskToByte(sNewState);
                    rulemask:=newvalue(k,rulemask,MIN(SHORTCARD), MAX(SHORTCARD));
                    byteToMask(sNewState,rulemask);

                    Graph.ClearScreen(Graph._GCLEARSCREEN);
                    initCA(ndx,inkbk,  initwi,xmax,forceonelit,sPat,sNewState,sInit);
                    ink:=white;
                    ycurr:=ymin;
                END;
            ELSE
                EXIT;
            END;
        END;
    END;

    IF HiresOFF()= FALSE THEN HALT; END;

    showparms(testmode,rulemask);

    abort(errNone);
END wolfRule.





(* // USED WHEN HACK WAS JUST A RULE 30 TESTER
        (* algo method *)
        vv:=XOR ( chToVal(l)  , ( chToVal(m) OR chToVal(r) ) );

        (* BFI method *)
        IF ( (m="0") AND (r="0") ) THEN
            v:=chToVal(l);
        ELSE
            IF l = "0" THEN
                v:=1;
            ELSE
                v:=0;
            END;
        END;

        IF v # vv THEN
            (*%T DBG *) dbgv("Mismatch between formulas",v); (*%E  *)
        END;
*)
