(* ---------------------------------------------------------------
Title         Q&D XOSL BootItem dump
Author        PhG
Overview      see help
Usage         see help
Notes         one of these days, check for BOOLEANs shown as INTEGERs
Bugs          yes, in XOSL !

              see "Fun with XOSL installer" !

Wish List

--------------------------------------------------------------- *)

MODULE XOSLdump;

IMPORT Lib;
IMPORT FIO;
IMPORT Str;
IMPORT IO;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "XOSLDUMP";
    ProgTitle     = "Q&D XOSL data dump";
    ProgVersion   = "v1.0b";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone         = 0;
    errHelp         = 1;
    errJoker        = 2;
    errNotFound     = 3;
    errFormat       = 4;
    errOption       = 5;
    errParmOverflow = 6;
CONST
    bootroot        = "C:\";
    itemsFile       = "BOOTITEM";
    dataFile        = "XOSLDATA";
    defaultItemsFile= bootroot+itemsFile;
    defaultDataFile = bootroot+dataFile;
    extXDF          = ".XDF";
    supportedVersion= "1.1.5";
CONST
    gremlin = CHR(254); (* was blank *)
    nl      = CHR(13)+CHR(10);

(* ------------------------------------------------------------ *)

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
    nl   = CHR(13)+CHR(10);
    help = Banner+nl+
nl+
"Syntax : "+ProgEXEname+" [ itemsFile["+extXDF+"] [ dataFile["+extXDF+"] ] ] [-z] [-v]"+nl+
nl+
'  -z truncate strings at CHR(0) (default is to show CHR(0) as "'+gremlin+'")'+nl+
"  -a dump all O.S. entries (defined or not)"+nl+
"  -x do not show user settings"+nl+
nl+
"This program dumps XOSL v"+supportedVersion+" infos about O.S. entries and user settings"+nl+
'(default files are "'+defaultItemsFile+'" and "'+defaultDataFile+'").'+nl;

VAR
    S : str128;
BEGIN
    CASE e OF
    | errHelp :     WrStr(help);
    | errJoker:     S:="Jokers are not supported !";
    | errNotFound : Str.Concat(S,einfo," does not exist !");
    | errFormat:    Str.Concat(S,einfo," does not have expected size !");
    | errOption:    Str.Concat(S,"Unknown ",einfo);Str.Append(S," option !");
    | errParmOverflow: S := "Too many parameters !";
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone,errHelp:
        ;
    ELSE
        WrStr(ProgEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
xos\xosl\items.h

class CBootItem {
	char ItemName[48];
	unsigned char PartIndex;
	unsigned short Hotkey;
	char Activate;
	char FixDriveNum;
	char Disabled;
		unsigned long Password;
	unsigned short Keys[16];
	unsigned char HideList[56];
	char SwapDrives;
	char Reserved[11]; // reserved for future use
    }; // sizeof(CBootItem) == 158

class CPartDesc {
	unsigned char Drive;
	unsigned long StartSector;
    }; // sizeof(CPartDesc) == 5

class CBootItemFile {
	CBootItem BootItems[24];
	CPartDesc PartList[56];
	unsigned char BootItemCount;
	unsigned char DefaultItem;
	unsigned char Reserved0; // was timeout
	unsigned char PartCount;
	unsigned short Timeout;
	char Reserved[18]; // reserved for future use
    }; //  sizeof(CBootItemFile) == 4096
*)

CONST
    countBootItem = 24;
    countPartDesc = 56;
TYPE
    CBootItem = RECORD (* 158 *)
	    ItemName    : ARRAY [0..47] OF CHAR;
	    PartIndex   : SHORTCARD;
	    Hotkey      : CARDINAL;
	    Activate    : SHORTCARD;
	    FixDriveNum : SHORTCARD;
	    Disabled    : SHORTCARD;
		Password    : LONGCARD;
	    Keys        : ARRAY [1..16] OF CARDINAL;
	    HideList    : ARRAY [1..56] OF SHORTCARD;
	    SwapDrives  : SHORTCARD;
	    Reserved    : ARRAY [1..11] OF SHORTCARD;
	END;
    CPartDesc = RECORD (* 5 *)
	    Drive       : SHORTCARD;
	    StartSector : LONGCARD;
    END;
    CBootItemFile = RECORD (* 4096 *)
        BootItems    : ARRAY [1..countBootItem] OF CBootItem;
	    PartList     : ARRAY [1..countPartDesc] OF CPartDesc;
	    BootItemCount: SHORTCARD;
	    DefaultItem  : SHORTCARD;
	    Reserved0    : SHORTCARD;
	    PartCount    : SHORTCARD;
	    Timeout      : CARDINAL;
	    Reserved     : ARRAY [1..18] OF SHORTCARD;
    END;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(*
xosl\include\graph.h

typedef enum {
	modeText = 0, mode640x480, mode800x600,
	mode1024x768, mode1280x1024, mode1600x1200
} TGraphMode;

*)

CONST
     modeText       = 0; sModeText      = "text";
     mode640x480    = 1; sMode640       = "640x480";
     mode800x600    = 2; sMode800       = "800x600";
     mode1024x768   = 3; sMode1024      = "1024x768";
     mode1280x1024  = 4; sMode1280      = "1280x1024";
     mode1600x1200  = 5; sMode16000x1200= "1600x1200";

(*
xosl\xosl\xosldata.h
install\install\xosldata.h seems an older version !!! //FIXME WARNING !!!

typedef struct {
        TGraphMode VideoMode;
        int FrameMove;
        int LinearFB;
        int Font9;
        int CursorInverted;
        int UseWallpaper;
        int DisableFadeOut;
        int DisableFadeIn;
        int DisplayItemIndex;
        int ClearScreen;
        int NoAnimation; //not in install
        int Reserved[5]; //[6] in install
} TGraphData;

typedef struct {
        int Reserved1; // can be used again
        int ColorScheme;
        int Hue;
        int Saturation;
        int Luminance;
        int FadeColorInt;
        int Reserved2[10];
} TColorData;

typedef struct {
        int MouseType;
        int MouseSpeed;
        int PS2Sampling;
        int KbdLayout;         //not in install
        int EnhancedSupport;   //not in install
        int Reserved[11];      //[13] in install
} TMouseData;

typedef struct {
	unsigned long Password;
	int Reserved[14];
} TPassword;

typedef struct {
	int AutoSave;
	int Reserved1;	// was 'BootOnSelect'. Not used anymore
	int KeyColdReboot;
	int KeyShutdown;
	int KeyRestart;
	int KeyPart;
	int SetAutoBoot;
	int SoftPowerOff;
	int KeyCycle;
	int ActiveAllow;
	int AntiVirus;
    int NoAutoBootDialog; //not in install
    int Reserved[4];      //[5] in install
} TMiscPref;

typedef struct {
    int XOSLVersion;      //"unsigned short" in install
	TGraphData GraphData;
	TColorData ColorData;
	TMouseData MouseData;
	TPassword PasswordData;
	TMiscPref MiscPref;
	int Reserved[175];
} TXOSLData;

*)

TYPE
	TGraphData = RECORD
		VideoMode         : INTEGER; (* TGraphicsMode; *)
		FrameMove         : INTEGER;
		LinearFB          : INTEGER;
		Font9             : INTEGER;
		CursorInverted    : INTEGER;
		UseWallpaper      : INTEGER;
		DisableFadeOut    : INTEGER;
		DisableFadeIn     : INTEGER;
		DisplayItemIndex  : INTEGER;
		ClearScreen       : INTEGER;
		NoAnimation       : INTEGER;
		Reserved          : ARRAY[1..5] OF INTEGER;
    END;
	TColorData = RECORD
		Reserved1         : INTEGER;
		ColorScheme       : INTEGER;
		Hue               : INTEGER;
		Saturation        : INTEGER;
		Luminance         : INTEGER;
		FadeColorInt      : INTEGER;
		Reserved2         : ARRAY [1..10] OF INTEGER;
    END;
	TMouseData = RECORD
		MouseType         : INTEGER;
		MouseSpeed        : INTEGER;
		PS2Sampling       : INTEGER;
		KbdLayout         : INTEGER;
		EnhancedSupport   : INTEGER;
		Reserved          : ARRAY [1..11] OF INTEGER;
	END;
	TPassword = RECORD
		Password          : LONGCARD;
		Reserved          : ARRAY [1..14] OF INTEGER;
	END;
	TMiscPref = RECORD
		AutoSave          : INTEGER;
		Reserved1         : INTEGER;
		KeyColdReboot     : INTEGER;
		KeyShutdown       : INTEGER;
		KeyRestart        : INTEGER;
		KeyPart           : INTEGER;
		SetAutoBoot       : INTEGER;
		SoftPowerOff      : INTEGER;
		KeyCycle          : INTEGER;
		ActiveAllow       : INTEGER;
		AntiVirus         : INTEGER;
        NoAutoBootDialog  : INTEGER;
		Reserved          : ARRAY [1..4] OF INTEGER;
    END;
TYPE
    TXOSLDataFile = RECORD
        Xoslversion   : INTEGER;
        graphdata     : TGraphData;
        colordata     : TColorData;
        mousedata     : TMouseData;
        passworddata  : TPassword;
        miscpref      : TMiscPref;
        reserved      : ARRAY [1..175] OF INTEGER;
    END;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE num2str (wi,base:CARDINAL; v : LONGCARD):str16;
VAR
    ok:BOOLEAN;
    i:CARDINAL;
    R:str16;
    pad:CHAR;
BEGIN
    IF base = 16 THEN
        pad:="0";
    ELSE
        pad:=" ";
    END;
    Str.CardToStr (v,R,base,ok);
    FOR i:=Str.Length(R)+1 TO wi DO
        Str.Prepend(R,pad);
    END;
    IF base = 16 THEN Str.Lows(R); Str.Prepend(R,"$");END;
    RETURN R;
END num2str;

PROCEDURE dmppad (S:ARRAY OF CHAR);
CONST
    wi = 16;
VAR
    i:CARDINAL;
BEGIN
    WrStr(S);
    FOR i:=Str.Length(S)+1 TO wi DO
        WrStr(" ");
    END;
    WrStr(" : ");
END dmppad;

PROCEDURE dmp (S:ARRAY OF CHAR; wi,base:CARDINAL; v:LONGCARD);
VAR
    R:str16;
BEGIN
    dmppad(S);
    R:=num2str(wi,base,v);
    WrStr(R);WrLn;
END dmp;

PROCEDURE dmpC (S:ARRAY OF CHAR;wi,base:CARDINAL; v : CARDINAL);
BEGIN
    dmp(S,wi,base, LONGCARD(v));
END dmpC;

PROCEDURE dmpI (S:ARRAY OF CHAR;wi,base:CARDINAL; v : INTEGER);
BEGIN
    dmp(S,wi,base, LONGCARD(v));
END dmpI;

PROCEDURE dmpb (S:ARRAY OF CHAR;wi,base:CARDINAL; v : SHORTCARD);
BEGIN
    dmp(S,wi,base, LONGCARD(v));
END dmpb;

PROCEDURE dmpbuf (S:ARRAY OF CHAR;wi,base:CARDINAL; buf:ARRAY OF SHORTCARD);
VAR
    i,last:CARDINAL;
BEGIN
    dmppad(S);
    last:=SIZE(buf)-1;
    FOR i := 0 TO last DO
        WrStr( num2str(wi,base,LONGCARD(buf[i]) ) );
        IF i < last THEN WrStr(" ");END;
    END;
    WrLn;
END dmpbuf;

PROCEDURE dmpbufcard (S:ARRAY OF CHAR; buf:ARRAY OF CARDINAL);
VAR
    i,last:CARDINAL;
BEGIN
    dmppad(S);
    last:=( SIZE(buf) DIV 2 )-1;
    FOR i := 0 TO last DO
        WrStr( num2str(4,16,LONGCARD(buf[i]) ) );
        IF i < last THEN WrStr(" ");END;
    END;
    WrLn;
END dmpbufcard;

PROCEDURE dmpstr (truncate:BOOLEAN; S:ARRAY OF CHAR; buf:ARRAY OF CHAR);
VAR
    i,last:CARDINAL;
BEGIN
    dmppad(S);
    WrStr('"');
    last:=SIZE(buf)-1;
    (* FOR i := 0 TO last DO *)
    i:=0;
    LOOP
        IF i > last THEN EXIT; END;
        IF buf[i] = 0C THEN
            IF truncate THEN EXIT; END;
            WrStr(gremlin);
        ELSE
            WrStr(buf[i]);
        END;
        INC(i);
    END;
    WrStr('"') ; WrLn;
END dmpstr;

PROCEDURE dmphexstr (S:ARRAY OF CHAR; buf:ARRAY OF CHAR);
VAR
    i,last:CARDINAL;
BEGIN
    dmppad(S);
    last:=SIZE(buf)-1;
    FOR i := 0 TO last DO
        WrStr( num2str(2,16,LONGCARD(buf[i]) ) );
        IF i < last THEN WrStr(" ");END;
    END;
    WrLn;
END dmphexstr;

PROCEDURE dmpItem (DEBUG,truncate:BOOLEAN; i:CARDINAL; d : CBootItem; H:str16  );
VAR
    S:str128;
BEGIN
    Str.Subst(H, "~" , num2str(2,10,LONGCARD(i)) );

    Str.Concat(S,H,"ItemName");
    dmpstr(truncate,S,d.ItemName);

    IF DEBUG THEN
        Str.Concat(S,H,"ItemNAME");
        dmphexstr(S,d.ItemName);
    END;

    Str.Concat(S,H,"PartIndex"); (* //WARNING ! 8-ung ! we must add 1 for 1-based partdescs *)
    dmpb(S,1,10,d.PartIndex+1);

    Str.Concat(S,H,"Hotkey");
    dmpC(S,4,16,d.Hotkey);

    Str.Concat(S,H,"Activate");
    dmpb(S,2,16,d.Activate);

    Str.Concat(S,H,"FixDriveNum");
    dmpb(S,2,16,d.FixDriveNum);

    Str.Concat(S,H,"Disabled");
    dmpb(S,2,16,d.Disabled);

    Str.Concat(S,H,"Password");
    dmp(S,8,16,d.Password);

    Str.Concat(S,H,"Keys");
    dmpbufcard(S,d.Keys);

    Str.Concat(S,H,"HideList");
    dmpbuf(S,1,10,d.HideList);

    Str.Concat(S,H,"SwapDrives");
    dmpb(S,2,16,d.SwapDrives);

    Str.Concat(S,H,"Reserved");
    dmpbuf(S,2,16,d.Reserved);

END dmpItem;

PROCEDURE dmpDesc (i:CARDINAL; d : CPartDesc; H:str16);
VAR
    S:str128;
BEGIN
    Str.Subst(H, "~" , num2str(2,10,LONGCARD(i)) );

    Str.Concat(S,H,"Drive");
    dmpb(S,2,16,d.Drive);

    Str.Concat(S,H,"StartSector");
    dmp(S,1,10,d.StartSector);
END dmpDesc;

PROCEDURE dmpVmode ( S:ARRAY OF CHAR;v:CARDINAL);
VAR
    Z:str80;
BEGIN
    dmppad(S);
    CASE v OF
    | modeText       : Z:= sModeText;
    | mode640x480    : Z:= sMode640;
    | mode800x600    : Z:= sMode800;
    | mode1024x768   : Z:= sMode1024;
    | mode1280x1024  : Z:= sMode1280;
    | mode1600x1200  : Z:= sMode16000x1200;
    ELSE
                       Z:="unexpected ~ video mode !";
                       Str.Subst(Z,"~",num2str (2,16,LONGCARD(v) ) );
    END;
    WrStr(Z);WrLn;
END dmpVmode;

PROCEDURE buildheaderstr (VAR H:str16; i:CARDINAL;maxi:SHORTCARD);
BEGIN
    IF i > CARDINAL(maxi) THEN
        Str.Copy(H,"(~)");
    ELSE
        Str.Copy(H,"[~]");
    END;
    Str.Append(H," ");
END buildheaderstr;

PROCEDURE dmpbool (S:ARRAY OF CHAR;DEBUG:BOOLEAN;ib:INTEGER);
VAR
    TF,Z:str16;
BEGIN
    dmppad(S);
    IF ib = 0 THEN
        TF:="no";
    ELSE
        TF:="YES";
    END;
    IF DEBUG THEN
        Z:="(~) ";
        Str.Subst(Z,"~",num2str(2,16,LONGCARD(ib)) );
        WrStr(Z);
    END;
    WrStr(TF);WrLn;
END dmpbool;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ioBufferSize  = (8 * 512) + FIO.BufferOverhead;
    firstIObuffer = 1;
    lastIObuffer  = ioBufferSize;
TYPE
    ioBufferType  = ARRAY[firstIObuffer..lastIObuffer] OF BYTE;
VAR
    ioBuffer    : ioBufferType;

CONST
    minparm = 1;
    maxparm = 2;
VAR
    parm : ARRAY [minparm..maxparm] OF str128;
    lastparm : CARDINAL;
VAR
    hnd         : FIO.File;
    last,opt,i,wanted,got,parmcount: CARDINAL;
    S,R,Fitems,Fdata  : str128;
    itemdata    : CBootItemFile;
    userdata    : TXOSLDataFile;
    fsize       : LONGCARD;
    items,parts : CARDINAL;
    truncate,showallitems,showusersettings,DEBUG : BOOLEAN;
    H:str16;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    WrLn;

    truncate         := FALSE;
    showallitems     := FALSE;
    showusersettings := TRUE;
    DEBUG            := FALSE;

    lastparm := minparm-1;
    parmcount := Lib.ParamCount();
    FOR i:=1 TO parmcount DO
        Lib.ParamStr(S,i);
        UpperCase(S); cleantabs(S);
        IF isOption(S) THEN
            opt := GetOptIndex(S, "?"+delim+"H"+delim+"HELP"+delim+
                                  "DEBUG"+delim+
                                  "A"+delim+"ALLENTRIES"+delim+"ALL"+delim+
                                  "Z"+delim+"TRUNCATE"+delim+
                                  "X"+delim+"SHOWUSERSETTINGS"+delim+"USER"
                                  );
            CASE opt OF
            | 1,2,3    : abort(errHelp,"");
            | 4        : DEBUG            := TRUE;
            | 5,6,7    : showallitems     := TRUE;
            | 8,9      : truncate         := TRUE;
            | 10,11,12 : showusersettings := FALSE;
            ELSE
                         abort(errOption,S);
            END;
        ELSE
           INC(lastparm);
           IF lastparm > maxparm THEN abort(errParmOverflow,"");END;
           Str.Copy(parm[lastparm],S);
        END;
    END;
    CASE lastparm OF
    | minparm-1:
        Str.Copy(Fitems,defaultItemsFile);
        Str.Copy(Fdata,defaultDataFile);
    | minparm:
        Str.Copy(Fitems,parm[minparm]);
        showusersettings := FALSE;
    | maxparm:
        Str.Copy(Fitems,parm[minparm]);
        Str.Copy(Fdata,parm[minparm+1]);
    END;
    IF showusersettings THEN
        last := 2;
    ELSE
        last := 1;
    END;

    FOR i:=1 TO last DO
        CASE i OF
        | 1: Str.Copy(S,Fitems); wanted:=SIZE(itemdata);
        | 2: Str.Copy(S,Fdata);  wanted:=SIZE(userdata);
        END;

        IF Str.CharPos(S,".") = MAX(CARDINAL) THEN Str.Append(S,extXDF);END;
        IF Str.CharPos(S,"?") # MAX(CARDINAL) THEN abort(errJoker,S);END;
        IF Str.CharPos(S,"*") # MAX(CARDINAL) THEN abort(errJoker,S);END;
        IF FIO.Exists(S)=FALSE THEN abort(errNotFound,S); END;

        hnd:=FIO.OpenRead(S);
        FIO.AssignBuffer(hnd,ioBuffer);
        fsize:=FIO.Size(hnd);
        CASE i OF
        | 1: got:=FIO.RdBin(hnd,itemdata,wanted);
        | 2: got:=FIO.RdBin(hnd,userdata,wanted);
        END;
        FIO.Close(hnd);
        IF fsize # LONGCARD(wanted) THEN abort(errFormat,S);END;

        CASE i OF
        | 1: Str.Copy(Fitems,S);
        | 2: Str.Copy(Fdata,S);
        END;
    END;

    IF showallitems THEN
        items:= countBootItem;
        parts:= countPartDesc;
    ELSE
        items:= CARDINAL( itemdata.BootItemCount );
        parts:= CARDINAL( itemdata.PartCount );
    END;

    S:='::: XOSL O.S. entries found in "~"'+nl+nl;
    Str.Subst(S,"~",Fitems);
    WrStr(S);

    FOR i:=1 TO items DO
        buildheaderstr(H, i, itemdata.BootItemCount);
        dmpItem(DEBUG,truncate, i,itemdata.BootItems[i], H);
        WrLn;
    END;
    FOR i:=1 TO parts DO
        buildheaderstr(H, i, itemdata.PartCount);
        dmpDesc(i,itemdata.PartList[i], H);
        WrLn;
    END;

	dmpb   ("BootItemCount" ,1,10,itemdata.BootItemCount);
	dmpb   ("DefaultItem"   ,2,16,itemdata.DefaultItem);
	IF DEBUG THEN dmpb   ("Reserved0"     ,2,16,itemdata.Reserved0); END;
	dmpb   ("PartCount"     ,1,10,itemdata.PartCount);
	dmpC   ("Timeout"       ,1,10,itemdata.Timeout);
	IF DEBUG THEN dmpbuf ("Reserved"      ,2,16,itemdata.Reserved); END;

  IF showusersettings THEN

    S:=nl+'::: XOSL user settings found in "~"'+nl+nl;
    Str.Subst(S,"~",Fdata);
    WrStr(S);

    dmpC   ("XoslVersion"   ,4,16,userdata.Xoslversion);
    WrLn;
	dmpVmode("VideoMode",userdata.graphdata.VideoMode);
    dmpbool("FrameMove"         ,DEBUG,userdata.graphdata.FrameMove);
	dmpbool("LinearFB"          ,DEBUG,userdata.graphdata.LinearFB);
	dmpbool("Font9"             ,DEBUG,userdata.graphdata.Font9);
	dmpbool("CursorInverted"    ,DEBUG,userdata.graphdata.CursorInverted);
	dmpbool("UseWallpaper"      ,DEBUG,userdata.graphdata.UseWallpaper);
	dmpbool("DisableFadeOut"    ,DEBUG,userdata.graphdata.DisableFadeOut);
	dmpbool("DisableFadeIn"     ,DEBUG,userdata.graphdata.DisableFadeIn);
	dmpbool("DisplayItemIndex"  ,DEBUG,userdata.graphdata.DisplayItemIndex);
	dmpbool("ClearScreen"       ,DEBUG,userdata.graphdata.ClearScreen);
	dmpbool("NoAnimation"       ,DEBUG,userdata.graphdata.NoAnimation);
    WrLn;
    dmpI("ColorScheme"          ,1,10,userdata.colordata.ColorScheme);
    dmp ("Hue"                  ,8,16, LONGCARD( userdata.colordata.Hue) );
    dmpI("Saturation"           ,1,10,userdata.colordata.Saturation);
    dmpI("Luminance"            ,1,10,userdata.colordata.Luminance);
    dmpI("FadeColorInt"         ,1,10,userdata.colordata.FadeColorInt);
    WrLn;
    dmpI("MouseType"            ,1,10,userdata.mousedata.MouseType);
    dmpI("MouseSpeed"           ,1,10,userdata.mousedata.MouseSpeed);
    dmpI("PS2Sampling"          ,1,10,userdata.mousedata.PS2Sampling);
    dmpI("KbdLayout"            ,1,10,userdata.mousedata.KbdLayout);
    dmpI("EnhancedSupport"      ,1,10,userdata.mousedata.EnhancedSupport); (* // bool ? *)
    WrLn;
    dmp("Password",8,16,userdata.passworddata.Password);
    WrLn;
	dmpbool("AutoSave"          ,DEBUG,userdata.miscpref.AutoSave);
    dmpI("KeyColdReboot"        ,4,16,userdata.miscpref.KeyColdReboot );
    dmpI("KeyShutdown"          ,4,16,userdata.miscpref.KeyShutdown);
    dmpI("KeyRestart"           ,4,16,userdata.miscpref.KeyRestart);
    dmpI("KeyPart"              ,4,16,userdata.miscpref.KeyPart);
    dmpbool("SetAutoBoot"       ,DEBUG,userdata.miscpref.SetAutoBoot);
    dmpbool("SoftPowerOff"      ,DEBUG,userdata.miscpref.SoftPowerOff);
    dmpI("KeyCycle"             ,4,16,userdata.miscpref. KeyCycle);
    dmpbool("ActiveAllow"       ,DEBUG,userdata.miscpref.ActiveAllow);
    dmpbool("AntiVirus"         ,DEBUG,userdata.miscpref.AntiVirus);
    dmpbool("NoAutoBootDialog"  ,DEBUG,userdata.miscpref.NoAutoBootDialog);
  END;

    abort(errNone,"");
END XOSLdump.
