(* ---------------------------------------------------------------
Title         Q&D Roman Numerals Converter
Overview
Notes         now, everyone will be able to check MGM's movies dates ;-)
              stand alone (no QD_* code) version of ROMAN.MOD
Bugs
Wish List
--------------------------------------------------------------- *)

MODULE RomanTS;

IMPORT Str;
IMPORT Lib;

FROM IO IMPORT WrStr,WrLn;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    cr            = CHR(13);
    lf            = CHR(10);
    nl            = cr+lf;
    digits        = "0123456789";
    nullchar      = 0C; (* CHR(0) *)
    charnull      = nullchar;
    space         = " ";
    slash         = "/";
    dash          = "-";
    delim         = ",";
    equal         = "=";
    colon         = ":";
    tabul         = CHR(9);

TYPE
    str256 = ARRAY [0..255] OF CHAR;
    str128 = ARRAY [0..127] OF CHAR;
    str16  = ARRAY [0..15] OF CHAR;
    str80  = ARRAY [0..79] OF CHAR;

PROCEDURE Belongs (S : ARRAY OF CHAR; c:CHAR  ) : BOOLEAN;
BEGIN
    RETURN ( Str.CharPos(S,c) # MAX(CARDINAL) );
END Belongs;

PROCEDURE verifyString (S,allowed:ARRAY OF CHAR):BOOLEAN;
VAR
    i,mismatches : CARDINAL;
    ch : CHAR;
BEGIN
    mismatches := 0;
    FOR i := 1 TO Str.Length(S) DO
        ch := S[i-1];
        IF Belongs(allowed,ch)=FALSE THEN INC(mismatches);END;
    END;
    RETURN (mismatches=0);
END verifyString;

PROCEDURE same(S1,S2 : ARRAY OF CHAR) : BOOLEAN;
BEGIN
    RETURN ( Str.Compare(S1,S2)=0 );
END same;

(* try and fix TopSpeed bug when tabs are in command line *)
(* try and fix our NEVER using tabs in any command line too ! *)

PROCEDURE cleantabs (VAR S : ARRAY OF CHAR);
VAR
    p : CARDINAL;
BEGIN
    LOOP
        p:=Str.CharPos(S,tabul);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        (* S[p]:=Str.Delete(S,p,1); (* delete unwanted tabulation *) *)
        S[p]:=space;
    END;
END cleantabs;

PROCEDURE map (c : CHAR ;table : ARRAY OF CHAR) : CHAR;
BEGIN
    RETURN table [ ORD (c) ];
END map;

CONST
    ignore = " ";
    Tupper = "                                "+
             "                                "+
             " ABCDEFGHIJKLMNOPQRSTUVWXYZ     "+
             " ABCDEFGHIJKLMNOPQRSTUVWXYZ     "+
             "CUEAAAACEEEIIIAAEOOOUUYOU     "+ (* use "ae" *)
             "AIOUNN                          "+
             "                                "+
             "                               ";

PROCEDURE UpperCase (VAR S : ARRAY OF CHAR);
VAR
    len : CARDINAL;
    i   : CARDINAL;
    c   : CHAR;
    ch  : CHAR;
BEGIN
    len := Str.Length(S);
    IF len = 0 THEN RETURN; END;         (* this WAS a problem ! *)
    FOR i := 0 TO (len - 1) DO
        c  := S[i];
        ch := map(c,Tupper);
        IF ch # ignore THEN S[i] := ch; END;
    END;
END UpperCase;

PROCEDURE isOption (S : ARRAY OF CHAR) : BOOLEAN;
BEGIN
CASE S[0] OF
| slash, dash :
    RETURN TRUE;
ELSE
    RETURN FALSE;
END;
END isOption;

PROCEDURE ReplaceChar (VAR S : ARRAY OF CHAR;
                       oldchar, newchar : CHAR);
VAR
    p : CARDINAL;
BEGIN
    IF oldchar=newchar THEN RETURN; END; (* silly... *)
    LOOP
        p := Str.CharPos(S,oldchar);
        IF p = MAX(CARDINAL) THEN EXIT; END;
        IF newchar = "" THEN
            Str.Delete(S,p,1);
        ELSE
            S[p]:=newchar;
        END;
    END;
END ReplaceChar;

PROCEDURE GetOptIndex (S : ARRAY OF CHAR; options : ARRAY OF CHAR) : CARDINAL;
VAR
    i   : CARDINAL;
    str : str80; (* option tested against s *)
BEGIN
    Str.Delete(S,0,1); (* remove leading - or / because we arrive if option *)
    Str.Append(options,delim);

    Str.Subst(S,equal,colon); (* command line option xxx= becomes xxx: *)
    ReplaceChar(options,equal,colon); (* all possibles option xxx= become xxx: *)

    i := 0;
    LOOP
        Str.ItemS(str,options,delim,i);
        IF Str.Length(str)=0 THEN RETURN 0; END;
        IF Str.CharPos(str,colon)=MAX(CARDINAL) THEN (* normal option *)
             IF Str.Compare(str,S)=0 THEN RETURN (i+1); END;
        ELSE (* option with a number *)
             Str.Append(str,"*"); (* match any number after : *)
             IF Str.Match(S, str )=TRUE THEN RETURN (i+1); END;
        END;
        INC(i);
    END;
END GetOptIndex;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    ProgEXEname   = "ROMAN";
    ProgTitle     = "Q&D Roman Numerals Converter";
    ProgVersion   = "v1.0";
    ProgCopyright = "by PhG";
    Banner        = ProgTitle+" "+ProgVersion+" "+ProgCopyright;

CONST
    errNone         = 0;
    errHelp         = 1;
    errOption       = 2;
    errParmOverflow = 3;
    errBadNum       = 4;
    errRomanRange   = 5;
    errBadRomanNum  = 6;

PROCEDURE abort (e : CARDINAL; einfo : ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)
    helpmsg =
Banner+nl+
nl+
"Syntax 1 : "+ProgEXEname+" <number>... [option]..."+nl+
"Syntax 2 : "+ProgEXEname+" <first..last> [option]..."+nl+
nl+
"With syntax 1, this program converts a number from decimal to roman numerals"+nl+
"and from roman numerals to decimal. With syntax 2, it converts a range,"+nl+
"provided it is a valid one."+nl+
nl+
"   -l show output roman as lowercase roman (default is uppercase)"+nl+
"   -u show input roman as uppercase (default is lowercase)"+nl+
nl+
"a) Note valid range for roman numerals is [1..3999]."+nl+
nl+
"Examples : "+ProgEXEname+" -l 1963 3888"+nl+
"           "+ProgEXEname+" -u mcmlxiii mmmdccclxxxviii"+nl+
"           "+ProgEXEname+" 1..3999"+nl;

VAR
    S : str256;
BEGIN
    CASE e OF
    | errHelp :
        WrStr(helpmsg);
    | errOption :
        S:='Illegal "~" option !';
    | errParmOverflow:
        S:='"~" is one parameter too many !';
    | errBadNum:
        S:='Illegal "~" number !';
    | errRomanRange:
        S:='"~" does not belong to valid roman numerals range !';
    | errBadRomanNum:
        S:='"~" does not look like a valid roman number !';
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp : ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(ProgEXEname+" : ");WrStr(S);WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    strERROR = "??? ERROR ???";

PROCEDURE padstr (VAR R:ARRAY OF CHAR;wi:INTEGER;padchar:CHAR   );
VAR
    i:CARDINAL;
BEGIN
    FOR i:=Str.Length(R)+1 TO ABS(wi) DO
        IF wi < 0 THEN
            Str.Append(R,padchar);
        ELSE
            Str.Prepend(R,padchar);
        END;
    END;
END padstr;

PROCEDURE bignum2str (v:LONGCARD;base:CARDINAL;wi:INTEGER;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
BEGIN
    Str.CardToStr( v,R,base,ok);
    IF ok THEN
        padstr(R,wi,padchar);
    ELSE
        R:=strERROR;
    END;
    RETURN R;
END bignum2str;

PROCEDURE num2str (v:CARDINAL;base:CARDINAL;wi:INTEGER;padchar:CHAR):str16;
BEGIN
    RETURN bignum2str( LONGCARD(v),base,wi,padchar);
END num2str;

PROCEDURE isValidDecimalNumber (VAR v:LONGCARD; S:ARRAY OF CHAR ):BOOLEAN;
CONST
    decdigits = digits;
VAR
    rc,ok:BOOLEAN;
BEGIN
    rc:=FALSE;
    IF verifyString(S,decdigits) THEN
        v:=Str.StrToCard( S, 10, ok);
        rc:=ok;
    END;
    RETURN rc;
END isValidDecimalNumber;

PROCEDURE isValidRomanNumber (S:ARRAY OF CHAR   ):BOOLEAN ;
CONST
    romandigits = "IVXLCDMivxlcdm"; (* how clever ! :-) *)
BEGIN
    RETURN verifyString(S,romandigits);
END isValidRomanNumber;

PROCEDURE isValidRange (VAR mini,maxi:LONGCARD;  S:ARRAY OF CHAR  ):BOOLEAN;
CONST
    seprange = "..";
VAR
    rc:BOOLEAN;
    p:CARDINAL;
    N:str128; (* oversized for safety *)
BEGIN
    rc:=FALSE;
    p:=Str.Pos(S,seprange);
    IF p # MAX(CARDINAL) THEN
        Str.Copy(N,S);
        N[p]:=charnull;
        IF isValidDecimalNumber(mini,N) THEN
            Str.Copy(N,S);
            Str.Delete(N,0,p+2); (* delete ".." too *)
            rc:=isValidDecimalNumber(maxi,N);
        END;
    END;
    RETURN rc;
END isValidRange;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minRomanNumber = 1;
    maxRomanNumber = 4000-1;

PROCEDURE isInRomanRange (v:LONGCARD):BOOLEAN;
BEGIN
    RETURN ( (v >= minRomanNumber) AND (v <= maxRomanNumber) );
END isInRomanRange;

(* adapted from my old 6502 PhilPrinter code, 1981 *)

CONST
    maxromanwidth    = 15; (*  *)
    maxdecromanwidth = 4;  (* 3999 *)

(* yes, we know there are a few inscriptions with faulty "iiii" instead of "iv" *)

PROCEDURE dec2rom (v:LONGCARD;wi:CARDINAL;flagLOWER:BOOLEAN):str16;
CONST
    numeral = "MCXI";
    special = " DLV";
VAR
    N,R:str16;
    i,j,k:CARDINAL;
BEGIN
    IF isInRomanRange(v) THEN
        Str.Copy(N,bignum2str(v,10,4,"0"));
        R:="";
        FOR i:=(1-1) TO (maxdecromanwidth-1) DO
            k:=ORD( N[i] ) - ORD( "0" );
            CASE k OF
            | 0:     ;
            | 1,2,3: FOR j:=1 TO k DO Str.Append(R,numeral[i]); END;
            | 4:     Str.Append(R,numeral[i]);
                     Str.Append(R,special[i]);
            | 5:     Str.Append(R,special[i]);
            | 6,7,8: Str.Append(R,special[i]);
                     FOR j:=1 TO (k-5) DO Str.Append(R,numeral[i]); END;
            | 9:     Str.Append(R,numeral[i]);
                     Str.Append(R,numeral[i-1]);
            END;
        END;
        padstr(R,wi,space);
        IF flagLOWER THEN Str.Lows(R);END;
    ELSE
        R:=strERROR;
    END;
    RETURN R;
END dec2rom;

(* yes, we could write a more clever sub but this ugly code WORKS ! :-) *)

PROCEDURE rom2dec (S:ARRAY OF CHAR ):LONGCARD;
VAR
    N:str16;
    i:CARDINAL;
BEGIN
    Str.Lows(S);
    FOR i := minRomanNumber TO maxRomanNumber DO
        Str.Copy(N,dec2rom( LONGCARD(i),1,TRUE));
        IF same(S,N) THEN RETURN LONGCARD(i);END;
    END;
    RETURN MAX(LONGCARD);
END rom2dec;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    minparm = 1;
    maxparm = 32;
VAR
    parm : ARRAY [minparm..maxparm] OF str128;
    lastparm,parmcount,i,opt:CARDINAL;
    S,R:str128;
    flagLOWERdst,flagLOWERsrc:BOOLEAN;
    mode : (convert,showrange);
    v,mini,maxi:LONGCARD;
    okdec,okrom:BOOLEAN;
    pass:CARDINAL;
    currval:LONGCARD;
    maxlen : CARDINAL;
BEGIN
    WrLn;
    parmcount := Lib.ParamCount();
    (* IF parmcount=0 THEN abort(errHelp,"");END; *)

    flagLOWERdst := FALSE;
    flagLOWERsrc := TRUE;

    lastparm  := minparm-1;
    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex (R, "?"+delim+"H"+delim+"HELP"+delim+
                                   "L"+delim+"LOWER"+delim+"LOWERCASE"+delim+"LOWS"+delim+
                                   "U"+delim+"UPPER"+delim+"UPPERCASE"+delim+"CAPS"
                               );
            CASE opt OF
            | 1,2,3   :    abort(errHelp,"");
            | 4,5,6,7 :    flagLOWERdst := TRUE;
            | 8,9,10,11:   flagLOWERsrc := FALSE;
            ELSE           abort(errOption,S); (* could be errHelp, eh eh ! *)
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errParmOverflow,S);END;
            Str.Copy(parm[lastparm],S); (* keep case *)
        END;
    END;

    mode:=convert;
    CASE lastparm OF
    | minparm-1 :
        abort(errHelp,"");
    | minparm :
        S:=parm[minparm];
        IF isValidRange(mini,maxi,S) THEN
           mode:=showrange;
        END;
    ELSE
        ;
    END;
    CASE mode OF
    | convert:
        FOR pass:=1 TO 2 DO
            FOR i:=minparm TO lastparm DO
                S:=parm[i];
                CASE pass OF
                | 1:
                    okdec:=isValidDecimalNumber(v,S);
                    IF okdec THEN
                        IF isInRomanRange(v)=FALSE THEN abort(errRomanRange,S);END;
                    END;
                    okrom:=isValidRomanNumber(S);
                    IF okrom THEN
                        IF rom2dec(S)=MAX(LONGCARD) THEN abort(errBadRomanNum,S);END;
                    END;
                    IF ( (okdec = FALSE) AND (okrom = FALSE) ) THEN abort(errBadNum,S);END;
                | 2:
                    (* both can't be true at the same time, eh eh *)
                    IF isValidDecimalNumber(v,S) THEN
                        okdec:=isValidDecimalNumber(v,S);
                        S:="~  ==>  ~";
                        Str.Subst(S,"~",bignum2str(v,10,maxdecromanwidth,space) );
                        Str.Subst(S,"~",dec2rom(v,1,flagLOWERdst) );
                        WrStr(S);WrLn;
                    END;
                    IF isValidRomanNumber(S) THEN
                        IF flagLOWERsrc THEN
                            Str.Lows(S);
                        ELSE
                            Str.Caps(S);
                        END;
                        v:=rom2dec(S);
                        padstr(S,maxromanwidth,space);
                        R:="~  ==>  ~";
                        Str.Subst(R,"~",S);
                        Str.Subst(R,"~",bignum2str(v,10,maxdecromanwidth,space));
                        WrStr(R);WrLn;
                    END;
                END;
            END;
        END;
    | showrange:
        S:=parm[minparm];
        IF isInRomanRange(mini)=FALSE THEN abort(errRomanRange,S);END;
        IF isInRomanRange(maxi)=FALSE THEN abort(errRomanRange,S);END;

        (* maxlen:=0; *)
        currval:=mini;
        LOOP
            (*
            Str.Copy(S,dec2rom(currval,1,flagLOWER));
            i:=Str.Length(S);
            IF i > maxlen THEN maxlen:=i;END;
            *)

            (*
            S:="~ ~ ~";
            Str.Subst(S,"~",bignum2str(currval,10,maxdecromanwidth,space));
            Str.Subst(S,"~",dec2rom(currval, maxromanwidth,TRUE));
            Str.Subst(S,"~",dec2rom(currval,-maxromanwidth,FALSE));
            WrStr(S);WrLn;
            *)

            S:="~  ==>  ~";
            Str.Subst(S,"~",bignum2str(currval,10,maxdecromanwidth,space));
            Str.Subst(S,"~",dec2rom(currval, 1,flagLOWERdst));
            WrStr(S);WrLn;

            IF maxi < mini THEN
                DEC(currval);
                IF currval < maxi THEN EXIT; END;
            ELSE
                INC(currval);
                IF currval > maxi THEN EXIT; END;
            END;
        END;
        (* WrStr("maxlen = ");WrStr(num2str(maxlen,10,1,""));WrLn; *)
    END;

    abort(errNone,"");
END RomanTS.