(* ---------------------------------------------------------------
Title         Q&D Modula-2 Cipher Saber
Overview
Usage
Notes         a mere "proof of concept"
              -t hex mode was not implemented : use XD utility instead
Bugs
Wish List     support for automatic filenames (.CS, .CS2)
              alternate header (possibly faking container data)
              wipe source file
              check passphrase for mini and maxi lengths, entropy, chars...
              allow escape chars in passphrase or hex string ($00 ?)
              handle redirection
              armored hex input and output
              select alternate rng and cipher
              gen pw (TEXTGEN-like)
              if passphrase+salt > MAX, silently shorten passphrase ?
              beware of key limits
              hash datetime for init ?

--------------------------------------------------------------- *)

MODULE mSaber;

IMPORT Str;
IMPORT Lib;
IMPORT FIO;
IMPORT IO;

FROM QD_ASCII IMPORT dash, slash, nullchar, tabchar, cr, lf, nl, bs,
space, dot, deg, degree, doublequote, quote, colon, percent, vbar,
blank, equal, dquote, charnull, singlequote, antislash, dollar,
star, backslash, coma, question, underscore, underbar, tabul, hbar,
comma, semicolon, diese, pound, openbracket, closebracket, tilde, exclam,
stardotstar, dotdot, escCh, escSet, letters, digits,
lettersUpp, lettersLow, openbrace, closebrace, ampersand;

FROM QD_Box IMPORT str80, str2, cmdInit, cmdShow, cmdStop, delim,
Work, video, Ltrim, Rtrim, UpperCase, LowerCase, ReplaceChar,
ChkEscape, Waitkey, WaitkeyDelay, Flushkey, IsRedirected, chkJoker,
isOption, GetOptIndex, GetLongCard, GetLongInt, GetString, CharCount,
same, aR, aH, aS, aD, aA, everything, isDirectory, fixDirectory,
str128, str256, Animation, allfiles, Belongs, FixAE, CodePhonetic,
CodeSoundex, CodeSoundexOrg, isReadOnly, LtrimBlanks, RtrimBlanks,
getStrIndex, cmdSHOW,BiosWaitkey,BiosWaitkeyShifted,BiosFlushkey,
str1024, isoleItemS, dmpTTX, str2048, Elapsed, TerminalReadString,
getDosVersion, DosVersion, warning95, runningWindows,
aV, reallyeverything, chkClassicTextMode, setClassicTextMode,
AltAnimation, str16, getCurrentDirectory, setReadWrite, setReadOnly,
getFileSize, verifyString, str4096, unfixDirectory,
animShow, animSHOW, animAdvance, animEnd, animClear,
animInit, animGetSdone, anim, cleantabs, UpperCaseAlt, LowerCaseAlt,
completedInit, completedShow, completedSHOW, completedEnd, completed,
removeDups, isValidHDunit, removePhantoms, removeFloppies,
getCDROMunits, getCDROMletters, removeCDROMs, getAllHDunits,
getAllLegalUnits, metaproc, getCli, argc, argv, qdSplitAllPath,
metaproc, UpperCaseFR, LowerCaseFR, ASCIIonly;

FROM QD_File IMPORT pathtype, w9XnothingRequired,
fileOpenRead, fileOpen, fileExists, fileExistsAlt,
fileIsRO, fileSetRW, fileSetRO,
fileErase, fileCreate, fileAppend,fileRename,
fileGetFileSize, fileGetFileStamp,
fileIsDirectorySpec, fileClose, fileFlush, fileSupportLFN;

FROM QD_rand IMPORT InitRnd, GetRnd,
GetRndCardRange, GetRndLngCardRange,
GetRndLngRealRange, GetRndRealRange,
GetRndIntRange, GetRndLngIntRange;

FROM IO IMPORT WrStr, WrLn;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    quotedpat     = '"*"';
    remark1       = semicolon;
    remark2       = pound;
    addrsep       = colon;
CONST
    zerocard      = MAX(CARDINAL);
    zeroshort     = MAX(SHORTCARD);
CONST
    SALTCS1       = 10; PASSESCS1 = 1;
    SALTCS2       = 10; PASSESCS2 = 20;
CONST
    (* MINPPLEN      = 0; MAXPPLEN   = 256; *)  sMAXPPLEN     = "256";
    MINSALTLEN    = 0; MAXSALTLEN = 128-1; sSALTLENrange = "[0..127]";
    MINPASSES     = 1; MAXPASSES  = 4096;  sPASSESrange  = "[1..4096]";
    DEFSALTLEN    = SALTCS2;
    DEFPASSES     = PASSESCS2;
CONST
    strDEFSALT    = "10";
    strDEFPASSES  = "20";
CONST
    progEXEname   = "MSABER";
    progTitle     = "Modula-2 Cipher Saber";
    progVersion   = "v1.0";
    progCopyright = "by PhG";
    progBanner    = progTitle+" "+progVersion+" "+progCopyright;
CONST
    errNone            = 0;
    errHelp            = 1;
    errExclusive       = 2;
    errUnknownOption   = 3;
    errTooManyParms    = 4;
    errExpecting       = 5;
    errNoMatch         = 6;
    errMissingCommand  = 7;
    errSALT            = 8;
    errPASSES          = 9;
    errJoker           = 10;
    errTargetPath      = 11;
    errTargetHere      = 12;
    errTargetRO        = 13;
    errIsDir           = 14;
    errMissingPPfilespec=15;
    errPPmismatch      = 16;
    errNoPPhere        = 17;
    errIllogicalInput  = 18;
    errEmptyPP         = 19;
    errPPSALTLEN       = 20;
    errBadSALTLEN      = 21;

PROCEDURE abort(e:CARDINAL;einfo:ARRAY OF CHAR);
CONST
(*
 00000000011111111112222222222333333333344444444445555555555666666666677777777778
 1...'....0....'....0....'....0....'....0....'....0....'....0....'....0....'....0
*)

errmsg =
progBanner+nl+
nl+
"Syntax 1 : "+progEXEname+" -e <original> <ciphered> [passphrase] [option]..."+nl+
"Syntax 2 : "+progEXEname+" -d <ciphered> <original> [passphrase] [option]..."+nl+
nl+
"-e[?] encrypt (-ee = -e -@)"+nl+
"-d[?] decrypt (-dd = -d -@)"+nl+
"-cs2  newer CipherSaber-2 method (default, same as -p:20 -i:10)"+nl+
"-cs1  older CipherSaber-1 method (legacy, same as -p:1 -i:10)"+nl+
"-i:#  size of initialization vector ("+sSALTLENrange+", default is "+strDEFSALT+")"+nl+
"-p:#  setup passes ("+sPASSESrange+", default is "+strDEFPASSES+")"+nl+
"-@|-f assume passphrase is plain text in specified passphrase file (see infra)"+nl+
"-s    secure entry of passphrase (do not betray length of passphrase)"+nl+
"-u    unwise (display passphrase)"+nl+
"-o[o] overwrite existing target file (-oo = -o -r)"+nl+
"-r    ignore existing read-only attribute when overwriting or erasing a file"+nl+
"-y    do not copy source date/time stamp to target"+nl+
(* "-t    hexadecimal text mode (default is binary)"+nl+ *)
"-x    disable LFN support even if available"+nl+
nl+
"This program implements Cipher Saber as designed by Arnold G. Reinhold."+nl+
"Note Cipher Saber, though clever in a few ways, is not clever in all ways. ;-)"+nl+
nl+
"a) If omitted from command line, passphrase will be asked from keyboard."+nl+
"b) Should spaces be needed, enclose passphrase with double quotes."+nl+
"c) Passphrase *is* case-sensitive. It should be created with extreme care."+nl+
"   The longer the better. Use at least one character of each group :"+nl+
"   lowercase letters, uppercase letters, digits and symbols."+nl+
"   A DiceWare passphrase may be a good idea. TEXTGEN program can help too."+nl+
"d) Compression prior to encryption is recommended."+nl+
"e) Encrypted and decrypted files are created in current directory,"+nl+
"   whatever their source location (this behavior is by design)."+nl+
"f) Any file decrypted with an invalid passphrase will contain garbage."+nl+
"g) With -@ option, actual passphrase is read from specified passphrase file :"+nl+
"   it will be first non-empty line found after left and right trimming"+nl+
"   (note it cannot be more than 256 bytes long)."+nl+
"   Should passphrase match _"+quotedpat+"_ pattern and thus be double quotes delimited,"+nl+
"   both outer _"+dquote+"_ will be removed."+nl+
(*
'h) -t hexadecimal processing ignores comments preceded by "'+remark1+'" or "'+remark2+'".'+nl+
'   It also ignore anything preceding "'+addrsep+'".'+nl+
*)
"h) Hexadecimal mode was not thought necessary : use XD utility instead."+nl+
"i) Both sender and receiver are required to agree on these parameters :"+nl+
"   -p:#, -i:#, passphrase and original filename."+nl+
nl+
"Examples : "+progEXEname+" -d -cs1 -o diploma1.cs1 diploma1.gif ThomasJefferson"+nl+
"           "+progEXEname+" -d -p:63 -i:19 diploma3.csx diploma3.gif Eris_1963"+nl+
"           "+progEXEname+" -d -cs2 diploma4.csx diploma4.gif -f testpp.asc"+nl;

VAR
    S : str1024; (* oversized for possible LFN *)
BEGIN
    CASE e OF
    | errHelp :            WrStr(errmsg);
    | errExclusive:        S := "-encrypt and -decrypt are mutually exclusive !";
    | errUnknownOption:    S := 'Unknown "~" option !';
    | errTooManyParms:     S := 'Useless "~"  parameter !';
    | errExpecting:        S := 'Missing "~" !';
    | errNoMatch:          S := 'No file matches "~" specification !';
    | errMissingCommand:   S := "Missing -encrypt or -decrypt command !";
    | errSALT:             S := 'Illegal -i:# value or out of "~" range !';
    | errPASSES:           S := 'Illegal -p:# value or out of "~" range !';
    | errJoker:            S := '"~" should not contain any joker !';
    | errTargetPath:       S := '"~" should not contain a path !';
    | errTargetHere:       S := '"~" target already exists !';
    | errTargetRO:         S := '"~" target is read-only !';
    | errIsDir:            S := '"~" cannot be a directory !';
    | errMissingPPfilespec:S :="Missing passphrase file parameter !";
    | errPPmismatch:       S :="First and second entries of passphrase do not match !";
    | errNoPPhere:         S :='No passphrase found in "~" file !';
    | errIllogicalInput:   S :="-s and -u options are mutually exclusive !";
    | errEmptyPP:          S :="No passphrase specified !";
    | errPPSALTLEN:        S :="Passphrase length plus -i:# value is greater than "+sMAXPPLEN+" !";
    | errBadSALTLEN:       S :='No initialization vector header in "~" !';
    ELSE
        S := "This is illogical, Captain !";
    END;
    CASE e OF
    | errNone, errHelp :
        ;
    ELSE
        Str.Subst(S,"~",einfo);
        WrStr(progEXEname+" : "); WrStr(S); WrLn;
    END;
    Lib.SetReturnCode(SHORTCARD(e));
    HALT;
END abort;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE XOR (b1,b2 : CARDINAL) : CARDINAL;
BEGIN
    RETURN CARDINAL ( (BITSET (b1) / BITSET (b2)) );
END XOR;

(* ------------------------------------------------------------ *)

CONST
    ioBufferSize    = (16 * 512) + FIO.BufferOverhead; (* was 8 *)
    firstBufferByte = 1;
    lastBufferByte  = ioBufferSize;
TYPE
    ioBufferType = ARRAY [firstBufferByte..lastBufferByte] OF BYTE;
VAR
    ioBufferIn  : ioBufferType;
    ioBufferOut : ioBufferType;

CONST
    dataBufferSize      = 256;
    firstDataBufferByte = 0;
    maxDataBufferByte   = dataBufferSize-1;
VAR
    databuffer : ARRAY [firstDataBufferByte..maxDataBufferByte] OF BYTE;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

(* rough input : only RETURN, DELETE/LEFT ARROW, ESCAPE *)

PROCEDURE input (VAR R : ARRAY OF CHAR;
                plain,secure:BOOLEAN; prompt:ARRAY OF CHAR);
CONST
    escape    = CHR(27);
    erasechar = bs+space+bs;
VAR
    i     : CARDINAL;
    c1,c2 : CHAR;
    hidechar : CHAR;
BEGIN
    IF secure THEN
        hidechar := space;
    ELSE
        hidechar := star;
    END;
    WrStr(prompt);
    Str.Copy(R,"");
    i:=0;
    LOOP
        c1 := IO.RdKey();
        CASE c1 OF
        | cr:
            EXIT;
        | bs: (* del key *)
            ;
        | escape: (* escape key *)
            WHILE i > 0 DO
                WrStr(erasechar);
                DEC(i);
                R[i]:=CHR(0);
            END;
            EXIT;
        | CHR(0):
            c2 := IO.RdKey();
            CASE c2 OF
            | CHR(75): (* left arrow *)
                c1 := bs;
            END;
        END;
        IF c1 = bs THEN
            IF i > 0 THEN
                WrStr(erasechar);
                DEC(i);
                R[i]:=CHR(0); (* force character delete *)
            END;
        ELSIF ORD(c1) >= ORD(space) THEN
            IF plain THEN
                WrStr(c1);
            ELSE
                IF hidechar # space THEN WrStr(hidechar); END; (* else, cursor would betray ! *)
            END;
            Str.Append(R,c1);
            INC(i);
        END;
    END;
    WrLn;
END input;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    TSrng     = 1;
    QDrng     = 2;

PROCEDURE getRndVal ( whatrng:CARDINAL;lowerlimit,upperlimit:CARDINAL):CARDINAL;
VAR
    v:CARDINAL;
BEGIN
    CASE whatrng OF
    | TSrng  : v:= lowerlimit+Lib.RANDOM(upperlimit-lowerlimit+1);
    | QDrng  : v:= GetRndCardRange(lowerlimit,upperlimit);
    END;
    RETURN v;
END getRndVal;

PROCEDURE initRNG ( whatrng:CARDINAL );
CONST
    warmup = 1024;
VAR
    i,k : CARDINAL;
BEGIN
    CASE whatrng OF
    | TSrng  : Lib.RANDOMIZE();
    | QDrng  : InitRnd;
    END;
    FOR i:=1 TO warmup DO
        k:=getRndVal(whatrng,MIN(CARDINAL),MAX(CARDINAL));
    END;
END initRNG;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

MODULE arcfour;

EXPORT MINPPLEN,MAXPPLEN,pptype,initRC4,getRC4;

CONST
    MINPPLEN      = 0; MAXPPLEN   = 256;
TYPE
    pptype = RECORD
        len : CARDINAL;
        dat : ARRAY[MINPPLEN..MAXPPLEN-1] OF CARDINAL; (* 0-based *)
    END;
CONST
    MINSTATE = MIN(SHORTCARD);
    MAXSTATE = MAX(SHORTCARD);
    MODSTATE = MAXSTATE+1;
VAR
    state    : ARRAY[MINSTATE..MAXSTATE] OF CARDINAL; (* bytes *)
    i,j      : CARDINAL; (* global for inner module *)

PROCEDURE swap (VAR a,b:CARDINAL);
VAR
    tmp:CARDINAL;
BEGIN
    tmp:=a;
    a:=b;
    b:=tmp;
END swap;

PROCEDURE initRC4 (warmloop:CARDINAL;userkey:pptype);
BEGIN
    FOR i := MINSTATE TO MAXSTATE DO state[i] := i; END;
    j := 0;
    LOOP
        IF warmloop = 0 THEN EXIT;END;
        FOR i:= MINSTATE TO MAXSTATE DO
            j := (j + state[i] + userkey.dat[i MOD userkey.len]) MOD MODSTATE;
            swap (state[i],state[j]);
        END;
        DEC(warmloop);
    END;
    i := 0;
    j := 0;
END initRC4;

PROCEDURE getRC4 (  ):CARDINAL;
BEGIN
    i := (i + 1) MOD MODSTATE;
    j := (j + state[i]) MOD MODSTATE;
    swap (state[i],state[j]);
    RETURN state[ (state[i] + state[j]) MOD MODSTATE ];
END getRC4;

END arcfour;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE preprocessLine (VAR S:ARRAY OF CHAR);
VAR
    len:CARDINAL;
BEGIN
    LtrimBlanks(S);
    RtrimBlanks(S);
    IF Str.Match(S,quotedpat) THEN
        Str.Delete(S,0,1);
        len:=Str.Length(S);
        S[len-1]:=charnull;
    END;
END preprocessLine;

(* passphrase enters as filename and exits as... passphrase *)

PROCEDURE grabPP (VAR pp:pptype;useLFN:BOOLEAN;F:pathtype):BOOLEAN;
CONST
    maxtries = 100;
    msg = "Working, please wait... ";
VAR
    S:str1024; (* almost pathtype *)
    hin:FIO.File;
    ok:BOOLEAN;
    i,tries:CARDINAL;
BEGIN
    video(msg,TRUE);

    tries:=1-1;
    ok:=FALSE;
    FIO.EOF:=FALSE;
    hin:=fileOpenRead(useLFN,F);
    FIO.AssignBuffer(hin,ioBufferIn);
    LOOP
        IF FIO.EOF THEN EXIT; END;
        FIO.RdStr(hin,S);
        IF FIO.EOF THEN EXIT; END;
        preprocessLine(S);
        IF same(S,"") = FALSE THEN
            pp.len:=Str.Length(S);
            FOR i:=1 TO pp.len DO pp.dat[i-1]:=ORD(S[i-1]); END;
            ok:=TRUE;
            EXIT;
        END;
        INC(tries);
        IF tries >= maxtries THEN ok:=FALSE; EXIT; END;
    END;
    fileClose(useLFN,hin);
    video(msg,FALSE);
    RETURN ok;
END grabPP;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    methodrc4 = 0;

PROCEDURE initCipher (whatcipher,warmloop:CARDINAL;userkey:pptype);
BEGIN
    CASE whatcipher OF
    | methodrc4: initRC4(warmloop,userkey);
    END;
END initCipher;

(* uses global databuffer *)

PROCEDURE doCipher (whatcipher,got:CARDINAL);
VAR
    i,b,v:CARDINAL;
BEGIN
    FOR i:= firstDataBufferByte TO (got-1) DO
        b:=CARDINAL(databuffer[i]);
        CASE whatcipher OF
        | methodrc4: v:=getRC4();
        END;
        databuffer[i] := BYTE ( XOR(b,v) );
    END;
END doCipher;

VAR
    salt     : ARRAY[MINSALTLEN..MAXSALTLEN-1] OF SHORTCARD; (* 0-based, bytes *)
    pp       : pptype;

(*
   check :
   - pplength + ivlength <= maxpplen
   - filelength >= SALT
*)

PROCEDURE procFile (doEncrypt,useLFN:BOOLEAN;
                   whatrng,whatcipher,SALTLEN,PASSES:CARDINAL;
                   pp:pptype;source,target:pathtype ):CARDINAL;
VAR
    hin,hout:FIO.File;
    got,i,p:CARDINAL;
    userkey:pptype;
BEGIN
    (* check passphrase + salt <= 255 *)
    IF (pp.len+SALTLEN) > MAXPPLEN THEN RETURN errPPSALTLEN;END;
    IF NOT(doEncrypt) THEN
        (* check salt exists *)
        IF fileGetFileSize(useLFN,source) < LONGCARD(SALTLEN) THEN RETURN errBadSALTLEN;END;
    END;

    hin :=fileOpenRead(useLFN,source);
    FIO.AssignBuffer(hin,ioBufferIn);
    hout :=fileCreate(useLFN,target);
    FIO.AssignBuffer(hout,ioBufferOut);

    FIO.EOF:=FALSE; (* useless safety *)

    IF doEncrypt THEN (* build salt *)
        i := MINSALTLEN;
        LOOP
            IF i >= SALTLEN THEN EXIT;END;
            salt[i]:=SHORTCARD( getRndVal(whatrng,MIN(SHORTCARD),MAX(SHORTCARD)) );
            INC(i);
        END;
        IF SALTLEN # 0 THEN FIO.WrBin(hout,salt,SALTLEN); END;
    ELSE              (* retrieve salt *)
        IF SALTLEN # 0 THEN got:=FIO.RdBin(hin,salt,SALTLEN); END;
    END;

    (* key = passphrase+salt *)
    p:=MINPPLEN;
    FOR i:=1 TO pp.len DO
        userkey.dat[p]:=pp.dat[i-1];
        INC(p);
        pp.dat[i-1]:=zerocard;
    END;
    i:=MINSALTLEN;
    LOOP
        IF i >= SALTLEN THEN EXIT;END;
        userkey.dat[p]:=CARDINAL( salt[i] );
        INC(p);
        salt[i]:=zeroshort;
        INC(i);
    END;
    userkey.len:=p;
    SALTLEN:=zerocard;

    initCipher(methodrc4,PASSES,userkey);

    LOOP
        got:=FIO.RdBin(hin,databuffer,dataBufferSize);
        IF got = 0 THEN EXIT; END;
        doCipher(methodrc4,got);
        FIO.WrBin(hout,databuffer,got);
        IF got # dataBufferSize THEN EXIT; END;
    END;

    FIO.Flush(hout);
    fileClose(useLFN,hout);
    fileClose(useLFN,hin);

    FOR i:= MINPPLEN TO MAXPPLEN-1 DO userkey.dat[i]:=zerocard;END;

    RETURN errNone;
END procFile;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

PROCEDURE grabvalue (VAR v:CARDINAL;
                    mini,maxi:CARDINAL;S:ARRAY OF CHAR):BOOLEAN;
VAR
    lc:LONGCARD;
    ok:BOOLEAN;
BEGIN
    ok:=GetLongCard(S,lc);
    IF ok THEN
        ok:=NOT ( lc > LONGCARD(maxi) );
        IF ok THEN
            ok:= ( lc >= LONGCARD(mini) );
            IF ok THEN v:=CARDINAL(lc);END;
        END;
    END;
    RETURN ok;
END grabvalue;

PROCEDURE chkNoPath (S:pathtype):BOOLEAN;
VAR
    u,d,n,e,R:pathtype;
BEGIN
    qdSplitAllPath(S,u,d,n,e);
    Lib.MakeAllPath(R,"","",n,e);
    RETURN same(S,R);
END chkNoPath;

PROCEDURE num2str (v,base:CARDINAL;wipad:INTEGER;padchar:CHAR):str16;
VAR
    R:str16;
    ok:BOOLEAN;
    i:CARDINAL;
BEGIN
    Str.CardToStr(LONGCARD(v),R,base,ok);
    IF ok THEN
        FOR i:=Str.Length(R)+1 TO ABS(wipad) DO
            IF wipad < 0 THEN
                Str.Append(R,padchar);
            ELSE
                Str.Prepend(R,padchar);
            END;
        END;
    ELSE
        R:="*** ERROR ***";
    END;
    RETURN R;
END num2str;

PROCEDURE msg2 (quoted:BOOLEAN;S1,S2:ARRAY OF CHAR );
CONST
    wi = 10;
VAR
    i:CARDINAL;
BEGIN
    WrStr(S1);
    FOR i:=Str.Length(S1)+1 TO wi DO WrStr(" ");END;WrStr(" : ");
    IF quoted THEN WrStr(dquote);END;
    WrStr(S2);
    IF quoted THEN WrStr(dquote);END;
    WrLn;
END msg2;

TYPE
    optype = (opNone,opEncrypt,opDecrypt);

PROCEDURE newop (VAR op:optype;wanted:optype ):BOOLEAN;
VAR
    ok:BOOLEAN;
BEGIN
    IF ( (op = opNone) OR (op = wanted) ) THEN
        op:=wanted;
        ok:=TRUE;
    ELSE
        ok:=FALSE;
    END;
    RETURN ok;
END newop;

PROCEDURE strToPP (VAR pp:pptype;S:ARRAY OF CHAR);
VAR
    i:CARDINAL;
BEGIN
    pp.len := Str.Length(S);
    FOR i:=1 TO pp.len DO pp.dat[i-1]:=ORD(S[i-1]);END;
END strToPP;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

TYPE
    dttype = RECORD
        CASE : BOOLEAN OF
        | TRUE  :
            t : CARDINAL; (* hms is low  *)
            d : CARDINAL; (* ymd is high *)
        | FALSE :
            dt : LONGCARD;
        END;
    END;

PROCEDURE getfilestamp (VAR h:dttype  ; useLFN:BOOLEAN;S:pathtype):BOOLEAN;
VAR
    hnd:FIO.File;
    rc:BOOLEAN;
BEGIN
    hnd:=fileOpenRead(useLFN,S);
    IF hnd = MAX(CARDINAL) THEN (* not trapped by fileGetFileStamp() *)
        rc:=FALSE;
    ELSE
        h.dt := FIO.GetFileDate(hnd);
        rc:=TRUE;
    END;
    fileClose(useLFN,hnd);
    RETURN rc;
END getfilestamp;

PROCEDURE setFileStamp (useLFN:BOOLEAN;h:dttype;S:pathtype):BOOLEAN;
VAR
    hnd:FIO.File;
    rc:BOOLEAN;
BEGIN
    hnd:=fileOpen(useLFN,S);
    IF hnd=MAX(CARDINAL) THEN
        rc:=FALSE;
    ELSE
        FIO.SetFileDate(hnd,h.dt);
        rc:=TRUE;
    END;
    fileClose(useLFN,hnd);
    RETURN rc;
END setFileStamp;

(* ------------------------------------------------------------ *)
(* ------------------------------------------------------------ *)

CONST
    msgPrompt1  = "Enter passphrase       : ";
    msgPrompt2  = "Enter passphrase again : ";
    strSource   = "::: Source";
    strTarget   = "+++ Target";
CONST
    msgWorking  = "Working, please wait...";
CONST
    firstparm = 1;
    maxparm   = 3;
VAR
    whatcipher,whatrng : CARDINAL;
    SALTLEN,PASSES : CARDINAL;
    op : optype;
    flagPPFROMFILE:BOOLEAN; (* if true, passphrase parm is a file *)
    flagHEX,flagREPLACEEXT,flagOVERWRITE,flagOVERWRITERO:BOOLEAN;
    flagCOPYSTAMP,flagSECUREINPUT,flagUNWISE:BOOLEAN;
VAR
    parm : ARRAY[firstparm..maxparm] OF pathtype;
    F,source,target,S,R: pathtype;
    lastparm,parmcount,i,got,opt : CARDINAL;
    ok,useLFN,DEBUG:BOOLEAN;
    ppsecurity:pptype;
    orgstamp:dttype;
BEGIN
    Lib.DisableBreakCheck();
    FIO.IOcheck := FALSE;
    FIO.ShareMode := FIO.ShareDenyNone; (* very, very important ! *)
    WrLn;

    parmcount := Lib.ParamCount();
    IF parmcount = 0 THEN abort(errHelp,"");END;

    DEBUG    := FALSE;
    useLFN   := TRUE;

    flagPPFROMFILE  := FALSE;
    flagOVERWRITE   := FALSE;
    flagOVERWRITERO := FALSE;
    flagHEX         := FALSE;
    flagREPLACEEXT  := FALSE;
    flagSECUREINPUT := FALSE;
    flagUNWISE      := FALSE;
    flagCOPYSTAMP   := TRUE;
    SALTLEN         := DEFSALTLEN;
    PASSES          := DEFPASSES;
    op              := opNone;
    whatrng         := QDrng;
    whatcipher      := methodrc4;

    lastparm := firstparm-1;

    FOR i := 1 TO parmcount DO
        Lib.ParamStr(S,i); cleantabs(S);
        Str.Copy(R,S);
        UpperCase(R);
        IF isOption(R) THEN
            opt := GetOptIndex(R, "?"+delim+"H"+delim+"HELP"+delim+
                                  "E"+delim+"ENCRYPT"+delim+"CIPHER"+delim+
                                  "D"+delim+"DECRYPT"+delim+"DECIPHER"+delim+
                                  "@"+delim+
                                  "X"+delim+"LFN"+delim+
                                  "DEBUG"+delim+

                                  "I:"+delim+"IV:"+delim+"SALT:"+delim+"VECTORSIZE:"+delim+
                                  "P:"+delim+"PASSES:"+delim+"LOOPS:"+delim+"N:"+delim+
                                  "CS1"+delim+"OLD"+delim+
                                  "CS2"+delim+"NEW"+delim+
                                  "EE"+delim+
                                  "DD"+delim+
                                  "O"+delim+"OVERWRITE"+delim+
                                  "R"+delim+"READONLY"+delim+
                                  "OO"+delim+

                                  "t"+delim+"hex"+delim+
                                  "et"+delim+
                                  "dt"+delim+
                                  "S"+delim+"SECURE"+delim+
                                  "U"+delim+"UNWISE"+delim+
                                  "F"+delim+"FILE"+delim+
                                  "Y"+delim+"STAMP"
                              );
            CASE opt OF
            | 1,2,3 : abort(errHelp,"");
            | 4,5,6 : IF newop(op,opEncrypt)=FALSE THEN abort(errExclusive,"");END;
            | 7,8,9 : IF newop(op,opDecrypt)=FALSE THEN abort(errExclusive,"");END;
            | 10 :    flagPPFROMFILE  := TRUE; (* passphrase is file *)
            | 11,12 : useLFN          := FALSE;
            | 13:     DEBUG           := TRUE;

            | 14,15,16,17:IF grabvalue(SALTLEN,MINSALTLEN,MAXSALTLEN,R)=FALSE THEN abort(errSALT,sSALTLENrange);END;
            | 18,19,20,21:IF grabvalue(PASSES,MINPASSES,MAXPASSES,R)=FALSE THEN abort(errPASSES,sPASSESrange);END;
            | 22,23:     SALTLEN := SALTCS1; PASSES:=PASSESCS1;
            | 24,25:     SALTLEN := SALTCS2; PASSES:=PASSESCS2;
            | 26:        IF newop(op,opEncrypt)=FALSE THEN abort(errExclusive,"");END;
                         flagPPFROMFILE  := TRUE;
            | 27:        IF newop(op,opDecrypt)=FALSE THEN abort(errExclusive,"");END;
                         flagPPFROMFILE  := TRUE;
            | 28,29:     flagOVERWRITE   := TRUE;
            | 30,31:     flagOVERWRITERO := TRUE;
            | 32:        flagOVERWRITE   := TRUE;
                         flagOVERWRITERO := TRUE;

            | 33,34:     flagHEX         := TRUE;
            | 35:        IF newop(op,opEncrypt)=FALSE THEN abort(errExclusive,"");END;
                         flagHEX         := TRUE;
            | 36:        IF newop(op,opDecrypt)=FALSE THEN abort(errExclusive,"");END;
                         flagHEX         := TRUE;
            | 37,38:     flagSECUREINPUT := TRUE;
            | 39,40:     flagUNWISE      := TRUE;
            | 41,42 :    flagPPFROMFILE  := TRUE; (* passphrase is file *) (* same as 10 *)
            | 43,44 :    flagCOPYSTAMP   := FALSE;
            ELSE
                abort(errUnknownOption,S);
            END;
        ELSE
            INC(lastparm);
            IF lastparm > maxparm THEN abort(errTooManyParms,S);END;
            parm[lastparm]:=S;
        END;
    END;
    IF op = opNone THEN abort(errMissingCommand,"");END;

    IF (flagSECUREINPUT AND flagUNWISE) THEN abort(errIllogicalInput,"");END;

    initRNG(whatrng);

    CASE lastparm OF
    | firstparm-1:
        abort(errHelp,"");
    | firstparm:
        CASE op OF
        | opEncrypt: S:="<ciphered> [passphrase]";
        | opDecrypt: S:="<original> [passphrase]";
        | opNone:    S:="<-e|-d>";
        END;
        abort(errExpecting,S);
    | firstparm+1:
        Str.Copy(source,parm[firstparm]);
        Str.Copy(target,parm[firstparm+1]);
        pp.len:=MINPPLEN;
    | maxparm:
        Str.Copy(source,parm[firstparm]);
        Str.Copy(target,parm[firstparm+1]);
        Str.Copy(F,parm[maxparm]);
        pp.len:=Str.Length(F);
        FOR i:= 1 TO pp.len DO pp.dat[i-1]:=ORD(F[i-1]); END;
    END;

    useLFN:=( useLFN AND fileSupportLFN() );

    IF same(source,dot) THEN source:=stardotstar;END;
    IF same(target,dot) THEN target:=stardotstar;END;

    IF chkJoker(source) THEN abort(errJoker,source);END;
    IF chkJoker(target) THEN abort(errJoker,target);END;
    IF chkNoPath(target)=FALSE THEN abort(errTargetPath,target);END;

    IF fileIsDirectorySpec(useLFN,source) THEN abort(errIsDir,source);END;
    IF fileIsDirectorySpec(useLFN,target) THEN abort(errIsDir,target);END;

    IF fileExists(useLFN,source)=FALSE THEN abort(errNoMatch,source);END;

    IF fileExists(useLFN,target) THEN
        IF flagOVERWRITE THEN
            IF fileIsRO(useLFN,target) THEN
                IF flagOVERWRITERO THEN
                    fileSetRW(useLFN,target);
                    fileErase(useLFN,target);
                ELSE
                    abort(errTargetRO,target);
                END;
            ELSE
                fileErase(useLFN,target);
            END;
        ELSE
            abort(errTargetHere,target);
        END;
    END;

    IF pp.len = MINPPLEN THEN (* no passphrase nor filevault specified *)
        IF flagPPFROMFILE THEN (* yes, we could ask it from kbd *)
            abort(errMissingPPfilespec,"");
        ELSE
            input (S, flagUNWISE,flagSECUREINPUT,msgPrompt1);
            preprocessLine(S);
            strToPP(pp,S);
            IF pp.len = MINPPLEN THEN abort(errEmptyPP,"");END;

            input (S, flagUNWISE,flagSECUREINPUT,msgPrompt2);
            preprocessLine(S);
            strToPP(ppsecurity,S);

            WrLn;
            IF pp.len # ppsecurity.len THEN abort(errPPmismatch,"");END;
            opt:=0;
            FOR i:=1 TO pp.len DO
                IF pp.dat[i-1] # ppsecurity.dat[i-1] THEN INC(opt); END;
            END;
            IF opt # 0 THEN abort(errPPmismatch,"");END;
        END;
    ELSE (* passphrase or filevault *)
        IF flagPPFROMFILE THEN
            IF same(F,dot) THEN F:=stardotstar;END;
            IF chkJoker(F) THEN abort(errJoker,F);END;
            IF fileIsDirectorySpec(useLFN,F) THEN abort(errIsDir,F);END;
            IF fileExists(useLFN,F)=FALSE THEN abort(errNoMatch,F);END;

            IF grabPP (pp,useLFN,F)= FALSE THEN abort(errNoPPhere,F);END;
        END;
    END;

    video(msgWorking,TRUE);
    IF flagCOPYSTAMP THEN ok:=getfilestamp(orgstamp, useLFN,source); END;
    i:=procFile( (op=opEncrypt),useLFN,
                 whatrng,whatcipher,SALTLEN,PASSES,pp,source,target);
    IF flagCOPYSTAMP THEN ok:=setFileStamp(useLFN,orgstamp,target); END;
    video(msgWorking,FALSE);

    IF i # errNone THEN abort(i,source);END;

    msg2(TRUE,strSource,source);
    msg2(TRUE,strTarget,target);

    IF DEBUG THEN
        msg2(TRUE,  "Source" ,source);
        msg2(TRUE,  "Target" ,target);
        IF flagCOPYSTAMP THEN
            S:="source date/time";
        ELSE
            S:="current date/time";
        END;
        msg2(FALSE, "Stamp"  ,S);
        msg2(FALSE, "IV"     ,num2str(SALTLEN,10,1," "));
        msg2(FALSE, "PASSES" ,num2str(PASSES,10,1," "));
(*
        IF flagHEX THEN
            S:="hex text";
        ELSE
            S:="binary";
        END;
        msg2(FALSE, "Mode"    ,S);
*)
        S:="";
        FOR i:= 1 TO pp.len DO Str.Append(S,CHR(pp.dat[i-1]));END; (* //FIXME later *)
        msg2(TRUE,  "Passphrase",S);
    END;

    SALTLEN        := zerocard;
    PASSES         := zerocard;
    pp.len         := zerocard;
    ppsecurity.len := zerocard;
    FOR i:=MINPPLEN TO MAXPPLEN-1 DO
        pp.dat[i]        :=zerocard;
        ppsecurity.dat[i]:=zerocard;
    END;

    abort(errNone,"");
END mSaber.





